/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.audit.destination;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.auth.AuthSchemeProvider;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.KerberosCredentials;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.config.AuthSchemes;
import org.apache.http.client.config.CookieSpecs;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.config.Lookup;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.auth.SPNegoSchemeFactory;
import org.apache.http.impl.client.*;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.protocol.HttpContext;
import org.apache.ranger.audit.model.AuditEventBase;
import org.apache.ranger.audit.model.AuthzAuditEvent;
import org.apache.ranger.audit.provider.MiscUtil;;

import java.io.*;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.Charset;
import java.security.*;
import java.security.cert.CertificateException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;

public class HTTPAuditDestination extends AuditDestination {
	private static final Log LOG = LogFactory.getLog(HTTPAuditDestination.class);

	public static final String PROP_HTTP_SERVER_URL                = "xasecure.audit.destination.http.url";
	public static final String PROP_HTTP_USER_NAME                 = "xasecure.audit.destination.http.username";
	public static final String PROP_HTTP_USER_PASSWORD 			   = "xasecure.audit.destination.http.password";
	public static final String PROP_HTTP_CLIENT_CONN_TIMEOUT_MS    = "xasecure.audit.destination.http.connection.timeout.ms";
	public static final String PROP_HTTP_CLIENT_READ_TIMEOUT_MS    = "xasecure.audit.destination.http.read.timeout.ms";
	public static final String PROP_HTTP_MAX_CONNECTION            = "xasecure.audit.destination.http.max.connections";
	public static final String PROP_HTTP_MAX_CONNECTION_PER_HOST   = "xasecure.audit.destination.http.max.connections.per.host";
	public static final String PROP_HTTP_VALIDATE_INACTIVE_MS      = "xasecure.audit.destination.http.validate.inactivity.ms";
	public static final String PROP_HTTP_POOL_RETRY_COUNT		   = "xasecure.audit.destination.http.pool.retry.count";
	public static final String GSON_DATE_FORMAT                    = "yyyy-MM-dd'T'HH:mm:ss.SSSZ";
	public static final String REST_ACCEPTED_MIME_TYPE_JSON        = "application/json";
	public static final String REST_CONTENT_TYPE_MIME_TYPE_JSON    = "application/json";
	public static final String REST_HEADER_ACCEPT                  = "Accept";
	public static final String REST_HEADER_CONTENT_TYPE            = "Content-type";
	public static final String REST_RELATIVE_PATH_POST			   = "/api/audit/post";

	private static final String PROTOCOL_HTTPS 		 = "https";
	private volatile CloseableHttpClient httpClient  = null;
	private volatile Gson				 gsonBuilder = null;
	private String  httpURL                          = null;

	@Override
	public void init(Properties props, String propPrefix) {
		LOG.info("==> HTTPAuditDestination:init()");
		super.init(props, propPrefix);
		this.httpClient  = buildHTTPClient();
		this.gsonBuilder = new GsonBuilder().setDateFormat(GSON_DATE_FORMAT).create();
		LOG.info("<== HTTPAuditDestination:init()");
	}

	@Override
	public void stop() {
		LOG.info("==> HTTPAuditDestination.stop() called..");
		logStatus();
		if (httpClient != null) {
			try {
				httpClient.close();
			} catch (IOException ioe) {
				LOG.error("Error while closing httpclient in HTTPAuditDestination!", ioe);
			} finally {
				httpClient = null;
			}
		}
	}

	@Override
	public boolean log(Collection<AuditEventBase> events) {
		boolean ret = false;
		try {
			logStatusIfRequired();
			addTotalCount(events.size());

			if (httpClient == null) {
				httpClient = buildHTTPClient();
				if (httpClient == null) {
					// HTTP Server is still not initialized. So need return error
					addDeferredCount(events.size());
					return ret;
				}
			}

			Map<String, String>  queryParams = new HashMap<>();
			UserGroupInformation ugi = UserGroupInformation.getCurrentUser();
			if (LOG.isDebugEnabled()) {
				LOG.debug("UserGroupInformation: " + ugi);
			}
			for (AuditEventBase event : events) {
				AuthzAuditEvent authzEvent = (AuthzAuditEvent) event;
				PrivilegedExceptionAction<String> action = getPrivilegedAction(REST_RELATIVE_PATH_POST, queryParams, authzEvent, String.class);
				executeAction(action, ugi);
				ret = true;
			}
		} catch (Throwable t) {
			addDeferredCount(events.size());
			logError("Error sending audit to HTTP Server", t);
		}
		return ret;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see org.apache.ranger.audit.provider.AuditProvider#flush()
	 */
	@Override
	public void flush() {

	}

	public boolean isAsync() {
		return true;
	}

	synchronized CloseableHttpClient buildHTTPClient() {
		Lookup<AuthSchemeProvider> authRegistry  = RegistryBuilder.<AuthSchemeProvider>create().register(AuthSchemes.SPNEGO, new SPNegoSchemeFactory(true, true)).build();
		HttpClientBuilder clientBuilder  = HttpClients.custom().setDefaultAuthSchemeRegistry(authRegistry);
		String userName                  = MiscUtil.getStringProperty(props, PROP_HTTP_USER_NAME);
		String passWord                  = MiscUtil.getStringProperty(props, PROP_HTTP_USER_PASSWORD);
		int    clientConnTimeOutMs       = MiscUtil.getIntProperty(props, PROP_HTTP_CLIENT_CONN_TIMEOUT_MS,1000);
		int    clientReadTimeOutMs       = MiscUtil.getIntProperty(props, PROP_HTTP_CLIENT_READ_TIMEOUT_MS,1000);
		int    maxConnections            = MiscUtil.getIntProperty(props, PROP_HTTP_MAX_CONNECTION,10);
		int    maxConnectionsPerHost     = MiscUtil.getIntProperty(props, PROP_HTTP_MAX_CONNECTION_PER_HOST,10);
		int    validateAfterInactivityMs = MiscUtil.getIntProperty(props, PROP_HTTP_VALIDATE_INACTIVE_MS,1000);
	    int    poolRetryCount			 = MiscUtil.getIntProperty(props, PROP_HTTP_POOL_RETRY_COUNT,5);
		httpURL               		     = MiscUtil.getStringProperty(props, PROP_HTTP_SERVER_URL);

		if (LOG.isDebugEnabled()) {
				LOG.debug("===>> Building apache HTTP client with configs={}");
			}

			try {
				BasicCredentialsProvider credentialsProvider = new BasicCredentialsProvider();

				if (isKerberosAuthenticated()) {
					credentialsProvider.setCredentials(AuthScope.ANY, new KerberosCredentials(null));
				} else if (StringUtils.isNotEmpty(userName) && StringUtils.isNotEmpty(passWord)) {
					credentialsProvider.setCredentials(AuthScope.ANY, new UsernamePasswordCredentials(userName, passWord));
				}

				clientBuilder.setDefaultCredentialsProvider(credentialsProvider);
			} catch (Exception excp) {
				LOG.error("Exception while adding credentials, skipping setting credentials for client.", excp); //log and proceed
			}

			PoolingHttpClientConnectionManager connectionManager;

			KeyManager[]   kmList     = getKeyManagers();
			TrustManager[] tmList     = getTrustManagers();
			SSLContext     sslContext = getSSLContext(kmList, tmList);

			if(sslContext != null) {
				SSLContext.setDefault(sslContext);
			}

			boolean isSSL = (httpURL != null && httpURL.contains("https://"));
			if (isSSL) {
				SSLConnectionSocketFactory sslsf = new SSLConnectionSocketFactory(sslContext, NoopHostnameVerifier.INSTANCE);
				Registry<ConnectionSocketFactory> socketFactoryRegistry = RegistryBuilder.<ConnectionSocketFactory>create().register(PROTOCOL_HTTPS, sslsf).build();

				connectionManager = new PoolingHttpClientConnectionManager(socketFactoryRegistry);

				clientBuilder.setSSLSocketFactory(sslsf);
			} else {
				connectionManager = new PoolingHttpClientConnectionManager();
			}

			connectionManager.setMaxTotal(maxConnections);
			connectionManager.setDefaultMaxPerRoute(maxConnectionsPerHost);
			connectionManager.setValidateAfterInactivity(validateAfterInactivityMs);

			RequestConfig customizedRequestConfig = RequestConfig.custom().setCookieSpec(CookieSpecs.DEFAULT)
					.setConnectTimeout(clientConnTimeOutMs)
					.setSocketTimeout(clientReadTimeOutMs).build();

			CloseableHttpClient httpClient = clientBuilder.setConnectionManager(connectionManager)
					.setRetryHandler(new AuditHTTPRetryHandler(poolRetryCount, true))
					.setDefaultRequestConfig(customizedRequestConfig).build();

			return httpClient;
	}

	private void close(InputStream str, String filename) {
		if (str != null) {
			try {
				str.close();
			} catch (IOException excp) {
				LOG.error("Error while closing file: [" + filename + "]", excp);
			}
		}
	}

	private <T> PrivilegedExceptionAction<T> getPrivilegedAction(String relativeUrl, Map<String, String> queryParams, Object postParam, Class<T> clazz) {
		return () -> executeHttpRequestPOST(relativeUrl, queryParams, postParam, clazz);
	}

	public <T> T executeHttpRequestPOST(String relativeUrl, Map<String, String> params, Object obj, Class<T> clazz) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==>  HTTPAuditDestination().executeHttpRequestPOST()");
		}

		T finalResponse = postAndParse(httpURL + relativeUrl, params, obj, clazz);

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== HTTPAuditDestination().executeHttpRequestPOST()");
		}
		return finalResponse;
	}

	public <T> T postAndParse(String url, Map<String, String> queryParams, Object obj, Class<T> clazz) throws Exception {
		HttpPost     httpPost = new HttpPost(buildURI(url, queryParams));
		StringEntity entity   = new StringEntity(gsonBuilder.toJson(obj));
		httpPost.setEntity(entity);
		return executeAndParseResponse(httpPost, clazz);
	}

	private <T extends HttpRequestBase, R> R executeAndParseResponse(T request, Class<R> responseClazz) throws Exception {
		R                   ret    = null;
		CloseableHttpClient client = getCloseableHttpClient();
		if (client != null) {
			addCommonHeaders(request);

			try (CloseableHttpResponse response = client.execute(request)) {
				ret = parseResponse(response, responseClazz);
			}
		} else {
			LOG.error("Cannot process request as Audit HTTPClient is null...");
		}
		return ret;
	}

	private <T extends HttpRequestBase> T addCommonHeaders(T t) {
		t.addHeader(REST_HEADER_ACCEPT, REST_ACCEPTED_MIME_TYPE_JSON);
		t.setHeader(REST_HEADER_CONTENT_TYPE, REST_CONTENT_TYPE_MIME_TYPE_JSON);
		return t;
	}

	@SuppressWarnings("unchecked")
	private <R> R parseResponse(HttpResponse response, Class<R> clazz) throws Exception {
		R type = null;
		if (response != null) {
			int    httpStatus  = response.getStatusLine().getStatusCode();
			if (httpStatus == HttpStatus.SC_OK) {
				InputStream responseInputStream = response.getEntity().getContent();
				if (clazz.equals(String.class)) {
					type = (R) IOUtils.toString(responseInputStream, Charset.defaultCharset());
				} else {
					type = gsonBuilder.fromJson(new InputStreamReader(responseInputStream), clazz);
				}
				responseInputStream.close();
			} else {
				String error = "Request failed : response= [+" + response + "], response.status= " + httpStatus;
				LOG.error(error);
				throw new Exception(error);
			}
		} else {
			String reponseError = "Received NULL response from server..";
			LOG.error(reponseError);
			throw new Exception(reponseError);
		}
		return type;
	}

	private <T> T executeAction(PrivilegedExceptionAction<T> action, UserGroupInformation owner) throws Exception {
		T ret = null;
		if (owner != null) {
			ret = owner.doAs(action);
		} else {
			ret = action.run();
		}
		return ret;
	}

	private URI buildURI(String url, Map<String, String> queryParams) throws URISyntaxException {
		URIBuilder builder = new URIBuilder(url);
		if (queryParams != null) {
			for (Map.Entry<String, String> param : queryParams.entrySet()) {
				builder.addParameter(param.getKey(), param.getValue());
			}
		}
		return builder.build();
	}

	boolean isKerberosAuthenticated() throws Exception {
		boolean status = false;
		try {
			UserGroupInformation loggedInUser           = UserGroupInformation.getLoginUser();
			boolean              isSecurityEnabled      = UserGroupInformation.isSecurityEnabled();
			boolean              hasKerberosCredentials = loggedInUser.hasKerberosCredentials();
			UserGroupInformation.AuthenticationMethod loggedInUserAuthMethod = loggedInUser.getAuthenticationMethod();

			status = isSecurityEnabled && hasKerberosCredentials && loggedInUserAuthMethod.equals(UserGroupInformation.AuthenticationMethod.KERBEROS);
		} catch (IOException e) {
			throw new Exception("Failed to get authentication details.", e);
		}
		return status;
	}

	private CloseableHttpClient getCloseableHttpClient() {
		CloseableHttpClient client = httpClient;
		if (client == null) {
			synchronized (this) {
				client = httpClient;

				if (client == null) {
					httpClient = client = buildHTTPClient();
				}
			}
		}
		return client;
	}

	private KeyManager[] getKeyManagers() {
		KeyManager[] kmList = null;
		String credentialProviderPath = MiscUtil.getStringProperty(props, RANGER_POLICYMGR_CLIENT_KEY_FILE_CREDENTIAL);
		String keyStoreAlias = RANGER_POLICYMGR_CLIENT_KEY_FILE_CREDENTIAL_ALIAS;
		String keyStoreFile = MiscUtil.getStringProperty(props, RANGER_POLICYMGR_CLIENT_KEY_FILE);
		String keyStoreFilepwd = MiscUtil.getCredentialString(credentialProviderPath, keyStoreAlias);
		if (StringUtils.isNotEmpty(keyStoreFile) && StringUtils.isNotEmpty(keyStoreFilepwd)) {
			InputStream in = null;
			try {
				in = getFileInputStream(keyStoreFile);

				if (in != null) {
					String keyStoreType = MiscUtil.getStringProperty(props, RANGER_POLICYMGR_CLIENT_KEY_FILE_TYPE);
					keyStoreType = StringUtils.isNotEmpty(keyStoreType) ? keyStoreType : RANGER_POLICYMGR_CLIENT_KEY_FILE_TYPE_DEFAULT;
					KeyStore keyStore = KeyStore.getInstance(keyStoreType);

					keyStore.load(in, keyStoreFilepwd.toCharArray());

					KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(RANGER_SSL_KEYMANAGER_ALGO_TYPE);

					keyManagerFactory.init(keyStore, keyStoreFilepwd.toCharArray());

					kmList = keyManagerFactory.getKeyManagers();
				} else {
					LOG.error("Unable to obtain keystore from file [" + keyStoreFile + "]");
				}
			} catch (KeyStoreException e) {
				LOG.error("Unable to obtain from KeyStore :" + e.getMessage(), e);
			} catch (NoSuchAlgorithmException e) {
				LOG.error("SSL algorithm is NOT available in the environment", e);
			} catch (CertificateException e) {
				LOG.error("Unable to obtain the requested certification ", e);
			} catch (FileNotFoundException e) {
				LOG.error("Unable to find the necessary SSL Keystore Files", e);
			} catch (IOException e) {
				LOG.error("Unable to read the necessary SSL Keystore Files", e);
			} catch (UnrecoverableKeyException e) {
				LOG.error("Unable to recover the key from keystore", e);
			} finally {
				close(in, keyStoreFile);
			}
		}
		return kmList;
	}

	private TrustManager[] getTrustManagers() {
		TrustManager[] tmList = null;
		String credentialProviderPath = MiscUtil.getStringProperty(props, RANGER_POLICYMGR_TRUSTSTORE_FILE_CREDENTIAL);
		String trustStoreAlias = RANGER_POLICYMGR_TRUSTSTORE_FILE_CREDENTIAL_ALIAS;
		String trustStoreFile = MiscUtil.getStringProperty(props, RANGER_POLICYMGR_TRUSTSTORE_FILE);
		String trustStoreFilepwd = MiscUtil.getCredentialString(credentialProviderPath, trustStoreAlias);
		if (StringUtils.isNotEmpty(trustStoreFile) && StringUtils.isNotEmpty(trustStoreFilepwd)) {
			InputStream in = null;
			try {
				in = getFileInputStream(trustStoreFile);

				if (in != null) {
					String trustStoreType = MiscUtil.getStringProperty(props, RANGER_POLICYMGR_TRUSTSTORE_FILE_TYPE);
					trustStoreType = StringUtils.isNotEmpty(trustStoreType) ? trustStoreType : RANGER_POLICYMGR_TRUSTSTORE_FILE_TYPE_DEFAULT;
					KeyStore trustStore = KeyStore.getInstance(trustStoreType);

					trustStore.load(in, trustStoreFilepwd.toCharArray());

					TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(RANGER_SSL_TRUSTMANAGER_ALGO_TYPE);

					trustManagerFactory.init(trustStore);

					tmList = trustManagerFactory.getTrustManagers();
				} else {
					LOG.error("Unable to obtain truststore from file [" + trustStoreFile + "]");
				}
			} catch (KeyStoreException e) {
				LOG.error("Unable to obtain from KeyStore", e);
			} catch (NoSuchAlgorithmException e) {
				LOG.error("SSL algorithm is NOT available in the environment :" + e.getMessage(), e);
			} catch (CertificateException e) {
				LOG.error("Unable to obtain the requested certification :" + e.getMessage(), e);
			} catch (FileNotFoundException e) {
				LOG.error("Unable to find the necessary SSL TrustStore File:" + trustStoreFile, e);
			} catch (IOException e) {
				LOG.error("Unable to read the necessary SSL TrustStore Files :" + trustStoreFile, e);
			} finally {
				close(in, trustStoreFile);
			}
		}
		return tmList;
	}

	private SSLContext getSSLContext(KeyManager[] kmList, TrustManager[] tmList) {
		SSLContext sslContext = null;
		try {
			sslContext = SSLContext.getInstance(RANGER_SSL_CONTEXT_ALGO_TYPE);
			if (sslContext != null) {
				sslContext.init(kmList, tmList, new SecureRandom());
			}
		} catch (NoSuchAlgorithmException e) {
			LOG.error("SSL algorithm is not available in the environment", e);
		} catch (KeyManagementException e) {
			LOG.error("Unable to initialise the SSLContext", e);
		}
		return sslContext;
	}

	private InputStream getFileInputStream(String fileName) throws IOException {
		InputStream in = null;
		if (StringUtils.isNotEmpty(fileName)) {
			File file = new File(fileName);
			if (file != null && file.exists()) {
				in = new FileInputStream(file);
			} else {
				in = ClassLoader.getSystemResourceAsStream(fileName);
			}
		}
		return in;
	}

	private class AuditHTTPRetryHandler extends StandardHttpRequestRetryHandler {

		public AuditHTTPRetryHandler(Integer poolRetryCount, boolean isIdempotentRequest) {
			super(poolRetryCount, isIdempotentRequest);
		}

		@Override
		public boolean retryRequest(IOException exception, int executionCount, HttpContext context) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("==> AuditHTTPRetryHandler.retryRequest" +  exception.getMessage() + "Execution Count = " + executionCount);
			}

			boolean ret = super.retryRequest(exception, executionCount, context);

			if (LOG.isDebugEnabled()) {
				LOG.debug("<== AuditHTTPRetryHandler.retryRequest(): ret= " + ret);
			}

			return ret;
		}
	}
}
