/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.plugin.contextenricher;

import com.sun.jersey.api.client.ClientResponse;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.admin.client.datatype.RESTResponse;
import org.apache.ranger.audit.provider.MiscUtil;
import org.apache.ranger.plugin.util.RangerRESTClient;
import org.apache.ranger.plugin.util.RangerRESTUtils;
import org.apache.ranger.plugin.util.RangerUserStore;

import javax.servlet.http.HttpServletResponse;
import java.nio.channels.ClosedByInterruptException;
import java.security.PrivilegedAction;
import java.util.HashMap;
import java.util.Map;

public class RangerRESTClientUserStoreRetriever extends RangerUserStoreRetriever {
    private static final Log LOG = LogFactory.getLog(RangerRESTClientUserStoreRetriever.class);

    public RangerRESTClientUserStoreRetriever(RangerRESTClient rangerRESTClient) {
        super(rangerRESTClient);
        init(null);
    }

    @Override
    public void init(Map<String, String> options) {

        if (rangerRESTClient == null) {
            LOG.error("FATAL: ranger rest client to retrieving userstore is null. Will NOT be able to retrieve userstore.");
        }
    }

    @Override
    public RangerUserStore retrieveUserStoreInfo(long lastKnownUserStoreVersion, long lastActivationTimeInMillis) throws Exception {

        RangerUserStore rangerUserStore = null;

        try {
            rangerUserStore = getUserStoreIfUpdated(lastKnownUserStoreVersion, lastActivationTimeInMillis);
        } catch (ClosedByInterruptException closedByInterruptException) {
            LOG.error("UserStore-retriever for raz thread was interrupted while blocked on I/O");
            throw new InterruptedException();
        } catch (Exception e) {
            LOG.error("UserStore-retriever for raz encounterd exception, exception=", e);
            LOG.error("Returning null userstore info");
        }
        return rangerUserStore;
    }

    private RangerUserStore getUserStoreIfUpdated(long lastKnownUserStoreVersion, long lastActivationTimeInMillis) throws Exception {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRESTClientUserStoreRetriever.getUserStoreIfUpdated(" + lastKnownUserStoreVersion + ", " + lastActivationTimeInMillis + ")");
        }

        final RangerUserStore ret;
        final UserGroupInformation user = MiscUtil.getUGILoginUser();
        final boolean isSecureMode = user != null && UserGroupInformation.isSecurityEnabled();
        final ClientResponse response;

        Map<String, String> queryParams = new HashMap<String, String>();
        queryParams.put(RangerRESTUtils.REST_PARAM_LAST_KNOWN_USERSTORE_VERSION, Long.toString(lastKnownUserStoreVersion));
        queryParams.put(RangerRESTUtils.REST_PARAM_LAST_ACTIVATION_TIME, Long.toString(lastActivationTimeInMillis));

        if (isSecureMode) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Checking UserStore updated as user : " + user);
            }
            PrivilegedAction<ClientResponse> action = new PrivilegedAction<ClientResponse>() {
                public ClientResponse run() {
                    ClientResponse clientRes = null;
                    String relativeURL = RangerRESTUtils.REST_URL_SERVICE_SERCURE_GET_USERSTORE;
                    try {
                        clientRes =  rangerRESTClient.get(relativeURL, queryParams);
                    } catch (Exception e) {
                        LOG.error("Failed to get response, Error is : "+e.getMessage());
                    }
                    return clientRes;
                }
            };
            response = user.doAs(action);
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Checking UserStore updated as user : " + user);
            }
            String relativeURL = RangerRESTUtils.REST_URL_SERVICE_SERCURE_GET_USERSTORE;
            response = rangerRESTClient.get(relativeURL, queryParams);
        }

        if (response == null || response.getStatus() == HttpServletResponse.SC_NOT_MODIFIED) {
            if (response == null) {
                LOG.error("Error getting UserStore; Received NULL response!!. secureMode=" + isSecureMode + ", user=" + user);
            } else {
                RESTResponse resp = RESTResponse.fromClientResponse(response);
                if (LOG.isDebugEnabled()) {
                    LOG.debug("No change in UserStore. secureMode=" + isSecureMode + ", user=" + user
                            + ", response=" + resp
                            + ", " + "lastKnownUserStoreVersion=" + lastKnownUserStoreVersion
                            + ", " + "lastActivationTimeInMillis=" + lastActivationTimeInMillis);
                }
            }
            ret = null;
        } else if (response.getStatus() == HttpServletResponse.SC_OK) {
            ret = response.getEntity(RangerUserStore.class);
        } else if (response.getStatus() == HttpServletResponse.SC_NOT_FOUND) {
            ret = null;
            LOG.error("Error getting UserStore; service not found. secureMode=" + isSecureMode + ", user=" + user
                    + ", response=" + response.getStatus()
                    + ", " + "lastKnownUserStoreVersion=" + lastKnownUserStoreVersion
                    + ", " + "lastActivationTimeInMillis=" + lastActivationTimeInMillis);
            String exceptionMsg = response.hasEntity() ? response.getEntity(String.class) : null;
            LOG.warn("Received 404 error code with body:[" + exceptionMsg + "], Ignoring");
        } else {
            RESTResponse resp = RESTResponse.fromClientResponse(response);
            LOG.warn("Error getting UserStore. secureMode=" + isSecureMode + ", user=" + user + ", response=" + resp);
            ret = null;
        }

        if(LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRESTClientUserStoreRetriever.getUserStoreIfUpdated(" + lastKnownUserStoreVersion + ", " + lastActivationTimeInMillis + "): ");
        }

        return ret;
    }

}

