/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.plugin.contextenricher;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.plugin.store.RangerUserStoreListener;
import org.apache.ranger.plugin.util.RangerPerfTracer;
import org.apache.ranger.plugin.util.DownloadTrigger;
import org.apache.ranger.plugin.util.RangerUserStore;
import org.apache.ranger.plugin.util.RangerServiceNotFoundException;

import java.io.File;
import java.io.Reader;
import java.io.Writer;
import java.io.FileWriter;
import java.io.FileReader;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.BlockingQueue;

public class RangerUserStoreRefresher extends Thread {
    private static final Log LOG = LogFactory.getLog(RangerUserStoreRefresher.class);
    private static final Log PERF_REFRESHER_INIT_LOG = RangerPerfTracer.getPerfLogger("userstore.init");

    private final RangerUserStoreRetriever userStoreRetriever;
    private final RangerUserStoreEnricher userStoreEnricher;
    private long lastKnownVersion;
    private final BlockingQueue<DownloadTrigger> userStoreDownloadQueue;
    private long lastActivationTimeInMillis;

    private final String cacheFile;
    private boolean hasProvidedUserStoreToReceiver;
    private Gson gson;
    private static Set<RangerUserStoreListener> listeners = new HashSet<>();

    public RangerUserStoreRefresher(RangerUserStoreRetriever userStoreRetriever, RangerUserStoreEnricher userStoreEnricher,
                                    RangerUserStoreListener rangerUserStoreListener, long lastKnownVersion,
                                    BlockingQueue<DownloadTrigger> userStoreDownloadQueue, String cacheFile) {
        this.userStoreRetriever = userStoreRetriever;
        this.userStoreEnricher = userStoreEnricher;
        this.lastKnownVersion = lastKnownVersion;
        this.userStoreDownloadQueue = userStoreDownloadQueue;
        this.cacheFile = cacheFile;
        registerUserStoreChangeListener(rangerUserStoreListener);
        try {
            gson = new GsonBuilder().setDateFormat("yyyyMMdd-HH:mm:ss.SSS-Z").create();
        } catch(Throwable excp) {
            LOG.fatal("failed to create GsonBuilder object", excp);
        }
        if (userStoreRetriever != null) {
            setName("RangerUserStoreRefresher(serviceName=" + userStoreRetriever.getServiceName() + ")-" + getId());
        } else {
            setName("RangerUserStoreRefresher(rangerraz)-" + getId());
        }
    }

    public long getLastActivationTimeInMillis() {
        return lastActivationTimeInMillis;
    }

    public void setLastActivationTimeInMillis(long lastActivationTimeInMillis) {
        this.lastActivationTimeInMillis = lastActivationTimeInMillis;
    }

    @Override
    public void run() {

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerUserStoreRefresher().run()");
        }

        while (true) {
            DownloadTrigger trigger = null;

            try {
                RangerPerfTracer perf = null;

                if(RangerPerfTracer.isPerfTraceEnabled(PERF_REFRESHER_INIT_LOG)) {
                    perf = RangerPerfTracer.getPerfTracer(PERF_REFRESHER_INIT_LOG,
                            "RangerUserStoreRefresher.run(lastKnownVersion=" + lastKnownVersion + ")");
                }
                trigger = userStoreDownloadQueue.take();
                populateUserStoreInfo();

                RangerPerfTracer.log(perf);

            } catch (InterruptedException excp) {
                LOG.debug("RangerUserStoreRefresher().run() : interrupted! Exiting thread", excp);
                break;
            } finally {
                if (trigger != null) {
                    trigger.signalCompletion();
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerUserStoreRefresher().run()");
        }
    }

    public void populateUserStoreInfo() throws InterruptedException {

        RangerUserStore rangerUserStore = null;
        if (userStoreEnricher != null && userStoreRetriever != null) {
            try {
                rangerUserStore = userStoreRetriever.retrieveUserStoreInfo(lastKnownVersion, lastActivationTimeInMillis);

                if (rangerUserStore == null) {
                    if (!hasProvidedUserStoreToReceiver) {
                        rangerUserStore = loadFromCache();
                    }
                }

                if (rangerUserStore != null) {
                    userStoreEnricher.setRangerUserStore(rangerUserStore);
                    if (rangerUserStore.getUserStoreVersion() != -1L) {
                        saveToCache(rangerUserStore);
                    }
                    LOG.info("RangerUserStoreRefresher.populateUserStoreInfo() - Updated userstore-cache to new version, lastKnownVersion=" + lastKnownVersion + "; newVersion="
                            + (rangerUserStore.getUserStoreVersion() == null ? -1L : rangerUserStore.getUserStoreVersion()));
                    hasProvidedUserStoreToReceiver = true;
                    lastKnownVersion = rangerUserStore.getUserStoreVersion() == null ? -1L : rangerUserStore.getUserStoreVersion();
                    setLastActivationTimeInMillis(System.currentTimeMillis());
                } else {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("RangerUserStoreRefresher.populateUserStoreInfo() - No need to update userstore-cache. lastKnownVersion=" + lastKnownVersion);
                    }
                }
            } catch (RangerServiceNotFoundException snfe) {
                LOG.error("Caught ServiceNotFound exception :", snfe);

                // Need to clean up local userstore cache
                if (userStoreEnricher.isDisableCacheIfServiceNotFound()) {
                    disableCache();
                    setLastActivationTimeInMillis(System.currentTimeMillis());
                    lastKnownVersion = -1L;
                }
            } catch (InterruptedException interruptedException) {
                throw interruptedException;
            } catch (Exception e) {
                LOG.error("Encountered unexpected exception. Ignoring", e);
            }
        } else if (userStoreRetriever != null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("RangerUserStoreRefresher.populateUserStoreInfo() for Ranger Raz");
            }
            try {
                rangerUserStore = userStoreRetriever.retrieveUserStoreInfo(lastKnownVersion, lastActivationTimeInMillis);

                if (rangerUserStore == null) {
                    if (!hasProvidedUserStoreToReceiver) {
                        rangerUserStore = loadFromCache();
                    }
                }

                if (rangerUserStore != null) {

                    for (RangerUserStoreListener listner : listeners) {
                        LOG.info("Updating userstore info for " + listner.getClass());
                        listner.updateRangerUserStore(rangerUserStore);
                    }
                    if (rangerUserStore.getUserStoreVersion() != -1L) {
                        saveToCache(rangerUserStore);
                    }
                    LOG.info("RangerUserStoreRefresher.populateUserStoreInfo() - Updated userstore-cache for raz to new version, lastKnownVersion=" + lastKnownVersion + "; newVersion="
                            + (rangerUserStore.getUserStoreVersion() == null ? -1L : rangerUserStore.getUserStoreVersion()));
                    hasProvidedUserStoreToReceiver = true;
                    lastKnownVersion = rangerUserStore.getUserStoreVersion() == null ? -1L : rangerUserStore.getUserStoreVersion();
                    setLastActivationTimeInMillis(System.currentTimeMillis());
                } else {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("RangerUserStoreRefresher.populateUserStoreInfo() - No need to update userstore-cache for raz. lastKnownVersion=" + lastKnownVersion);
                    }
                }
            }catch (InterruptedException interruptedException) {
                throw interruptedException;
            } catch (Exception e) {
                LOG.error("Encountered unexpected exception. Ignoring", e);
            }
        }
        else {
            LOG.error("RangerUserStoreRefresher.populateUserStoreInfo() - no userstore receiver to update userstore-cache");
        }
    }

    public void cleanup() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerUserStoreRefresher.cleanup()");
        }

        stopRefresher();

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerUserStoreRefresher.cleanup()");
        }
    }

    public void startRefresher() {
        try {
            super.start();
        } catch (Exception excp) {
            LOG.error("RangerUserStoreRefresher.startRetriever() - failed to start, exception=" + excp);
        }
    }

    public void stopRefresher() {

        if (super.isAlive()) {
            super.interrupt();

            boolean setInterrupted = false;
            boolean isJoined = false;

            while (!isJoined) {
                try {
                    super.join();
                    isJoined = true;
                } catch (InterruptedException excp) {
                    LOG.warn("RangerUserStoreRefresher(): Error while waiting for thread to exit", excp);
                    LOG.warn("Retrying Thread.join(). Current thread will be marked as 'interrupted' after Thread.join() returns");
                    setInterrupted = true;
                }
            }
            if (setInterrupted) {
                Thread.currentThread().interrupt();
            }
        }
    }


    private RangerUserStore loadFromCache() {
        RangerUserStore rangerUserStore = null;

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerUserStoreRefreher.loadFromCache()");
        }

        File cacheFile = StringUtils.isEmpty(this.cacheFile) ? null : new File(this.cacheFile);

        if (cacheFile != null && cacheFile.isFile() && cacheFile.canRead()) {
            Reader reader = null;

            try {
                reader = new FileReader(cacheFile);

                rangerUserStore = gson.fromJson(reader, RangerUserStore.class);

            } catch (Exception excp) {
                LOG.error("failed to load userstore information from cache file " + cacheFile.getAbsolutePath(), excp);
            } finally {
                if (reader != null) {
                    try {
                        reader.close();
                    } catch (Exception excp) {
                        LOG.error("error while closing opened cache file " + cacheFile.getAbsolutePath(), excp);
                    }
                }
            }
        } else {
            LOG.warn("cache file does not exist or not readable '" + (cacheFile == null ? null : cacheFile.getAbsolutePath()) + "'");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerUserStoreRefreher.loadFromCache()");
        }

        return rangerUserStore;
    }

    public void saveToCache(RangerUserStore rangerUserStore) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerUserStoreRefreher.saveToCache()");
        }

        if (rangerUserStore != null) {
            File cacheFile = StringUtils.isEmpty(this.cacheFile) ? null : new File(this.cacheFile);

            if (cacheFile != null) {
                Writer writer = null;

                try {
                    writer = new FileWriter(cacheFile);

                    gson.toJson(rangerUserStore, writer);
                } catch (Exception excp) {
                    LOG.error("failed to save userstore information to cache file '" + cacheFile.getAbsolutePath() + "'", excp);
                } finally {
                    if (writer != null) {
                        try {
                            writer.close();
                        } catch (Exception excp) {
                            LOG.error("error while closing opened cache file '" + cacheFile.getAbsolutePath() + "'", excp);
                        }
                    }
                }
            }
        } else {
            LOG.info("userstore information is null. Nothing to save in cache");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerUserStoreRefreher.saveToCache()");
        }
    }

    private void disableCache() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerUserStoreRefreher.disableCache()");
        }

        File cacheFile = StringUtils.isEmpty(this.cacheFile) ? null : new File(this.cacheFile);
        if (cacheFile != null && cacheFile.isFile() && cacheFile.canRead()) {
            LOG.warn("Cleaning up local userstore cache");
            String renamedCacheFile = cacheFile.getAbsolutePath() + "_" + System.currentTimeMillis();
            if (!cacheFile.renameTo(new File(renamedCacheFile))) {
                LOG.error("Failed to move " + cacheFile.getAbsolutePath() + " to " + renamedCacheFile);
            } else {
                LOG.warn("moved " + cacheFile.getAbsolutePath() + " to " + renamedCacheFile);
            }
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("No local userstore cache found. No need to disable it!");
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerUserStoreRefreher.disableCache()");
        }
    }

    public void registerUserStoreChangeListener(RangerUserStoreListener listener) {
        if (listener != null) {
            listeners.add(listener);
        }
    }

    public void unRegisterUserStoreChangeListener(RangerUserStoreListener listener) {
        if (listener != null) {
            listeners.remove(listener);
        }
    }
}
