/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.audit.server;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.*;

import org.apache.catalina.Context;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.WebResourceRoot;
import org.apache.catalina.connector.Connector;
import org.apache.catalina.core.StandardContext;
import org.apache.catalina.startup.Tomcat;
import org.apache.catalina.valves.AccessLogValve;
import org.apache.catalina.webresources.StandardRoot;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.alias.BouncyCastleFipsKeyStoreProvider;
import org.apache.hadoop.security.alias.CredentialProvider;
import org.apache.hadoop.security.alias.CredentialProviderFactory;
import org.apache.hadoop.security.alias.JavaKeyStoreProvider;
import org.apache.hadoop.security.alias.LocalBouncyCastleFipsKeyStoreProvider;
import org.apache.log4j.PropertyConfigurator;
import org.apache.ranger.authorization.hadoop.utils.RangerCredentialProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;


public class EmbeddedServer {
    private static final Logger LOG = LoggerFactory.getLogger(EmbeddedServer.class);
    private static final String RANGER_KEYSTORE_FILE_TYPE_DEFAULT   = "jks";
    private static final String RANGER_TRUSTSTORE_FILE_TYPE_DEFAULT = "jks";
    private static final String RANGER_SSL_CONTEXT_ALGO_TYPE        = "TLSv1.2";
    private static final String RANGER_SSL_KEYMANAGER_ALGO_TYPE     = KeyManagerFactory.getDefaultAlgorithm();
    private static final String RANGER_SSL_TRUSTMANAGER_ALGO_TYPE   = TrustManagerFactory.getDefaultAlgorithm();
    private static final String ACCESS_LOG_PREFIX                   = "accesslog.prefix";
    private static final String ACCESS_LOG_DATE_FORMAT              = "accesslog.dateformat";
    private static final String ACCESS_LOG_PATTERN                  = "accesslog.pattern";
    private static final String ACCESS_LOG_ROTATE_MAX_DAYS          = "accesslog.rotate.max.days";
    private static final String DEFAULT_LOG_DIR                     = "/tmp";
    private static final int    DEFAULT_HTTP_PORT                   = 6081;
    private static final int    DEFAULT_HTTPS_PORT                  = -1;
    private static final int    DEFAULT_SHUTDOWN_PORT               = 6185;
    private static final String DEFAULT_SHUTDOWN_COMMAND            = "SHUTDOWN";
    private static final String DEFAULT_WEBAPPS_ROOT_FOLDER         = "webapps";
    private static final String DEFAULT_ENABLED_PROTOCOLS           = "TLSv1.2";
    public static final String  KEYSTORE_FILE_TYPE_DEFAULT          = KeyStore.getDefaultType();
    public static final String  TRUSTSTORE_FILE_TYPE_DEFAULT        = KeyStore.getDefaultType();
    public static final String  HADOOP_AUTHENTICATION_TYPE          = "hadoop.security.authentication";

    private final Configuration configuration;
    private final String        appName;
    private final String        configPrefix;

    public static void main(String[] args) {
        Configuration config       = AuditConfig.getInstance();
        String        appName      = "ranger-audit";
        String        configPrefix = "ranger.audit.";

        new EmbeddedServer(config, appName, configPrefix).start();
    }

    public EmbeddedServer(Configuration configuration, String appName, String configPrefix) {
        LOG.info("==> EmbeddedServer(appName={}, configPrefix={})", appName, configPrefix);

        try {
            // load log configuration file dynamically if log4j.properties changed
            if (StringUtils.isNotBlank(System.getProperty("log4j.configuration"))) {
                String logPropFile = StringUtils.splitByWholeSeparator(System.getProperty("log4j.configuration"), ":")[1];
                PropertyConfigurator.configureAndWatch(logPropFile, 10000L);
            }
        } catch (Exception ignored) {
            LOG.warn("Failed to get log4j.configuration  Reason: " + ignored.toString());
        }

        this.configuration = configuration;
        this.appName       = appName;
        this.configPrefix  = configPrefix;

        LOG.info("<== EmbeddedServer(appName={}, configPrefix={}", appName, configPrefix);
    }

    @SuppressWarnings("deprecation")
    public void start() {
        LOG.info("==> EmbeddedServer.start(appName={})", appName);

        SSLContext sslContext = getSSLContext();

        if (sslContext != null) {
            SSLContext.setDefault(sslContext);
        }

        final Tomcat server = new Tomcat();

        String  logDir          = getConfig("log.dir", DEFAULT_LOG_DIR);
        String  hostName        = getConfig("service.host");
        int     serverPort      = getIntConfig("service.http.port", DEFAULT_HTTP_PORT);
        int     sslPort         = getIntConfig("service.https.port", DEFAULT_HTTPS_PORT);
        int     shutdownPort    = getIntConfig("service.shutdown.port", DEFAULT_SHUTDOWN_PORT);
        String  shutdownCommand = getConfig("service.shutdown.command", DEFAULT_SHUTDOWN_COMMAND);
        boolean isHttpsEnabled  = getBooleanConfig("service.https.attrib.ssl.enabled", false);
        boolean ajpEnabled      = getBooleanConfig("ajp.enabled", false);

        server.setHostname(hostName);
        server.setPort(serverPort);
        server.getServer().setPort(shutdownPort);
        server.getServer().setShutdown(shutdownCommand);

        if (ajpEnabled) {
            Connector ajpConnector = new Connector("org.apache.coyote.ajp.AjpNioProtocol");

            ajpConnector.setPort(serverPort);
            ajpConnector.setProperty("protocol", "AJP/1.3");

            server.getService().addConnector(ajpConnector);

            // Making this as a default connector
            server.setConnector(ajpConnector);

            LOG.info("Created AJP Connector");
        } else if (isHttpsEnabled && sslPort > 0) {
            String clientAuth              = getConfig("service.https.attrib.clientAuth", "false");
            String providerPath            = getConfig("credential.provider.path");
            String keyAlias                = getConfig("service.https.attrib.keystore.credential.alias", "keyStoreCredentialAlias");
            String keystorePass            = null;
            String enabledProtocols        = getConfig("service.https.attrib.ssl.enabled.protocols", DEFAULT_ENABLED_PROTOCOLS);
            String ciphers                 = getConfig("tomcat.ciphers");

            if (StringUtils.equalsIgnoreCase(clientAuth, "false")) {
                clientAuth = getConfig("service.https.attrib.client.auth", "want");
            }

            if (providerPath != null && keyAlias != null) {
                keystorePass = getDecryptedString(providerPath.trim(), keyAlias.trim(), getConfig("ranger-raz.keystore.file.type", KEYSTORE_FILE_TYPE_DEFAULT));

                if (StringUtils.isBlank(keystorePass) || StringUtils.equalsIgnoreCase(keystorePass.trim(), "none")) {
                    keystorePass = getConfig("service.https.attrib.keystore.pass");
                }
            }

            Connector ssl = new Connector();
            String sslKeystoreKeyAlias = getConfig("service.https.attrib.keystore.keyalias","");
            ssl.setPort(sslPort);
            ssl.setSecure(true);
            ssl.setScheme("https");
            ssl.setAttribute("SSLEnabled", "true");
            ssl.setAttribute("sslProtocol", getConfig("service.https.attrib.ssl.protocol", "TLSv1.2"));
            ssl.setAttribute("clientAuth",clientAuth);
            if (StringUtils.isNotBlank(sslKeystoreKeyAlias)) {
                ssl.setAttribute("keyAlias", sslKeystoreKeyAlias);
            }
            ssl.setAttribute("keystorePass", keystorePass);
            ssl.setAttribute("keystoreFile", getKeystoreFile());
            ssl.setAttribute("sslEnabledProtocols", enabledProtocols);
            ssl.setAttribute("keystoreType", getConfig("audit.keystore.file.type", KEYSTORE_FILE_TYPE_DEFAULT));
            ssl.setAttribute("truststoreType", getConfig("audit.truststore.file.type", TRUSTSTORE_FILE_TYPE_DEFAULT));

            if (StringUtils.isNotBlank(ciphers)) {
                ssl.setAttribute("ciphers", ciphers);
            }

            server.getService().addConnector(ssl);

            //
            // Making this as a default connector
            //
            server.setConnector(ssl);
        }

        updateHttpConnectorAttribConfig(server);

        File logDirectory = new File(logDir);

        if (!logDirectory.exists()) {
            logDirectory.mkdirs();
        }

        String logPattern = getConfig(ACCESS_LOG_PATTERN, "%h %l %u %t \"%r\" %s %b");

        AccessLogValve valve = new AccessLogValve();

        valve.setRotatable(true);
        valve.setAsyncSupported(true);
        valve.setBuffered(false);
        valve.setEnabled(true);
        valve.setPrefix(getConfig(ACCESS_LOG_PREFIX,"access_log-" + hostName +"-"));
        valve.setFileDateFormat(getConfig(ACCESS_LOG_DATE_FORMAT, "yyyy-MM-dd.HH"));
        valve.setDirectory(logDirectory.getAbsolutePath());
        valve.setSuffix(".log");
        valve.setPattern(logPattern);
        valve.setMaxDays(getIntConfig(ACCESS_LOG_ROTATE_MAX_DAYS, 15));

        server.getHost().getPipeline().addValve(valve);

        try {
            String webappDir = getConfig("webapp.dir");

            if (StringUtils.isBlank(webappDir)) {
                LOG.error("Tomcat Server failed to start: {}.webapp.dir is not set", configPrefix);

                System.exit(1);
            }

            String webContextName = getConfig("contextName", "");

            if (StringUtils.isBlank(webContextName)) {
                webContextName = "";
            } else if (!webContextName.startsWith("/")) {
                LOG.info("Context Name [{}] is being loaded as [ /{}]", webContextName, webContextName);

                webContextName = "/" + webContextName;
            }

            File wad = new File(webappDir);

            if (wad.isDirectory()) {
                LOG.info("Webapp dir={}, webAppName={}", webappDir, webContextName);
            } else if (wad.isFile()) {
                File webAppDir = new File(DEFAULT_WEBAPPS_ROOT_FOLDER);

                if (!webAppDir.exists()) {
                    webAppDir.mkdirs();
                }

                LOG.info("Webapp file={}, webAppName={}", webappDir, webContextName);
            }

            LOG.info("Adding webapp [{}] = path [{}] .....", webContextName, webappDir);

            Context webappCtx = server.addWebapp(webContextName, new File(webappDir).getAbsolutePath());
            if (webappCtx instanceof StandardContext) {
                boolean allowLinking = getBooleanConfig("allow.linking", true);
                StandardContext standardContext = (StandardContext) webappCtx;
                String workDirPath = getConfig("tomcat.work.dir", "");
                if (!workDirPath.isEmpty() && new File(workDirPath).exists()) {
                    standardContext.setWorkDir(workDirPath);
                } else {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Skipping to set tomcat server work directory, '" + workDirPath +
                                "', as it is blank or directory does not exist.");
                    }
                }
                WebResourceRoot resRoot = new StandardRoot(standardContext);
                webappCtx.setResources(resRoot);
                webappCtx.getResources().setAllowLinking(allowLinking);
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Tomcat Configuration - allowLinking=[{}]", allowLinking);
                }
            } else {
                LOG.error("Tomcat Context [{}] is either NULL OR it's NOT instanceof StandardContext", webappCtx);
            }

            webappCtx.init();

            LOG.info("Finished init of webapp [{}] = path [{}].", webContextName, webappDir);
        } catch (LifecycleException lce) {
            LOG.error("Tomcat Server failed to start webapp", lce);

            lce.printStackTrace();
        }

        LOG.info("Starting server in non-kerberos mode");

        startServer(server);

        LOG.info("<== EmbeddedServer.start(appName={})", appName);
    }


    private void startServer(final Tomcat server) {
        LOG.info("==> EmbeddedServer.startServer(appName={})", appName);

        try {
            server.start();

            server.getServer().await();

            shutdownServer();
        } catch (LifecycleException e) {
            LOG.error("Tomcat Server failed to start", e);

            e.printStackTrace();
        } catch (Exception e) {
            LOG.error("Tomcat Server failed to start", e);

            e.printStackTrace();
        }

        LOG.info("<== EmbeddedServer.startServer(appName={})", appName);
    }

    private void shutdownServer() {
        LOG.info("==> EmbeddedServer.shutdownServer(appName={})", appName);

        int timeWaitForShutdownInSeconds = getIntConfig("service.waitTimeForForceShutdownInSeconds", 0);

        if (timeWaitForShutdownInSeconds > 0) {
            long endTime = System.currentTimeMillis() + (timeWaitForShutdownInSeconds * 1000L);

            LOG.info("Will wait for all threads to shutdown gracefully. Final shutdown Time: " + new Date(endTime));

            while (System.currentTimeMillis() < endTime) {
                int activeCount = Thread.activeCount();

                if (activeCount == 0) {
                    LOG.info("Number of active threads = " + activeCount + ".");

                    break;
                } else {
                    LOG.info("Number of active threads = {}. Waiting for all threads to shutdown ...", activeCount);

                    try {
                        Thread.sleep(5000L);
                    } catch (InterruptedException e) {
                        LOG.warn("shutdownServer process is interrupted with exception", e);

                        break;
                    }
                }
            }
        }

        LOG.info("Shuting down the Server.");

        LOG.info("<== EmbeddedServer.shutdownServer(appName={})", appName);

        System.exit(0);
    }

    @SuppressWarnings("deprecation")
    private void updateHttpConnectorAttribConfig(Tomcat server) {
        server.getConnector().setAllowTrace(getBooleanConfig("service.http.connector.attrib.allowTrace", false));
        server.getConnector().setAsyncTimeout(getLongConfig("service.http.connector.attrib.asyncTimeout", 10000));
        server.getConnector().setEnableLookups(getBooleanConfig("service.http.connector.attrib.enableLookups", false));
        server.getConnector().setMaxParameterCount(getIntConfig("service.http.connector.attrib.maxParameterCount", 10000));
        server.getConnector().setMaxPostSize(getIntConfig("service.http.connector.attrib.maxPostSize", 2097152));
        server.getConnector().setMaxSavePostSize(getIntConfig("service.http.connector.attrib.maxSavePostSize", 4096));
        server.getConnector().setParseBodyMethods(getConfig("service.http.connector.attrib.methods", "POST"));
        server.getConnector().setURIEncoding(getConfig("service.http.connector.attrib.URIEncoding", "UTF-8"));
        server.getConnector().setProperty("acceptCount", getConfig("service.http.connector.attrib.acceptCount", "1024"));
        server.getConnector().setXpoweredBy(false);
        server.getConnector().setAttribute("server", getConfig("servername", "Audit Server"));

        int maxThreads                  = getIntConfig("service.http.connector.attrib.maxThreads",                  2000);                      // (default 2000)
        int maxKeepAliveRequests        = getIntConfig("service.http.connector.attrib.maxKeepAliveRequests",        1000);                      // (default 1000)
        int minSpareThreads             = getIntConfig("service.http.connector.attrib.minSpareThreads",             (int) (maxThreads * 0.8));  // (default 80% of maxThreads)
        boolean prestartminSpareThreads = getBooleanConfig("service.http.connector.attrib.prestartminSpareThreads", true);                      // (default TRUE)

        server.getConnector().setAttribute("maxThreads",              maxThreads);
        server.getConnector().setAttribute("maxKeepAliveRequests",    maxKeepAliveRequests);
        server.getConnector().setAttribute("minSpareThreads",         minSpareThreads);
        server.getConnector().setAttribute("prestartminSpareThreads", prestartminSpareThreads);

        server.getConnector().setProperty("sendReasonPhrase",getConfig(configPrefix + "service.http.connector.property.sendReasonPhrase", "true"));

        for (Map.Entry<String, String> entry : configuration) {
            String key = entry.getKey();
            String value = entry.getValue();

            if (key != null && key.startsWith(configPrefix + "service.http.connector.property.")) {
                String property = key.replace(configPrefix + "service.http.connector.property.", "");

                server.getConnector().setProperty(property, value);

                LOG.info("{}:{}", property, server.getConnector().getProperty(property));
            }
        }
    }

    public String getDecryptedString(String crendentialProviderPath, String alias, String storeType) {
        String ret = null;

        if (StringUtils.isNotBlank(crendentialProviderPath) && StringUtils.isNotBlank(alias)) {
            try {
                Configuration conf             = new Configuration();
                String        prefixJceks      = JavaKeyStoreProvider.SCHEME_NAME + "://file";
                String        prefixLocalJceks = "localjceks://file";
                String        prefixBcfks      = BouncyCastleFipsKeyStoreProvider.SCHEME_NAME + "://file";
                String        prefixLocalBcfks = LocalBouncyCastleFipsKeyStoreProvider.SCHEME_NAME + "://file";

                if (StringUtils.startsWithIgnoreCase(crendentialProviderPath, prefixJceks) ||
                        StringUtils.startsWithIgnoreCase(crendentialProviderPath, prefixLocalJceks) ||
                        StringUtils.startsWithIgnoreCase(crendentialProviderPath, prefixBcfks) ||
                        StringUtils.startsWithIgnoreCase(crendentialProviderPath, prefixLocalBcfks)) {
                    conf.set(CredentialProviderFactory.CREDENTIAL_PROVIDER_PATH, crendentialProviderPath);
                } else {
                    if (crendentialProviderPath.startsWith("/")) {
                        if (StringUtils.equalsIgnoreCase(storeType, "bcfks")) {
                            conf.set(CredentialProviderFactory.CREDENTIAL_PROVIDER_PATH, crendentialProviderPath);
                        } else {
                            conf.set(CredentialProviderFactory.CREDENTIAL_PROVIDER_PATH, prefixJceks + crendentialProviderPath);
                        }
                    } else {
                        conf.set(CredentialProviderFactory.CREDENTIAL_PROVIDER_PATH, prefixJceks + "/" + crendentialProviderPath);
                    }
                }

                List<CredentialProvider> providers = CredentialProviderFactory.getProviders(conf);

                if (LOG.isDebugEnabled()) {
                    LOG.debug("crendentialProviderPath="+ crendentialProviderPath + " alias="+ alias + " storeType="+ storeType);
                    LOG.debug("List of CredentialProvider = " + providers.toString());
                }

                for (CredentialProvider provider : providers) {
                    //System.out.println("Credential Provider :" + provider);
                    List<String> aliasesList = provider.getAliases();

                    if (CollectionUtils.isNotEmpty(aliasesList) && aliasesList.contains(alias)) {
                        CredentialProvider.CredentialEntry credEntry = provider.getCredentialEntry(alias);
                        char[]                             pass      = credEntry.getCredential();

                        if (pass != null && pass.length > 0) {
                            ret = String.valueOf(pass);

                            break;
                        }
                    }
                }
            } catch (Exception ex) {
                LOG.error("CredentialReader failed while decrypting provided string. Reason: " + ex.toString());

                ret = null;
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("getDecryptedString() : ret = " + ret);
        }

        return ret;
    }

    private SSLContext getSSLContext() {
        SSLContext     ret    = null;
        KeyManager[]   kmList = getKeyManagers();
        TrustManager[] tmList = getTrustManagers();

        if (tmList != null) {
            try {
                ret = SSLContext.getInstance(RANGER_SSL_CONTEXT_ALGO_TYPE);

                ret.init(kmList, tmList, new SecureRandom());
            } catch (NoSuchAlgorithmException e) {
                LOG.error("SSL algorithm is not available in the environment", e);
            } catch (KeyManagementException e) {
                LOG.error("Unable to initials the SSLContext", e);
            }
        }

        return ret;
    }

    private KeyManager[] getKeyManagers() {
        KeyManager[] ret                    = null;
        String       keyStoreFile           = getConfig("keystore.file");
        String       keyStoreAlias          = getConfig("keystore.alias");
        String       credentialProviderPath = getConfig("credential.provider.path");
        String       keyStoreFilepwd        = getCredential(credentialProviderPath, keyStoreAlias);

        if (StringUtils.isNotBlank(keyStoreFile) && StringUtils.isNotBlank(keyStoreFilepwd)) {
            InputStream in = null;

            try {
                in = getFileInputStream(keyStoreFile);

                if (in != null) {
                    KeyStore keyStore = KeyStore.getInstance(RANGER_KEYSTORE_FILE_TYPE_DEFAULT);

                    keyStore.load(in, keyStoreFilepwd.toCharArray());

                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(RANGER_SSL_KEYMANAGER_ALGO_TYPE);

                    keyManagerFactory.init(keyStore, keyStoreFilepwd.toCharArray());

                    ret = keyManagerFactory.getKeyManagers();
                } else {
                    LOG.error("Unable to obtain keystore from file [{}]", keyStoreFile);
                }
            } catch (KeyStoreException e) {
                LOG.error("Unable to obtain from KeyStore", e);
            } catch (NoSuchAlgorithmException e) {
                LOG.error("SSL algorithm is NOT available in the environment", e);
            } catch (CertificateException e) {
                LOG.error("Unable to obtain the requested certification", e);
            } catch (FileNotFoundException e) {
                LOG.error("Unable to find the necessary SSL Keystore Files", e);
            } catch (IOException e) {
                LOG.error("Unable to read the necessary SSL Keystore Files", e);
            } catch (UnrecoverableKeyException e) {
                LOG.error("Unable to recover the key from keystore", e);
            } finally {
                close(in, keyStoreFile);
            }
        }

        return ret;
    }

    private TrustManager[] getTrustManagers() {
        TrustManager[] ret                    = null;
        String         truststoreFile         = getConfig("truststore.file");
        String         truststoreAlias        = getConfig("truststore.alias");
        String         credentialProviderPath = getConfig("credential.provider.path");
        String         trustStoreFilepwd      = getCredential(credentialProviderPath, truststoreAlias);

        if (StringUtils.isNotBlank(truststoreFile) && StringUtils.isNotBlank(trustStoreFilepwd)) {
            InputStream in = null;

            try {
                in = getFileInputStream(truststoreFile);

                if (in != null) {
                    KeyStore trustStore = KeyStore.getInstance(RANGER_TRUSTSTORE_FILE_TYPE_DEFAULT);

                    trustStore.load(in, trustStoreFilepwd.toCharArray());

                    TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(RANGER_SSL_TRUSTMANAGER_ALGO_TYPE);

                    trustManagerFactory.init(trustStore);

                    ret = trustManagerFactory.getTrustManagers();
                } else {
                    LOG.error("Unable to obtain truststore from file [{}]", truststoreFile);
                }
            } catch (KeyStoreException e) {
                LOG.error("Unable to obtain from KeyStore", e);
            } catch (NoSuchAlgorithmException e) {
                LOG.error("SSL algorithm is NOT available in the environment", e);
            } catch (CertificateException e) {
                LOG.error("Unable to obtain the requested certification", e);
            } catch (FileNotFoundException e) {
                LOG.error("Unable to find the necessary SSL TrustStore File:{}", truststoreFile, e);
            } catch (IOException e) {
                LOG.error("Unable to read the necessary SSL TrustStore Files:{}", truststoreFile, e);
            } finally {
                close(in, truststoreFile);
            }
        }

        return ret;
    }

    private String getKeystoreFile() {
        String ret = getConfig("service.https.attrib.keystore.file");

        if (StringUtils.isBlank(ret)) {
            // new property not configured, lets use the old property
            ret = getConfig("https.attrib.keystore.file");
        }

        return ret;
    }

    private String getCredential(String url, String alias) {
        return RangerCredentialProvider.getInstance().getCredentialString(url, alias);
    }

    private String getConfig(String key) {
        String propertyWithPrefix = configPrefix + key;
        String value              = configuration.get(propertyWithPrefix);
        if (value == null) { // config-with-prefix not found; look at System properties
            value = System.getProperty(propertyWithPrefix);
            if (value == null) { // config-with-prefix not found in System properties; look for config-without-prefix
                value = configuration.get(key);
                if (value == null) { // config-without-prefix not found; look at System properties
                    value = System.getProperty(key);
                }
            }
        }

        return value;
    }

    private String getConfig(String key, String defaultValue) {
        String ret = getConfig(key);

        if (ret == null) {
            ret = defaultValue;
        }

        return ret;
    }

    private int getIntConfig(String key, int defaultValue) {
        int    ret      = defaultValue;
        String strValue = getConfig(key);

        try {
            if (strValue != null) {
                ret = Integer.parseInt(strValue);
            }
        } catch (Exception err) {
            LOG.warn(strValue + " can't be parsed to int. Reason: " + err.toString());
        }

        return ret;
    }

    protected long getLongConfig(String key, long defaultValue) {
        long   ret    = defaultValue;
        String retStr = getConfig(key);

        try {
            if (retStr != null) {
                ret = Long.parseLong(retStr);
            }
        } catch(Exception err){
            LOG.warn(retStr + " can't be parsed to long. Reason: " + err.toString());
        }

        return ret;
    }

    private boolean getBooleanConfig(String key, boolean defaultValue) {
        boolean ret      = defaultValue;
        String  strValue = getConfig(key);

        if (StringUtils.isNotBlank(strValue)) {
            ret = Boolean.valueOf(strValue);
        }

        return ret;
    }

    private InputStream getFileInputStream(String fileName) throws IOException {
        InputStream ret = null;

        if (StringUtils.isNotEmpty(fileName)) {
            File f = new File(fileName);

            if (f.exists()) {
                ret = new FileInputStream(f);
            } else {
                ret = ClassLoader.getSystemResourceAsStream(fileName);
            }
        }

        return ret;
    }

    private void close(InputStream str, String filename) {
        if (str != null) {
            try {
                str.close();
            } catch (IOException excp) {
                LOG.error("Error while closing file: [{}]", filename, excp);
            }
        }
    }
}
