/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.chainedplugin.hdfs.hive;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.authorization.hadoop.RangerHdfsAuthorizer;
import org.apache.ranger.authorization.utils.StringUtil;
import org.apache.ranger.plugin.contextenricher.RangerTagForEval;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.policyengine.*;
import org.apache.ranger.plugin.policyevaluator.RangerPolicyEvaluator;
import org.apache.ranger.plugin.service.RangerBasePlugin;
import org.apache.ranger.plugin.service.RangerChainedMapperPlugin;
import org.apache.ranger.plugin.util.RangerAccessRequestUtil;
import org.apache.ranger.plugin.util.RangerPerfTracer;
import org.apache.ranger.plugin.util.RangerRMSAccessResourceImpl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class RangerHdfsHiveChainedPlugin extends RangerChainedMapperPlugin {
    private static final Log LOG = LogFactory.getLog(RangerHdfsHiveChainedPlugin.class);

    private static final Log PERF_CHAINED_PLUGIN_POLICY_EVAL_LOG = RangerPerfTracer.getPerfLogger("chainedplugin.policyeval");

    private static final String PROP_HDFS_TO_HIVE_AUTHORIZE_ONLY_WITH_CHAINED_POLICIES = ".mapping.hive.authorize.with.only.chained.policies";

    private static final String PROP_HDFS_TO_HIVE_TABLE_ACCESSTYPE_MAPPING_READ = ".accesstype.mapping.read";
    private static final String PROP_HDFS_TO_HIVE_TABLE_ACCESSTYPE_MAPPING_WRITE = ".accesstype.mapping.write";
    private static final String PROP_HDFS_TO_HIVE_TABLE_ACCESSTYPE_MAPPING_EXECUTE = ".accesstype.mapping.execute";

    private static final String PROP_HDFS_TO_HIVE_DATABASE_ACCESSTYPE_MAPPING_READ = "db.accesstype.mapping.read";
    private static final String PROP_HDFS_TO_HIVE_DATABASE_ACCESSTYPE_MAPPING_WRITE = "db.accesstype.mapping.write";
    private static final String PROP_HDFS_TO_HIVE_DATABASE_ACCESSTYPE_MAPPING_EXECUTE = "db.accesstype.mapping.execute";

    private static final String CHAINED_SERVICE_TYPE     = "hive";
    private static final String HDFS_ACCESS_TYPE_READ    = "read";
    private static final String HDFS_ACCESS_TYPE_WRITE   = "write";
    private static final String HDFS_ACCESS_TYPE_EXECUTE = "execute";
    private static final String HIVE_ACCESS_TYPE_SELECT  = "select";
    private static final String HIVE_ACCESS_TYPE_UPDATE  = "update";
    private static final String HIVE_ACCESS_TYPE_ALTER   = "alter";
    private static final String HIVE_ACCESS_TYPE_CREATE  = "create";
    private static final String HIVE_ACCESS_TYPE_DROP    = "drop";
    private static final String ACCESS_TYPE_SEPARATOR    = ",";

    private static final String KEY_DATABASE             = "database";
    private static final String KEY_TABLE                = "table";

    private final Map<String, List<String>> accessTypeMapHdfsToHiveForTable = new HashMap<>();
    private final Map<String, List<String>> accessTypeMapHdfsToHiveForDatabase = new HashMap<>();
    private final boolean                   isAuthorizeOnlyWithChainedPolicies;
    private final List<String>              hiveServiceUserNames;
    private final Map<String, String>       accessTypeMapHiveToHdfsForTable = new HashMap<>();
    private final Map<String, String>       accessTypeMapHiveToHdfsForDatabase = new HashMap<>();
    private final List<String>              hivePrivilegedUserNames;

    public RangerHdfsHiveChainedPlugin(RangerBasePlugin rootPlugin, String serviceName) {
        super(rootPlugin, CHAINED_SERVICE_TYPE, serviceName);

        initAccessTypesMap(rootPlugin);
        isAuthorizeOnlyWithChainedPolicies = rootPlugin.getConfig().getBoolean(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_AUTHORIZE_ONLY_WITH_CHAINED_POLICIES, true);

        String hiveServiceUsersPropName = plugin.getConfig().getPropertyPrefix() + ".service.names";
        String hiveServiceUsers = plugin.getConfig().get(hiveServiceUsersPropName, "hive,impala");
        hiveServiceUserNames = StringUtils.isBlank(hiveServiceUsers) ? new ArrayList<>() : StringUtil.toList(hiveServiceUsers);

        String hivePrivilegedUsersPropName = plugin.getConfig().getPropertyPrefix() + ".privileged.user.names";
        String hivePrivilegedUsers = plugin.getConfig().get(hivePrivilegedUsersPropName, "admin,dpprofiler,hue,beacon,hive,impala");
        hivePrivilegedUserNames = StringUtils.isBlank(hivePrivilegedUsers) ? new ArrayList<>() : StringUtil.toList(hivePrivilegedUsers);

        LOG.info("RangerHdfsHiveChainedPlugin(" + serviceType + ", " + serviceName + ")");
    }

    @Override
    public void init() {
        LOG.info("==> RangerHdfsHiveChainedPlugin.init(" + serviceType + ", " + serviceName + ")");

        if (getIsInited()) {
            super.init();
            if (LOG.isDebugEnabled()) {
                LOG.debug("Configuration for RangerHdfsHiveChainedPlugin is :[" + plugin.getConfig() + "]");
            }
        } else {
            LOG.error("RangerHdfsHiveChainedPlugin for service [" + serviceName + "] was not initialized correctly. It will not be operational!");
        }

        LOG.info("<== RangerHdfsHiveChainedPlugin.init(" + serviceType + ", " + serviceName + ")");
    }

    @Override
    public RangerAccessResult isAccessAllowed(RangerAccessRequest hdfsRequest) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.isAccessAllowed(" + hdfsRequest + ")");
        }

        RangerAccessResult         ret             = null;

        if (getIsInited()) {

            RangerPerfTracer perf = null;

            if (RangerPerfTracer.isPerfTraceEnabled(PERF_CHAINED_PLUGIN_POLICY_EVAL_LOG)) {
                perf = RangerPerfTracer.getPerfTracer(PERF_CHAINED_PLUGIN_POLICY_EVAL_LOG, "RangerHdfsHiveChainedPlugin.isAccessAllowed(resource=" + hdfsRequest.getResource().getAsString() + ")");
            }

            List<RangerAccessResource> hiveResources = getHiveResources(hdfsRequest);

            List<RangerAccessResource> defaultDatabaseResource = extractDefaultDatabaseResource(hiveResources);

            if (CollectionUtils.isNotEmpty(defaultDatabaseResource) && CollectionUtils.isNotEmpty(hiveResources)) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Mapped hive resources contain both default-db-resource and other resources. Excluding default-db-resource for access evaluation");
                    LOG.debug("hiveResources:[" + hiveResources + "], defaultDatabaseResource:[" + defaultDatabaseResource + "]");
                }
            } else if (CollectionUtils.isEmpty(defaultDatabaseResource) && CollectionUtils.isEmpty(hiveResources)) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Mapped hive resources do not contain either default-db-resource or other resources");
                }
            } else if (CollectionUtils.isNotEmpty(hiveResources)) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Mapped hive resources do not contain default-db-resource, mapped hiveResources:[" + hiveResources + "] will be checked for access evaluation");
                }
            } else { // Only mapped resource must be default-db so oheck for access on it
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Mapped hive resource is only the default-db-resource:[" + defaultDatabaseResource + "]. Checking default-db-resource for access evaluation");
                }
                hiveResources = defaultDatabaseResource;
            }

            if (CollectionUtils.isNotEmpty(hiveResources)) {
                Set<RangerTagForEval> hdfsTags = RangerAccessRequestUtil.getRequestTagsFromContext(hdfsRequest.getContext());
                Set<RangerTagForEval> allTags = hdfsTags == null ? null : new HashSet<>(hdfsTags);
                boolean               isInError = false;

                for (RangerAccessResource hiveResource : hiveResources) {
                    RangerAccessResourceImpl hiveResourceImpl = (RangerAccessResourceImpl) hiveResource;
                    hiveResourceImpl.setServiceDef(plugin.getServiceDef());
                    Set<RangerTagForEval> tagsForResource      = null;

                    String  leafResourceName     = hiveResourceImpl.getLeafName();
                    boolean isRowFilterEvaluated = false;
                    boolean isDataMaskEvaluated  = false;
                    boolean isDatabaseResource   = false;

                    isInError = true;

                    if (StringUtils.equals(leafResourceName, KEY_DATABASE) || RangerHdfsAuthorizer.isGetAclOperation(hdfsRequest.getAction())) {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("leafResourceName:[" + leafResourceName + "], action:[" + hdfsRequest.getAction() + "]");
                            LOG.debug("Row-filtering and Data-masking policies will not be applied - either resource is not a table or operation is harmless!");
                        }
                        isRowFilterEvaluated = isDataMaskEvaluated = isDatabaseResource = true;
                    }

                    List<String> hiveAccessTypes = getHiveAccessTypesForResource(hdfsRequest, hiveResource, isDatabaseResource);
                    if (CollectionUtils.isEmpty(hiveAccessTypes)) {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("List of access-types to be checked for resource [" + hiveResource + "] is empty. Denying access!");
                        }
                        ret = null;
                        break;
                    }

                    if (hivePrivilegedUserNames.contains(hdfsRequest.getUser())) {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Requesting user[" + hdfsRequest.getUser() + "] is one of the privileged users:[" + hivePrivilegedUserNames + "] who are granted all access to Hive resources by Ranger default policies. Returning without evaluating Ranger policies.");
                        }
                        isInError = false;
                        continue;
                    }

                    for (String hiveAccessType : hiveAccessTypes) {
                        RangerAccessRequestImpl hiveAccessRequest = new RangerAccessRequestImpl(hiveResource, hiveAccessType, hdfsRequest.getUser(), hdfsRequest.getUserGroups(), hdfsRequest.getUserRoles());

                        hiveAccessRequest.setClientIPAddress(hdfsRequest.getClientIPAddress());
                        hiveAccessRequest.setClientType(hdfsRequest.getClientType());
                        hiveAccessRequest.setClusterName(hdfsRequest.getClusterName());
                        hiveAccessRequest.setClusterType(hdfsRequest.getClusterType());
                        hiveAccessRequest.setAction(hdfsRequest.getAction());
                        hiveAccessRequest.setAccessTime(hdfsRequest.getAccessTime());
                        hiveAccessRequest.setRemoteIPAddress(hdfsRequest.getRemoteIPAddress());
                        hiveAccessRequest.setRequestData(hdfsRequest.getRequestData());
                        hiveAccessRequest.setSessionId(hdfsRequest.getSessionId());
                        hiveAccessRequest.setContext(hdfsRequest.getContext());

                        if (isDatabaseResource) {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Mapped resource is a database, setting resource-match-scope to SELF");
                            }
                            hiveAccessRequest.setResourceMatchingScope(RangerAccessRequest.ResourceMatchingScope.SELF);
                        } else {
                            // set the scope to consider columns as well
                            hiveAccessRequest.setResourceMatchingScope(RangerAccessRequest.ResourceMatchingScope.SELF_OR_DESCENDANTS);
                        }

                        // check if row-filter is applied
                        if (!isRowFilterEvaluated) {
                            hiveAccessRequest.setAccessType(HIVE_ACCESS_TYPE_SELECT);
                            RangerAccessResult rowFilteringResult = plugin.evalRowFilterPolicies(hiveAccessRequest, null);

                            if (LOG.isDebugEnabled()) {
                                LOG.debug("hiveAccessRequest:[" + hiveAccessRequest + "], Row-filter-result:[" + rowFilteringResult + "]");
                            }

                            isRowFilterEvaluated = true;
                            tagsForResource = RangerAccessRequestUtil.getRequestTagsFromContext(hiveAccessRequest.getContext());

                            if (rowFilteringResult != null && rowFilteringResult.isRowFilterEnabled()) {
                                if (ret == null) {
                                    ret = new RangerAccessResult(RangerPolicy.POLICY_TYPE_ROWFILTER, plugin.getServiceName(), plugin.getServiceDef(), hiveAccessRequest);
                                }
                                ret.setIsAllowed(false);
                                ret.setPolicyId(rowFilteringResult.getPolicyId());
                                ret.setPolicyVersion(rowFilteringResult.getPolicyVersion());
                                ret.setPolicyPriority(rowFilteringResult.getPolicyPriority());
                                ret.setZoneName(rowFilteringResult.getZoneName());

                                break; // no need to evaluate further policies
                            }
                        }

                        // check if column-masking is applied
                        if (!isDataMaskEvaluated) {
                            hiveAccessRequest.setAccessType(HIVE_ACCESS_TYPE_SELECT);
                            RangerAccessResult maskingResult = plugin.evalDataMaskPolicies(hiveAccessRequest, null);

                            if (LOG.isDebugEnabled()) {
                                LOG.debug("hiveAccessRequest:[" + hiveAccessRequest + "], Data-masking-result:[" + maskingResult + "]");
                            }

                            isDataMaskEvaluated = true;
                            tagsForResource = RangerAccessRequestUtil.getRequestTagsFromContext(hiveAccessRequest.getContext());

                            if (maskingResult != null && StringUtils.isNotEmpty(maskingResult.getMaskType())) {
                                if (ret == null) {
                                    ret = new RangerAccessResult(RangerPolicy.POLICY_TYPE_DATAMASK, plugin.getServiceName(), plugin.getServiceDef(), hiveAccessRequest);
                                }
                                ret.setIsAllowed(false);
                                ret.setPolicyId(maskingResult.getPolicyId());
                                ret.setPolicyVersion(maskingResult.getPolicyVersion());
                                ret.setPolicyPriority(maskingResult.getPolicyPriority());
                                ret.setZoneName(maskingResult.getZoneName());

                                break; // no need to evaluate further policies
                            }
                        }

                        hiveAccessRequest.setAccessType(hiveAccessType);

                        RangerAccessResult hiveAccessResult = plugin.isAccessAllowed(hiveAccessRequest);

                        if (LOG.isDebugEnabled()) {
                            LOG.debug("hiveAccessRequest:[" + hiveAccessRequest + "], Access-result:[" + hiveAccessResult + "]");
                        }

                        tagsForResource = RangerAccessRequestUtil.getRequestTagsFromContext(hiveAccessRequest.getContext());

                        if (hiveAccessResult == null || !hiveAccessResult.getIsAccessDetermined() || (!hiveAccessResult.getIsAllowed() && hiveAccessResult.getPolicyId() == -1L)) { // RANGER-3294 sets isAccessDetermined to true in certain cases.
                            continue;
                        }

                        if (isDatabaseResource) {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Mapped resource is a database, No need to check for resource-matching-scope of SELF again!");
                            }
                        } else {
                            if (hiveAccessResult.getIsAllowed()) {
                                // CDPD-17227 - This problem is essentially not solvable as RMS mappings do not keep track of the columns in the table.
                                // An approximation is as follows:
                                // Check if there is a policy allowing access for all columns. This means the policy either has '*' in column resource-spec or is a table level policy.
                                // This is still not completely correct, as there may be several Hive policies which together grant access to all columns in the table. In such case, the access will be denied.
                                // Easiest but most inefficient way to do this is to check again but with scope set to SELF

                                // Set the resource-matching scope to SELF
                                RangerAccessRequest.ResourceMatchingScope oldScope = hiveAccessRequest.getResourceMatchingScope();
                                hiveAccessRequest.setResourceMatchingScope(RangerAccessRequest.ResourceMatchingScope.SELF);

                                hiveAccessResult = plugin.isAccessAllowed(hiveAccessRequest);

                                if (LOG.isDebugEnabled()) {
                                    LOG.debug("hiveAccessRequest:[" + hiveAccessRequest + "], Access-result:[" + hiveAccessResult + "]");
                                }

                                // Reset the resource-matching scope
                                hiveAccessRequest.setResourceMatchingScope(oldScope);

                                if (hiveAccessResult == null || !hiveAccessResult.getIsAccessDetermined() || (!hiveAccessResult.getIsAllowed() && hiveAccessResult.getPolicyId() == -1L)) { // RANGER-3294 sets isAccessDetermined to true in certain cases.
                                    continue;
                                }
                            }
                        }

                        boolean overrideResult = ret == null || hiveAccessResult.getPolicyPriority() > ret.getPolicyPriority();

                        if (!overrideResult) {
                            // override if current result is from the same policy priority, and if denies access
                            if (hiveAccessResult.getPolicyPriority() == ret.getPolicyPriority() && !hiveAccessResult.getIsAllowed()) {
                                overrideResult = true;
                            }
                        }

                        if (overrideResult) {
                            ret = hiveAccessResult;
                        } else if (ret != null) {
                            ret.setZoneName(hiveAccessResult.getZoneName());
                        }
                    }

                    if (CollectionUtils.isNotEmpty(tagsForResource)) {
                        if (CollectionUtils.isEmpty(allTags)) {
                            allTags = tagsForResource;
                        } else {
                            allTags.addAll(tagsForResource);
                        }
                    }

                }

                if (isAuthorizeOnlyWithChainedPolicies && ret == null && isInError) {
                    ret = new RangerAccessResult(RangerPolicy.POLICY_TYPE_ACCESS, plugin.getServiceName(), plugin.getServiceDef(), hdfsRequest);
                    ret.setIsAllowed(false);
                    ret.setPolicyPriority(RangerPolicy.POLICY_PRIORITY_OVERRIDE);
                }

                RangerAccessRequestUtil.setRequestTagsInContext(hdfsRequest.getContext(), allTags);
            }

            RangerPerfTracer.logAlways(perf);

        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("isAccessAllowed() is a not operational as the chained plugin was not correctly initialized");
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.isAccessAllowed(" + hdfsRequest + "): ret=" + ret);
        }

        return ret;
    }

    @Override
    public Collection<RangerAccessResult> isAccessAllowed(Collection<RangerAccessRequest> hdfsRequests) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.isAccessAllowed(" + hdfsRequests + ")");
        }

        Collection<RangerAccessResult> ret = new ArrayList<>();

        for (RangerAccessRequest hdfsRequest : hdfsRequests) {
            ret.add(isAccessAllowed(hdfsRequest));
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.isAccessAllowed(" + hdfsRequests + "): ret=" + ret);
        }

        return ret;
    }

    @Override
    public RangerResourceACLs getResourceACLs(RangerAccessRequest request) {
        return getResourceACLs(request, RangerPolicy.POLICY_TYPE_ACCESS);
    }

    @Override
    public RangerResourceACLs getResourceACLs(RangerAccessRequest request, Integer policyType) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.getResourceACLs(" + request + ", policyType=" + policyType + ")");
        }
        RangerResourceACLs ret = null;

        List<RangerAccessResource> hiveResources = getHiveResources(request);
        if (CollectionUtils.isEmpty(hiveResources)) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("No Mapped Hive-Resources for HDFS resource:[" + request.getResource().getAsString() + "]");
            }
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug(hiveResources.size() + " Mapped Hive-Resources for HDFS resource:[" + request.getResource().getAsString() + "]");
            }
            ret = new RangerResourceACLs();

            for (RangerAccessResource hiveResource : hiveResources) {

                RangerAccessRequestImpl hiveRequest = new RangerAccessRequestImpl(hiveResource, RangerPolicyEngine.ANY_ACCESS, null, null, null);

                RangerResourceACLs accessACLs = plugin.getResourceACLs(hiveRequest, RangerPolicy.POLICY_TYPE_ACCESS);

                if (LOG.isDebugEnabled()) {
                    LOG.debug("Hive ACLs for resource:[" + hiveResource.getAsString() + "] are [" + accessACLs + "]");
                }
                // Get additional acls for row-filtering and column-masking if HiveResource is a table
                final RangerResourceACLs resourceACLs;
                final String             leafResourceName = hiveResource.getLeafName();

                if (StringUtils.equals(leafResourceName, KEY_TABLE)) {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Hive-Resource is a table - Row-filtering and Data-masking policies will be applied.");
                    }
                    hiveRequest.setResourceMatchingScope(RangerAccessRequest.ResourceMatchingScope.SELF_OR_DESCENDANTS);
                    hiveRequest.setAccessType(HDFS_ACCESS_TYPE_READ);

                    RangerResourceACLs rowFilterACLs = plugin.getResourceACLs(hiveRequest, RangerPolicy.POLICY_TYPE_ROWFILTER);
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Hive row-masking ACLs for resource:[" + hiveResource.getAsString() + "] are [" + rowFilterACLs + "]");
                    }

                    RangerResourceACLs dataMaskingACLs = plugin.getResourceACLs(hiveRequest, RangerPolicy.POLICY_TYPE_DATAMASK);
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Hive row-masking ACLs for resource:[" + hiveResource.getAsString() + "] are [" + dataMaskingACLs + "]");
                    }
                    // Map Hive ResourceACLs to HDFS ResourceACLs
                    boolean isExternalTable = true;
                    if (hiveResource instanceof RangerRMSAccessResourceImpl) {
                        Map<String, String> additionalInfo = ((RangerRMSAccessResourceImpl) hiveResource).getAdditionalInfo();
                        if (MapUtils.isNotEmpty(additionalInfo)) {
                            String tableType = additionalInfo.get("tableType");
                            isExternalTable = StringUtils.equals(tableType, "EXTERNAL_TABLE");
                        }
                    }
                    resourceACLs = mapHiveACLsToHdfsACLs(accessACLs, rowFilterACLs, dataMaskingACLs, false, isExternalTable);
                } else {
                    RangerResourceACLs emptyACLs = new RangerResourceACLs();
                    resourceACLs = mapHiveACLsToHdfsACLs(accessACLs, emptyACLs, emptyACLs, true, false);
                }

                RangerBasePlugin.getMergedResourceACLs(ret, resourceACLs);
            }
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.getResourceACLs(policyType=" + policyType + ") : ret :[" + ret + "]");
        }
        return ret;
    }

    @Override
    public boolean isAuthorizeOnlyWithChainedPlugin() { return isAuthorizeOnlyWithChainedPolicies; }

    private List<RangerAccessResource> getHiveResources(RangerAccessRequest hdfsRequest) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.getHiveResources(" + hdfsRequest + ")");
        }

        List<RangerAccessResource> ret = null;

        if (isNotWhitelisted(hdfsRequest)) {
            if (hdfsRequest.getResource() != null) {

                ret = mappingProvider.getMappedResources(hdfsRequest);

            } else {
                LOG.warn("Unexpected null resource found in the request, ignoring");
            }
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("user/groups are whitelisted :[" + hdfsRequest + "]");
            }
        }

        /*
        if (!isHiveResourcesValid(ret)) {
            LOG.warn("Hive Resources are not valid. resources:[" + ret + "] returning null!!");
            ret = null;
        }
         */

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.getHiveResources(" + hdfsRequest + "): ret=" + ret);
        }

        return ret;
    }

    private List<RangerAccessResource> extractDefaultDatabaseResource(List<RangerAccessResource> resources) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.extractDefaultDatabaseResource(" + resources + ")");
        }
        List<RangerAccessResource> ret = null;

        if (CollectionUtils.isNotEmpty(resources)) {
            if (resources.size() > 1) {
                Iterator<RangerAccessResource> iterator = resources.iterator();

                while (iterator.hasNext()) {
                    RangerAccessResource resource = iterator.next();

                    RangerAccessResourceImpl hiveResourceImpl = (RangerAccessResourceImpl) resource;
                    hiveResourceImpl.setServiceDef(plugin.getServiceDef());
                    String                leafResourceName     = hiveResourceImpl.getLeafName();
                    if (StringUtils.equals(leafResourceName, KEY_DATABASE)) {
                        Object values = hiveResourceImpl.getValue(KEY_DATABASE);

                        List<String> stringValues = new ArrayList<>();
                        if (stringValues.getClass().isAssignableFrom(values.getClass())) {
                            stringValues = (List<String>) values;
                            if (stringValues.contains("default")) {
                                // add to ret  and possibly take this out of the resources list
                                if (ret == null) {
                                    ret = new ArrayList<>();
                                }
                                ret.add(resource);
                                iterator.remove();
                            }
                        } else {
                            LOG.warn("Could not cast values to List of Strings, returning null!!");
                            return null;
                        }
                    }
                }
            }
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.extractDefaultDatabaseResource(" + resources + ") : " + (CollectionUtils.isEmpty(ret) ? "No default db/table resources" : "[" + ret.size() + " defauult db/table resources found" ));
        }
        return ret;
    }

    private List<String> getHiveAccessTypes(String hdfsAccessType, RangerAccessResource hiveResource, boolean isDatabase) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.getHiveAccessTypes(" + hdfsAccessType + ", " + hiveResource + ", " + isDatabase + ")");
        }

        List<String> ret = isDatabase? accessTypeMapHdfsToHiveForDatabase.get(hdfsAccessType) : accessTypeMapHdfsToHiveForTable.get(hdfsAccessType);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.getHiveAccessTypes(" + hdfsAccessType + ", " + hiveResource + ", " + isDatabase + ") : ret:[" + ret + "]");
        }

        return ret;
    }

    private List<String> getHiveAccessTypesForResource(RangerAccessRequest hdfsRequest, RangerAccessResource hiveResource, boolean isDatabase) {
        List<String> ret = getHiveAccessTypes(hdfsRequest.getAccessType(), hiveResource, isDatabase);

        String user = hdfsRequest.getUser();

        if (!hiveServiceUserNames.contains(user)) {
            if (hiveResource instanceof RangerRMSAccessResourceImpl) {
                Map<String, String> additionalInfo = ((RangerRMSAccessResourceImpl) hiveResource).getAdditionalInfo();
                if (MapUtils.isNotEmpty(additionalInfo)) {
                    String tableType = additionalInfo.get("tableType");
                    if (!isDatabase && !StringUtils.equals(tableType, "EXTERNAL_TABLE")) {
                        // Deny write access
                        if (hdfsRequest.getAccessType().equals(HDFS_ACCESS_TYPE_WRITE)) {
                            LOG.error("User [" + user + "] is allowed only [" + HDFS_ACCESS_TYPE_READ + " or " + HDFS_ACCESS_TYPE_EXECUTE + "] access to " + hdfsRequest.getResource().getAsString() + " which maps to table of type [" + tableType + "]. Access is not allowed.]");
                            ret = Collections.emptyList();
                        }
                    }
                }
            }
        }
        return ret;
    }

    private void initAccessTypesMap(RangerBasePlugin rootPlugin) {

        //String hdfsReadToHiveAccessTypes    = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_READ, HIVE_ACCESS_TYPE_SELECT + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_REPLADMIN);
        String hdfsReadToHiveAccessTypes    = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_TABLE_ACCESSTYPE_MAPPING_READ, HIVE_ACCESS_TYPE_SELECT);
        //String hdfsWriteToHiveAccessTypes   = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_ACCESSTYPE_MAPPING_WRITE, HIVE_ACCESS_TYPE_UPDATE + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_ALTER + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_DROP + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_REPLADMIN);
        String hdfsWriteToHiveAccessTypes   = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_TABLE_ACCESSTYPE_MAPPING_WRITE, HIVE_ACCESS_TYPE_UPDATE + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_ALTER);
        String hdfsExecuteToHiveAccessTypes = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_TABLE_ACCESSTYPE_MAPPING_EXECUTE, RangerPolicyEngine.ANY_ACCESS);

        accessTypeMapHdfsToHiveForTable.put(HDFS_ACCESS_TYPE_READ,         Arrays.asList(hdfsReadToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        accessTypeMapHdfsToHiveForTable.put(HDFS_ACCESS_TYPE_WRITE,        Arrays.asList(hdfsWriteToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        accessTypeMapHdfsToHiveForTable.put(HDFS_ACCESS_TYPE_EXECUTE,      Arrays.asList(hdfsExecuteToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        accessTypeMapHdfsToHiveForTable.put(RangerPolicyEngine.ANY_ACCESS, Collections.singletonList(RangerPolicyEngine.ANY_ACCESS));

        String[] hiveAccessesForRead = hdfsReadToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR);
        for (String hiveAccess : hiveAccessesForRead) {
            accessTypeMapHiveToHdfsForTable.put(hiveAccess, HDFS_ACCESS_TYPE_READ);
        }
        String[] hiveAccessesForWrite = hdfsWriteToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR);
        for (String hiveAccess : hiveAccessesForWrite) {
            accessTypeMapHiveToHdfsForTable.put(hiveAccess, HDFS_ACCESS_TYPE_WRITE);
        }

        hdfsReadToHiveAccessTypes    = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_DATABASE_ACCESSTYPE_MAPPING_READ, RangerPolicyEngine.ANY_ACCESS);
        hdfsWriteToHiveAccessTypes   = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_DATABASE_ACCESSTYPE_MAPPING_WRITE, HIVE_ACCESS_TYPE_CREATE + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_DROP + ACCESS_TYPE_SEPARATOR + HIVE_ACCESS_TYPE_ALTER);
        hdfsExecuteToHiveAccessTypes = rootPlugin.getConfig().get(rootPlugin.getConfig().getPropertyPrefix() + PROP_HDFS_TO_HIVE_DATABASE_ACCESSTYPE_MAPPING_EXECUTE, RangerPolicyEngine.ANY_ACCESS);

        accessTypeMapHdfsToHiveForDatabase.put(HDFS_ACCESS_TYPE_READ,         Arrays.asList(hdfsReadToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        accessTypeMapHdfsToHiveForDatabase.put(HDFS_ACCESS_TYPE_WRITE,        Arrays.asList(hdfsWriteToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        accessTypeMapHdfsToHiveForDatabase.put(HDFS_ACCESS_TYPE_EXECUTE,      Arrays.asList(hdfsExecuteToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR)));
        accessTypeMapHdfsToHiveForDatabase.put(RangerPolicyEngine.ANY_ACCESS, Collections.singletonList(RangerPolicyEngine.ANY_ACCESS));

        hiveAccessesForRead = hdfsReadToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR);
        for (String hiveAccess : hiveAccessesForRead) {
            accessTypeMapHiveToHdfsForDatabase.put(hiveAccess, HDFS_ACCESS_TYPE_READ);
        }
        hiveAccessesForWrite = hdfsWriteToHiveAccessTypes.split(ACCESS_TYPE_SEPARATOR);
        for (String hiveAccess : hiveAccessesForWrite) {
            accessTypeMapHiveToHdfsForDatabase.put(hiveAccess, HDFS_ACCESS_TYPE_WRITE);
        }
    }

    private RangerResourceACLs mapHiveACLsToHdfsACLs(RangerResourceACLs hiveACLs, RangerResourceACLs rowFilterACLs, RangerResourceACLs datamaskACLs, final boolean isDatabase, final boolean isExternalTable) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerHdfsHiveChainedPlugin.mapHiveACLsToHdfsAcls(hiveACLs [" + hiveACLs + "] )");
        }
        RangerResourceACLs ret = new RangerResourceACLs();

        for (Map.Entry<String, Map<String, RangerResourceACLs.AccessResult>> entry : hiveACLs.getUserACLs().entrySet()) {
            boolean      isAnyAccessAllowed = false;
            RangerPolicy policyForAnyAccess = null;
            String       user               = entry.getKey();

            for (Map.Entry<String, RangerResourceACLs.AccessResult> permission : entry.getValue().entrySet()) {
                if (permission.getValue().getResult() == RangerPolicyEvaluator.ACCESS_ALLOWED || permission.getValue().getResult() == RangerPolicyEvaluator.ACCESS_DENIED) {
                    String hiveAccessType = permission.getKey();
                    String hdfsAccessType = isDatabase? accessTypeMapHiveToHdfsForDatabase.get(hiveAccessType) : accessTypeMapHiveToHdfsForTable.get(hiveAccessType);
                    int    accessResult   = permission.getValue().getResult();
                    RangerPolicy policy   = permission.getValue().getPolicy();

                    if (hdfsAccessType != null && hdfsAccessType.equals(HDFS_ACCESS_TYPE_WRITE) && !isDatabase && !isExternalTable && !hiveServiceUserNames.contains(user)) {
                        RangerPolicy newPolicy = new RangerPolicy();
                        newPolicy.updateFrom(policy);

                        newPolicy.setId(policy.getId());
                        newPolicy.setGuid(policy.getGuid());
                        newPolicy.setVersion(policy.getVersion());
                        newPolicy.setServiceType(policy.getServiceType());

                        newPolicy.setPolicyPriority(Integer.MAX_VALUE);

                        policy       = newPolicy;
                        accessResult = RangerPolicyEvaluator.ACCESS_DENIED;
                    }
                    if (!isAnyAccessAllowed && accessResult == RangerPolicyEvaluator.ACCESS_ALLOWED) {
                        isAnyAccessAllowed = true;
                        policyForAnyAccess = policy;
                    }
                    if (hdfsAccessType != null) {
                        ret.setUserAccessInfo(user, hdfsAccessType, accessResult, policy);
                    }
                }
            }
            // TBD - If some access is allowed and some other access (which maps to _any)  is denied, we should not deny execute access.
            if (isAnyAccessAllowed) {
                ret.setUserAccessInfo(user, HDFS_ACCESS_TYPE_EXECUTE, RangerPolicyEvaluator.ACCESS_ALLOWED, policyForAnyAccess);
                if (isDatabase) {
                    ret.setUserAccessInfo(user, HDFS_ACCESS_TYPE_READ, RangerPolicyEvaluator.ACCESS_ALLOWED, policyForAnyAccess);
                }
            }
        }

        for (Map.Entry<String, Map<String, RangerResourceACLs.AccessResult>> entry : hiveACLs.getGroupACLs().entrySet()) {
            boolean      isAnyAccessAllowed = false;
            RangerPolicy policyForAnyAccess = null;
            String       group              = entry.getKey();

            for (Map.Entry<String, RangerResourceACLs.AccessResult> permission : entry.getValue().entrySet()) {
                if (permission.getValue().getResult() == RangerPolicyEvaluator.ACCESS_ALLOWED || permission.getValue().getResult() == RangerPolicyEvaluator.ACCESS_DENIED) {
                    String hiveAccessType = permission.getKey();
                    String hdfsAccessType = isDatabase ? accessTypeMapHiveToHdfsForDatabase.get(hiveAccessType) : accessTypeMapHiveToHdfsForTable.get(hiveAccessType);
                    if (!isAnyAccessAllowed && permission.getValue().getResult() == RangerPolicyEvaluator.ACCESS_ALLOWED) {
                        isAnyAccessAllowed = true;
                        policyForAnyAccess = permission.getValue().getPolicy();
                    }
                    if (hdfsAccessType != null) {
                        ret.setGroupAccessInfo(group, hdfsAccessType, permission.getValue().getResult(), permission.getValue().getPolicy());
                    }
                }
                // TBD - If some access is allowed and some other access (which maps to _any)  is denied, we should not deny execute access.
                if (isAnyAccessAllowed) {
                    ret.setGroupAccessInfo(group, HDFS_ACCESS_TYPE_EXECUTE, RangerPolicyEvaluator.ACCESS_ALLOWED, policyForAnyAccess);
                    if (isDatabase) {
                        ret.setGroupAccessInfo(group, HDFS_ACCESS_TYPE_READ, RangerPolicyEvaluator.ACCESS_ALLOWED, policyForAnyAccess);
                    }
                }
            }
        }

        overrideReadAccess(ret.getUserACLs(), getRowFilteredUsersOrGroups(rowFilterACLs.getRowFilters(), true));
        overrideReadAccess(ret.getUserACLs(), getRowFilteredUsersOrGroups(rowFilterACLs.getRowFilters(), false));

        overrideReadAccess(ret.getUserACLs(), getMaskedUsersOrGroups(datamaskACLs.getDataMasks(), true));
        overrideReadAccess(ret.getUserACLs(), getMaskedUsersOrGroups(datamaskACLs.getDataMasks(), false));

        ret.finalizeAcls();

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerHdfsHiveChainedPlugin.mapHiveACLsToHdfsAcls() : ret :[" + ret + "]");
        }

        return ret;
    }

    private List<String> getMaskedUsersOrGroups(final List<RangerResourceACLs.DataMaskResult> dataMaskResults, final boolean isUser) {
        List<String> ret = null;
        for (RangerResourceACLs.DataMaskResult result : dataMaskResults) {
            if (!StringUtils.equals(result.getMaskInfo().getDataMaskType(), RangerPolicy.MASK_TYPE_NONE)) {
                if (ret == null) {
                    ret = new ArrayList<>();
                }
                ret.addAll(isUser ? result.getUsers() : result.getGroups());
            }
        }
        return ret;
    }

    private List<String> getRowFilteredUsersOrGroups(final List<RangerResourceACLs.RowFilterResult> rowFilterResults, final boolean isUser) {
        List<String> ret = null;
        for (RangerResourceACLs.RowFilterResult result : rowFilterResults) {
            if (StringUtils.isNotEmpty(result.getFilterInfo().getFilterExpr())) {
                if (ret == null) {
                    ret = new ArrayList<>();
                }
                ret.addAll(isUser ? result.getUsers() : result.getGroups());
            }
        }
        return ret;
    }

    private void overrideReadAccess(Map<String, Map<String, RangerResourceACLs.AccessResult>> oldACLs, List<String> usersOrGroups) {
        if (CollectionUtils.isNotEmpty(usersOrGroups)) {
            for (String userOrGroup : usersOrGroups) {
                Map<String, RangerResourceACLs.AccessResult> oldAccessResults = oldACLs.get(userOrGroup);

                if (oldAccessResults == null) {
                    oldAccessResults = new HashMap<>();
                    RangerResourceACLs.AccessResult newAccessResult = new RangerResourceACLs.AccessResult();
                    newAccessResult.setResult(RangerPolicyEvaluator.ACCESS_DENIED);
                    newAccessResult.setPolicy(null);
                    oldAccessResults.put(HDFS_ACCESS_TYPE_READ, newAccessResult);
                    oldACLs.put(userOrGroup, oldAccessResults);
                } else {
                    RangerResourceACLs.AccessResult oldAccessResult = oldAccessResults.get(HDFS_ACCESS_TYPE_READ);
                    if (oldAccessResult == null) {
                        RangerResourceACLs.AccessResult newAccessResult = new RangerResourceACLs.AccessResult();
                        newAccessResult.setResult(RangerPolicyEvaluator.ACCESS_DENIED);
                        newAccessResult.setPolicy(null);
                        oldAccessResults.put(HDFS_ACCESS_TYPE_READ, newAccessResult);
                    } else {
                        if (oldAccessResult.getResult() != RangerPolicyEvaluator.ACCESS_DENIED) {
                            oldAccessResult.setResult(RangerPolicyEvaluator.ACCESS_DENIED);
                            oldAccessResult.setPolicy(null);
                        }
                    }
                }
            }
        }
    }

}
