/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.plugin.util;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.admin.client.RangerAdminClient;
import org.apache.ranger.plugin.model.RangerRole;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.List;

public class RangerFileBasedPoliciesRetriever implements RangerAdminClient {
    private static final Log LOG = LogFactory.getLog(RangerFileBasedPoliciesRetriever.class);

    private URL    servicePoliciesFileURL;
    private String servicePoliciesFileName;
    private Gson   gsonBuilder;

    @Override
    public void init(String serviceName, String appId, String configPropertyPrefix, Configuration config) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> init()" );
        }

        gsonBuilder = new GsonBuilder().setDateFormat("yyyyMMdd-HH:mm:ss.SSS-Z")
                .setPrettyPrinting()
                .create();

        String servicePoliciesFileNameProperty = configPropertyPrefix + ".policy.source.file.name";
        String servicePoliciesDefaultFileName = "/testdata/test_servicepolicies.json";

        if (StringUtils.isNotBlank(serviceName) && StringUtils.isNotBlank(appId)) {
            InputStream servicePoliciesFileStream = null;

            // Open specified file from options- it should contain service-policies

            servicePoliciesFileName = config != null? config.get(servicePoliciesFileNameProperty) : null;

            servicePoliciesFileName = servicePoliciesFileName == null ? servicePoliciesDefaultFileName : servicePoliciesFileName;

            File f = new File(servicePoliciesFileName);

            if (f.exists() && f.isFile() && f.canRead()) {
                try {
                    servicePoliciesFileStream = new FileInputStream(f);
                    servicePoliciesFileURL = f.toURI().toURL();
                } catch (FileNotFoundException exception) {
                    LOG.error("Error processing input file:" + servicePoliciesFileName + " or no privilege for reading file " + servicePoliciesFileName, exception);
                } catch (MalformedURLException malformedException) {
                    LOG.error("Error processing input file:" + servicePoliciesFileName + " cannot be converted to URL " + servicePoliciesFileName, malformedException);
                }
            } else {
                URL fileURL = getClass().getResource(servicePoliciesFileName);
                if (fileURL == null && !servicePoliciesFileName.startsWith("/")) {
                    fileURL = getClass().getResource("/" + servicePoliciesFileName);
                }

                if (fileURL == null) {
                    fileURL = ClassLoader.getSystemClassLoader().getResource(servicePoliciesFileName);
                    if (fileURL == null && !servicePoliciesFileName.startsWith("/")) {
                        fileURL = ClassLoader.getSystemClassLoader().getResource("/" + servicePoliciesFileName);
                    }
                }

                if (fileURL != null) {
                    try {
                        servicePoliciesFileStream = fileURL.openStream();
                        servicePoliciesFileURL = fileURL;
                    } catch (Exception exception) {
                        LOG.error(servicePoliciesFileName + " is not a file", exception);
                    }
                } else {
                    LOG.warn("Error processing input file: URL not found for " + servicePoliciesFileName + " or no privilege for reading file " + servicePoliciesFileName);
                }
            }

            if (servicePoliciesFileStream != null) {
                try {
                    servicePoliciesFileStream.close();
                } catch (Exception e) {
                    // Ignore
                }
            }

        } else {
            LOG.error("FATAL: Cannot find service/servicePoliciesFile to use for retrieving policies. Will NOT be able to retrieve policies.");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== init() : servicePoliciesFileName=" + servicePoliciesFileName);
        }
    }

    @Override
    public ServicePolicies getServicePoliciesIfUpdated(long lastKnownVersion, long lastActivationTimeInMillis) throws Exception {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> getServicePoliciesIfUpdated(lastKnownVersion=" + lastKnownVersion + ", lastActivationTimeInMillis=" + lastActivationTimeInMillis + ")");
        }

        ServicePolicies servicePolicies;

        if (servicePoliciesFileURL != null) {
            try (
                    InputStream servicePoliciesFileStream = servicePoliciesFileURL.openStream();
                    Reader reader = new InputStreamReader(servicePoliciesFileStream, StandardCharsets.UTF_8)
            ) {

                servicePolicies = gsonBuilder.fromJson(reader, ServicePolicies.class);

                if (servicePolicies.getPolicyVersion() <= lastKnownVersion) {
                    // No change in servicePolicies
                    servicePolicies = null;
                }
            } catch (IOException e) {
                LOG.warn("Error processing input file: or no privilege for reading file " + servicePoliciesFileName);
                throw e;
            }
        } else {
            LOG.error("Error reading file: " + servicePoliciesFileName);
            throw new Exception("servicePoliciesFileURL is null!");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== getServicePoliciesIfUpdated(lastKnownVersion=" + lastKnownVersion + ", lastActivationTimeInMillis=" + lastActivationTimeInMillis + ")");
        }

        return servicePolicies;
    }

    @Override
    public RangerRoles getRolesIfUpdated(long lastKnownRoleVersion, long lastActivationTimeInMills) throws Exception {
        return null;
    }

    @Override
    public RangerRole createRole(RangerRole request) {
        return null;
    }

    @Override
    public void dropRole(String execUser, String roleName) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.dropRole() is not supported");
    }

    @Override
    public List<String> getAllRoles(String execUser) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.getAllRoles() is not supported");
    }

    @Override
    public List<String> getUserRoles(String execUser) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.getUserRoles() is not supported");
    }

    @Override
    public RangerRole getRole(String execUser, String roleName) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.getRole() is not supported");
    }

    @Override
    public void grantRole(GrantRevokeRoleRequest request) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.grantRole() is not supported");
    }

    @Override
    public void revokeRole(GrantRevokeRoleRequest request) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.revokeRole() is not supported");
    }

    @Override
    public void grantAccess(GrantRevokeRequest request) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.grantAccess() is not supported");
    }

    @Override
    public void revokeAccess(GrantRevokeRequest request) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.revokeAccess() is not supported");
    }

    @Override
    public ServiceTags getServiceTagsIfUpdated(long lastKnownVersion, long lastActivationTimeInMillis) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.getServiceTagsIfUpdated() is not supported");
    }

    @Override
    public List<String> getTagTypes(String tagTypePattern) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.getTagTypes() is not supported");
    }

    @Override
    public RangerUserStore getUserStoreIfUpdated(long lastKnownUserStoreVersion, long lastActivationTimeInMillis) throws Exception {
        throw new Exception("RangerFileBasedPoliciesRetriever.getUserStoreIfUpdated() is not supported");
    }

}
