/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.plugin.util;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.ranger.authorization.hadoop.config.RangerPluginConfig;
import org.apache.ranger.plugin.service.RangerChainedMapperPlugin;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.StandardCharsets;

public class RangerMappingFileRetriever extends RangerMappingRetriever {
    private static final Log LOG = LogFactory.getLog(RangerMappingFileRetriever.class);

    String       mappingFileName;
    private URL  mappingFileURL;
    private Gson gsonBuilder;

    @Override
    public boolean init(String sourceFileName, String targetServiceName, RangerPluginConfig configuration) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingFileRetriever.init()");
        }
        super.init(sourceServiceName, targetServiceName, configuration);

        gsonBuilder = new GsonBuilder().setDateFormat("yyyyMMdd-HH:mm:ss.SSS-Z").setPrettyPrinting().create();

        mappingFileName = configuration.get(configuration.getPropertyPrefix() + RangerChainedMapperPlugin.ACL_MAPPING_SOURCE_FILE_PARAMETER);

        if (LOG.isDebugEnabled()) {
            LOG.debug("mappingFileName=[" + mappingFileName + "]");
        }

        if (StringUtils.isNotBlank(mappingFileName)) {
            InputStream mappingFileStream = null;

            File f = new File(mappingFileName);

            if (f.exists() && f.isFile() && f.canRead()) {
                try {
                    mappingFileStream = new FileInputStream(f);
                    mappingFileURL = f.toURI().toURL();
                } catch (FileNotFoundException exception) {
                    LOG.error("Error processing input file:" + mappingFileName + " or no privilege for reading file " + mappingFileName, exception);
                } catch (MalformedURLException malformedException) {
                    LOG.error("Error processing input file:" + mappingFileName + " cannot be converted to URL " + mappingFileName, malformedException);
                }
            } else {

                URL fileURL = getClass().getResource(mappingFileName);
                if (fileURL == null) {
                    if (!mappingFileName.startsWith("/")) {
                        fileURL = getClass().getResource("/" + mappingFileName);
                    }
                }

                if (fileURL == null) {
                    fileURL = ClassLoader.getSystemClassLoader().getResource(mappingFileName);
                    if (fileURL == null) {
                        if (!mappingFileName.startsWith("/")) {
                            fileURL = ClassLoader.getSystemClassLoader().getResource("/" + mappingFileName);
                        }
                    }
                }

                if (fileURL != null) {

                    try {
                        mappingFileStream = fileURL.openStream();
                        mappingFileURL = fileURL;
                    } catch (Exception exception) {
                        LOG.error(mappingFileName + " is not a file", exception);
                    }
                } else {
                    LOG.warn("Error processing input file: URL not found for " + mappingFileName + " or no privilege for reading file " + mappingFileName);
                }
            }

            if (mappingFileStream != null) {
                try {
                    mappingFileStream.close();
                } catch (Exception e) {
                    // Ignore
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingFileRetriever.init()");
        }
        return true;
    }

    @Override
    public ServiceResourceMappings retrieve(long lastKnownVersion, long lastActivationTimeInMillis) throws InterruptedException {
        return readFromFile();
    }

    private ServiceResourceMappings readFromFile() throws InterruptedException {

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingFileRetriever.readFromFile: sourceFileName=" + mappingFileName);
        }

        ServiceResourceMappings ret = null;

        if (mappingFileURL != null) {
            try (
                    InputStream serviceTagsFileStream = mappingFileURL.openStream();
                    Reader reader = new InputStreamReader(serviceTagsFileStream, StandardCharsets.UTF_8)
            ) {

                ret = gsonBuilder.fromJson(reader, ServiceResourceMappings.class);

            } catch (IOException e) {
                LOG.warn("Error processing input file: or no privilege for reading file " + mappingFileName, e);
            }
        } else {
            LOG.error("Error reading file: " + mappingFileName);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingFileRetriever.readFromFile: sourceFileName=" + mappingFileName);
        }

        return ret;
    }

}
