/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.services.hdfs;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonDeserializationContext;
import com.google.gson.JsonDeserializer;
import com.google.gson.JsonElement;
import com.google.gson.JsonParseException;
import org.apache.commons.collections.MapUtils;
import org.apache.ranger.authorization.hadoop.RangerHdfsAuthorizer;
import org.apache.ranger.authorization.hadoop.config.RangerPluginConfig;
import org.apache.ranger.plugin.policyengine.RangerAccessRequestImpl;
import org.apache.ranger.plugin.policyengine.RangerAccessResource;
import org.apache.ranger.plugin.policyengine.RangerAccessResourceImpl;
import org.apache.ranger.plugin.policyengine.RangerPluginContext;
import org.apache.ranger.plugin.policyengine.RangerResourceACLs;
import org.apache.ranger.plugin.policyevaluator.RangerPolicyEvaluator;
import org.apache.ranger.plugin.service.RangerBasePlugin;
import org.apache.ranger.plugin.util.RangerRoles;
import org.apache.ranger.plugin.util.ServicePolicies;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Type;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.assertTrue;

public class HdfsHiveACLProviderTest {
    private static Gson gsonBuilder;
    private static RangerBasePlugin plugin;

    @BeforeClass
    public static void setUpBeforeClass() {
        gsonBuilder = new GsonBuilder().setDateFormat("yyyyMMdd-HH:mm:ss.SSS-Z")
                .setPrettyPrinting()
                .registerTypeAdapter(RangerAccessResource.class, new RangerResourceDeserializer())
                .create();

        RangerPluginContext pluginContext = new RangerPluginContext(new RangerPluginConfig("hdfs", "hdfsdev", "test-chained-plugin", "cl1", "on-prem", null));
        RangerPluginConfig pluginConfig = pluginContext.getConfig();

        // To ensure that RangerRESTAdminClient does not throw outOfBounds exception
        pluginConfig.set(pluginConfig.getPropertyPrefix() + ".policy.rest.url", "http://localhost:6080");

        // To chain this plugin with hive plugin
        pluginConfig.set(pluginConfig.getPropertyPrefix() + ".chained.services", "mapped_hive");
        pluginConfig.set(pluginConfig.getPropertyPrefix() + ".chained.services.mapped_hive.impl", "org.apache.ranger.chainedplugin.hdfs.hive.RangerHdfsHiveChainedPlugin");

        // To get Hive policies
        pluginConfig.set("ranger.plugin.hive.policy.source.impl", "org.apache.ranger.plugin.util.RangerFileBasedPoliciesRetriever");
        pluginConfig.set("ranger.plugin.hive.policy.source.file.name", "test_servicepolicies_hive.json");

        // To get hdfs->hive mappings
        pluginConfig.set("ranger.plugin.hive.mapping.source.type", "file");
        pluginConfig.set("ranger.plugin.hive.mapping.source.impl", "org.apache.ranger.plugin.util.RangerMappingFileRetriever");
        pluginConfig.set("ranger.plugin.hive.mapping.source.file.name", "resourceMappings.json");

        plugin = new RangerBasePlugin(pluginContext.getConfig());
        plugin.init();

        //plugin = new RangerBasePlugin("hdfs", "HdfsHiveACLProviderTest");
    }

    @AfterClass
    public static void tearDownAfterClass() {
    }

    @Before
    public void setUp() {
    }

    @After
    public void tearDown() {
    }

    @Test
    public void testResourceMatcher_default() {
        String[] tests = { "/aclprovider/test_aclprovider_default.json" };

        runTestsFromResourceFiles(tests);
    }

    private void runTestsFromResourceFiles(String[] resourceNames) {
        for(String resourceName : resourceNames) {
            InputStream inStream = this.getClass().getResourceAsStream(resourceName);
            InputStreamReader reader   = new InputStreamReader(inStream);

            runTests(reader, resourceName);
        }
    }

    private void runTests(InputStreamReader reader, String testName) {
        HdfsHiveACLProviderTests testCases = gsonBuilder.fromJson(reader, HdfsHiveACLProviderTests.class);

        assertTrue("invalid input: " + testName, testCases != null && testCases.testCases != null);

        for(HdfsHiveACLProviderTests.TestCase testCase : testCases.testCases) {
            if (testCase.roles != null) {
                plugin.setRoles(testCase.roles);
            }
            plugin.setPolicies(testCase.servicePolicies);

            for(HdfsHiveACLProviderTests.TestCase.OneTest oneTest : testCase.tests) {
                if(oneTest == null) {
                    continue;
                }
                RangerAccessRequestImpl accessRequest = new RangerAccessRequestImpl();
                accessRequest.setResource(oneTest.resource);
                RangerResourceACLs acls = plugin.getResourceACLs(accessRequest);
                boolean userACLsMatched = true, groupACLsMatched = true;

                for (Map.Entry<String, Map<String, RangerResourceACLs.AccessResult>> entry : acls.getUserACLs().entrySet()) {
                    String userName = entry.getKey();
                    Map<String, RangerResourceACLs.AccessResult> expected = oneTest.userPermissions.get(userName);
                    if (MapUtils.isNotEmpty(entry.getValue()) && MapUtils.isNotEmpty(expected)) {
                        // Compare
                        for (Map.Entry<String, RangerResourceACLs.AccessResult> privilege : entry.getValue().entrySet()) {
                            RangerResourceACLs.AccessResult expectedResult = expected.get(privilege.getKey());
                            if (expectedResult == null) {
                                if (privilege.getValue().getResult() != RangerPolicyEvaluator.ACCESS_ALLOWED) {
                                    continue;
                                } else {
                                    userACLsMatched = false;
                                    break;
                                }
                            } else if (!expectedResult.equals(privilege.getValue())) {
                                userACLsMatched = false;
                                break;
                            }
                        }
                        if (entry.getValue().size() != expected.size()) {
                            // Compare
                            for (Map.Entry<String, RangerResourceACLs.AccessResult> privilege : expected.entrySet()) {
                                RangerResourceACLs.AccessResult actualResult = entry.getValue().get(privilege.getKey());
                                if (actualResult == null) {
                                    if (privilege.getValue().getResult() != RangerPolicyEvaluator.ACCESS_ALLOWED) {
                                        continue;
                                    } else {
                                        userACLsMatched = false;
                                        break;
                                    }
                                } else if (!actualResult.equals(privilege.getValue())) {
                                    userACLsMatched = false;
                                    break;
                                }
                            }
                        }
                    } else if (!(MapUtils.isEmpty(entry.getValue()) && MapUtils.isEmpty(expected))) {
                        userACLsMatched = false;
                        break;
                    }
                    if (!userACLsMatched) {
                        break;
                    }
                }

                for (Map.Entry<String, Map<String, RangerResourceACLs.AccessResult>> entry : acls.getGroupACLs().entrySet()) {
                    String groupName = entry.getKey();
                    Map<String, RangerResourceACLs.AccessResult> expected = oneTest.groupPermissions.get(groupName);
                    if (MapUtils.isNotEmpty(entry.getValue()) && MapUtils.isNotEmpty(expected)) {
                        // Compare
                        for (Map.Entry<String, RangerResourceACLs.AccessResult> privilege : entry.getValue().entrySet()) {
                            RangerResourceACLs.AccessResult expectedResult = expected.get(privilege.getKey());
                            if (expectedResult == null) {
                                if (privilege.getValue().getResult() != RangerPolicyEvaluator.ACCESS_ALLOWED) {
                                    continue;
                                } else {
                                    groupACLsMatched = false;
                                    break;
                                }
                            } else if (!expectedResult.equals(privilege.getValue())) {
                                groupACLsMatched = false;
                                break;
                            }
                        }
                        if (entry.getValue().size() != expected.size()) {
                            // Compare
                            for (Map.Entry<String, RangerResourceACLs.AccessResult> privilege : expected.entrySet()) {
                                RangerResourceACLs.AccessResult actualResult = entry.getValue().get(privilege.getKey());
                                if (actualResult == null) {
                                    if (privilege.getValue().getResult() != RangerPolicyEvaluator.ACCESS_ALLOWED) {
                                        continue;
                                    } else {
                                        groupACLsMatched = false;
                                        break;
                                    }
                                } else if (!actualResult.equals(privilege.getValue())) {
                                    groupACLsMatched = false;
                                    break;
                                }
                            }
                        }
                    } else if (!(MapUtils.isEmpty(entry.getValue()) && MapUtils.isEmpty(expected))) {
                        groupACLsMatched = false;
                        break;
                    }
                    if (!groupACLsMatched) {
                        break;
                    }
                }

                assertTrue("getResourceACLs() failed! " + testCase.name + ":" + oneTest.name, userACLsMatched && groupACLsMatched);
            }
        }
    }

    static class HdfsHiveACLProviderTests {
        List<TestCase> testCases;

        static class TestCase {
            String               name;
            RangerRoles          roles;
            ServicePolicies      servicePolicies;
            List<OneTest>        tests;

            static class OneTest {
                String               name;
                RangerAccessResource resource;
                Map<String, Map<String, RangerResourceACLs.AccessResult>> userPermissions;
                Map<String, Map<String, RangerResourceACLs.AccessResult>> groupPermissions;
            }
        }
    }

    static class RangerResourceDeserializer implements JsonDeserializer<RangerAccessResource> {
        @Override
        public RangerAccessResource deserialize(JsonElement jsonObj, Type type,
                                                JsonDeserializationContext context) throws JsonParseException {
            return gsonBuilder.fromJson(jsonObj, RangerHdfsResource.class);
        }
    }

    static class RangerHdfsResource extends RangerAccessResourceImpl {

        public RangerHdfsResource(String path, String owner) {
            super.setValue(RangerHdfsAuthorizer.KEY_RESOURCE_PATH, path);
            super.setOwnerUser(owner);
        }
    }
}
