/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.services.hdfs;

import static org.junit.Assert.*;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Type;
import java.util.Date;
import java.util.List;

import com.google.gson.JsonDeserializationContext;
import com.google.gson.JsonDeserializer;
import com.google.gson.JsonElement;
import com.google.gson.JsonParseException;
import org.apache.ranger.authorization.hadoop.config.RangerPluginConfig;
import org.apache.ranger.plugin.audit.RangerDefaultAuditHandler;
import org.apache.ranger.plugin.policyengine.RangerAccessRequest;
import org.apache.ranger.plugin.policyengine.RangerAccessRequestImpl;
import org.apache.ranger.plugin.policyengine.RangerAccessResource;
import org.apache.ranger.plugin.policyengine.RangerAccessResourceImpl;
import org.apache.ranger.plugin.policyengine.RangerAccessResult;
import org.apache.ranger.plugin.policyengine.RangerAccessResultProcessor;
import org.apache.ranger.plugin.policyengine.RangerPluginContext;
import org.apache.ranger.plugin.service.RangerBasePlugin;
import org.apache.ranger.plugin.util.ServicePolicies;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

public class TestChainedPlugin {
    private static Gson gsonBuilder;
    private static RangerBasePlugin plugin;
    static RangerPluginContext pluginContext;

    @BeforeClass
    public static void setUpBeforeClass() {
        gsonBuilder = new GsonBuilder().setDateFormat("yyyyMMdd-HH:mm:ss.SSS-Z")
                .setPrettyPrinting()
                .registerTypeAdapter(RangerAccessRequest.class, new RangerAccessRequestDeserializer())
                .registerTypeAdapter(RangerAccessResource.class, new RangerResourceDeserializer())
                .create();

        pluginContext = new RangerPluginContext(new RangerPluginConfig("hdfs", "hdfsdev", "test-chained-plugin", "cl1", "on-prem", null));
        RangerPluginConfig pluginConfig = pluginContext.getConfig();

        // To ensure that RangerRESTAdminClient does not throw outOfBounds exception
        pluginConfig.set(pluginConfig.getPropertyPrefix() + ".policy.rest.url", "http://localhost:6080");

        // To chain this plugin with hive plugin
        pluginConfig.set(pluginConfig.getPropertyPrefix() + ".chained.services", "mapped_hive");
        pluginConfig.set(pluginConfig.getPropertyPrefix() + ".chained.services.mapped_hive.impl", "org.apache.ranger.chainedplugin.hdfs.hive.RangerHdfsHiveChainedPlugin");

        // To get Hive policies
        pluginConfig.set("ranger.plugin.hive.policy.source.impl", "org.apache.ranger.plugin.util.RangerFileBasedPoliciesRetriever");
        pluginConfig.set("ranger.plugin.hive.policy.source.file.name", "test_servicepolicies_hive.json");

        // To get hdfs->hive mappings
        pluginConfig.set("ranger.plugin.hive.mapping.source.type", "file");
        pluginConfig.set("ranger.plugin.hive.mapping.source.impl", "org.apache.ranger.plugin.util.RangerMappingFileRetriever");
        pluginConfig.set("ranger.plugin.hive.mapping.source.file.name", "resourceMappings.json");

        plugin = new RangerBasePlugin(pluginContext.getConfig());
        plugin.init();
    }

    @AfterClass
    public static void tearDownAfterClass() {
    }

    @Before
    public void setUp() {
    }

    @After
    public void tearDown() {
    }

    @Test
    public void testHdfsHiveChaining() {
        String[] tests = {"/test_hdfs_hive_chaining.json"};

        runTestsFromResourceFiles(tests);
    }

    @Test
    public void testHdfsHiveChaining2() {
        String[] tests = {"/test_hdfs_hive_chaining-2.json"};

        runTestsFromResourceFiles(tests);
    }

    private void runTestsFromResourceFiles(String[] resourceNames) {
        for(String resourceName : resourceNames) {
            InputStream       inStream = this.getClass().getResourceAsStream(resourceName);
            InputStreamReader reader   = new InputStreamReader(inStream);

            runTests(reader);
        }
    }

    private void runTests(InputStreamReader reader) {
        RangerChainedPluginTests testCases = gsonBuilder.fromJson(reader, RangerChainedPluginTests.class);

        for (RangerChainedPluginTests.TestCase testCase : testCases.testCases) {

            plugin.setPolicies(testCase.servicePolicies);

            RangerAccessRequest request;

            for (RangerChainedPluginTests.TestCase.TestData test : testCase.tests) {
                request = test.request;

                RangerAccessResultProcessor auditHandler = new RangerDefaultAuditHandler();

                if (test.result != null) {
                    RangerAccessResult expected = test.result;
                    RangerAccessResult result;

                    result = plugin.isAccessAllowed(request, auditHandler);

                    assertNotNull("result was null! - " + testCase.name + ":" + test.name, result);
                    assertEquals("isAllowed mismatched! - " + test.name, expected.getIsAllowed(), result.getIsAllowed());
                    assertEquals("isAudited mismatched! - " + test.name, expected.getIsAudited(), result.getIsAudited());
                }
            }
        }
    }

    static class RangerChainedPluginTests {
        public List<TestCase> testCases;

        static class TestCase {
            public String               name;
            public ServicePolicies      servicePolicies;
            public List<TestData> tests;

            static class TestData {
                public String              name;
                public RangerAccessRequest request;
                public RangerAccessResult  result;
            }
        }
    }

    static class RangerAccessRequestDeserializer implements JsonDeserializer<RangerAccessRequest> {
        @Override
        public RangerAccessRequest deserialize(JsonElement jsonObj, Type type,
                                               JsonDeserializationContext context) throws JsonParseException {
            RangerAccessRequestImpl ret = gsonBuilder.fromJson(jsonObj, RangerAccessRequestImpl.class);

            ret.setAccessType(ret.getAccessType()); // to force computation of isAccessTypeAny and isAccessTypeDelegatedAdmin
            if (ret.getAccessTime() == null) {
                ret.setAccessTime(new Date());
            }

            return ret;
        }
    }

    static class RangerResourceDeserializer implements JsonDeserializer<RangerAccessResource> {
        @Override
        public RangerAccessResource deserialize(JsonElement jsonObj, Type type,
                                                JsonDeserializationContext context) throws JsonParseException {
            return gsonBuilder.fromJson(jsonObj, RangerAccessResourceImpl.class);
        }
    }
}


