/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.hadoop.crypto.key;

import java.io.FileOutputStream;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.apache.ranger.plugin.util.SearchFilter;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import javax.security.auth.Subject;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.crypto.key.kms.server.KMSConfiguration;
import org.apache.hadoop.security.SecureClientLogin;
import org.apache.ranger.authorization.utils.StringUtil;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyItem;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyItemAccess;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyResource;
import org.apache.ranger.plugin.model.RangerService;
import org.apache.ranger.plugin.util.RangerRESTClient;
import com.nimbusds.oauth2.sdk.http.HTTPResponse;
import com.sun.jersey.api.client.ClientResponse;

import static org.apache.ranger.plugin.model.RangerPolicy.POLICY_PRIORITY_NORMAL;
import static org.apache.ranger.plugin.model.RangerPolicy.POLICY_PRIORITY_OVERRIDE;

public class KeyTrusteeACLToRangerPolicy {
	private static final Logger LOG = Logger.getLogger(KeyTrusteeACLToRangerPolicy.class.getName());

	private static final String HADOOP_KMS_PREFIX            = "hadoop.kms.";
	private static final String HADOOP_KMS_BLACKLIST_PREFIX  = "hadoop.kms.blacklist.";
	private static final String DEFAULT_KEY_PREFIX           = "default.key.";
	private static final String WHITELIST_KEY_PREFIX         = "whitelist.key.";
	private static final String KEY_ACL_PREFIX               = "key.acl.";

	private static final String KEY_ACL_MANAGEMENT             = "MANAGEMENT";
	private static final String KEY_ACL_READ                   = "READ";
	private static final String KEY_ACL_GENERATE_EEK           = "GENERATE_EEK";
	private static final String KEY_ACL_DECRYPT_EEK            = "DECRYPT_EEK";
	private static final String KEY_ACL_ALL                    = "ALL";
	private static final String KEY_ACL_CREATE                 = "CREATE";
	private static final String KEY_ACL_DELETE                 = "DELETE";
	private static final String KEY_ACL_ROLLOVER               = "ROLLOVER";
	private static final String KEY_ACL_SET_KEY_MATERIAL       = "SET_KEY_MATERIAL";
	private static final String KEY_ACL_GET                    = "GET";
	private static final String KEY_ACL_GET_KEYS               = "GET_KEYS";
	private static final String KEY_ACL_GET_METADATA           = "GET_METADATA";

	private static final String ACCESS_TYPE_DECRYPT_EEK        = "decrypteek";
	private static final String ACCESS_TYPE_GENERATE_EEK       = "generateeek";
	private static final String ACCESS_TYPE_GET_METADATA       = "getmetadata";
	private static final String ACCESS_TYPE_GET                = "get";
	private static final String ACCESS_TYPE_GET_KEYS           = "getkeys";
	private static final String ACCESS_TYPE_CREATE             = "create";
	private static final String ACCESS_TYPE_ROLLOVER           = "rollover";
	private static final String ACCESS_TYPE_DELETE             = "delete";
	private static final String ACCESS_TYPE_SET_KEY_MATERIAL   = "setkeymaterial";

	private static final String SERVICE_TYPE_ACL               = "serviceTypeAcl";
	private static final String POLICY_TYPE_ACL                = "policyTypeAcl";
	private static final String WHITELIST_TYPE_ACL             = "whiteListTypeAcl";
	private static final String BLACKLIST_TYPE_ACL             = "blackListTypeAcl";
	private static final String KEY_SPECIFIC_ACL               = "keySpecificAcl";
	private static final String KEY_TRUSTEE_ACL                = "keyTrusteeAcl";

	private static final String RANGER_KMS_REST_URL             = "ranger.plugin.kms.policy.rest.url";
	private static final String DEFAULT_RANGER_KMS_SERVICE_NAME = "cm_kms";
	private static final String RANGER_KMS_SERVICE_TYPE         = "kms";
	private static final String DEFAULT_KEY_POLICY_RESOURCE     = "*";
	private static final String RESOURCE_KEY_NAME               = "keyname";
	private static final String WHITELIST_RESOURCE_KEY_NAME     = "***whitelist***";
	private static final String BLACKLIST_RESOURCE_KEY_NAME     = "***blacklist***";
	private static final String GLOBAL_OVERRIDE_POLICY_NAME     = "Global overrides";

	private static final String AUTHENTICATION_TYPE                     = "hadoop.kms.authentication.type";
	private static final String KMS_USER_PRINCIPAL                      = "ranger.ks.kerberos.principal";
	private static final String RANGER_SERVICE_HOST                     = "ranger.service.host";
	private static final String KMS_USER_KEYTAB                         = "ranger.ks.kerberos.keytab";
	private static final String RANGER_SERVICE_HTTPS_ATTRIB_SSL_ENABLED = "ranger.service.https.attrib.ssl.enabled";
	private static final String KMS_NAME_RULES                          = "hadoop.security.auth_to_local";
	private static final String RANGER_KMS_SECURITY                     = "ranger-kms-security.xml";
	private static final String RANGER_KMS_POLICY_MGR_SSL               = "ranger-kms-policymgr-ssl.xml";
	private static final String RANGER_KMS_SITE                         = "ranger-kms-site.xml";
	private static final String AUTH_KERBEROS                           = "kerberos";
	private static final String USER_NOBODY                             = "nobody";
	private static final String REST_POLICIES                           = "/service/plugins/policies";
	private static final String REST_GET_USERS                          = "/service/xusers/lookup/users";
	private static final String REST_GET_GROUPS                         = "/service/xusers/lookup/groups";
	private static final String REST_GET_KMS_SERVICE                    = "/service/plugins/services/name/" + DEFAULT_RANGER_KMS_SERVICE_NAME;

	private static final Set<String> READ_PERMISSIONS             = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_GET, ACCESS_TYPE_GET_KEYS, ACCESS_TYPE_GET_METADATA)));
	private static final Set<String> DECRYPT_EEK_PERMISSIONS      = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_DECRYPT_EEK)));
	private static final Set<String> GENERATE_EEK_PERMISSIONS     = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_GENERATE_EEK)));
	private static final Set<String> ACL_MANAGEMENT_PERMISSIONS   = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_CREATE, ACCESS_TYPE_ROLLOVER, ACCESS_TYPE_DELETE)));
	private static final Set<String> CREATE_PERMISSIONS           = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_CREATE)));
	private static final Set<String> DELETE_PERMISSIONS           = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_DELETE)));
	private static final Set<String> ROLLOVER_PERMISSIONS         = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_ROLLOVER)));
	private static final Set<String> SET_KEY_MATERIAL_PERMISSIONS = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_SET_KEY_MATERIAL)));
	private static final Set<String> GET_PERMISSIONS              = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_GET)));
	private static final Set<String> GET_KEYS_PERMISSIONS         = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_GET_KEYS)));
	private static final Set<String> GET_METADATA_PERMISSIONS     = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_GET_METADATA)));
	private static final Set<String> ALL_PERMISSIONS              = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(ACCESS_TYPE_CREATE, ACCESS_TYPE_ROLLOVER, ACCESS_TYPE_DELETE, ACCESS_TYPE_SET_KEY_MATERIAL, ACCESS_TYPE_GET, ACCESS_TYPE_GET_KEYS, ACCESS_TYPE_GET_METADATA, ACCESS_TYPE_GENERATE_EEK, ACCESS_TYPE_DECRYPT_EEK)));

	private final Path                                         dbks_path;
	private final Path                                         kms_acl_path;
	private final String                                       principal;
	private final String                                       keytab;
	private final String                                       nameRules;
	private final String                                       baseURL;
	private final RangerRESTClient                             restClient;
	private final List<String>                                 existingUserList        = new ArrayList<String>();
	private final List<String>                                 existingGroupList       = new ArrayList<String>();
	private final List<String>                                 nonExistingUserList     = new ArrayList<String>();
	private final List<String>                                 nonExistingGroupList    = new ArrayList<String>();
	private final Map<String, RangerKMSKeyUserGroupPermission> existingUserGroupkeyMap = new HashMap<String, RangerKMSKeyUserGroupPermission>();
	private       boolean                                      isServiceWideACLPresent = false;
	private       boolean                                      isPolicyACLPresent      = false;

	public static void main(String[] args) {
		try {
			if (args.length == 0) {
				logErrorDetails("Invalid number of parameter found.");

				logErrorDetails("USAGE: java " + KeyTrusteeACLToRangerPolicy.class.getName() + " <Path for Conf Directory>");

				System.exit(1);
			}

			KeyTrusteeACLToRangerPolicy kt = new KeyTrusteeACLToRangerPolicy(args[0]);

			kt.execute();

			System.exit(0);
		} catch (Exception e) {
			logErrorDetails("Error : ", e);

			System.exit(1);
		}
	}

	public KeyTrusteeACLToRangerPolicy() {
		kms_acl_path = null;
		principal = null;
		keytab    = null;
		nameRules = null;
		restClient = null;
		dbks_path = null;
		baseURL = null;
	}
	private KeyTrusteeACLToRangerPolicy(String path) throws Exception {
		Configuration config = RangerKeyStoreProvider.getDBKSConf();

		config.addResource(KMSConfiguration.getKMSConf());
		config.addResource(KMSConfiguration.getConfiguration(true, RANGER_KMS_SECURITY));
		config.addResource(KMSConfiguration.getConfiguration(true, RANGER_KMS_SITE));

		dbks_path    = Paths.get(path, "dbks-site.xml");
		kms_acl_path = Paths.get(path, "kms-acls.xml");
		baseURL      = config.get(RANGER_KMS_REST_URL);

		String  authenticationType = config.get(AUTHENTICATION_TYPE);
		boolean isSSLEnabled       = config.getBoolean(RANGER_SERVICE_HTTPS_ATTRIB_SSL_ENABLED, false);

		if (!StringUtil.isEmpty(authenticationType) && !AUTH_KERBEROS.equals(authenticationType)) {
			logErrorDetails("Authentication Type : " + authenticationType);
			logErrorDetails("It is recommended to setup Ranger and Ranger KMS in Kerberize mode");

			System.exit(1);
		}
		if (isSSLEnabled) {
			String sslConfigFile = Paths.get(path, RANGER_KMS_POLICY_MGR_SSL).toString();
			restClient = new RangerRESTClient(baseURL, sslConfigFile, config);
		} else {
			restClient = new RangerRESTClient(baseURL, "", config);
		}

		if (!StringUtil.isEmpty(authenticationType) && AUTH_KERBEROS.equals(authenticationType)) {
			try {
				String localhost = config.get(RANGER_SERVICE_HOST);

				principal = SecureClientLogin.getPrincipal(config.get(KMS_USER_PRINCIPAL), localhost);
				keytab    = config.get(KMS_USER_KEYTAB);
				nameRules = config.get(KMS_NAME_RULES);
			} catch (Exception ex) {
				logErrorDetails("Error while setting kerberos parameter", ex);

				throw new Exception("Error while setting kerberos parameter", ex);
			}
		} else {
			principal = null;
			keytab    = null;
			nameRules = null;
		}
	}

	private void execute() throws Exception {
		if (SecureClientLogin.isKerberosCredentialExists(principal, keytab)) {
			try {
				logInfoDetails("Provided Kerberos Credential : Principal = " + principal + " and Keytab = " + keytab);

				Subject sub = SecureClientLogin.loginUserFromKeytab(principal, keytab, nameRules);

				Subject.doAs(sub, new PrivilegedAction<Void>() {
					@Override
					public Void run() {
						logInfoDetails("Starting ACL migration using kerberos credential");

						try {
							operation();
						} catch (Exception e) {
							logErrorDetails("Migration fail in kerberos mode", e);
							throw new RuntimeException("Migration fail in kerberos mode", e);
						}

						return null;
					}
				});
			} catch (Exception e) {
				logErrorDetails("Migration fail in kerberos mode", e);

				throw new Exception("Migration fail in kerberos mode", e);
			}
		}
	}

	private void operation() throws Exception {
		try {
			validateAclConfigFilePath();
			verifyRangerKmsServiceExists();
			getLookUpUserAndGroup();

			Document dbksDoc = parseAclXmlFile();

			if(!nonExistingUserList.isEmpty() || !nonExistingGroupList.isEmpty()) {
				messageForNonExistingUserAndGroup();

				System.exit(1);
			} else {
				if (isServiceWideACLPresent) {
					writeToDBKSFile(dbksDoc);
				}

				if (isPolicyACLPresent) {
					createOrUpdatePolicies();
				}
			}
		} catch (Exception ex) {
			logErrorDetails("Error while migrating ACL", ex);

			throw new Exception("Error while migrating ACL", ex);
		}
	}

	private void validateAclConfigFilePath() {
		if (!dbks_path.toFile().exists()) {
			logErrorDetails("dbks-site.xml file not found at: " + dbks_path);

			System.exit(1);
		}

		if (!kms_acl_path.toFile().exists()) {
			logErrorDetails("kms-acls.xml file not found at: " + kms_acl_path);

			System.exit(1);
		}
	}

	private void verifyRangerKmsServiceExists() throws Exception {
		try {
			ClientResponse response = restClient.get(REST_GET_KMS_SERVICE, null);
			RangerService  service  = response.getEntity(RangerService.class);

			if (response.getStatus() == HTTPResponse.SC_OK && service != null && service.getName().equals(DEFAULT_RANGER_KMS_SERVICE_NAME)) {
				logInfoDetails("Found Ranger KMS Service with name " + DEFAULT_RANGER_KMS_SERVICE_NAME);
			} else {
				logErrorDetails("Default Ranger KMS Service does not exist. Kindly create it with name " + DEFAULT_RANGER_KMS_SERVICE_NAME);

				System.exit(1);
			}
		} catch (Exception ex) {
			logErrorDetails("Error while getting Ranger KMS default service", ex);
			throw new Exception("Error while getting Ranger KMS default service", ex);
		}
	}

	private void getLookUpUserAndGroup() throws Exception {
		try {
			int     pageSize                = 200;
			long    totalCount              = -1;
			long    retrievedCount          = 0;
			boolean isLastPage              = false;
			Map<String, String> queryParams = new HashMap<>();

			queryParams.put("isVisible", "1");
			queryParams.put("pageSize", Integer.toString(pageSize));

			for (int startIdx = 0; !isLastPage; startIdx += pageSize) {
				logInfoDetails("Retrieving users from Ranger: startIdx=" + startIdx + ", pageSize=" + pageSize);

				queryParams.put("startIndex", Integer.toString(startIdx));

				ClientResponse userResponse   = restClient.get(REST_GET_USERS, queryParams);
				VXStringList   userLookupList = userResponse.getEntity(VXStringList.class);

				if (userResponse.getStatus() == HTTPResponse.SC_OK) {
					if (totalCount == -1) {
						totalCount = userLookupList.getTotalCount();

						logInfoDetails("Total number of users found in Ranger: " + totalCount);
					}

					List<VXString> userList = userLookupList.getvXStrings();

					if (!userList.isEmpty()) {
						for (VXString userValue : userList) {
							existingUserList.add(userValue.getValue());
						}

						retrievedCount += userList.size();
						isLastPage     =  retrievedCount >= totalCount;
					} else {
						isLastPage = true;
					}
				} else {
					logErrorDetails("Failed to get User from Ranger: " + userResponse.getStatus());

					System.exit(1);
				}
			}

			if (existingUserList.isEmpty()) {
				logInfoDetails("No users are present in Ranger.");
			}

			totalCount     = -1;
			retrievedCount = 0;
			isLastPage     = false;

			queryParams.clear();
			queryParams.put("isVisible", "1");
			queryParams.put("pageSize", Integer.toString(pageSize));

			for (int startIdx = 0; !isLastPage; startIdx += pageSize) {
				logInfoDetails("Retrieving groups from Ranger: startIdx=" + startIdx + ", pageSize=" + pageSize);

				queryParams.put("startIndex", Integer.toString(startIdx));

				ClientResponse groupResponse   = restClient.get(REST_GET_GROUPS, queryParams);
				VXStringList   groupLookupList = groupResponse.getEntity(VXStringList.class);

				if (groupResponse.getStatus() == HTTPResponse.SC_OK) {
					if (totalCount == -1) {
						totalCount = groupLookupList.getTotalCount();

						logInfoDetails("Total number of groups found in Ranger: " + totalCount);
					}

					List<VXString> groupList = groupLookupList.getvXStrings();

					if (!groupList.isEmpty()) {
						for (VXString groupValue : groupList) {
							existingGroupList.add(groupValue.getValue());
						}

						retrievedCount += groupList.size();
						isLastPage     =  retrievedCount >= totalCount;
					} else {
						isLastPage = true;
					}
				} else {
					logErrorDetails("Failed to get group from Ranger: " + groupResponse.getStatus());

					System.exit(1);
				}
			}

			if (existingGroupList.isEmpty()) {
				logInfoDetails("No groups are present in Ranger.");
			}
		} catch (Exception ex) {
			logErrorDetails("Error while getting exiting user / groups", ex);

			throw new Exception("Error while getting exiting user / groups", ex);
		}
	}

	public Path getKmsAclPath() {
		return kms_acl_path;
	}

	public Path getDbksPath() {
		return dbks_path;
	}

	public List<String> getExistingUserList(){
		return existingUserList;
	}

	public List<String> getExistingGroupList(){
		return existingGroupList;
	}

	public Map<String, RangerKMSKeyUserGroupPermission> getUserGroupMap() {
		return existingUserGroupkeyMap;
	}


	public Document parseAclXmlFile() throws Exception {
		try {
			DocumentBuilderFactory factory   = DocumentBuilderFactory.newInstance();
			DocumentBuilder        builder   = factory.newDocumentBuilder();
			Document               kmsAclDoc = builder.parse(getKmsAclPath().toFile());

			kmsAclDoc.getDocumentElement().normalize();

			NodeList ktsNodeList = kmsAclDoc.getElementsByTagName("property");
			Document dbksDoc     = builder.parse(getDbksPath().toFile());

			dbksDoc.getDocumentElement().normalize();

			NodeList dbksNodeList    = dbksDoc.getElementsByTagName("property");
			Element parentConfigNode = dbksDoc.getDocumentElement();

			for (int idx = 0; idx < ktsNodeList.getLength(); idx++) {
				Node ktsNode = ktsNodeList.item(idx);

				if (ktsNode.getNodeType() == Node.ELEMENT_NODE) {
					Element ktsElement   = (Element) ktsNode;
					Node    nodeName     = ktsElement.getElementsByTagName("name").item(0);
					Node    nodeValue    = ktsElement.getElementsByTagName("value").item(0);
					String  ktsPropName  = (nodeName != null) ? nodeName.getTextContent() : null;
					String  ktsPropValue = (nodeValue != null) ? nodeValue.getTextContent() : "";
					String  aclType      = getACLType(ktsPropName);

					if (StringUtils.isEmpty(ktsPropName) || StringUtil.isEmpty(ktsPropValue)) {
						continue;
					}

					if (aclType.equals(SERVICE_TYPE_ACL)) {
						isServiceWideACLPresent = true;

						Element existingElement = null;

						for (int i = 0; i < dbksNodeList.getLength(); i++) {
							Node dbksNode = dbksNodeList.item(i);

							if (dbksNode.getNodeType() == Node.ELEMENT_NODE) {
								Element dbksElement  = (Element) dbksNode;
								String  dbksPropName = dbksElement.getElementsByTagName("name").item(0).getTextContent();

								// property already exists
								if (dbksPropName.equals(ktsPropName)) {
									existingElement = dbksElement;

									break;
								}
							}

						}

						Element propertyElement = dbksDoc.createElement("property");
						Element e               = dbksDoc.createElement("name");

						e.appendChild(dbksDoc.createTextNode(ktsPropName));
						propertyElement.appendChild(e);

						e = dbksDoc.createElement("value");

						e.appendChild(dbksDoc.createTextNode(ktsPropValue));
						propertyElement.appendChild(e);

						if (existingElement == null) {
							logInfoDetails("Adding new property : "
									+ propertyElement.getElementsByTagName("name").item(0).getTextContent() + " = "
									+ propertyElement.getElementsByTagName("value").item(0).getTextContent());

							parentConfigNode.appendChild(propertyElement);
						} else {
							logInfoDetails("Replacing existing property with new value : ");
							logInfoDetails("  Old Config : "
									+ existingElement.getElementsByTagName("name").item(0).getTextContent() + " = "
									+ existingElement.getElementsByTagName("value").item(0).getTextContent());
							logInfoDetails("  New Config : "
									+ propertyElement.getElementsByTagName("name").item(0).getTextContent() + " = "
									+ propertyElement.getElementsByTagName("value").item(0).getTextContent());

							parentConfigNode.replaceChild(propertyElement, existingElement);
						}

					} else if (aclType.equals(POLICY_TYPE_ACL) || aclType.equals(KEY_SPECIFIC_ACL) || aclType.equals(WHITELIST_TYPE_ACL) || aclType.equals(BLACKLIST_TYPE_ACL)) {
						isPolicyACLPresent = true;

						final Set<String>              userSet;
						final Set<String>              groupSet;
						final Map<String, Set<String>> usrPermissionMap = new HashMap<String, Set<String>>();
						final Map<String, Set<String>> grpPermissionMap = new HashMap<String, Set<String>>();
						final int                      keyNameIndex     = 2;
						final int                      accessTypeIndex  = 3;
						final String[]                 ktsPropNameArray = ktsPropName.split("\\.");
						final String                   accessType       = ktsPropNameArray[accessTypeIndex];
						final String                   keyname;
						final Set<String>              permissionSet;


						if (aclType.equals(POLICY_TYPE_ACL)) {
							keyname       = DEFAULT_KEY_POLICY_RESOURCE;
							permissionSet = createPermissionSet(accessType);
						} else if (aclType.equals(KEY_SPECIFIC_ACL)) {
							keyname       = ktsPropNameArray[keyNameIndex];
							permissionSet = createPermissionSet(accessType);
						} else if (aclType.equals(WHITELIST_TYPE_ACL)) {
							keyname       = WHITELIST_RESOURCE_KEY_NAME;
							permissionSet = createPermissionSet(accessType);
						} else if (aclType.equals(BLACKLIST_TYPE_ACL)) {
							keyname       = BLACKLIST_RESOURCE_KEY_NAME;
							permissionSet = createPermissionSet(accessType);
						} else {
							keyname       = null;
							permissionSet = Collections.emptySet();
						}

						if (ktsPropValue.equals("*")) {
							userSet  = null;
							groupSet = new HashSet<>();

							groupSet.add("public");
						} else {
							String[] userGroupArray = ktsPropValue.split(" ");

							if (userGroupArray.length == 2) {
								userSet  = new HashSet<>(Arrays.asList(userGroupArray[0].split(",")));
								groupSet = new HashSet<>(Arrays.asList(userGroupArray[1].split(",")));
							} else {
								userSet  = new HashSet<>(Arrays.asList(ktsPropValue.split(",")));
								groupSet = null;
							}
						}

						if (userSet != null && !userSet.isEmpty()) {
							for (String user : userSet) {
								user = user.trim();
								if(!StringUtil.isEmpty(user)) {
									/*user: nobody means there is only set of groups present.
									 * We should not consider value "nobody".
									 * Ref Link: https://docs.cloudera.com/documentation/enterprise/6/6.3/topics/cdh_sg_kms_acl_config.html */
									if (!user.equalsIgnoreCase(USER_NOBODY)) {
										if (getExistingUserList().contains(user.trim())) {
											usrPermissionMap.put(user, new HashSet<String>(permissionSet));
										} else {
											nonExistingUserList.add(user);
										}
									}
								}
							}
						}

						if (groupSet != null && !groupSet.isEmpty()) {
							for (String group : groupSet) {
								group = group.trim();
								if(!StringUtil.isEmpty(group)) {
									if (getExistingGroupList().contains(group)) {
										grpPermissionMap.put(group, new HashSet<String>(permissionSet));
									} else {
										nonExistingGroupList.add(group);
									}
								}
							}
						}

						collectUserGroupPermissions(keyname, usrPermissionMap, grpPermissionMap);
					}
				}
			}

			return dbksDoc;
		} catch (Exception e) {
			logErrorDetails("Error while parsing ACL XML file", e);

			throw new Exception("Error while parsing ACL XML file", e);
		}
	}

	private void messageForNonExistingUserAndGroup() {
		if(!nonExistingUserList.isEmpty()) {
			logErrorDetails("Following users do not exist in Ranger DB: " + nonExistingUserList.stream().distinct().collect(Collectors.toList()));
		}

		if(!nonExistingGroupList.isEmpty()) {
			logErrorDetails("Following groups do not exist in Ranger DB: " + nonExistingGroupList.stream().distinct().collect(Collectors.toList()));
		}

		logErrorDetails("To fix this problem, either add the above listed users/groups to your user management system and re-sync the users/groups, or create the listed users/groups in Ranger using the Ranger Admin Web UI : " + baseURL);
	}

	public void writeToDBKSFile(Document dom) throws Exception {
		logInfoDetails("Updating " + getDbksPath().toString());

		try {
			Transformer tr = TransformerFactory.newInstance().newTransformer();

			tr.setOutputProperty(OutputKeys.INDENT, "yes");
			tr.setOutputProperty(OutputKeys.METHOD, "xml");
			tr.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
			tr.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");

			// send DOM to file
			tr.transform(new DOMSource(dom), new StreamResult(new FileOutputStream(getDbksPath().toString())));
		} catch (Exception ex) {
			logErrorDetails("Error while writing in dbks-site.xml of Ranger KMS file", ex);

			throw new Exception("Error while writing in dbks-site.xml of Ranger KMS file", ex);
		}
	}

	private void createOrUpdatePolicies() throws Exception {
		logInfoDetails("Updating Ranger with KMS policies for " + existingUserGroupkeyMap.size() + " keys");

		for (Map.Entry<String, RangerKMSKeyUserGroupPermission> entry : existingUserGroupkeyMap.entrySet()) {
			String                          resourceName   = entry.getKey();
			RangerKMSKeyUserGroupPermission permissions    = entry.getValue();
			RangerPolicy                    existingPolicy = getRangerKMSPolicyForResource(resourceName);

			if (existingPolicy == null) {
				createPolicy(permissions);
			} else {
				updatePolicy(existingPolicy, permissions);
			}
		}
	}

	private RangerPolicy getRangerKMSPolicyForResource(String keyname) throws Exception {
		RangerPolicy        ret               = null;
		Map<String, String> queryParams       = new HashMap<String, String>();
		int                 keyPolicyPriority = POLICY_PRIORITY_NORMAL;

		queryParams.put("serviceName", DEFAULT_RANGER_KMS_SERVICE_NAME);
		queryParams.put("serviceType", RANGER_KMS_SERVICE_TYPE);

		if (StringUtils.equals(keyname, WHITELIST_RESOURCE_KEY_NAME) || StringUtils.equals(keyname, BLACKLIST_RESOURCE_KEY_NAME)) {
			keyPolicyPriority = POLICY_PRIORITY_OVERRIDE;

			queryParams.put(SearchFilter.POLICY_PRIORITY, Integer.toString(keyPolicyPriority));

			keyname = DEFAULT_KEY_POLICY_RESOURCE;
		}

		queryParams.put(SearchFilter.RESOURCE_PREFIX + RESOURCE_KEY_NAME, keyname);

		try {
			ClientResponse res = restClient.get(REST_POLICIES, queryParams);

			if (res.getStatus() == HTTPResponse.SC_OK) {
				RangerKMSPolicyList kmsPolicyList = res.getEntity(RangerKMSPolicyList.class);
				List<RangerPolicy>  policies      = kmsPolicyList != null ? kmsPolicyList.getPolicies() : null;

				if (policies != null) {
					for (RangerPolicy policy : policies) {
						int policyPriority = policy.getPolicyPriority() == null ? POLICY_PRIORITY_NORMAL : policy.getPolicyPriority();

						if (policyPriority != keyPolicyPriority) {
							continue;
						}

						RangerPolicyResource policyResource = policy.getResources() != null ? policy.getResources().get(RESOURCE_KEY_NAME) : null;
						List<String>         policyKeyNames = policyResource != null ? policyResource.getValues() : null;

						if (policyKeyNames != null &&  policyKeyNames.size() == 1 && StringUtils.equals(keyname, policyKeyNames.get(0))) {
							ret = policy;

							break;
						}
					}
				}
			} else {
				logErrorDetails("Failed to get Ranger KMS policy for resource: " + keyname + "; http-status: " + res.getStatus());
			}
		} catch (Exception ex) {
			logErrorDetails("Error while getting Ranger KMS policy for resource: " + keyname, ex);

			throw new Exception("Error while getting Ranger KMS policy for resource: " + keyname, ex);
		}

		if (ret == null) {
			logInfoDetails("Ranger KMS policy does not exist for resource: " + keyname + ", priority=" + (keyPolicyPriority == POLICY_PRIORITY_OVERRIDE ? "override" : "normal"));
		}

		return ret;
	}

	private void createPolicy(RangerKMSKeyUserGroupPermission keyUsrGrpPermObj) throws Exception {
		RangerPolicy policy = getPolicyObj(keyUsrGrpPermObj);

		try {
			ClientResponse response      = restClient.post(REST_POLICIES, null, policy);
			RangerPolicy   createdPolicy = response.getEntity(RangerPolicy.class);

			if (response.getStatus() == HTTPResponse.SC_OK && createdPolicy != null) {
				logInfoDetails("Created Ranger KMS policy: name='" + createdPolicy.getName() + "', id=" + createdPolicy.getId());
			} else {
				logErrorDetails("Failed to create Ranger KMS policy: key-name='" + keyUsrGrpPermObj.getKeyname() + "'. Status : " + response.getStatus());
			}
		} catch (Exception ex) {
			logErrorDetails("Error while creating policy for keyname: " + keyUsrGrpPermObj.getKeyname(), ex);

			throw new Exception("Error while creating policy for keyname: " + keyUsrGrpPermObj.getKeyname(), ex);
		}
	}

	private void updatePolicy(RangerPolicy policy, RangerKMSKeyUserGroupPermission userGroupPermissions) throws Exception {
		boolean isUpdated = updatePolicyItems(policy, userGroupPermissions);

		if (isUpdated) {
			try {
				ClientResponse response = restClient.put(REST_POLICIES + "/" + policy.getId(), null, policy);

				if (response.getStatus() == HTTPResponse.SC_OK) {
					RangerPolicy updatedPolicy = response.getEntity(RangerPolicy.class);

					logInfoDetails("Updated Ranger KMS policy: name='" + updatedPolicy.getName() + "', id=" + updatedPolicy.getId());
				} else {
					logErrorDetails("Failed to update Ranger KMS policy: name='" + policy.getName() + "', id=" + policy.getId() + ". Status : " + response.getStatus());
				}

			} catch (Exception ex) {
				logErrorDetails("Error while updating Ranger KMS policy: name='" + policy.getName() + "', id=" + policy.getId(), ex);

				throw new Exception("Error while updating Ranger KMS policy: name='" + policy.getName() + "', id=" + policy.getId(), ex);
			}
		} else {
			logInfoDetails("No update is required for Ranger KMS Policy name: '" + policy.getName() + "', id=" + policy.getId());
		}
	}

	private RangerPolicy getPolicyObj(RangerKMSKeyUserGroupPermission keyUsrGrpPermObj) {
		RangerPolicy             policy             = new RangerPolicy();
		String                   keyname            = keyUsrGrpPermObj.getKeyname();
		String                   policyName         = keyname;
		boolean                  isDenyAllElse      = true;
		boolean                  isBlackListPolicy  = false;
		String                   description        = "ACLs for key '" + keyname + "', migrated from KT-KMS";
		Map<String, Set<String>> userPermissionMap  = keyUsrGrpPermObj.getUserPermissionMap();
		Map<String, Set<String>> groupPermissionMap = keyUsrGrpPermObj.getGroupPermissionMap();

		if (StringUtils.equals(keyname, WHITELIST_RESOURCE_KEY_NAME) || StringUtils.equals(keyname, BLACKLIST_RESOURCE_KEY_NAME)) {
			if (StringUtils.equals(keyname, BLACKLIST_RESOURCE_KEY_NAME)) {
				isBlackListPolicy = true;
			}

			policy.setPolicyPriority(POLICY_PRIORITY_OVERRIDE);

			keyname       = DEFAULT_KEY_POLICY_RESOURCE;
			policyName    = GLOBAL_OVERRIDE_POLICY_NAME;
			isDenyAllElse = false;
			description   = "Global override policy, migrated from KT-KMS";
		}

		policy.setService(DEFAULT_RANGER_KMS_SERVICE_NAME);
		policy.setServiceType(RANGER_KMS_SERVICE_TYPE);
		policy.setName(policyName);
		policy.setDescription(description);
		policy.setIsEnabled(true);
		policy.setPolicyType(0);
		policy.setResources(Collections.singletonMap(RESOURCE_KEY_NAME, new RangerPolicyResource(keyname, false, false)));
		policy.setIsDenyAllElse(isDenyAllElse);

		List<RangerPolicyItem> policyItems = new ArrayList<>();

		for (Map.Entry<String, Set<String>> entry : userPermissionMap.entrySet()) {
			String           userName    = entry.getKey();
			Set<String>      permissions = entry.getValue();
			RangerPolicyItem policyItem  = new RangerPolicyItem();

			policyItem.setDelegateAdmin(false);
			policyItem.getUsers().add(userName);

			for (String accessType : permissions) {
				policyItem.getAccesses().add(new RangerPolicyItemAccess(accessType, true));
			}

			policyItems.add(policyItem);
		}

		for (Map.Entry<String, Set<String>> entry : groupPermissionMap.entrySet()) {
			String           groupName   = entry.getKey();
			Set<String>      permissions = entry.getValue();
			RangerPolicyItem policyItem  = new RangerPolicyItem();

			policyItem.setDelegateAdmin(false);
			policyItem.getGroups().add(groupName);

			for (String accessType : permissions) {
				policyItem.getAccesses().add(new RangerPolicyItemAccess(accessType, true));
			}

			policyItems.add(policyItem);
		}

		if (isBlackListPolicy) {
			policy.setDenyPolicyItems(policyItems);
		} else {
			policy.setPolicyItems(policyItems);
		}

		return policy;
	}

	private void collectUserGroupPermissions(String keyname, Map<String, Set<String>> usrPermissionMap, Map<String, Set<String>> grpPermissionMap) {
		RangerKMSKeyUserGroupPermission keyPermissions = existingUserGroupkeyMap.get(keyname);

		if (keyPermissions != null) {
			for (Map.Entry<String, Set<String>> entry : usrPermissionMap.entrySet()) {
				String      username            = entry.getKey();
				Set<String> permissions         = entry.getValue();
				Set<String> existingPermissions = keyPermissions.getUserPermissionMap().get(username);

				if (existingPermissions != null) {
					existingPermissions.addAll(permissions);
				} else {
					keyPermissions.getUserPermissionMap().put(username, permissions);
				}
			}

			for (Map.Entry<String, Set<String>> entry : grpPermissionMap.entrySet()) {
				String      groupname           = entry.getKey();
				Set<String> permissions         = entry.getValue();
				Set<String> existingPermissions = keyPermissions.getGroupPermissionMap().get(groupname);

				if (existingPermissions != null) {
					existingPermissions.addAll(permissions);
				} else {
					keyPermissions.getGroupPermissionMap().put(groupname, permissions);
				}
			}
		} else {
			keyPermissions = new RangerKMSKeyUserGroupPermission();

			if(!grpPermissionMap.isEmpty()) {
				keyPermissions.setGroupPermissionMap(grpPermissionMap);
			}

			if(!usrPermissionMap.isEmpty()) {
				keyPermissions.setUserPermissionMap(usrPermissionMap);
			}

			keyPermissions.setKeyname(keyname);

			existingUserGroupkeyMap.put(keyname, keyPermissions);
		}
	}

	private RangerPolicyItem getPolicyItemForUser(List<RangerPolicyItem> policyItems, String userName) {
		RangerPolicyItem ret = null;

		for (RangerPolicyItem item : policyItems) {
			if (item.getUsers().size() == 1 && item.getGroups().isEmpty() && item.getUsers().contains(userName)) {
				ret = item;

				break;
			}
		}

		return ret;
	}

	private RangerPolicyItem getPolicyItemForGroup(List<RangerPolicyItem> policyItems, String groupName) {
		RangerPolicyItem ret = null;

		for (RangerPolicyItem item : policyItems) {
			if (item.getUsers().isEmpty() &&  item.getGroups().size() == 1 && item.getGroups().contains(groupName)) {
				ret = item;

				break;
			}
		}

		return ret;
	}

	private boolean addAccessType(RangerPolicyItem policyItem, String accessType) {
		boolean exists = false;

		for (RangerPolicyItemAccess access : policyItem.getAccesses()) {
			if (StringUtils.equalsIgnoreCase(access.getType(), accessType)) {
				exists = true;

				break;
			}
		}

		if (!exists) {
			policyItem.getAccesses().add(new RangerPolicyItemAccess(accessType, true));
		}

		return !exists;
	}

	private boolean updatePolicyItems(RangerPolicy policy, RangerKMSKeyUserGroupPermission userGroupPermissions) {
		boolean ret = false;

		Map<String, Set<String>> userPermissionMap  = userGroupPermissions.getUserPermissionMap();
		Map<String, Set<String>> groupPermissionMap = userGroupPermissions.getGroupPermissionMap();
		List<RangerPolicyItem>   policyItems        = StringUtils.equals(userGroupPermissions.getKeyname(), BLACKLIST_RESOURCE_KEY_NAME) ? policy.getDenyPolicyItems() : policy.getPolicyItems();

		for (Map.Entry<String, Set<String>> entry : userPermissionMap.entrySet()) {
			String      userName    = entry.getKey();
			Set<String> permissions = entry.getValue();

			if (permissions == null || permissions.isEmpty()) {
				continue;
			}

			RangerPolicyItem policyItem = getPolicyItemForUser(policyItems, userName);

			if (policyItem == null) {
				policyItem = new RangerPolicyItem();

				policyItem.setDelegateAdmin(false);
				policyItem.getUsers().add(userName);

				policyItems.add(policyItem);
			}

			for (String accessType : permissions) {
				ret = addAccessType(policyItem, accessType) || ret;
			}
		}

		for (Map.Entry<String, Set<String>> entry : groupPermissionMap.entrySet()) {
			String           groupName   = entry.getKey();
			Set<String>      permissions = entry.getValue();

			if (permissions == null || permissions.isEmpty()) {
				continue;
			}

			RangerPolicyItem policyItem = getPolicyItemForGroup(policyItems, groupName);

			if (policyItem == null) {
				policyItem = new RangerPolicyItem();

				policyItem.setDelegateAdmin(false);
				policyItem.getGroups().add(groupName);

				policyItems.add(policyItem);
			}

			for (String accessType : permissions) {
				ret = addAccessType(policyItem, accessType) || ret;
			}
		}

		return ret;
	}

	private String getACLType(String property) {
		if (property.startsWith(HADOOP_KMS_BLACKLIST_PREFIX)) { // this should be prior to HADOOP_KMS_PREFIX
			return BLACKLIST_TYPE_ACL;
		} else if (property.startsWith(HADOOP_KMS_PREFIX)) {
			return SERVICE_TYPE_ACL;
		} else if (property.startsWith(DEFAULT_KEY_PREFIX)) {
			return POLICY_TYPE_ACL;
		} else if (property.startsWith(WHITELIST_KEY_PREFIX)) {
			return WHITELIST_TYPE_ACL;
		} else if (property.startsWith(KEY_ACL_PREFIX)) {
			return KEY_SPECIFIC_ACL;
		}

		return KEY_TRUSTEE_ACL;
	}

	private Set<String> createPermissionSet(String accessType) {
		switch (accessType) {
			case KEY_ACL_READ:
				return READ_PERMISSIONS;

			case KEY_ACL_DECRYPT_EEK:
				return DECRYPT_EEK_PERMISSIONS;

			case KEY_ACL_GENERATE_EEK:
				return GENERATE_EEK_PERMISSIONS;

			case KEY_ACL_MANAGEMENT:
				return ACL_MANAGEMENT_PERMISSIONS;

			case KEY_ACL_CREATE:
				return CREATE_PERMISSIONS;

			case KEY_ACL_DELETE:
				return DELETE_PERMISSIONS;

			case KEY_ACL_ROLLOVER:
				return ROLLOVER_PERMISSIONS;

			case KEY_ACL_SET_KEY_MATERIAL:
				return SET_KEY_MATERIAL_PERMISSIONS;

			case KEY_ACL_GET:
				return GET_PERMISSIONS;

			case KEY_ACL_GET_KEYS:
				return GET_KEYS_PERMISSIONS;

			case KEY_ACL_GET_METADATA:
				return GET_METADATA_PERMISSIONS;

			case KEY_ACL_ALL:
				return ALL_PERMISSIONS;
		}

		return Collections.emptySet();
	}

	private static void logErrorDetails(String message) {
		LOG.error(message);
		System.err.println(message);
	}
	
	private static void logErrorDetails(String message, Throwable excp) {
		LOG.error(message, excp);

		System.err.println(message);
		excp.printStackTrace();
	}

	private static void logInfoDetails(String message) {
		LOG.info(message);

		System.out.println(message);
	}
}
