/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.hadoop.crypto.key.kms;

import org.junit.Test;
import org.mockito.Mockito;
import org.w3c.dom.Document;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import org.apache.hadoop.crypto.key.KeyTrusteeACLToRangerPolicy;
import org.apache.hadoop.crypto.key.RangerKMSKeyUserGroupPermission;
import org.junit.Assert;
public class TestKeyTrusteeACLToRangerPolicy 
{
	
	@Test
	public void testUserGroupPermissionPolicyMapAndParseXML() throws Exception {
		
		List<String> userList = Arrays.asList("user01", "hdfs", "yarn", "hive", "oozie", "hue", "solr", 
				"user2", "systest", "hbase", "user1","defaultuser1");
		List<String> groupList = Arrays.asList("user01", "hdfs", "yarn", "hive", "oozie", "hue", "solr", 
				"user2", "systest", "hbase", "user1", "usergroup1", "usergroup2","hadoop", "public","defaultgrp1");
		Path kmsAclpath = Paths.get("./src/test/resources/kms/kms-acls.xml");
		Path dbksSitepath = Paths.get("./src/test/resources/kms/dbks-site.xml");
		
		KeyTrusteeACLToRangerPolicy k = new KeyTrusteeACLToRangerPolicy();
		KeyTrusteeACLToRangerPolicy k1 = Mockito.spy(k);
		
		Mockito.doReturn(kmsAclpath).when(k1).getKmsAclPath();
		Mockito.doReturn(dbksSitepath).when(k1).getDbksPath();
		Mockito.doReturn(userList).when(k1).getExistingUserList();
		Mockito.doReturn(groupList).when(k1).getExistingGroupList();
		Document parseDoc = k1.parseAclXmlFile();
		Map<String, RangerKMSKeyUserGroupPermission> userGrpPermMap = k1.getUserGroupMap();
		
		k1.writeToDBKSFile(parseDoc);
		
		/*Assert for Key specific ACL for key : hue*/
		RangerKMSKeyUserGroupPermission map = userGrpPermMap.get("hue");
		Assert.assertEquals(map.getKeyname(), "hue");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("oozie"));
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hue"));
		
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("oozie"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hue"));
		
		Assert.assertTrue(map.getUserPermissionMap().containsKey("oozie"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("hue"));
		
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("oozie"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("hue"));
		
		/*Assert for Key specific ACL for key : hue-key*/
		map = userGrpPermMap.get("hue-key");
		Assert.assertEquals(map.getKeyname(), "hue-key");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("oozie"));
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hue"));
		
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("oozie"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hue"));
		
		Assert.assertTrue(map.getUserPermissionMap().containsKey("oozie"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("hue"));
		
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("oozie"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("hue"));
		
		/*Assert for Key specific ACL for key : hive*/
		map = userGrpPermMap.get("hive");
		Assert.assertEquals(map.getKeyname(), "hive");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hive"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("get", "getmetadata", "getkeys", "decrypteek")), map.getGroupPermissionMap().get("hive"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("hive"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("get", "getmetadata", "getkeys", "decrypteek")), map.getUserPermissionMap().get("hive"));
		
		/*Assert for Key specific ACL for key : hive-key*/
		map = userGrpPermMap.get("hive-key");
		Assert.assertEquals(map.getKeyname(), "hive-key");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hive"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("get", "getmetadata", "getkeys", "decrypteek")), map.getGroupPermissionMap().get("hive"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("hive"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("get", "getmetadata", "getkeys", "decrypteek")), map.getUserPermissionMap().get("hive"));
		
		/*Assert for Key specific ACL for key : user1-key*/
		map = userGrpPermMap.get("user1-key");
		Assert.assertEquals(map.getKeyname(), "user1-key");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hive"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hive"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("usergroup1"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("usergroup1"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hbase"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hbase"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("solr"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("solr"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("oozie"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("oozie"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hue"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hue"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("yarn"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("yarn"));
		
		Assert.assertTrue(map.getUserPermissionMap().containsKey("user1"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("user1"));
		
		/*Assert for Key specific ACL for key : hbase*/
		map = userGrpPermMap.get("hbase");
		Assert.assertEquals(map.getKeyname(), "hbase");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hbase"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hbase"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("hbase"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("hbase"));
		
		/*Assert for Key specific ACL for key : hbase-key*/
		map = userGrpPermMap.get("hbase-key");
		Assert.assertEquals(map.getKeyname(), "hbase-key");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hbase"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hbase"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("hbase"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("hbase"));
		
		/*Assert for Key specific ACL for key : solr*/
		map = userGrpPermMap.get("solr");
		Assert.assertEquals(map.getKeyname(), "solr");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("solr"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("solr"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("solr"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("solr"));
		
		/*Assert for Key specific ACL for key : solr-key*/
		map = userGrpPermMap.get("solr-key");
		Assert.assertEquals(map.getKeyname(), "solr-key");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("solr"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("solr"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("solr"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("solr"));
		
		/*Assert for Key specific ACL for key : user2-key*/
		map = userGrpPermMap.get("user2-key");
		Assert.assertEquals(map.getKeyname(), "user2-key");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hive"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hive"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("usergroup1"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("usergroup1"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("usergroup2"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("usergroup2"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hbase"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hbase"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("solr"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("solr"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("oozie"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("oozie"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hue"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("hue"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("yarn"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("yarn"));
		
		Assert.assertTrue(map.getUserPermissionMap().containsKey("user2"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("user2"));
		
		/*Assert for whitelist Key ACL */
		map = userGrpPermMap.get("***whitelist***");
		Assert.assertEquals(map.getKeyname(), "***whitelist***");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hadoop"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("get", "getmetadata", "getkeys","generateeek")), map.getGroupPermissionMap().get("hadoop"));
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("systest"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("create", "rollover", "delete","decrypteek")), map.getGroupPermissionMap().get("systest"));
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("usergroup1"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("create", "rollover", "delete","decrypteek")), map.getGroupPermissionMap().get("usergroup1"));
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("usergroup2"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("create", "rollover", "delete","decrypteek")), map.getGroupPermissionMap().get("usergroup2"));
		
		Assert.assertTrue(map.getUserPermissionMap().containsKey("hdfs"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("get", "getmetadata", "getkeys","generateeek")), map.getUserPermissionMap().get("hdfs"));
		Assert.assertTrue(map.getUserPermissionMap().containsKey("systest"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("create", "rollover", "delete","decrypteek")), map.getUserPermissionMap().get("systest"));
		
		/*Assert for blacklist Key ACL */
		map = userGrpPermMap.get("***blacklist***");
		Assert.assertEquals(map.getKeyname(), "***blacklist***");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("hadoop"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("create", "rollover", "decrypteek", "delete")), map.getGroupPermissionMap().get("hadoop"));
		
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("public"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("get", "setkeymaterial")), map.getGroupPermissionMap().get("public"));
		
		Assert.assertTrue(map.getUserPermissionMap().containsKey("hdfs"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("create", "rollover", "decrypteek", "delete")), map.getUserPermissionMap().get("hdfs"));
		
		/*Assert for default Key ACL */
		map = userGrpPermMap.get("*");
		Assert.assertEquals(map.getKeyname(), "*");
		Assert.assertTrue(map.getGroupPermissionMap().containsKey("defaultgrp1"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getGroupPermissionMap().get("defaultgrp1"));
		
		Assert.assertTrue(map.getUserPermissionMap().containsKey("defaultuser1"));
		Assert.assertEquals(new HashSet<String>(Arrays.asList("decrypteek")), map.getUserPermissionMap().get("defaultuser1"));
		
	}
}
