/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.services.adls;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyItem;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyResource;
import org.apache.ranger.plugin.model.RangerService;
import org.apache.ranger.plugin.model.RangerServiceDef;
import org.apache.ranger.plugin.service.RangerBaseService;
import org.apache.ranger.plugin.service.ResourceLookupContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.reflect.TypeToken;

/**
 * RangerService for ADLS (Azure Data Lake Storage).
 */
public class RangerServiceADLS extends RangerBaseService {
    private static final Logger LOG = LoggerFactory.getLogger(RangerServiceADLS.class);

    public static final String STORAGE_ACCOUNT             = "storageaccount";
    public static final String CONTAINER                   = "container";
    public static final String RELATIVE_PATH               = "relativepath";
    public static final String ABFS_PROTOCOL               = "abfs://";
    public static final String CONTAINER_ACCOUNT_SEPARATOR = "@";
    public static final String ABFS_FILE_ACCOUNT_SUFFIX    = ".dfs.core.windows.net";
    public static final String ADLS_DEFAULT_POLICY_FILE    = "ranger-adls-default-policies.json";
    private static final String RELATIVE_PATH_ROOT         = "/";

    public static final String CONF_ADLS_DATA_LOC             = "adls.url.data";
    public static final String CONF_ADLS_LOG_LOC              = "adls.url.logs";
    public static final String CONF_ADLS_RANGER_AUDIT_LOC     = "adls.url.audits.ranger";
    public static final String CONF_ADLS_SETUP_DEFAULT_POLICY = "setup.additional.adls.default.policies";
    private static final String LOCATION_PREFIX               = "cloud.storage.url.";

    private static final String STR_RANGER_AUDIT                      = "RANGER_AUDIT";
    private static final String STR_HIVE_METASTORE_WAREHOUSE          = "HIVE_METASTORE_WAREHOUSE";
    private static final String STR_HIVE_REPLICA_WAREHOUSE            = "HIVE_REPLICA_WAREHOUSE";
    private static final String STR_HIVE_METASTORE_EXTERNAL_WAREHOUSE = "HIVE_METASTORE_EXTERNAL_WAREHOUSE";
    private static final String STR_DATA                              = "DATA";
    private static final String STR_LOG                               = "LOG";
    private static final String STR_ACCOUNT                           = "_ACCOUNT";
    private static final String STR_CONTAINER                         = "_CONTAINER";
    private static final String STR_RELATIVE_PATH                     = "_RELATIVE_PATH";
    private static final String STR_BOOLEAN_TRUE                      = "true";
    private static final String STR_OPERATOR_EQUAL                    = "=";
    private static final String STR_RANGER_AUDIT_PATH                 = "/ranger/audit";
    private static final String STR_HBASE_ROOT                        = "HBASE_ROOT";

    private static final Pattern TEMPLATE_TOKEN_PATTERN = Pattern.compile("\\{\\{([a-zA-Z_=\\.\\*/-]*)\\}\\}");
    private final Map<String, AdlsLocation> adlsLocationMap = new HashMap<>();

    @Override
    public HashMap<String, Object> validateConfig() throws Exception {
        // TODO: implement configure validation for ADLS policies.
        return new HashMap<>();
    }

    @Override
    public List<String> lookupResource(ResourceLookupContext context) throws Exception {
        // TODO: implement resource lookup for ADLS policies.
        return new ArrayList<>();
    }

    @Override
    public void init(RangerServiceDef serviceDef, RangerService service) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.init({}, {})", serviceDef, service);
        }

        super.init(serviceDef, service);

        Map<String, String> serviceConfig = service.getConfigs();
        String auditLocStr = null;

        for (Map.Entry<String, String> entry : serviceConfig.entrySet()) {
            if (entry.getKey().startsWith(LOCATION_PREFIX)) {
                String keyForMap = entry.getKey().substring(LOCATION_PREFIX.length());
                if (StringUtils.isNotBlank(keyForMap)) {
                    adlsLocationMap.put(keyForMap, new AdlsLocation(entry.getValue()));
                    if (StringUtils.equals(keyForMap, STR_RANGER_AUDIT)) {
                        auditLocStr = entry.getValue();
                    }
                }
            }
        }

        if (auditLocStr == null) {
            auditLocStr = serviceConfig.get(CONF_ADLS_RANGER_AUDIT_LOC);
        }

        String auditLocWithoutRelativePath = null;

        if (StringUtils.isNotBlank(auditLocStr)) {
            int auditRelativePathIndex = auditLocStr.indexOf(STR_RANGER_AUDIT_PATH);
            if (auditRelativePathIndex < 0) { // Audit path does not contains "/ranger/audit"
                auditRelativePathIndex = auditLocStr.indexOf(RELATIVE_PATH_ROOT, auditLocStr.indexOf(ABFS_FILE_ACCOUNT_SUFFIX) + (ABFS_FILE_ACCOUNT_SUFFIX.length()));
            }

            if (auditRelativePathIndex > 0) {
                auditLocWithoutRelativePath = auditLocStr.substring(0, auditRelativePathIndex);
            } else {
                auditLocWithoutRelativePath = auditLocStr ;
            }
        }

        // Add IF RANGER_AUDIT is missing && adlsLocataion is not BLANK
        addMissingAdlsLocation(STR_RANGER_AUDIT, auditLocStr);

        // Add IF DATA is missing && adlsLocataion is not BLANK
        if (!adlsLocationMap.containsKey(STR_DATA)) {
            String dataLocStr  = serviceConfig.get(CONF_ADLS_DATA_LOC);
            if (StringUtils.isBlank(dataLocStr)) {
                dataLocStr = auditLocWithoutRelativePath;
            }
            addMissingAdlsLocation(STR_DATA, dataLocStr);
        }

        // Add IF LOG is missing && adlsLocataion is not BLANK
        addMissingAdlsLocation(STR_LOG, serviceConfig.get(CONF_ADLS_LOG_LOC));

        if (StringUtils.isNotBlank(auditLocWithoutRelativePath)) {
            // Add IF HIVE_METASTORE_WAREHOUSE is missing && adlsLocataion is not BLANK
            addMissingAdlsLocation(STR_HIVE_METASTORE_WAREHOUSE, auditLocWithoutRelativePath + "/warehouse/tablespace/managed");

            // Add IF HIVE_REPLICA_WAREHOUSE is missing && adlsLocataion is not BLANK
            addMissingAdlsLocation(STR_HIVE_REPLICA_WAREHOUSE, auditLocWithoutRelativePath + "/hive_replica_functions_dir");

            // Add IF HIVE_METASTORE_EXTERNAL_WAREHOUSE is missing && adlsLocataion is not BLANK
            addMissingAdlsLocation(STR_HIVE_METASTORE_EXTERNAL_WAREHOUSE, auditLocWithoutRelativePath + "/warehouse/tablespace/external");

            // Add IF HBASE_ROOT is missing && adlsLocataion is not BLANK
            addMissingAdlsLocation(STR_HBASE_ROOT, auditLocWithoutRelativePath + "/hbase");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.init(): adlsLocationMap={}", adlsLocationMap);
        }
    }

    @Override
    public List<RangerPolicy> getDefaultRangerPolicies() throws Exception {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.getDefaultRangerPolicies()");
        }

        final List<RangerPolicy> ret                     = super.getDefaultRangerPolicies();
        final Boolean            setupAdditionalPolicies = Boolean.valueOf(configs.getOrDefault(CONF_ADLS_SETUP_DEFAULT_POLICY, STR_BOOLEAN_TRUE));

        if (setupAdditionalPolicies && areRequiredConfigsPresent()) {
            // Fetch policies from JSON
            InputStream       inStream = getClass().getClassLoader().getResourceAsStream(ADLS_DEFAULT_POLICY_FILE);
            InputStreamReader reader   = new InputStreamReader(inStream);

            readPoliciesFromJsonFile(reader, ret);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.getDefaultRangerPolicies(): ret={}", ret);
        }

        return ret;
    }

    private boolean areRequiredConfigsPresent() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.areRequiredConfigsPresent()");
        }

        boolean ret = areAdlsLocationsValid(adlsLocationMap);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.areRequiredConfigsPresent(): ret={}", ret);
        }

        return ret;
    }

    private boolean areAdlsLocationsValid(Map<String, AdlsLocation> adlsLocationMapParam) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.isAdlsLocationValid({})", adlsLocationMapParam);
        }

//        boolean ret = adlsLocationMapParam.size() > 0 && adlsLocationMapParam.get(STR_RANGER_AUDIT) != null;
        boolean ret = true; //FIXME currently CDEP clusters, do not send RANGER AUDIT location

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.isAdlsLocationValid({}): {}", adlsLocationMapParam, ret);
        }

        return ret;
    }

    private void readPoliciesFromJsonFile(Reader reader, List<RangerPolicy> ret) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.readPoliciesFromJsonFile({}, {})", reader, ret);
        }

        List<RangerPolicy>  transformedPolicies = new ArrayList<>();
        List<RangerPolicy>  defaultPolicies     = readJson(reader);
        Map<String, String> adlsLocationConfig  = getAdlsLocationMap();

        if (!defaultPolicies.isEmpty()) {
            for (RangerPolicy policy : defaultPolicies) {
                RangerPolicy policyTransformed = transformPolicy(policy, adlsLocationConfig);

                if (policyTransformed != null) {
                    transformedPolicies.add(policyTransformed);
                }
            }
        }

        if (!transformedPolicies.isEmpty()) {
            ret.addAll(transformedPolicies);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.readPoliciesFromJsonFile(): ret={}", ret);
        }
    }

    private List<RangerPolicy> readJson(Reader reader) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.readJson({})", reader);
        }

        List<RangerPolicy> policies = new ArrayList<>();

        if (reader != null) {
            Gson gson     = new GsonBuilder().setPrettyPrinting().create();
            Type listType = new TypeToken<ArrayList<RangerPolicy>>(){}.getType();

            policies = gson.fromJson(reader, listType);
        } else {
            LOG.warn("===>>> File "+reader+", does not exits.");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.readJson(): policies={}", policies);
        }

        return policies;
    }

    private RangerPolicy transformPolicy(RangerPolicy policy, Map<String, String> adlsLocationConfig) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.transformPolicy({}, {})", policy, adlsLocationConfig);
        }

        RangerPolicy                      ret              = null;
        Map<String, RangerPolicyResource> resourceMap      = policy.getResources();
        RangerPolicyResource              storageAccounts  = resourceMap.get(STORAGE_ACCOUNT);
        RangerPolicyResource              containers       = resourceMap.get(CONTAINER);
        RangerPolicyResource              relativePaths    = resourceMap.get(RELATIVE_PATH);
        List<String>                      accountList      = parseValuesFromTemplate(storageAccounts.getValues(), adlsLocationConfig);
        List<String>                      containerList    = parseValuesFromTemplate(containers.getValues(), adlsLocationConfig);
        List<String>                      relativePathList = parseValuesFromTemplate(relativePaths.getValues(), adlsLocationConfig);

        if (CollectionUtils.isNotEmpty(accountList) || CollectionUtils.isNotEmpty(containerList) || CollectionUtils.isNotEmpty(relativePathList)) {
            policy.setService(service.getName());

            storageAccounts.setValues(accountList);
            containers.setValues(containerList);
            relativePaths.setValues(relativePathList);

            for (RangerPolicyItem policyItem : policy.getPolicyItems()) {
                List<String> users  = parseValuesFromTemplate(policyItem.getUsers(), service.getConfigs());
                List<String> groups = parseValuesFromTemplate(policyItem.getGroups(), service.getConfigs());
                List<String> roles  = parseValuesFromTemplate(policyItem.getRoles(), service.getConfigs());

                policyItem.setUsers(users);
                policyItem.setGroups(groups);
                policyItem.setRoles(roles);
            }

            ret = policy;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.transformPolicy(): ret={}", ret);
        }

        return ret;
    }

    private static class AdlsLocation {
        private final String storageAccount;
        private final String container;
        private final String relativePath;

        public AdlsLocation(String location) {
            if (StringUtils.contains(location, ABFS_PROTOCOL) &&
                StringUtils.contains(location, CONTAINER_ACCOUNT_SEPARATOR) &&
                StringUtils.contains(location, ABFS_FILE_ACCOUNT_SUFFIX)) {
                this.storageAccount = location.substring(location.indexOf(CONTAINER_ACCOUNT_SEPARATOR) + CONTAINER_ACCOUNT_SEPARATOR.length(), location.indexOf(ABFS_FILE_ACCOUNT_SUFFIX));
                this.container      = location.substring(location.indexOf(ABFS_PROTOCOL) + ABFS_PROTOCOL.length(), location.indexOf(CONTAINER_ACCOUNT_SEPARATOR));

                int    idxRelativePath = location.indexOf(RELATIVE_PATH_ROOT, location.indexOf(ABFS_FILE_ACCOUNT_SUFFIX) + (ABFS_FILE_ACCOUNT_SUFFIX.length()));
                String relativePath    = (idxRelativePath > 0) ? location.substring(idxRelativePath) : RELATIVE_PATH_ROOT;

                if (relativePath.length() > 1 && relativePath.endsWith(RELATIVE_PATH_ROOT)) {
                    this.relativePath = relativePath.substring(0, relativePath.length() - 1);
                } else {
                    this.relativePath = relativePath;
                }
            } else {
                this.storageAccount = null;
                this.container      = null;
                this.relativePath   = null;
            }
        }

        public String getStorageAccount() {
            return storageAccount;
        }

        public String getContainer() {
            return container;
        }

        public String getRelativePath() {
            return relativePath;
        }

        public boolean isRelativePathNotEmpty() {
            return StringUtils.isNotBlank(relativePath) && relativePath.length() > 1;
        }

        @Override
        public String toString() {
            return "{storageAccount="+storageAccount+", container="+container+", relativePath="+relativePath+"}";
        }
    }

    private Map<String, String> getAdlsLocationMap() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.getAdlsLocationMap()");
        }

        Map<String, String> ret = new HashMap<>();
        for(Map.Entry<String, AdlsLocation> entry : adlsLocationMap.entrySet()) {
            AdlsLocation adlsLocation = entry.getValue();
            addValueIfNotNull(ret, entry.getKey() + STR_ACCOUNT,   adlsLocation.getStorageAccount());
            addValueIfNotNull(ret, entry.getKey() + STR_CONTAINER, adlsLocation.getContainer());

            if (adlsLocation.isRelativePathNotEmpty()) {
                addValueIfNotNull(ret, entry.getKey() + STR_RELATIVE_PATH, adlsLocation.getRelativePath());
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.getAdlsLocationMap(): ret={}", ret);
        }

        return ret;
    }

    /**
     * This method, if present replace first (single) token with appropriate value.
     * @param inputList
     * @param config
     * @return final value after replacement
     */
    private List<String> parseValuesFromTemplate(List<String> inputList, Map<String, String> config) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceADLS.parseValuesFromTemplate({}, {})", inputList, config);
        }

        List<String> ret = new ArrayList<>();
        Set<String> uniqueValues = new HashSet<>();

        if (inputList != null) {

            for (String input : inputList) {
                Matcher matcher = TEMPLATE_TOKEN_PATTERN.matcher(input);

                // example input: {{service.user.hive=hive}}, {{service.user.hive}}, hive
                if (matcher.find()) {
                    String   keyWithDefaultValue = matcher.group(1);
                    String[] keyValue            = keyWithDefaultValue.split(STR_OPERATOR_EQUAL);
                    String   value               = config.get(keyValue[0]);

                    if (StringUtils.isNotBlank(value)) {
                        input = matcher.replaceFirst(value);
                    } else if (keyValue.length > 1 && StringUtils.isNotBlank(keyValue[1])) {
                        input = matcher.replaceFirst(keyValue[1]);
                    } else {
                        input = null; //token exists, but nothing matches
                    }
                }

                if (StringUtils.isNotBlank(input)) {
                    uniqueValues.add(input);
                }
            }
        }
        ret.addAll(uniqueValues);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceADLS.parseValuesFromTemplate(): ret={}", ret);
        }

        return ret;
    }

    private void addValueIfNotNull(Map<String, String> map, String key, String value) {
        if (value != null) {
            map.put(key, value);
        }
    }

    private void addMissingAdlsLocation(String key, String adlsLocation) {
        if (StringUtils.isNotBlank(adlsLocation)) {
            adlsLocationMap.putIfAbsent(key, new AdlsLocation(adlsLocation));
        }
    }
}
