
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.authorization.ozone.authorizer;

import com.google.common.collect.Sets;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.ozone.om.helpers.OzoneFileStatus;
import org.apache.hadoop.ozone.security.acl.IAccessAuthorizer;
import org.apache.hadoop.ozone.security.acl.IOzoneObj;
import org.apache.hadoop.ozone.security.acl.OzoneObj;
import org.apache.hadoop.ozone.security.acl.OzonePrefixPath;
import org.apache.hadoop.ozone.security.acl.RequestContext;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.audit.model.AuthzAuditEvent;
import org.apache.ranger.audit.provider.MiscUtil;
import org.apache.ranger.plugin.audit.RangerDefaultAuditHandler;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.policyengine.RangerAccessRequestImpl;
import org.apache.ranger.plugin.policyengine.RangerAccessResource;
import org.apache.ranger.plugin.policyengine.RangerAccessResourceImpl;
import org.apache.ranger.plugin.policyengine.RangerAccessResult;
import org.apache.ranger.plugin.policyengine.RangerPolicyEngine;
import org.apache.ranger.plugin.policyengine.RangerPolicyEngineImpl;
import org.apache.ranger.plugin.service.RangerBasePlugin;
import org.apache.ranger.plugin.util.RangerPerfTracer;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class RangerOzoneAuthorizer implements IAccessAuthorizer {
	public static final String ACCESS_TYPE_READ = "read";
	public static final String ACCESS_TYPE_WRITE = "write";
	public static final String ACCESS_TYPE_CREATE = "create";
	public static final String ACCESS_TYPE_LIST = "list";
	public static final String ACCESS_TYPE_DELETE = "delete";
	public static final String ACCESS_TYPE_READ_ACL = "read_acl";
	public static final String ACCESS_TYPE_WRITE_ACL = "write_acl";
	public static final String WILDCARD_ASTERISK = "*";
	public static final String WILDCARD_QUESTIONMARK = "?";
	public static final String pathSeparatorChar = "/";


    public static final String KEY_RESOURCE_VOLUME = "volume";
	public static final String KEY_RESOURCE_BUCKET = "bucket";
	public static final String KEY_RESOURCE_KEY = "key";
	private static String OPTIMIZED_SUBACCESSPATH_ENABLED = ".optimized.subaccesspath.enabled";

	private static final Log PERF_OZONEAUTH_REQUEST_LOG = RangerPerfTracer.getPerfLogger("ozoneauth.request");

    private static final Log LOG = LogFactory.getLog(RangerOzoneAuthorizer.class);

	private static volatile RangerOzonePlugin rangerPlugin = null;
	RangerDefaultAuditHandler auditHandler = null;
	Boolean isOptimizedSubAccessPathEnabled;

	public RangerOzoneAuthorizer() {
		rangerPlugin = new RangerOzonePlugin();

		try {
			rangerPlugin.init(); // this will initialize policy engine and policy refresher
		} catch (Throwable t) {
			LOG.fatal("Error creating and initializing RangerBasePlugin()", t);
		}
		auditHandler = new RangerDefaultAuditHandler();
		rangerPlugin.setResultProcessor(auditHandler);
		isOptimizedSubAccessPathEnabled = rangerPlugin.getConfig().getBoolean(
				rangerPlugin.getConfig().getPropertyPrefix()+OPTIMIZED_SUBACCESSPATH_ENABLED, false);
	}

	@Override
	public boolean checkAccess(IOzoneObj ozoneObject, RequestContext context) {
		boolean returnValue = false;
		if (ozoneObject == null) {
			LOG.error("Ozone object is null!!");
			return returnValue;
		}
		OzoneObj ozoneObj = (OzoneObj) ozoneObject;
		UserGroupInformation ugi = context.getClientUgi();
		ACLType operation = context.getAclRights();
		String resource = ozoneObj.getPath();

		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RangerOzoneAuthorizer.checkAccess with operation = " + operation + ", resource = " +
					resource + ", store type = " + OzoneObj.StoreType.values() + ", ugi = " + ugi + ", ip = " +
					context.getIp() + ", resourceType = " + ozoneObj.getResourceType() + ", owner = " + context.getOwnerName() +
					" isRecursiveAccessCheck = " + context.isRecursiveAccessCheck() + ")");
		}

		if (rangerPlugin == null) {
			MiscUtil.logErrorMessageByInterval(LOG,
					"Authorizer is still not initialized");
			return returnValue;
		}

		//TODO: If sorce type is S3 and resource is volume, then allow it by default
		if (ozoneObj.getStoreType() == OzoneObj.StoreType.S3 && ozoneObj.getResourceType() == OzoneObj.ResourceType.VOLUME) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("If store type is s3 and resource is volume, then we allow it by default!  Returning true");
			}
			LOG.warn("Allowing access by default since source type is S3 and resource type is Volume!!");
			return true;
		}

		RangerPerfTracer perf = null;

		if (RangerPerfTracer.isPerfTraceEnabled(PERF_OZONEAUTH_REQUEST_LOG)) {
			perf = RangerPerfTracer.getPerfTracer(PERF_OZONEAUTH_REQUEST_LOG, "RangerOzoneAuthorizer.authorize(resource=" + resource + ")");
		}

		Date eventTime = new Date();
		String accessType = mapToRangerAccessType(operation);
		if (accessType == null) {
			MiscUtil.logErrorMessageByInterval(LOG,
					"Unsupported access type. operation=" + operation) ;
			LOG.fatal("Unsupported access type. operation=" + operation + ", resource=" + resource);
			return returnValue;
		}
		String action = accessType;
		String clusterName = rangerPlugin.getClusterName();

		RangerAccessRequestImpl rangerRequest = new RangerAccessRequestImpl();
		rangerRequest.setUser(ugi.getShortUserName());
		rangerRequest.setUserGroups(Sets.newHashSet(ugi.getGroupNames()));
		rangerRequest.setClientIPAddress(context.getIp().getHostAddress());
		rangerRequest.setRemoteIPAddress(context.getIp().getHostAddress());
		rangerRequest.setAccessTime(eventTime);

		RangerAccessResourceImpl rangerResource = new RangerAccessResourceImpl();
		rangerRequest.setResource(rangerResource);
		rangerRequest.setAccessType(accessType);
		rangerRequest.setAction(action);
		rangerRequest.setRequestData(resource);
		rangerRequest.setClusterName(clusterName);
		rangerResource.setOwnerUser(context.getOwnerName());

		if (ozoneObj.getResourceType() == OzoneObj.ResourceType.VOLUME) {
			rangerResource.setValue(KEY_RESOURCE_VOLUME, ozoneObj.getVolumeName());
		} else if (ozoneObj.getResourceType() == OzoneObj.ResourceType.BUCKET || ozoneObj.getResourceType() == OzoneObj.ResourceType.KEY) {
			if (ozoneObj.getStoreType() == OzoneObj.StoreType.S3) {
				rangerResource.setValue(KEY_RESOURCE_VOLUME, "s3Vol");
			} else {
				rangerResource.setValue(KEY_RESOURCE_VOLUME, ozoneObj.getVolumeName());
			}
			rangerResource.setValue(KEY_RESOURCE_BUCKET, ozoneObj.getBucketName());
			if (ozoneObj.getResourceType() == OzoneObj.ResourceType.KEY) {
				rangerResource.setValue(KEY_RESOURCE_KEY, ozoneObj.getKeyName());
			}
		} else {
			LOG.fatal("Unsupported resource = " + resource);
			MiscUtil.logErrorMessageByInterval(LOG, "Unsupported resource type " + ozoneObj.getResourceType() + " for resource = " + resource
					+ ", request=" + rangerRequest);
			return returnValue;
		}

		try {
				if (!context.isRecursiveAccessCheck()) {
					RangerAccessResult result = rangerPlugin
							.isAccessAllowed(rangerRequest);
					if (result == null) {
						LOG.error("Ranger Plugin returned null. Returning false");
					} else {
						returnValue = result.getIsAllowed();
					}
				} else {
					if (LOG.isDebugEnabled()) {
						LOG.debug("Evaluating policies for recursive delete with isOptimizedSubAccessPathEnabled set to " +  isOptimizedSubAccessPathEnabled);
						LOG.debug("Resource scope = " + rangerRequest.getResourceMatchingScope());
					}
					if (isOptimizedSubAccessPathEnabled) {
						returnValue = isAccessAllowedForSubPathsOptimized(ozoneObj, rangerResource, rangerRequest);
					} else {
						returnValue = isAccessAllowedForSubPaths(ozoneObj, rangerResource, rangerRequest);
					}
				}
		} catch (Throwable t) {
			LOG.error("Error while calling isAccessAllowed(). request="
					+ rangerRequest, t);
		}
		RangerPerfTracer.log(perf);

		if (LOG.isDebugEnabled()) {
			LOG.debug("rangerRequest=" + rangerRequest + ", return="
					+ returnValue);
		}
		return returnValue;
	}

	private String mapToRangerAccessType(ACLType operation) {
		String rangerAccessType = null;
		switch (operation) {
			case READ:
				rangerAccessType = ACCESS_TYPE_READ;
				break;
			case WRITE:
				rangerAccessType = ACCESS_TYPE_WRITE;
				break;
			case CREATE:
				rangerAccessType = ACCESS_TYPE_CREATE;
				break;
			case DELETE:
				rangerAccessType = ACCESS_TYPE_DELETE;
				break;
			case LIST:
				rangerAccessType = ACCESS_TYPE_LIST;
				break;
			case READ_ACL:
				rangerAccessType = ACCESS_TYPE_READ_ACL;
				break;
			case WRITE_ACL:
				rangerAccessType = ACCESS_TYPE_WRITE_ACL;
				break;
		}
		return rangerAccessType;
	}

	private boolean isValidSubPath(String subPath, String parentDirPath) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("subPath = " + subPath + " parentDirPath = " + parentDirPath);
		}
		// Subpaths are computed by evaluating likely matched policies for a given parent path with Self and immediate Child scope.
		// Since Ranger supports wildcards in the policy resources, we want to ignore cases where resource has only wildcards ("*" or "?")
		if ((StringUtils.equals(subPath,WILDCARD_ASTERISK) || StringUtils.equals(subPath, WILDCARD_QUESTIONMARK) || StringUtils.equals(subPath,parentDirPath))) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("Skipping invalid subPath = " + subPath);
			}
			return  false;
		}
		return isSubDirForParentDir(subPath, parentDirPath);

	}

	/**
	 * returns true if sub directory path pattern matches the given parent path
	 *
	 * @param subDirPath
	 * @param parentDirPath
	 * @return
	 */

	private boolean isSubDirForParentDir(String subDirPath, String parentDirPath) {
		String subDirPathUpdated = subDirPath.replace(WILDCARD_QUESTIONMARK, "."+WILDCARD_QUESTIONMARK).replace(WILDCARD_ASTERISK, "."+WILDCARD_ASTERISK);
		if (LOG.isDebugEnabled()) {
			LOG.debug("updated subDirPathUpdated = " + subDirPathUpdated);
		}

		String[] parentSplits = parentDirPath.split(pathSeparatorChar);
		String[] childSplits = subDirPathUpdated.split(pathSeparatorChar);
		if (childSplits.length > parentSplits.length) {
			for (int i=0; i<parentSplits.length; i++) {
				Pattern pattern = Pattern.compile(childSplits[i]);
				Matcher matcher = pattern.matcher(parentSplits[i]);
				if (!matcher.matches()) {
					if (LOG.isDebugEnabled()) {
						LOG.debug("childSplit[" + i + "] = " + childSplits[i] + " parentSplit[" + i + "] = " + parentSplits[i] + " don't match");
					}
					return false;
				}
			}
		} else {
			return false;
		}
		return true;
	}

	private Set<String> getSubPathsFromMatchingPolicies(RangerAccessResource resource) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> getSubPathsFromMatchingPolicies(" + resource + ")");
		}

		Set<String> policyResources = new HashSet<>();

		List<RangerPolicy> matchingPolicies = rangerPlugin.getMatchingPoliciesForResource(resource);

		for (RangerPolicy rangerPolicy : matchingPolicies) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("rangerPolicy " + rangerPolicy.getName() + ": " + rangerPolicy.getResources().values());
			}
			// Ranger policy resources contain Volume, Bucket, and Key. Checking only the policies that have Key resources and retrieve subpaths from the key resource.
			if (rangerPolicy.getResources().values().size() > 2) {
				RangerPolicy.RangerPolicyResource policyResource = (RangerPolicy.RangerPolicyResource) CollectionUtils.get(rangerPolicy.getResources().values(), 2);
				policyResources.addAll(policyResource.getValues());
			}
		}
		if (LOG.isDebugEnabled()) {
			LOG.debug("<==  getSubPathsFromMatchingPolicies(" + resource + ", " + ") : " + policyResources.size());
		}
		return policyResources;
	}

	// isAccessAllowedForSubPathsOptimized() recursively checks all the matching key resources or sub directory paths to determine if the delete is allowed on the parent directory.
	private boolean isAccessAllowedForSubPathsOptimized(OzoneObj ozoneObj, RangerAccessResourceImpl rangerResource, RangerAccessRequestImpl rangerRequest) {
		boolean returnValue = false;
		String dirPath = ozoneObj.getKeyName();
		if (LOG.isDebugEnabled()) {
			LOG.debug("Evaluating policies for " + dirPath );
		}
		rangerResource.setValue(KEY_RESOURCE_KEY, dirPath);
		RangerAccessResult result;
		result = rangerPlugin.isAccessAllowed(rangerRequest);
		if (result == null || !result.getIsAllowed()) {
			returnValue = false;
			LOG.error("Deleting " + dirPath + " is not allowed. Hence returning false");
		} else {
			returnValue = true;
			Set<String> subPaths = getSubPathsFromMatchingPolicies(rangerResource);
			if (LOG.isDebugEnabled()) {
				LOG.debug("All sub paths for " + rangerResource.getValue(KEY_RESOURCE_KEY) + " = " + subPaths);
			}
			if (CollectionUtils.isNotEmpty(subPaths)) {
				Iterator<? extends String> cList = subPaths.iterator();

				while (cList.hasNext()) {
					String subPath = cList.next();
					if (isValidSubPath(subPath, dirPath)) {
						if (LOG.isDebugEnabled()) {
							LOG.debug("Added " + subPath + " for policy evaluation");
						}
						rangerResource.setValue(KEY_RESOURCE_KEY, subPath);
						RangerDefaultAuditHandler auditHandler = new RangerOzoneAuditHandler();
						result = rangerPlugin.isAccessAllowed(rangerRequest, auditHandler);
						if (result == null || !result.getIsAllowed()) {
							if (result.getPolicyId() != -1) {
								LOG.error("Deleting " + dirPath + " is not allowed. Hence returning false");
								returnValue = false;
								break;
							} else {
								returnValue = false;
							}
						}
					}
				}
			}
		}

		return returnValue;
	}

	// isAccessAllowedForSubPaths() recursively checks all the matching key resources or sub directory paths to determine if the delete is allowed on the parent directory.
	private boolean isAccessAllowedForSubPaths(OzoneObj ozoneObj, RangerAccessResourceImpl rangerResource, RangerAccessRequestImpl rangerRequest) throws IOException {
		boolean returnValue = false;
		if (ozoneObj.getOzonePrefixPathViewer() != null) {
			OzonePrefixPath ozonePrefixPath = ozoneObj.getOzonePrefixPathViewer();

			Stack<OzoneFileStatus> directories = new Stack<>();

			directories.push(ozonePrefixPath.getOzoneFileStatus());
			while (!directories.isEmpty()) {
				OzoneFileStatus ozoneFileStatus = directories.pop();
				if (ozoneFileStatus == null) {
					break;
				}
				String dirPath = ozoneFileStatus.getTrimmedName();
				if (LOG.isDebugEnabled()) {
					LOG.debug("OzonePrefixPath = " + dirPath +
							" isDirectory = " + ozoneFileStatus.isDirectory());
				}

				rangerResource.setValue(KEY_RESOURCE_KEY, dirPath);
				RangerAccessResult result = rangerPlugin.isAccessAllowed(rangerRequest);

				if (result == null || !result.getIsAllowed()) {
					LOG.error("Deleting " + dirPath + " is not allowed. Hence returning false");
					returnValue = false;
					break;
				} else {
					returnValue = true;
				}

				if (ozoneFileStatus.isDirectory()) {
					Iterator<? extends OzoneFileStatus> cList = ozonePrefixPath.getChildren(dirPath);

					while (cList.hasNext()) {
						OzoneFileStatus child = cList.next();
						if (child.isDirectory()) {
							directories.push(child);
						} else {
							if (LOG.isDebugEnabled()) {
								LOG.debug("child.getTrimmedName() = " + child.getTrimmedName());
							}
							rangerResource.setValue(KEY_RESOURCE_KEY, child.getTrimmedName());
							result = rangerPlugin.isAccessAllowed(rangerRequest);

							if (result == null || !result.getIsAllowed()) {
								LOG.error("Deleting " + child.getTrimmedName() + " is not allowed. Hence returning false");
								returnValue = false;
								break;
							} else {
								returnValue = true;
							}
						}
					}
				}
			}
		}
		return returnValue;
	}
}

class RangerOzonePlugin extends RangerBasePlugin {
	private static final Log LOG = LogFactory.getLog(RangerOzoneAuthorizer.class);
	RangerOzonePlugin() {
		super("ozone", "ozone");
	}
	public List<RangerPolicy> getMatchingPoliciesForResource(RangerAccessResource resource) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RangerOzonePlugin.getMatchingPoliciesForResource(" + resource + ")");
		}
		List<RangerPolicy> matchingPolicies = new ArrayList<>();

		RangerPolicyEngine policyEngine = getPolicyEngine();

		if(policyEngine != null) {
			RangerPolicyEngineImpl rangerPolicyEngine = (RangerPolicyEngineImpl) policyEngine;
			matchingPolicies = rangerPolicyEngine.getLikelyMatchingResourcePolicies(resource);
		}
		if (LOG.isDebugEnabled()) {
			LOG.debug("<==  RangerOzonePlugin.getMatchingPoliciesForResource(" + resource + ", " + ") : " + matchingPolicies.size());
		}
		return matchingPolicies;
	}
}

class RangerOzoneAuditHandler extends RangerDefaultAuditHandler {
	private AuthzAuditEvent auditEvent     = null;

	public RangerOzoneAuditHandler() {

	}

	@Override
	public void processResult(RangerAccessResult result) {
		// Audit subpath access checks for recursive delete operations only when the access is not allowed.
		if (!isAuditingNeeded(result)) {
			return;
		}
		auditEvent = super.getAuthzEvents(result);
		super.logAuthzAudit(auditEvent);
	}

	private boolean isAuditingNeeded(final RangerAccessResult result) {
		boolean ret = false;

		if (!result.getIsAllowed()) {
			ret = true;
		}
		return ret;
	}
}

