/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.services.s3;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyItem;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyResource;
import org.apache.ranger.plugin.model.RangerService;
import org.apache.ranger.plugin.model.RangerServiceDef;
import org.apache.ranger.plugin.service.RangerBaseService;
import org.apache.ranger.plugin.service.ResourceLookupContext;

import com.google.common.annotations.VisibleForTesting;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.reflect.TypeToken;

/**
 * RangerService for S3 (Azure Data Lake Storage).
 */
public class RangerServiceS3 extends RangerBaseService {
    private static final Logger LOG = LoggerFactory.getLogger(
        RangerServiceS3.class);

    public static final String BUCKET = "bucket";
    public static final String PATH = "path";
    public static final String S3_PROTOCOL = "s3a://";
    public static final String S3_DEFAULT_POLICY_FILE = "ranger-s3-default-policies.json";
    private static final String RELATIVE_PATH_ROOT = "/";

    public static final String CONF_S3A_DATA_LOC = "s3.url.data";
    public static final String CONF_S3A_LOG_LOC = "s3.url.logs";
    public static final String CONF_S3A_RANGER_AUDIT_LOC = "s3.url.audits.ranger";
    public static final String CONF_S3A_SETUP_DEFAULT_POLICY = "setup.additional.s3.default.policies";
    private static final String LOCATION_PREFIX = "cloud.storage.url.";

    private static final String STR_RANGER_AUDIT = "RANGER_AUDIT";
    private static final String STR_HIVE_METASTORE_WAREHOUSE = "HIVE_METASTORE_WAREHOUSE";
    private static final String STR_HIVE_REPLICA_WAREHOUSE = "HIVE_REPLICA_WAREHOUSE";
    private static final String STR_HIVE_METASTORE_EXTERNAL_WAREHOUSE = "HIVE_METASTORE_EXTERNAL_WAREHOUSE";
    private static final String STR_HBASE_ROOT = "HBASE_ROOT";
    private static final String STR_DATA = "DATA";
    private static final String STR_LOG = "LOG";
    private static final String STR_BUCKET = "_BUCKET";
    private static final String STR_RELATIVE_PATH = "_RELATIVE_PATH";
    private static final String STR_BOOLEAN_TRUE = "true";
    private static final String STR_OPERATOR_EQUAL = "=";
    private static final String STR_RANGER_AUDIT_PATH = "/ranger/audit";

    private static final Pattern TEMPLATE_TOKEN_PATTERN = Pattern.compile("\\{\\{([a-zA-Z_=\\.\\*/-]*)\\}\\}");
    private final Map<String, S3ALocation> s3aLocationMap = new HashMap<>();

    private static final Pattern s3aBucketPattern = Pattern.compile("s3a:\\/\\/([^\\/]*)");
    private static final Pattern s3aPathPattern = Pattern.compile("s3a:\\/\\/[^\\/]*\\/(.*)");

    @Override
    public HashMap<String, Object> validateConfig() throws Exception {
        // TODO: implement configure validation for S3 policies.
        return new HashMap<>();
    }

    @Override
    public List<String> lookupResource(ResourceLookupContext context)
        throws Exception {
        // TODO: implement resource lookup for S3 policies.
        return new ArrayList<>();
    }

    @Override
    public void init(RangerServiceDef serviceDef, RangerService service) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.init({}, {})", serviceDef, service);
        }

        super.init(serviceDef, service);

        Map<String, String> serviceConfig = service.getConfigs();
        String auditLocStr = null;

        for (Map.Entry<String, String> entry : serviceConfig.entrySet()) {
            if (entry.getKey().startsWith(LOCATION_PREFIX)) {
                String keyForMap = entry.getKey().substring(LOCATION_PREFIX.length());
                if (StringUtils.isNotBlank(keyForMap)) {
                    s3aLocationMap.put(keyForMap, new S3ALocation(entry.getValue()));
                    if (StringUtils.equals(keyForMap, STR_RANGER_AUDIT)) {
                        auditLocStr = entry.getValue();
                    }
                }
            }
        }

        if (auditLocStr == null) {
            auditLocStr = serviceConfig.get(CONF_S3A_RANGER_AUDIT_LOC);
        }

        String auditLocWithoutPath = null;

        if (StringUtils.isNotBlank(auditLocStr)) {
            int auditPathIndex = auditLocStr.indexOf(STR_RANGER_AUDIT_PATH);
            if (auditPathIndex < 0) { // Audit path does not contain "/ranger/audit"
                auditPathIndex = auditLocStr.indexOf(RELATIVE_PATH_ROOT, auditLocStr.indexOf(S3_PROTOCOL) + (S3_PROTOCOL.length()));
            }

            if (auditPathIndex > 0) {
                auditLocWithoutPath = auditLocStr.substring(0, auditPathIndex);
            } else {
                auditLocWithoutPath = auditLocStr;
            }
        }

        // Add IF RANGER_AUDIT is missing && s3Locataion is not BLANK
        addMissingS3ALocation(STR_RANGER_AUDIT, auditLocStr);

        // Add IF DATA is missing && s3Locataion is not BLANK
        if (!s3aLocationMap.containsKey(STR_DATA)) {
            String dataLocStr = serviceConfig.get(CONF_S3A_DATA_LOC);
            if (StringUtils.isBlank(dataLocStr)) {
                dataLocStr = auditLocWithoutPath;
            }
            addMissingS3ALocation(STR_DATA, dataLocStr);
        }

        // Add IF LOG is missing && s3Locataion is not BLANK
        addMissingS3ALocation(STR_LOG, serviceConfig.get(CONF_S3A_LOG_LOC));

        if (StringUtils.isNotBlank(auditLocWithoutPath)) {
            // Add IF HIVE_METASTORE_WAREHOUSE is missing && s3Locataion is not BLANK
            addMissingS3ALocation(STR_HIVE_METASTORE_WAREHOUSE,
                auditLocWithoutPath + "/warehouse/tablespace/managed");

            // Add IF HIVE_REPLICA_WAREHOUSE is missing && s3Locataion is not BLANK
            addMissingS3ALocation(STR_HIVE_REPLICA_WAREHOUSE,
                auditLocWithoutPath + "/hive_replica_functions_dir");

            // Add IF HIVE_METASTORE_EXTERNAL_WAREHOUSE is missing && s3Locataion is not BLANK
            addMissingS3ALocation(STR_HIVE_METASTORE_EXTERNAL_WAREHOUSE,
                auditLocWithoutPath + "/warehouse/tablespace/external");

            // Add IF HBASE is missing && s3Locataion is not BLANK
            addMissingS3ALocation(STR_HBASE_ROOT,
                    auditLocWithoutPath + "/hbase");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.init(): s3aLocationMap={}", s3aLocationMap);
        }
    }

    @Override
    public List<RangerPolicy> getDefaultRangerPolicies() throws Exception {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.getDefaultRangerPolicies()");
        }

        final List<RangerPolicy> ret = super.getDefaultRangerPolicies();
        final Boolean setupAdditionalPolicies = Boolean.valueOf(
            configs.getOrDefault(CONF_S3A_SETUP_DEFAULT_POLICY, STR_BOOLEAN_TRUE));

        if (setupAdditionalPolicies && areRequiredConfigsPresent()) {
            // Fetch policies from JSON
            InputStream inStream = getClass().getClassLoader().getResourceAsStream(S3_DEFAULT_POLICY_FILE);
            InputStreamReader reader = new InputStreamReader(inStream);
            readPoliciesFromJsonFile(reader, ret);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.getDefaultRangerPolicies(): ret={}", ret);
        }

        return ret;
    }

    private boolean areRequiredConfigsPresent() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.areRequiredConfigsPresent()");
        }

        boolean ret = areS3ALocationsValid(s3aLocationMap);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.areRequiredConfigsPresent(): ret={}", ret);
        }

        return ret;
    }

    private boolean areS3ALocationsValid(Map<String, S3ALocation> s3aLocationMapParam) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.isS3ALocationValid({})",
                s3aLocationMapParam);
        }

//        boolean ret = s3aLocationMapParam.size() > 0 && s3aLocationMapParam.get(STR_RANGER_AUDIT) != null;
        boolean ret = true; //FIXME currently CDEP clusters, do not send RANGER AUDIT location

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.isS3ALocationValid({}): {}", s3aLocationMapParam, ret);
        }

        return ret;
    }

    private void readPoliciesFromJsonFile(Reader reader, List<RangerPolicy> ret) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.readPoliciesFromJsonFile({}, {})", reader,
                ret);
        }

        List<RangerPolicy> transformedPolicies = new ArrayList<>();
        List<RangerPolicy> defaultPolicies = readJson(reader);
        Map<String, String> s3aLocationConfig = getS3ALocationMap();

        if (!defaultPolicies.isEmpty()) {
            for (RangerPolicy policy : defaultPolicies) {
                RangerPolicy policyTransformed = transformPolicy(policy, s3aLocationConfig);

                if (policyTransformed != null) {
                    transformedPolicies.add(policyTransformed);
                }
            }
        }

        if (!transformedPolicies.isEmpty()) {
            ret.addAll(transformedPolicies);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.readPoliciesFromJsonFile(): ret={}", ret);
        }
    }

    private List<RangerPolicy> readJson(Reader reader) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.readJson({})", reader);
        }

        List<RangerPolicy> policies = new ArrayList<>();

        if (reader != null) {
            Gson gson = new GsonBuilder().setPrettyPrinting().create();
            Type listType = new TypeToken<ArrayList<RangerPolicy>>() {
            }.getType();

            policies = gson.fromJson(reader, listType);
        } else {
            LOG.warn("===>>> File " + reader + ", does not exits.");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.readJson(): policies={}", policies);
        }

        return policies;
    }

    private RangerPolicy transformPolicy(RangerPolicy policy,
        Map<String, String> s3aLocationConfig) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.transformPolicy({}, {})", policy,
                s3aLocationConfig);
        }

        RangerPolicy ret = null;
        Map<String, RangerPolicyResource> resourceMap = policy.getResources();
        RangerPolicyResource buckets = resourceMap.get(BUCKET);
        RangerPolicyResource relativePaths = resourceMap.get(PATH);
        List<String> bucketList = parseValuesFromTemplate(buckets.getValues(), s3aLocationConfig);
        List<String> relativePathList = parseValuesFromTemplate(relativePaths.getValues(), s3aLocationConfig);

        if (CollectionUtils.isNotEmpty(bucketList) ||
            CollectionUtils.isNotEmpty(relativePathList)) {
            policy.setService(service.getName());

            buckets.setValues(bucketList);
            relativePaths.setValues(relativePathList);

            for (RangerPolicyItem policyItem : policy.getPolicyItems()) {
                List<String> users = parseValuesFromTemplate(policyItem.getUsers(), service.getConfigs());
                List<String> groups = parseValuesFromTemplate(policyItem.getGroups(), service.getConfigs());
                List<String> roles = parseValuesFromTemplate(policyItem.getRoles(), service.getConfigs());

                policyItem.setUsers(users);
                policyItem.setGroups(groups);
                policyItem.setRoles(roles);
            }

            ret = policy;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.transformPolicy(): ret={}", ret);
        }

        return ret;
    }

    @VisibleForTesting
    public static class S3ALocation {
        private final String bucket;
        private final String path;

        public S3ALocation(String location) {
            Matcher bucketMatcher = s3aBucketPattern.matcher(location);
            Matcher pathMatcher = s3aPathPattern.matcher(location);

            this.bucket = bucketMatcher.find() ? bucketMatcher.group(1) : null;
            this.path = pathMatcher.find() ? RELATIVE_PATH_ROOT + pathMatcher.group(1) : null;
        }

        public String getBucket() {
            return bucket;
        }

        public String getPath() {
            return path;
        }

        public boolean pathNotEmpty() {
            return path != null && path.length() > 1;
        }

        @Override
        public String toString() {
            return "{bucket=" + bucket + ", path=" + path + "}";
        }
    }

    private Map<String, String> getS3ALocationMap() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.getS3ALocationMap()");
        }

        Map<String, String> ret = new HashMap<>();
        for (Map.Entry<String, S3ALocation> entry : s3aLocationMap.entrySet()) {
            S3ALocation s3aLocation = entry.getValue();
            addValueIfNotNull(ret, entry.getKey() + STR_BUCKET,
                s3aLocation.getBucket());

            if (s3aLocation.pathNotEmpty()) {
                addValueIfNotNull(ret, entry.getKey() + STR_RELATIVE_PATH,
                    s3aLocation.getPath());
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.getS3ALocationMap(): ret={}", ret);
        }

        return ret;
    }

    /**
     * This method, if present replace first (single) token with appropriate value.
     * @param inputList
     * @param config
     * @return final value after replacement
     */
    private List<String> parseValuesFromTemplate(List<String> inputList,
        Map<String, String> config) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerServiceS3.parseValuesFromTemplate({}, {})",
                inputList, config);
        }

        List<String> ret = new ArrayList<>();
        Set<String> uniqueValues = new HashSet<>();

        if (inputList != null) {

            for (String input : inputList) {
                Matcher matcher = TEMPLATE_TOKEN_PATTERN.matcher(input);

                // example input: {{service.user.hive=hive}}, {{service.user.hive}}, hive
                if (matcher.find()) {
                    String keyWithDefaultValue = matcher.group(1);
                    String[] keyValue = keyWithDefaultValue.split(STR_OPERATOR_EQUAL);
                    String value = config.get(keyValue[0]);

                    if (StringUtils.isNotBlank(value)) {
                        input = matcher.replaceFirst(value);
                    } else if (keyValue.length > 1 &&
                        StringUtils.isNotBlank(keyValue[1])) {
                        input = matcher.replaceFirst(keyValue[1]);
                    } else if (keyValue.length == 1) {
                        // If there is no value replace with empty string.
                        // check Hadoop App Logs policy.
                        input = matcher.replaceFirst("");
                    }
                    else {
                        input = null; //token exists, but nothing matches
                    }
                }

                if (StringUtils.isNotBlank(input)) {
                    uniqueValues.add(input);
                }
            }
        }
        ret.addAll(uniqueValues);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerServiceS3.parseValuesFromTemplate(): ret={}", ret);
        }

        return ret;
    }

    private void addValueIfNotNull(Map<String, String> map,
        String key,
        String value) {
        if (value != null) {
            map.put(key, value);
        }
    }

    private void addMissingS3ALocation(String key, String s3aLocation) {
        if (StringUtils.isNotBlank(s3aLocation)) {
            s3aLocationMap.putIfAbsent(key, new S3ALocation(s3aLocation));
        }
    }
}
