/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.services.s3;

import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerService;
import org.apache.ranger.plugin.model.RangerServiceDef;
import org.junit.Test;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class TestRangerServiceS3 {

  private static final String CUSTOM_PATH = "/custom1/custom2";
  private static final String BASE_BUCKET_PATH = "s3a://base-bucket";
  private static final String RANGER_AUDIT = "/ranger/audit";
  private static final String DEFAULT_BACKUP_PATH = "/backups";
  private static final String BACKUP_LOCATION = BASE_BUCKET_PATH
          + CUSTOM_PATH + DEFAULT_BACKUP_PATH;

  private static final String PATH_KEY = "path";

  private static final String HADOOP_APP_LOGS_POLICY_NAME = "Default: Hadoop App Logs";
  private static final String HADOOP_APP_LOGS_POLICIY_PATH = "/oplogs/yarn-app-logs";
  private static final String HADOOP_APP_LOGS_POLICY_PATH_CUSTOM =
          CUSTOM_PATH + HADOOP_APP_LOGS_POLICIY_PATH;

  private static final String HADOOP_APP_LOGS_USER_POLICY_NAME = "Default: Hadoop App Logs - user";
  private static final String HADOOP_APP_LOGS_USER_POLICY_PATH = "/oplogs/yarn-app-logs/{USER}";
  private static final String HADOOP_APP_LOGS_USER_POLICY_PATH_CUSTOM =
          CUSTOM_PATH + HADOOP_APP_LOGS_USER_POLICY_PATH;

  private static final String DATA_CATALOG_APP_POLICY_NAME = "Default: Data Catalog";
  private static final String DATA_CATALOG_APP_POLICY_PATH = "/dpprofiler";
  private static final String DATA_CATALOG_APP_POLICY_PATH_CUSTOM =
      CUSTOM_PATH + DATA_CATALOG_APP_POLICY_PATH;

  private static final String TMP_POLICY_NAME = "Default: Temporary files";
  private static final String TMP_PATH = "/tmp/{USER}";
  private static final String TMP_CUSTOM_PATH =
      CUSTOM_PATH + TMP_PATH;

  private static final String DEFAULT_USER_HOME_POLICY_NAME = "Default: User Home";
  private static final String DEFAULT_USER_HOME_POLICY_PATH = "/user/{USER}";
  private static final String DEFAULT_USER_HOME_POLICY_PATH_CUSTOM =
          CUSTOM_PATH + DEFAULT_USER_HOME_POLICY_PATH;

  private static final String DEFAULT_SCHEMA_REGISTRY_POLICY_NAME = "Default: Schema Registry";
  private static final String DEFAULT_SCHEMA_REGISTRY_POLICY_PATH = "/*/schema-registry";

  private static final String DEFAULT_FLINK_SSB_POLICY_NAME = "Default: flink/ssb";
  private static final String DEFAULT_FLINK_SSB_POLICY_PATH = "/flink";
  private static final String DEFAULT_FLINK_SSB_POLICY_PATH_CUSTOM =
      CUSTOM_PATH + DEFAULT_FLINK_SSB_POLICY_PATH;

  private static final String DEFAULT_FLINK_USER_POLICY_NAME = "Default: flink - user";
  private static final String DEFAULT_FLINK_USER_POLICY_PATH = "/flink/*/checkpoints/{USER}";
  private static final String DEFAULT_FLINK_USER_POLICY_PATH_CUSTOM =
      CUSTOM_PATH + DEFAULT_FLINK_USER_POLICY_PATH;

  private static final String DEFAULT_BACKUP_POLICY_NAME = "Default: Automatic backup";
  private static final String DEFAULT_BACKUP_POLICY_PATH = DEFAULT_BACKUP_PATH;
  private static final String DEFAULT_BACKUP_POLICY_PATH_CUSTOM =
          CUSTOM_PATH + DEFAULT_BACKUP_PATH;

  @Test
  public void testS3Location() {
    String protocol = "s3a://";
    String bucket = "my-bucket";
    String path = "/my/relative/path";

    String location = protocol + bucket + path;
    RangerServiceS3.S3ALocation s3ALocation = new RangerServiceS3.S3ALocation(location);

    assertEquals(bucket, s3ALocation.getBucket());
    assertEquals(path, s3ALocation.getPath());
  }

  @Test
  public void testS3LocationPathNull() {
    String protocol = "s3a://";
    String bucket = "my-bucket";
    String path = "";

    String location = protocol + bucket + path;
    RangerServiceS3.S3ALocation s3ALocation = new RangerServiceS3.S3ALocation(location);

    assertEquals(bucket, s3ALocation.getBucket());
    assertEquals(null, s3ALocation.getPath());
  }


  /**
   * Validates whether default ranger policies are created correctly
   * for default s3 path.
   */
  @Test
  public void testS3DefaultLocation() throws Exception {
    Map<String, String> config = new HashMap<>();
    config.put("cluster.name", "random-dl-cluster");
    config.put("cloud.storage.url.RANGER_AUDIT", BASE_BUCKET_PATH + RANGER_AUDIT);
    RangerServiceS3 serviceS3 = new RangerServiceS3();
    RangerService rangerService = new RangerService();
    rangerService.setConfigs(config);
    serviceS3.init(new RangerServiceDef(), rangerService);
    List<RangerPolicy> defaultRangerPolicies = serviceS3.getDefaultRangerPolicies();
    assertPathContains(defaultRangerPolicies,
            HADOOP_APP_LOGS_POLICY_NAME, HADOOP_APP_LOGS_POLICIY_PATH);

    assertPathContains(defaultRangerPolicies,
            HADOOP_APP_LOGS_USER_POLICY_NAME, HADOOP_APP_LOGS_USER_POLICY_PATH);

    assertPathContains(defaultRangerPolicies,
            DEFAULT_USER_HOME_POLICY_NAME, DEFAULT_USER_HOME_POLICY_PATH);

    assertPathContains(defaultRangerPolicies,
            DATA_CATALOG_APP_POLICY_NAME, DATA_CATALOG_APP_POLICY_PATH);

    assertPathContains(defaultRangerPolicies,
            TMP_POLICY_NAME, TMP_PATH);

    assertPathContains(defaultRangerPolicies,
            DEFAULT_SCHEMA_REGISTRY_POLICY_NAME, DEFAULT_SCHEMA_REGISTRY_POLICY_PATH);

    assertPathContains(defaultRangerPolicies,
        DEFAULT_FLINK_SSB_POLICY_NAME, DEFAULT_FLINK_SSB_POLICY_PATH);

    assertPathContains(defaultRangerPolicies,
        DEFAULT_FLINK_USER_POLICY_NAME, DEFAULT_FLINK_USER_POLICY_PATH);

    assertPathContains(defaultRangerPolicies,
            DEFAULT_BACKUP_POLICY_NAME, DEFAULT_BACKUP_POLICY_PATH);
  }

  /**
   * Validates whether default ranger policies are created correctly
   * for custom s3 path.
   */
  @Test
  public void testS3CustomLocation() throws Exception {
    Map<String, String> config = new HashMap<>();
    config.put("cluster.name", "random-dl-cluster");
    config.put("cloud.storage.url.RANGER_AUDIT", BASE_BUCKET_PATH + CUSTOM_PATH + RANGER_AUDIT);
    config.put("cloud.storage.url.BACKUP_LOCATION", BACKUP_LOCATION);
    RangerServiceS3 serviceS3 = new RangerServiceS3();
    RangerService rangerService = new RangerService();
    rangerService.setConfigs(config);
    serviceS3.init(new RangerServiceDef(), rangerService);
    List<RangerPolicy> defaultRangerPolicies = serviceS3.getDefaultRangerPolicies();
    assertPathContains(defaultRangerPolicies,
            HADOOP_APP_LOGS_POLICY_NAME, HADOOP_APP_LOGS_POLICY_PATH_CUSTOM);

    assertPathContains(defaultRangerPolicies,
            HADOOP_APP_LOGS_USER_POLICY_NAME, HADOOP_APP_LOGS_USER_POLICY_PATH_CUSTOM);

    assertPathContains(defaultRangerPolicies,
            DEFAULT_USER_HOME_POLICY_NAME, DEFAULT_USER_HOME_POLICY_PATH_CUSTOM);

    assertPathContains(defaultRangerPolicies,
            DATA_CATALOG_APP_POLICY_NAME, DATA_CATALOG_APP_POLICY_PATH_CUSTOM);

    assertPathContains(defaultRangerPolicies,
            TMP_POLICY_NAME, TMP_CUSTOM_PATH);

    assertPathContains(defaultRangerPolicies,
            DEFAULT_SCHEMA_REGISTRY_POLICY_NAME, DEFAULT_SCHEMA_REGISTRY_POLICY_PATH);

    assertPathContains(defaultRangerPolicies,
        DEFAULT_FLINK_SSB_POLICY_NAME, DEFAULT_FLINK_SSB_POLICY_PATH_CUSTOM);

    assertPathContains(defaultRangerPolicies,
        DEFAULT_FLINK_USER_POLICY_NAME, DEFAULT_FLINK_USER_POLICY_PATH_CUSTOM);

    assertPathContains(defaultRangerPolicies,
            DEFAULT_BACKUP_POLICY_NAME, DEFAULT_BACKUP_POLICY_PATH_CUSTOM);

  }

  private void assertPathContains(List<RangerPolicy> defaultRangerPolicies,
                                  String policyName,
                                  String policyPath) {
    for (RangerPolicy defaultRangerPolicy : defaultRangerPolicies) {
      if (defaultRangerPolicy.getName().equals(policyName)) {
        List<String> values = defaultRangerPolicy.getResources().get(PATH_KEY).getValues();
        assertTrue(policyName + " values should contain " + policyPath
                        + "\nCurrent values " + values,
                values.contains(policyPath));
      }
    }
  }
}