#!/bin/bash
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------------
START=$(date +%s)
usage() {
        [ "$*" ] && echo "$0: $*"
        sed -n '/^##/,/^$/s/^## \{0,1\}//p' "$0"
        exit 2
} 2>/dev/null

log() {
    local prefix="$(date +%Y-%m-%d\ %H:%M:%S,%3N) "
    echo "${prefix} $@" >> $LOGFILE
    echo "${prefix} $@"
}

source env.sh

policymgr_external_url=${POLICYMGR_EXTERNAL_URL}
policymgr_kerberos_enabled=${POLICYMGR_KERBEROS_ENABLED}
policymgr_kerberos_enabled=`echo $policymgr_kerberos_enabled | tr '[:upper:]' '[:lower:]'`
policymgr_https_enabled=${POLICYMGR_HTTPS_ENABLED}
policymgr_https_enabled=`echo $policymgr_https_enabled | tr '[:upper:]' '[:lower:]'`
policymgr_https_cert_file=${POLICYMGR_HTTPS_CERT_FILE}

policymgr_admin=${POLICYMGR_ADMIN}
policymgr_pwd=${POLICYMGR_PWD}

source_service_name=${SOURCE_SERVICE_NAME}
source_zone_name=${SOURCE_ZONE_NAME}

policy_export_output_location=${EXPORT_OUTPUT_LOCATION}
service_export_output_location=''

ARG_requestParams=''
ARG_policyExportLocation=''
ARG_serviceExportLocation=''
ARG_sslParams='-k'
ARG_kerberosParams=''
ARG_knoxsso=''
ARG_userpwd=''

init_variables(){
    if [ "${policymgr_external_url}" == "" ]
    then
        log "[E] Please provide value for POLICYMGR_EXTERNAL_URL"
        exit 1
    fi

    regex='^(https?|http)://[-A-Za-z0-9\+&@#/%?=~_|!:.]*[-A-Za-z0-9\+&@#/%=~_|]\.[-A-Za-z0-9\+&@#/%?=~_|!:.]*[-A-Za-z0-9\+&@#/%=~_|]$'
    if ! [[ $policymgr_external_url =~ $regex ]]
    then
        log "[E] Invalid Ranger Admin URL provided in config POLICYMGR_EXTERNAL_URL"
        exit 1
    fi

    if [ "${source_service_name}" != "all" ]  && [ "${source_service_name}" != "" ]
    then
        ARG_requestParams="?serviceName=${source_service_name}"
    else
        source_service_name="all"
        ARG_requestParams=''
    fi

    if [ "${source_zone_name}" != "" ] && [ "${ARG_requestParams}" != "" ]
    then
        ARG_requestParams="${ARG_requestParams}"'&zoneName='${source_zone_name}
    fi

    if [ "${source_zone_name}" != "" ] && [ "${ARG_requestParams}" == "" ]
    then
        ARG_requestParams="?zoneName=${source_zone_name}"
    fi

    if [ "${policy_export_output_location}" == "" ]
    then
        policy_export_output_location=${PWD}
        service_export_output_location=${PWD}
    fi

    if [[ $policy_export_output_location != *.json ]]
    then
        suffix=$(date +'_%Y%m%d_%H%M%S')
        service_export_output_location="${policy_export_output_location}/exportedRangerServices${suffix}.json"
        policy_export_output_location="${policy_export_output_location}/exportedRangerPolicies${suffix}.json"
    else
        parent_dir="$(dirname -- "$(realpath -- "$policy_export_output_location")")"
        service_export_output_location="${parent_dir}/exportedRangerServices.json"
    fi
    ARG_policyExportLocation="-o ${policy_export_output_location}"
    ARG_serviceExportLocation="-o ${service_export_output_location}"

    if [[ "$policymgr_https_enabled" == "true" ]]
    then
        if [[ "$policymgr_https_cert_file" == "" ]]
        then
            log "[E] Please provide value in config POLICYMGR_HTTPS_CERT_FILE"
            exit 1
        fi
        if ! test -f "$policymgr_https_cert_file"; then
            log "[E] Please provide valid file path in config POLICYMGR_HTTPS_CERT_FILE"
            exit 1
        fi
        ARG_sslParams="--cacert ${policymgr_https_cert_file}"
    fi
 
    if [[ "$policymgr_kerberos_enabled" == "true" ]]
    then
        ARG_kerberosParams='--negotiate'
        policymgr_admin=""
        policymgr_pwd=""
    fi

    ARG_userpwd="-u ${policymgr_admin}:${policymgr_pwd}"
    if [[ "$POLICYMGR_KNOXSSO_TOKEN" != "" ]]
    then
        ARG_knoxsso="--cookie hadoop-jwt=${POLICYMGR_KNOXSSO_TOKEN}"
        policymgr_admin=""
        policymgr_pwd=""
        ARG_userpwd=""
    fi

    if [[ "$policymgr_kerberos_enabled" != "true" && "$POLICYMGR_KNOXSSO_TOKEN" == "" ]] && [[ "${policymgr_admin}" == "" || "${policymgr_pwd}" == "" ]]
    then
        log "[E] Please provide value for POLICYMGR_ADMIN and POLICYMGR_PWD"
        exit 1
    fi
}

init_variables

set -x
curl ${ARG_sslParams} ${ARG_knoxsso} --header "text/json" -H "Content-Type: text/json" ${ARG_policyExportLocation} ${ARG_kerberosParams} ${ARG_userpwd} -X GET "${policymgr_external_url}/service/plugins/policies/exportJson${ARG_requestParams}"
cmdReturn=$?
set +x
if [ "$cmdReturn" == "0" ]
then
    set -x
    curl ${ARG_sslParams} ${ARG_knoxsso} --header "text/json" -H "Content-Type: text/json" ${ARG_serviceExportLocation} ${ARG_kerberosParams} ${ARG_userpwd} -X GET "${policymgr_external_url}/service/plugins/services"
    cmdReturn2=$?
    set +x
    if [ "$cmdReturn2" != "0" ]
    then
        echo "Ranger service export request failed!"
    fi

    if test -f "$policy_export_output_location"; then
        echo "Policies exported at location: $policy_export_output_location"
    fi

    if test -f "$service_export_output_location"; then
        echo "Services exported at location: $service_export_output_location"
    fi

    END=$(date +%s)
    DIFF=$(( $END - $START ))
    echo "Time taken: $DIFF seconds"
else
    echo "Ranger policy export request failed!"
fi