/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.policymigration;

import static org.apache.ranger.policymigration.cli.RangerPolicyMigrationCmdOptions.CONVERT_COMMAND;
import static org.apache.ranger.policymigration.cli.RangerPolicyMigrationCmdOptions.TRANSFORM_COMMAND;
import static org.apache.ranger.policymigration.common.RangerPolicyMigrationConstants.POLICYMIGRATION_CONF;

import java.io.File;
import java.net.MalformedURLException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;

import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.policymigration.cli.RangerPolicyMigrationCmdOptions;
import org.apache.ranger.policymigration.common.PolicyMigrationUtils;
import org.apache.ranger.policymigration.common.RangerPolicyMigrationConfig;
import org.apache.ranger.policymigration.reader.CSVFileReader;
import org.apache.ranger.policymigration.reader.NativeAclPolicy;
import org.apache.ranger.policymigration.transformer.NativeAclToRangerPolicyConverter;
import org.apache.ranger.policymigration.transformer.TransformTask;
import org.apache.ranger.policymigration.view.RangerExportPolicyList;
import org.apache.ranger.policymigration.view.RangerServiceList;

public class RangerExportedPolicyTransformer {

	private static final Log LOG = LogFactory.getLog(RangerExportedPolicyTransformer.class);

	private static void addResourceToConfiguration(Configuration configObj, String dir, String fileName)
			throws MalformedURLException {

		if (dir.isEmpty() || fileName.isEmpty()) {
			throw new IllegalArgumentException("Path not specified");
		}
		Path pathObj = Paths.get(dir, fileName);
		File configFile = new File(pathObj.toString());
		if (configFile.isFile() && configFile.canRead()) {
			configObj.addResource(configFile.toURI().toURL());
		} else {
			throw new IllegalArgumentException("Cannot read configuration file " + configFile);
		}
	}

	static void buildConfiguration(String type, String dir, Configuration configObj)
			throws MalformedURLException {
		//This block can be removed if same settings can be used for all commands
		switch (type) {
		case CONVERT_COMMAND:
		case TRANSFORM_COMMAND:
			addResourceToConfiguration(configObj, dir, POLICYMIGRATION_CONF);
			break;
		default:
			return;
		}
	}

	public static void main(String[] args) {
		RangerPolicyMigrationCmdOptions policymigrationCmdOptions = new RangerPolicyMigrationCmdOptions();
		int status = 1;// Default to failure
		Configuration config = new Configuration(true);
		try {
			String outputFileLocation = null;
			policymigrationCmdOptions.parse(args);
			String policyFileLocation = policymigrationCmdOptions.getPolicyLocation();
			String command = policymigrationCmdOptions.getCommand();
			switch (command) {
			case TRANSFORM_COMMAND:
				buildConfiguration(TRANSFORM_COMMAND, policymigrationCmdOptions.getConfDir(), config);
				outputFileLocation = RangerPolicyMigrationConfig.getProperty(config, RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_OUTPUTFILE_PATH);
				try {
					RangerExportPolicyList rangerExportPolicyList = null;
					LOG.info("==========Ranger policy transformation start==========");
					try {
						LOG.info("Received ranger policies json file: " + policyFileLocation);
						rangerExportPolicyList = PolicyMigrationUtils.readRangerExportPolicyListJsonFile(policyFileLocation);
						if (LOG.isDebugEnabled()) {
							LOG.debug("Received ranger policies json content: " + rangerExportPolicyList);
						}
					} catch (Exception e) {
						LOG.error("Error reading ranger policy json file:"+policyFileLocation, e);
					}

					String serviceFileLocation = policymigrationCmdOptions.getServiceLocation();
					RangerServiceList rangerServiceList = null;
					if (StringUtils.isNotBlank(serviceFileLocation)) {
						try {
							LOG.info("Received ranger service json file: " + serviceFileLocation);
							rangerServiceList = PolicyMigrationUtils.readRangerServiceListJsonFile(serviceFileLocation);
							if (LOG.isDebugEnabled()) {
								LOG.debug("Received ranger service json content: " + rangerServiceList);
							}
						} catch (Exception e) {
							LOG.error("Error reading ranger service json file:"+serviceFileLocation, e);
						}
					}

					if (rangerExportPolicyList != null) {
						TransformTask transformTask = new TransformTask(config);

						rangerExportPolicyList = transformTask.getTransformedPolicies(rangerExportPolicyList, rangerServiceList);
						String transformedFileLocation = PolicyMigrationUtils.getFileOutPutLocation(policyFileLocation, outputFileLocation, command);

						PolicyMigrationUtils.saveRangerExportPolicyListToFile(config, rangerExportPolicyList, transformedFileLocation);

						if (rangerExportPolicyList != null) {
							LOG.info("Transformed Ranger Policies json file: " + transformedFileLocation);
							System.out.println("Transformed Ranger Policies json file: " + transformedFileLocation);
							status = 0;
						}
					}
					if (status == 0) {
						LOG.info("Transformation of the ranger policies complete");
					} else {
						LOG.error("Transformation of the ranger policies failed");
					}
				} catch (Exception exception) {
					LOG.error("Transformation could not be performed because of an exception", exception);
				}
				break;
			case CONVERT_COMMAND:
				buildConfiguration(CONVERT_COMMAND, policymigrationCmdOptions.getConfDir(), config);
				outputFileLocation = RangerPolicyMigrationConfig.getProperty(config, RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_OUTPUTFILE_PATH);
				try {
					LOG.info("==========Raw ACL conversion to RangerPolicy start ==========");
					try {
						CSVFileReader reader = new CSVFileReader();
						List<NativeAclPolicy> rawAcls = reader.getNativeACLPolicies(reader.getCSVReader(policyFileLocation, CSVFileReader.SEPARATOR_PIPE));
						if (LOG.isDebugEnabled()) {
							LOG.debug("Received input csv content: " + rawAcls);
						}

						if (rawAcls != null) {
							NativeAclToRangerPolicyConverter converterTask = new NativeAclToRangerPolicyConverter(config);

							RangerExportPolicyList rangerExportPolicyList = converterTask.getConvertedPolicies(rawAcls);

							String convertedFileLocation = PolicyMigrationUtils.getFileOutPutLocation(policyFileLocation, outputFileLocation, command);

							PolicyMigrationUtils.saveRangerExportPolicyListToFile(config, rangerExportPolicyList, convertedFileLocation);

							if (rangerExportPolicyList != null) {
								LOG.info("Converted Ranger Policies json file: " + convertedFileLocation);
								System.out.println("Converted Ranger Policies json file: " + convertedFileLocation);
								status = 0;
							}
						}
					} catch (Exception e) {
						LOG.error(e);
					}

					if (status == 0) {
						LOG.info("Conversion of the hdfs acl to ranger policies complete");
					} else {
						LOG.error("Conversion of the hdfs acl to ranger policies failed");
					}
				} catch (Exception exception) {
					LOG.error("Conversion could not be performed because of an exception", exception);
				}
				break;
			default:
				throw new ParseException("Unknown command");
			}
		} catch (ParseException e) {
			LOG.error("ParseException occurred during ranger policies policymigration", e);
		} catch (MalformedURLException e) {
			LOG.error("MalformedURLException occurred during ranger policies policymigration", e);
		} catch (Exception e) {
			LOG.error("Exception occurred during ranger policies policymigration", e);
		} catch (Throwable t) {
			LOG.error("Throwables occurred during ranger policies policymigration", t);
		} finally {
			System.exit(status);
		}
	}
}