/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.policymigration.cli;

import java.util.HashSet;
import java.util.Set;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;

public class RangerPolicyMigrationCmdOptions {

	public static final String HELP_LONG 		  = "help";
	public static final String HELP_SHORT 		  = "h";
	public static final String COMMAND_LONG 	  = "command";
	public static final String COMMAND_SHORT 	  = "c";
	public static final String CONF_DIR_LONG 	  = "confdir";
	public static final String CONF_DIR_SHORT 	  = "dr";
	public static final String POLICY_JSON_LONG   = "policy";
	public static final String POLICY_JSON_SHORT  = "p";
	public static final String SERVICE_JSON_LONG  = "service";
	public static final String SERVICE_JSON_SHORT = "s";

	public static final String TRANSFORM_COMMAND  = "transform";
	public static final String CONVERT_COMMAND    = "convert";

	Options options = new Options();

	private String command;
	private String confDir;
	private String policyLocation;
	private String serviceLocation=null;

	public static Set<String> VALID_COMMANDS = new HashSet<String>();

	static {
		VALID_COMMANDS.add(TRANSFORM_COMMAND);
		VALID_COMMANDS.add(CONVERT_COMMAND);
	}

	public RangerPolicyMigrationCmdOptions() {
		initializeOptions();
	}

	private void initializeOptions() {
		// add help option
		options.addOption(HELP_SHORT, HELP_LONG, false, "Print this help text");

		// add command option
		String commandMessage = "Specify the policymigration commmand to be executed.";
		options.addOption(COMMAND_SHORT, COMMAND_LONG, true, commandMessage);

		String dirMessage = "Path to directory with all required config/properties files. " + "Expected list of files includes core-site.xml, ranger-policymigration-site.xml";
		options.addOption(CONF_DIR_SHORT, CONF_DIR_LONG, true, dirMessage);

		String inputMessage = "Path of exported ranger policy json file which should contain a RangerPolicyExportList object in json format";
		options.addOption(POLICY_JSON_SHORT, POLICY_JSON_LONG, true, inputMessage);

		inputMessage = "Path of exported ranger service json file which should contain a RangerServiceList object in json format";
		options.addOption(SERVICE_JSON_SHORT, SERVICE_JSON_LONG, true, inputMessage);
	}

	public CommandLine parse(String... args) throws ParseException {
		try {
			CommandLine cmdLine = new GnuParser().parse(options, args);

			if (hasOption(cmdLine, HELP_LONG, HELP_SHORT)) {
				printHelp(options, null);
				System.exit(1);
			} else {
				initializeParams(cmdLine);
			}

			return cmdLine;

		} catch (ParseException e) {
			printHelp(options, e.getMessage());
			throw e;
		}
	}

	private void initializeParams(CommandLine cmdLine) throws ParseException {

		this.command		 = getOptionArguments(cmdLine, COMMAND_LONG, COMMAND_SHORT);
		this.confDir		 = getOptionArguments(cmdLine, CONF_DIR_LONG, CONF_DIR_SHORT);
		this.policyLocation	 = getOptionArguments(cmdLine, POLICY_JSON_LONG, POLICY_JSON_SHORT);

		if (hasOption(cmdLine, SERVICE_JSON_LONG, SERVICE_JSON_SHORT)) {
			this.serviceLocation = getOptionArguments(cmdLine, SERVICE_JSON_LONG, SERVICE_JSON_SHORT);
		}

		if (StringUtils.isBlank(this.command) && StringUtils.isBlank(confDir)) {
			throw new ParseException("No inputs provided. Please include [" + COMMAND_LONG + "] and [" + CONF_DIR_LONG + "] options");
		}

		if (StringUtils.isBlank(command)) {
			throw new ParseException("Missing/incorrect required option: [" + COMMAND_LONG + "]");
		}

		if (!VALID_COMMANDS.contains(command)) {
			throw new ParseException("Unknown command [" + command + "]");
		}

		if (command.equalsIgnoreCase(TRANSFORM_COMMAND) || command.equalsIgnoreCase(CONVERT_COMMAND)) {
			if (StringUtils.isBlank(policyLocation)) {
				throw new ParseException("Missing/incorrect required option: [" + POLICY_JSON_LONG + "]");
			}
		}

	}

	private boolean hasOption(CommandLine cmdLine, String longString, String shortString) {
		return cmdLine.hasOption(longString) || cmdLine.hasOption(shortString);
	}

	private String getOptionArguments(CommandLine cmdLine, String longString, String shortString) throws ParseException {
		String value = cmdLine.getOptionValue(longString);

		if (value != null) {
			if (hasMultiple(cmdLine, longString)) {
				throw new ParseException("Multiple option values for [" + longString + "]");
			}
			return value;
		}

		value = cmdLine.getOptionValue(shortString);

		if (value != null) {
			if (hasMultiple(cmdLine, shortString)) {
				throw new ParseException("Multiple option values for [" + longString + "]");
			}
		}

		return null;
	}

	private boolean hasMultiple(CommandLine cmdLine, String option) {
		return cmdLine.getOptionValues(option) != null && cmdLine.getOptionValues(option).length > 1;
	}

	private void printHelp(Options options, String msg) {
		String messageStr = "policymigration\n";
		if (msg != null) {
			messageStr = msg + " " + messageStr;
		}
		(new HelpFormatter()).printHelp(messageStr, options);
	}

	public String getCommand() {
		return this.command;
	}

	public String getConfDir() {
		return this.confDir;
	}

	public String getPolicyLocation() {
		return policyLocation;
	}

	public String getServiceLocation() {
		return serviceLocation;
	}

}