/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.policymigration.common;

import java.io.BufferedWriter;
import java.io.Closeable;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.ranger.policymigration.view.RangerExportPolicyList;

import com.google.common.base.Strings;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

/**
 * Client to write/read the ranger policy information from filesystem
 * <p>
 * Implements the serialization/de-serialization of the java objects before
 * writing/reading from the file.
 * <p>
 * Filesystem can be a local filesystem or HDFS based on the inputs provided.
 */

public class FSClient implements Closeable {

	private static final Log LOG = LogFactory.getLog(FSClient.class);


	private FileSystem 	fileSystem = null;
	private Path 		path;

	public Path getPath() {
		return path;
	}

	/**
	 * Added for tests. Should not be used.
	 * 
	 * @param newPath
	 */
	void setPath(Path newPath) {
		path = newPath;
	}

	FileSystem getFileSystem() {
		return fileSystem;
	}

	/**
	 * Constructor to be used if the filesystem is localfs.
	 *
	 * @param location where the data should written/read.
	 */
	FSClient(String location) {
		this(location, new Configuration(true));
	}

	/**
	 * @param location location where the data should written/read.
	 * @param conf     Configuration should have information of filesystem to be
	 *                 used.
	 */
	public FSClient(String location, Configuration conf) {
		try {
			path = new Path(location);
			fileSystem = path.getFileSystem(conf);
		} catch (Exception e) {
			LOG.error("Failed to initialize filesystem", e);
		}
	}

	public boolean write(RangerExportPolicyList rangerExportPolicyList) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> FSClient.write()");
		}

		boolean ret = false;
		if (rangerExportPolicyList == null) {
			LOG.error("Input is empty, Nothing to write");
			return ret;
		}

		if (CollectionUtils.isEmpty(rangerExportPolicyList.getPolicies())) {
			LOG.warn("There are no policies in the export data.");
		}

		// Serialize data
		Gson gson = new GsonBuilder().setPrettyPrinting().create();
		String rangerExportPolicyListStr = gson.toJson(rangerExportPolicyList);
		if (Strings.isNullOrEmpty(rangerExportPolicyListStr)) {
			LOG.error("Serialized data empty, nothing to write");
			return ret;
		}

		BufferedWriter br = null;
		try {
			// write the string into the file
			LOG.info("[Started] Writing policy information to [" + path + "]");

			OutputStream os = fileSystem.create(path);
			br = new BufferedWriter(new OutputStreamWriter(os, "UTF-8"));
			br.write(rangerExportPolicyListStr);

			LOG.info("[Completed] Writing policy information to " + path);
		} catch (Exception ex) {
			LOG.error("Exception while exporting policy information to file, located at " + path, ex);
		} finally {
			if (br != null) {
				try {
					br.close();
					ret = true;
				} catch (Exception ex) {
					LOG.error("Exception while closing file at " + path, ex);
				}
			}
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== FSClient.write()");
		}
		return ret;
	}

	public void close() throws IOException {
		if (fileSystem != null) {
			fileSystem.close();
		}
	}

	/**
	 * Used to delete the export that is created.
	 */
	public void deleteExportFile() {
		try {
			if (fileSystem.exists(path)) {
				fileSystem.delete(path, true);
				LOG.info("Delete file " + path);
			}
		} catch (Exception e) {
			LOG.error("Failed to delete the file " + path);
		}
	}
}
