/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.policymigration.common;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.plugin.util.JsonUtilsV2;
import org.apache.ranger.policymigration.reader.CSVFileReader;
import org.apache.ranger.policymigration.view.RangerExportPolicyList;
import org.apache.ranger.policymigration.view.RangerServiceList;

import com.google.common.base.Splitter;

public class PolicyMigrationUtils {

	private static final Log LOG 			= LogFactory.getLog(PolicyMigrationUtils.class);
	public static String DELIMITER_COMMA = ",";
	public static String DELIMITER_DEFAULT = ",";
	public static String DELIMITER_COLON = ":";
	public static String DELIMITER_PIPE = "|";

	public static RangerExportPolicyList readRangerExportPolicyListJsonFile(String fileName) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> PolicyMigrationUtils.readRangerExportPolicyListJsonFile("+fileName+")");
		}

		if (StringUtils.isBlank(fileName)) {
			throw new IllegalArgumentException("Path not specified");
		}

		RangerExportPolicyList dataset = null;

		Path pathObj 	= Paths.get(fileName);
		File configFile = new File(pathObj.toString());

		if (configFile.isFile() && configFile.canRead()) {

			FileInputStream uploadedInputStream = new FileInputStream(configFile);
			String inputJsonString 				= IOUtils.toString(uploadedInputStream, "UTF-8");
			inputJsonString 					= inputJsonString.trim();

			if (StringUtils.isNotEmpty(inputJsonString)) {
				dataset = JsonUtilsV2.jsonToObj(inputJsonString, RangerExportPolicyList.class);
			} else {
				LOG.error("Provided json file is empty!!");
			}
		} else {
			throw new IllegalArgumentException("Cannot read json file " + configFile);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== PolicyMigrationUtils.readRangerExportPolicyListJsonFile()");
		}
		return dataset;
	}

	public static RangerServiceList readRangerServiceListJsonFile(String fileName) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> PolicyMigrationUtils.readRangerServiceListJsonFile("+fileName+")");
		}

		if (StringUtils.isBlank(fileName)) {
			return null;
		}

		RangerServiceList dataset = null;

		Path pathObj 	= Paths.get(fileName);
		File configFile = new File(pathObj.toString());

		if (configFile.isFile() && configFile.canRead()) {

			FileInputStream uploadedInputStream = new FileInputStream(configFile);
			String inputJsonString 				= IOUtils.toString(uploadedInputStream, "UTF-8");
			inputJsonString 					= inputJsonString.trim();

			if (StringUtils.isNotEmpty(inputJsonString)) {
				dataset = JsonUtilsV2.jsonToObj(inputJsonString, RangerServiceList.class);
			} else {
				LOG.error("Provided json file is empty!!");
			}
		} else {
			throw new IllegalArgumentException("Cannot read json file " + configFile);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== PolicyMigrationUtils.readRangerServiceListJsonFile()");
		}
		return dataset;
	}

	public static boolean saveRangerExportPolicyListToFile(Configuration conf, RangerExportPolicyList rangerExportPolicyList, String exportLocation) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> PolicyMigrationUtils.saveRangerExportPolicyListToFile()");
		}

		FSClient fsClient 		= null;
		boolean writeComplete 	= false;

		try {
			if (LOG.isDebugEnabled()) {
				LOG.debug("File path:" + exportLocation);
			}

			if (StringUtils.isNotBlank(exportLocation)) {
				fsClient		= new FSClient(exportLocation, conf);
				writeComplete	= fsClient.write(rangerExportPolicyList);
			}
		} catch (Exception e) {
			LOG.error("Failed while closing the stream for " + fsClient);
		} finally {
			if (fsClient != null) {
				try {
					fsClient.close();
				} catch (IOException e) {
					LOG.error("Failed while closing the stream for " + fsClient);
				}
				fsClient = null;
			}
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== PolicyMigrationUtils.saveRangerExportPolicyListToFile()");
		}
		return writeComplete;
	}

	public static List<String> stringToList(String data) {
		List<String> items = new ArrayList<String>();
		if (StringUtils.isNotBlank(data)) {
			data = data.trim().replace("\"", "");
			String separator = DELIMITER_COMMA;
			if (!data.contains(separator)) {
				separator = DELIMITER_DEFAULT;
			}
			//split and remove spaces if any
			String[] itemsArray = Arrays.stream(data.split(separator)).map(String::trim).toArray(String[]::new);
			if (itemsArray != null && itemsArray.length > 0) {
		        // Now convert string into ArrayList
				items = new ArrayList<String>(Arrays.asList(itemsArray));
			}
		}
		return items;
	}

	public static Map<String, String> convertStringToMap(String str, String delimiter) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> PolicyMigrationUtils.convertStringToMap()" + str);
		}

		Map<String, String> ret = new HashMap<String, String>();
		if (StringUtils.isNotBlank(str)) {
			str = str.replace("{", "").replace("}", "");
			try {
				ret = Splitter.on(DELIMITER_COMMA).withKeyValueSeparator(delimiter).split(str);
				//ret = Arrays.stream(str.split(DELIMITER_COMMA)).map(s -> s.split(delimiter)).collect(Collectors.toMap(s -> s[0], s -> s[1]));
			} catch(Exception ex) {
				LOG.error("Error while converting string to map:"+str, ex);
			}
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== PolicyMigrationUtils.convertStringToMap()" + Arrays.asList(ret));
		}
		return ret;
	}
	
	public static String getFileOutPutLocation(String inputFilePath, String fileOutPutLocation, String command) {
		String suffix = "_" + command + ".json";

		String fileExt = FilenameUtils.getExtension(inputFilePath);
		fileExt = (fileExt.startsWith(".")) ? fileExt : "." +fileExt;
		if (StringUtils.isBlank(fileOutPutLocation)) {
			if (StringUtils.isNotBlank(inputFilePath)) {
				fileOutPutLocation = inputFilePath.replace(fileExt, suffix);
			}
		} else {
			if (!StringUtils.endsWith(fileOutPutLocation, ".json")) {
				Path pathObj = Paths.get(inputFilePath);
				String fileName = pathObj.getFileName().toString();
				if (StringUtils.isNotBlank(fileName)) {
					fileName = fileName.replace(fileExt, suffix);
				}
				fileOutPutLocation = StringUtils.endsWith(fileOutPutLocation, "/") ? fileOutPutLocation + fileName : fileOutPutLocation + "/" + fileName;
			}
		}
			
		if (StringUtils.startsWith(fileOutPutLocation, "hdfs:")) {
			fileOutPutLocation = fileOutPutLocation.replaceFirst("hdfs:", "");
			while (fileOutPutLocation.startsWith("/")) {
				fileOutPutLocation = fileOutPutLocation.replaceFirst("/", "");
			}
			fileOutPutLocation = "hdfs://" + fileOutPutLocation;
		} else {
			fileOutPutLocation = fileOutPutLocation.replaceFirst("file:", "");
			while (fileOutPutLocation.startsWith("/")) {
				fileOutPutLocation = fileOutPutLocation.replaceFirst("/", "");
			}
			fileOutPutLocation = "file:///" + fileOutPutLocation;
		}
		return fileOutPutLocation;
	}

	public static Map<String, String> readMappingFile(String fileName) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> PolicyMigrationUtils.readMappingFile():"+ fileName);
		}

		Map<String, String> ret = new HashMap<String, String>();

		if (StringUtils.isBlank(fileName)) {
			return ret;
		}

		String fileExt = FilenameUtils.getExtension(fileName);

		Path pathObj = Paths.get(fileName);
		File mappingFilePath = new File(pathObj.toString());

		if (mappingFilePath.isFile() && mappingFilePath.canRead()) {
			if ("json".equalsIgnoreCase(fileExt)) {
				FileInputStream uploadedInputStream = new FileInputStream(mappingFilePath);
				String inputJsonString = IOUtils.toString(uploadedInputStream, "UTF-8");
				inputJsonString = inputJsonString.trim();
				if (StringUtils.isNotEmpty(inputJsonString)) {
					ret = JsonUtilsV2.jsonToMap(inputJsonString);
				} else {
					LOG.error("Provided json file is empty!!");
				}
			} else {
				CSVFileReader reader = new CSVFileReader();
				ret = reader.getMappingEntries(reader.getCSVReader(fileName, CSVFileReader.SEPARATOR_PIPE));
			}
		} else {
			throw new IllegalArgumentException("Cannot read file " + mappingFilePath);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== PolicyMigrationUtils.readMappingFile()");
		}
		return ret;
	}

	public static void main(String[] args) throws Exception {
		//System.out.println(convertStringToMap("{/policy1/path:/hdfs/policy/path,/policy2/path:/hdfs/policy2/path}",DELIMITER_COLON));
		//System.out.println(convertStringToMap("{/policy1/path|/hdfs/policy/path,/policy3/path|/hdfs/policy3/path}",DELIMITER_PIPE));
		//System.out.println(convertStringToMap("{}",DELIMITER_PIPE));
		//System.out.println(stringToList("READ, WRITE"));
		//String inputFile = "/Users/pagrawal/Desktop/users.json";
		//readUsersMappingFile(inputFile);
		//System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "/abc", "transform"));
		/*
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "file://tmp", "transform"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "file://tmp/", "transform"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "hdfs://tmp/", "transform"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "hdfs://tmp/", "transform"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "hdfs://tmp/output.json", "transform"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "file:/tmp/output.json", "transform"));
		
		inputFile = "/tmp/hdfspolicy.csv";
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "", "convert"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "file://tmp", "convert"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "file://tmp/", "convert"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "hdfs://tmp/", "convert"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "hdfs://tmp/", "convert"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "hdfs://tmp/output.json", "convert"));
		System.out.println(PolicyMigrationUtils.getFileOutPutLocation(inputFile, "file:/tmp/output.json", "convert"));
		*/
		//CSVFileReader reader = new CSVFileReader();
		//reader.readDataFromCustomSeparator(ACL_POLICY_FILE_PATH);
		//System.out.println("_______________________________________________");
	}

}
