/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.policymigration.reader;

import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.policymigration.common.PolicyMigrationUtils;

import com.opencsv.CSVParser;
import com.opencsv.CSVParserBuilder;
import com.opencsv.CSVReader;
import com.opencsv.CSVReaderBuilder;

public class CSVFileReader {

	private static final Log LOG = LogFactory.getLog(CSVFileReader.class);
	public static char SEPARATOR_PIPE = '|';
	public static char SEPARATOR_EQUAL = '=';
	public static char SEPARATOR_COLON = ':';

	public static void main(String[] args) {
		/*String inputFile = "/Users/pagrawal/Desktop/sample_native_hdfs_permissions_5.csv";
		CSVFileReader reader = new CSVFileReader();
		reader.readDataFromCustomSeparator(inputFile);
		inputFile = "/Users/pagrawal/Desktop/users.csv";
		reader.getUsersMappingFromCSVFile(inputFile);
		System.out.println("_______________________________________________");
		*/
	}

	public List<String[]> getCSVReader(String file, char separator) {
		List<String[]> csvContent = new ArrayList<String[]>();
		CSVReader csvReader = null;
		FileReader filereader = null;
		try {
			// Create object of filereader class with csv file as parameter.
			filereader = new FileReader(file);

			// create csvParser object with custom separator
			CSVParser parser = new CSVParserBuilder()
								.withSeparator(separator)
								.build();

			// create csvReader object with parameter filereader and parser
			csvReader = new CSVReaderBuilder(filereader)
									.withCSVParser(parser)
									.build();
			String[] line;
			while ((line = csvReader.readNext()) != null) {
				csvContent.add(line);
			}
		}
		catch (Exception e) {
			LOG.error("Error while reading and parsing the  file: " + file, e);
		} finally {
			if (csvReader != null) {
				try {
					csvReader.close();
					csvReader = null;
				} catch (Exception ex) {
					LOG.error("Error while closing the file: " + file, ex);
				}
			}

			if (filereader != null) {
				try {
					filereader.close();
					filereader = null;
				} catch (Exception ex) {
					LOG.error("Error while closing the file: " + file, ex);
				}
			}
		}
		return csvContent;
	}

	public List<NativeAclPolicy> getNativeACLPolicies(List<String[]> csvContent) {
		List<NativeAclPolicy> ret = new ArrayList<NativeAclPolicy>();
		try {
			// we are going to read data line by line
			if (csvContent != null) {
				for(String[] line : csvContent) {
					List<String> resources = PolicyMigrationUtils.stringToList(line[0]);
					List<String> users = PolicyMigrationUtils.stringToList(line[1]);
					List<String> groups = PolicyMigrationUtils.stringToList(line[2]);
					List<String> permissions = PolicyMigrationUtils.stringToList(line[3]);
					if (CollectionUtils.isEmpty(resources) || CollectionUtils.isEmpty(permissions) || (CollectionUtils.isEmpty(users) && CollectionUtils.isEmpty(groups))) {
						continue;
					}
					NativeAclPolicy policy = new NativeAclPolicy(resources, users, groups, permissions);
					ret.add(policy);
				}
			}
		}
		catch (Exception e) {
			LOG.error("Error while reading and parsing content", e);
		}
		return ret;
	}

	public Map<String, String> getMappingEntries(List<String[]> csvContent) {
		Map<String, String> ret = new HashMap<String, String>();
		try {
			// we are going to read data line by line
			if (csvContent != null) {
				for(String[] line : csvContent) {
					String sourceUserName = line[0];
					String targetUserName = line[1];
					if (StringUtils.isNotBlank(sourceUserName) && StringUtils.isNotBlank(targetUserName)) {
						sourceUserName = sourceUserName.trim();
						targetUserName = targetUserName.trim();
						if (sourceUserName.contains(" ") || targetUserName.contains(" ")) {
							LOG.error("invalid mapping found for {" + sourceUserName + ":" + targetUserName.trim() +"}");
							continue;
						}
						ret.put(sourceUserName, targetUserName);
					}
				}
			}
		}
		catch (Exception e) {
			LOG.error("Error while reading and parsing content", e);
		}
		return ret;
	}
}

