/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.policymigration.transformer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerPolicyResourceSignature;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyItem;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyItemAccess;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyResource;
import org.apache.ranger.policymigration.common.RangerPolicyMigrationConfig;
import org.apache.ranger.policymigration.common.RangerPolicyMigrationConstants;
import org.apache.ranger.policymigration.reader.NativeAclPolicy;
import org.apache.ranger.policymigration.view.RangerExportPolicyList;

public class NativeAclToRangerPolicyConverter {
	private static final Log LOG = LogFactory.getLog(NativeAclToRangerPolicyConverter.class);
	private Configuration 	conf = null;

	public NativeAclToRangerPolicyConverter(Configuration config) {
		this.conf = config;
	}

	/**
	 * Converts the hdfs native policies to ranger policies and returns the
	 * converted policies.
	 *
	 * @return ranger policy list.
	 * @throws Exception If there is error while performing the translation.
	 */
	public RangerExportPolicyList getConvertedPolicies(List<NativeAclPolicy> nativePolicies) throws Exception {
		return performConversion(nativePolicies);
	}

	private RangerExportPolicyList performConversion(List<NativeAclPolicy> nativePolicies) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> NativeAclToRangerPolicyConverter.performConversion()");
		}

		long totalPermissionsCount = 0l;
		long convertedRangerPoliciesCount = 0l;
		RangerExportPolicyList ret = null;
		if (nativePolicies != null) {
			if (CollectionUtils.isNotEmpty(nativePolicies)) {
				List<RangerPolicy> rangerPolicies = new ArrayList<RangerPolicy>();
				String serviceName = conf.get(RangerPolicyMigrationConfig.NATIVE_POLICIES_SERVICE_NAME, "cm_hdfs");
				totalPermissionsCount = nativePolicies.size();
				for (NativeAclPolicy nativeAclPolicy : nativePolicies) {
					//path string to RangerPolicyResource
					Map<String, RangerPolicyResource> policyResources = new HashMap<String, RangerPolicyResource>();
					RangerPolicyResource resource = new RangerPolicyResource(nativeAclPolicy.getResourcePaths(), false, true);
					policyResources.put(RangerPolicyMigrationConstants.RESOURCE_TYPE_PATH, resource);
					
					//permissions string to RangerPolicyItemAccess
					List<RangerPolicyItemAccess> accesses = new ArrayList<RangerPolicyItemAccess>();
					for (String permission : nativeAclPolicy.getPermissions()) {
						accesses.add(new RangerPolicyItemAccess(permission, true));
					}

					//RangerPolicyItem with permissions, users and groups
					RangerPolicyItem  policyItem = new RangerPolicyItem(accesses, nativeAclPolicy.getUsers(), nativeAclPolicy.getGroups(), null, null, null);
					List<RangerPolicyItem> policyItems = new ArrayList<RangerPolicyItem>();
					policyItems.add(policyItem);

					RangerPolicy rangerPolicy = new RangerPolicy();
					rangerPolicy.setId(null);
					rangerPolicy.setResourceSignature(null);
					rangerPolicy.setCreatedBy(null);
					rangerPolicy.setUpdatedBy(null);
					rangerPolicy.setCreateTime(null);
					rangerPolicy.setUpdateTime(null);
					rangerPolicy.setService(serviceName);
					rangerPolicy.setServiceType(RangerPolicyMigrationConstants.LITERAL_HDFS);
					rangerPolicy.setResources(policyResources);
					rangerPolicy.setPolicyItems(policyItems);
					setPolicyNameAndGUID(rangerPolicy);
					rangerPolicies.add(rangerPolicy);
					if (LOG.isDebugEnabled()) {
						LOG.debug("Converted RangerPolicy:"+rangerPolicy);
					}

				}

				rangerPolicies = performCompaction(rangerPolicies);

				if (CollectionUtils.isNotEmpty(rangerPolicies)) {
					ret = new RangerExportPolicyList();
					ret.setPolicies(rangerPolicies);
					convertedRangerPoliciesCount = rangerPolicies.size();
				}
			}
		}

		LOG.info("Native ACL entries count="+totalPermissionsCount+", Converted Ranger policies count="+convertedRangerPoliciesCount);
		System.out.println("Native ACL entries count="+totalPermissionsCount+", Converted Ranger policies count="+convertedRangerPoliciesCount);

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== NativeAclToRangerPolicyConverter.performConversion()");
		}
		return ret;
	}

	private void setPolicyNameAndGUID(RangerPolicy policy) {
		if (StringUtils.isBlank(policy.getName())) {
			String policyNamePrefix  = conf.get(RangerPolicyMigrationConfig.NATIVE_POLICIES_POLICY_NAME_PREFIX, StringUtils.EMPTY);
			String policyDescription = conf.get(RangerPolicyMigrationConfig.NATIVE_POLICIES_POLICY_DESCRIPTION, StringUtils.EMPTY);

			String guid = policy.getGuid();
			if (StringUtils.isBlank(guid)) {
				guid = UUID.randomUUID().toString();
				policy.setGuid(guid);
			}

			if (StringUtils.isBlank(policyNamePrefix)) {
				policyNamePrefix = policy.getService();
			}
			String policyName = policyNamePrefix + "-" + guid;
			policy.setName(policyName);
			policy.setDescription(policyDescription);
		}
	}

	private List<RangerPolicy> performCompaction(List<RangerPolicy> rangerPolicies) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> NativeAclToRangerPolicyConverter.performCompaction():" + rangerPolicies);
		}

		if (CollectionUtils.isNotEmpty(rangerPolicies)) {
			LOG.info("Number of Ranger Policies before compaction:" + rangerPolicies.size());
			Iterator<RangerPolicy> itrRangerPolicies = rangerPolicies.iterator();
			LinkedHashMap<String, RangerPolicy> signAndRangerPolicyMap = new LinkedHashMap<String, RangerPolicy>();
			while (itrRangerPolicies.hasNext()) {
				RangerPolicy newRangerPolicy = itrRangerPolicies.next();
				String signature = (new RangerPolicyResourceSignature(newRangerPolicy)).getSignature();
				if (signAndRangerPolicyMap.containsKey(signature)) {
					RangerPolicy existingRangerPolicy = signAndRangerPolicyMap.get(signature);
					existingRangerPolicy.getPolicyItems().addAll(newRangerPolicy.getPolicyItems());
				} else {
					signAndRangerPolicyMap.put(signature, newRangerPolicy);
				}
			}
			rangerPolicies = new ArrayList<>(signAndRangerPolicyMap.values());
			LOG.info("Number of Ranger Policies after compaction:" + rangerPolicies.size());
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== NativeAclToRangerPolicyConverter.performCompaction():" + rangerPolicies);
		}
		return rangerPolicies;
	}

}