/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.policymigration.transformer;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerPolicy.RangerDataMaskPolicyItem;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyItem;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyItemAccess;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyResource;
import org.apache.ranger.plugin.model.RangerPolicy.RangerRowFilterPolicyItem;
import org.apache.ranger.plugin.model.RangerService;
import org.apache.ranger.policymigration.common.PolicyMigrationUtils;
import org.apache.ranger.policymigration.common.RangerPolicyMigrationConfig;
import org.apache.ranger.policymigration.common.RangerPolicyMigrationConstants;
import org.apache.ranger.policymigration.view.RangerExportPolicyList;
import org.apache.ranger.policymigration.view.RangerServiceList;

public class TransformTask {
	private static final Log LOG = LogFactory.getLog(TransformTask.class);
	private Configuration conf = null;
	private Map<String, String> serviceTypeMappings;
	private Map<String, String> serviceNameMappings;
	private Map<String, String> zoneNameMappings;
	private Map<String, String> serviceResourceMappings;
	private Map<String, String> hiveURLMappings;
	private Map<String, String> userMappings;
	private List<String> excludePoliciesFromServiceTypes;

	public TransformTask(Configuration conf) {
		this.conf 		= conf;
		this.serviceTypeMappings = PolicyMigrationUtils.convertStringToMap(conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_SERVICE_TYPE_MAPPING, ""), PolicyMigrationUtils.DELIMITER_COLON);
		LOG.info("serviceTypeMappings:"+Collections.singletonList(this.serviceTypeMappings)); 
		this.serviceNameMappings = PolicyMigrationUtils.convertStringToMap(conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_SERVICE_NAME_MAPPING, ""), PolicyMigrationUtils.DELIMITER_COLON);
		LOG.info("serviceNameMappings:"+Collections.singletonList(this.serviceNameMappings)); 
		this.zoneNameMappings = PolicyMigrationUtils.convertStringToMap(conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_ZONE_NAME_MAPPING, ""), PolicyMigrationUtils.DELIMITER_COLON);
		LOG.info("zoneNameMappings:"+Collections.singletonList(this.zoneNameMappings)); 

		//load resource mapping
		String resourceMappingFile = conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_RESOURCE_MAPPING_FILE, "");
		try {
			this.serviceResourceMappings = PolicyMigrationUtils.readMappingFile(resourceMappingFile);
		} catch (Exception e) {
			LOG.error("Error reading file: "+ resourceMappingFile, e);
		}
		LOG.info("serviceResourceMappings:"+Collections.singletonList(this.serviceResourceMappings));

		//load hive url mapping
		String hiveurlMappingFile = conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_HIVE_URL_MAPPING_FILE, "");
		try {
			this.hiveURLMappings = PolicyMigrationUtils.readMappingFile(hiveurlMappingFile);
		} catch (Exception e) {
			LOG.error("Error reading file: "+ hiveurlMappingFile, e);
		}
		LOG.info("hiveURLMappings:"+Collections.singletonList(this.hiveURLMappings));

		//load users mapping
		String usersMappingFile = RangerPolicyMigrationConfig.getProperty(conf, RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_USERS_MAPPING_FILE);
		try {
			userMappings = PolicyMigrationUtils.readMappingFile(usersMappingFile);
		} catch (Exception e) {
			LOG.error("Error reading file:"+usersMappingFile, e);
		}

		this.excludePoliciesFromServiceTypes = PolicyMigrationUtils.stringToList(conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_EXCLUDE_SERVICE_TYPES, RangerPolicyMigrationConstants.LITERAL_STORM));
		LOG.info("excludeServiceNames:"+Collections.singletonList(this.excludePoliciesFromServiceTypes)); 
	}

	/**
	 * Transforms the ranger policies to target env ranger policies and returns the
	 * Transformed policies.
	 *
	 * @return ranger policy list.
	 * @throws Exception If there is error while performing the Transformation.
	 */
	public RangerExportPolicyList getTransformedPolicies(RangerExportPolicyList rangerExportPolicyList, RangerServiceList rangerServiceList) throws Exception {
		return performTransformation(rangerExportPolicyList, rangerServiceList);
	}

	private RangerExportPolicyList performTransformation(RangerExportPolicyList rangerExportPolicyList, RangerServiceList rangerServiceList) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> TransformTask.performTransformation()");
		}
		long totalPolicyCount = 0l;
		long skippedPolicyCount = 0l;
		Map<String, String> serviceNameAndTypeMap = loadServiceNameAndTypeMapping(rangerServiceList);
		if (rangerExportPolicyList != null) {
			if (CollectionUtils.isNotEmpty(rangerExportPolicyList.getPolicies())) {
				totalPolicyCount = rangerExportPolicyList.getPolicies().size();
				LOG.info("Total Ranger policies:"+totalPolicyCount);
				Iterator<RangerPolicy> itrRangerPolicy = rangerExportPolicyList.getPolicies().iterator();
				while (itrRangerPolicy.hasNext()) {
					RangerPolicy rangerPolicy = itrRangerPolicy.next();
					if (LOG.isDebugEnabled()) {
						LOG.debug("Transforming Ranger Policy:"+ rangerPolicy);
					}
					if (StringUtils.isBlank(rangerPolicy.getServiceType())) {
						if (serviceNameAndTypeMap == null || serviceNameAndTypeMap.isEmpty()) {
							throw new Exception("Please provide Ranger service json file path with the transform command to update the service types in the Ranger policies");
						}
						if (serviceNameAndTypeMap.containsKey(rangerPolicy.getService())) {
							rangerPolicy.setServiceType(serviceNameAndTypeMap.get(rangerPolicy.getService()));
						} else {
							LOG.warn("ServiceType field is empty in the ranger policy:"+ rangerPolicy.getName());
							skippedPolicyCount++;
							continue;
						}
					} else if(RangerPolicyMigrationConstants.SERVICE_DEF_ID_NAME_MAP.containsKey(rangerPolicy.getServiceType())) {
						rangerPolicy.setServiceType(RangerPolicyMigrationConstants.SERVICE_DEF_ID_NAME_MAP.get(rangerPolicy.getServiceType()));
					}

					String oldServiceType = rangerPolicy.getServiceType();

					//skip storm policy
					if (excludePoliciesFromServiceTypes.contains(oldServiceType)) {
						itrRangerPolicy.remove();
						skippedPolicyCount++;
						continue;
					}

					//replace service type mapping
					if (RangerPolicyMigrationConstants.LITERAL_HDFS.equalsIgnoreCase(oldServiceType)) {
						if (serviceTypeMappings.containsKey(oldServiceType)) {
							rangerPolicy.setServiceType(serviceTypeMappings.get(oldServiceType).toLowerCase());
						}
					}

					//replace service name mappings
					if (serviceNameMappings.containsKey(rangerPolicy.getService())) {
						rangerPolicy.setService(serviceNameMappings.get(rangerPolicy.getService()));
						if (LOG.isDebugEnabled()) {
							LOG.debug("Target policy service name:"+rangerPolicy.getService());
						}
					}

					//replace zone mappings
					if (zoneNameMappings.containsKey(rangerPolicy.getZoneName())) {
						rangerPolicy.setZoneName(zoneNameMappings.get(rangerPolicy.getZoneName()));
					}

					//replace resource mappings
					Map<String, RangerPolicyResource> policyResources = rangerPolicy.getResources();
					for (Map.Entry<String, RangerPolicyResource> entry : policyResources.entrySet()) {
						List<String> listResources = entry.getValue().getValues();
						// to replace hive url resource start string with mapped s3 string.
						if (RangerPolicyMigrationConstants.LITERAL_HIVE.equalsIgnoreCase(oldServiceType) && "url".equalsIgnoreCase(entry.getKey())) {
							Set<String> resourceMappingKeys = hiveURLMappings.keySet();
							if (resourceMappingKeys != null && resourceMappingKeys.size() >0) {
								for (int index=0; index < listResources.size(); index ++) {
									for (String sourceStartStr : resourceMappingKeys) {
										if (StringUtils.startsWith(listResources.get(index), sourceStartStr)) {
											String targetStartStr = hiveURLMappings.get(sourceStartStr);
											String newResourceStr = listResources.get(index).replaceFirst(sourceStartStr, targetStartStr);
											listResources.set(index, newResourceStr);
										}
									}
								}
							}
						} else {
							for (int index=0; index < listResources.size(); index ++) {
								if (serviceResourceMappings.containsKey(listResources.get(index))) {
									LOG.info("matchedresource:"+serviceResourceMappings.get(listResources.get(index)));
									listResources.set(index, serviceResourceMappings.get(listResources.get(index)));
								}
							}
						}
					}

					//remove invalid access types (storm accesses) from tag service def
					removeInvalidAccessTypesFromPolicyItems(oldServiceType, rangerPolicy.getPolicyItems());
					removeInvalidAccessTypesFromPolicyItems(oldServiceType, rangerPolicy.getDenyPolicyItems());
					removeInvalidAccessTypesFromPolicyItems(oldServiceType, rangerPolicy.getAllowExceptions());
					removeInvalidAccessTypesFromPolicyItems(oldServiceType, rangerPolicy.getDenyExceptions());
					removeInvalidAccessTypesFromDataMaskPolicyItems(oldServiceType, rangerPolicy.getDataMaskPolicyItems());
					removeInvalidAccessTypesFromRowFilterPolicyItems(oldServiceType, rangerPolicy.getRowFilterPolicyItems());
					
					//S3 : add bucket
					if (RangerPolicyMigrationConstants.LITERAL_S3.equalsIgnoreCase(serviceTypeMappings.get(oldServiceType))) {
						String s3BucketName = conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_S3_BUCKET_NAME, "myS3bucket");
						RangerPolicyResource res = new RangerPolicyResource(s3BucketName);
						rangerPolicy.getResources().put("bucket", res);
					}

					//ADLS : add account and container
					if (RangerPolicyMigrationConstants.LITERAL_ADLS.equalsIgnoreCase(serviceTypeMappings.get(oldServiceType))) {
						String storageAccount = conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_ADLS_STORAGE_ACCOUNT, "mystorage");
						String storageContainer = conf.get(RangerPolicyMigrationConfig.RANGER_POLICYMIGRATION_ADLS_STORGAGE_CONTAINER, "mycontainer");
						RangerPolicyResource resourceStorageAccount = new RangerPolicyResource(storageAccount);
						RangerPolicyResource resourceStorageContainer = new RangerPolicyResource(storageContainer);
						rangerPolicy.getResources().put("storageaccount", resourceStorageAccount);
						rangerPolicy.getResources().put("container", resourceStorageContainer);

						//convert resource type string path to relative path
						if (rangerPolicy.getResources().containsKey("path")) {
							RangerPolicyResource relativePath = rangerPolicy.getResources().get("path");
							rangerPolicy.getResources().put("relativepath", relativePath);
							rangerPolicy.getResources().remove("path");
						}
					}

					//replace user name mappings
					replaceUserNamesFromPolicyItems(rangerPolicy.getPolicyItems());
					replaceUserNamesFromPolicyItems(rangerPolicy.getDenyPolicyItems());
					replaceUserNamesFromPolicyItems(rangerPolicy.getAllowExceptions());
					replaceUserNamesFromPolicyItems(rangerPolicy.getDenyExceptions());
					replaceUserNamesFromDataMaskPolicyItems(rangerPolicy.getDataMaskPolicyItems());
					replaceUserNamesFromRowFilterPolicyItem(rangerPolicy.getRowFilterPolicyItems());

					rangerPolicy.setId(null);
					rangerPolicy.setGuid(null);
					rangerPolicy.setResourceSignature(null);
					rangerPolicy.setCreatedBy(null);
					rangerPolicy.setUpdatedBy(null);
					rangerPolicy.setCreateTime(null);
					rangerPolicy.setUpdateTime(null);
				}
			}
		}

		LOG.info("Ranger Policies Count:[Total="+totalPolicyCount+", Skipped="+skippedPolicyCount+"]");
		System.out.println("Ranger Policies Count:[Total="+totalPolicyCount+", Skipped="+skippedPolicyCount+"]");

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== TransformTask.performTransformation()");
		}
		return rangerExportPolicyList;
	}

	private void replaceUserNamesFromPolicyItems(List<RangerPolicyItem> rangerPolicyItems) {
		for (RangerPolicyItem  policyItem : rangerPolicyItems) {
			policyItem.setUsers(replaceUserNames(policyItem.getUsers()));
		}
	}

	private void replaceUserNamesFromDataMaskPolicyItems(List<RangerDataMaskPolicyItem> rangerPolicyItems) {
		for (RangerDataMaskPolicyItem  policyItem : rangerPolicyItems) {
			policyItem.setUsers(replaceUserNames(policyItem.getUsers()));
		}
	}

	private void replaceUserNamesFromRowFilterPolicyItem(List<RangerRowFilterPolicyItem> rangerPolicyItems) {
		for (RangerRowFilterPolicyItem  policyItem : rangerPolicyItems) {
			policyItem.setUsers(replaceUserNames(policyItem.getUsers()));
		}
	}

	private List<String> replaceUserNames(List<String> users) {
		for (int index=0; index < users.size(); index ++) {
			String userName = users.get(index);
			if (MapUtils.isNotEmpty(userMappings) && userMappings.containsKey(userName)) {
				users.set(index, userMappings.get(userName));
			}
		}
		return users;
	}

	private void removeInvalidAccessTypesFromPolicyItems(String serviceType, List<RangerPolicyItem> rangerPolicyItems) {
		Iterator<RangerPolicyItem> itrRangerPolicyItems = rangerPolicyItems.iterator();
		while (itrRangerPolicyItems.hasNext()) {
			RangerPolicyItem  policyItem = itrRangerPolicyItems.next();
			removeInvalidAccessTypes(serviceType, policyItem.getAccesses());
			if (CollectionUtils.isEmpty(policyItem.getAccesses())) {
				itrRangerPolicyItems.remove();
			}
		}
	}

	private void removeInvalidAccessTypesFromDataMaskPolicyItems(String serviceType, List<RangerDataMaskPolicyItem> rangerPolicyItems) {
		Iterator<RangerDataMaskPolicyItem> itrRangerPolicyItems = rangerPolicyItems.iterator();
		while (itrRangerPolicyItems.hasNext()) {
			RangerPolicyItem  policyItem = itrRangerPolicyItems.next();
			removeInvalidAccessTypes(serviceType, policyItem.getAccesses());
			if (CollectionUtils.isEmpty(policyItem.getAccesses())) {
				itrRangerPolicyItems.remove();
			}
		}
	}

	private void removeInvalidAccessTypesFromRowFilterPolicyItems(String serviceType, List<RangerRowFilterPolicyItem> rangerPolicyItems) {
		Iterator<RangerRowFilterPolicyItem> itrRangerPolicyItems = rangerPolicyItems.iterator();
		while (itrRangerPolicyItems.hasNext()) {
			RangerPolicyItem  policyItem = itrRangerPolicyItems.next();
			removeInvalidAccessTypes(serviceType, policyItem.getAccesses());
			if (CollectionUtils.isEmpty(policyItem.getAccesses())) {
				itrRangerPolicyItems.remove();
			}
		}
	}

	private void removeInvalidAccessTypes(String serviceType, List<RangerPolicyItemAccess> accesses) {
		if (RangerPolicyMigrationConstants.LITERAL_TAG.equalsIgnoreCase(serviceType)) {
			updateAccessTypesForADLSandS3(accesses);
		}
		if (RangerPolicyMigrationConstants.NONSUPPORTED_ACCESS_TYPES_MAP.containsKey(serviceType)) {
			Set<String> invalidAccessTypes = RangerPolicyMigrationConstants.NONSUPPORTED_ACCESS_TYPES_MAP.get(serviceType);
			Iterator<RangerPolicyItemAccess> itrAccesses = accesses.iterator();
			while (itrAccesses.hasNext()) {
				RangerPolicyItemAccess access = itrAccesses.next();
				if (invalidAccessTypes.contains(access.getType())) {
					if (LOG.isDebugEnabled()) {
						LOG.debug("removing permission:"+access.getType());
					}
					itrAccesses.remove();
				}
			}
		}
	}

	private void updateAccessTypesForADLSandS3(List<RangerPolicyItemAccess> accesses) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> TransformTask.updateAccessTypesForADLSandS3(accesses):" + accesses);
		}
		Iterator<RangerPolicyItemAccess> itrAccesses = accesses.iterator();
		int index = -1;
		while (itrAccesses.hasNext()) {
			index = index + 1;
			RangerPolicyItemAccess access = itrAccesses.next();

			if (serviceTypeMappings.containsKey(RangerPolicyMigrationConstants.LITERAL_HDFS)) {
				String targetServiceType = serviceTypeMappings.get(RangerPolicyMigrationConstants.LITERAL_HDFS).toLowerCase();
				if (RangerPolicyMigrationConstants.LITERAL_S3.equalsIgnoreCase(targetServiceType) || RangerPolicyMigrationConstants.LITERAL_ADLS.equalsIgnoreCase(targetServiceType)) {
					//remove if access type is hdfs:execute
					if (RangerPolicyMigrationConstants.LITERAL_S3.equalsIgnoreCase(targetServiceType) && RangerPolicyMigrationConstants.ACCESS_TYPE_HDFS_EXECUTE.equalsIgnoreCase(access.getType())) {
						if (LOG.isDebugEnabled()) {
							LOG.debug("removing permission:"+access.getType());
						}
						itrAccesses.remove();
					} else if (access.getType().startsWith(RangerPolicyMigrationConstants.LITERAL_HDFS+PolicyMigrationUtils.DELIMITER_COLON)) {
						if (LOG.isDebugEnabled()) {
							LOG.info("replacing access Type:"+access.getType());
						}
						//replace if access Types are of s3 or adls
						access.setType(access.getType().replaceFirst(RangerPolicyMigrationConstants.LITERAL_HDFS+PolicyMigrationUtils.DELIMITER_COLON, targetServiceType+PolicyMigrationUtils.DELIMITER_COLON));
						accesses.set(index, access);
					}
				}
			}
		}
		if (LOG.isDebugEnabled()) {
			LOG.debug("<== TransformTask.performTransformation(accesses):" + accesses);
		}
	}

	private Map<String, String> loadServiceNameAndTypeMapping(RangerServiceList rangerServiceList) {
		Map<String, String> ret = new HashMap<String, String>();
		if (rangerServiceList != null) {
			List<RangerService> rangerServices= rangerServiceList.getServices();
			if (CollectionUtils.isNotEmpty(rangerServices)) {
				for (RangerService rangerService : rangerServices) {
					ret.put(rangerService.getName(), rangerService.getType());
				}
			}
		}
		return ret;
	}
}