/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.ranger.raz.processor.abfsutil;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.xml.XmlMapper;
import com.azure.storage.blob.models.UserDelegationKey;
import com.sun.jersey.api.client.ClientResponse;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.audit.provider.MiscUtil;
import org.apache.ranger.raz.intg.RangerRazErrorCode;
import org.apache.ranger.raz.intg.client.RangerRazClient;
import org.apache.ranger.raz.intg.RangerRazException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.security.PrivilegedAction;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.*;

public class AbfsTokenProvider {
    private static final Logger LOG = LoggerFactory.getLogger(AbfsTokenProvider.class);

    private static final String REST_SSL_FILE_LOC                                     = "ranger.raz.policy.rest.ssl.config.file";
    private static final String REST_CLIENT_CONN_TIMEOUT_MS                           = "ranger.raz.rest.client.read.timeoutMs";
    private static final String REST_CLIENT_READ_TIMEOUT_MS                           = "ranger.raz.rest.client.read.timeoutMs";
    private static final String REST_IDBROKER_INIT_RETRY_INTERVAL_SECONDS             = "ranger.raz.rest.idbroker.init.retry.interval.seconds";
    private static final String AZURE_STORAGE_ACCOUNT                                 = "ranger.raz.azure.storage.accounts";
    private static final String USER_DELEGATION_KEY_VALIDITY_MINUTES                  = "ranger.raz.azure.userdelegationkey.validity.minutes";
    private static final String KNOX_DELEGATION_TOKEN_EXPIRY_TIME_PRECHECK_IN_SECONDS = "ranger.raz.knox.delegation.token.expiry.time.precheck.in.seconds";
    private static final String AZURE_ACCESSTOKEN_EXPIRY_TIME_PRECHECK_IN_SECONDS     = "ranger.raz.azure.access.token.expiry.time.precheck.in.seconds";
    private static final String USE_USER_DELEGATION_KEY_FROM_CONFIG                   = "ranger.raz.azure.use.user.delegation.key.from.config";
    private static final String SAS_EXPIRY_TIME_IN_SECONDS                            = "ranger.raz.azure.sas.expiry.time.in.seconds";
    private static final String ID_BROKER_HOST_URL                                    = "fs.azure.ext.cab.address";
    private static final String ID_BROKER_API_GET_DELEGATION_TOKEN                    = "/dt/knoxtoken/api/v1/token";
    private static final String ID_BROKER_API_GET_ACCESS_TOKEN                        = "/azure-cab/cab/api/v1/credentials";
    private static final String URL_SEPARATOR                                         = "/";
    private static final String STR_ACCESS_TOKEN_EQALS                                = "acessToken = ";

    private final Configuration                       conf;
    private final String                              sslConfigFileName;
    private final int                                 restClientConnTimeOutMs;
    private final int                                 idBrokerInitRetryIntervalSeconds;
    private final int                                 restClientReadTimeOutMs;
    private final int                                 userDelegationKeyValidityMinutes;
    private final boolean                             useUserDelegationKeyFromConfig;
    private final Map<String, DelegationSASGenerator> dsasGeneratorCache = new ConcurrentHashMap<>();
    private final Map<String,Object>                  locks              = new ConcurrentHashMap<>();

    private volatile RESTClient              idBrokerClient          = null;
    private volatile IdBrokerDelegationToken idBrokerDelegationToken = null;
    private volatile AccessToken             accessToken             = null;


    public AbfsTokenProvider(Configuration conf) throws RangerRazException {
        if (conf == null) {
            throw new RangerRazException(RangerRazErrorCode.INTERNAL_ERROR, "AbfsTokenProvider(): conf is null");
        }

        if (LOG.isDebugEnabled()) {
            Iterator<Map.Entry<String, String>> confIterator = conf.iterator();

            while (confIterator.hasNext()) {
                Map.Entry<String, String> entry = confIterator.next();

                LOG.debug("AbfsTokenProvider: " + entry.getKey() + "=" + entry.getValue());
            }
        }

        this.conf                             = conf;
        this.sslConfigFileName                = conf.get(REST_SSL_FILE_LOC);
        this.restClientConnTimeOutMs          = conf.getInt(REST_CLIENT_CONN_TIMEOUT_MS, 120 * 1000);           // 120 seconds;
        this.restClientReadTimeOutMs          = conf.getInt(REST_CLIENT_READ_TIMEOUT_MS, 30 * 1000);            //  30 seconds;
        this.idBrokerInitRetryIntervalSeconds = conf.getInt(REST_IDBROKER_INIT_RETRY_INTERVAL_SECONDS, 15);
        this.userDelegationKeyValidityMinutes = conf.getInt(USER_DELEGATION_KEY_VALIDITY_MINUTES, 7 * 24 * 60); //   7 days
        this.useUserDelegationKeyFromConfig   = conf.getBoolean(USE_USER_DELEGATION_KEY_FROM_CONFIG, false);

         Thread initTask = new Thread(new IdBrokerInitTask());
         initTask.setDaemon(true);
         initTask.start();
    }

    public DelegationSASGenerator getDelegationSASGenerator(String storageAccount) throws IOException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> AbfsTokenProvider.getDelegationSASGenerator(storageAccount=" + storageAccount + ")");
        }

        DelegationSASGenerator ret = dsasGeneratorCache.get(storageAccount);

        if (ret == null || isUserDelegationKey_renewalRequired(ret.getUserDelegationKey())) {
            //get a new UserDelegationKey.Synchorize this call so only one request from the client per storage account goes
            //to abfs for userDelegationKey fetch.

            Object lock = locks.get(storageAccount);

            if (lock == null) {
                synchronized (locks) {
                    lock = locks.get(storageAccount);

                    if (lock == null) {
                        lock = new Object();

                        locks.put(storageAccount, lock);
                    }
                }
            }

            synchronized (lock) {
                ret = dsasGeneratorCache.get(storageAccount);

                if (ret == null || isUserDelegationKey_renewalRequired(ret.getUserDelegationKey())) {
                    UserDelegationKey newDelgationKey = fetchUserDelegationKey(storageAccount);

                    if (newDelgationKey != null) {
                        ret = new DelegationSASGenerator(storageAccount, newDelgationKey);

                        cacheDelegationSASGenerator(ret);
                    } else {
                        if (ret != null) {
                            LOG.warn("Failed to get userDelegationKey for storageAccount:[" + storageAccount + "}, will continue to use existing key");
                        } else {
                            LOG.warn("Failed to get userDelegationKey for storageAccount:[" + storageAccount + "}y");
                        }
                    }
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== AbfsTokenProvider.getDelegationSASGenerator(storageAccount=" + storageAccount + ") : " + ret == null ? " failed" : "successful");
        }

        return ret;
    }

    private AccessToken getAccessToken() throws IOException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> AbfsTokenProvider.getAccessToken()");
        }

        AccessToken ret = accessToken;

        if (ret == null || ret.isRenewalRequired()) {
            IdBrokerDelegationToken dt = getIdBrokerDelegationToken();

            if (dt == null) {
                LOG.error("Failed to get IdBroker delegationToken");
            } else {
                synchronized (this) {
                    ret = accessToken;

                    if (ret == null || ret.isRenewalRequired()) {
                        ret = accessToken = fetchAccessToken(dt);
                    }
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== AbfsTokenProvider.getAccessToken() : " + ret == null ? "failed" : "successful");
        }

        return ret;
    }

    private IdBrokerDelegationToken getIdBrokerDelegationToken() throws IOException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> AbfsTokenProvider.getIdBrokerDelegationToken()");
        }

        final IdBrokerDelegationToken ret;

        if (idBrokerDelegationToken != null && !idBrokerDelegationToken.isRenewalRequired()) {
            ret = idBrokerDelegationToken;
        } else {
            synchronized (this) {
                if (idBrokerDelegationToken != null && !idBrokerDelegationToken.isRenewalRequired()) {
                    ret = idBrokerDelegationToken;
                } else {
                    ret = idBrokerDelegationToken = fetchIdbrokerDelegationToken();
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== AbfsTokenProvider.getIdBrokerDelegationToken() : " + ret == null ? "failed" : "successful");
        }

        return ret;
    }

    private IdBrokerDelegationToken fetchIdbrokerDelegationToken() throws IOException {
        LOG.info("==> AbfsTokenProvider.fetchIdbrokerDelegationToken()");

        IdBrokerDelegationToken ret            = null;
        ObjectMapper                      mapper         = new ObjectMapper();
        UserGroupInformation              ugi            = UserGroupInformation.getCurrentUser();
        PrivilegedAction<ClientResponse>  action         = getIdBrokerResponse(ID_BROKER_API_GET_DELEGATION_TOKEN, null, null, null);
        ClientResponse                    clientResponse = executeAction(ugi, action);
        String                            response       = parseResponse(clientResponse, String.class);

        if (response != null) {
            if (StringUtils.isNotEmpty(response)) {
                JsonNode responseObj = mapper.readTree(response);

                if (responseObj != null) {
                    String token    = responseObj.findValue("access_token").asText();
                    String expiryDt = responseObj.findValue("expires_in").asText();

                    ret = new IdBrokerDelegationToken(token, new Date(new Long(expiryDt)));
                } else {
                    LOG.error("Cannot parse response string received from IDBroker:[" + response + "]");
                }
            } else {
                LOG.error("Empty response string received from IDBroker");
            }
        } else {
            LOG.error("Null response object received from IDBroker");
        }

        LOG.info("<== AbfsTokenProvider.fetchIdbrokerDelegationToken(): {}", (ret == null ? " failed" : "successful"));

        return ret;
    }

    private AccessToken fetchAccessToken(IdBrokerDelegationToken idBrokerDelegationToken) throws IOException {
        LOG.info("==> AbfsTokenProvider.fetchAccessToken()");

        AccessToken          ret     = null;
        ObjectMapper         mapper  = new ObjectMapper();
        UserGroupInformation ugi     = UserGroupInformation.getCurrentUser();
        Map<String, String>  headers = new HashMap<>();
        String               value   = AbfsUtilConstants.HEADER_BEARER + idBrokerDelegationToken.getDelegationToken();

        headers.put(AbfsUtilConstants.HEADER_AUTHORIZATION, value);

        PrivilegedAction<ClientResponse>  action         = getIdBrokerResponse(ID_BROKER_API_GET_ACCESS_TOKEN, headers, null, null);
        ClientResponse                    clientResponse = executeAction(ugi, action);
        String                            response       = parseResponse(clientResponse, String.class);

        if (response != null) {
            if (StringUtils.isNotEmpty(response)) {
                JsonNode responseObj = mapper.readTree(response);

                if (responseObj != null) {
                    String token      = responseObj.findValue("access_token").asText();
                    String expires_in = StringUtils.EMPTY;
                    JsonNode jsonNode = responseObj.findValue("expires_in");
                    if (jsonNode != null) {
                        expires_in = jsonNode.asText();
                    }
                    String expires_on = responseObj.findValue("expires_on").asText();

                    ret = new AccessToken(token, expires_on, expires_in);
                } else {
                    LOG.error("fetchAccessToken(): failed to parse response received from IDBroker: {}", response);
                }
            } else {
                LOG.error("fetchAccessToken(): empty response received from IDBroker");
            }
        } else {
            LOG.error("fetchAccessToken(): null response received from IDBroker");
        }

        LOG.info("<== AbfsTokenProvider.fetchAccessToken(): {}", (ret == null ? " failed" : "successful"));

        return ret;
    }

    private UserDelegationKey fetchUserDelegationKey(String storageAccount) throws IOException {
        LOG.info("==> AbfsTokeProvider.fetchUserDelegationKey(storageAccount={})", storageAccount);

        UserDelegationKey ret         = null;
        RESTClient        restClient  = buildAzureClient(storageAccount);
        AccessToken       accessToken = getAccessToken();

        if (accessToken != null && !accessToken.isRenewalRequired()) {
            ret = fetchUserDelegationKeyWithAccessToken(restClient, "", accessToken);
        } else if (useUserDelegationKeyFromConfig) {
            ret = getUserDelegationKeyFromConfig(storageAccount);
        }

        if (ret == null) {
            LOG.error("fetchUserDelegationKey(storageAccount={}): invalid accessToken");
        }

        LOG.info("<== AbfsTokeProvider.fetchUserDelegationKey(storageAccount={}): {}", storageAccount, (ret == null ? "failed" : "successful"));

        return ret;
    }

    private UserDelegationKey getUserDelegationKeyFromConfig(String storageAccount) {
        String prefix = "ranger.raz.adls.user.delegation.key.";

        String         signedOid        = conf.get(prefix + storageAccount + ".oid");
        String         signedTid        = conf.get(prefix + storageAccount + ".tid");
        String         signedService    = conf.get(prefix + storageAccount + ".service");
        String         signedStart      = conf.get(prefix + storageAccount + ".start");
        String         signedExpiry     = conf.get(prefix + storageAccount + ".expiry");
        String         signedVersion    = conf.get(prefix + storageAccount + ".version", "2018-11-09");
        String         signedValue      = conf.get(prefix + storageAccount + ".value");
        OffsetDateTime signedStartDate  = OffsetDateTime.parse(signedStart, DateTimeFormatter.ISO_OFFSET_DATE_TIME);
        OffsetDateTime signedExpiryDate = OffsetDateTime.parse(signedExpiry, DateTimeFormatter.ISO_OFFSET_DATE_TIME);

        UserDelegationKey ret =  new UserDelegationKey()
                                            .setSignedObjectId(signedOid)
                                            .setSignedTenantId(signedTid)
                                            .setSignedService(signedService)
                                            .setSignedStart(signedStartDate)
                                            .setSignedExpiry(signedExpiryDate)
                                            .setSignedVersion(signedVersion)
                                            .setValue(signedValue);

        return ret;
    }

    private UserDelegationKey fetchUserDelegationKeyWithAccessToken(RESTClient azureRestClient, String relativeUrl, AccessToken accessToken) throws IOException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> AbfsTokenProvider.fetchUserDelegationKeyWithAccessToken()" + " relativeUrl =" +  relativeUrl, "AccessToken =" + accessToken );
        }

        UserDelegationKey    ret     = null;
        UserGroupInformation ugi     = UserGroupInformation.getCurrentUser();
        Map<String, String>  headers = new HashMap<>();
        Map<String, String>  params  = new HashMap<>();
        String               keyInfo = getExpiryKeyInfo(userDelegationKeyValidityMinutes);

        headers.put(AbfsUtilConstants.HEADER_AUTHORIZATION, AbfsUtilConstants.HEADER_BEARER + accessToken.getAccessToken());
        headers.put(AbfsUtilConstants.HEADER_XML_VERSION, AbfsUtilConstants.API_VERSION);

        params.put(AbfsUtilConstants.ABFS_PARAM_TYPE, AbfsUtilConstants.ABFS_PARAM_TYPE_VALUE);
        params.put(AbfsUtilConstants.ABFS_PARAM_COMP, AbfsUtilConstants.ABFS_PARAM_COMP_VALUE);

        PrivilegedAction<ClientResponse> action         = getAzureClientResponse(azureRestClient, relativeUrl, headers, params, keyInfo);
        ClientResponse                   clientResponse = executeAction(ugi, action);
        String                           response       = parseResponse(clientResponse, String.class);

        if (response != null) {
            if (StringUtils.isNotEmpty(response)) {
                XmlMapper xmlMapper  = new XmlMapper();
                JsonNode responseObj = xmlMapper.readTree(response.getBytes());

                if (responseObj != null) {
                    String         signedOid        = responseObj.findValue("SignedOid").asText();
                    String         signedTid        = responseObj.findValue("SignedTid").asText();
                    String         signedService    = responseObj.findValue("SignedService").asText();
                    String         signedStart      = responseObj.findValue("SignedStart").asText();
                    String         signedExpiry     = responseObj.findValue("SignedExpiry").asText();
                    String         signedVersion    = responseObj.findValue("SignedVersion").asText();
                    String         signedValue      = responseObj.findValue("Value").asText();
                    OffsetDateTime signedStartDate  = OffsetDateTime.parse(signedStart, DateTimeFormatter.ISO_OFFSET_DATE_TIME);
                    OffsetDateTime signedExpiryDate = OffsetDateTime.parse(signedExpiry, DateTimeFormatter.ISO_OFFSET_DATE_TIME);

                    ret = new UserDelegationKey()
                            .setSignedObjectId(signedOid)
                            .setSignedTenantId(signedTid)
                            .setSignedService(signedService)
                            .setSignedStart(signedStartDate)
                            .setSignedExpiry(signedExpiryDate)
                            .setSignedVersion(signedVersion)
                            .setValue(signedValue);
                } else {
                    LOG.error("Cannot parse response string received from Azure:[" + response + "]");
                }
            } else {
                LOG.error("Empty response string received from Azure");
            }
        } else {
            LOG.error("Null response object received from Azure");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== AbfsTokenProvider.fetchUserDelegationKeyWithAccessToken()");
        }
        return ret;
    }

    private void cacheDelegationSASGenerator(DelegationSASGenerator dsasGenerator) {
        String storageAccount = dsasGenerator.getStorageAccountName();

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> AbfsTokenProvider.setUserDelegationKeyToCache()" + storageAccount);
        }

        synchronized (this) {
            DelegationSASGenerator existing = dsasGeneratorCache.get(storageAccount);

            if (existing == null) {
                dsasGeneratorCache.put(storageAccount, dsasGenerator);
            } else {
                dsasGeneratorCache.replace(storageAccount, dsasGenerator);
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== AbfsTokenProvider.setUserDelegationKeyToCache()" + storageAccount);
        }
    }

    private class IdBrokerInitTask implements Runnable {
        public void run() {
            LOG.info("==> IdBrokerInitTask.run()");

            try {
                idBrokerClient = buildIdBrokerClient();

                initIdBrokerDelegationToken();

                accessToken = getAccessToken();

                initUserDelegationKeyCache();
            } catch (Exception excp) {
                LOG.error("IdBroker initialization failed. Aborting", excp);
            }

            LOG.info("<== IdBrokerInitTask.run()");
        }

        private RESTClient buildIdBrokerClient() {
            String idBrokerUrl = conf.get(ID_BROKER_HOST_URL);

            if (StringUtils.isEmpty(idBrokerUrl)) {
                idBrokerUrl = "";
            } else {
                idBrokerUrl = idBrokerUrl.trim();
            }

            if (idBrokerUrl.endsWith(URL_SEPARATOR)) {
                idBrokerUrl = idBrokerUrl.substring(0, idBrokerUrl.length() - 1);
            }

            RESTClient ret = new RESTClient(idBrokerUrl, sslConfigFileName, conf);

            ret.setRestClientConnTimeOutMs(restClientConnTimeOutMs);
            ret.setRestClientReadTimeOutMs(restClientReadTimeOutMs);

            return ret;
        }

        private void initIdBrokerDelegationToken()  throws InterruptedException {
            LOG.info("==> IdBrokerInitTask.initIdBrokerDelegationToken()");

            for (int attempt = 1; ; attempt++) {
                try {
                    idBrokerDelegationToken = fetchIdbrokerDelegationToken();

                    if (idBrokerDelegationToken != null) {
                        break;
                    }
                } catch (Throwable excp) {
                    LOG.warn("initIdBrokerDelegationToken() failed", excp);
                }

                LOG.info("Failed to get IdBroker delegation token. Will retry after {} seconds. attempt# {}", idBrokerInitRetryIntervalSeconds, attempt);

                Thread.sleep(idBrokerInitRetryIntervalSeconds * 1000);
            }

            LOG.info("<== IdBrokerInitTask.initIdBrokerDelegationToken()");
        }

        private void initUserDelegationKeyCache() throws IOException {
            LOG.info("==> IdBrokerInitTask.initUserDelegationKeyCache()");

            String[] storageAccounts = conf.getStrings(AZURE_STORAGE_ACCOUNT);

            if (storageAccounts != null) {
                for (String storageAccount : storageAccounts) {
                    DelegationSASGenerator dsasGenerator = getDelegationSASGenerator(storageAccount);

                    if (dsasGenerator != null) {
                        cacheDelegationSASGenerator(dsasGenerator);
                    }
                }
            }

            LOG.info("<== IdBrokerInitTask.initUserDelegationKeyCache()");
        }
    }

    class IdBrokerDelegationToken {
        private final String delegationToken;
        private final Date   expiryDate;

        IdBrokerDelegationToken(String delegationToken, Date expiryDate) {
            this.delegationToken = delegationToken;
            this.expiryDate      = expiryDate;
        }

        String getDelegationToken() {
            return this.delegationToken;
        }

        boolean isRenewalRequired() {
            Date    now          = MiscUtil.getUTCDateForLocalDate(new Date());
            Long    nowUtc       = now.getTime();
            Long    expiryDate   = MiscUtil.getUTCDateForLocalDate(this.expiryDate).getTime();
            int     precheckTime = conf.getInt(KNOX_DELEGATION_TOKEN_EXPIRY_TIME_PRECHECK_IN_SECONDS, AbfsUtilConstants.DEFAULT_KNOXDT_EXPIRY_TIME_PRECHECK_IN_SECONDS)*1000;
            boolean ret          = ((nowUtc - precheckTime) >= expiryDate);

            return ret;
        }
    }

    class AccessToken {
        private final String accessToken ;
        private final String expires_on_inSeconds;
        private final String expires_in_inSeconds;
        private       Long   accessTokenExpiryTime_inSeconds;

        AccessToken(String accessToken, String expires_on_inSeconds, String expires_in_inSeconds) {
            this.accessToken = accessToken;
            this.expires_on_inSeconds  = expires_on_inSeconds;
            this.expires_in_inSeconds  = expires_in_inSeconds;
            /*
             * If expires_on is set, use it over expires_in value.  If neither
             * are set, assume token is expired.
             * expires_on = number of seconds from 1 January 1970 (Midnight UTC/GMT).
             * expires_in = number of seconds from current time.
             */
            if (StringUtils.isNotBlank(this.expires_on_inSeconds)) {
                accessTokenExpiryTime_inSeconds = Long.parseLong(this.expires_on_inSeconds);
            } else if (StringUtils.isNotBlank(this.expires_in_inSeconds)) {
                Instant instant = Instant.now().plus(Long.parseLong(this.expires_in_inSeconds), ChronoUnit.SECONDS);
                accessTokenExpiryTime_inSeconds = instant.getEpochSecond();
            }
        }

        String getAccessToken() {
            return this.accessToken;
        }

        boolean isRenewalRequired() {
            if (accessTokenExpiryTime_inSeconds == null) {
                return true;
            }

            int precheckTime = conf.getInt(AZURE_ACCESSTOKEN_EXPIRY_TIME_PRECHECK_IN_SECONDS, AbfsUtilConstants.DEFAULT_AZURE_ACCESS_TOKEN_EXPIRY_TIME_PRECHECK_IN_SECONDS);

            return Instant.ofEpochSecond(accessTokenExpiryTime_inSeconds - precheckTime).isBefore(Instant.now());
        }

        @Override
        public String toString() {
            StringBuilder sb =  new StringBuilder();
            if ( this.accessToken != null) {
                sb = sb.append(STR_ACCESS_TOKEN_EQALS);
                sb.append(RangerRazClient.MASKED_TOKEN_STR);
            }
            return  sb.toString();
        }
    }

    boolean isUserDelegationKey_renewalRequired(UserDelegationKey userDelegationKey) {
        // UserDelegationKey renewal needs to be done before sasExpires. expiryTimeInHours <= sasExpiryTimeInHours * 2 ) ensures that.
        OffsetDateTime expiryDate                            = userDelegationKey.getSignedExpiry();
        long           userDelegationKeyexpiryTime_inSeconds = getUserDelegationKeyexpiryTime_inSeconds(expiryDate);
        long           sasExpiryTime_inSeconds               = getSasExpiryTime_inSeconds();
        boolean        ret                                   = (userDelegationKeyexpiryTime_inSeconds <= sasExpiryTime_inSeconds  * 2);

        if (LOG.isDebugEnabled()) {
            LOG.debug("AdlsGen2RazProcessor.isUserDelegationKey_renewalRequired(): {} , sasExpiryTime_inSeconds: {} , userDelegationKeyexpiryTime_inSeconds : {}", ret, sasExpiryTime_inSeconds, userDelegationKeyexpiryTime_inSeconds );
        }

        return ret;
    }

    private long getUserDelegationKeyexpiryTime_inSeconds(OffsetDateTime expiryDate) {
        long userDelegationKeyexpiry_inEpocSeconds  = expiryDate.toEpochSecond();
        long currentTime_inEpocSeconds              = OffsetDateTime.now().toEpochSecond();
        return (userDelegationKeyexpiry_inEpocSeconds - currentTime_inEpocSeconds) ;
    }

    private long getSasExpiryTime_inSeconds() {
        long sasExpiryTimeInSeconds = conf.getInt(SAS_EXPIRY_TIME_IN_SECONDS, SASGenerator.ONE_HOUR_IN_SECONDS);
        long ret                    = (sasExpiryTimeInSeconds <= 0) ? SASGenerator.ONE_HOUR_IN_SECONDS : sasExpiryTimeInSeconds;
        return ret;
    }

    private PrivilegedAction<ClientResponse> getIdBrokerResponse(final String url, final Map<String, String> headers, final Map<String, String> queryParams, final Object postParam) {
        return () -> {
            if (idBrokerClient != null) {
                try {
                    return idBrokerClient.get(url, headers, queryParams);
                } catch (Exception excp) {
                    LOG.error("AbfsTokenProvider.getIdBrokerResponse() failed", excp);
                }
            }

            return null;
        };
    }

    private static PrivilegedAction<ClientResponse> getAzureClientResponse(RESTClient azureRestClient, final String url, final Map<String, String> headers, final Map<String, String> queryParams, final Object postParam) {
        return () -> {
            if (azureRestClient != null) {
                try {
                    return azureRestClient.post(url, headers, queryParams, postParam);
                } catch (Exception excp) {
                    LOG.error("AbfsTokenProvider.getAzureClientResponse() failed", excp);
                }
            }

            return null;
        };
    }

    private static ClientResponse executeAction(UserGroupInformation ugi, final PrivilegedAction<ClientResponse> action) {
        ClientResponse ret = ugi.doAs(action);

        return ret;
    }

    private static <T> T parseResponse(final ClientResponse clientResponse, final Class<T> clazz) {
        T t = null;

        if (clientResponse != null && clientResponse.getStatus() == 200) {
            t = clientResponse.getEntity(clazz);
        } else {
            LOG.error("failed to parse ClientResponse {}", clientResponse);
        }

        return t;
    }

    private static String getExpiryKeyInfo(int expiryDurationMinutes) {
        Date     now = MiscUtil.getUTCDateForLocalDate(new Date());
        Calendar cal = Calendar.getInstance();

        cal.setTime(now);
        cal.add(Calendar.MINUTE, expiryDurationMinutes);

        Date             expiryTime = cal.getTime();
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
        String           expiry     = dateFormat.format(expiryTime);
        String           ret        = String.format("<KeyInfo><Expiry>%s</Expiry></KeyInfo>", expiry);

        if (LOG.isDebugEnabled()) {
            LOG.debug("getExpiryKeyInfo({}): ret={}", expiryDurationMinutes, ret);
        }

        return ret;
    }

    private RESTClient buildAzureClient(String storageAccount) {
        String azureUrl = String.format(AbfsUtilConstants.ABFS_USER_DELEGATION_KEY_PROVIDER, storageAccount);

        if (StringUtils.isNotEmpty(azureUrl)) {
            azureUrl = azureUrl.trim();
        }

        RESTClient ret = new RESTClient(azureUrl, sslConfigFileName, conf);

        ret.setRestClientConnTimeOutMs(restClientConnTimeOutMs);
        ret.setRestClientReadTimeOutMs(restClientReadTimeOutMs);

        return ret;
    }
}
