package org.apache.ranger.raz.processor.abfsutil;

/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

import com.azure.storage.blob.models.UserDelegationKey;
import org.apache.commons.lang3.StringUtils;
import org.apache.hadoop.fs.azurebfs.services.AbfsUriQueryBuilder;
import org.apache.hadoop.fs.azurebfs.utils.Base64;

import java.time.Instant;

public class DelegationSASGenerator extends SASGenerator {
    private final String            storageAccountName;
    private final UserDelegationKey userDelegationKey;
    private final String            skoid;
    private final String            sktid;
    private final String            skt;
    private final String            ske;
    private final String            sks = "b";
    private final String            skv;

    public DelegationSASGenerator(String storageAccountName, UserDelegationKey userDelegationKey) {
        super(Base64.decode(userDelegationKey.getValue()));

        this.storageAccountName = storageAccountName;
        this.userDelegationKey  = userDelegationKey;
        this.skoid              = userDelegationKey.getSignedObjectId();
        this.sktid              = userDelegationKey.getSignedTenantId();
        this.skt                = SASGenerator.ISO_8601_FORMATTER.format(userDelegationKey.getSignedStart());
        this.ske                = SASGenerator.ISO_8601_FORMATTER.format(userDelegationKey.getSignedExpiry());
        this.skv                = userDelegationKey.getSignedVersion();
    }

    public String getStorageAccountName() {
        return storageAccountName;
    }

    public UserDelegationKey getUserDelegationKey() {
        return userDelegationKey;
    }

    public String getDelegationSAS(String containerName, String path, String operation,
                                   String saoid, String suoid, String scid, int sasExpiryTime) {
        final String sv  = AuthenticationVersion.Feb20.toString();
        final String st  = ISO_8601_FORMATTER.format(Instant.now().minusSeconds(FIVE_MINUTES));
        final String se  = ISO_8601_FORMATTER.format(Instant.now().plusSeconds(sasExpiryTime));
        String       sr  = "b";
        String       sdd = null;
        String       sp  = null;

        switch (operation) {
            case SASTokenConstants.CREATE_FILE_OPERATION:
            case SASTokenConstants.CREATE_DIRECTORY_OPERATION:
            case SASTokenConstants.WRITE_OPERATION:
            case SASTokenConstants.SET_PROPERTIES_OPERATION:
                sp = "w";
                break;
            case SASTokenConstants.DELETE_OPERATION:
                sp = "d";
                break;
            case SASTokenConstants.DELETE_RECURSIVE_OPERATION:
                sp = "d";
                sr = "d";
                sdd = Integer.toString(StringUtils.countMatches(path, "/"));
                break;
            case SASTokenConstants.CHECK_ACCESS_OPERATION:
            case SASTokenConstants.GET_ACL_OPERATION:
            case SASTokenConstants.GET_STATUS_OPERATION:
                sp = "e";
                break;
            case SASTokenConstants.LIST_OPERATION:
                sp = "l";
                break;
            case SASTokenConstants.GET_PROPERTIES_OPERATION:
            case SASTokenConstants.READ_OPERATION:
                sp = "r";
                break;
            case SASTokenConstants.RENAME_DESTINATION_OPERATION:
            case SASTokenConstants.RENAME_SOURCE_OPERATION:
                sp = "m";
                break;
            case SASTokenConstants.SET_ACL_OPERATION:
            case SASTokenConstants.SET_PERMISSION_OPERATION:
                sp = "p";
                break;
            case SASTokenConstants.SET_OWNER_OPERATION:
                sp = "o";
                break;
            default:
                throw new IllegalArgumentException(operation);
        }

        String              signature = computeSignatureForSAS(sp, st, se, sv, sr, containerName, path, saoid, suoid, scid);
        AbfsUriQueryBuilder qb        = new AbfsUriQueryBuilder();

        qb.addQuery("skoid", skoid);
        qb.addQuery("sktid", sktid);
        qb.addQuery("skt", skt);
        qb.addQuery("ske", ske);
        qb.addQuery("sks", sks);
        qb.addQuery("skv", skv);
        if (saoid != null) {
            qb.addQuery("saoid", saoid);
        }
        if (suoid != null) {
            qb.addQuery("suoid", suoid);
        }
        if (scid != null) {
            qb.addQuery("scid", scid);
        }
        qb.addQuery("sp", sp);
        qb.addQuery("st", st);
        qb.addQuery("se", se);
        qb.addQuery("sv", sv);
        qb.addQuery("sr", sr);
        if (sdd != null) {
            qb.addQuery("sdd", sdd);
        }
        qb.addQuery("sig", signature);
        return qb.toString().substring(1);
    }

    private String computeSignatureForSAS(String sp, String st, String se, String sv,
                                          String sr, String containerName,
                                          String path, String saoid, String suoid, String scid) {

        StringBuilder sb = new StringBuilder();
        sb.append(sp);
        sb.append("\n");
        sb.append(st);
        sb.append("\n");
        sb.append(se);
        sb.append("\n");
        // canonicalized resource
        sb.append("/blob/");
        sb.append(storageAccountName);
        sb.append("/");
        sb.append(containerName);
        if (path != null && sr != "c") {
            sb.append(path);
        }
        sb.append("\n");
        sb.append(skoid);
        sb.append("\n");
        sb.append(sktid);
        sb.append("\n");
        sb.append(skt);
        sb.append("\n");
        sb.append(ske);
        sb.append("\n");
        sb.append(sks);
        sb.append("\n");
        sb.append(skv);
        sb.append("\n");
        if (saoid != null) {
            sb.append(saoid);
        }
        sb.append("\n");
        if (suoid != null) {
            sb.append(suoid);
        }
        sb.append("\n");
        if (scid != null) {
            sb.append(scid);
        }
        sb.append("\n");

        sb.append("\n"); // sip
        sb.append("\n"); // spr
        sb.append(sv);
        sb.append("\n");
        sb.append(sr);
        sb.append("\n");
        sb.append("\n"); // - For optional : rscc - ResponseCacheControl
        sb.append("\n"); // - For optional : rscd - ResponseContentDisposition
        sb.append("\n"); // - For optional : rsce - ResponseContentEncoding
        sb.append("\n"); // - For optional : rscl - ResponseContentLanguage
        sb.append("\n"); // - For optional : rsct - ResponseContentType

        String stringToSign = sb.toString();
        if(LOG.isDebugEnabled()) {
            LOG.debug("Delegation SAS stringToSign: " + stringToSign.replace("\n", "."));
        }
        return computeHmac256(stringToSign);
    }
}