/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.raz.processor.abfsutil;

import com.sun.jersey.api.client.*;
import com.sun.jersey.api.client.config.ClientConfig;
import com.sun.jersey.api.client.config.DefaultClientConfig;
import com.sun.jersey.client.urlconnection.HTTPSProperties;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.authorization.utils.StringUtil;
import org.apache.ranger.plugin.token.TokenRetriever;
import org.codehaus.jackson.jaxrs.JacksonJsonProvider;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyStore;
import java.security.SecureRandom;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManagerFactory;
import javax.servlet.http.HttpServletResponse;
import javax.ws.rs.core.Cookie;
import javax.ws.rs.core.MediaType;

public class RESTClient {
    private static final Logger LOG = LoggerFactory.getLogger(RESTClient.class);

    public static final String RANGER_RAZ_TRUSTSTORE_FILE_CREDENTIAL         = "xasecure.policymgr.clientssl.truststore";
    public static final String RANGER_RAZ_SSL_TRUSTMANAGER_ALGO_TYPE         = TrustManagerFactory.getDefaultAlgorithm();
    public static final String RANGER_RAZ_SSL_CONTEXT_ALGO_TYPE              = "TLS";
    public static final String RANGER_PROP_JWT_TOKEN_RETRIEVER_CLASS         = "ranger.auth.jwt.retriever.class";
    public static final String RANGER_PROP_JWT_TOKEN_RETRIEVER_CLASS_DEFAULT = "org.apache.ranger.plugin.token.JwTokenRetrieverEnv";
    public static final String AUTHORIZATION_HEADER                          = "Authorization";
    public static final String JWT_COOKIE_NAME_DEFAULT                       = "hadoop-jwt";
    public static final String RANGER_PROP_JWT_SERVER_COOKIE_NAME            = "ranger.auth.jwt.server.cookie.name";

    private String  mUrl;
    private String  mSslConfigFileName;
    private boolean mIsSSL;


    private String mKeyStoreFile;
    private int    mRestClientConnTimeOutMs;
    private int    mRestClientReadTimeOutMs;
    private int    lastKnownActiveUrlIndex;

    private final List<String> configuredURLs;

    private volatile Client        client;
    private TokenRetriever<String> tokenRetriever = null;
    private String                 jwtServerCookieName;

    public RESTClient(String url, String sslConfigFileName, Configuration config) {
        mUrl               = url;
        mSslConfigFileName = sslConfigFileName;
        configuredURLs     = StringUtil.getURLs(mUrl);

        setLastKnownActiveUrlIndex((new Random()).nextInt(getConfiguredURLs().size()));
        init(config);
    }

    public void setRestClientConnTimeOutMs(int mRestClientConnTimeOutMs) {
        this.mRestClientConnTimeOutMs = mRestClientConnTimeOutMs;
    }

    public void setRestClientReadTimeOutMs(int mRestClientReadTimeOutMs) {
        this.mRestClientReadTimeOutMs = mRestClientReadTimeOutMs;
    }

    public Client getClient() {
        Client result = client;

        if(result == null) {
            synchronized(this) {
                result = client;
                if(result == null) {
                    client = result = buildClient();
                }
            }
        }

        return result;
    }

    private Client buildClient() {
        Client client = null;

        if (mIsSSL) {
            SSLContext   sslContext  = getSSLContextFromKeyStoreFile(mKeyStoreFile);
            ClientConfig config      = new DefaultClientConfig();

            config.getClasses().add(JacksonJsonProvider.class);

            HostnameVerifier hv = new HostnameVerifier() {
                public boolean verify(String urlHostName, SSLSession session) {
                    return session.getPeerHost().equals(urlHostName);
                }
            };

            config.getProperties().put(HTTPSProperties.PROPERTY_HTTPS_PROPERTIES, new HTTPSProperties(hv, sslContext));
            client = Client.create(config);
        }

        if(client == null) {
            ClientConfig config = new DefaultClientConfig();
            config.getClasses().add(JacksonJsonProvider.class);
            client = Client.create(config);
        }

        client.setConnectTimeout(mRestClientConnTimeOutMs);
        client.setReadTimeout(mRestClientReadTimeOutMs);

        return client;
    }

    private void init(Configuration config) {

        mIsSSL = isSsl(mUrl);

        if (mIsSSL) {

            InputStream in = null;

            try {
                in = getFileInputStream(mSslConfigFileName);
                if (LOG.isDebugEnabled()) {
                    LOG.info("==> RESTClient.init() mSslConfigFileName" + mSslConfigFileName);
                }
                if (in != null) {
                    config.addResource(in);
                }

                mKeyStoreFile  = config.get(RANGER_RAZ_TRUSTSTORE_FILE_CREDENTIAL);
            } catch (IOException ioe) {
                LOG.error("Unable to load SSL Config FileName: [" + mSslConfigFileName + "]", ioe);
            } finally {
                close(in, mSslConfigFileName);
            }
        }

        tokenRetriever      = getJwtTokenRetriever(config);
        jwtServerCookieName = config.get(RANGER_PROP_JWT_SERVER_COOKIE_NAME, JWT_COOKIE_NAME_DEFAULT);
    }

    @SuppressWarnings("unchecked")
    private TokenRetriever<String> getJwtTokenRetriever(Configuration config) {
        String                        clsName   = config.get(RANGER_PROP_JWT_TOKEN_RETRIEVER_CLASS, RANGER_PROP_JWT_TOKEN_RETRIEVER_CLASS_DEFAULT);
        ClassLoader                   clsLoader = Thread.currentThread().getContextClassLoader();
        TokenRetriever<String>        ret       = null;
        Class<TokenRetriever<String>> cls;

        try {
            cls = (Class<TokenRetriever<String>>) clsLoader.loadClass(clsName.trim());
            ret = cls.getConstructor(Configuration.class).newInstance(config);
        } catch (Exception e) {
            LOG.error("RangerRESTClient.getJwtTokenRetriever(): Failed to initialize JWT token retriever.", e);
        }

        return ret;
    }

    private boolean isSsl(String url) {
        return !StringUtils.isEmpty(url) && url.toLowerCase().startsWith("https");
    }

    private InputStream getFileInputStream(String fileName)  throws IOException {
        InputStream in = null;

        if(StringUtils.isNotEmpty(fileName)) {
            File f = new File(fileName);

            if (f.exists()) {
                in = new FileInputStream(f);
            }
            else {
                in = ClassLoader.getSystemResourceAsStream(fileName);
            }
        }

        return in;
    }

    private void close(InputStream str, String filename) {
        if (str != null) {
            try {
                str.close();
            } catch (IOException excp) {
                LOG.error("Error while closing file: [" + filename + "]", excp);
            }
        }
    }

    public ClientResponse get(String relativeUrl, Map<String,String> headers, Map<String, String> params) throws Exception {
        ClientResponse finalResponse = null;
        int startIndex = this.lastKnownActiveUrlIndex;
        int currentIndex = 0;

        for (int index = 0; index < configuredURLs.size(); index++) {
            try {
                currentIndex = (startIndex + index) % configuredURLs.size();

                finalResponse = createWebResourceForGET(relativeUrl,headers,params,index);

                if (!shouldTryNextUrl(finalResponse, currentIndex)) {
                    break;
                }
            } catch (ClientHandlerException ex) {
                LOG.warn("Failed to communicate with Service with URL : " + configuredURLs.get(currentIndex));
                processException(index, ex);
            }
        }
        return finalResponse;
    }

    public ClientResponse post(String relativeUrl, Map<String,String> headers, Map<String, String> params, Object postParam) throws Exception {
        ClientResponse finalResponse = null;
        int startIndex = this.lastKnownActiveUrlIndex;
        int currentIndex = 0;

        for (int index = 0; index < configuredURLs.size(); index++) {
            try {
                currentIndex = (startIndex + index) % configuredURLs.size();

                finalResponse = createWebResourceForPOST(relativeUrl,headers,params,postParam,index);

                if (!shouldTryNextUrl(finalResponse, index)) {
                    break;
                }
            } catch (ClientHandlerException ex) {
                LOG.warn("Failed to communicate with Service with URL : " + configuredURLs.get(currentIndex));
                processException(index, ex);
            }
        }
        return finalResponse;
    }

    ClientResponse createWebResourceForGET(String relativeUrl, Map<String, String> headers, Map<String,String> params, int currentIndex) {

        ClientResponse ret = null;

        WebResource webResource = getClient().resource(configuredURLs.get(currentIndex) + relativeUrl);
        if (webResource != null && params != null) {
            webResource = setQueryParams(webResource, params);
        }

        String              token   = null;
        WebResource.Builder builder = null;
        if (headers != null) {
            token   = headers.get(AbfsUtilConstants.HEADER_AUTHORIZATION);
            builder = handleJwt(webResource.getRequestBuilder())
                    .accept(MediaType.WILDCARD)
                    .header(AbfsUtilConstants.HEADER_AUTHORIZATION, token);
            if (LOG.isDebugEnabled()) {
                LOG.info("==> RESTClient.createWebResourceForGET() URL: " + mSslConfigFileName);
            }
            ret = builder.get(ClientResponse.class);
        } else {
            webResource.accept(MediaType.WILDCARD).type(MediaType.WILDCARD);
            ret = webResource.get(ClientResponse.class);

        }

        return ret;
    }

    ClientResponse createWebResourceForPOST(String relativeUrl, Map<String, String> headers, Map<String,String> params,Object postParam, int currentIndex) {
        ClientResponse ret = null;

        WebResource webResource = getClient().resource(configuredURLs.get(currentIndex) + relativeUrl);
        if (webResource != null && params != null) {
            webResource = setQueryParams(webResource, params);
        }

        String              token      = null;
        String              apiVersion = null;
        WebResource.Builder builder    = null;
        if (headers != null) {
            token      = headers.get(AbfsUtilConstants.HEADER_AUTHORIZATION);
            apiVersion = headers.get(AbfsUtilConstants.HEADER_XML_VERSION);

            builder = handleJwt(webResource.getRequestBuilder())
                    .accept(MediaType.WILDCARD)
                    .type(MediaType.WILDCARD)
                    .header(AbfsUtilConstants.HEADER_AUTHORIZATION, token);
            builder.header(AbfsUtilConstants.HEADER_XML_VERSION,apiVersion);

            if (LOG.isDebugEnabled()) {
                LOG.info("==> RESTClient.createWebResourceForPOST() URL: " + webResource.getURI());
            }

            ret = builder.post(ClientResponse.class, postParam);
        }

        return ret;
    }

    protected static WebResource setQueryParams(WebResource webResource, Map<String, String> params) {
        WebResource ret = webResource;
        if (ret != null && params != null) {
            Set<Map.Entry<String, String>> entrySet= params.entrySet();
            for (Map.Entry<String, String> entry : entrySet) {
                ret = ret.queryParam(entry.getKey(), entry.getValue());
            }
        }
        return ret;
    }

    private WebResource.Builder handleJwt(WebResource.Builder builder) {
        if (tokenRetriever != null) {
            Optional<String> jwtOptional = tokenRetriever.retrieve();

            if (jwtOptional.isPresent()) {
                builder.cookie(new Cookie(jwtServerCookieName, jwtOptional.get()));
                builder.header(AUTHORIZATION_HEADER, jwtOptional.get());
            }
        } else {
            LOG.warn("Since JWTokenRetriver init failed, skipping JWT auth.");
        }

        return builder;
    }

    protected void processException(int index, ClientHandlerException e) throws Exception {
        if (index == configuredURLs.size() - 1) {
            LOG.error("Failed to communicate with all URL's : [ " + configuredURLs + " ]", e);
            throw e;
        }
    }

    protected void setLastKnownActiveUrlIndex(int lastKnownActiveUrlIndex) {
        this.lastKnownActiveUrlIndex = lastKnownActiveUrlIndex;
    }

    public List<String> getConfiguredURLs() {
        return configuredURLs;
    }


    public static SSLContext getSSLContextFromKeyStoreFile(String fileName) {
        SSLContext ret  = null;

        try {
            KeyStore keyStore = getKeyStore(fileName);
            ret = SSLContext.getInstance(RANGER_RAZ_SSL_CONTEXT_ALGO_TYPE);
            TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(RANGER_RAZ_SSL_TRUSTMANAGER_ALGO_TYPE);
            trustManagerFactory.init(keyStore);
            ret.init(null, trustManagerFactory.getTrustManagers(), new SecureRandom());
        } catch (Exception e) {
            String msg = "Error Creating SSLContext:" + "" ;
            LOG.error(msg, e);
        }

        return ret;
    }

    private static KeyStore getKeyStore(String fileName) {
        KeyStore        keyStore = null;
        FileInputStream fis      = null;

        try {
            String keyStoreType = KeyStore.getDefaultType();
            keyStore = KeyStore.getInstance(keyStoreType);
            fis = new FileInputStream(fileName);
            keyStore.load(fis, null);
            fis.close();
        } catch (Exception e) {
            LOG.error("Error during getting keystore", e);
        }

        return keyStore;
    }

    private boolean shouldTryNextUrl(ClientResponse response, int urlIndex) {
        boolean ret = response == null || response.getStatus() == HttpServletResponse.SC_NOT_FOUND;
        if (!ret) {
            if (response.getStatus() == HttpServletResponse.SC_OK) {
                setLastKnownActiveUrlIndex(urlIndex);
            }
        }
        return ret;
    }
}
