/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.processor.adls;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.plugin.policyengine.RangerAccessResource;
import org.apache.ranger.plugin.policyengine.RangerAccessResourceImpl;
import org.apache.ranger.plugin.policyengine.RangerPolicyEngine;
import org.apache.ranger.plugin.util.RangerUserStore;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.model.RangerRazRequestBase;
import org.apache.ranger.raz.model.RangerRazResultBase;
import org.apache.ranger.raz.model.RangerRazResultBase.ResourceAccessResult;
import org.apache.ranger.raz.processor.RangerRazRequestContext;
import org.apache.ranger.raz.processor.RangerRazContext;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.apache.ranger.raz.processor.RangerDefaultRazProcessor;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.plugin.policyengine.RangerAccessRequest;
import org.apache.ranger.plugin.policyengine.RangerAccessResult;
import org.apache.ranger.raz.processor.abfsutil.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_ADLS_UNAUTHORIZED_ACCESS;
import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_INVALID_RESPONSE;

public class AdlsGen2RazProcessor extends RangerDefaultRazProcessor {
	private static final Logger LOG = LoggerFactory.getLogger(AdlsGen2RazProcessor.class);

	private static final String SAS_EXPIRY_TIME_IN_SECONDS = "ranger.raz.azure.sas.expiry.time.in.seconds";
	private static final String CONF_PREFIX                = "ranger.raz.service-type.adls.";
	private static final String DEFAULT_SERVICE_NAME       = "cm_adls";
	private static final String CLOUD_IDENTITY_NAME        = "cloud_id";

	private static final String ACCESS_TYPE_READ               = "read";
	private static final String ACCESS_TYPE_WRITE              = "write";
	private static final String ACCESS_TYPE_EXECUTE            = "execute";
	private static final String ACCESS_TYPE_DELETE             = "delete";
	private static final String ACCESS_TYPE_DELETE_RECURSIVE   = "delete-recursive";
	private static final String ACCESS_TYPE_LIST               = "list";
	private static final String ACCESS_TYPE_MOVE               = "move";
	private static final String ACCESS_TYPE_MODIFY_PERMISSIONS = "modify-permissions";
	private static final String ADLS_ACCESS_ENFORCER           = "azure-acl";

	private static final Map<String, Set<String>> OPER_ACCESSTYPES = initOperAccessTypesMap();

	private AbfsTokenProvider       abfsTokenProvider      = null;
	private String                  adlsServiceName        = null;
	int     						sasExpiryTime          = 0;

	@Override
	public void init(RangerRazContext context) throws RangerRazException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> AdlsGen2RazProcessor.init(context={})", context);

			for (Iterator<Map.Entry<String, String>> confIterator = context.getRazConfig().iterator(); confIterator.hasNext(); ) {
				Map.Entry<String,String> entry = confIterator.next();
				LOG.debug("AdlsGen2RazProcessor: " + "Key = " +  entry.getKey() + " value = " + entry.getValue());
			}
		}

		try {
			Configuration conf     = context.getRazConfig();
			adlsServiceName        = conf.get(CONF_PREFIX + "default.service.name", DEFAULT_SERVICE_NAME);
			sasExpiryTime          = conf.getInt(SAS_EXPIRY_TIME_IN_SECONDS, SASGenerator.ONE_HOUR_IN_SECONDS);
			abfsTokenProvider      = new AbfsTokenProvider(conf);
		} catch (Throwable excp) {
			LOG.error("AdlsGen2RazProcessor.init() failed", excp);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== AdlsGen2RazProcessor.init(context={})", context);
		}
	}

	@Override
	public List<RangerAccessRequest> preProcess(RangerRazRequest request, RangerRazRequestContext context) throws RangerRazException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> AdlsGen2RazProcessor.preProcess(request={}, context={})", request, context);
		}

		request.setServiceName(adlsServiceName);                  // ignore the value in request.serviceName
		request.setUser(context.getRemoteUser());                 // ignore the value in request.user
		request.setUserGroups(context.getRemoteUserGroups());     // ignore the value in request.userGroups
		request.setClientIpAddress(context.getRemoteIpAddress()); // ignore the value in request.clientIpAddress
		request.setAccessTime(new Date());                        // ignore the value in request.accessTime

		request.getOperation().setAccessTypes(getAccessTypes(request.getOperation().getAction()));

		List<RangerAccessRequest> ret = super.preProcess(request, context);

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== AdlsGen2RazProcessor.preProcess(request={}, context={}): ret={}", request, context, ret);
		}

		return ret;
	}

	@Override
	public RangerRazResult postProcess(RangerRazRequest request, List<RangerAccessResult> accessResults, RangerRazRequestContext context) throws RangerRazException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> AdlsGen2RazProcessor.postProcess(request={}, accessResults={}, context={})", request, accessResults, context);
		}

		RangerRazResult      ret = super.postProcess(request, accessResults, context);
		ResourceAccessResult rar = ret != null ? ret.getOperResult() : null;

		if (rar != null) {
			switch (rar.getResult()) {
				case ALLOWED:
					createRazResult(RangerRazResultBase.AccessResult.ALLOWED, request, rar, context);
					break;
				case NOT_DETERMINED:
					createRazResult(RangerRazResultBase.AccessResult.NOT_DETERMINED, request, rar, context);
					break;
				case DENIED:
					createRazResult(RangerRazResultBase.AccessResult.DENIED, request, rar, context);
					break;
			}
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== AdlsGen2RazProcessor.postProcess(request={}, accessResults={}, context={}): ret={}", request, accessResults, context, ret);
		}

		return ret;
	}

	@Override
	public String getFallbackAclEnforcerName() {
		return ADLS_ACCESS_ENFORCER;
	}

	@Override
	public RangerAccessResource getRangerAccessResource(RangerRazRequest.ResourceAccess razResourceAccess) {
		return new RangerAdlsResource(razResourceAccess.getResource(), razResourceAccess.getResourceOwner());
	}

	private void createRazResult(RangerRazResultBase.AccessResult accessResult, RangerRazRequest request, ResourceAccessResult rar, RangerRazRequestContext requestContext) throws  RangerRazException {

		String             adlsDSASToken  = generateDSASToken(request, rar, accessResult, requestContext);
		Map<String,String> additionalInfo = rar.getAdditionalInfo();
		if (additionalInfo == null) {
			additionalInfo = new HashMap<>();
		}
		additionalInfo.put("ADLS_DSAS",          adlsDSASToken);

		if (requestContext != null && requestContext.getRangerUserStore() != null
				&& requestContext.getRangerUserStore().getUserStoreVersion() != null) {
			additionalInfo.put("RAZ_ID_MAP_VERSION", requestContext.getRangerUserStore().getUserStoreVersion().toString());
		}
		rar.setAdditionalInfo(additionalInfo);
	}

	private String generateDSASToken(RangerRazRequest request, ResourceAccessResult rar, RangerRazResultBase.AccessResult accessResult, RangerRazRequestContext requestContext) throws RangerRazException {
		String ret = null;

		if (LOG.isDebugEnabled()) {
			LOG.debug("==> AdlsGen2RazProcessor.generateDSASToken(uri={})", getStorageURI(request));
		}

		try {
			String                 user                   = request.getUser();
			String                 action                 = request.getOperation().getAction();
			String                 storageURI             = getStorageURI(request);
			String                 requestStorageAccount  = getResource(request, AbfsUtilConstants.KEY_STORAGE_ACCOUNT);
			DelegationSASGenerator delegationSASGenerator = abfsTokenProvider.getDelegationSASGenerator(requestStorageAccount);

			if (delegationSASGenerator != null) {
				String saoid = null;
				String suoid = null;
				switch (accessResult) {
					case ALLOWED:
						if (!(isOperationModifyOwner(action) ||
								isOperationModifyPermission(action))) {
							saoid = getOidForUser(user,requestContext);
						}
						break;

					case DENIED:
						throw new RangerRazException(RAZ_ADLS_UNAUTHORIZED_ACCESS, action, storageURI);

					case NOT_DETERMINED:
						suoid = getOidForUser(user,requestContext);

						if (suoid == null) {
							throw new RangerRazException(RAZ_ADLS_UNAUTHORIZED_ACCESS, action, storageURI);
						}
						break;
				}

				String auditLogId = CollectionUtils.isNotEmpty(rar.getAuditLogs()) ? rar.getAuditLogs().get(0).getAuditId() : null;

				ret = delegationSASGenerator.getDelegationSAS(
						getResource(request, AbfsUtilConstants.KEY_CONTAINER),
						getResource(request, AbfsUtilConstants.KEY_RELATIVE_PATH),
						action,
						saoid,
						suoid,
						auditLogId,
						sasExpiryTime);

				// if the token starts with '?', strip it
				if (StringUtils.startsWith(ret, "?")) {
					ret = ret.substring(1);
				}
			} else {
				if (LOG.isDebugEnabled()) {
					LOG.debug("<== AdlsGen2RazProcessor.generateDSASToken() " +"request user: " + user + "result: " + accessResult);
				}

				throw new RangerRazException(RAZ_CLIENT_INVALID_RESPONSE, "Error fetching UserDelegationKey for StorageAccount: " + requestStorageAccount);
			}

			if (LOG.isDebugEnabled()) {
				LOG.debug("<== AdlsGen2RazProcessor.generateDSASToken() " +"request user: " + user + "result: " + accessResult);
			}
		} catch (RangerRazException re) {
			//Exceptions related to Access Control are handled here.
			throw re;
		} catch (Exception e) {
			//Exceptions related to UserDelegationKey Fetch from IDBroker and other are handled here.
			throw new RangerRazException("Unable to generate DSAS Token....", e);
		}

		return ret;
	}

	private String getStorageURI(RangerRazRequest request) {
		StringBuilder ret = new StringBuilder("abfs://");

		RangerRazRequestBase.ResourceAccess resourceAccess = request.getOperation();

		Map<String,String> resource = resourceAccess.getResource();

		String storageAccount  = resource.get(AbfsUtilConstants.KEY_STORAGE_ACCOUNT) + AbfsUtilConstants.ABFS_RESOURCE_DOMAIN;
		String container       = resource.get(AbfsUtilConstants.KEY_CONTAINER);
		String relativePath    = resource.get(AbfsUtilConstants.KEY_RELATIVE_PATH);

		ret.append(container);
		ret.append("@");
		ret.append(storageAccount);
		ret.append(relativePath);

		return ret.toString();

	}

	private String getResource(RangerRazRequest request, String abfsUtilConstants ) {
		String ret = null;
		RangerRazRequestBase.ResourceAccess resourceAccess = request.getOperation();
		Map<String,String>                  resource       = resourceAccess.getResource();
		ret												   = resource.get(abfsUtilConstants);
		return ret;
	}

	private static Map<String, Set<String>> initOperAccessTypesMap() {
		Map<String, Set<String>> ret = new HashMap<>();

		ret.put(SASTokenConstants.CHECK_ACCESS_OPERATION,       Collections.singleton(ACCESS_TYPE_EXECUTE));
		ret.put(SASTokenConstants.CREATE_FILE_OPERATION,        Collections.singleton(ACCESS_TYPE_WRITE));
		ret.put(SASTokenConstants.DELETE_OPERATION,             Collections.singleton(ACCESS_TYPE_DELETE));
		ret.put(SASTokenConstants.DELETE_RECURSIVE_OPERATION,   Collections.singleton(ACCESS_TYPE_DELETE_RECURSIVE));
		ret.put(SASTokenConstants.GET_ACL_OPERATION,            Collections.singleton(ACCESS_TYPE_EXECUTE));
		ret.put(SASTokenConstants.GET_STATUS_OPERATION,         Collections.singleton(RangerPolicyEngine.ANY_ACCESS));
		ret.put(SASTokenConstants.GET_PROPERTIES_OPERATION,     Collections.singleton(ACCESS_TYPE_READ));
		ret.put(SASTokenConstants.LIST_OPERATION,               Collections.singleton(ACCESS_TYPE_LIST));
		ret.put(SASTokenConstants.CREATE_DIRECTORY_OPERATION,   Collections.singleton(ACCESS_TYPE_WRITE));
		ret.put(SASTokenConstants.READ_OPERATION,               Collections.singleton(ACCESS_TYPE_READ));
		ret.put(SASTokenConstants.RENAME_SOURCE_OPERATION,      Collections.singleton(ACCESS_TYPE_MOVE));
		ret.put(SASTokenConstants.RENAME_DESTINATION_OPERATION, Collections.singleton(ACCESS_TYPE_MOVE));
		ret.put(SASTokenConstants.SET_ACL_OPERATION,            Collections.singleton(ACCESS_TYPE_MODIFY_PERMISSIONS));
		ret.put(SASTokenConstants.SET_OWNER_OPERATION,          Collections.singleton(RangerPolicyEngine.SUPER_USER_ACCESS));
		ret.put(SASTokenConstants.SET_PERMISSION_OPERATION,     Collections.singleton(ACCESS_TYPE_MODIFY_PERMISSIONS));
		ret.put(SASTokenConstants.SET_PROPERTIES_OPERATION,     Collections.singleton(ACCESS_TYPE_WRITE));
		ret.put(SASTokenConstants.WRITE_OPERATION,              Collections.singleton(ACCESS_TYPE_WRITE));

		return ret;
	}

	private Set<String> getAccessTypes(String operation) {
		Set<String> ret = OPER_ACCESSTYPES.get(operation);

		if (ret == null) {
			LOG.error("Unknown operation:[" + operation + "]");

			ret = Collections.singleton(operation);
		}

		return ret;
	}

	private String getOidForUser(String user, RangerRazRequestContext requestContext) {
		String userOid = null;
		RangerUserStore rangerUserStore = requestContext.getRangerUserStore();
		if (rangerUserStore != null) {
			Map<String, Map<String,String>>  userAttrMapping = rangerUserStore.getUserAttrMapping();
			if (MapUtils.isNotEmpty(userAttrMapping)) {
				Map<String,String> cloudAttr = userAttrMapping.get(user);
				if (MapUtils.isNotEmpty(cloudAttr)) {
					userOid = cloudAttr.get(CLOUD_IDENTITY_NAME);
				}
			}
		}
		return userOid;
	}

	private static class RangerAdlsResource extends RangerAccessResourceImpl {

		public RangerAdlsResource(Map<String, String> resource, String ownerUser) {
			setValue(AbfsUtilConstants.KEY_STORAGE_ACCOUNT, resource.get(AbfsUtilConstants.KEY_STORAGE_ACCOUNT));
			setValue(AbfsUtilConstants.KEY_CONTAINER, resource.get(AbfsUtilConstants.KEY_CONTAINER));
			String relativePath = resource.get(AbfsUtilConstants.KEY_RELATIVE_PATH);
			if (StringUtils.equals(relativePath, AbfsUtilConstants.ABFS_RELATIVE_PATH_FOR_FILEACLSTATUS_OPERATION)) {
				relativePath = AbfsUtilConstants.ABFS_RELATIVE_PATH_FOR_ANY_ACCESS;
			}
			setValue(AbfsUtilConstants.KEY_RELATIVE_PATH, relativePath);

			setOwnerUser(ownerUser);
		}

		@Override
		public String getAsString() {
			String ret = getStringifiedValue();

			if(ret == null) {
				// format: myContainer@myStorageAccount:/myDir/myFile
				Object storageAccount = getValue(AbfsUtilConstants.KEY_STORAGE_ACCOUNT);
				Object container      = getValue(AbfsUtilConstants.KEY_CONTAINER);
				Object relativePath   = getValue(AbfsUtilConstants.KEY_RELATIVE_PATH);

				StringBuilder sb = new StringBuilder();

				if (container != null) {
					sb.append(container.toString());
				}
				sb.append('@');

				if (storageAccount != null) {
					sb.append(storageAccount.toString());
				}
				sb.append(':');

				if (relativePath != null) {
					sb.append(relativePath.toString());
				}

				ret = sb.toString();

				setStringifiedValue(ret);
			}
			return ret;
		}

	}

	private boolean isOperationModifyOwner(String operation) {
		return SASTokenConstants.SET_OWNER_OPERATION.equals(operation);
	}

	private boolean isOperationModifyPermission(String operation) {
		return SASTokenConstants.SET_PERMISSION_OPERATION.equals(operation);
	}
}
