/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.processor.abfsutil;

import com.azure.storage.blob.models.UserDelegationKey;
import org.apache.hadoop.conf.Configuration;
import org.junit.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.OffsetDateTime;

public class TestUserDelegationKeyExpiryTime {
    private static final Logger LOG = LoggerFactory.getLogger(TestUserDelegationKeyExpiryTime.class);
    private static Configuration     conf;
    private static AbfsTokenProvider abfsTokenProvider;
    private static UserDelegationKey userDelegationKey;

    @After
    public void after() {
        conf = null;
        abfsTokenProvider = null;
        userDelegationKey = null;
    }

    @Test
    public void runTests_test1() {
        LOG.info("runTests_test1: Both UserDelegationKey and SAS expiries in 5 hours, in that case isRenewalRequired should be true.");
        setUpTestData(5,3600*5L);
        boolean isRenewalRequired = abfsTokenProvider.isUserDelegationKey_renewalRequired(userDelegationKey);
        LOG.info("isRenewalRequired: " + isRenewalRequired);
        Assert.assertEquals(isRenewalRequired,true);
    }

    @Test
    public void runTests_test2() {
        LOG.info("runTests_test2: UserDelegationKey expiry in 7 hours and SAS expiries in 1 hours, in that case isRenewalRequired should be false.");
        setUpTestData(7,3600L);
        boolean isRenewalRequired = abfsTokenProvider.isUserDelegationKey_renewalRequired(userDelegationKey);
        LOG.info("isRenewalRequired: " + isRenewalRequired);
        Assert.assertEquals(isRenewalRequired,false);
    }

    @Test
    public void runTests_test3() {
        LOG.info("runTests_test3: UserDelegationKey expiry in 2 hours and SAS expiries in 1 hours, in that case isRenewalRequired should be false.");
        setUpTestData(1,3600*2L);
        boolean isRenewalRequired = abfsTokenProvider.isUserDelegationKey_renewalRequired(userDelegationKey);
        LOG.info("isRenewalRequired: " + isRenewalRequired);
        Assert.assertEquals(isRenewalRequired,true);
    }

    @Test
    public void runTests_test4() {
        LOG.info("runTests_test4: Both UserDelegationKey and SAS expiries in 1 hours, in that case isRenewalRequired should be true.");
        setUpTestData(1,3600*5L);
        boolean isRenewalRequired = abfsTokenProvider.isUserDelegationKey_renewalRequired(userDelegationKey);
        LOG.info("isRenewalRequired: " + isRenewalRequired);
        Assert.assertEquals(isRenewalRequired,true);
    }

    @Test
    public void runTests_test5() {
        LOG.info("runTests_test5: UserDelegationKey expiry in 1 hours and SAS expiries with in 1 hours, in that case isRenewalRequired should be true.");
        setUpTestData(1,3200L);
        boolean isRenewalRequired = abfsTokenProvider.isUserDelegationKey_renewalRequired(userDelegationKey);
        LOG.info("isRenewalRequired: " + isRenewalRequired);
        Assert.assertEquals(isRenewalRequired,true);
    }

    @Test
    public void runTests_test6() {
        LOG.info("runTests_test5: UserDelegationKey expiry in 1 hours and SAS expiries configure as 0 hrs, default value of 1 hr will be taken and in that case isRenewalRequired should be true.");
        setUpTestData(1,0L);
        boolean isRenewalRequired = abfsTokenProvider.isUserDelegationKey_renewalRequired(userDelegationKey);
        LOG.info("isRenewalRequired: " + isRenewalRequired);
        Assert.assertEquals(isRenewalRequired,true);
    }

    @Test
    public void runTests_test7() {
        LOG.info("runTests_test7: UserDelegationKey expiry in 1 hours and SAS expiries with in 1/2 hours, in that case isRenewalRequired should be false.");
        setUpTestData(1,1500L);
        boolean isRenewalRequired = abfsTokenProvider.isUserDelegationKey_renewalRequired(userDelegationKey);
        LOG.info("isRenewalRequired: " + isRenewalRequired);
        Assert.assertEquals(isRenewalRequired,false);
    }

    private void setUpTestData(long userDelegationKeyExpiryTimeinHours, Long sasExpiryInSeconds) {
        OffsetDateTime signedStartDate  = OffsetDateTime.now();
        OffsetDateTime signedExpiryDate = signedStartDate.plusHours(userDelegationKeyExpiryTimeinHours);
        try {
            conf = new Configuration();
            conf.set("ranger.raz.azure.sas.expiry.time.in.seconds", sasExpiryInSeconds.toString() );
            conf.set("fs.azure.ext.cab.address","http://dummy:7080");
            abfsTokenProvider = new AbfsTokenProvider(conf);
            userDelegationKey = new UserDelegationKey()
                    .setSignedObjectId("869aa174-6e4c-4517-948f-5051003b8b12")
                    .setSignedTenantId("fce84054-b9ea-4efc-af17-1d08b8156f22")
                    .setSignedService("b")
                    .setSignedStart(signedStartDate)
                    .setSignedExpiry(signedExpiryDate)
                    .setSignedVersion("2018-11-09")
                    .setValue("flzQBid79vhAJkXH3fgpVUmcz90MfkbF/0pE4GxNZSE=");
        } catch (Exception e) {
        }
    }
}
