/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.abfs;

import java.io.IOException;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.azurebfs.oauth2.IdentityTransformer;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.token.Token;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRazClient;
import org.apache.ranger.raz.intg.client.RangerRazClientLogger;
import org.apache.ranger.raz.model.RangerRazIdentityMappingsUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class RangerRazIdentityTransformer extends IdentityTransformer {
	private static final Logger LOG                           = LoggerFactory.getLogger(RangerRazIdentityTransformer.class);
	private static final String ADLS_CONFIG_PREFIX            = "fs.azure.ext.raz.prefix";
	private static final String ADLS_CONFIG_PREFIX_DEFAULT    = "fs.azure.ext.raz.";
	private static final String CONF_DELEGATION_TOKEN_KIND    = "delegation-token.token-kind";
	private static final String DELEGATION_TOKEN_KIND_DEFAULT = "raz-dt";

	private final RangerRazClient rangerRazClient;
	private final String          delegationToken;

	public RangerRazIdentityTransformer(Configuration configuration) throws IOException {
		super(configuration);
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RangerRazIdentityTransformer({})", configuration);
		}

		UserGroupInformation ugi = UserGroupInformation.getCurrentUser();

		if (ugi == null) {
			String errorMsg = "no user is logged in";

			RangerRazClientLogger.error(LOG, "RangerRazIdentityTransformer(): {}", errorMsg);

			throw new IOException("RangerRazIdentityTransformer initialization failed: " + errorMsg);
		}

		String configPrefix = configuration.get(ADLS_CONFIG_PREFIX, ADLS_CONFIG_PREFIX_DEFAULT);

		configuration.set(RangerRazClient.RAZ_CLIENT_PREFIX_KEY, configPrefix);

		String tokenKindStr = configuration.get(configPrefix + CONF_DELEGATION_TOKEN_KIND, DELEGATION_TOKEN_KIND_DEFAULT);

		delegationToken = getDelegationTokenFromUGI(ugi, tokenKindStr);
		rangerRazClient = RangerRazClient.getInstance(configuration, ugi);

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== RangerRazIdentityTransformer()");
		}
	}

	@Override
	public String transformIdentityForGetRequest(String originalIdentity, boolean isUserName, String localIdentity) throws IOException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RangerRazIdentityTransformer.transformIdentityForGetRequest({}, {}, {})", originalIdentity, isUserName, localIdentity);
		}

		String clusterId = null;
		try {
			RangerRazIdentityMappingsUtil razIdMappingsUtil = rangerRazClient.getIdentityMappingUtil(delegationToken);

			if (isUserName) {
				clusterId = razIdMappingsUtil.getUserClusterId(originalIdentity);
			} else {
				clusterId = razIdMappingsUtil.getGroupClusterId(originalIdentity);
			}
		} catch (RangerRazException e) {
			LOG.error("RangerRazIdentityTransformer.transformIdentityForGetRequest(): Exception occured while transforming identity: ", e);
		}

		if (StringUtils.isBlank(clusterId)) {
			clusterId = super.transformIdentityForGetRequest(originalIdentity, isUserName, localIdentity);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== RangerRazIdentityTransformer.transformIdentityForGetRequest(): clusterId={}", clusterId);
		}

		return clusterId;
	}

	@Override
	public String transformUserOrGroupForSetRequest(String userOrGroup) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RangerRazIdentityTransformer.transformUserOrGroupForSetRequest({})", userOrGroup);
		}

		String cloudId = null;
		try {
			RangerRazIdentityMappingsUtil razIdMappingsUtil = rangerRazClient.getIdentityMappingUtil(delegationToken);

			cloudId = razIdMappingsUtil.getUserCloudId(userOrGroup);
			if (StringUtils.isBlank(cloudId)) {
				cloudId = razIdMappingsUtil.getGroupCloudId(userOrGroup);
			}
		} catch (RangerRazException e) {
			LOG.error("RangerRazIdentityTransformer.transformUserOrGroupForSetRequest(): Exception occured while transforming user & group: ", e);
		}

		if (StringUtils.isBlank(cloudId)) {
			cloudId = super.transformUserOrGroupForSetRequest(userOrGroup);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== RangerRazIdentityTransformer.transformUserOrGroupForSetRequest(): cloudId=", cloudId);
		}

		return cloudId;
	}

    private static String getDelegationTokenFromUGI(UserGroupInformation ugi, String tokenKindStr) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazIdentityTransformer.getDelegationTokenFromUGI({}, {})", ugi, tokenKindStr);
        }

        String razDToken       = null;
        String maskedRazDToken = null;

        try {
            for (Token<?> token : ugi.getTokens()) {
                String tokenKind = (token != null && token.getKind() != null) ? token.getKind().toString() : null;

                if (LOG.isDebugEnabled()) {
                    LOG.debug("RangerRazIdentityTransformer.getDelegationTokenFromUGI(): tokenKind={}", tokenKind);
                }

                if (tokenKindStr.equalsIgnoreCase(tokenKind)) {
                    razDToken = token.encodeToUrlString();

                    if (LOG.isDebugEnabled()) {
                        maskedRazDToken = RangerRazClient.maskParam(razDToken);

                        LOG.debug("RangerRazIdentityTransformer.getDelegationTokenFromUGI(): found DT=[{}]", maskedRazDToken);
                    }

                    break;
                }
            }
        } catch (IOException excp) {
            RangerRazClientLogger.error(LOG, "RangerRazIdentityTransformer.getDelegationTokenFromUGI(): failed", excp);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRazIdentityTransformer.getDelegationTokenFromUGI(): ret={}", maskedRazDToken);
        }

        return razDToken;
    }
}
