/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.abfs;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.azurebfs.AzureBlobFileSystem;
import org.apache.hadoop.fs.azurebfs.extensions.SASTokenProvider;
import org.apache.hadoop.security.AccessControlException;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.token.Token;
import org.apache.ranger.raz.intg.RangerRazErrorCode;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRazClient;
import org.apache.ranger.raz.intg.client.RangerRazClientLogger;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.apache.ranger.raz.model.RangerRazRequestBase.ResourceAccess;
import org.apache.ranger.raz.model.RangerRazResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class RangerRazTokenProvider implements SASTokenProvider{
	private static final Logger LOG = LoggerFactory.getLogger(RangerRazTokenProvider.class);

	public static final String ADLS_RESOURCE_STORAGE_ACCOUNT = "storageaccount";
	public static final String ADLS_RESOURCE_CONTAINER       = "container";
	public static final String ADLS_RESOURCE_RELATIVE_PATH   = "relativepath";
	public static final String CONF_DELEGATION_TOKEN_KIND    = "delegation-token.token-kind";
	public static final String CONF_ADLS_SERVICE_TYPE        = "adls.service.type";
	public static final String DELEGATION_TOKEN_KIND_DEFAULT = "raz-dt";
	public static final String ADLS_SERVICE_TYPE_DEFAULT     = "adls";
	public static final String ADLS_CONFIG_PREFIX            = "fs.azure.ext.raz.prefix";
	public static final String ADLS_CONFIG_PREFIX_DEFAULT    = "fs.azure.ext.raz.";
	public static final String CONF_CLUSTER_NAME             = ".access.cluster.name";
	public static final String CONF_CLUSTER_TYPE             = ".access.cluster.type";
	public static final String ADDL_INFO_KET_ADLS_DSAS       = "ADLS_DSAS";

	private String          serviceType  = ADLS_SERVICE_TYPE_DEFAULT;
	private String          clusterName;
	private String          clusterType;
	private String          userName;
	private String          delegationToken;
	private RangerRazClient razClient;


	@Override
	public void initialize(final Configuration configuration, final String accountName) throws IOException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RangerRazTokenProvider.initialize(accountName={})", accountName);
		}

		UserGroupInformation ugi = UserGroupInformation.getCurrentUser();

		if (ugi == null) {
			String errorMsg = "no user is logged in";

			RangerRazClientLogger.error(LOG, "RangerRazTokenProvider(): {}", errorMsg);

			throw new IOException("RangerRazTokenProvider initialization failed: " + errorMsg);
		}

		String configPrefix = configuration.get(ADLS_CONFIG_PREFIX, ADLS_CONFIG_PREFIX_DEFAULT);

		configuration.set(RangerRazClient.RAZ_CLIENT_PREFIX_KEY, configPrefix);

		String tokenKindStr = configuration.get(configPrefix + CONF_DELEGATION_TOKEN_KIND, DELEGATION_TOKEN_KIND_DEFAULT);

		serviceType     = configuration.get(configPrefix + CONF_ADLS_SERVICE_TYPE, ADLS_SERVICE_TYPE_DEFAULT);
		clusterName     = configuration.get(configPrefix + serviceType + CONF_CLUSTER_NAME, "");
		clusterType     = configuration.get(configPrefix + serviceType + CONF_CLUSTER_TYPE, "");
		userName        = getUserNameFromUGI(ugi);
		delegationToken = getDelegationTokenFromUGI(ugi, tokenKindStr);
		razClient       = RangerRazClient.getInstance(configuration, ugi);

		if (LOG.isDebugEnabled()) {
			LOG.debug("RangerRazTokenProvider.initialize(): configPrefix={}", configPrefix);
			LOG.debug("RangerRazTokenProvider.initialize(): {}={}", CONF_DELEGATION_TOKEN_KIND, tokenKindStr);
			LOG.debug("RangerRazTokenProvider.initialize(): {}={}", CONF_ADLS_SERVICE_TYPE, serviceType);
			LOG.debug("RangerRazTokenProvider.initialize(): {}={}", serviceType + CONF_CLUSTER_NAME, clusterName);
			LOG.debug("RangerRazTokenProvider.initialize(): {}={}", serviceType + CONF_CLUSTER_TYPE, clusterType);

			LOG.debug("<== RangerRazTokenProvider.initialize(accountName={})", accountName);
		}
	}

	@Override
	public String getSASToken(String account, String fileSystem, String path, String operation) throws IOException, AccessControlException {
		if(LOG.isDebugEnabled()) {
			LOG.debug("==> RangerRazTokenProvider.getSASToken({}, {}, {}, {})", account, fileSystem, path, operation);

			AzureBlobFileSystem.printStatistics();
		}

		String              ret         = null;
		Map<String, String> resource    = createRazResource(account, fileSystem, path);
		Set<String>         accessTypes = new HashSet<>();

		accessTypes.add(operation);

		ResourceAccess      resourceAccess = new ResourceAccess(resource, account, operation, accessTypes);//TODO verify all parameters
		RangerRazRequest    razRequest     = new RangerRazRequest(serviceType, resourceAccess, userName);

		razRequest.setClusterName(clusterName);
		razRequest.setClusterType(clusterType);

		try {
			RangerRazResult result = razClient.checkPrivilege(razRequest, delegationToken);

			if (result != null && result.getOperResult() != null) {
				if (LOG.isDebugEnabled()) {
					LOG.debug("RangerRazTokenProvider.getSASToken(): result=" + result);
				}

				Map<String,String> additionalInfo = result.getOperResult().getAdditionalInfo();

				if (additionalInfo != null) {
					ret = additionalInfo.get(ADDL_INFO_KET_ADLS_DSAS);
				}
			}
		} catch (RangerRazException excp) {
			RangerRazClientLogger.error(LOG, "Failed to get DSAS token from Raz", excp);

			if (excp.getErrorCode() == RangerRazErrorCode.RAZ_CLIENT_ACCESS_DENIED) {
				throw new AccessControlException();
			} else {
				throw new AccessControlException(excp);
			}
		}

		return ret;
	}

	private static Map<String, String> createRazResource(String account, String container, String path) {
		Map<String, String> ret = new HashMap<>();

		if (!path.startsWith(Path.SEPARATOR)) {
			path = Path.SEPARATOR + path;
		}

		ret.put(ADLS_RESOURCE_STORAGE_ACCOUNT, account);
		ret.put(ADLS_RESOURCE_CONTAINER, container);
		ret.put(ADLS_RESOURCE_RELATIVE_PATH, path);

		return ret;
	}

	private static String getUserNameFromUGI(UserGroupInformation ugi) throws AccessControlException {
		String ret = ugi != null ? ugi.getShortUserName() : null;

		if (ret == null) {
			String errorMsg = "Failed to obtain currently logged in user";

			RangerRazClientLogger.error(LOG, errorMsg);

			throw new AccessControlException(errorMsg);
		}

		return ret;
	}

    private static String getDelegationTokenFromUGI(UserGroupInformation ugi, String tokenKindStr) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazTokenProvider.getDelegationTokenFromUGI()");
        }

        String razDToken       = null;
        String maskedRazDToken = null;

        try {
            for (Token<?> token : ugi.getTokens()) {
                String tokenKind = (token != null && token.getKind() != null) ? token.getKind().toString() : null;

                if (LOG.isDebugEnabled()) {
                    LOG.debug("RangerRazTokenProvider.getDelegationTokenFromUGI(): tokenKind={}", tokenKind);
                }

                if (tokenKindStr.equalsIgnoreCase(tokenKind)) {
                    razDToken = token.encodeToUrlString();

                    if (LOG.isDebugEnabled()) {
                        maskedRazDToken = RangerRazClient.maskParam(razDToken);

                        LOG.debug("RangerRazTokenProvider.getDelegationTokenFromUGI(): found DT=[{}]", maskedRazDToken);
                    }

                    break;
                }
            }
        } catch (IOException excp) {
            RangerRazClientLogger.error(LOG, "RangerRazTokenProvider.getDelegationTokenFromUGI(): failed", excp);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRazTokenProvider.getDelegationTokenFromUGI(): ret={}", maskedRazDToken);
        }

        return razDToken;
    }
}
