/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.abfs;

import java.io.IOException;
import java.net.URI;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.azurebfs.extensions.BoundDTExtension;
import org.apache.hadoop.fs.azurebfs.extensions.CustomDelegationTokenManager;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.token.Token;
import org.apache.hadoop.security.token.delegation.web.DelegationTokenIdentifier;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRazClient;
import org.apache.ranger.raz.intg.client.RangerRazClientLogger;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class RazDelegationTokenManager implements CustomDelegationTokenManager, BoundDTExtension {
    private static final Logger LOG = LoggerFactory.getLogger(RazDelegationTokenManager.class);

    public static final String ADLS_CONFIG_PREFIX         = "fs.azure.ext.raz.prefix";
    public static final String ADLS_CONFIG_PREFIX_DEFAULT = "fs.azure.ext.raz.";

    public static final String CANONICAL_SERVICE_NAME_STRATEGY                       = "canonical-service-name-strategy";
    public static final String CANONICAL_SERVICE_NAME_STRATEGY_ACCOUNT_ONLY          = "ACCOUNT_ONLY";
    public static final String CANONICAL_SERVICE_NAME_STRATEGY_ACCOUNT_AND_CONTAINER = "ACCOUNT_AND_CONTAINER";
    public static final String CANONICAL_SERVICE_NAME_STRATEGY_DEFAULT               = CANONICAL_SERVICE_NAME_STRATEGY_ACCOUNT_ONLY;

    private RangerRazClient rangerRazClient;
    private String canonicalServiceName;
    private String canonicalServiceNameStrategy;

    @Override
    public void initialize(Configuration configuration) throws IOException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RazDelegationTokenManager.initialize()");
        }

        UserGroupInformation ugi = UserGroupInformation.getCurrentUser();

        if (ugi == null) {
            String errorMsg = "No user is logged in";

            RangerRazClientLogger.error(LOG, "RazDelegationTokenManager(): {}", errorMsg);

            throw new IOException("RazDelegationTokenManager initialization failed: " + errorMsg);
        }

        String configPrefix = configuration.get(ADLS_CONFIG_PREFIX, ADLS_CONFIG_PREFIX_DEFAULT);

        configuration.set(RangerRazClient.RAZ_CLIENT_PREFIX_KEY, configPrefix);
        canonicalServiceNameStrategy = configuration.get(configPrefix + CANONICAL_SERVICE_NAME_STRATEGY, CANONICAL_SERVICE_NAME_STRATEGY_DEFAULT);

        if (LOG.isDebugEnabled()) {
            LOG.debug("RazDelegationTokenManager.initialize(): configPrefix=[{}]", configPrefix);
        }

        rangerRazClient = RangerRazClient.getInstance(configuration, ugi);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RazDelegationTokenManager.initialize()");
        }
    }

	@Override
	public Token<DelegationTokenIdentifier> getDelegationToken(String renewer) throws IOException {
		if(LOG.isDebugEnabled()) {
			LOG.debug("==> RazDelegationTokenManager.getDelegationToken({})", renewer);
		}

		Token<DelegationTokenIdentifier> ret = new Token<>();
		String maskedRazDToken = null;

		try {
			String tokenStr = rangerRazClient.getDelegationToken(renewer);

			ret.decodeFromUrlString(tokenStr);

			if (StringUtils.isNotBlank(tokenStr)) {
				maskedRazDToken = RangerRazClient.MASKED_TOKEN_STR;
			}
		} catch (RangerRazException excp) {
			RangerRazClientLogger.error(LOG, "failed to get delegation token", excp);

			throw new IOException(excp);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== RazDelegationTokenManager.getDelegationToken({}): ret={}", renewer, maskedRazDToken);
		}

		return ret;
	}

	@Override
	public long renewDelegationToken(Token<?> token) throws IOException {
		String maskedRazDToken = null;
		long   ret             = 0;
		String tokenStr        = token.encodeToUrlString();

		if (StringUtils.isNotBlank(tokenStr)) {
			maskedRazDToken = RangerRazClient.MASKED_TOKEN_STR;
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RazDelegationTokenManager.renewDelegationToken({})", maskedRazDToken);
		}

		try {
			ret = rangerRazClient.renewDelegationToken(tokenStr);
		} catch (RangerRazException excp) {
			RangerRazClientLogger.error(LOG, "failed to renew delegation token", excp);

			throw new IOException(excp);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== RazDelegationTokenManager.renewDelegationToken(): ret={}", ret);
		}

		return ret;
	}

	@Override
	public void cancelDelegationToken(Token<?> token) throws IOException {
		String maskedRazDToken = null;
		String tokenStr = token.encodeToUrlString();

		if (StringUtils.isNotBlank(tokenStr)) {
			maskedRazDToken = RangerRazClient.MASKED_TOKEN_STR;
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RazDelegationTokenManager.cancelDelegationToken({})", maskedRazDToken);
		}

		try {
			rangerRazClient.cancelDelegationToken(tokenStr);
		} catch (RangerRazException excp) {
			RangerRazClientLogger.error(LOG, "failed to cancel delegation token", excp);

			throw new IOException(excp);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== RazDelegationTokenManager.cancelDelegationToken()");
		}
	}

    @Override
    public void close() throws IOException {
        //NO-OP method
        if (LOG.isDebugEnabled()) {
            LOG.debug("===>> RazDelegationTokenManager.close(): Nothing to clean related to Delegation token.");
        }
    }

    @Override
    public void bind(URI fsURI, Configuration conf) throws IOException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("===>> RazDelegationTokenManager.bind({}, {})", fsURI, conf);
        }

        if (CANONICAL_SERVICE_NAME_STRATEGY_ACCOUNT_AND_CONTAINER.equalsIgnoreCase(this.canonicalServiceNameStrategy)) {
            this.canonicalServiceName = fsURI.toString();
        } else {
            this.canonicalServiceName = "abfs://" + fsURI.getHost() + "/";
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<<=== RazDelegationTokenManager.bind(): setting this.canonicalServiceName={}", this.canonicalServiceName);
        }
    }

    @Override
    public String getCanonicalServiceName() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("===>> RazDelegationTokenManager.getCanonicalServiceName()");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<<=== RazDelegationTokenManager.getCanonicalServiceName(): ret={}", this.canonicalServiceName);
        }
        return canonicalServiceName;
    }
}
