/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.s3;

import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.AWSCredentialsProvider;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.raz.hook.s3.cache.SignerCacheImpl;
import org.apache.ranger.raz.hook.s3.cache.SignerCache;
import org.apache.ranger.raz.intg.client.RangerRazClient;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsRequestCacheKey;

import java.util.Map;

import static org.apache.ranger.raz.hook.s3.cache.SignerCacheImpl.CACHE_KEY_EXPIRY_IN_SECONDS_DEFAULT;
import static org.apache.ranger.raz.hook.s3.cache.SignerCacheImpl.CACHE_MAX_SIZE_DEFAULT;

/**
 * Special AWSCredentialsProvider which does not contain any aws credentials.
 * This is instantiated during creation of the filesystem
 * thus ensuring the lifecycle of this instance matches the lifecycle of filesystem.
 * This is actually used to pass following information from filesystem to
 * signer {@code org.apache.ranger.raz.hook.s3.RazS3SignerPlugin}.
 * <li>FS owner</li>
 * <li>FS configuration</li>
 * <li>Delegation token binding which contain raz token to communicate with RAZ</li>
 * <li>RazClient instance to communicate with RAZ</li>
 *
 * This also stores the signature cache thus ensuring that the cache is not shared
 * across FileSystem instances.
 */
public class RazAnonymousAWSCredentialsProvider implements AWSCredentialsProvider {
  private final RazDelegationTokenBinding
      razDelegationTokenBinding;
  private final UserGroupInformation owner;
  private final Configuration configuration;

  /**
   * Configuration to configure signature cache max size. Can
   * be configured in core-site.xml. If not configured, default
   * value will be used from {@code SignerCacheImpl#SIGNATURE_CACHE_MAX_SIZE}.
   * Cache will evict the value once this limit is reached
   * ie. keys will be removed once cache size is equal to max configured size.
   * Eviction policy will be based on the underlying cache implementation.
   */
  private static final String SIGNATURE_CACHE_MAX_SIZE = "fs.s3a.signature.cache.max.size";

  /**
   * Configuration to configure signature cache expiry in seconds.
   * Can be configured in core-site.xml. If not configured, default
   * value will be used from {@code SignerCacheImpl#SIGNATURE_CACHE_EXPIRY_SECONDS}.
   * Cache will evict the value once the keys are expired
   * ie. current_time - insertion_time > configured_expiry_time.
   */
  private static final String SIGNATURE_CACHE_EXPIRY_SECONDS = "fs.s3a.signature.cache.expiry.sec";

  /**
   * SignerCache to store signed headers for AWS requests and reuse them.
   * The reason it is stored here is because we want cache to be per fileSystem
   * instance.
   */
  private final SignerCache<AwsRequestCacheKey, Map<String, String>> signerCache;

  /**
   * RangerRazClient instance created in the RazDelegationTokenBinding,
   * so its lifecycle matches the S3AFileSystem's lifecycle and owner.
   */
  private final RangerRazClient razClient;

  public RazAnonymousAWSCredentialsProvider(RazDelegationTokenBinding razDT,
      UserGroupInformation owner, Configuration conf, RangerRazClient razClient) {
    this.razDelegationTokenBinding = razDT;
    this.owner = owner;
    this.configuration = conf;
    this.razClient = razClient;
    long maxSize = conf.getLong(SIGNATURE_CACHE_MAX_SIZE, CACHE_MAX_SIZE_DEFAULT);
    long expiryInSec = conf.getLong(SIGNATURE_CACHE_EXPIRY_SECONDS, CACHE_KEY_EXPIRY_IN_SECONDS_DEFAULT);
    this.signerCache = new SignerCacheImpl(maxSize, expiryInSec);
  }

  @Override
  public AWSCredentials getCredentials() {
    // there are no credentials here, only a reference to the fs
    return new RazAnonymousAWSCredentials();
  }

  @Override
  public void refresh() {
    // do nothing
  }

  public RazDelegationTokenBinding getRazDelegationTokenBinding() {
    return razDelegationTokenBinding;
  }

  public UserGroupInformation getOwner() {
    return owner;
  }

  /**
   * AWSCredentials that the signer with end up using.
   */
  public class RazAnonymousAWSCredentials implements AWSCredentials {
    public RangerRazClient getRazClient() {
      return razClient;
    }

    @Override
    public String getAWSAccessKeyId() {
      return "";
    }

    @Override
    public String getAWSSecretKey() {
      return "";
    }

    public RazDelegationTokenBinding getRazDelegationTokenBinding() {
      return razDelegationTokenBinding;
    }

    public UserGroupInformation getOwner() {
      return owner;
    }

    public Configuration getConfig() {
      return configuration;
    }

    public SignerCache<AwsRequestCacheKey, Map<String, String>> getSignerCache() {
      return signerCache;
    }
  }
}
