/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.s3;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.s3a.AWSCredentialProviderList;
import org.apache.hadoop.fs.s3a.Retries;
import org.apache.hadoop.fs.s3a.S3AFileSystem;
import org.apache.hadoop.fs.s3a.auth.RoleModel;
import org.apache.hadoop.fs.s3a.auth.delegation.AbstractDelegationTokenBinding;
import org.apache.hadoop.fs.s3a.auth.delegation.AbstractS3ATokenIdentifier;
import org.apache.hadoop.fs.s3a.auth.delegation.EncryptionSecrets;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.token.Token;
import org.apache.hadoop.security.token.delegation.web.DelegationTokenIdentifier;
import org.apache.hadoop.service.ServiceStateException;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRazClient;

import java.io.IOException;
import java.util.Optional;

import static org.apache.ranger.raz.hook.s3.RazS3ATokenIdentifier.RAZ_TOKEN_KIND;

/**
 * Implementation of {@code AbstractDelegationTokenBinding } for
 * RAZ delegation tokens.
 */
public class RazDelegationTokenBinding extends AbstractDelegationTokenBinding {

  /**
   * Wire name of this binding: {@value}.
   */
  private static final String NAME = "RazDelegationToken";
  public static final String RAZ_S3_CONFIG_PREFIX_NAME = "fs.s3a.ext.raz.prefix";
  public static final String RAZ_S3_CONFIG_PREFIX_DEFAULT = "fs.s3a.ext.raz.";

  private static final Logger LOG =
      LoggerFactory.getLogger(
          org.apache.ranger.raz.hook.s3.RazDelegationTokenBinding.class);

  /**
   * The Raz token for this binding
   */
  private RazToken razToken;

  /**
   * There's only one credential provider; this ensures that
   * its synchronized calls really do get locks.
   */
  private AWSCredentialProviderList credentialProviders;

  /**
   * Raz client is used to get the delegation token from
   * RAZ server.
   */
  RangerRazClient razClient;

  /**
   * Constructor.
   *  @param name as passed to superclass for use in log messages.
   * @param kind token kind.
   */
  protected RazDelegationTokenBinding(String name,
      Text kind) {
    super(name, kind);
  }

  /**
   * Reflection-based constructor.
   */
  public RazDelegationTokenBinding() {
    this(NAME, RAZ_TOKEN_KIND);
  }

  /**
   * Collect the RAZ token
   *
   * @param policy minimum policy to use, if known.
   * @param encryptionSecrets encryption secrets for the token.
   * @param renewer the principal permitted to renew the token.
   * @return the token identifier for the DT
   * @throws IOException failure to collect a DT.
   */
  @Override
  public AbstractS3ATokenIdentifier createTokenIdentifier(Optional<RoleModel.Policy> policy,
      EncryptionSecrets encryptionSecrets,
      Text renewer) throws IOException {
    LOG.debug("Creating token identifier.");
    checkRazClientInitialized();

    RazToken razTokenBonded = requestRazAccessToken(renewer.toString());

    final String dtServiceName = razClient.getDelegationTokenServiceName();

    RazS3ATokenIdentifier identifier = new RazS3ATokenIdentifier(
        RAZ_TOKEN_KIND,
        getOwnerText(),
        renewer,
        getCanonicalUri(),
        "Created from " + dtServiceName,
        encryptionSecrets,
        razTokenBonded
    );

    LOG.debug("Created token identifier. Owner: {}; Bucket: {}",
        identifier.getOwner(), identifier.getBucket());

    return identifier;
  }

  /**
   * {@inheritDoc}
   * Deploying unbonded means the RAZ DT is bounded to some
   * local user obtained using {@code getOwnerText()}.
   */
  @Override
  public AWSCredentialProviderList deployUnbonded() throws IOException {
    LOG.debug("Deploy Unbonded.");
    // In unbonded mode kerberos auth will be used.
    // Thus not requesting any raz token.
    this.razToken = null;
    return getAwsCredentialProviderList();
  }

  @Override
  public AbstractS3ATokenIdentifier createEmptyIdentifier() {
    return new RazS3ATokenIdentifier();
  }

  @Override
  public AWSCredentialProviderList bindToTokenIdentifier(
      AbstractS3ATokenIdentifier retrievedIdentifier) throws IOException {
    if (LOG.isDebugEnabled()) {
      LOG.debug("Binding to retrieved token: {}", retrievedIdentifier.toString());
    }

    RazS3ATokenIdentifier tokenIdentifier =
        convertTokenIdentifier(retrievedIdentifier,
            RazS3ATokenIdentifier.class);
    tokenIdentifier.validate();

    this.razToken = tokenIdentifier.getRazToken();


    return getAwsCredentialProviderList();
  }

  private AWSCredentialProviderList getAwsCredentialProviderList() {
    credentialProviders = new AWSCredentialProviderList();
    final S3AFileSystem fs = this.getFileSystem();
    final UserGroupInformation owner = fs.createStoreContext().getOwner();
    final RazAnonymousAWSCredentialsProvider razAnonymousAWSCredentialsProvider
        = new RazAnonymousAWSCredentialsProvider(this, owner, getConfig(), razClient);
    credentialProviders.add(razAnonymousAWSCredentialsProvider);

    return credentialProviders;
  }

  @Override
  protected void serviceStop() throws Exception {
    super.serviceStop();
    this.razClient = null;
  }

  @Override
  protected void serviceStart() throws Exception {
    super.serviceStart();
    this.razClient = createRazClient(getConfig(), getOwner());
  }

  /**
   * If a new access token needed, collect one and bond to it.
   *
   * @throws IOException if a token could not be requested.
   */
  @Retries.OnceRaw
  private RazToken requestRazAccessToken(String renewer) throws IOException {
    LOG.debug("Requesting initial Raz delegation token");
    checkRazClientInitialized();

    RazToken retrievedToken = null;
    Token<DelegationTokenIdentifier> ret = new Token<>();
    try {
      String razDT = razClient.getDelegationToken(renewer);
      ret.decodeFromUrlString(razDT);
      DelegationTokenIdentifier delegationTokenIdentifier = ret.decodeIdentifier();
      retrievedToken = new RazToken(delegationTokenIdentifier.getIssueDate(),
              delegationTokenIdentifier.getMaxDate(),
              razDT);
      LOG.debug("Got token from raz server via client.");
    } catch (RangerRazException e) {
      throw new IOException("Error while getting token from raz server.", e);
    }

    return retrievedToken;
  }

  private void checkRazClientInitialized() {
    if (this.razClient == null) {
      throw new ServiceStateException("Raz client is not initialized. " +
          "The service should be started with " + this.getClass().getSimpleName() + ".start() " +
          "so the raz client will be initialized properly.");
    }
  }

  protected static RangerRazClient createRazClient(
      Configuration config,
      UserGroupInformation owner
      ) throws IOException {
    String configPrefix = config.get(RAZ_S3_CONFIG_PREFIX_NAME, RAZ_S3_CONFIG_PREFIX_DEFAULT);
    config.set(RangerRazClient.RAZ_CLIENT_PREFIX_KEY, configPrefix);
    return RangerRazClient.getInstance(config, owner);
  }

  public RazToken getRazToken() {
    return razToken;
  }
}
