/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apache.ranger.raz.hook.s3;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.token.Token;
import org.apache.hadoop.security.token.TokenIdentifier;
import org.apache.hadoop.security.token.TokenRenewer;
import org.apache.hadoop.security.token.delegation.web.DelegationTokenIdentifier;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRazClient;

import java.io.IOException;
import java.util.Locale;

import static org.apache.ranger.raz.hook.s3.RazS3ATokenIdentifier.RAZ_TOKEN_KIND;

/**
 * Implementation for renewing and cancelling RAZ
 * delegation tokens which will be used by the renewer
 * for renewing a RAZ DT for a long running job and cancelling
 * RAZ DT once the job has finished. Renewer figure out the
 * implementation using the config defined in path
 * {raz-hook-s3/src/main/resources/META-INF/services/
 * org.apache.hadoop.security.token.TokenRenewer}.
 */
public class RazS3ATokenRenewer extends TokenRenewer {

  private static final Logger LOG = LoggerFactory.getLogger(RazS3ATokenRenewer.class);

  /**
   * {@inheritDoc}
   */
  @Override
  public boolean handleKind(Text text) {
    return RAZ_TOKEN_KIND.equals(text);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public boolean isManaged(Token<?> token)
          throws IOException {
    // If tokens are handled then those can be renewed and canceled.
    return handleKind(token.getKind());
  }

  /**
   * {@inheritDoc}
   * Renews token by calling RAZ service through {@code RangerRazClient}.
   */
  @Override
  public long renew(Token<?> token, Configuration configuration)
          throws IOException, InterruptedException {
    long expiry = 0;

    TokenIdentifier identifier = token.decodeIdentifier();

    if (handleKind(identifier.getKind())) {
      LOG.info("Renewing {}", identifier);

      DelegationTokenIdentifier dtIdentifier = (DelegationTokenIdentifier) identifier;
      LOG.debug("Token: {}", dtIdentifier);

      // Default to the token's original expiration
      expiry = getTokenExpiration(dtIdentifier);

      UserGroupInformation user = UserGroupInformation.getCurrentUser();

      if (validateRenewer(user, dtIdentifier)) {
        final RazS3ATokenIdentifier razTokenId = (RazS3ATokenIdentifier) dtIdentifier;

        final RazToken razToken = razTokenId.getRazToken();
        RangerRazClient razClient = getRazClient(user, configuration);

        try {
          expiry = razClient.renewDelegationToken(razToken.getAccessToken());
        } catch (RangerRazException e) {
          throw new IOException("Error during DT renewal from RAZ server: " + e, e);
        }

        if (expiry == -1L) {
          LOG.error("Token {} could not be renewed.", identifier.getKind());
          throw new IOException(String.format("Token {%s} could not be renewed",
                  identifier.getKind()));
        }
      } else {
        throwInvalidRenewerException(user.getShortUserName(), identifier.getKind());
      }
    } else {
      LOG.error("Token kind {} can not be handled ", identifier.getKind());
    }

    return expiry;
  }

  private void throwInvalidRenewerException(String user, Text tokenKind) throws IOException {
    throw new IOException(String.format(ERR_INVALID_RENEWER,
            user, tokenKind));
  }
  private RangerRazClient getRazClient(UserGroupInformation owner, Configuration configuration)
      throws IOException {
    return RazDelegationTokenBinding.createRazClient(configuration, owner);
  }

  private static final String ERR_INVALID_RENEWER =
      "The user (%s) does not match the renewer declared for the token: %s";

  /**
   * Validate the user requesting the token lifecycle action against the renewer specified for that token.
   *
   * @param candidate  The user requesting the change.
   * @param identifier The identifier for the token to be changed.
   *
   * @return true, if the user is valid according to the token identifier; Otherwise, false.
   *
   * @throws IllegalArgumentException
   */
  private static boolean validateRenewer(UserGroupInformation candidate, DelegationTokenIdentifier identifier)
      throws IllegalArgumentException {
    boolean isValid = true;

    Text declaredRenewer = identifier.getRenewer();
    if (declaredRenewer != null && declaredRenewer.getLength() > 0) {
      if (!declaredRenewer.toString().equals(candidate.getShortUserName())) {
        LOG.error(String.format(Locale.getDefault(),
            ERR_INVALID_RENEWER,
            candidate.getUserName(),
            declaredRenewer));
        isValid = false;
      }
    } else {
      LOG.error("Operation not permitted. No renewer is specified in the identifier.");
      isValid = false;
    }

    return isValid;
  }

  private long getTokenExpiration(DelegationTokenIdentifier dtIdentifier) {
    return ((RazS3ATokenIdentifier) dtIdentifier).getExpiryTime();
  }

  /**
   * {@inheritDoc}
   * Cancels token by calling RAZ service through {@code RangerRazClient}.
   */
  @Override
  public void cancel(Token<?> token, Configuration configuration) throws IOException {
    TokenIdentifier identifier = token.decodeIdentifier();

    if (handleKind(identifier.getKind())) {
      LOG.info("Cancelling {}", identifier);

      DelegationTokenIdentifier dtIdentifier = (DelegationTokenIdentifier) identifier;
      LOG.debug("Token: {}", dtIdentifier);

      UserGroupInformation user = UserGroupInformation.getCurrentUser();

      if (validateRenewer(user, dtIdentifier)) {
        final RazS3ATokenIdentifier razTokenId = (RazS3ATokenIdentifier) dtIdentifier;

        final RazToken razToken = razTokenId.getRazToken();
        RangerRazClient razClient = getRazClient(user, configuration);

        Boolean cancellationResult = false;
        try {
          cancellationResult = razClient.cancelDelegationToken(razToken.getAccessToken());
          if (!cancellationResult) {
            throw new IOException(String.format("Token kind %s could not be cancelled.",
                    identifier.getKind()));
          }
        } catch (RangerRazException e) {
          throw new IOException("Error during DT renewal from RAZ server: " + e, e);
        }

      } else {
        throwInvalidRenewerException(user.getShortUserName(), identifier.getKind());
      }
    } else {
      LOG.error("Token kind {} can not be handled", identifier.getKind());
    }
  }
}
