/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.s3;

import com.amazonaws.AmazonServiceException;
import com.amazonaws.HandlerContextAware;
import com.amazonaws.Request;
import com.amazonaws.SignableRequest;
import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.CanHandleNullCredentials;
import com.amazonaws.auth.EndpointPrefixAwareSigner;
import com.amazonaws.auth.ServiceAwareSigner;
import com.amazonaws.auth.internal.SignerConstants;
import com.amazonaws.services.s3.Headers;
import com.amazonaws.services.signer.model.AccessDeniedException;
import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Preconditions;

import org.apache.hadoop.fs.s3a.audit.AWSAuditEventCallbacks;
import org.apache.ranger.raz.hook.s3.cache.SignerCache;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsRequest;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsRequestCacheKey;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.s3a.CredentialInitializationException;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.raz.hook.s3.utils.S3Utils;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.s3.lib.RangerRazExceptionS3Headers;
import org.apache.ranger.raz.intg.client.RangerRazClient;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.raz.model.RangerRazResultBase;
import org.apache.ranger.raz.s3.lib.HttpMethod;
import org.apache.ranger.raz.s3.lib.aws.exceptions.RCSAWSException;
import org.apache.ranger.raz.s3.lib.aws.parsers.AwsRequestParser;
import org.apache.ranger.raz.s3.lib.signer.Constants;
import org.apache.ranger.raz.s3.lib.signer.SignResponse;
import org.apache.ranger.raz.s3.lib.signer.impl.SignRequestImpl;
import org.apache.ranger.raz.s3.lib.signer.util.ProtoConverters;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.UUID;

import static org.apache.hadoop.fs.s3a.audit.AuditIntegration.retrieveAttachedSpan;
import static org.apache.ranger.raz.model.RangerRazRequestBase.*;

/**
 * An AWS signer implementation which makes a (remote) call to a trusted service for Raz policy
 * authorization and signing of the request.
 *
 * Some of the restrictions with the signer. - Content will never be signed / sent over to the
 * remote signer. - Supports the v4 protocol only as of now. - Several client config parameters are
 * not honored.
 */
public class RazS3SignerPlugin
    implements com.amazonaws.auth.Signer, ServiceAwareSigner,
    EndpointPrefixAwareSigner, CanHandleNullCredentials {

  private static final Logger LOG = LoggerFactory.getLogger(
      org.apache.ranger.raz.hook.s3.RazS3SignerPlugin.class);

  private static final Set<String>
      listOfSignerHeadersToLogWhenReplaced = new HashSet<>(
      Arrays.asList(SignerConstants.HOST, SignerConstants.X_AMZ_DATE,
          SignerConstants.X_AMZ_CONTENT_SHA256,
          SignerConstants.AUTHORIZATION));

  private String endpointPrefix;
  private String serviceName;

  static final String DEFAULT_SERVICE_TYPE = "s3";
  static final String DEFAULT_SERVICE_NAME = "cm_s3";

  static final String SERVICE_TYPE_CONFIG_KEY = "s3a.service.type";
  static final String SERVICE_NAME_CONFIG_KEY = "s3a.service.name";
  static final String CLUSTER_NAME_CONFIG_KEY = ".access.cluster.name";

  static final String ACCESS_DENIED_EXCEPTION = "AccessDeniedException";

  static final String UNKNOWN_OP_NAME = "UNKNOWN_AWS_OPERATION";
  static final String PREFIX_TO_TRUNCATE = "op_";

  public RazS3SignerPlugin() {
    if (LOG.isDebugEnabled()) {
      LOG.debug("==> Raz S3 Signer client constructor");
    }

    if (LOG.isDebugEnabled()) {
      LOG.debug("<== Raz S3 Signer client constructor");
    }
  }

  /**
   * This signs all request towards S3.
   * Signs using remote RAZ service.
   *
   * @param request to be signed.
   * @param credentials should be RazAnonymousAWSCredentials with DT
   *                    information required for the singing.
   */
  @Override
  public void sign(SignableRequest<?> request, AWSCredentials credentials) {
    if (LOG.isDebugEnabled()) {
      LOG.debug("==> Sign request: {}, AWSCredentials: {}", request, credentials);
    }

    // skip signing if credentials are null (default behaviour in aws signers)
    if (credentials.getAWSSecretKey() == null) {
      if (LOG.isDebugEnabled()) {
        LOG.debug("Secret key is null from credentials. Returning from the signer without signing the request.");
      }
      return;
    }

    String razToken = null;
    UserGroupInformation owner = null;
    String userName = null;
    Configuration configuration = null;
    RangerRazClient razClient = null;
    SignerCache<AwsRequestCacheKey, Map<String, String>> cache = null;
    String operation = retrieveOperationName(request);

    // Check if credentials is the required type
    if (credentials instanceof RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials) {
      RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials
          razAnonymousAWSCredentials =
          (RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials) credentials;

      final RazDelegationTokenBinding razDelegationTokenBinding
          = razAnonymousAWSCredentials.getRazDelegationTokenBinding();

      if (LOG.isDebugEnabled()) {
        LOG.debug("Getting raz delegation token from the filesystem.");
      }

      // Raz token will be null in case of unbonded mode.
      if (razDelegationTokenBinding.getRazToken() != null) {
        razToken = razDelegationTokenBinding.getRazToken().getAccessToken();
      }
      owner = razAnonymousAWSCredentials.getOwner();
      userName = owner.getUserName();
      razClient = razAnonymousAWSCredentials.getRazClient();
      cache = razAnonymousAWSCredentials.getSignerCache();

      if (razClient == null) {
        throw new RuntimeException("Raz client from DT binding is null.");
      }

      //  The configuration is patched with the S3_CONFIG_PREFIX.
      //  This prefix will be passed to the RAZ client, so it can read the endpoint
      //  for the RAZ service in environments such as Tez.
      configuration = razAnonymousAWSCredentials.getConfig();

      // RangerRazClient instance is stored in the RazAnonymousAWSCredentials,
      // so it's lifecycle matches the S3AFileSystem's lifecycle.
      // It will be strongly coupled with the file system and the owner.
      if (LOG.isDebugEnabled()) {
        if (!StringUtils.isEmpty(razToken)) {
          LOG.debug("razToken: {}, owner: {}, userName: {}", razToken.substring(0, 6), owner, userName);
        } else {
          LOG.debug("No razToken found, owner: {}, userName: {}", owner, userName);
        }
      }
    } else {
      final String eMsg = String.format(
          "Credentials are not an instance of %s, classname: %s, toString: %s",
          RazAnonymousAWSCredentialsProvider.class.getName(),
          credentials.getClass().getName(),
          credentials.toString());
      throw new CredentialInitializationException(eMsg);
    }

    String bucketName = S3Utils
        .getBucketNameFromHost(request.getEndpoint().getHost());
    Objects.requireNonNull(bucketName,
        "Unable to look up bucketname from hostname=[" + request.getEndpoint()
            .getHost() + "]");
    String regionName = S3Utils.lookupRegionFromCache(bucketName);

    AwsRequestParser requestParser = new AwsRequestParser();

    byte[] contentBytes = null;
    if (requestParser.isContentSigningRequired((Request) request)) {
      if (LOG.isDebugEnabled()) {
        LOG.debug("Content signing required for Request: {}", request);
      }
      try {
        contentBytes = requestParser.getRequestContents((Request) request);
      } catch (IOException e) {
        throw new RCSAWSException(RCSAWSException.StatusCode.OTHER,
            "Unable to extract request contents for request: [" + request + "]",
            e);
      }
    }

    AwsRequest awsRequest = requestParser.parseAwsRequest((Request) request);
    AwsRequestCacheKey awsRequestCacheKey = new AwsRequestCacheKey(awsRequest, userName);
    if (LOG.isDebugEnabled()) {
      LOG.debug("Incoming request: {} Cache key generated: {} ", awsRequest, awsRequestCacheKey);
    }
    Map<String, String> cachedSignedHeaders = cache.get(awsRequestCacheKey);
    if (cachedSignedHeaders == null) {
      SignRequestImpl signRequest = new SignRequestImpl(request.getEndpoint(),
              HttpMethod.valueOf(request.getHttpMethod().name()),
              request.getHeaders(),
              request.getParameters(), request.getResourcePath(), contentBytes,
              request.getTimeOffset(),
              endpointPrefix, serviceName);
      if (LOG.isDebugEnabled()) {
        LOG.debug("Sign request: {}", signRequest);
      }

      // Serialize into a base64(protobuf)
      Map<String, String> rContext = new HashMap<>();
      try {
        final String signReq = ProtoConverters.toBase64(signRequest);
        rContext.put(Constants.S3_SIGN_REQUEST_FIELD, signReq);
      } catch (IOException e) {
        throw new RuntimeException("Can't convert sign request to protobuffed base64.", e);
      }

      // Add to raz request, which will be sent over the wire
      RangerRazRequest razRequest = getRangerRazRequestWithSetup(configuration,
              rContext,
              userName,
              operation);

      // Check privilege with the raz client
      RangerRazResult rangerRazResult = null;
      SignResponse signResponse = null;


      try {
        rangerRazResult = razClient.checkPrivilege(razRequest, razToken);
        // Handle RangerRazResult being null
        if (rangerRazResult == null) {
          // user hasn't kinit'ed.
          if (!isKerberosAuthenticated()) {
            throw new RuntimeException("User: " + userName + " doesn't have Kerberos Authentication.");
          }
          // user has kinit'ed but RAZ server is down which gives empty RAZ DT.
          else {
            throw new RuntimeException("User: " + userName + " have kinited but"
                    + " RAZ server may be down. Please verify if RAZ server is up or not.");
          }
        }
      } catch (RangerRazException e) {
        // Throw an AmazonServiceException(retryable or non-retryable) else a
        // RazS3ClientCredentialException(non-retryable).
        rangerRazExceptionToASEAndThrow(e, razRequest.getRequestId());

        LOG.error("Exception while checking privilege, RazRequestID: " +
                razRequest.getRequestId(), e);
        throw new RazS3ClientCredentialsException(e.getMessage(), e);
      }

      // Check if we received an access denied from the RANGER side
      // if yes, construct an exception with a message
      final RangerRazResultBase.AccessResult accessResult
              = rangerRazResult.getOperResult().getResult();
      switch (accessResult) {
        case ALLOWED:
          break;
        case DENIED:
        case NOT_DETERMINED:
          throw constructAccessDeniedException(rangerRazResult, userName);
        default:
          throw new UnsupportedOperationException("Unknown operation result from RAZ: " +
                  accessResult);
      }

      try {
        final String signResponseStr = rangerRazResult.getOperResult()
                .getAdditionalInfo()
                .get(Constants.S3_SIGN_RESPONSE_FIELD);
        signResponse = ProtoConverters.respFromBase64(signResponseStr);
      } catch (IOException e) {
        throw new UncheckedIOException("RAZ result cannot be read. " +
                "request id: " + razRequest.getRequestId() +
                " exception message: " + e.toString(),
                e);
      }
      maybeUpdateRegionNameInCache(bucketName, regionName, signResponse);

      Map<String, String> signerGeneratedHeaders = signResponse.getSignerGeneratedHeaders();

      if (awsRequest.isSignatureCachingAllowed()) {
        if (LOG.isDebugEnabled()) {
          LOG.debug("Caching aws request key : {}", awsRequestCacheKey);
        }
        cache.put(awsRequestCacheKey, signerGeneratedHeaders);
      }
      addSignedHeadersToRequest(request, signerGeneratedHeaders);
      if (LOG.isDebugEnabled()) {
        LOG.debug("<== Sign request end. Request: {}, Credentials: {}, Result: {}", request, credentials, rangerRazResult);
      }
    }
    else {
      if (LOG.isDebugEnabled()) {
        LOG.debug("Reusing aws request key from cache : {}", awsRequestCacheKey);
      }
      addSignedHeadersToRequest(request, cachedSignedHeaders);
    }

  }

  private String retrieveOperationName(SignableRequest<?> request) {
    if (request instanceof HandlerContextAware) {
      AWSAuditEventCallbacks awsAuditEventCallbacks = retrieveAttachedSpan((HandlerContextAware) request);
      if (awsAuditEventCallbacks != null) {
        String opName = awsAuditEventCallbacks.getOperationName();
        return getOpNameWithoutPrefix(opName);
      }
    }
    LOG.debug("Unable to figure out operation name from incoming request." +
            "So setting to {}", UNKNOWN_OP_NAME);
    return UNKNOWN_OP_NAME;
  }

  @VisibleForTesting
  public String getOpNameWithoutPrefix(String opName) {
    return opName.startsWith(PREFIX_TO_TRUNCATE)
            ? opName.substring(PREFIX_TO_TRUNCATE.length())
            : opName;
  }

  /**
   * A method used to check if the RangerRazException raised is an instance
   * of AmazonServiceException and to re-construct this exception to be
   * further translated in S3AUtils.
   *
   * @param e            RangerRazException raised.
   * @param razRequestId RazRequestID associated with this request.
   */
  private void rangerRazExceptionToASEAndThrow(RangerRazException e,
      String razRequestId) {
    Map<String, String> exceptionHeaderMap = e.getRangerRazExceptionHeadersMap();
    // An exception with headers was parsed in the client and contains an AWS
    // header.
    if (exceptionHeaderMap != null &&
        exceptionHeaderMap.containsKey(RangerRazExceptionS3Headers.AWS_STATUS_CODE)) {
      AmazonServiceException ase =
          new AmazonServiceException(
              e.getMessage() + ", RazRequestID: " + razRequestId);
      ase.setServiceName("Amazon S3");

      // Check the headers and set them in ASE.
      if (exceptionHeaderMap
          .containsKey(RangerRazExceptionS3Headers.AWS_ERROR_CODE)) {
        ase.setErrorCode(
            exceptionHeaderMap.get(RangerRazExceptionS3Headers.AWS_ERROR_CODE));
      }
      if (exceptionHeaderMap
          .containsKey(RangerRazExceptionS3Headers.AWS_STATUS_CODE)) {
        String statusCodeStr =
            exceptionHeaderMap.get(RangerRazExceptionS3Headers.AWS_STATUS_CODE);
        try{
          ase.setStatusCode(Integer.parseInt(statusCodeStr));
        } catch (NumberFormatException numberFormatException) {
          LOG.error("String: " + statusCodeStr + " can't be parsed into an Integer.");
          // Can't parse status code string so set to 400.
          ase.setStatusCode(400);
        }
      }
      if (exceptionHeaderMap
          .containsKey(RangerRazExceptionS3Headers.AWS_REQUEST_ID)) {
        ase.setRequestId(
            exceptionHeaderMap.get(RangerRazExceptionS3Headers.AWS_REQUEST_ID));
      }
      if (exceptionHeaderMap
          .containsKey(RangerRazExceptionS3Headers.AWS_EXTENDED_REQUEST_ID)) {
        // ExtendedRequestID is a header in ASE
        Map<String, String> extendedRequestIDHeader = new HashMap<>();
        extendedRequestIDHeader.put(Headers.EXTENDED_REQUEST_ID,
            exceptionHeaderMap
                .get(RangerRazExceptionS3Headers.AWS_EXTENDED_REQUEST_ID));
        ase.setHttpHeaders(extendedRequestIDHeader);
      }

      throw ase;
    }
  }

  /**
   * Method to verify if the user is kerberos authenticated or not.
   *
   * @return true if user is kerberos authenticated else false.
   */
  private boolean isKerberosAuthenticated() {
    try {
      UserGroupInformation loggedInUser = UserGroupInformation.getLoginUser();
      boolean isSecurityEnabled = UserGroupInformation.isSecurityEnabled();
      boolean hasKerberosCredentials = loggedInUser.hasKerberosCredentials();
      UserGroupInformation.AuthenticationMethod loggedInUserAuthMethod = loggedInUser.getAuthenticationMethod();

      return isSecurityEnabled && hasKerberosCredentials
          && loggedInUserAuthMethod.equals(UserGroupInformation.AuthenticationMethod.KERBEROS);
    } catch (IOException e) {
      throw new RuntimeException("Failed to get authentication details, Exception: " + e.toString());
    }
  }

  @Override
  public void setEndpointPrefix(String endpointPrefix) {
    this.endpointPrefix = endpointPrefix;
  }

  @Override
  public void setServiceName(String serviceName) {
    this.serviceName = serviceName;
  }

  private AccessDeniedException constructAccessDeniedException(RangerRazResult razResult, String username) {
    // The userName, group etc in the error response is populated by the Ranger Authorizer itself.
    final String msg =
        String.format("Ranger result: %s, Audit: %s, Username: %s",
            razResult.getOperResult().getResult(),
            razResult.getOperResult().getAuditLogs(),
            username);
    AccessDeniedException ade = new AccessDeniedException(msg);
    ade.setStatusCode(RCSAWSException.StatusCode.ACCESS_DENIED.getCode());
    ade.setErrorCode(ACCESS_DENIED_EXCEPTION);
    ade.setRequestId(razResult.getRequestId());
    if (LOG.isDebugEnabled()) {
      LOG.debug("Constructing AccessDeniedException for RangerRazResult (for raz, not for signer).");
    }
    return ade;
  }

  private void maybeUpdateRegionNameInCache(String bucketName,
      String regionName,
      SignResponse signResponse) {
    if (signResponse.getRegion() != null) {
      if (regionName != null) {
        Preconditions
            .checkState(regionName.equalsIgnoreCase(signResponse.getRegion()),
                "region and region from signedResponse must match");
        return;
      }
      if (LOG.isDebugEnabled()) {
        LOG.debug(
                "Updating region in cache for bucket=[{}], from sign response to [{}] ",
                bucketName,
                signResponse.getRegion());
      }
      S3Utils.updateBucketToRegionCache(bucketName, signResponse.getRegion());
    }
  }

  private void addSignedHeaderToRequest(SignableRequest<?> request,
      String headerName,
      String value) {
    if (listOfSignerHeadersToLogWhenReplaced.contains(headerName) &&
        request.getHeaders().containsKey(headerName)) {
      LOG.warn("Original request already contains the header: {}. Replacing...",
          headerName);
    }
    request.addHeader(headerName, value);
  }

  private void addSignedHeadersToRequest(SignableRequest<?> request,
      Map<String, String> signedHeaders) {
    for (Map.Entry<String, String> entry : signedHeaders.entrySet()) {
      addSignedHeaderToRequest(request, entry.getKey(), entry.getValue());
    }
  }

  public static RangerRazRequest getRangerRazRequestWithSetup(
      Configuration configuration,
      Map<String, String> rContext,
      String userName,
      String operation) {
    RangerRazRequest razRequest = new RangerRazRequest();
    razRequest.setContext(rContext);
    final String S3_CONFIG_PREFIX = configuration.get(RazDelegationTokenBinding.RAZ_S3_CONFIG_PREFIX_NAME,
        RazDelegationTokenBinding.RAZ_S3_CONFIG_PREFIX_DEFAULT);

    String serviceType = configuration.get(
        S3_CONFIG_PREFIX + SERVICE_TYPE_CONFIG_KEY, DEFAULT_SERVICE_TYPE);
    razRequest.setServiceType(serviceType);
    razRequest.setServiceName(configuration.get(
        S3_CONFIG_PREFIX + SERVICE_NAME_CONFIG_KEY, DEFAULT_SERVICE_NAME));
    razRequest.setUser(userName);
    razRequest.setKeysToRedact(RangerRazClient.RAZ_CLIENT_KEYS_TO_REDACT);
    ResourceAccess resourceAccess = new ResourceAccess(new HashMap<>(), operation);
    razRequest.setOperation(resourceAccess);

    String clusterName = configuration.get(S3_CONFIG_PREFIX + serviceType + CLUSTER_NAME_CONFIG_KEY, "");
    razRequest.setClusterName(clusterName);
    // this is for debugging - corresponding request could be found easier
    razRequest.setRequestId(String.valueOf(UUID.randomUUID()));
    return razRequest;
  }

  /**
   * Method used to test the translation of RangerRazException to
   * AmazonServiceException.
   * @param e RangerRazException.
   * @param razRequestId RequestID for a specific Raz request.
   */
  @VisibleForTesting
  public void translateToASEAndThrow(RangerRazException e,
      String razRequestId) {
    rangerRazExceptionToASEAndThrow(e, razRequestId);
  }
}