/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.s3;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.Serializable;
import java.time.Duration;
import java.time.Instant;
import java.time.temporal.TemporalUnit;

import static java.time.temporal.ChronoUnit.SECONDS;

/**
 * This is the delegation token for Raz S3.
 *
 * This is loaded in YARN RM and must not contain dependencies on things
 * which may not be on the classpath.
 */
public class RazToken implements Writable, Serializable {
  private static final long serialVersionUID = 7508655772386801692L;

  public static final int MAX_TEXT_LENGTH = 32768;
  public static final int RAZTOKEN_EXPIRATION_OFFSET_SECONDS = 120;

  private static final Logger LOG = LoggerFactory.getLogger(
      org.apache.ranger.raz.hook.s3.RazToken.class);

  /**
   * Timestamp when the token was issued.
   */
  long issueTime;

  /**
   * Expiry time, UTC seconds since the epoch.
   */
  long expiryTime;

  /**
   * Raz token used.
   */
  String accessToken;

  public RazToken() {
  }

  public RazToken(long issueTime, long expiryTime, String accessToken) {
    this.issueTime = issueTime;
    this.expiryTime = expiryTime;
    this.accessToken = accessToken;
  }

  public long getIssueTime() {
    return issueTime;
  }

  public long getExpiryTime() {
    return expiryTime;
  }

  public String getAccessToken() {
    return accessToken;
  }

  @Override
  public void write(DataOutput out) throws IOException {
    out.writeLong(issueTime);
    out.writeLong(expiryTime);
    Text.writeString(out, accessToken);
  }

  @Override
  public void readFields(DataInput in) throws IOException {
    this.issueTime = in.readLong();
    this.expiryTime = in.readLong();
    this.accessToken = Text.readString(in, MAX_TEXT_LENGTH);
  }

  public boolean isAboutToExpire() {
    return isAboutToExpire(this.expiryTime, RAZTOKEN_EXPIRATION_OFFSET_SECONDS, SECONDS);
  }

  /*
   * Todo add tests
   */
  public boolean isAboutToExpire(long expiryTime, long offset, TemporalUnit unit) {
    Instant now = Instant.now();

    LOG.debug("Raz token expiration:" +
            "\n\tExpires in:\t{} seconds" +
            "\n\tOffset:\t{} seconds" +
            "\n\tTime left:\t{} seconds",
        expiryTime - now.getEpochSecond(),
        Duration.of(offset, unit).getSeconds(),
        expiryTime - now.plus(offset, unit).getEpochSecond());

    // Expired if the expiration time is the same as or before the current time plus any offset.
    return !Instant.ofEpochSecond(expiryTime).isAfter(now.plus(offset, unit));
  }

  @Override
  public String toString() {
    StringBuilder str = new StringBuilder();
    str.append("RazToken{")
        .append("issueTime=").append(issueTime)
        .append(", expiryTime=").append(expiryTime);
    str.append("}");
    return str.toString();
  }
}
