/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.s3.cache;

import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsRequestCacheKey;

import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 * Implementation of signer cache which internally uses google common
 * cache libraries. All the operations are threadsafe.
 */
public class SignerCacheImpl implements SignerCache<AwsRequestCacheKey, Map<String, String>> {

    /**
     * Cache which will hold {@code AwsRequestCacheKey} as key and
     * map of signed headers as value.
     */
    private final Cache<AwsRequestCacheKey, Map<String, String>> signerCache;

    /**
     * Default value of cache max size.
     */
    public static final long CACHE_MAX_SIZE_DEFAULT = 100;

    /**
     * Default value of cache expiry in seconds.
     */
    public static final long CACHE_KEY_EXPIRY_IN_SECONDS_DEFAULT = 30;

    public SignerCacheImpl() {
        this(CACHE_MAX_SIZE_DEFAULT, CACHE_KEY_EXPIRY_IN_SECONDS_DEFAULT);
    }

    /**
     *
     * @param maxSize maximum size of cache.
     * @param expiry key expiry time in seconds.
     */
    public SignerCacheImpl(long maxSize, long expiry) {
        this.signerCache = CacheBuilder
                .newBuilder()
                .maximumSize(maxSize)
                .expireAfterWrite(expiry, TimeUnit.SECONDS)
                .build();
    }

    public void put(AwsRequestCacheKey key, Map<String, String> value) {
        signerCache.put(key, value);
    }

    public Map<String, String> get(AwsRequestCacheKey key) {
        return signerCache.getIfPresent(key);
    }

    public long size() {
        return signerCache.size();
    }

}
