/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.s3;

import java.io.IOException;
import java.nio.file.AccessDeniedException;
import java.util.HashMap;
import java.util.Map;

import com.amazonaws.AmazonServiceException;
import org.junit.Assert;
import org.junit.Test;

import org.apache.hadoop.fs.s3a.AWSBadRequestException;
import org.apache.hadoop.fs.s3a.S3AUtils;
import org.apache.hadoop.fs.s3a.UnknownStoreException;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.s3.lib.RangerRazExceptionS3Headers;

/**
 * Test to verify RazS3 Exception Handling is translated correctly.
 */
public class TestRazS3ExceptionHandling {

  private static final String ERROR_CODE_NO_SUCH_BUCKET = "NoSuchBucket";
  private static final int STATUS_CODE_NO_SUCH_BUCKET = 404;
  private static final String ERROR_CODE_INVALID_BUCKET = "InvalidBucketName";
  private static final int STATUS_CODE_NO_INVALID_BUCKET = 400;
  private static final String ERROR_CODE_ACCESS_DENIED = "AccessDenied";
  private static final int STATUS_CODE_ACCESS_DENIED = 403;

  /**
   * Test to verify the correct construction of RangerRazException containing
   * S3 headers to AmazonServiceException correctly and then to map to
   * correct exceptions in S3AUtils.
   */
  @Test
  public void testRangerRazExceptionWithHeadersToASE() {
    testTranslatingRangerRazExceptionWithHeadersToASE(ERROR_CODE_NO_SUCH_BUCKET,
        STATUS_CODE_NO_SUCH_BUCKET);
    testTranslatingRangerRazExceptionWithHeadersToASE(ERROR_CODE_INVALID_BUCKET,
        STATUS_CODE_NO_INVALID_BUCKET);
    testTranslatingRangerRazExceptionWithHeadersToASE(ERROR_CODE_ACCESS_DENIED,
        STATUS_CODE_ACCESS_DENIED);
  }

  /**
   * Test to verify RangerRazException without headers won't be converted to
   * AmazonServiceException.
   */
  @Test
  public void testRangerRazExceptionWithoutHeadersToASE() {
    RangerRazException rangerRazExceptionWithoutHeaders =
        new RangerRazException("RangerRazException without headers");
    testTranslatingRangerRazExceptionToASE(rangerRazExceptionWithoutHeaders);
  }

  /**
   * Test to verify RangerRazException with null header-map won't be converted
   * to AmazonServiceException.
   */
  @Test
  public void testRangerRazExceptionWithNullHeadersToASE() {
    RangerRazException rangerRazExceptionWithNullHeaders =
        new RangerRazException("RangerRazException with null headers");
    rangerRazExceptionWithNullHeaders.setRangerRazExceptionHeadersMap(null);
    testTranslatingRangerRazExceptionToASE(rangerRazExceptionWithNullHeaders);
  }

  /**
   * Method to assert that RangerRazException with no valid headers wouldn't
   * be converted into ASE.
   * @param rangerRazException RangerRazException to be converted.
   */
  private void testTranslatingRangerRazExceptionToASE(RangerRazException rangerRazException) {
    try {
      RazS3SignerPlugin s3SignerPlugin = new RazS3SignerPlugin();
      s3SignerPlugin.translateToASEAndThrow(rangerRazException,
          "raz-request-id");
    } catch (AmazonServiceException ase) {
      Assert.fail("RangerRazException without headers shouldn't throw ASE");
    }
  }

  /**
   * Method to verify the correct translating to RangerRazException with
   * headers to ASE and then the ASE to correct mapped exception in S3AUtils.
   * @param errorCode S3 error code.
   * @param statusCode S3 status code.
   */
  private void testTranslatingRangerRazExceptionWithHeadersToASE(String errorCode,
      int statusCode) {

    // Creating a Mock RangerRazException to emulate parsed state.
    Map<String, String> headersInRazException = new HashMap<>();
    headersInRazException.put(RangerRazExceptionS3Headers.AWS_ERROR_CODE,
        errorCode);
    headersInRazException.put(RangerRazExceptionS3Headers.AWS_STATUS_CODE,
        String.valueOf(statusCode));
    headersInRazException.put(RangerRazExceptionS3Headers.AWS_REQUEST_ID,
        "some-request-id");
    headersInRazException.put(RangerRazExceptionS3Headers.AWS_EXTENDED_REQUEST_ID,
            "some-ext-request-id");
    RangerRazException rangerRazException = new RangerRazException("some error");
    rangerRazException.setRangerRazExceptionHeadersMap(headersInRazException);

    // Trying to translate a RangerRazException into ASE in Signer.
    String razRequestId = "raz-request-id";

    // if AmazonServiceException is thrown, this test would pass.
    try {
      RazS3SignerPlugin s3SignerPlugin = new RazS3SignerPlugin();
      s3SignerPlugin.translateToASEAndThrow(rangerRazException,
          razRequestId);
      // No exception was thrown, so test failed.
      Assert.fail("Expected to throw AmazonServiceException.class");
    } catch (AmazonServiceException ase) {
      // assert if ASE is translated properly in S3AUtils.
      testTranslatingASEToS3AUtils(ase);
    }
  }

  /**
   * Method to test if raised ASE is translating properly in S3AUtils.
   *
   * @param ase AmazonServiceException raised.
   */
  private void testTranslatingASEToS3AUtils(
      AmazonServiceException ase) {
    IOException ioe = S3AUtils.translateException("testOp", "testPath", ase);
    if (ase.getErrorCode().equals(ERROR_CODE_NO_SUCH_BUCKET)
        && ase.getStatusCode() == STATUS_CODE_NO_SUCH_BUCKET) {
      Assert.assertTrue("Mismatch in exception type: expected "
              + "UnknownStoreException.class"
              + " but got " + ioe.getClass(),
          ioe instanceof UnknownStoreException);
    } else if (ase.getErrorCode().equals(ERROR_CODE_INVALID_BUCKET)
        && ase.getStatusCode() == STATUS_CODE_NO_INVALID_BUCKET) {
      Assert.assertTrue("Mismatch in exception type: expected "
              + "AWSBadRequestException.class"
              + " but got " + ioe.getClass(),
          ioe instanceof AWSBadRequestException);
    } else if (ase.getErrorCode().equals(ERROR_CODE_ACCESS_DENIED)
        && ase.getStatusCode() == STATUS_CODE_ACCESS_DENIED) {
      Assert.assertTrue("Mismatch in exception type: expected "
          + "AccessDeniedException.class"
          + " but got " + ioe.getClass(),
          ioe instanceof AccessDeniedException);
    }
  }
}
