/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.s3;

import com.amazonaws.Request;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.raz.hook.s3.cache.SignerCacheImpl;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRazClient;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.raz.model.RangerRazResultBase;
import org.apache.ranger.raz.s3.lib.aws.util.AwsRequestGenerator;
import org.apache.ranger.raz.s3.lib.signer.SignResponse;
import org.apache.ranger.raz.s3.lib.signer.impl.SignResponseImpl;
import org.apache.ranger.raz.s3.lib.signer.util.ProtoConverters;
import org.junit.Test;

import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.mockito.Matchers;

import java.util.HashMap;
import java.util.Map;

import static org.apache.ranger.raz.hook.s3.RazDelegationTokenBinding.RAZ_S3_CONFIG_PREFIX_DEFAULT;
import static org.apache.ranger.raz.s3.lib.signer.Constants.S3_SIGN_RESPONSE_FIELD;
import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.*;

public class TestRazS3SignerPlugin {

  private static final String REGION_NAME = "us-west-2";
  private static final String FAKE_BUCKET_NAME = "fakebucketname";
  private static final String FAKE_PATH = "path/under/test";
  private static final String FAKE_PATH_2 = "path/under/test2";

  @Test
  public void testGetRangerRazRequestWithSetup() throws Exception{
    Configuration configuration = new Configuration();
    final String serviceType = "serviceType";
    final String serviceName = "serviceName";
    final String clusterName = "clusterName";
    final String userName = "userName";
    Map<String, String> rContext = new HashMap<>();
    final String testKey = "testKey";
    final String testValue = "testValue";
    final String operation ="dummy_op";

    configuration.set(RAZ_S3_CONFIG_PREFIX_DEFAULT + RazS3SignerPlugin.SERVICE_TYPE_CONFIG_KEY, serviceType);
    configuration.set(RAZ_S3_CONFIG_PREFIX_DEFAULT + RazS3SignerPlugin.SERVICE_NAME_CONFIG_KEY, serviceName);
    configuration.set(RAZ_S3_CONFIG_PREFIX_DEFAULT + serviceType + RazS3SignerPlugin.CLUSTER_NAME_CONFIG_KEY, clusterName);
    rContext.put(testKey, testValue);

    final RangerRazRequest razRequest
        = RazS3SignerPlugin.getRangerRazRequestWithSetup(configuration, rContext, userName, operation);

    assertEquals("ServiceType should match",serviceType, razRequest.getServiceType());
    assertEquals("ServiceName should match", serviceName, razRequest.getServiceName());
    assertEquals("ClusterName should match", clusterName, razRequest.getClusterName());
    assertEquals("UserName should match", userName, razRequest.getUser());
    assertEquals("Map kv should match", testValue, razRequest.getContext().get(testKey));
    assertEquals("Operation name should match ", operation, razRequest.getOperation().getAction());
  }

  @Test
  public void TestOpNameTruncation() {
    RazS3SignerPlugin plugin = new RazS3SignerPlugin();
    assertEquals("Truncated operation name should match",
            "file_status", plugin.getOpNameWithoutPrefix("op_file_status"));
    assertEquals("Truncated operation name should match",
            "rename", plugin.getOpNameWithoutPrefix("op_rename"));
    assertEquals("Truncated operation name should match",
            "rename", plugin.getOpNameWithoutPrefix("op_rename"));
    assertEquals("Truncated operation name should match",
            "_test", plugin.getOpNameWithoutPrefix("_test"));
    assertEquals("Truncated operation name should match",
            "unknown", plugin.getOpNameWithoutPrefix("unknown"));

  }

  @Test
  public void testSignatureCaching() throws Exception {
    RazS3SignerPlugin razS3SignerPlugin = new RazS3SignerPlugin();
    SignerCacheImpl signerCache = new SignerCacheImpl();

    final AwsRequestGenerator requestGenerator = new AwsRequestGenerator(REGION_NAME);
    Request<?> getObjectRequest = requestGenerator.generateGetObjectRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    Request<?> getObjectRequestDiffPath = requestGenerator.generateGetObjectRequest(FAKE_BUCKET_NAME, FAKE_PATH_2);

    Pair<RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials, RangerRazClient>
            razAnonymousAWSCredentialsRangerRazClientPair = prepareMocks(signerCache);

    RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials credentials =
            razAnonymousAWSCredentialsRangerRazClientPair.getLeft();
    RangerRazClient razClient = razAnonymousAWSCredentialsRangerRazClientPair.getRight();

    // initial read request should call RAZ for checking privilege.
    razS3SignerPlugin.sign(getObjectRequest, credentials);
    assertCachingDetails(razClient, signerCache, 1, 1);

    // same read request shouldn't call RAZ for checking privilege.
    razS3SignerPlugin.sign(getObjectRequest, credentials);
    assertCachingDetails(razClient, signerCache, 0, 1);

    // read request on a different path should call RAZ for checking privilege.
    razS3SignerPlugin.sign(getObjectRequestDiffPath, credentials);
    assertCachingDetails(razClient, signerCache, 1, 2);

    //Non read-requests should call RAZ for checking privilege.
    Request<?> deleteObjReq = requestGenerator.generateDeleteObjectRequest(FAKE_BUCKET_NAME, FAKE_PATH_2);
    razS3SignerPlugin.sign(deleteObjReq, credentials);
    assertCachingDetails(razClient, signerCache, 1, 2);

    //Read request on same path but by different user should RAZ for checking privilege.
    prepareUGIMock(credentials, "user2");
    razS3SignerPlugin.sign(getObjectRequest, credentials);
    assertCachingDetails(razClient, signerCache, 1, 3);


  }
  private void assertCachingDetails(RangerRazClient razClient, SignerCacheImpl signerCache,
                                    int expectedCheckPrivInvocations, int expectedCurrentCacheSize) throws RangerRazException {
    verify(razClient, times(expectedCheckPrivInvocations)).checkPrivilege(Matchers.any(), Matchers.any());
    assertEquals("Cache size mismatch", expectedCurrentCacheSize, signerCache.size());
    clearInvocations(razClient);
  }

  private Pair<RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials, RangerRazClient>
      prepareMocks(SignerCacheImpl signerCache)
            throws Exception {
    RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials credentials = prepareCredentialsMock(signerCache);
    RangerRazClient rangerRazClient = prepareRazClientMocks(credentials);
    prepareRazResultMocks(rangerRazClient);
    prepareUGIMock(credentials, "user1");
    return Pair.of(credentials, rangerRazClient);
  }

  private void prepareRazResultMocks(RangerRazClient rangerRazClient) throws Exception {
    RangerRazResult rangerRazResult = mock(RangerRazResult.class);
    RangerRazResultBase.ResourceAccessResult resourceAccessResult = mock(RangerRazResultBase.ResourceAccessResult.class);
    when(rangerRazClient.checkPrivilege(Matchers.any(), Matchers.any())).thenReturn(rangerRazResult);
    when(rangerRazResult.getOperResult()).thenReturn(resourceAccessResult);
    when(resourceAccessResult.getResult()).thenReturn(RangerRazResultBase.AccessResult.ALLOWED);
    HashMap<String, String> signedHeaders = new HashMap<>();
    signedHeaders.put("key", "value");
    SignResponse.AuthorizationResponse authorizationResponse = mock(SignResponse.AuthorizationResponse.class);
    when(authorizationResponse.getResponseCode()).thenReturn(SignResponse.AuthorizationResponse.ResponseCode.AUTHZ_AUTHORIZED);
    SignResponse signResponse = new SignResponseImpl(authorizationResponse, signedHeaders, REGION_NAME);
    String signResStr = ProtoConverters.toBase64(signResponse);
    Map<String, String> addInfo = new HashMap<>();
    addInfo.put(S3_SIGN_RESPONSE_FIELD, signResStr);
    when(rangerRazResult.getOperResult().getAdditionalInfo()).thenReturn(addInfo);
  }

  private RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials prepareCredentialsMock(
          SignerCacheImpl signerCache) {
    RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials credentials =
            mock(RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials.class);

    when(credentials.getSignerCache()).thenReturn(signerCache);
    when(credentials.getAWSSecretKey()).thenReturn("random_access_key");
    RazDelegationTokenBinding razDelegationTokenBinding = mock(RazDelegationTokenBinding.class);
    when(credentials.getRazDelegationTokenBinding()).thenReturn(razDelegationTokenBinding);
    when(credentials.getConfig()).thenReturn(new Configuration());
    return credentials;
  }

  private RangerRazClient prepareRazClientMocks(RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials credentials) {
    RangerRazClient rangerRazClient = mock(RangerRazClient.class);
    when(credentials.getRazClient()).thenReturn(rangerRazClient);
    return rangerRazClient;
  }

  private void prepareUGIMock(RazAnonymousAWSCredentialsProvider.RazAnonymousAWSCredentials credentials,
                              String userName) {
    UserGroupInformation owner = mock(UserGroupInformation.class);
    when(credentials.getOwner()).thenReturn(owner);
    when(owner.getUserName()).thenReturn(userName);
  }
}
