/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.hook.s3;

import org.apache.ranger.raz.hook.s3.cache.SignerCacheImpl;
import org.apache.ranger.raz.s3.lib.RazS3Constants;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsOperationType;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsRequestCacheKey;
import org.junit.Before;
import org.junit.Test;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.*;

public class TestSignerCacheImpl {

    private AwsRequestCacheKey keyPath1Owner1;
    private AwsRequestCacheKey keyPath1Owner2;
    private AwsRequestCacheKey keyPath2Owner1;

    private final Map<String, String> valKeyPath1Owner1 = new HashMap<>();
    private final Map<String, String> valKeyPath1Owner2 = new HashMap<>();
    private final Map<String, String> valKeyPath2Owner1 = new HashMap<>();

    @Before
    public void setup() throws Exception {
        keyPath1Owner1 = generateKeyWithPathAndOwner("path1", "owner1");
        keyPath1Owner2 = generateKeyWithPathAndOwner("path1", "owner2");
        keyPath2Owner1 = generateKeyWithPathAndOwner("path2", "owner1");

        valKeyPath1Owner1.put("signature", "signValPath1Owner1");
        valKeyPath1Owner2.put("signature", "signValPath1Owner2");
        valKeyPath2Owner1.put("signature", "signValPath2Owner1");
    }
    @Test
    public void testCacheOperations() throws Exception {
        SignerCacheImpl signerCacheImpl = new SignerCacheImpl();

        assertEquals("Initial cache size should be 0", 0, signerCacheImpl.size());

        signerCacheImpl.put(keyPath1Owner1, valKeyPath1Owner1);
        assertEquals("Cache size should be 1", 1, signerCacheImpl.size());
        assertTrue("Cache should contain key keyPath1Owner1 ",
                signerCacheImpl.get(keyPath1Owner1) != null);
        assertEquals("Cache should return right value ",
                valKeyPath1Owner1, signerCacheImpl.get(keyPath1Owner1));

        assertNull("Cache should return null for same path but different owner",
                signerCacheImpl.get(keyPath1Owner2));

        assertNull("Cache should return null for different path but same owner",
                signerCacheImpl.get(keyPath2Owner1));

        signerCacheImpl.put(keyPath1Owner2, valKeyPath1Owner2);

        assertEquals("Cache size should be 2", 2, signerCacheImpl.size());
        assertEquals("Cache should return right value ",
                valKeyPath1Owner2, signerCacheImpl.get(keyPath1Owner2));
        assertTrue("Cache shouldn't contain key keyPath2Owner1 ",
                signerCacheImpl.get(keyPath2Owner1) == null);


        signerCacheImpl.put(keyPath2Owner1, valKeyPath2Owner1);

        assertEquals("Cache size should be 3", 3, signerCacheImpl.size());
        assertEquals("Cache should return right value ",
                valKeyPath2Owner1, signerCacheImpl.get(keyPath2Owner1));

    }

    private AwsRequestCacheKey generateKeyWithPathAndOwner(String path,
                                                           String owner)
            throws URISyntaxException {
        return new AwsRequestCacheKey(
                AwsOperationType.READ_OBJECT,
                new URI("dummy"),
                new HashMap<String, List<String>>(),
                "bucket",
                new String[]{path},
                RazS3Constants.AccessType.READ,
                owner);
    }
}
