/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.intg;


import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.util.HashMap;
import java.util.Map;

import com.sun.jersey.core.spi.factory.ResponseBuilderImpl;

public class RangerRazException extends Exception {
    private RangerRazErrorCode errorCode;

    // Map of headers.
    private Map<String, String> rangerRazExceptionHeadersMap = new HashMap<>();

    /** ETag for WebApplicationException. */
    public static final String WEBAPPLICATIONEXCP_TAG =
        "webApplicationExcp";

    /** Custom ETag for WebApplicationException raised with headers. */
    public static final String WEBAPPLICATIONEXCP_WITH_HEADERS_TAG =
        "webApplicationExcp_with_headers";

    public RangerRazException() {
        this(RangerRazErrorCode.INTERNAL_ERROR);
    }

    public RangerRazException(String message) {
        super(message);

        this.errorCode = RangerRazErrorCode.INTERNAL_ERROR;
    }

    public RangerRazException(String message, Throwable cause) {
        super(message, cause);

        this.errorCode = RangerRazErrorCode.INTERNAL_ERROR;
    }

    public RangerRazException(Throwable cause) {
        super(cause);

        this.errorCode = RangerRazErrorCode.INTERNAL_ERROR;
    }

    public RangerRazException(RangerRazErrorCode errorCode, Object ... params) {
        super(errorCode.getFormattedErrorMessage(params));

        this.errorCode = errorCode;
    }

    public RangerRazException(RangerRazErrorCode errorCode, Throwable cause, Object ... params) {
        super(errorCode.getFormattedErrorMessage(params), cause);

        this.errorCode = errorCode;
    }

    /**
     * Constructor to convert an exception with a header map containing
     * params into a RangerRazException.
     * @param cause Exception raised.
     * @param rangerRazExceptionHeadersMap Map of headers containing exception
     *                                     params.
     */
    public RangerRazException(Throwable cause,
        Map<String, String> rangerRazExceptionHeadersMap) {
        super(cause);
        this.errorCode = RangerRazErrorCode.INTERNAL_ERROR;
        this.rangerRazExceptionHeadersMap = rangerRazExceptionHeadersMap;
    }

    /**
     * Converts a RangerRazException into a WebApplicationException with a
     * response in JSON to be parsed in client.
     *
     * @return an instance of WebApplicationException.
     */
    public WebApplicationException toWebApplicationException() {
        Response response = Response.status(errorCode.getHttpCode())
            .entity(getMessage())
            .tag(WEBAPPLICATIONEXCP_TAG)
            .type(MediaType.APPLICATION_JSON_TYPE).build();

        return new WebApplicationException(response);
    }

    /**
     * Converts a RangerRazException into a WebApplicationException with a
     * response containing several headers in JSON to be parsed in client and
     * translated further in FileSystem.
     *
     * @return an instance of WebApplicationException.
     */
    public WebApplicationException toWebApplicationExceptionWithHeaders() {
        Response.ResponseBuilder responseBuilder = new ResponseBuilderImpl();

        // Build a Response in JSON with appropriate params.
        responseBuilder.status(errorCode.getHttpCode())
            .entity(getMessage())
            .tag(WEBAPPLICATIONEXCP_WITH_HEADERS_TAG)
            .type(MediaType.APPLICATION_JSON_TYPE);

        // Add headers to the WebApplicationException if map isn't null
        if (rangerRazExceptionHeadersMap != null) {
            for (Map.Entry<String, String> entry : rangerRazExceptionHeadersMap.entrySet()) {
                responseBuilder.header(entry.getKey(), entry.getValue());
            }
        }

        Response response = responseBuilder.build();
        return new WebApplicationException(response);
    }

    public RangerRazErrorCode getErrorCode() {
        return errorCode;
    }

    public void setErrorCode(RangerRazErrorCode errorCode) {
        this.errorCode = errorCode;
    }

    public Map<String, String> getRangerRazExceptionHeadersMap() {
        return rangerRazExceptionHeadersMap;
    }

    public void setRangerRazExceptionHeadersMap(
        Map<String, String> rangerRazExceptionHeadersMap) {
        this.rangerRazExceptionHeadersMap = rangerRazExceptionHeadersMap;
    }
}
