/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.intg.client;

import com.sun.jersey.api.client.ClientHandlerException;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.raz.intg.RangerRazErrorCode;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.executor.HttpClientExecutor;
import org.apache.ranger.raz.intg.client.executor.RestClientExecutorConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.KeyManager;
import javax.net.ssl.TrustManager;
import javax.net.ssl.SSLContext;
import java.io.InputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.File;
import java.io.FileInputStream;
import java.net.InetAddress;
import java.net.URL;
import java.net.UnknownHostException;
import java.net.URISyntaxException;
import java.net.ConnectException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.KeyManagementException;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.util.Random;
import java.util.Set;
import java.util.Map;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;

public class RangerRESTClient {
    private static final Logger LOG = LoggerFactory.getLogger(RangerRESTClient.class);

    public static enum HTTP_METHOD {
        GET, POST, PUT, DELETE
    };

    public static final String REST_EXPECTED_MIME_TYPE = "application/json";
    public static final String REST_MIME_TYPE_JSON     = "application/json";

    public static final String RANGER_PROP_POLICYMGR_URL                = "ranger.service.store.rest.url";
    public static final String RANGER_PROP_POLICYMGR_SSLCONFIG_FILENAME = "ranger.service.store.rest.ssl.config.file";
    public static final String RANGER_PROP_REST_CLIENT_CLASS            = "ranger.raz.client.executor.class";
    public static final String RANGER_DEFAULT_REST_CLIENT_CLASS         = "org.apache.ranger.raz.intg.client.executor.RestClientExecutorApacheHttp";
    public static final String RANGER_PROP_REST_CLIENT_MAX_RETRY        = "ranger.raz.client.max.retry";
    public static final String RANGER_PROP_JWT_FILENAME                 = "ranger.auth.jwt.file.name";
    public static final String RANGER_PROP_JWT_ENV_VAR_NAME             = "ranger.auth.jwt.env.var.name";
    public static final String DEFAULT_JWT_ENV_VAR_NAME                 = "RAZ_AUTH_JWT";
    public static final String RANGER_PROP_JWT_SERVER_COOKIE_NAME       = "ranger.auth.jwt.server.cookie.name";

    public static final String RANGER_PROP_REST_CLIENT_REQ_PARAMS_TO_MASK    = "ranger.raz.client.request.params.to.mask";
    public static final String RANGER_DEFAULT_REST_CLIENT_REQ_PARAMS_TO_MASK = "token,delegation";
    public static final String RANGER_PROP_JWT_TOKEN_RETRIEVER_CLASS         = "ranger.auth.jwt.retriever.class";
    public static final String RANGER_PROP_JWT_TOKEN_RETRIEVER_CLASS_DEFAULT = "org.apache.ranger.raz.intg.token.JwTokenRetrieverEnv";

    public static final String RANGER_POLICYMGR_CLIENT_KEY_FILE                  = "xasecure.policymgr.clientssl.keystore";
    public static final String RANGER_POLICYMGR_CLIENT_KEY_FILE_TYPE             = "xasecure.policymgr.clientssl.keystore.type";
    public static final String RANGER_POLICYMGR_CLIENT_KEY_FILE_CREDENTIAL       = "xasecure.policymgr.clientssl.keystore.credential.file";
    public static final String RANGER_POLICYMGR_TRUSTSTORE_FILE_PWD              = "xasecure.policymgr.clientssl.keystore.credential.file.pwd";
    public static final String RANGER_POLICYMGR_CLIENT_KEY_FILE_CREDENTIAL_ALIAS = "sslKeyStore";
    public static final String RANGER_POLICYMGR_CLIENT_KEY_FILE_TYPE_DEFAULT     = "jks";

    public static final String RANGER_POLICYMGR_TRUSTSTORE_FILE                  = "xasecure.policymgr.clientssl.truststore";
    public static final String RANGER_POLICYMGR_TRUSTSTORE_FILE_TYPE             = "xasecure.policymgr.clientssl.truststore.type";
    public static final String RANGER_POLICYMGR_TRUSTSTORE_FILE_CREDENTIAL       = "xasecure.policymgr.clientssl.truststore.credential.file";
    public static final String RANGER_POLICYMGR_TRUSTSTORE_FILE_CREDENTIAL_ALIAS = "sslTrustStore";
    public static final String RANGER_POLICYMGR_TRUSTSTORE_FILE_TYPE_DEFAULT     = "jks";

    private static final String HADOOP_SSL_CLIENT_CONF_FILE         = "hadoop.ssl.client.conf";
    private static final String HADOOP_SSL_CLIENT_CONF_FILE_DEFAULT = "ssl-client.xml";
    private static final String HADOOP_SSL_CLIENT_TRUSTSTORE_LOC    = "ssl.client.truststore.location";
    private static final String HADOOP_SSL_CLIENT_TRUSTSTORE_PWD    = "ssl.client.truststore.password";

    public static final String RANGER_SSL_KEYMANAGER_ALGO_TYPE   = KeyManagerFactory.getDefaultAlgorithm();
    public static final String RANGER_SSL_TRUSTMANAGER_ALGO_TYPE = TrustManagerFactory.getDefaultAlgorithm();
    public static final String RANGER_SSL_CONTEXT_ALGO_TYPE      = "TLS";

    private String  mUrl;
    private String  mSslConfigFileName;
    private String  mUsername;
    private String  mPassword;
    private boolean mIsSSL;

    private String mKeyStoreURL;
    private String mKeyStoreAlias;
    private String mKeyStoreFile;
    private String mKeyStoreType;
    private String mTrustStoreURL;
    private String mTrustStoreAlias;
    private String mTrustStoreFile;
    private String mTrustStoreFilePwd;
    private String mTrustStoreType;
    private String clientClassName;
    private String jwtEnvVarName;
    private String jwtFileName;
    private String jwtTokenRetrieverClass;
    private String jwtServerCookieName;

    private int    mRestClientConnTimeOutMs;
    private int    mRestClientReadTimeOutMs;
    private int    lastKnownActiveUrlIndex;
    private int    mMaxConnections;
    private int    mMaxConnectionsPerHost;
    private int    mMaxAttempts;
    private int    validateAfterInactivityMs;
    private int    poolRetryCount;

    private final List<String> configuredURLs;
    private final Set<String>  listOfRequestParamsToMask = new HashSet<String>();

    private volatile HttpClientExecutor client;

    public RangerRESTClient(String url, String sslConfigFileName, Configuration config, boolean useLocalRazService) throws IOException {
        mUrl               = url;
        mSslConfigFileName = sslConfigFileName;
        configuredURLs     = getURLs(mUrl, useLocalRazService);
        mMaxAttempts       = config.getInt(RANGER_PROP_REST_CLIENT_MAX_RETRY, 3); // Fail safe if configured negative value

        init(config);
    }

    public String getUsername() {
        return mUsername;
    }

    public String getPassword() {
        return mPassword;
    }

    public int getRestClientConnTimeOutMs() {
        return mRestClientConnTimeOutMs;
    }

    public void setRestClientConnTimeOutMs(int mRestClientConnTimeOutMs) {
        this.mRestClientConnTimeOutMs = mRestClientConnTimeOutMs;
    }

    public int getRestClientReadTimeOutMs() {
        return mRestClientReadTimeOutMs;
    }

    public void setRestClientReadTimeOutMs(int mRestClientReadTimeOutMs) {
        this.mRestClientReadTimeOutMs = mRestClientReadTimeOutMs;
    }

    public void setBasicAuthInfo(String username, String password) {
        mUsername = username;
        mPassword = password;
    }

    public HttpClientExecutor getClient() {
        // result saves on access time when client is built at the time of the call
        HttpClientExecutor result = client;
        if (result == null) {
            synchronized (this) {
                result = client;
                if (result == null) {
                    client = result = buildClient();
                }
            }
        }

        return result;
    }

    public void resetClient() {
        client = null;
    }

    public int getLastKnownActiveUrlIndex() {
        return lastKnownActiveUrlIndex;
    }

    public List<String> getConfiguredURLs() {
        return configuredURLs;
    }

    public boolean isSSL() {
        return mIsSSL;
    }

    public void setSSL(boolean mIsSSL) {
        this.mIsSSL = mIsSSL;
    }

    public int getmMaxConnections() {
        return mMaxConnections;
    }

    public void setmMaxConnections(int maxConnections) {
        this.mMaxConnections = maxConnections;
    }

    public int getMaxConnectionsPerHost() {
        return mMaxConnectionsPerHost;
    }

    public void setMaxConnectionsPerHost(int maxConnectionsPerHost) {
        this.mMaxConnectionsPerHost = maxConnectionsPerHost;
    }

    public <T> T executeHttpRequest(HTTP_METHOD method, String relativeUrl, Map<String, String> params, Object obj, Class<T> clazz) throws Exception {
        if (LOG.isTraceEnabled()) {
            Map<String, String> maskedParamSet = new HashMap<String, String>(params);

            for(String paramToMask : listOfRequestParamsToMask) {
                String paramOrgValue = maskedParamSet.get(paramToMask);
                maskedParamSet.put(paramToMask, RangerRazClient.maskParam(paramOrgValue));
            }

            LOG.trace("==> RangerRESTClient.executeHttpRequest({}, {}, {}, {}, {})", method, relativeUrl, maskedParamSet, obj, clazz);
        }

        T   finalResponse = null;
        int startIndex    = this.lastKnownActiveUrlIndex;
        int noOfRazHost   = configuredURLs.size();

        for (int urlIndex = 0; urlIndex < noOfRazHost; urlIndex++) {
            for (int attempt = 0; attempt < mMaxAttempts; attempt++) {
                boolean isLastAttempt = attempt == (mMaxAttempts - 1) && urlIndex == (noOfRazHost - 1);
                int     currentIndex  = (startIndex + urlIndex) % noOfRazHost;
                try {
                    switch (method) {
                        case GET:
                            finalResponse = (T) getClient().getAndParse(configuredURLs.get(currentIndex) + relativeUrl, params, clazz);
                            break;

                        case POST:
                            finalResponse = (T) getClient().postAndParse(configuredURLs.get(currentIndex) + relativeUrl, params, obj, clazz);
                            break;

                        case PUT:
                            finalResponse = (T) getClient().putAndParse(configuredURLs.get(currentIndex) + relativeUrl, params, obj, clazz);
                            break;

                        case DELETE:
                            finalResponse = (T) getClient().deleteAndParse(configuredURLs.get(currentIndex) + relativeUrl, params, clazz);
                            break;
                        default:
                            break;
                    }

                    if (finalResponse != null) {
                        setLastKnownActiveUrlIndex(currentIndex);
                        break;
                    }
                } catch (ClientHandlerException ex) {
                    RangerRazClientLogger.warn(LOG, "Failed to communicate with Raz, URL: {} {}", configuredURLs.get(currentIndex),"ClientHandlerException...trying again to connect" );
                    processException(isLastAttempt, ex);
                } catch (RangerRazException ex) {
                    processRazException(isLastAttempt, ex, currentIndex);
                } catch (UnknownHostException ue) {
                    RangerRazClientLogger.warn(LOG, "Failed to communicate with Raz, URL: {} {}", configuredURLs.get(currentIndex),"UnknownHostException...trying next raz host");
                    processException(isLastAttempt, ue);
                    break;
                } catch (URISyntaxException urise) {
                    RangerRazClientLogger.warn(LOG, "Failed to communicate with Raz, URL: {} {}", configuredURLs.get(currentIndex), "URISyntaxException...trying next raz host");
                    processException(isLastAttempt, urise);
                    break;
                } catch (ConnectException ce) {
                    RangerRazClientLogger.warn(LOG, "Failed to communicate with Raz, URL: {} {}", configuredURLs.get(currentIndex), "ConnectException...trying next raz host");
                    processException(isLastAttempt, ce);
                    break;
                } catch (IOException ex) {//Most probably network issue, hence do not break keep trying
                    RangerRazClientLogger.warn(LOG, "Failed to communicate with Raz, URL: {} {}", configuredURLs.get(currentIndex), "Exception...trying next raz host");
                    processExceptionFinal(isLastAttempt, ex);
                } catch (Exception e) {
                    RangerRazClientLogger.warn(LOG, "Failed to communicate with Raz, URL: {} {}", configuredURLs.get(currentIndex), "Exception...trying next raz host");
                    processExceptionFinal(isLastAttempt, e);
                    break;
                }
            }
            if (finalResponse != null) {
                //got response from a raz host, so no need to try other hosts.
                break;
            }
        }
        if (LOG.isTraceEnabled()) {
            LOG.trace("<== RangerRESTClient.executeHttpRequest()"); //Don't print response, some might have sensitive info
        }
        return finalResponse;
    }

    //protected
    protected SSLContext getSSLContext(KeyManager[] kmList, TrustManager[] tmList) {
        SSLContext ret = null;

        try {
            if (tmList == null) {
                // if TrustManager is not, set if from default Hadoop Conf File.
                tmList = setSSLContextFromHadoopConfFile();
            }
            SSLContext sslContext = SSLContext.getInstance(RANGER_SSL_CONTEXT_ALGO_TYPE);
            sslContext.init(kmList, tmList, new SecureRandom());
            ret = sslContext;
        } catch (NoSuchAlgorithmException e) {
            RangerRazClientLogger.error(LOG, "SSL algorithm is not available in the environment", e);
            throw new IllegalStateException("SSL algorithm is not available in the environment: " + e.getMessage(), e);
        } catch (KeyManagementException e) {
            RangerRazClientLogger.error(LOG, "Unable to initials the SSLContext", e);
            throw new IllegalStateException("Unable to initials the SSLContex: " + e.getMessage(), e);
        } catch (IOException ioe) {
            RangerRazClientLogger.error(LOG, "Unable to fetch the SSLContext from default Hadoop SSL conf: {}", HADOOP_SSL_CLIENT_CONF_FILE, ioe);
            throw new IllegalStateException("Unable to fetch the SSLContext from default Hadoop SSL conf: " + ioe.getMessage(), ioe);
        }

        return ret;
    }

    protected void processException(boolean isLastAttempt, ClientHandlerException e) throws Exception {
        if (isLastAttempt) {
            LOG.error("Failed to communicate with all Raz URLs: [{}]", configuredURLs);
            LOG.error("Verify that, the URLs are correct and corresponding services are running & accessible.");
            throw e;
        }
    }

    protected void processException(boolean isLastAttempt, UnknownHostException e) throws Exception {
        if (isLastAttempt) {
            LOG.error("Failed to communicate with all Raz URLs: [{}]", configuredURLs);
            LOG.error("Verify that, the URLs are correct and corresponding services are running & accessible.");
            throw e;
        }
    }

    protected void processException(boolean isLastAttempt, URISyntaxException e) throws Exception {
        if (isLastAttempt) {
            LOG.error("Failed to communicate with all Raz URLs: [{}]", configuredURLs);
            LOG.error("Verify that, the URLs are correct and corresponding services are running & accessible.");
            throw e;
        }
    }

    protected void processException(boolean isLastAttempt, ConnectException e) throws Exception {
        if (isLastAttempt) {
            LOG.error("Failed to communicate with all Raz URLs: [{}]", configuredURLs);
            LOG.error("Verify that, the URLs are correct and corresponding services are running & accessible.");
            throw e;
        }
    }

    protected void processExceptionFinal(boolean isLastAttempt, Exception e) throws Exception {
        if (isLastAttempt) {
            LOG.error("Failed to communicate with all Raz URLs: [{}]", configuredURLs);
            LOG.error("Verify that, the URLs are correct and corresponding services are running & accessible.");
            throw e;
        }
    }

    protected void processRazException(boolean isLastAttempt, RangerRazException e, int currentIndex) throws Exception {
        if(e.getErrorCode().equals(RangerRazErrorCode.RAZ_CLIENT_REQUEST_FAILED)) {
            RangerRazClientLogger.warn(LOG, "Failed to communicate with Raz, URL: {}", configuredURLs.get(currentIndex));
            if (isLastAttempt) {
                LOG.error("Failed to communicate with all Raz URLs: [{}]", configuredURLs);
                LOG.error("Verify that, the URLs are correct and corresponding services are running & accessible.");
                throw e;
            }
        } else {
            if(e.getErrorCode().equals(RangerRazErrorCode.RAZ_CLIENT_ACCESS_DENIED)) {
                IOException ioe = new IOException("Permission denied", e);
                throw ioe;
            }
            throw e; //No need to retry
        }
    }

    protected void setLastKnownActiveUrlIndex(int lastKnownActiveUrlIndex) {
        this.lastKnownActiveUrlIndex = lastKnownActiveUrlIndex;
    }

    protected void setKeyStoreType(String mKeyStoreType) {
        this.mKeyStoreType = mKeyStoreType;
    }

    protected void setTrustStoreType(String mTrustStoreType) {
        this.mTrustStoreType = mTrustStoreType;
    }

    public int getValidateAfterInactivityMs() {
        return validateAfterInactivityMs;
    }

    public void setValidateAfterInactivityMs(int validateAfterInactivityMs) {
        this.validateAfterInactivityMs = validateAfterInactivityMs;
    }

    public int getPoolRetryCount() {
        return poolRetryCount;
    }

    public void setPoolRetryCount(int poolRetryCount) {
        this.poolRetryCount = poolRetryCount;
    }

    //private
    @SuppressWarnings("unchecked")
    private HttpClientExecutor buildClient() {
        RestClientExecutorConfig config = new RestClientExecutorConfig();
        config.setUsername(mUsername);
        config.setPassword(mPassword);
        config.setRestClientConnTimeOutMs(mRestClientConnTimeOutMs);
        config.setRestClientReadTimeOutMs(mRestClientReadTimeOutMs);
        config.setMaxConnections(mMaxConnections);
        config.setMaxConnectionsPerHost(mMaxConnectionsPerHost);
        config.setSsl(mIsSSL);
        config.setValidateAfterInactivityMs(validateAfterInactivityMs);
        config.setPoolRetryCount(poolRetryCount);

        if (mIsSSL) {
            KeyManager[]   kmList     = getKeyManagers();
            TrustManager[] tmList     = getTrustManagers();
            SSLContext     sslContext = getSSLContext(kmList, tmList);

            config.setSslContext(sslContext);
        }

        // JWT properties
        config.setJwtEnvVarName(jwtEnvVarName);
        config.setJwtFileName(jwtFileName);
        config.setJwtTokenRetrieverClass(jwtTokenRetrieverClass);
        config.setJwtServerCookieName(jwtServerCookieName);

        HttpClientExecutor clientExecutor = null;

        Class<HttpClientExecutor> cls;
        try {
            cls            = (Class<HttpClientExecutor>) Class.forName(clientClassName);
            clientExecutor = cls.getDeclaredConstructor(RestClientExecutorConfig.class).newInstance(config);
        } catch (Exception e) {
            LOG.error("Exception occured while creating instance of executor class[{}]. Exception={}", clientClassName, e);
        }

        return clientExecutor;
    }

    private void init(Configuration config) {
        mIsSSL = isSsl(mUrl);

        if (mIsSSL) {

            InputStream in                = null;
            InputStream inHadoopSsl       = null;
            String      hadoopSslConfFile = config.get(HADOOP_SSL_CLIENT_CONF_FILE, HADOOP_SSL_CLIENT_CONF_FILE_DEFAULT);

            try {
                in = getFileInputStream(mSslConfigFileName);

                if (in != null) {
                    config.addResource(in);
                }

                mKeyStoreURL   = config.get(RANGER_POLICYMGR_CLIENT_KEY_FILE_CREDENTIAL);
                mKeyStoreAlias = RANGER_POLICYMGR_CLIENT_KEY_FILE_CREDENTIAL_ALIAS;
                mKeyStoreType  = config.get(RANGER_POLICYMGR_CLIENT_KEY_FILE_TYPE, RANGER_POLICYMGR_CLIENT_KEY_FILE_TYPE_DEFAULT);
                mKeyStoreFile  = config.get(RANGER_POLICYMGR_CLIENT_KEY_FILE);

                mTrustStoreURL   = config.get(RANGER_POLICYMGR_TRUSTSTORE_FILE_CREDENTIAL);
                mTrustStoreAlias = RANGER_POLICYMGR_TRUSTSTORE_FILE_CREDENTIAL_ALIAS;
                mTrustStoreType  = config.get(RANGER_POLICYMGR_TRUSTSTORE_FILE_TYPE, RANGER_POLICYMGR_TRUSTSTORE_FILE_TYPE_DEFAULT);
                mTrustStoreFile  = config.get(RANGER_POLICYMGR_TRUSTSTORE_FILE);

                // Read SSL configurations from hadoop
                if (StringUtils.isNotEmpty(hadoopSslConfFile)) {
                    inHadoopSsl = getFileInputStream(hadoopSslConfFile);
                    Configuration hadoopSslConfiguration = new Configuration();
                    if (inHadoopSsl != null) {
                        hadoopSslConfiguration.addResource(inHadoopSsl);
                    }
                    if (StringUtils.isEmpty(mTrustStoreFile)) {
                        mTrustStoreFile = hadoopSslConfiguration.get(HADOOP_SSL_CLIENT_TRUSTSTORE_LOC);
                    }
                    mTrustStoreFilePwd = hadoopSslConfiguration.get(HADOOP_SSL_CLIENT_TRUSTSTORE_PWD);
                }
            } catch (IOException ioe) {
                RangerRazClientLogger.error(LOG, "Unable to load SSL Config FileName: [{}]", mSslConfigFileName, ioe);
            } finally {
                close(in, mSslConfigFileName);
                close(inHadoopSsl, hadoopSslConfFile);
            }
        }
        clientClassName = config.get(RANGER_PROP_REST_CLIENT_CLASS, RANGER_DEFAULT_REST_CLIENT_CLASS);


        String requestParamsToMask = config.get(RANGER_PROP_REST_CLIENT_REQ_PARAMS_TO_MASK, RANGER_DEFAULT_REST_CLIENT_REQ_PARAMS_TO_MASK);

        for (String requestParamToMask : requestParamsToMask.split(",")) {
            listOfRequestParamsToMask.add(requestParamToMask);
        }

        // JWT configs
        jwtEnvVarName          = config.get(RANGER_PROP_JWT_ENV_VAR_NAME, DEFAULT_JWT_ENV_VAR_NAME);
        jwtFileName            = config.get(RANGER_PROP_JWT_FILENAME);
        jwtTokenRetrieverClass = config.get(RANGER_PROP_JWT_TOKEN_RETRIEVER_CLASS, RANGER_PROP_JWT_TOKEN_RETRIEVER_CLASS_DEFAULT);
        jwtServerCookieName    = config.get(RANGER_PROP_JWT_SERVER_COOKIE_NAME);
    }

    private boolean isSsl(String url) {
        return !StringUtils.isEmpty(url) && url.toLowerCase().startsWith("https");
    }

    private KeyManager[] getKeyManagers() {
        KeyManager[] kmList = null;

        String keyStoreFilepwd = getCredential(mKeyStoreURL, mKeyStoreAlias);

        kmList = getKeyManagers(mKeyStoreFile, keyStoreFilepwd);
        return kmList;
    }

    private KeyManager[] getKeyManagers(String keyStoreFile, String keyStoreFilePwd) {
        KeyManager[] kmList = null;

        if (StringUtils.isNotEmpty(keyStoreFile) && StringUtils.isNotEmpty(keyStoreFilePwd)) {
            InputStream in = null;

            try {
                in = getFileInputStream(keyStoreFile);

                if (in != null) {
                    KeyStore keyStore = KeyStore.getInstance(mKeyStoreType);

                    keyStore.load(in, keyStoreFilePwd.toCharArray());

                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(RANGER_SSL_KEYMANAGER_ALGO_TYPE);

                    keyManagerFactory.init(keyStore, keyStoreFilePwd.toCharArray());

                    kmList = keyManagerFactory.getKeyManagers();
                } else {
                    RangerRazClientLogger.error(LOG, "Unable to obtain keystore from file [{}]", keyStoreFile);
                    throw new IllegalStateException("Unable to find keystore file :" + keyStoreFile);
                }
            } catch (KeyStoreException e) {
                RangerRazClientLogger.error(LOG, "Unable to obtain from KeyStore: {}", e.getMessage(), e);
                throw new IllegalStateException("Unable to init keystore:" + e.getMessage(), e);
            } catch (NoSuchAlgorithmException e) {
                RangerRazClientLogger.error(LOG, "SSL algorithm is NOT available in the environment", e);
                throw new IllegalStateException("SSL algorithm is NOT available in the environment :" + e.getMessage(), e);
            } catch (CertificateException e) {
                RangerRazClientLogger.error(LOG, "Unable to obtain the requested certification ", e);
                throw new IllegalStateException("Unable to obtain the requested certification :" + e.getMessage(), e);
            } catch (FileNotFoundException e) {
                RangerRazClientLogger.error(LOG, "Unable to find the necessary SSL Keystore Files", e);
                throw new IllegalStateException("Unable to find keystore file :" + keyStoreFile + ", error :" + e.getMessage(), e);
            } catch (IOException e) {
                RangerRazClientLogger.error(LOG, "Unable to read the necessary SSL Keystore Files", e);
                throw new IllegalStateException("Unable to read keystore file :" + keyStoreFile + ", error :" + e.getMessage(), e);
            } catch (UnrecoverableKeyException e) {
                RangerRazClientLogger.error(LOG, "Unable to recover the key from keystore", e);
                throw new IllegalStateException("Unable to recover the key from keystore :" + keyStoreFile + ", error :" + e.getMessage(), e);
            } finally {
                close(in, keyStoreFile);
            }
        }

        return kmList;
    }

    private TrustManager[] getTrustManagers() {
        TrustManager[] tmList = null;

        String trustStoreFilepwd = getCredential(mTrustStoreURL, mTrustStoreAlias);
        if (StringUtils.isEmpty(trustStoreFilepwd)) {
            trustStoreFilepwd = this.mTrustStoreFilePwd;
        }

        tmList = getTrustManagers(mTrustStoreFile, trustStoreFilepwd);
        return tmList;
    }

    private TrustManager[] getTrustManagers(String trustStoreFile, String trustStoreFilepwd) {
        TrustManager[] tmList = null;

        if (StringUtils.isNotEmpty(trustStoreFile) && StringUtils.isNotEmpty(trustStoreFilepwd)) {
            InputStream in = null;

            try {
                in = getFileInputStream(trustStoreFile);

                if (in != null) {
                    KeyStore trustStore = KeyStore.getInstance(mTrustStoreType);

                    trustStore.load(in, trustStoreFilepwd.toCharArray());

                    TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(RANGER_SSL_TRUSTMANAGER_ALGO_TYPE);

                    trustManagerFactory.init(trustStore);

                    tmList = trustManagerFactory.getTrustManagers();
                } else {
                    RangerRazClientLogger.error(LOG, "Unable to obtain truststore from file [{}]", trustStoreFile);
                    throw new IllegalStateException("Unable to find truststore file :" + trustStoreFile);
                }
            } catch (KeyStoreException e) {
                RangerRazClientLogger.error(LOG, "Unable to obtain from KeyStore", e);
                throw new IllegalStateException("Unable to init keystore:" + e.getMessage(), e);
            } catch (NoSuchAlgorithmException e) {
                RangerRazClientLogger.error(LOG, "SSL algorithm is NOT available in the environment: {}", e.getMessage(), e);
                throw new IllegalStateException("SSL algorithm is NOT available in the environment :" + e.getMessage(), e);
            } catch (CertificateException e) {
                RangerRazClientLogger.error(LOG, "Unable to obtain the requested certification: {}", e.getMessage(), e);
                throw new IllegalStateException("Unable to obtain the requested certification :" + e.getMessage(), e);
            } catch (FileNotFoundException e) {
                RangerRazClientLogger.error(LOG, "Unable to find the necessary SSL TrustStore File: {}", trustStoreFile, e);
                throw new IllegalStateException("Unable to find trust store file :" + trustStoreFile + ", error :" + e.getMessage(), e);
            } catch (IOException e) {
                RangerRazClientLogger.error(LOG, "Unable to read the necessary SSL TrustStore Files: {}", trustStoreFile, e);
                throw new IllegalStateException("Unable to read the trust store file :" + trustStoreFile + ", error :" + e.getMessage(), e);
            } finally {
                close(in, trustStoreFile);
            }
        }

        return tmList;
    }

    private TrustManager[] setSSLContextFromHadoopConfFile() throws IOException {
        TrustManager[] ret               = null;
        Configuration  config            = new Configuration();
        String         hadoopSslConfFile = config.get(HADOOP_SSL_CLIENT_CONF_FILE, HADOOP_SSL_CLIENT_CONF_FILE_DEFAULT);

        URL fUrl = getFileLocation(hadoopSslConfFile, config);

        if (fUrl != null) {
            RangerRazClientLogger.info(LOG, "addResourceIfReadable({}): resource file is {}", hadoopSslConfFile, fUrl);

            config.addResource(fUrl);
            mTrustStoreFile    = config.get(HADOOP_SSL_CLIENT_TRUSTSTORE_LOC);
            mTrustStoreFilePwd = config.get(HADOOP_SSL_CLIENT_TRUSTSTORE_PWD);
            ret                = getTrustManagers(mTrustStoreFile, mTrustStoreFilePwd);
        } else {
            RangerRazClientLogger.error(LOG, "addResourceIfReadable({}): couldn't find resource file location", hadoopSslConfFile);
        }

        return ret;
    }

    private URL getFileLocation(String fileName, Configuration config) {
        URL lurl = config.getClassLoader().getResource(fileName);

        if (lurl == null) {
            lurl = config.getClassLoader().getResource("/" + fileName);
        }

        return lurl;
    }

    private String getCredential(String url, String alias) {
        return RangerCredentialProvider.getInstance().getCredentialString(url, alias);
    }

    private InputStream getFileInputStream(String fileName) throws IOException {
        InputStream in = null;

        if (StringUtils.isNotEmpty(fileName)) {
            File f = new File(fileName);

            if (f.exists()) {
                in = new FileInputStream(f);
            } else {
                in = ClassLoader.getSystemResourceAsStream(fileName);
            }
        }

        return in;
    }

    private void close(InputStream str, String filename) {
        if (str != null) {
            try {
                str.close();
            } catch (IOException excp) {
                RangerRazClientLogger.error(LOG, "Error while closing file: [{}]", filename, excp);
            }
        }
    }

    private List<String> getURLs(String configURLs, boolean useLocalRazService) throws UnknownHostException {
        List<String> configuredURLs = Collections.emptyList();

        if (configURLs != null) {
            String[] urls  = configURLs.trim().split(",");
            configuredURLs = new ArrayList<String>(urls.length + 2); // configureUrls + 1 localUrl + 1 buffer
            for (String strUrl : urls) {
                strUrl = cleanRestUrl(strUrl);
                if (StringUtils.isNotBlank(strUrl)) {
                    configuredURLs.add(strUrl);
                }
            }

            if (useLocalRazService && (!configuredURLs.isEmpty())) {
                String [] urlParts = configuredURLs.get(0).split(":");
                configuredURLs.add(0, cleanRestUrl(urlParts[0] + "://" + InetAddress.getLocalHost().getHostName() + ":" + urlParts[2])); // Add local RAZ-URL at 0th place
                setLastKnownActiveUrlIndex(0); // Prioritize local RAZ-URL
            } else {
                setLastKnownActiveUrlIndex((new Random()).nextInt(configuredURLs.size()));
            }
        }

        return configuredURLs;
    }

    private static String cleanRestUrl(String restUrl) {
        String ret = restUrl;
        if (!StringUtils.isEmpty(StringUtils.trimToEmpty(ret))) {
            ret = ret.trim();
            if (ret.endsWith(RangerRazClient.HTTP_URL_SEPARATOR)) {
                ret = ret.substring(0, ret.length() - 1).trim();
            }
        }

        return ret;
    }
}