/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.intg.client;

import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_INVALID_RESPONSE;
import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_NULL_RESPONSE;
import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_UNAUTHORIZED_ACCESS;

import java.io.IOException;
import java.net.URI;
import java.security.PrivilegedExceptionAction;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.SecurityUtil;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.UserGroupInformation.AuthenticationMethod;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.ranger.raz.intg.RangerRazErrorCode;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRESTClient.HTTP_METHOD;
import org.apache.ranger.raz.model.RangerRazMultiOperationRequest;
import org.apache.ranger.raz.model.RangerRazMultiOperationResult;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.raz.model.RangerRazResultBase.ResourceAccessResult;
import org.apache.ranger.raz.model.RangerRazIdentityMappings;
import org.apache.ranger.raz.model.RangerRazIdentityMappingsUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

public class RangerRazClient {
    private static final Logger LOG = LoggerFactory.getLogger(RangerRazClient.class);

    public static final String HTTP_URL_SEPARATOR                 = "/";
    public static final String RAZ_CLIENT_PREFIX_KEY              = "ranger.raz.client.prefix";
    public static final String RAZ_CLIENT_PREFIX_DEFAULT          = "ranger.raz.client.";
    public static final String RAZ_SSL_FILE_LOC                   = "rest.ssl.config.file";
    public static final String RAZ_CLIENT_CONN_TIMEOUT_MS         = "rest.client.connection.timeoutMs";
    public static final String RAZ_CLIENT_READ_TIMEOUT_MS         = "rest.client.read.timeoutMs";
    public static final String RAZ_REST_HOST_URL                  = "rest.host.url";
    public static final String RAZ_CLIENT_LOG_QUIET_MODE          = "log.quiet.mode";
    public static final String MASKED_TOKEN_STR                   = "********";
    public static final String RAZ_REST_OID_API                   = "rest.oid.map.url";
    public static final String RAZ_CLIENT_SKIP_MASK_FOR_FIRST_KEY = "client.skip.mask.for.first";
    public static final String RAZ_CLIENT_KEYS_TO_REDACT_PROP     = "keys.to.redact";
    public static Set<String>  RAZ_CLIENT_KEYS_TO_REDACT          = new HashSet<>(Arrays.asList("ADLS_DSAS", "S3_SIGN_REQUEST", "S3_SIGN_RESPONSE"));

    private static final String PARAM_OP                               = "op";
    private static final String PARAM_RENEWER                          = "renewer";
    private static final String PARAM_TOKEN                            = "token";
    private static final String PARAM_DELEGATION                       = "delegation";
    private static final String PARAM_DOAS                             = "doAs";
    private static final String PARAM_DT_SERVICENAME                   = "service";
    private static final String OP_GETDELEGATIONTOKEN                  = "GETDELEGATIONTOKEN";
    private static final String OP_RENEWDELEGATIONTOKEN                = "RENEWDELEGATIONTOKEN";
    private static final String OP_CANCELDELEGATIONTOKEN               = "CANCELDELEGATIONTOKEN";
    private static final String RAZ_API_SINGLE_RESOURCE                = "/api/authz/%s/access";
    private static final String RAZ_API_MULTI_RESOURCE                 = "/api/authz/%s/accesses";
    private static final String RAZ_API_DOWNLOAD_OID_MAP               = "/api/authz/identity-mappings";
    private static final String DEFAULT_DT_SERVICE_NAME                = "RANGER_RAZ";
    private static final int    DEFAULT_RAZ_SERVER_PORT                = 6181;
    private static final Object DUMMY_POST_PARAM                       = new Object();
    private static Long         latestRazIdMappingVersion              = Long.valueOf(-1);
    private static int          RAZ_CLIENT_SKIP_MASK_FOR_FIRST_DEFAULT = 5;


    private final ObjectMapper         mapper = new ObjectMapper();
    private final RangerRESTClient     restClient;
    private final UserGroupInformation ugi;
    private final String               dtServiceName;

    private RangerRazIdentityMappingsUtil razIdMappUtil = null;

    private static final RangerRazException MISSING_KERBEROS_GET_DT    = new RangerRazException(RangerRazErrorCode.RAZ_CLIENT_UNAUTHORIZED_ACCESS,
    "Kerberos Authentication is Required, to get RAZ delegation token");
    private static final RangerRazException MISSING_KERBEROS_RENEW_DT  = new RangerRazException(RangerRazErrorCode.RAZ_CLIENT_UNAUTHORIZED_ACCESS,
    "Kerberos Authentication is Required, to renew RAZ delegaton token");
    private static final RangerRazException MISSING_KERBEROS_CANCEL_DT = new RangerRazException(RangerRazErrorCode.RAZ_CLIENT_UNAUTHORIZED_ACCESS,
    "Kerberos Authentication is Required, to cancel RAZ delegation token");

    // private constructor
    private RangerRazClient(Configuration configuration, UserGroupInformation ugi) throws IOException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazClient()");
        }

        // Get all required parameters
        String confPrefix                = configuration.get(RAZ_CLIENT_PREFIX_KEY, RAZ_CLIENT_PREFIX_DEFAULT);
        String sslConfigFileName         = configuration.get(confPrefix + RAZ_SSL_FILE_LOC);
        int    restClientConnTimeOutMs   = configuration.getInt(confPrefix + RAZ_CLIENT_CONN_TIMEOUT_MS, 120 * 1000);
        int    restClientReadTimeOutMs   = configuration.getInt(confPrefix + RAZ_CLIENT_READ_TIMEOUT_MS, 30 * 1000);
        int    maxConnections            = configuration.getInt(confPrefix + "rest.client.connection.pool.max", 10);
        int    maxConnectionsPerHost     = configuration.getInt(confPrefix + "rest.client.connection.pool.max-per-host", 10);
        int    validateAfterInactivityMs = configuration.getInt(confPrefix + "rest.client.connection.pool.validate-after-inactivity", 2000);
        int    poolRetryCount            = configuration.getInt(confPrefix + "rest.client.connection.pool.retry-count", 3);
        String razURL                    = configuration.get(confPrefix + RAZ_REST_HOST_URL);
        String initialzationError        = null;

        RangerRazClientLogger.setIsQuietMode(configuration.getBoolean(confPrefix + RAZ_CLIENT_LOG_QUIET_MODE, true));

        this.ugi = ugi;

        // validate parameters; throw runtime error if missing.
        if (ugi == null) {
            String errorMsg = "no user is logged in";

            RangerRazClientLogger.error(LOG, "RangerRazClient(): {}", errorMsg);

            initialzationError = (initialzationError == null ? errorMsg : (initialzationError + ", " + errorMsg));
        }

        if (StringUtils.isBlank(razURL)) {
            String errorMsg = "configuration '" + confPrefix + RAZ_REST_HOST_URL + "' is not set";

            RangerRazClientLogger.error(LOG, "RangerRazClient(): {}", errorMsg);

            initialzationError = (initialzationError == null ? errorMsg : (initialzationError + ", " + errorMsg));
        }

        if (initialzationError != null) {
            throw new IOException("RangerRazClient initialization failed: " + initialzationError);
        }

        boolean useLocalRazService = configuration.getBoolean(confPrefix + "client.use.local.raz", false);
        RAZ_CLIENT_SKIP_MASK_FOR_FIRST_DEFAULT = configuration.getInt(confPrefix + RAZ_CLIENT_SKIP_MASK_FOR_FIRST_KEY, RAZ_CLIENT_SKIP_MASK_FOR_FIRST_DEFAULT);

        String redactKeys = configuration.get(confPrefix + RAZ_CLIENT_KEYS_TO_REDACT_PROP);
        if (StringUtils.isNotEmpty(redactKeys)) {
            String[] keys = redactKeys.split(",");
            if (ArrayUtils.isNotEmpty(keys)) {
                RAZ_CLIENT_KEYS_TO_REDACT = new HashSet<>(Arrays.asList(keys));
            }
        }

        // Rest client setup
        restClient = new RangerRESTClient(razURL, sslConfigFileName, configuration, useLocalRazService);

        restClient.setRestClientConnTimeOutMs(restClientConnTimeOutMs);
        restClient.setRestClientReadTimeOutMs(restClientReadTimeOutMs);
        restClient.setmMaxConnections(maxConnections);
        restClient.setMaxConnectionsPerHost(maxConnectionsPerHost);
        restClient.setValidateAfterInactivityMs(validateAfterInactivityMs);
        restClient.setPoolRetryCount(poolRetryCount);

        dtServiceName = getDTServiceName(restClient.getConfiguredURLs());

        // Override Apache HTTP client log level to prevent logging sensitive information.
        if (configuration.getBoolean(confPrefix + "override.apache.http.log.level", true)) {
            LogManager.getLogger("org.apache.http").setLevel(Level.toLevel(configuration.get(confPrefix + "log.level.org.apache.http"), Level.INFO));
        }

        if (LOG.isDebugEnabled()) {
            String              propPrefix = configuration.get(RAZ_CLIENT_PREFIX_KEY, RAZ_CLIENT_PREFIX_DEFAULT);
            Map<String, String> propMap    = configuration.getPropsWithPrefix(propPrefix);

            for (Map.Entry<String, String> prop : propMap.entrySet()) {
                LOG.debug("RangerRazClient.init(): config: {}={}", prop.getKey(), prop.getValue());
            }

            UserGroupInformation.logAllUserInfo(LOG, ugi);

            LOG.debug("RangerRazClient(): dtServiceName={}", dtServiceName);

            LOG.debug("<== RangerRazClient()");
        }
    }

    /**
     * This method acts as static builder method, every time this method is invoked,
     * it will return new instance.
     * 
     * @param configuration
     * @param ugi
     * @return
     * @throws IOException
     */
    public static RangerRazClient getInstance(Configuration configuration, UserGroupInformation ugi) throws IOException {
        RangerRazClient instance = new RangerRazClient(configuration, ugi);

        return instance;
    }

    public String getDelegationTokenServiceName() {
        return dtServiceName;
    }

    /**
     * Get Ranger RAZ delegation token for currently logged in user user.
     * <p>
     * Sample CURL command: curl -ikv --negotiate -u : -X GET -H
     * "Accept:application/json" -H "Content-Type:application/json"
     * "http://RANGER_RAZ_SERVER?op=GETDELEGATIONTOKEN&renewer=OWNER"
     * </p>
     * 
     * @return {@link String string} representation of Ranger RAZ delegation token.
     * @throws Exception
     */
    public String getDelegationToken(String renewer) throws RangerRazException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazClient.getDelegationToken(renewer={})", renewer);
        }

        String ret                   = "";
        String maskedDelegationToken = null;

        if (isKerberosAuthenticated()) {
            Map<String, String>               queryParams  = createMap(PARAM_OP, OP_GETDELEGATIONTOKEN, PARAM_RENEWER, renewer, PARAM_DT_SERVICENAME, dtServiceName);
            UserGroupInformation              user         = getUgi(queryParams);
            PrivilegedExceptionAction<String> action       = getPrivilegedAction(HTTP_METHOD.GET, "", queryParams, DUMMY_POST_PARAM, String.class);
            String                            response     = executeAction(action, user);
            JsonNode                          node         = getJsonNode(response, "urlString");

            ret = node.asText();
            if (StringUtils.isNotBlank(ret)) {
                maskedDelegationToken = MASKED_TOKEN_STR;
            }

        } else {
            LOG.error("Kerberos Authentication is Required, to get RAZ delegation token.");
            throw MISSING_KERBEROS_GET_DT;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRazClient.getDelegationToken(renewer={}): delegationToken=[{}]", renewer, maskedDelegationToken);
        }

        return ret;
    }

    /**
     * Renew provided Ranger RAZ delegation token.
     * <p>
     * Sample CURL command: curl -ikv --negotiate -u : -X PUT -H
     * "Accept:application/json" -H "Content-Type:application/json"
     * "http://RANGER_RAZ_SERVER?op=RENEWDELEGATIONTOKEN&token=DELEGATION_TOKEN"
     * </p>
     * 
     * @return {@link Long updated} token expire time-stamp.
     * @throws Exception
     */
    public Long renewDelegationToken(String delegationToken) throws RangerRazException {
        String maskedDelegationToken = null;
        if (StringUtils.isNotBlank(delegationToken)) {
            maskedDelegationToken = MASKED_TOKEN_STR;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazClient.renewDelegationToken({})", maskedDelegationToken);
        }

        Long ret = -1l;
        if (isKerberosAuthenticated()) {
            Map<String, String>               queryParams = createMap(PARAM_OP, OP_RENEWDELEGATIONTOKEN, PARAM_TOKEN, delegationToken);
            UserGroupInformation              user        = getUgi(queryParams);
            PrivilegedExceptionAction<String> action      = getPrivilegedAction(HTTP_METHOD.PUT, "", queryParams, DUMMY_POST_PARAM, String.class);
            String                            response    = executeAction(action, user);
            JsonNode                          node        = getJsonNode(response, "long");

            ret = node.asLong();
        } else {
            LOG.error("Kerberos Authentication is Required, to renew RAZ delegaton token.");
            throw MISSING_KERBEROS_RENEW_DT;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRazClient.renewDelegationToken({}): DT nextExpiry={}", maskedDelegationToken, ret);
        }

        return ret;
    }

    /**
     * Cancel/ Delete provided Ranger RAZ delegation token.
     * <p>
     * Sample CURL command: curl -ikv --negotiate -u : -X PUT -H
     * "Accept:application/json" -H "Content-Type:application/json"
     * "http://RANGER_RAZ_SERVER?op=CANCELDELEGATIONTOKEN&token=DELEGATION_TOKEN"
     * </p>
     * 
     * @return
     */
    public Boolean cancelDelegationToken(String delegationToken) throws RangerRazException {
        String maskedDelegationToken = null;
        if (StringUtils.isNotBlank(delegationToken)) {
            maskedDelegationToken = MASKED_TOKEN_STR;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazClient.cancelDelegationToken({})", maskedDelegationToken);
        }

        Boolean ret = false;
        if (isKerberosAuthenticated()) {
            Map<String, String>               queryParams = createMap(PARAM_OP, OP_CANCELDELEGATIONTOKEN, PARAM_TOKEN, delegationToken);
            UserGroupInformation              user        = getUgi(queryParams);
            PrivilegedExceptionAction<String> action      = getPrivilegedAction(HTTP_METHOD.PUT, "", queryParams, DUMMY_POST_PARAM, String.class);
            executeAction(action, user);

            ret = true;
        } else {
            LOG.error("Kerberos Authentication is Required, to cancel RAZ delegation token.");
            throw MISSING_KERBEROS_CANCEL_DT;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRazClient.cancelDelegationToken(): ret={}", ret);
        }

        return ret;
    }

    public RangerRazMultiOperationResult checkPrivileges(RangerRazMultiOperationRequest requests, String delegationToken) throws RangerRazException {
        String maskedDelegationToken = null;
        if (StringUtils.isNotBlank(delegationToken)) {
            maskedDelegationToken = MASKED_TOKEN_STR;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazClient.checkPrivileges({}, {})", requests, maskedDelegationToken);
        }

        RangerRazMultiOperationResult ret         = null;
        String                        url         = String.format(RAZ_API_MULTI_RESOURCE, requests.getServiceType());
        Map<String, String>           queryParams = new HashMap<>();
        UserGroupInformation          user        = null;

        if (!StringUtils.isEmpty(delegationToken)) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("==> RangerRazClient.checkPrivileges(): RAZ Delegation token will be used to authenticate.");
            }

            queryParams.put(PARAM_DELEGATION, delegationToken);
        } else if (isKerberosAuthenticated()) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("==> RangerRazClient.checkPrivileges(): Since RAZ Delegation token is missing, Kerberos Authentication will be used to authenticate.");
            }

            user = getUgi(queryParams);
        }

        PrivilegedExceptionAction<RangerRazMultiOperationResult> action = getPrivilegedAction(HTTP_METHOD.POST, url, queryParams, requests, RangerRazMultiOperationResult.class);

        try {
            ret = executeAction(action, user);
        } catch (RangerRazException rre) {
            if (rre.getErrorCode() == RAZ_CLIENT_UNAUTHORIZED_ACCESS) {
                printAvailableAuthDetails(delegationToken);
            }
            throw rre;
        }

        if (ret != null && ret.getOperResults() != null && ret.getOperResults().size() > 0) {
            updateLatestRazIdMappingsVersion(ret.getOperResults().get(0));
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRazClient.checkPrivileges(): ret={}", ret);
        }

        return ret;
    }

    public RangerRazResult checkPrivilege(RangerRazRequest request, String delegationToken) throws RangerRazException {
        String maskedDelegationToken = null;
        if (StringUtils.isNotBlank(delegationToken)) {
            maskedDelegationToken = MASKED_TOKEN_STR;
        }

        if (LOG.isTraceEnabled()) {
            LOG.trace("==> RangerRazClient.checkPrivilege({}, {})", request, maskedDelegationToken);
        }

        RangerRazResult      ret         = null;
        String               url         = String.format(RAZ_API_SINGLE_RESOURCE, request.getServiceType());
        Map<String, String>  queryParams = new HashMap<>();
        UserGroupInformation user        = null;

        if (!StringUtils.isEmpty(delegationToken)) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("==> RangerRazClient.checkPrivilege(): RAZ Delegation token will be used to authenticate.");
            }

            queryParams.put(PARAM_DELEGATION, delegationToken);
        } else if (isKerberosAuthenticated()) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("==> RangerRazClient.checkPrivilege(): Since RAZ Delegation token is missing Kerberos TGT will be used to authenticate.");
            }

            user = getUgi(queryParams);
        }

        PrivilegedExceptionAction<RangerRazResult> action = getPrivilegedAction(HTTP_METHOD.POST, url, queryParams, request, RangerRazResult.class);

        try {
            ret = executeAction(action, user);
        } catch (RangerRazException rre) {
            if (rre.getErrorCode() == RAZ_CLIENT_UNAUTHORIZED_ACCESS) {
                printAvailableAuthDetails(delegationToken);
            }
            throw rre;
        }

        if (ret != null && ret.getOperResult() != null) {
            updateLatestRazIdMappingsVersion(ret.getOperResult());
        }

        if (LOG.isTraceEnabled()) {
            LOG.trace("<== RangerRazClient.checkPrivilege(): ret={}", ret);
        }

        return ret;
    }

    private static String getDTServiceName(List<String> razUrls) {
        String ret    = null;

        if (razUrls != null && razUrls.size() > 0) {
            for (String razUrl : razUrls) {
                try {
                    ret = SecurityUtil.buildDTServiceName(URI.create(razUrl), DEFAULT_RAZ_SERVER_PORT);
                    break;
                } catch (Exception e) {
                    // Try next valid razURL
                }
            }
        }

        if (StringUtils.isEmpty(ret)) {
            ret = DEFAULT_DT_SERVICE_NAME;
        }

        return ret;
    }

    private <T> PrivilegedExceptionAction<T> getPrivilegedAction(HTTP_METHOD method, String relativeUrl, Map<String, String> queryParams, Object postParam, Class<T> clazz) {
        return () -> restClient.executeHttpRequest(method, relativeUrl, queryParams, postParam, clazz);
    }

    private <T> T executeAction(PrivilegedExceptionAction<T> action, UserGroupInformation owner) throws RangerRazException {
        T ret = null;

        try {
            if (owner != null) {
                ret = owner.doAs(action);
            } else {
                ret = action.run();
            }
        } catch (Exception excp) {
            // in case we get our exceptions wrapped in
            // PrivilegedActionException, we check the cause of exception
            // else we check the exception.
            if (excp.getCause() instanceof RangerRazException) {
                throw (RangerRazException) excp.getCause();
            } else if (excp instanceof RangerRazException) {
                throw (RangerRazException) excp;
            } else {
                throw new RangerRazException(excp);
            }
        }

        if (ret == null) {
            throw new RangerRazException(RAZ_CLIENT_NULL_RESPONSE);
        }

        return ret;
    }

    private JsonNode getJsonNode(String jsonString, String value) throws RangerRazException {
        JsonNode ret = null;

        if (jsonString != null) {
            try {
                JsonNode node = mapper.readTree(jsonString);

                if (node != null) {
                    ret = node.findValue(value);
                }
            } catch (IOException excp) {
                RangerRazClientLogger.error(LOG, "RangerRazClient.getJsonNode({}): failed to parse the response", jsonString, excp);

                throw new RangerRazException(RAZ_CLIENT_INVALID_RESPONSE, jsonString);
            }
        }

        if (ret == null) {
            RangerRazClientLogger.error(LOG, "RangerRazClient.getJsonNode({}): {} not found", jsonString, value);

            throw new RangerRazException(RAZ_CLIENT_INVALID_RESPONSE, value + " not found");
        }

        return ret;
    }

    private UserGroupInformation getUgi(Map<String, String> queryParams) throws RangerRazException {
        UserGroupInformation ret = this.ugi;

        /*
         * If ugi is a proxy user, return the ugi of the real-user, and add doAs to
         * queryParams
         */
        UserGroupInformation realUser = ret.getRealUser();

        if (realUser != null) {
            if (queryParams != null && !StringUtils.equals(realUser.getShortUserName(), ret.getShortUserName())) {
                String proxyUser = ret.getShortUserName();

                queryParams.put(PARAM_DOAS, proxyUser);

                if (LOG.isDebugEnabled()) {
                    LOG.debug("RangerRazClient.getUgi(): owner = {} and proxy user = {}", realUser.getShortUserName(), proxyUser);
                }
            }

            ret = realUser;
        }

        return ret;
    }

    private boolean isKerberosAuthenticated() {
        boolean isKerberosAuthenticated = false;
        try {
            UserGroupInformation loggedInUser           = UserGroupInformation.getLoginUser();
            boolean              isSecurityEnabled      = UserGroupInformation.isSecurityEnabled();
            boolean              hasKerberosCredentials = loggedInUser.hasKerberosCredentials();
            AuthenticationMethod loggedInUserAuthMethod = loggedInUser.getAuthenticationMethod();

            isKerberosAuthenticated = isSecurityEnabled && hasKerberosCredentials && loggedInUserAuthMethod.equals(AuthenticationMethod.KERBEROS);

            if (isKerberosAuthenticated) {
                loggedInUser.checkTGTAndReloginFromKeytab();
            }
        } catch (IOException e) {
            LOG.error("Failed to get authentication details. Exception : {}", e);
        }

        return isKerberosAuthenticated;
    }

    @SuppressWarnings("serial")
    private Map<String, String> createMap(String key1, String value1, String key2, String value2) {
        return new HashMap<String, String>() {
            {
                put(key1, value1);
                put(key2, value2);
            }
        };
    }

    @SuppressWarnings("serial")
    private Map<String, String> createMap(String key1, String value1, String key2, String value2, String key3, String value3) {
        return new HashMap<String, String>() {
            {
                put(key1, value1);
                put(key2, value2);
                put(key3, value3);
            }
        };
    }

    /**
     * Get latest {@link RangerRazIdentityMappings}.</br>
     * Use {@link RangerRazClient#getIdentityMappingUtil(delegationToken)} instead
     * for performance.
     * 
     * @param delegationToken
     * @return RangerRazIdentityMappings retrieved from RAZ server
     * @throws RangerRazException
     */
    public RangerRazIdentityMappings getRangerRazIdentityMappings(String delegationToken) throws RangerRazException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazClient.getRangerRazIdentityMappings()");
        }

        RangerRazIdentityMappings ret         = null;
        Map<String, String>       queryParams = new HashMap<>();
        UserGroupInformation      user        = null;

        if (!StringUtils.isEmpty(delegationToken)) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("==> RangerRazClient.getRangerRazIdentityMappings(): RAZ Delegation token will be used to authenticate.");
            }

            queryParams.put(PARAM_DELEGATION, delegationToken);
        } else if (isKerberosAuthenticated()) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("==> RangerRazClient.getRangerRazIdentityMappings(): Since RAZ Delegation token is missing Kerberos TGT will be used to authenticate.");
            }

            user = getUgi(queryParams);
        }

        PrivilegedExceptionAction<RangerRazIdentityMappings> action = getPrivilegedAction(HTTP_METHOD.GET, RAZ_API_DOWNLOAD_OID_MAP, queryParams, null, RangerRazIdentityMappings.class);
        ret = executeAction(action, user);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRazClient.getRangerRazIdentityMappings(): ret={}", ret);
        }

        return ret;
    }

    /**
     * Keeps latest {@link RangerRazIdentityMappingsUtil} in cache.</br>
     * Checks if cached version is latest, update cache if not latest.
     * 
     * @param delegationToken
     * @return latest {@link RangerRazIdentityMappingsUtil}
     * @throws RangerRazException
     */
    public RangerRazIdentityMappingsUtil getIdentityMappingUtil(String delegationToken) throws RangerRazException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRazClient.getIdentityMappingUtil()");
        }

        RangerRazIdentityMappingsUtil ret = this.razIdMappUtil;
        if (ret == null || !latestRazIdMappingVersion.equals(ret.getVersion())) {
            synchronized (DUMMY_POST_PARAM) {
                ret = this.razIdMappUtil;
                if (ret == null || !latestRazIdMappingVersion.equals(ret.getVersion())) {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("==> RangerRazClient.getIdentityMappingUtil(): updating cached razIdMappUtil");
                    }

                    // update cache
                    RangerRazIdentityMappings razIdMappings = getRangerRazIdentityMappings(delegationToken);

                    latestRazIdMappingVersion = razIdMappings.getVersion();
                    this.razIdMappUtil        = new RangerRazIdentityMappingsUtil(razIdMappings);
                    ret                       = this.razIdMappUtil;
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRazClient.getIdentityMappingUtil(): ret={}", ret);
        }
        return ret;
    }

    /**
     * Updates latest available version of {@link RangerRazIdentityMappings}
     * 
     * @param rangerAccessResult
     */
    private void updateLatestRazIdMappingsVersion(ResourceAccessResult rangerAccessResult) {
        if (rangerAccessResult != null) {
            Map<String, String> map = rangerAccessResult.getAdditionalInfo();

            if (MapUtils.isNotEmpty(map)) {
                String razMappingVersionStr = map.get("RAZ_ID_MAP_VERSION");

                if (StringUtils.isNotBlank(razMappingVersionStr)) {
                    Long latestRazIdMappsVersion = Long.valueOf(razMappingVersionStr);

                    if (!latestRazIdMappsVersion.equals(latestRazIdMappingVersion)) {
                        synchronized (DUMMY_POST_PARAM) {
                            if (!latestRazIdMappsVersion.equals(latestRazIdMappingVersion)) {
                                latestRazIdMappingVersion = Long.valueOf(razMappingVersionStr);
                            }
                        }
                    }
                }
            }
        }
    }

    private void printAvailableAuthDetails(final String delegationToken) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("RAZ Delegation Token: IsAuthInfoPresent ? {}", StringUtils.isNotBlank(delegationToken));
            LOG.debug("Kerberos TGT        : IsAuthInfoPresent ? {}", isKerberosAuthenticated());
        }
    }

    public static String maskParam(final String orgParam) {
        String ret = null;
        if (StringUtils.isNotBlank(orgParam)) {
            StringBuffer tempRazDToken = new StringBuffer();
            if (orgParam.length() > RangerRazClient.RAZ_CLIENT_SKIP_MASK_FOR_FIRST_DEFAULT) {
                tempRazDToken.append(orgParam.substring(0, RangerRazClient.RAZ_CLIENT_SKIP_MASK_FOR_FIRST_DEFAULT));
            }
            tempRazDToken.append(RangerRazClient.MASKED_TOKEN_STR);
            ret = tempRazDToken.toString();
        }

        return ret;
    }
}