/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.intg.client.executor;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.UserGroupInformation.AuthenticationMethod;
import org.apache.ranger.raz.intg.token.TokenRetriever;

public interface RestClientExecutor<C> {
    public static final String REST_ACCEPTED_MIME_TYPE_JSON     = "application/json";
    public static final String REST_CONTENT_TYPE_MIME_TYPE_JSON = "application/json";
    public static final String REST_HEADER_ACCEPT               = "Accept";
    public static final String REST_HEADER_CONTENT_TYPE         = "Content-type";
    public static final String GSON_DATE_FORMAT                 = "yyyy-MM-dd'T'HH:mm:ss.SSSZ";
    public static final String JWT_COOKIE_NAME_DEFAULT          = "hadoop-jwt";
    public static final String AUTHORIZATION_HEADER             = "Authorization";

    C build();

    default List<String> getURLs(String configURLs) {
        List<String> configuredURLs = new ArrayList<>();
        if (configURLs != null) {
            String[] urls = configURLs.split(",");
            for (String strUrl : urls) {
                if (!StringUtils.isEmpty(StringUtils.trimToEmpty(strUrl))) {
                    if (strUrl.endsWith("/")) {
                        strUrl = strUrl.substring(0, strUrl.length() - 1);
                    }
                    configuredURLs.add(strUrl);
                }
            }
        }
        return configuredURLs;
    }

    default boolean isKerberosAuthenticated() throws Exception {
        boolean status = false;
        try {
            UserGroupInformation loggedInUser           = UserGroupInformation.getLoginUser();
            boolean              isSecurityEnabled      = UserGroupInformation.isSecurityEnabled();
            boolean              hasKerberosCredentials = loggedInUser.hasKerberosCredentials();
            AuthenticationMethod loggedInUserAuthMethod = loggedInUser.getAuthenticationMethod();

            status = isSecurityEnabled && hasKerberosCredentials && loggedInUserAuthMethod.equals(AuthenticationMethod.KERBEROS);
        } catch (IOException e) {
            throw new Exception("Failed to get authentication details.", e);
        }

        return status;
    }

    @SuppressWarnings("unchecked")
    default TokenRetriever<String> getJwtTokenRetriever(RestClientExecutorConfig clientConfig) throws Exception {
        String                        clsName   = clientConfig.getJwtTokenRetrieverClass();
        ClassLoader                   clsLoader = Thread.currentThread().getContextClassLoader();
        TokenRetriever<String>        ret       = null;
        Class<TokenRetriever<String>> cls;

        cls = (Class<TokenRetriever<String>>) clsLoader.loadClass(clsName.trim());
        ret = cls.getConstructor(RestClientExecutorConfig.class).newInstance(clientConfig);

        return ret;
    }
}