/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.intg.client.executor;

import javax.ws.rs.core.MediaType;

import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_ACCESS_DENIED;
import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_REQUEST_FAILED;
import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_UNAUTHORIZED_ACCESS;
import static org.apache.ranger.raz.intg.RangerRazException.WEBAPPLICATIONEXCP_WITH_HEADERS_TAG;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.auth.AuthSchemeProvider;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.KerberosCredentials;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.config.AuthSchemes;
import org.apache.http.client.config.CookieSpecs;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.config.Lookup;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.auth.SPNegoSchemeFactory;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.client.StandardHttpRequestRetryHandler;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.protocol.HttpContext;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRazClientLogger;
import org.apache.ranger.raz.intg.token.TokenRetriever;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.Map.Entry;

public class RestClientExecutorApacheHttp implements RestClientExecutor<CloseableHttpClient>, HttpClientExecutor {
    private static final Logger LOG            = LoggerFactory.getLogger(RestClientExecutorApacheHttp.class);
    private static final String PROTOCOL_HTTPS = "https";
    private static final String RESPONSE_ETAG  = "ETag";

    private volatile CloseableHttpClient      httpClient;
    private final    RestClientExecutorConfig clientConfig;
    private final    Gson                     gsonBuilder;

    private TokenRetriever<String> tokenRetriever      = null;
    private String                 jwtServerCookieName = null;

    public RestClientExecutorApacheHttp(RestClientExecutorConfig config) {
        this.clientConfig = config;
        this.gsonBuilder  = new GsonBuilder().setDateFormat(GSON_DATE_FORMAT).create();
    }

    public void resetClient() {
        httpClient = null;
    }

    @Override
    public CloseableHttpClient build() {
        Lookup<AuthSchemeProvider> authRegistry  = RegistryBuilder.<AuthSchemeProvider>create().register(AuthSchemes.SPNEGO, new SPNegoSchemeFactory(true, true)).build();
        HttpClientBuilder          clientBuilder = HttpClients.custom().setDefaultAuthSchemeRegistry(authRegistry);

        if (LOG.isDebugEnabled()) {
            LOG.debug("===>> Building apache HTTP client with configs={}", clientConfig);
        }

        try {
            BasicCredentialsProvider credentialsProvider = new BasicCredentialsProvider();

            if (isKerberosAuthenticated()) {
                credentialsProvider.setCredentials(AuthScope.ANY, new KerberosCredentials(null));
            } else if (StringUtils.isNotEmpty(clientConfig.getUsername()) && StringUtils.isNotEmpty(clientConfig.getPassword())) {
                credentialsProvider.setCredentials(AuthScope.ANY, new UsernamePasswordCredentials(clientConfig.getUsername(), clientConfig.getPassword()));
            }

            clientBuilder.setDefaultCredentialsProvider(credentialsProvider);
        } catch (Exception excp) {
            LOG.error("Exception while adding credentials, skipping setting credentials for client.", excp); //log and proceed
        }

        PoolingHttpClientConnectionManager connectionManager;

        if (clientConfig.isSsl()) {
            SSLConnectionSocketFactory        sslsf                 = new SSLConnectionSocketFactory(clientConfig.getSslContext(), NoopHostnameVerifier.INSTANCE);
            Registry<ConnectionSocketFactory> socketFactoryRegistry = RegistryBuilder.<ConnectionSocketFactory>create().register(PROTOCOL_HTTPS, sslsf).build();

            connectionManager = new PoolingHttpClientConnectionManager(socketFactoryRegistry);

            clientBuilder.setSSLSocketFactory(sslsf);
        } else {
            connectionManager = new PoolingHttpClientConnectionManager();
        }

        connectionManager.setMaxTotal(clientConfig.getMaxConnections());
        connectionManager.setDefaultMaxPerRoute(clientConfig.getMaxConnectionsPerHost());
        connectionManager.setValidateAfterInactivity(clientConfig.getValidateAfterInactivityMs());

        RequestConfig       customizedRequestConfig = RequestConfig.custom().setCookieSpec(CookieSpecs.DEFAULT)
                                                        .setConnectTimeout(clientConfig.getRestClientConnTimeOutMs())
                                                        .setSocketTimeout(clientConfig.getRestClientReadTimeOutMs()).build();

        CloseableHttpClient httpClient              = clientBuilder.setConnectionManager(connectionManager)
                                                        .setRetryHandler(new RazRetryHandlerApacheHttp(clientConfig.getPoolRetryCount(), true))
                                                        .setDefaultRequestConfig(customizedRequestConfig).build();

        jwtServerCookieName = clientConfig.getJwtServerCookieName();
        if (StringUtils.isBlank(jwtServerCookieName)) {
            jwtServerCookieName = JWT_COOKIE_NAME_DEFAULT;
        }

        try {
            tokenRetriever = getJwtTokenRetriever(clientConfig);
        } catch (Exception e) {
            LOG.error("RestClientExecutorApacheHttp.build(): Failed to initialize JWT token retriever.", e);
        }

        return httpClient;

    }

    @Override
    public <T> T getAndParse(String url, Map<String, String> queryParams, Class<T> clazz) throws Exception {
        HttpGet httpGet = new HttpGet(buildURI(url, queryParams));

        return executeAndParseResponse(httpGet, clazz);
    }

    @Override
    public <T> T postAndParse(String url, Map<String, String> queryParams, Object obj, Class<T> clazz) throws Exception {
        HttpPost     httpPost = new HttpPost(buildURI(url, queryParams));
        StringEntity entity   = new StringEntity(gsonBuilder.toJson(obj));

        httpPost.setEntity(entity);

        return executeAndParseResponse(httpPost, clazz);
    }

    @Override
    public <T> T putAndParse(String url, Map<String, String> queryParams, Object obj, Class<T> clazz) throws Exception {
        HttpPut      httpPut = new HttpPut(buildURI(url, queryParams));
        StringEntity entity  = new StringEntity(gsonBuilder.toJson(obj));

        httpPut.setEntity(entity);

        return executeAndParseResponse(httpPut, clazz);
    }

    @Override
    public <T> T deleteAndParse(String url, Map<String, String> queryParams, Class<T> clazz) throws Exception {
        HttpDelete httpDelete = new HttpDelete(buildURI(url, queryParams));

        return executeAndParseResponse(httpDelete, clazz);
    }

    @SuppressWarnings("unchecked")
    private <R> R parseResponse(HttpResponse response, Class<R> clazz) throws RangerRazException, UnsupportedOperationException, IOException {
        R type = null;

        if (response != null) {
            int    httpStatus  = response.getStatusLine().getStatusCode();
            String contentType = "";
            if (response.getEntity().getContentType() != null) {
                contentType = response.getEntity().getContentType().getValue();
            }

            if (httpStatus == HttpStatus.SC_OK) {
                InputStream responseInputStream = response.getEntity().getContent();

                if (clazz.equals(String.class)) {
                    type = (R) IOUtils.toString(responseInputStream, Charset.defaultCharset());
                } else {
                    type = gsonBuilder.fromJson(new InputStreamReader(responseInputStream), clazz);
                }
                responseInputStream.close();
            } else {
                RangerRazClientLogger.error(LOG, "Request failed : response=[{}], response.status={}", response, httpStatus);
                Map<String, String> exceptionHeaders = new HashMap<>();
                // Extracting content from Response body.
                String errorMessage = "";
                try (InputStream responseInputStream = response.getEntity().getContent()) {
                    errorMessage = IOUtils.toString(responseInputStream, Charset.defaultCharset());
                }

                if (contentType.equals(MediaType.APPLICATION_JSON)) {

                    // The HttpResponse ETag would help in identifying what kind of response is it.
                    String responseTag = null;
                    if (response.containsHeader(RESPONSE_ETAG)) {
                        responseTag = response.getFirstHeader(RESPONSE_ETAG).getValue();
                    }

                    // This string have double quotes, which needs to be stripped before comparing
                    responseTag = responseTag != null ? responseTag.replace("\"", "") : null;

                    // An exception with headers was raised in RazServer
                    if (responseTag != null && responseTag.equals(WEBAPPLICATIONEXCP_WITH_HEADERS_TAG)) {

                        // Putting all the headers from the response into the map.
                        for (Header header : response.getAllHeaders()) {
                            exceptionHeaders.put(header.getName(), header.getValue());
                        }
                    }
                }

                // Create a new RangerRazException with the errorMessage and desired ErrorCode and throw.
                RangerRazException rangerRazException = new RangerRazException(errorMessage + "; HttpStatus: " + httpStatus);
                if (httpStatus == HttpStatus.SC_FORBIDDEN) {
                    rangerRazException.setErrorCode(RAZ_CLIENT_ACCESS_DENIED);
                } else if (httpStatus == HttpStatus.SC_UNAUTHORIZED) {
                    LOG.warn("Server did not process this request due to in-sufficient auth details.");
                    rangerRazException.setErrorCode(RAZ_CLIENT_UNAUTHORIZED_ACCESS);
                } else {
                    rangerRazException.setErrorCode(RAZ_CLIENT_REQUEST_FAILED);
                }

                rangerRazException.setRangerRazExceptionHeadersMap(exceptionHeaders);
                throw rangerRazException;
            }
        } else {
            RangerRazClientLogger.error(LOG, "Received NULL response from server.");

            throw new RangerRazException(RAZ_CLIENT_REQUEST_FAILED);
        }

        return type;
    }

    private URI buildURI(String url, Map<String, String> queryParams) throws URISyntaxException {
        URIBuilder builder = new URIBuilder(url);

        if (queryParams != null) {
            for (Entry<String, String> param : queryParams.entrySet()) {
                builder.addParameter(param.getKey(), param.getValue());
            }
        }

        return builder.build();
    }

    private CloseableHttpClient getCloseableHttpClient() {
        // result saves on access time when client is built at the time of the call
        CloseableHttpClient result = httpClient;

        if (result == null) {
            synchronized (this) {
                result = httpClient;

                if (result == null) {
                    httpClient = result = build();
                }
            }
        }

        return result;
    }

    private <T extends HttpRequestBase> T addCommonHeaders(T t) {
        t.addHeader(REST_HEADER_ACCEPT, REST_ACCEPTED_MIME_TYPE_JSON);
        t.setHeader(REST_HEADER_CONTENT_TYPE, REST_CONTENT_TYPE_MIME_TYPE_JSON);

        t = handleJwt(t);

        return t;
    }

    private <T extends HttpRequestBase> T handleJwt(T t) {
        if (tokenRetriever != null) {
            Optional<String> jwtOptional = tokenRetriever.retrieve();

            if (jwtOptional.isPresent()) {
                t.setHeader(AUTHORIZATION_HEADER, jwtOptional.get());

                StringBuffer jwtWithCookieName = new StringBuffer();
                jwtWithCookieName.append(jwtServerCookieName);
                jwtWithCookieName.append("=");
                jwtWithCookieName.append(jwtOptional.get());

                t.setHeader("Cookie", jwtWithCookieName.toString());
            }
        } else {
            LOG.warn("Since JWTokenRetriver init failed, skipping JWT auth.");
        }

        return t;
    }

    private <T extends HttpRequestBase, R> R executeAndParseResponse(T request, Class<R> responseClazz) throws Exception {
        R                   ret    = null;
        CloseableHttpClient client = getCloseableHttpClient();

        if (client != null) {
            addCommonHeaders(request);

            try (CloseableHttpResponse response = client.execute(request)) {
                ret = parseResponse(response, responseClazz);
            }
        } else {
            RangerRazClientLogger.error(LOG, "Can not procees request as client is null.");
        }

        return ret;
    }

    private class RazRetryHandlerApacheHttp extends StandardHttpRequestRetryHandler {

        public RazRetryHandlerApacheHttp(Integer poolRetryCount, boolean isIdempotentRequest) {
            super(poolRetryCount, isIdempotentRequest);
        }

        @Override
        public boolean retryRequest(IOException exception, int executionCount, HttpContext context) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("===>> RazRetryHandlerApacheHttp.retryRequest({},{})", exception.getMessage(), executionCount);
            }

            boolean ret = super.retryRequest(exception, executionCount, context);

            if (LOG.isDebugEnabled()) {
                LOG.debug("<<=== RazRetryHandlerApacheHttp.retryRequest(): ret={}", ret);
            }

            return ret;
        }
    }
}
