/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.intg.client.executor;

import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_ACCESS_DENIED;
import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_REQUEST_FAILED;
import static org.apache.ranger.raz.intg.RangerRazErrorCode.RAZ_CLIENT_UNAUTHORIZED_ACCESS;

import java.util.Map;
import java.util.Optional;
import java.util.Set;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.SSLSession;
import javax.ws.rs.core.Cookie;

import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.intg.client.RangerRazClientLogger;
import org.apache.ranger.raz.intg.token.TokenRetriever;
import org.codehaus.jackson.jaxrs.JacksonJsonProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.ClientResponse;
import com.sun.jersey.api.client.WebResource;
import com.sun.jersey.api.client.WebResource.Builder;
import com.sun.jersey.api.client.config.ClientConfig;
import com.sun.jersey.api.client.config.DefaultClientConfig;
import com.sun.jersey.api.client.filter.HTTPBasicAuthFilter;
import com.sun.jersey.client.urlconnection.HTTPSProperties;

public class RestClientExecutorJersey implements RestClientExecutor<Client>, HttpClientExecutor {
    private static final Logger LOG = LoggerFactory.getLogger(RestClientExecutorJersey.class);

    private volatile Client client;

    private final RestClientExecutorConfig clientConfig;
    private final Gson                     gsonBuilder;

    private TokenRetriever<String> tokenRetriever      = null;
    private String                 jwtServerCookieName = null;

    public RestClientExecutorJersey(RestClientExecutorConfig config) {
        this.clientConfig = config;
        this.gsonBuilder  = new GsonBuilder().setDateFormat(GSON_DATE_FORMAT).create();
    }

    public Client getClient() {
        // result saves on access time when client is built at the time of the call
        Client result = client;
        if (result == null) {
            synchronized (this) {
                result = client;
                if (result == null) {
                    client = result = build();
                }
            }
        }

        return result;
    }

    @Override
    public Client build() {
        Client ret = null;
        if (clientConfig.isSsl()) {
            ClientConfig config = new DefaultClientConfig();

            config.getClasses().add(JacksonJsonProvider.class); // to handle List<> unmarshalling

            HostnameVerifier hv = new HostnameVerifier() {
                public boolean verify(String urlHostName, SSLSession session) {
                    return session.getPeerHost().equals(urlHostName);
                }
            };

            config.getProperties().put(HTTPSProperties.PROPERTY_HTTPS_PROPERTIES, new HTTPSProperties(hv, clientConfig.getSslContext()));

            ret = Client.create(config);
        }

        if (ret == null) {
            ClientConfig config = new DefaultClientConfig();

            config.getClasses().add(JacksonJsonProvider.class); // to handle List<> unmarshalling

            ret = Client.create(config);
        }

        if (StringUtils.isNotEmpty(clientConfig.getUsername()) && StringUtils.isNotEmpty(clientConfig.getPassword())) {
            ret.addFilter(new HTTPBasicAuthFilter(clientConfig.getUsername(), clientConfig.getPassword()));
        }

        // Set Connection Timeout and ReadTime for the PolicyRefresh
        ret.setConnectTimeout(clientConfig.getRestClientConnTimeOutMs());
        ret.setReadTimeout(clientConfig.getRestClientReadTimeOutMs());

        jwtServerCookieName = clientConfig.getJwtServerCookieName();
        if (StringUtils.isBlank(jwtServerCookieName)) {
            jwtServerCookieName = JWT_COOKIE_NAME_DEFAULT;
        }

        try {
            tokenRetriever = getJwtTokenRetriever(clientConfig);
        } catch (Exception e) {
            LOG.error("RestClientExecutorJersey.build(): Failed to initialize JWT token retriever.", e);
        }

        return ret;
    }

    @Override
    public <T> T getAndParse(String url, Map<String, String> queryParams, Class<T> clazz) throws Exception {
        ClientResponse resp = getWebResourceBuilder(url, queryParams).get(ClientResponse.class);
        return parseResponse(resp, clazz);
    }

    @Override
    public <T> T postAndParse(String url, Map<String, String> queryParams, Object obj, Class<T> clazz) throws Exception {
        ClientResponse resp = getWebResourceBuilder(url, queryParams).post(ClientResponse.class, gsonBuilder.toJson(obj));
        return parseResponse(resp, clazz);
    }

    @Override
    public <T> T putAndParse(String url, Map<String, String> queryParams, Object obj, Class<T> clazz) throws Exception {
        ClientResponse resp = getWebResourceBuilder(url, queryParams).put(ClientResponse.class, gsonBuilder.toJson(obj));
        return parseResponse(resp, clazz);
    }

    @Override
    public <T> T deleteAndParse(String url, Map<String, String> queryParams, Class<T> clazz) throws Exception {
        ClientResponse resp = getWebResourceBuilder(url, queryParams).delete(ClientResponse.class);
        return parseResponse(resp, clazz);
    }

    private Builder getWebResourceBuilder(String url, Map<String, String> queryParams) {
        WebResource webResource = getClient().resource(url);
        webResource = setQueryParams(webResource, queryParams);

        Builder builder = webResource.accept(REST_ACCEPTED_MIME_TYPE_JSON).type(REST_CONTENT_TYPE_MIME_TYPE_JSON);
        return builder;
    }

    private WebResource setQueryParams(WebResource webResource, Map<String, String> queryParams) {
        WebResource ret = webResource;
        WebResource.Builder builder = ret.getRequestBuilder();

        builder = handleJwt(builder);

        if (webResource != null && queryParams != null) {
            Set<Map.Entry<String, String>> entrySet = queryParams.entrySet();
            for (Map.Entry<String, String> entry : entrySet) {
                ret = ret.queryParam(entry.getKey(), entry.getValue());
            }
        }

        return ret;
    }

    private WebResource.Builder handleJwt(WebResource.Builder builder) {
        if (tokenRetriever != null) {
            Optional<String> jwtOptional = tokenRetriever.retrieve();

            if (jwtOptional.isPresent()) {
                builder.cookie(new Cookie(jwtServerCookieName, jwtOptional.get()));
                builder.header(AUTHORIZATION_HEADER, jwtOptional.get());
            }
        } else {
            LOG.warn("Since JWTokenRetriver init failed, skipping JWT auth.");
        }

        return builder;
    }

    private <R> R parseResponse(ClientResponse response, Class<R> clazz) throws RangerRazException {
        R type = null;

        if (response != null) {
            try {
                if (response.getStatus() == HttpStatus.SC_OK) {
                    type = response.getEntity(clazz);
                } else {
                    int httpStatus = response.getStatus();

                    RangerRazClientLogger.error(LOG, "Request failed : response=[{}], response.status={}", response, httpStatus);

                    if (httpStatus == HttpStatus.SC_FORBIDDEN || httpStatus == HttpStatus.SC_UNAUTHORIZED) {
                        throw new RangerRazException(RAZ_CLIENT_ACCESS_DENIED);
                    } else if (httpStatus == HttpStatus.SC_UNAUTHORIZED) {
                        LOG.warn("Server did not process this request due to in-sufficient auth details.");
                        throw new RangerRazException(RAZ_CLIENT_UNAUTHORIZED_ACCESS);
                    } else {
                        throw new RangerRazException(RAZ_CLIENT_REQUEST_FAILED, httpStatus);
                    }
                }
            } finally {
                try {
                    response.close();
                } catch (Exception excp) {
                    // ignore
                }
            }
        } else {
            RangerRazClientLogger.error(LOG, "Received NULL response from server.");
            throw new RangerRazException(RAZ_CLIENT_REQUEST_FAILED);
        }

        return type;
    }
}