/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.intg.token;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Optional;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.raz.intg.client.executor.RestClientExecutorConfig;

public class JwTokenRetrieverFile implements TokenRetriever<String> {
    private static final Log LOG = LogFactory.getLog(JwTokenRetrieverFile.class);

    private File             jwtFile      = null;
    private Optional<String> cachedJWT    = Optional.empty();
    private Long             lastModified = -1L;

    public JwTokenRetrieverFile(RestClientExecutorConfig config) {
        init(config);
    }

    private void init(RestClientExecutorConfig config) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> JwTokenRetrieverFile.init()");
        }

        String fileName = config.getJwtFileName();
        if (StringUtils.isNotBlank(fileName)) {
            jwtFile = new File(fileName);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== JwTokenRetrieverFile.init(): fileName=" + fileName);
        }
    }

    @Override
    public Optional<String> retrieve() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> JwTokenRetrieverFile.retrieve()");
        }

        Optional<String> ret = cachedJWT;

        if (jwtFile != null && jwtFile.lastModified() > this.lastModified) {
            try (BufferedReader reader = new BufferedReader(new FileReader(jwtFile))) {
                String line = null;
                while ((line = reader.readLine()) != null) {
                    if (StringUtils.isNotBlank(line) && !line.startsWith("#")) {
                        ret               = Optional.of(line);
                        this.lastModified = jwtFile.lastModified();
                        break;
                    }
                }
            } catch (IOException e) {
                LOG.error("JwTokenRetrieverFile.retrieve(): Failed to get JWT token.", e);
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== JwTokenRetrieverFile.retrieve(): isJwTokenPresent=" + ret.isPresent());
        }

        return ret;
    }
}