/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.processor;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.ranger.authorization.hadoop.constants.RangerHadoopConstants;
import org.apache.ranger.plugin.policyengine.*;
import org.apache.ranger.plugin.policyengine.RangerAccessRequest.ResourceMatchingScope;
import org.apache.ranger.raz.intg.*;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.raz.model.RangerRazResultBase.AccessResult;
import org.apache.ranger.raz.model.RangerRazResultBase.ResourceAccessResult;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;


public class RangerDefaultRazProcessor implements RangerRazProcessor {
	private static final Logger LOG = LoggerFactory.getLogger(RangerDefaultRazProcessor.class);

	public RangerDefaultRazProcessor() {
		if (LOG.isDebugEnabled()) {
			LOG.debug("RangerDefaultRazProcessor()");
		}
	}

	@Override
	public void init(RangerRazContext context) throws RangerRazException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> RangerDefaultRazProcessor.init(context={})", context);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== RangerDefaultRazProcessor.init(context={})", context);
		}
	}

	@Override
	public List<RangerAccessRequest> preProcess(RangerRazRequest request, RangerRazRequestContext context) throws RangerRazException {
		if (LOG.isTraceEnabled()) {
			LOG.trace("==> RangerDefaultRazProcessor.preProcess(request={}, context={})", request, context);
		}

		RangerRazRequest.ResourceAccess resourceAccess = request.getOperation();

		if (StringUtils.isEmpty(request.getServiceName())) {
			throw new RangerRazException(RangerRazErrorCode.INVALID_PARAMETERS, "serviceName must be provided");
		}

		if (resourceAccess == null || MapUtils.isEmpty(resourceAccess.getResource())) {
			throw new RangerRazException(RangerRazErrorCode.INVALID_PARAMETERS, "resource must be provided");
		}

		if (CollectionUtils.isEmpty(resourceAccess.getAccessTypes())) {
			throw new RangerRazException(RangerRazErrorCode.INVALID_PARAMETERS, "accessTypes must be provided");
		}

		List<RangerAccessRequest> ret = new ArrayList<>();

		// TODO:
		// validate that
		// 1. if request.userName is not same as context.remoteUser, ensure that context.remoteUser is a service-admin
		//
		String      userName   = request.getUser();
		Set<String> userGroups = request.getUserGroups();

		for (String accessType : resourceAccess.getAccessTypes()) {
			Map<String, Object>      reqContext     = request.getContext() != null ? new HashMap<>(request.getContext()) : null;
			RangerAccessResource     accessResource = getRangerAccessResource(resourceAccess);
			RangerAccessRequestImpl  accessRequest  = new RangerAccessRequestImpl(accessResource, accessType, userName, userGroups, null);

			accessRequest.setAction(resourceAccess.getAction());
			accessRequest.setAccessTime(request.getAccessTime());
			accessRequest.setClientIPAddress(request.getClientIpAddress());
			accessRequest.setClientType(request.getClientType());
			accessRequest.setClusterName(request.getClusterName());
			accessRequest.setClusterType(request.getClientType());
			accessRequest.setSessionId(request.getSessionId());
			accessRequest.setContext(reqContext);
			accessRequest.setResourceMatchingScope(ResourceMatchingScope.SELF_OR_CHILD);

			ret.add(accessRequest);
		}

		if (LOG.isTraceEnabled()) {
			LOG.trace("<== RangerDefaultRazProcessor.preProcess(request={}, context={}): ret={}", request, context, ret);
		}

		return ret;
	}

	@Override
	public RangerRazResult postProcess(RangerRazRequest request, List<RangerAccessResult> accessResults, RangerRazRequestContext context) throws RangerRazException {
		if (LOG.isTraceEnabled()) {
			LOG.trace("==> RangerDefaultRazProcessor.postProcess(request={}, accessResults={}, context={})", request, accessResults, context);
		}

		ResourceAccessResult rar = new ResourceAccessResult(AccessResult.ALLOWED, false, new ArrayList<>(), null, context.getKeysToRedact());

		for (RangerAccessResult accessResult : accessResults) {
			final AccessResult result;

			if (accessResult.getIsAccessDetermined()) {
				if (accessResult.getIsAllowed()) {
					result = AccessResult.ALLOWED;
				} else {
					result = AccessResult.DENIED;
				}
			} else {
				result = AccessResult.NOT_DETERMINED;
			}

			if (rar.getResult() != AccessResult.DENIED) { // don't overwrite deny
				if (result == AccessResult.DENIED) {
					rar.setResult(AccessResult.DENIED);
				} else if (result == AccessResult.NOT_DETERMINED) {
					rar.setResult(AccessResult.NOT_DETERMINED);
				}
			}

			if (accessResult.getIsAudited()) {
				rar.setIsAudited(true);
			}

			RangerRazResult.AuditInfo audit = new RangerRazResult.AuditInfo(accessResult.getAuditLogId(), accessResult.getAccessRequest().getAccessType(), result, accessResult.getPolicyId(), accessResult.getPolicyVersion());

			rar.getAuditLogs().add(audit);

			// pick the first non-null accessResult.additionalInfo
			if (rar.getAdditionalInfo() == null && accessResult.getAdditionalInfo() != null) {
				Map<String, String> additionalInfo = new HashMap<>(accessResult.getAdditionalInfo().size());

				for (Map.Entry<String, Object> e : accessResult.getAdditionalInfo().entrySet()) {
					additionalInfo.put(e.getKey(), Objects.toString(e.getValue()));
				}

				rar.setAdditionalInfo(additionalInfo);
			}
		}

		RangerRazResult ret = new RangerRazResult(request.getRequestId(), rar);

		if (LOG.isTraceEnabled()) {
			LOG.trace("==> RangerDefaultRazProcessor.postProcess(request={}, accessResults={}, context={}): ret={}", request, accessResults, context, ret);
		}

		return ret;
	}

	@Override
	public String getFallbackAclEnforcerName() {
		return RangerHadoopConstants.DEFAULT_RANGER_MODULE_ACL_NAME;
	}

	public RangerAccessResource getRangerAccessResource(RangerRazRequest.ResourceAccess razResourceAccess) {
	    Map<String, Object> resource  = new HashMap<>(razResourceAccess.getResource());
		String              ownerUser = razResourceAccess.getResourceOwner();

		return new RangerAccessResourceImpl(resource, ownerUser);
	}
}
