/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.aws;

import com.amazonaws.services.s3.internal.AWSS3V4Signer;

import java.util.Arrays;
import java.util.List;
import java.util.Locale;

/**
 * <p>
 * A custom implementation of V4Signer which allows skipping of some headers
 * for signature calculation. This is useful to cache the signature and reuse
 * them in consecutive requests. For example, suppose client are requesting
 * different ranges of date from an S3 object in different request, we can just
 * calculate the signature first time and keep reusing it for next requests thus
 * improving performance. Currently headers listed under {@link
 * AWSS3V4SignerWithoutRange#listOfHeadersToIgnoreInLowerCase} are skipped as
 * these headers may change during consecutive request.
 * </p>
 * <p>
 * <b> range </b> can be different depending on requested data for a same file
 * especially during seek and read operations. </p>
 * <p>
 * <b> amz-sdk-request and amz-sdk-retry </b> headers are set by aws sdk used for
 * analysis of retry information.
 * </p>
 * <p>
 * <b> amz-sdk-invocation-id </b> is a UUID for same requests across retries set
 * by aws sdk.
 * </p>
 * <p>
 * <b> if-match </b> is used to identify if there is a version mismatch for a
 * s3 object. There maybe scenarios where if-match is not set during first
 * read and set in consecutive read requests.
 * </p>
 * <p>
 * <b> connection and x-amzn-trace-id </b> headers are already ignored by
 * {@code com.amazonaws.auth.AWS4Signer#listOfHeadersToIgnoreInLowerCase}.
 * </p>
 * <p>
 * <b> referer </b> header is used to track audit information.
 * </p>
 * <p>
 * Note: This works when accessing files on S3, but is not explicitly documented
 * in the official AWS docs.
 * </p>
 *
 */
public class AWSS3V4SignerWithoutRange extends AWSS3V4Signer {

  private static final List<String> listOfHeadersToIgnoreInLowerCase = Arrays
      .asList("connection", "x-amzn-trace-id", "range",
              "amz-sdk-invocation-id", "amz-sdk-request",
              "amz-sdk-retry", "if-match", "referer");

  @Override
  protected boolean shouldExcludeHeaderFromSigning(String header) {
    return listOfHeadersToIgnoreInLowerCase.contains(header.toLowerCase(Locale.ENGLISH));
  }
}
