/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.aws.fsrequests;

import com.amazonaws.http.HttpMethodName;

import org.apache.ranger.raz.s3.lib.RazS3Constants.AccessType;
import org.apache.ranger.raz.s3.lib.authorizer.AuthorizationRequest;
import org.apache.ranger.raz.s3.lib.signer.ContentSigningRequired;
import org.apache.ranger.raz.s3.lib.utils.S3Utils;
import com.google.common.base.Preconditions;

import java.net.URI;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Objects;

// TODO: Refactor this to be re-usable for GCS.
@ContentSigningRequired(false)
public abstract class AwsRequest implements AuthorizationRequest {

//  private static final Logger LOG = LoggerFactory.getLogger(AwsRequest.class.getName());

  private final AwsOperationType operationType;

  private final URI endpoint;
  private final Map<String, String> headers;
  private final Map<String, List<String>> parameters;
  private final String bucket;
  private final String[] resourcePaths;

  private final AccessType accessType;


  protected AwsRequest(AwsOperationType operationType, URI endpoint, Map<String, String> headers,
      Map<String, List<String>> parameters, String resourcePath, AccessType accessType) {
    this(operationType, endpoint, headers, parameters, new String[]{resourcePath}, accessType);
  }

  protected AwsRequest(AwsOperationType operationType, URI endpoint,
      Map<String, String> headers,
      Map<String, List<String>> parameters, String[] resourcePaths,
      AccessType accessType) {
    Objects.requireNonNull(endpoint);
    Objects.requireNonNull(operationType);
    Objects.requireNonNull(headers);
    Objects.requireNonNull(parameters);
    Objects.requireNonNull(resourcePaths);
    Objects.requireNonNull(accessType);
    this.operationType = operationType;
    this.endpoint = endpoint;
    String bucketName = S3Utils.getBucketNameFromHost(endpoint.getHost());
    Preconditions.checkState(bucketName != null, "Could not extract bucket name from host: [" + endpoint.getHost() + "]");
    this.bucket = bucketName;
    this.headers = headers;
    this.parameters = parameters;
    // Add a leading "/" to all requests
    this.resourcePaths = new String[resourcePaths.length];
    int count = 0;
    for (String p :resourcePaths) {
      this.resourcePaths[count++] = "/" + p;
    }
    this.accessType = accessType;
  }

  public AwsOperationType getOperationType() {
    return operationType;
  }

  public URI getEndpoint() {
    return endpoint;
  }

  public HttpMethodName getHttpMethodName() {
    return operationType.getHttpMethodName();
  }

  public Map<String, String> getHeaders() {
    return headers;
  }

  public Map<String, List<String>> getParameters() {
    return parameters;
  }

  @Override
  public String getBucket() {
    return bucket;
  }

  @Override
  public String[] getPathsToAuthorize() {
    return resourcePaths;
  }

  @Override
  public AccessType getRequiredAuthorization() {
    return accessType;
  }

  /**
   * Method to check if signature caching is allowed for this
   * AWS request. Currently only signature of read object requests
   * are cached.
   * @return T/F.
   */
  public boolean isSignatureCachingAllowed() {
    return this.getOperationType().equals(AwsOperationType.READ_OBJECT);
  }

  @Override
  public String toString() {
    return "AwsRequest{" +
        "operationType=" + operationType +
        ", endpoint=" + endpoint +
        ", headers=" + headers +
        ", parameters=" + parameters +
        ", bucket=" + bucket +
        ", resourcePaths=" + Arrays.toString(resourcePaths) +
        ", accessType=" + accessType +
        '}';
  }
}
