/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.aws.fsrequests;

import com.google.common.annotations.VisibleForTesting;
import org.apache.ranger.raz.s3.lib.RazS3Constants;

import java.net.URI;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Objects;

/**
 * Class used to represent a key used during
 * signature caching.
 * IMP: We are not adding headers in the cache key as they contains
 * values like amz-sdk-invocation-id which are unique among requests.
 * Also since we are just using the cache for caching signature of read
 * requests, parameters like encryption settings won't matter as they are
 * not sent while making READ requests. But if in future we think of
 * extending this to WRITE or other requests, we should revisit matching
 * of headers as well.
 */
public class AwsRequestCacheKey {

    private final AwsOperationType operationType;
    private final URI endpoint;
    private final Map<String, List<String>> parameters;
    private final String bucket;
    private final String[] resourcePaths;
    private final RazS3Constants.AccessType accessType;
    private final String requestOwner;

    public AwsRequestCacheKey(AwsRequest awsRequest, String requestOwner) {
        this.operationType = awsRequest.getOperationType();
        this.endpoint = awsRequest.getEndpoint();
        this.parameters = awsRequest.getParameters();
        this.bucket = awsRequest.getBucket();
        this.resourcePaths = awsRequest.getPathsToAuthorize();
        this.accessType = awsRequest.getRequiredAuthorization();
        this.requestOwner = requestOwner;
    }

    @VisibleForTesting
    public AwsRequestCacheKey(AwsOperationType operationType,
                              URI endpoint,
                              Map<String, List<String>> parameters,
                              String bucket,
                              String[] resourcePaths,
                              RazS3Constants.AccessType accessType,
                              String requestOwner) {
        this.operationType = operationType;
        this.endpoint = endpoint;
        this.parameters = parameters;
        this.bucket = bucket;
        this.resourcePaths = resourcePaths;
        this.accessType = accessType;
        this.requestOwner = requestOwner;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        AwsRequestCacheKey that = (AwsRequestCacheKey) o;
        return operationType == that.operationType
                && Objects.equals(endpoint, that.endpoint)
                && Objects.equals(parameters, that.parameters)
                && Objects.equals(bucket, that.bucket)
                && Arrays.equals(resourcePaths, that.resourcePaths)
                && accessType == that.accessType
                && requestOwner.equals(that.requestOwner);
    }

    @Override
    public int hashCode() {
        int result = Objects.hash(operationType, endpoint, parameters, bucket, accessType, requestOwner);
        result = 31 * result + Arrays.hashCode(resourcePaths);
        return result;
    }

    @Override
    public String toString() {
        return "AwsRequestCacheKey{" +
                "operationType=" + operationType +
                ", endpoint=" + endpoint +
                ", parameters=" + parameters +
                ", bucket='" + bucket + '\'' +
                ", resourcePaths=" + Arrays.toString(resourcePaths) +
                ", accessType=" + accessType +
                ", requestOwner='" + requestOwner + '\'' +
                '}';
    }
}
