/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.aws.parsers;

import com.amazonaws.Request;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.commons.io.IOUtils;
import org.apache.ranger.raz.s3.lib.aws.S3Parameters;
import org.apache.ranger.raz.s3.lib.aws.exceptions.RCSAWSException;
import org.apache.ranger.raz.s3.lib.aws.exceptions.RCSAWSException.StatusCode;
import org.apache.ranger.raz.s3.lib.exceptions.RCSUncheckedException;
import org.apache.ranger.raz.s3.lib.exceptions.RCSUncheckedException.Code;
import org.apache.ranger.raz.s3.lib.signer.ContentSigningRequired;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Map;

// Private and Unstable
public class AwsRequestParser {

  private static final Logger LOG = LoggerFactory.getLogger(AwsRequestParser.class);

  private static final String CREATE_METHOD_NAME = "create";

  public AwsRequest parseAwsRequest(Request<?> request) {
    LOG.debug("Parsing incoming request: {} ", request);
    AwsRequestHolder awsRequestHolder = getAwsRequestHolder(request, true);
    return awsRequestHolder.getAwsRequest();
  }

  /**
   * Checks whether a request requires content signing
   */
  public boolean isContentSigningRequired(Request<?> request) {
    // For delete requests, the body contains the actual objects being accessed. Without signing
    // the contents, it's possible for the client to change the body after being signed by this service.
    // Not signing the body is OK in case of AWS IAM based authorization - not in this case.

    AwsRequestHolder awsRequestHolder;
    try {
      awsRequestHolder = getAwsRequestHolder(request, false);
    } catch (RCSUncheckedException e) {
      throw new RCSAWSException(StatusCode.BAD_REQUEST, "Unable to interpret request", e);
    }
    ContentSigningRequired annotation = awsRequestHolder.getRequestClazz()
        .getAnnotation(ContentSigningRequired.class);
    if (annotation == null) {
      LOG.warn(
          "Expected" + ContentSigningRequired.class + " annotation not found on " + awsRequestHolder
              .getRequestClazz());
      throw new IllegalStateException(
          "Expected" + ContentSigningRequired.class + " annotation not found on " + awsRequestHolder
              .getRequestClazz());
    } else {
      return annotation.value();
    }
  }

  private AwsRequestHolder getAwsRequestHolder(Request<?> request, boolean instantiate)
      throws RCSUncheckedException {

    // This can likely be simplified by using request.getOriginalRequest and checking type
    Map<String, List<String>> parameters = request.getParameters();
    //Map<String, String> headers = request.getHeaders();
    String resourcePath = request.getResourcePath();

    switch (request.getHttpMethod()) {
      case GET:
        if (resourcePath.equals("")) {
          if (parameters.containsKey(S3Parameters.LOCATION)) {
            return new AwsRequestHolder(AwsGetBucketLocationRequest.class, request, instantiate);
          } else if (parameters.containsKey(S3Parameters.LIST_TYPE)) {
            return new AwsRequestHolder(AwsListV2Request.class, request, instantiate);
          } else if (parameters.containsKey(S3Parameters.UPLOADS)) {
            return new AwsRequestHolder(AwsMultiPartListUploadsRequest.class, request, instantiate);
          } else if (parameters.containsKey(S3Parameters.ACL)) {
            return new AwsRequestHolder(AwsGetBucketAclRequest.class, request, instantiate);
          } else {
            return new AwsRequestHolder(AwsListV1Request.class, request, instantiate);
          }
        } else if (parameters == null || parameters.isEmpty()) {
          // TODO: Will this be empty for encryption
          return new AwsRequestHolder(AwsReadObjectRequest.class, request, instantiate);
        } else if(parameters.containsKey(S3Parameters.UPLOAD_ID)) {
          return new AwsRequestHolder(AwsListPartsRequest.class, request, instantiate);
        } else {
          handleUnknownRequestType(request);
        }
        break;
      case POST:
        if (parameters.containsKey(S3Parameters.DELETE)) {
          return new AwsRequestHolder(AwsDeleteObjectsRequest.class, request, instantiate);
        } else if (parameters.containsKey(S3Parameters.UPLOADS)) {
          return new AwsRequestHolder(AwsMultipartInitiateRequest.class, request, instantiate);
        } else if (parameters.containsKey(S3Parameters.UPLOAD_ID)) {
          return new AwsRequestHolder(AwsMultipartCommitRequest.class, request, instantiate);
        } else if (parameters.containsKey(S3Parameters.SELECT) && parameters
            .containsKey(S3Parameters.SELECT_TYPE)) {
          return new AwsRequestHolder(AwsSelectObjectContentRequest.class, request, instantiate);
        } else {
          handleUnknownRequestType(request);
        }
        break;
      case PUT:
        if (parameters.containsKey(S3Parameters.UPLOAD_ID)) {
          return new AwsRequestHolder(AwsMultipartUploadRequest.class, request, instantiate);
        } else if (parameters == null || parameters.isEmpty()) {
          // TODO: Will this be empty for encryption
          return new AwsRequestHolder(AwsPutObjectRequest.class, request, instantiate);
        } else {
          handleUnknownRequestType(request);
        }
        break;
      case HEAD:
        if (resourcePath.equals("")) {
          return new AwsRequestHolder(AwsHeadBucketRequest.class, request, instantiate);
        } else {
          return new AwsRequestHolder(AwsHeadObjectRequest.class, request, instantiate);
        }
      case DELETE:
        if (parameters.containsKey(S3Parameters.UPLOAD_ID)) {
          return new AwsRequestHolder(AwsMultiPartAbortRequest.class, request, instantiate);
        } else {
          return new AwsRequestHolder(AwsDeleteObjectRequest.class, request, instantiate);
        } // Anything else with the DELETE operation?
      case PATCH:
      case OPTIONS:
      default:
        handleUnknownRequestType(request);
        break;
    }
    return null; // Will never happen.
  }

  // Private
  public byte[] getRequestContents(Request<?> request) throws IOException {
    InputStream inputStream = request.getContent();
    if (inputStream == null) {
      return null;
    }

    if (!inputStream.markSupported()) {
      AwsRequestHolder requestHolder = null;
      try {
        requestHolder = getAwsRequestHolder(request, false);
      } catch (RCSUncheckedException e) {
        // This may be hit for an unsupported operation. Ignore in this case.
        LOG.warn("Failed to extract AWS operation", e);
      }
      throw new IllegalStateException(
          "AWS SDK is supposed to be using streams which support mark. RequestType: "
              + (requestHolder == null ? "Unknown, see previous trace in log"
              : requestHolder.getRequestClazz()));
    }

    // TODO: Maybe set a limit on this.
    inputStream.mark(Integer.MAX_VALUE);
    byte[] bytes = IOUtils.toByteArray(inputStream);
    inputStream.reset();
    return bytes;
  }

  /*
  Handle unknown requests. Look for recent changes to S3A API usage in S3AFileSystem. File a bug.
   */
  private void handleUnknownRequestType(Request<?> request) {
    throw new RCSUncheckedException(Code.UNRECOGNIZED_REQUEST_TYPE, String.format(
        "Unknown request type. Method=[%s], Headers=[%s], Parameters=[%s], ResourcePath=[%s]",
        request.getHttpMethod(), request.getHeaders(), request.getParameters(),
        request.getResourcePath()));
  }


  private static class AwsRequestHolder {

    Class<? extends AwsRequest> requestClazz;
    AwsRequest awsRequest;

    public AwsRequestHolder(
        Class<? extends AwsRequest> requestClazz,
        Request<?> request, boolean instantiate) {
      this.requestClazz = requestClazz;
      if (instantiate) {
        try {
          LOG.debug("AwsRequestHolder instantiate class: {}", requestClazz.getName());
          Method method = requestClazz.getMethod(CREATE_METHOD_NAME, Request.class);
          this.awsRequest = (AwsRequest) method.invoke(null, request);
        } catch (NoSuchMethodException | IllegalAccessException | InvocationTargetException e) {
          throw new RCSUncheckedException(Code.GENERAL_SERVER_ERROR,
              "Unable to instantiate: [" + requestClazz + "]", e);
        }
      }
    }

    public Class<? extends AwsRequest> getRequestClazz() {
      return requestClazz;
    }

    public AwsRequest getAwsRequest() {
      return awsRequest;
    }
  }
}
