/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.aws.parsers;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedList;
import java.util.List;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

public class RequestBodyParser {


  public DeleteObjects parseDeleteObjectsRequest(InputStream is) {
    SAXParser parser = getParser();
    DeleteObjectsHandler deleteHandler = new DeleteObjectsHandler();
    try {
      parser.parse(is, deleteHandler);
    } catch (SAXException | IOException e) {
      throw new RuntimeException("Failed to parse Delete request", e);
    }
    return deleteHandler.getDeleteObjects();
  }


  private SAXParser getParser() {
    SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();
    try {
      return saxParserFactory.newSAXParser();
    } catch (ParserConfigurationException | SAXException e) {
      throw new RuntimeException("Failed to create SAXParser", e);
    }
  }


  static class DeleteObjectsHandler extends DefaultHandler {

    private DeleteObjects deleteObjects = null;

    private DeleteObjectEntity currentDeleteObject = null;
    private boolean q = false;
    private boolean k = false;
    private boolean v = false;

    public DeleteObjects getDeleteObjects() {
      return deleteObjects;
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes)
        throws SAXException {
      if (qName.equalsIgnoreCase("delete")) {
        deleteObjects = new DeleteObjects();
      } else if (qName.equalsIgnoreCase("quiet")) {
        q = true;
      } else if (qName.equalsIgnoreCase("object")) {
        currentDeleteObject = new DeleteObjectEntity();
      } else if (qName.equalsIgnoreCase("key")) {
        k = true;
      } else if (qName.equalsIgnoreCase("versionid")) {
        v = true;
      }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {

      if (q == true) {
        q = false;
        String quietString = new String(ch, start, length);
        boolean quiet = Boolean.valueOf(quietString.toLowerCase());
        deleteObjects.setQuiet(quiet);
      } else if (k == true) {
        k = false;
        String deleteKey = new String(ch, start, length);
        currentDeleteObject.setKey(deleteKey);
      } else if (v == true) {
        v = false;
        String versionIdString = new String(ch, start, length);
        currentDeleteObject.setVersionId(versionIdString);
      }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
      if (qName.equalsIgnoreCase("object")) {
        deleteObjects.addDeleteEntity(currentDeleteObject);
        currentDeleteObject = null;
      }
    }
  }


  public static class DeleteObjects {

    private boolean quiet = false;
    private List<DeleteObjectEntity> deleteEntities = new LinkedList<>();

    public boolean isQuiet() {
      return quiet;
    }

    public void setQuiet(boolean quiet) {
      this.quiet = quiet;
    }

    public List<DeleteObjectEntity> getDeleteEntities() {
      return deleteEntities;
    }

    public void addDeleteEntity(DeleteObjectEntity deleteEntity) {
      deleteEntities.add(deleteEntity);
    }

    @Override
    public String toString() {
      return "DeleteObjects{" +
          "quiet=" + quiet +
          ", deleteEntities=" + deleteEntities +
          '}';
    }
  }

  public static class DeleteObjectEntity {

    private String key;
    private String versionId;

    public String getKey() {
      return key;
    }

    public void setKey(String key) {
      this.key = key;
    }

    public String getVersionId() {
      return versionId;
    }

    public void setVersionId(String versionId) {
      this.versionId = versionId;
    }

    @Override
    public String toString() {
      return "DeleteObjectEntity{" +
          "key='" + key + '\'' +
          ", versionId='" + versionId + '\'' +
          '}';
    }
  }

}
