/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.perflogger;

import org.apache.ranger.raz.s3.lib.perflogger.PerfLogger.PerfLoggerInterface;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Actual implementation of the perf tracker
 */
public class PerfLoggerImpl implements PerfLoggerInterface {

  private static final Logger LOG = LoggerFactory.getLogger(PerfLoggerImpl.class);

  private static final Logger PERFLOG = LoggerFactory.getLogger("RCS.PERFLOG");

  private PerfLoggerImpl() {
    beginTimes = new ConcurrentHashMap<>();
  }

  private enum Instance {
    INSTANCE(new PerfLoggerImpl());

    private final PerfLoggerInterface perfTracker;

    Instance(PerfLoggerInterface perfTracker) {
      this.perfTracker = perfTracker;
    }
  }

  public static PerfLoggerInterface get() {
    return Instance.INSTANCE.perfTracker;
  }

  private final ConcurrentHashMap<Key, Long> beginTimes;

  @Override
  public void opBegin(String context, String opName) {
    Key key = new Key(context, opName);
    long beginTime = System.nanoTime();
    Long old = beginTimes.put(key, beginTime);
    if (old != null) {
      LOG.debug("Duplicate beginOp for key={}", key);
    }
    PERFLOG.info("PERFLOG BEGIN: context=[{}], opName=[{}], beginTime=[{}]", context, opName,
        beginTime);
  }

  // TODO Add support to indicate whether the operation completely successfully or not.
  //  Also add an Invoker which can replace the generic try { BEGIN. OPERATION} finally {END}
  //  block
  @Override
  public long opEnd(String context, String opName) {
    Key key = new Key(context, opName);
    Long beginTime = beginTimes.remove(key);
    if (beginTime == null) {
      LOG.debug("No begin information found for key: {}", key);
    }
    long endTime = System.nanoTime();
    long duration = (endTime - beginTime) / 1_000_000L;
    PERFLOG
        .info("PERFLOG END: context=[{}], opName=[{}], duration=[{}], beginTime=[{}], endTime=[{}]",
            context, opName, duration, beginTime, endTime);
    return duration;
  }

  private static class Key {

    private String context;
    private String opName;

    public Key(String context, String opName) {
      this.context = context;
      this.opName = opName;
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) {
        return true;
      }
      if (o == null || getClass() != o.getClass()) {
        return false;
      }
      Key key = (Key) o;
      return Objects.equals(context, key.context) &&
          Objects.equals(opName, key.opName);
    }

    @Override
    public int hashCode() {
      return Objects.hash(context, opName);
    }

    @Override
    public String toString() {
      return "Key{" +
          "context='" + context + '\'' +
          ", opName='" + opName + '\'' +
          '}';
    }
  }
}
