/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.signer;

import com.google.common.base.Preconditions;
import java.util.Arrays;
import java.util.Map;

public interface SignResponse {

  /**
   * Get the authorization response including information like whether the request was allowed,
   * denied, etc.
   *
   * @return the authorization response
   */
  AuthorizationResponse getAuthorizationResponse();

  /**
   * @return signed headers, which need to be included as part of the request
   */
  Map<String, String> getSignerGeneratedHeaders();

  /**
   * @return the region for the specific request. This value can be null if the region is not
   * determined.
   */
  String getRegion();


  class AuthorizationResponse {

    private final ResponseCode responseCode;
    private final String[] unauthorizedPaths;
    private final String responseMessage;

    public enum ResponseCode {
      AUTHZ_AUTHORIZED,
      AUTHZ_UNAUTHORIZED,
      AUTHZ_UNKNOWN,
      ERROR
    }

    public static AuthorizationResponse AUTHORIZED_INSTANCE = new AuthorizationResponse(
        ResponseCode.AUTHZ_AUTHORIZED, null, null);

    /**
     * Return the default AUTHORIZED response.
     */
    public static AuthorizationResponse getAuthorizedInstance() {
      return AUTHORIZED_INSTANCE;
    }

    public AuthorizationResponse(
        ResponseCode responseCode, String[] unauthorizedPaths, String responseMessage) {
      Preconditions.checkNotNull(responseCode, "ResponseCode must be set");
      this.responseCode = responseCode;
      this.unauthorizedPaths = unauthorizedPaths;
      this.responseMessage = responseMessage;
    }

    /**
     * Get the response code from the authorization request
     *
     * @return the ResponseCode
     */
    public ResponseCode getResponseCode() {
      return this.responseCode;
    }

    /**
     * Get the response message from the authorization request. This will typically be populated for
     * UNAUTHROZED requests and ERRORs
     *
     * @return A message containing human readable debug information related to the response
     */
    public String getResponseMessage() {
      return this.responseMessage;
    }

    /**
     * Get the paths which failed the authorization check
     */
    public String[] getUnauthorizedPaths() {
      return unauthorizedPaths;
    }

    @Override
    public String toString() {
      return "AuthorizationStatus{" +
          "responseCode=" + responseCode +
          ", unauthorizedPaths=" + Arrays.toString(unauthorizedPaths) +
          ", responseMessage='" + responseMessage + '\'' +
          '}';
    }
  }
}
