/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.signer.impl;

import org.apache.ranger.raz.s3.lib.signer.SignResponse;

import com.google.common.base.Preconditions;

import java.util.Collections;
import java.util.Map;

public class SignResponseImpl implements SignResponse {

  private final AuthorizationResponse authorizationResponse;
  private final Map<String, String> signerGeneratedHeaders;
  private final String region;

  public SignResponseImpl(AuthorizationResponse authorizationResponse,
      Map<String, String> signerGeneratedHeaders, String region) {
    // TODO region is not used here?
    validateResponseParameters(authorizationResponse, signerGeneratedHeaders);

    this.authorizationResponse = authorizationResponse;
    this.signerGeneratedHeaders = signerGeneratedHeaders;
    this.region = region;
  }

  @Override
  public AuthorizationResponse getAuthorizationResponse() {
    return authorizationResponse;
  }

  @Override
  public Map<String, String> getSignerGeneratedHeaders() {
    return signerGeneratedHeaders == null ? Collections.EMPTY_MAP : signerGeneratedHeaders;
  }

  @Override
  public String getRegion() {
    return region;
  }

  @Override
  public String toString() {
    return "SignResponseImpl{" +
        "authorizationResponse=" + authorizationResponse +
        ", signerGeneratedHeaders=" + signerGeneratedHeaders +
        ", region='" + region + '\'' +
        '}';
  }

  private void validateResponseParameters(AuthorizationResponse authorizationResponse,
      Map<String, String> signerGeneratedHeaders) {
    Preconditions.checkNotNull(authorizationResponse, "AuthorizationStatus cannot be empty");
    switch (authorizationResponse.getResponseCode()) {

      case AUTHZ_AUTHORIZED:
        // TODO: This will need to change when supporting GCS, along with the ProtoConverter tests
        Preconditions
            .checkArgument(signerGeneratedHeaders != null && signerGeneratedHeaders.size() != 0,
                "Signer generated headers must be populated for Authorized requests");
        break;
      case AUTHZ_UNAUTHORIZED:
      case AUTHZ_UNKNOWN:
      case ERROR:
      default:
        Preconditions.checkState(signerGeneratedHeaders == null,
            "Signer generated headers should not be present for ResponseCode: "
                + authorizationResponse.getResponseCode());
        break;
    }
  }
}
