/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.signer.util;

import org.apache.ranger.raz.s3.lib.HttpMethod;
import org.apache.ranger.raz.s3.lib.signer.SignRequest;
import org.apache.ranger.raz.s3.lib.signer.SignResponse;
import org.apache.ranger.raz.s3.lib.signer.SignResponse.AuthorizationResponse;
import org.apache.ranger.raz.s3.lib.signer.SignResponse.AuthorizationResponse.ResponseCode;
import org.apache.ranger.raz.s3.lib.signer.impl.SignRequestImpl;
import org.apache.ranger.raz.s3.lib.signer.impl.SignResponseImpl;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.AuthorizationResponseProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.HttpMethodProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.ResponseCodeProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.SignRequestProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.SignResponseProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.StringListStringMapProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.StringStringMapProto;
import com.google.common.base.Preconditions;
import com.google.protobuf.ByteString;
import com.google.protobuf.CodedOutputStream;

import org.apache.commons.lang3.ArrayUtils;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.Base64;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

// Null and empty checks are placed based on which fields could be null.
public class ProtoConverters {

  private static final String CONST_NULL = "_NULL_";

  public static HttpMethod fromProto(HttpMethodProto proto) {
    return HttpMethod.valueOf(proto.name());
  }

  public static HttpMethodProto toProto(HttpMethod arg) {
    return HttpMethodProto.valueOf(arg.name());
  }

  public static ResponseCode fromProto(ResponseCodeProto proto) {
    return ResponseCode.valueOf(proto.name());
  }

  public static ResponseCodeProto toProto(ResponseCode arg) {
    return ResponseCodeProto.valueOf(arg.name());
  }

  public static AuthorizationResponse fromProto(AuthorizationResponseProto proto) {
    AuthorizationResponse authorizationResponse = new AuthorizationResponse(
        fromProto(proto.getResponseCode()),
        proto.getUnauthorizedPathsCount() == 0 ? null : proto.getUnauthorizedPathsList()
            .toArray(new String[proto.getUnauthorizedPathsCount()]),
        proto.hasResponseMessage() ? proto.getResponseMessage() : null);
    return authorizationResponse;
  }

  public static AuthorizationResponseProto toProto(AuthorizationResponse arg) {
    AuthorizationResponseProto.Builder builder = AuthorizationResponseProto.newBuilder();
    builder.setResponseCode(toProto(arg.getResponseCode()));
    if (ArrayUtils.isNotEmpty(arg.getUnauthorizedPaths())) {
      builder.addAllUnauthorizedPaths(Arrays.asList(arg.getUnauthorizedPaths()));
    }
    if (arg.getResponseMessage() != null) { // Allow empty messages
      builder.setResponseMessage(arg.getResponseMessage());
    }
    return builder.build();
  }

  // At the moment, assuming maps / headers will not have NULL values.
  // If that assumption changes, this will be similar to toProtoParams and fromProtoParams
  public static Map<String, String> fromProto(List<StringStringMapProto> list) {
    if (list == null) {
      return null;
    }
    Map<String, String> map = new HashMap<>();
    list.forEach((l) -> map.put(l.getKey(), l.getValue()));
    return map;
  }

  public static List<StringStringMapProto> toProto(Map<String, String> arg) {
    // This could be implemented as an iterator over the map. Doesn't save on too much other than
    // the list object itself. The StringStringMapProto instances will still be created.
    Preconditions.checkState(!arg.isEmpty());
    List<StringStringMapProto> list = new LinkedList<>();
    arg.forEach(
        (k, v) -> list.add(StringStringMapProto.newBuilder().setKey(k).setValue(v).build()));
    return list;
  }

  public static Map<String, List<String>> fromProtoParams(List<StringListStringMapProto> list) {
    if (list == null) {
      return null;
    }
    Map<String, List<String>> map = new HashMap<>(list.size());

    for (StringListStringMapProto p : list) {
      if (p.getValueCount() == 0) {
        map.put(p.getKey(), null);
      } else {
        if (p.getValueList().contains(CONST_NULL)) {
          List<String> newList = new LinkedList<>();
          for (String s : p.getValueList()) {
            if (s.equals(CONST_NULL)) {
              newList.add(null);
            } else {
              newList.add(s);
            }
          }
          map.put(p.getKey(), newList);
        } else {
          map.put(p.getKey(), p.getValueList());
        }
      }
    }
    return map;
  }

  public static List<StringListStringMapProto> toProtoParams(Map<String, List<String>> arg) {
    Preconditions.checkState(!arg.isEmpty());

    List<StringListStringMapProto> list = new LinkedList<>();
    for (Entry<String, List<String>> e : arg.entrySet()) {
      StringListStringMapProto.Builder builder = StringListStringMapProto.newBuilder();
      builder.setKey(e.getKey());
      if (e.getValue() != null) {
        if (e.getValue().contains(null)) {
          for (String s : e.getValue()) {
            if (s == null) {
              // Protobuf doesn't like null values anywhere. AWS SDK seems to like putting a null value. e.g. delete=null
              builder.addValue(CONST_NULL);
            } else {
              builder.addValue(s);
            }
          }
        } else {
          builder.addAllValue(e.getValue());
        }
      }
      list.add(builder.build());
    }

    return list;
  }

  public static byte[] fromProto(ByteString byteString) {
    if (byteString == null || byteString.size() == 0) {
      return null;
    }
    return byteString.toByteArray();
  }

  public static ByteString toProto(byte[] args) {
    if (args == null) {
      return ByteString.EMPTY;
    }
    return ByteString.copyFrom(args);
  }

  public static SignRequest fromProto(SignRequestProto proto) {
    URI uri;
    try {
      uri = new URI(proto.getEndpoint());
    } catch (URISyntaxException e) {
      throw new RuntimeException(e);
    }
    SignRequestImpl signRequest = new SignRequestImpl(
        uri,
        fromProto(proto.getHttpMethod()),
        proto.getHeadersCount() == 0 ? null : fromProto(proto.getHeadersList()),
        proto.getParametersCount() == 0 ? null : fromProtoParams(proto.getParametersList()),
        proto.getResourcePath(),
        proto.hasContentToSign() ? fromProto(proto.getContentToSign()) : null,
        proto.getTimeOffset(),
        proto.getEndpointPrefix(),
        proto.getServiceName()
    );
    return signRequest;
  }


  public static SignRequestProto toProto(SignRequest signRequest) {
    SignRequestProto.Builder builder = SignRequestProto.newBuilder();
    builder.setEndpoint(signRequest.getEndpoint().toString());
    builder.setHttpMethod(toProto(signRequest.getHttpMethod()));
    if (!signRequest.getHeaders().isEmpty()) {
      builder.addAllHeaders(toProto(signRequest.getHeaders()));
    }
    if (!signRequest.getParameters().isEmpty()) {
      builder.addAllParameters(toProtoParams(signRequest.getParameters()));
    }
    builder.setResourcePath(signRequest.getResourcePath());
    if (signRequest.getContentToSign() != null) {
      builder.setContentToSign(toProto(signRequest.getContentToSign()));
    }
    builder.setTimeOffset(signRequest.getTimeOffset());
    if (signRequest.getEndpointPrefix() != null) {
      builder.setEndpointPrefix(signRequest.getEndpointPrefix());
    }
    if (signRequest.getServiceName() != null) {
      builder.setServiceName(signRequest.getServiceName());
    }
    return builder.build();
  }

  public static SignResponse fromProto(SignResponseProto proto) {

    AuthorizationResponse authorizationResponse = fromProto(proto.getAuthorizationResponse());
    Map<String, String> signedHeaders = proto.getSignerGeneratedHeadersCount() == 0 ? null
        : fromProto(proto.getSignerGeneratedHeadersList());

    return new SignResponseImpl(authorizationResponse, signedHeaders,
        proto.hasRegion() ? proto.getRegion() : null);
  }

  public static SignResponseProto toProto(SignResponse arg) {
    SignResponseProto.Builder builder = SignResponseProto.newBuilder();
    builder.setAuthorizationResponse(toProto(arg.getAuthorizationResponse()));
    if (!arg.getSignerGeneratedHeaders().isEmpty()) {
      builder.addAllSignerGeneratedHeaders(toProto(arg.getSignerGeneratedHeaders()));
    }
    if (arg.getRegion() != null) {
      builder.setRegion(arg.getRegion());
    }
    return builder.build();
  }

  public static String toBase64(SignResponse signResponse) throws IOException {
    final SignResponseProto srp = ProtoConverters.toProto(signResponse);
    final ByteArrayOutputStream bArrOs = new ByteArrayOutputStream();
    final CodedOutputStream stream = CodedOutputStream.newInstance(bArrOs);
    srp.writeTo(stream);
    stream.flush();
    return Base64.getEncoder().encodeToString(bArrOs.toByteArray());
  }

  public static String toBase64(SignRequest signRequest) throws IOException {
    final SignRequestProto srp = ProtoConverters.toProto(signRequest);
    final ByteArrayOutputStream bArrOs = new ByteArrayOutputStream();
    final CodedOutputStream stream = CodedOutputStream.newInstance(bArrOs);
    srp.writeTo(stream);
    stream.flush();
    return Base64.getEncoder().encodeToString(bArrOs.toByteArray());
  }

  public static SignRequest reqFromBase64(String srpString) throws IOException {
    final byte[] responseProtoBarr = Base64.getDecoder().decode(srpString);
    final SignRequestProto srp = SignRequestProto.parseFrom(responseProtoBarr);
    return fromProto(srp);
  }

  public static SignResponse respFromBase64(String srpString) throws IOException {
    final byte[] responseProtoBarr = Base64.getDecoder().decode(srpString);
    final SignResponseProto srp = SignResponseProto.parseFrom(
        responseProtoBarr);
    return fromProto(srp);
  }
}
