/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.utils;

import com.amazonaws.AmazonServiceException;
import com.amazonaws.ClientConfiguration;
import com.amazonaws.SdkClientException;
import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilder;
import com.google.common.base.Preconditions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

public class S3Utils {

  private static final Logger LOG = LoggerFactory.getLogger(S3Utils.class);

  private static final int BUCKET_REGION_CACHE_SIZE = 300;
  private static final Map<String, String> bucketRegionCache =
      Collections
          .synchronizedMap(new LinkedHashMap<String, String>(BUCKET_REGION_CACHE_SIZE, 1.1f, true) {
            private static final long serialVersionUID = 23453L;

            @Override
            protected boolean removeEldestEntry(Map.Entry<String, String> eldest) {
              return size() > BUCKET_REGION_CACHE_SIZE;
            }
          });

  /**
   * Look up the region for a bucket from the cache. If not present, look up the region via an S3 call,
   * and updates the cache with the result.
   * @param bucketName  the bucket name
   * @param credentialsProvider credentials for an S3 invocation
   * @param awsConf aws conf used to create S3 client.
   * @return
   */
  // TODO Change this over to an AWSCredentialChain, as well as usage.
  public static String lookupRegion(String bucketName,
                                    AWSCredentialsProvider credentialsProvider,
                                    ClientConfiguration awsConf) throws SdkClientException, AmazonServiceException {

    // The Amazon client special cases the s3v4signer to update the region. That functionality
    // is not available for custom signers, without modifying the amazon java sdk - which may be a possibility.
    // The mechanism used to determine the region there is by issuing a HEAD request and handling errors.
    // TODO - check if that mechanism is more efficient, and why they did not use a getLocation call.
    // The GetLocation call requires allow global regions to be enabled. It makes an explicit call to
    // a specific region for this though, and ideally that should be the closest location.
    // TODO: Allow the region for bucket lookup to be configurable.

    AmazonS3 s3 = AmazonS3ClientBuilder.standard()
            .withCredentials(credentialsProvider)
            .withForceGlobalBucketAccessEnabled(true)
            .withClientConfiguration(awsConf)
            .withRegion("us-east-1").build();

    String regionName = s3.getBucketLocation(bucketName);
    //  See: https://forums.aws.amazon.com/thread.jspa?messageID=796829&tstart=0
    if (regionName.equals("US")) {
      regionName = "us-east-1";
    }
    bucketRegionCache.put(bucketName, regionName);
    LOG.info("Region for bucket {}={}", bucketName, regionName);
    return regionName;
  }

  /**
   * Lookup the bucket name from the hostname
   * @param host the S3 hostname
   * @return
   */
  // TODO: Some of this will not work for Glacier.
  // TODO Are there Amazon utility classes for this?
  public static String getBucketNameFromHost(String host) {
    // Assuming format: <bucket_name>.s3.<region_name>.amazonaws.com
    // OR <bucket_name>.s3.amazonaws.com
    String[] parts = host.split("\\.");
    if (parts.length > 0) {
      return parts[0];
    } else {
      return null;
    }
  }


  // The following methods are commented out to avoid confusion caused by AWS not having an explicit
  // region in the hostname for us-east-1 requests. The region must be looked up via a remote call,
  // and then cached.
//  /**
//   * Try looking up the region from the bucketRegionCache or the hostname. Updates the bucketRegion
//   * cache in the process.
//   * @param host S3 hostname. Must be non null
//   * @param bucketName The bucket name. Can be null
//   */
//  public static String lookupRegion(String host, String bucketName) {
//    Preconditions.checkNotNull(host, "hostname must be provided");
//
//    String region;
//    if (bucketName == null) {
//      bucketName = getBucketNameFromHost(host);
//    }
//
//    // BucketName provided or looked up. Check in the cache.
//    region = bucketRegionCache.get(bucketName);
//    if (region != null) {
//      return region;
//    }
//
//    // Bucket not found in cache. Try looking up the region from the hostname, and update the cache if found.
//    region = getBucketRegionExplicitFromHost(host);
//    if (region != null) {
//      bucketRegionCache.put(bucketName, region);
//      return region;
//    }
//    return null;
//  }
//
//  /**
//   * Lookup the bucket region from the hostname. Returns null if the hostname doesn't explicitly
//   * mention a region (this would otherwise default to us-east-1.
//   */
//  private static String getBucketRegionExplicitFromHost(String host) {
//    // Assuming format: <bucket_name>.s3.<region_name>.amazonaws.com
//    String[] parts = host.split("\\.");
//    if (parts.length == 5) {
//      return parts[2];
//    } else {
//      return null;
//    }
//    // The following would return the region as us-east-1 if the hostname does not contain a region.
//    // return AwsHostNameUtils.parseRegion(host, null);
//  }

  /**
   * Looks up the region for a bucket from the cache
   * @param bucketName  the bucket name.
   * @return  the region for the bucket, or null if not found
   */
  public static String lookupRegionFromCache(String bucketName) {
    Preconditions.checkNotNull(bucketName, "BucketName must be provided");
    return bucketRegionCache.get(bucketName);
  }

  /**
   * Updates the bucket to region cache
   * @param bucketName  the bucket name
   * @param regionName  the associated region name
   */
  // TODO: This is not the best way to implement this, since the cache can be corrupted by
  // any user code. Ideally make the cache an instance.
  public static void updateBucketToRegionCache(String bucketName, String regionName) {
    Preconditions.checkNotNull(bucketName, "BucketName must be specified");
    Preconditions.checkNotNull(regionName, "RegionName must be specified");
    bucketRegionCache.put(bucketName, regionName);
  }

}