/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package com.amazonaws.services.s3;

import com.amazonaws.client.AwsSyncClientParams;
import com.amazonaws.internal.SdkFunction;
import com.amazonaws.regions.AwsRegionProvider;
import com.amazonaws.regions.DefaultAwsRegionProviderChain;

/**
 * Not meant for everyday usage.
 *
 * This is part of a series of specialized classes used to extract the {@link
 * com.amazonaws.SignableRequest} generated by the AWS SDK.
 *
 * The underlying AmazonHttpClient is replaced with a custom client, which throws an instance of
 * {@link RequestProvidingException} which contains the {@link com.amazonaws.Request} object
 * generated by the AmazonS3Client.
 *
 * This can be fairly brittle and is dependent on AWS SDK internals.
 */
public class AmazonS3ClientBuilderForRequestInterception extends
    AmazonS3Builder<AmazonS3ClientBuilderForRequestInterception, AmazonS3> {

  private static final SdkFunction<AmazonS3ClientParamsWrapper, AmazonS3> DEFAULT_CLIENT_FACTORY =
      new SdkFunction<AmazonS3ClientParamsWrapper, AmazonS3>() {
        @Override
        public AmazonS3 apply(AmazonS3ClientParamsWrapper params) {
          return new AmazonS3ClientForRequestInterception(params);
        }
      };

  private static final AmazonS3ClientConfigurationFactory CLIENT_CONFIG_FACTORY =
      new AmazonS3ClientConfigurationFactory();

  private static final AwsRegionProvider DEFAULT_REGION_PROVIDER =
      new DefaultAwsRegionProviderChain();

  private AmazonS3ClientBuilderForRequestInterception() {
    super(DEFAULT_CLIENT_FACTORY, CLIENT_CONFIG_FACTORY, DEFAULT_REGION_PROVIDER);
  }

  /**
   * @return Create new instance of builder with all defaults set.
   */
  public static AmazonS3ClientBuilderForRequestInterception standard() {
    return new AmazonS3ClientBuilderForRequestInterception()
        .withCredentials(new S3CredentialsProviderChain());
  }

  /**
   * @return Default client using the {@link com.amazonaws.regions.DefaultAwsRegionProviderChain}
   * chain.
   */
  public static AmazonS3 defaultClient() {
    return standard().build();
  }

  /**
   * Construct a synchronous implementation of AmazonS3 using the current builder configuration.
   *
   * @return Fully configured implementation of AmazonS3.
   */
  @Override
  protected AmazonS3 build(AwsSyncClientParams clientParams) {
    return clientFactory
        .apply(new AmazonS3ClientParamsWrapper(clientParams, resolveS3ClientOptions()));
  }
}
