/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.aws.parsers;

import com.amazonaws.Request;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.*;
import org.assertj.core.api.Assertions;
import org.junit.BeforeClass;
import org.junit.Test;

import org.apache.ranger.raz.s3.lib.aws.util.AwsRequestGenerator;

import java.util.HashSet;
import java.util.Set;

public class TestAwsRequestParser {

  private static final String FAKE_BUCKET_NAME = "fakebucketname";
  private static final String FAKE_PATH = "path/under/test";
  private static final String REGION_NAME = "us-west-2";

  private final AwsRequestParser requestParser = new AwsRequestParser();
  private final AwsRequestGenerator requestGenerator = new AwsRequestGenerator(REGION_NAME);

  private static final Set<AwsOperationType> seenAwsRequestTypes = new HashSet<>();

  // TODO Maybe add additional validation like
  //  - expected authorization
  //  - headers, parameters
  //  - httpMethodName

  @BeforeClass
  public static void beforeClass() {
    seenAwsRequestTypes.clear();
  }

  @Test(timeout = 10_000L)
  public void testHeadBucketRequest() {
    Request<?> request = requestGenerator.generateHeadBucketRequest(FAKE_BUCKET_NAME);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsHeadBucketRequest.class, "/");
  }

  @Test(timeout = 10_000L)
  public void testGetBucketLocationRequest() {
    Request<?> request = requestGenerator.generateGetBucketLocationRequest(FAKE_BUCKET_NAME);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsGetBucketLocationRequest.class, "/");
  }

  @Test(timeout = 10_000L)
  public void testGetBucketAclRequest() {
    Request<?> request = requestGenerator.generateGetBucketAclRequest(FAKE_BUCKET_NAME);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsGetBucketAclRequest.class, "/");
  }

  @Test(timeout = 10_000L)
  public void testGetObjectMetadataRequest() {
    Request<?> request = requestGenerator
        .generateGetObjectMetadataRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsHeadObjectRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testListObjectsV1Request() {
    Request<?> request = requestGenerator.generateListObjectsV1Request(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsListV1Request.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testListRootBucketNullPrefixV1Request() {
    final String rootPathPrefix = null;
    Request<?> request = requestGenerator.generateListObjectsV1Request(FAKE_BUCKET_NAME, rootPathPrefix);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);
    verifyRequestBasic(request, awsRequest, AwsListV1Request.class, "/");
  }

  @Test(timeout = 10_000L)
  public void testListRootBucketEmptyPrefixV1Request() {
    final String rootPathPrefix = "";
    Request<?> request = requestGenerator.generateListObjectsV1Request(FAKE_BUCKET_NAME, rootPathPrefix);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);
    verifyRequestBasic(request, awsRequest, AwsListV1Request.class, rootPathPrefix);
  }

  @Test(timeout = 10_000L)
  public void testListObjectsV2Request() {
    Request<?> request = requestGenerator.generateListObjectsV2Request(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsListV2Request.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testListRootBucketNullPrefixV2Request() {
    final String rootPathPrefix = null;
    Request<?> request = requestGenerator.generateListObjectsV2Request(FAKE_BUCKET_NAME, rootPathPrefix);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);
    verifyRequestBasic(request, awsRequest, AwsListV2Request.class, "/");
  }

  @Test(timeout = 10_000L)
  public void testListRootBucketEmptyPrefixV2Request() {
    final String rootPathPrefix = "";
    Request<?> request = requestGenerator.generateListObjectsV2Request(FAKE_BUCKET_NAME, rootPathPrefix);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);
    verifyRequestBasic(request, awsRequest, AwsListV2Request.class, rootPathPrefix);
  }


  @Test(timeout = 10_000L)
  public void testDeleteObjectRequest() {
    Request<?> request = requestGenerator.generateDeleteObjectRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsDeleteObjectRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testDeleteObjectsRequest() {
    String[] paths = new String[]{FAKE_PATH, "anotherPath", "yet/another/path"};
    Request<?> request = requestGenerator.generateDeleteObjectsRequest(FAKE_BUCKET_NAME, paths);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsDeleteObjectsRequest.class, paths);
  }

  /**
   * This is a fix for S3A sending empty content with delete when deleting
   * directory markers during the cleanup
   */
  @Test(timeout = 10_000L)
  public void testDeleteObjectsRequestNullContent() {
    String[] paths = new String[]{};
    Request<?> request = requestGenerator.generateDeleteObjectsRequest(FAKE_BUCKET_NAME, paths);
    request.setContent(null);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsDeleteObjectsRequest.class, paths);
  }

  @Test(timeout = 10_000L)
  public void testPutObjectRequest() {
    Request<?> request = requestGenerator.generatePutObjectRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsPutObjectRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testMultipartUploadPartRequest() {
    Request<?> request = requestGenerator.generateUploadPartRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsMultipartUploadRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testMultipartListRequest() {
    Request<?> request = requestGenerator
        .generateListMultipartUploadRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsMultiPartListUploadsRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testListPartsRequest() {
    Request<?> request = requestGenerator
            .generateListPartsUploadRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsListPartsRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testMultipartInitiateRequest() {
    Request<?> request = requestGenerator
        .generateInitiateMultipartUploadRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsMultipartInitiateRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testMultipartAbortRequest() {
    Request<?> request = requestGenerator
        .generateAbortMultipartUploadRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsMultiPartAbortRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testMultipartCommitRequest() {
    Request<?> request = requestGenerator
        .generateMultipartCommitRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsMultipartCommitRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testGetObjectRequest() {
    Request<?> request = requestGenerator.generateGetObjectRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsReadObjectRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testSelectObjectContentRequest() {
    Request<?> request = requestGenerator
        .generateSelectObjectContentRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);

    verifyRequestBasic(request, awsRequest, AwsSelectObjectContentRequest.class, FAKE_PATH);
  }

  @Test(timeout = 10_000L)
  public void testAwsRequestCachingAllowed() {
    Request<?> request = requestGenerator.generateGetObjectRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);
    Assertions.assertThat(awsRequest.isSignatureCachingAllowed())
            .as("Signature caching should be allowed for GetObjectRequest")
            .isTrue();
  }

  @Test(timeout = 10_000L)
  public void testAwsRequestCachingNotAllowed() {
    Request<?> request = requestGenerator.generatePutObjectRequest(FAKE_BUCKET_NAME, FAKE_PATH);
    AwsRequest awsRequest = requestParser.parseAwsRequest(request);
    Assertions.assertThat(awsRequest.isSignatureCachingAllowed())
            .as("Signature caching should be only be allowed for GetObjectRequest")
            .isFalse();
  }

  private void verifyRequestBasic(Request<?> request, AwsRequest awsRequest,
      Class<? extends AwsRequest> expectedClazz, String expectedResourcePath) {
    verifyRequestBasic(request, awsRequest, expectedClazz, new String[]{expectedResourcePath});
  }

  private void verifyRequestBasic(Request<?> request, AwsRequest awsRequest,
      Class<? extends AwsRequest> expectedClazz, String[] expectedResourcePaths) {
    if(request == null) {
      // TODO do we need the request here?
      request = null;
    }
    for (int i = 0; i < expectedResourcePaths.length; i++) {
      expectedResourcePaths[i] = (expectedResourcePaths[i].startsWith("/")
          ? expectedResourcePaths[i]
          : "/" + expectedResourcePaths[i]);
    }

    Assertions.assertThat(awsRequest).as("Type mismatch").isInstanceOf(expectedClazz);

    Assertions.assertThat(awsRequest.getBucket()).as("Bucket name mismatch")
        .isEqualTo(FAKE_BUCKET_NAME);

    Assertions.assertThat(awsRequest.getEndpoint().toString()).as("Endpoint mismatch")
        .isEqualTo(getExpectedEndpoint());

    Assertions.assertThat(awsRequest.getPathsToAuthorize()).as("Incorrect paths to authorize")
        .containsExactlyInAnyOrder(expectedResourcePaths);

    seenAwsRequestTypes.add(awsRequest.getOperationType());
  }


  private String getExpectedEndpoint() {
    return ("https://" + FAKE_BUCKET_NAME + ".s3." + REGION_NAME + ".amazonaws.com");
  }

}
