/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.aws.util;

import com.amazonaws.ClientConfiguration;
import com.amazonaws.Request;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilderForRequestInterception;
import com.amazonaws.services.s3.RequestProvidingException;
import com.amazonaws.services.s3.model.*;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.Random;
import java.util.UUID;
import javax.annotation.Nullable;

/**
 * Utility class to generate {@link Request} instances for various operations. Intended to be used
 * in testing.
 */
// Private
public class AwsRequestGenerator {

  private static final ByteArrayInputStream EMPTY_INPUT_STREAM = new ByteArrayInputStream(
      new byte[0]);

  private final AmazonS3 s3Client;
  private final Random random = new Random();

  public AwsRequestGenerator(@Nullable String region) {
    region = (region == null ? "us-east-1" : region);
    s3Client = AmazonS3ClientBuilderForRequestInterception.standard().withRegion(region)
        .withClientConfiguration(new ClientConfiguration())
        // Path style vs host style access can cause havoc with the authorizer.
        // Path style: 'getResourcePath' -> 'bucketName/path'
        // Host style: 'getResourcePath' -> 'path'
        //.withPathStyleAccessEnabled(true)
        .build();
  }

  /**
   * @param bucketName the bucket name.
   */
  public Request<?> generateHeadBucketRequest(String bucketName) {
    return invoke(() -> s3Client.doesBucketExist(bucketName));
  }

  /**
   * @param bucketName the bucket name.
   */
  public Request<?> generateGetBucketLocationRequest(String bucketName) {
    return invoke(() -> s3Client.getBucketLocation(bucketName));
  }

  /**
   * @param bucketName the bucket name.
   */
  public Request<?> generateGetBucketAclRequest(String bucketName) {
    return invoke(() -> s3Client.getBucketAcl(bucketName));
  }

  /**
   * @param bucketName the bucket name.
   * @param path the object path (Leading / will be remove).
   */
  public Request<?> generateGetObjectMetadataRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    GetObjectMetadataRequest request = new GetObjectMetadataRequest(bucketName, path);
    return invoke(() -> s3Client.getObjectMetadata(request));
  }

  /**
   * @param bucketName the bucket name.
   * @param path the object path (Leading / will be remove).
   */
  public Request<?> generateListObjectsV1Request(String bucketName, String path) {
    path = (path!= null && path.startsWith("/") ? path.substring(1) : path);
    ListObjectsRequest request = new ListObjectsRequest();
    request.setBucketName(bucketName);
    request.setPrefix(path);
    request.setDelimiter("/");
    return invoke(() -> s3Client.listObjects(request));
  }

  /**
   * @param bucketName the bucket name.
   * @param path the object path (Leading / will be remove).
   */
  public Request<?> generateListObjectsV2Request(String bucketName, String path) {
    path = (path!= null && path.startsWith("/") ? path.substring(1) : path);
    ListObjectsV2Request request = new ListObjectsV2Request().withBucketName(bucketName)
        .withPrefix(path).withDelimiter("/");
    return invoke(() -> s3Client.listObjectsV2(request));
  }

  /**
   * @param bucketName the bucket name.
   * @param path the object path (Leading / will be remove).
   */
  public Request<?> generateDeleteObjectRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    String finalPath = path;
    return invoke(() -> s3Client.deleteObject(bucketName, finalPath));
  }

  /**
   * @param bucketName the bucket name.
   * @param paths the object paths (Leading / will be remove).
   */
  public Request<?> generateDeleteObjectsRequest(String bucketName, String[] paths) {
    for (int i = 0; i < paths.length; i++) {
      paths[i] = (paths[i].startsWith("/") ? paths[i].substring(1) : paths[i]);
    }
    DeleteObjectsRequest deleteObjectsRequest = new DeleteObjectsRequest(bucketName).withKeys(paths)
        .withQuiet(true);
    return invoke(() -> s3Client.deleteObjects(deleteObjectsRequest));
  }

  /**
   * Creates a Request for a PutObjectRequest with an empty upload stream.
   *
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generatePutObjectRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    ObjectMetadata objectMetadata = new ObjectMetadata();
    objectMetadata.setContentLength(0);
    PutObjectRequest putObjectRequest = new PutObjectRequest(bucketName, path, EMPTY_INPUT_STREAM,
        objectMetadata);
    putObjectRequest.setMetadata(objectMetadata);
    return invoke(() -> s3Client.putObject(putObjectRequest));
  }

  /**
   * Creates a Request for a UploadPartRequest with a random uploadId and an empty upload stream.
   *
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generateUploadPartRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    UploadPartRequest uploadPartRequest = new UploadPartRequest().withBucketName(bucketName)
        .withKey(path).withUploadId(
            UUID.randomUUID().toString()).withPartNumber(random.nextInt(10)).withPartSize(0)
        .withInputStream(EMPTY_INPUT_STREAM);
    return invoke(() -> s3Client.uploadPart(uploadPartRequest));
  }

  /**
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generateListMultipartUploadRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    ListMultipartUploadsRequest request = new ListMultipartUploadsRequest(bucketName);
    request.setPrefix(path);
    return invoke(() -> s3Client.listMultipartUploads(request));
  }

  /**
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generateListPartsUploadRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    ListPartsRequest request = new ListPartsRequest(bucketName, path, UUID.randomUUID().toString());
    return invoke(() -> s3Client.listParts(request));
  }

  /**
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generateInitiateMultipartUploadRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    InitiateMultipartUploadRequest request = new InitiateMultipartUploadRequest(bucketName, path,
        new ObjectMetadata());
    return invoke(() -> s3Client.initiateMultipartUpload(request));
  }

  /**
   * Creates a request for a AbortMultiPartUpload with a random uploadId
   *
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generateAbortMultipartUploadRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    AbortMultipartUploadRequest request = new AbortMultipartUploadRequest(bucketName, path,
        UUID.randomUUID().toString());
    return invoke(() -> s3Client.abortMultipartUpload(request));
  }

  /**
   * Creates a request for a CompleteMultipartUpload with a random uploadId
   *
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generateMultipartCommitRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    CompleteMultipartUploadRequest request = new CompleteMultipartUploadRequest(bucketName, path,
        UUID.randomUUID().toString(), new ArrayList<>());
    return invoke(() -> s3Client.completeMultipartUpload(request));
  }

  /**
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generateGetObjectRequest(String bucketName, String path) {
    path = (path.startsWith("/") ? path.substring(1) : path);
    GetObjectRequest request = new GetObjectRequest(bucketName, path);
    return invoke(() -> s3Client.getObject(request));
  }

  /**
   * Creates a SelectObjectRequest with a fixed sql string
   *
   * @param bucketName the bucket name.
   * @param path the object paths (Leading / will be remove).
   */
  public Request<?> generateSelectObjectContentRequest(String bucketName, String path) {
    SelectObjectContentRequest request = new SelectObjectContentRequest();
    request.setBucketName(bucketName);
    request.setKey(path);
    request.setExpressionType(ExpressionType.SQL);
    request.setExpression("SELECT * FROM S3OBJECT");
    return invoke(() -> s3Client.selectObjectContent(request));
  }


  private static Request<?> invoke(Runnable op) {
    try {
      op.run();
    } catch (RequestProvidingException e) {
      return e.getRequest();
    }
    throw new RuntimeException("Failed to extract request");
  }
}
