/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.s3.lib.signer.util;

import org.apache.ranger.raz.s3.lib.HttpMethod;
import org.apache.ranger.raz.s3.lib.signer.SignRequest;
import org.apache.ranger.raz.s3.lib.signer.SignResponse;
import org.apache.ranger.raz.s3.lib.signer.SignResponse.AuthorizationResponse;
import org.apache.ranger.raz.s3.lib.signer.SignResponse.AuthorizationResponse.ResponseCode;
import org.apache.ranger.raz.s3.lib.signer.impl.SignRequestImpl;
import org.apache.ranger.raz.s3.lib.signer.impl.SignResponseImpl;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.AuthorizationResponseProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.HttpMethodProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.ResponseCodeProto;
import org.apache.ranger.raz.s3.lib.signer.protos.SignerProtocolProtos.SignResponseProto;

import com.google.common.collect.Lists;
import java.net.URI;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import org.junit.Assert;
import org.junit.Test;

public class TestProtoConverters {

  @Test(timeout = 10_000l)
  public void testEnumHttpMethod() {
    // Mainly to ensure new enums are not added in one place only

    for (HttpMethod httpMethod : HttpMethod.values()) {
      ProtoConverters.toProto(httpMethod);
    }

    for (HttpMethodProto httpMethodProto : HttpMethodProto.values()) {
      ProtoConverters.fromProto(httpMethodProto);
    }
  }

  @Test(timeout = 10_000l)
  public void testEnumResponseCode() {
    // Mainly to ensure new enums are not added in one place only
    for (ResponseCode responseCode : ResponseCode.values()) {
      ProtoConverters.toProto(responseCode);
    }

    for (ResponseCodeProto responseCodeProto : ResponseCodeProto.values()) {
      ProtoConverters.fromProto(responseCodeProto);
    }
  }

  @Test(timeout = 10_000l)
  public void testAuthorizationResponse() {
    AuthorizationResponse authzResponse = null;

    // Authorized Response
    authzResponse = new AuthorizationResponse(ResponseCode.AUTHZ_AUTHORIZED, null, null);
    convertAndValidateAuthorizationResponse(authzResponse);

    // Unauthorized Response, with null pth (probably never hit), and null message
    authzResponse = new AuthorizationResponse(ResponseCode.AUTHZ_UNAUTHORIZED, null, null);
    convertAndValidateAuthorizationResponse(authzResponse);

    // Unauthorized Response with paths and message populated.
    authzResponse = new AuthorizationResponse(ResponseCode.AUTHZ_UNAUTHORIZED,
        new String[]{"/path1", "/path2/baz"}, "Unauthorized blah");
    convertAndValidateAuthorizationResponse(authzResponse);

    // Unknown with no paths + response message
    authzResponse = new AuthorizationResponse(ResponseCode.AUTHZ_UNKNOWN, null, "Unknown blah");
    convertAndValidateAuthorizationResponse(authzResponse);

    authzResponse = new AuthorizationResponse(ResponseCode.AUTHZ_UNKNOWN, null, "");
    convertAndValidateAuthorizationResponse(authzResponse);
  }

  @Test(timeout = 10_000l)
  public void testSignResponse() {
    AuthorizationResponse authorizationResponse = new AuthorizationResponse(
        ResponseCode.AUTHZ_AUTHORIZED, null, null);
    Map<String, String> signerGeneratedHeaders = null;

    String region = null;
    SignResponse signResponse = null;

    // Simple response.
    signerGeneratedHeaders = new HashMap<>();
    signerGeneratedHeaders.put("h1", "v1");
    signResponse = new SignResponseImpl(authorizationResponse, signerGeneratedHeaders, null);
    convertAndValidateSignResponse(signResponse);

    // Include region, multiple headers
    signerGeneratedHeaders = new HashMap<>();
    signerGeneratedHeaders.put("h1", "v1");
    signerGeneratedHeaders.put("h2", "v2");
    region = "us-east2";
    signResponse = new SignResponseImpl(authorizationResponse, signerGeneratedHeaders, region);
    convertAndValidateSignResponse(signResponse);

    // Empty header map not allowed, so not tested.
  }

  @Test(timeout = 10_000)
  public void testSignRequest() throws Exception {
    URI endpoint = new URI("http://example.com");
    HttpMethod httpMethod = HttpMethod.GET;
    Map<String, String> headers = null;
    Map<String, List<String>> parameters = null;
    String resourcePath = "";
    byte[] contentToSign = null;
    int timeOffset = 0;
    String endporintPrefix = "prefix";
    String serviceName = "service";

    SignRequest signRequest;

    // Empty headers and parameters.
    signRequest = new SignRequestImpl(endpoint, httpMethod, headers, parameters, resourcePath,
        contentToSign, timeOffset, endporintPrefix, serviceName);
    convertAndValidateSignRequest(signRequest);

    // Headers - empty map, parameters empty
    headers = new HashMap<>();
    parameters = null;
    signRequest = new SignRequestImpl(endpoint, httpMethod, headers, parameters, resourcePath,
        contentToSign, timeOffset, endporintPrefix, serviceName);
    convertAndValidateSignRequest(signRequest);

    // Headers and Parameters - empty maps
    headers = new HashMap<>();
    parameters = new HashMap<>();
    signRequest = new SignRequestImpl(endpoint, httpMethod, headers, parameters, resourcePath,
        contentToSign, timeOffset, endporintPrefix, serviceName);
    convertAndValidateSignRequest(signRequest);

    // Headers populated, null parameters
    headers = new HashMap<>();
    headers.put("h1", "v1");
    headers.put("h2", "v2");
    parameters = null;
    signRequest = new SignRequestImpl(endpoint, httpMethod, headers, parameters, resourcePath,
        contentToSign, timeOffset, endporintPrefix, serviceName);
    convertAndValidateSignRequest(signRequest);

    // Headers populated, null parameters, content populated, path set
    headers.put("h1", "v1");
    headers.put("h2", "v2");
    parameters = null;
    resourcePath = "resourcePath/blah";
    contentToSign = "content".getBytes();
    signRequest = new SignRequestImpl(endpoint, httpMethod, headers, parameters, resourcePath,
        contentToSign, timeOffset, endporintPrefix, serviceName);
    convertAndValidateSignRequest(signRequest);

    // Headers populated, parameter populated, null content, path set
    headers.put("h1", "v1");
    headers.put("h2", "v2");
    parameters = new HashMap<>();
    parameters.put("list-type", Lists.newArrayList("2"));
    parameters.put("delimitter", Lists.newArrayList("/"));
    resourcePath = "resourcePath/blah";
    contentToSign = null;
    signRequest = new SignRequestImpl(endpoint, httpMethod, headers, parameters, resourcePath,
        contentToSign, timeOffset, endporintPrefix, serviceName);
    convertAndValidateSignRequest(signRequest);

    // Headers populated, parameters with 'null' value in list, null content, path set
    List<String> values = new LinkedList<>();
    values.add(null);
    headers.put("h1", "v1");
    headers.put("h2", "v2");
    parameters = new HashMap<>();
    parameters.put("delete", values);
    resourcePath = "resourcePath/blah";
    contentToSign = null;
    signRequest = new SignRequestImpl(endpoint, httpMethod, headers, parameters, resourcePath,
        contentToSign, timeOffset, endporintPrefix, serviceName);
    convertAndValidateSignRequest(signRequest);

    // Parameters with null list.
    values.add(null);
    headers.put("h2", "v2");
    parameters = new HashMap<>();
    parameters.put("random", null);
    parameters.put("delimitter", Lists.newArrayList("/"));
    resourcePath = "resourcePath/blah";
    contentToSign = null;
    signRequest = new SignRequestImpl(endpoint, httpMethod, headers, parameters, resourcePath,
        contentToSign, timeOffset, endporintPrefix, serviceName);
    convertAndValidateSignRequest(signRequest);
  }


  private void convertAndValidateAuthorizationResponse(AuthorizationResponse authzResponse1) {

    AuthorizationResponseProto proto = ProtoConverters.toProto(authzResponse1);
    AuthorizationResponse authzResponse2 = ProtoConverters.fromProto(proto);

    validateAuthorizationResponseEquivalence(authzResponse1, authzResponse2);

  }

  private void validateAuthorizationResponseEquivalence(AuthorizationResponse authzResponse1,
      AuthorizationResponse authzResponse2) {
    Assert.assertEquals(authzResponse1.getResponseCode(), authzResponse2.getResponseCode());
    if (authzResponse1.getUnauthorizedPaths() == null) {
      Assert.assertNull(authzResponse2.getUnauthorizedPaths());
    } else {
      Assert.assertArrayEquals(authzResponse1.getUnauthorizedPaths(),
          authzResponse2.getUnauthorizedPaths());
    }
    Assert.assertEquals(authzResponse1.getResponseMessage(), authzResponse2.getResponseMessage());
  }

  private void convertAndValidateSignResponse(SignResponse signResponse1) {
    SignResponseProto proto = ProtoConverters.toProto(signResponse1);
    SignResponse signResponse2 = ProtoConverters.fromProto(proto);

    validateAuthorizationResponseEquivalence(signResponse1.getAuthorizationResponse(),
        signResponse2.getAuthorizationResponse());
    Assert.assertEquals(signResponse1.getSignerGeneratedHeaders(),
        signResponse2.getSignerGeneratedHeaders());
    Assert.assertEquals(signResponse1.getRegion(), signResponse2.getRegion());
  }

  private void convertAndValidateSignRequest(SignRequest signRequest1) throws Exception {
    final String base64 = ProtoConverters.toBase64(signRequest1);
    final SignRequest signRequest2 = ProtoConverters.reqFromBase64(base64);

    Assert.assertEquals(signRequest1.getEndpoint(), signRequest2.getEndpoint());
    Assert.assertEquals(signRequest1.getHttpMethod(), signRequest2.getHttpMethod());
    Assert.assertEquals(signRequest1.getHeaders(), signRequest2.getHeaders());
    Assert.assertEquals(signRequest1.getParameters(), signRequest2.getParameters());
    Assert.assertEquals(signRequest1.getResourcePath(), signRequest2.getResourcePath());
    if (signRequest1.getContentToSign() == null) {
      Assert.assertNull(signRequest2.getContentToSign());
    } else {
      Assert.assertArrayEquals(signRequest1.getContentToSign(), signRequest2.getContentToSign());
    }
    Assert.assertEquals(signRequest1.getTimeOffset(), signRequest2.getTimeOffset());
    Assert.assertEquals(signRequest1.getEndpointPrefix(), signRequest2.getEndpointPrefix());
    Assert.assertEquals(signRequest1.getServiceName(), signRequest2.getServiceName());

  }
}