/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.ranger.raz.processor.s3;

import java.io.IOException;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.auth.AWSCredentialsProviderChain;
import com.amazonaws.auth.AWSSessionCredentials;
import com.amazonaws.auth.BasicSessionCredentials;
import com.amazonaws.client.builder.AwsClientBuilder;
import com.amazonaws.services.securitytoken.AWSSecurityTokenService;
import com.amazonaws.services.securitytoken.AWSSecurityTokenServiceClientBuilder;
import com.amazonaws.services.securitytoken.model.AWSSecurityTokenServiceException;
import com.amazonaws.services.securitytoken.model.AssumeRoleRequest;
import com.amazonaws.services.securitytoken.model.AssumeRoleResult;
import com.amazonaws.services.securitytoken.model.MalformedPolicyDocumentException;
import com.amazonaws.services.securitytoken.model.PackedPolicyTooLargeException;
import com.amazonaws.services.securitytoken.model.RegionDisabledException;
import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Preconditions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.raz.s3.lib.exceptions.RCSUncheckedException;

import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS;
import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS_DEFAULT;

/**
 * A Credential Provider class for RAZ to call assumeRole on an ARN provided
 * via fs configs to assume that role and get the session credentials needed
 * to sign a request.
 */
public class RazAssumeRoleCredentialProvider implements AWSCredentialsProvider {

  private static final Logger LOG =
      LoggerFactory.getLogger(RazAssumeRoleCredentialProvider.class);

  /** Session credentials to be used to sign the request. */
  private AWSSessionCredentials credentials;

  private final Configuration configuration;

  /** ARN to be assumed. */
  private final String arn;

  /** Duration of the STS token. In Seconds. */
  private final int duration;

  /** Name of the session. */
  private final String roleSessionName;

  /** Offset time to expire the token before the actual expiry time. */
  private final long awsCredentialExpiryOffsetSeconds;

  /** STS client used to call AWS to do the AssumeRole call. */
  private AWSSecurityTokenService stsClient;

  /** Session credentials expiry time. */
  private long awsSessionCredentialsExpiryTime;

  /** Session credentials expiry date. */
  private Date awsSessionCredentialsExpiryDate;

  /** Read Write Lock to keep the credential fetching thread safe. */
  private final ReentrantReadWriteLock credentialsRWLock = new ReentrantReadWriteLock();

  /** Result of the AssumeRole call */
  private AssumeRoleResult assumeRoleResult;

  /** A config to add a list of custom credential providers for source creds. */
  private static final String CUSTOM_SOURCE_CREDENTIAL_PROVIDER_LIST =
      "raz.s3.source.credential.provider.list";

  /** A config to specify which role ARN to assume. */
  private static final String ROLE_ARN = "raz.s3.ext.assumed.role.arn";

  /** A config to specify the role session duration. */
  private static final String ROLE_SESSION_DURATION =
      "raz.s3.ext.assumed.role.session.duration";

  /** A config to specify role session name. */
  private static final String ROLE_SESSION_NAME =
      "raz.s3.ext.assumed.role.session.name";

  /** A config to set the region for the STS client. */
  private static final String STS_REGION = "raz.s3.ext.sts.region";

  /** Default role session name if no session name is provided in the config. */
  private static final String DEFAULT_ROLE_SESSION_NAME = "rangerraz";

  /**
   * Credential Provider class for RAZ S3 where the user gets the Session
   * Credentials by Assuming another role provided as a config.
   *
   * @param configuration FS Config containing the configs to assume a role.
   * @throws IOException
   */
  public RazAssumeRoleCredentialProvider(final Configuration configuration)
      throws IOException {
    this.configuration = configuration;
    arn = configuration.getTrimmed(ROLE_ARN, "");
    Preconditions.checkState(StringUtils.isNotEmpty(arn),
        "Empty ARN, please set " + ROLE_ARN);

    duration = configuration.getInt(ROLE_SESSION_DURATION, 3600);
    roleSessionName = configuration.getTrimmed(ROLE_SESSION_NAME, DEFAULT_ROLE_SESSION_NAME);
    awsCredentialExpiryOffsetSeconds =
        configuration.getInt(RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS,
        RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS_DEFAULT);
    LOG.info("Configured value of param {} is {} ", RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS,
        awsCredentialExpiryOffsetSeconds);

    try{
      // build up the STS client.
      stsClient = getSTSClient();
      getAssumedRoleResult();
    } catch (Exception e) {
      LOG.error("Failed to initiate RazAssumeRoleCredentialProvider class", e);
      return;
    }

    LOG.info("RazAssumeRoleCredentialProvider initialized..");
  }

  @Override
  public AWSCredentials getCredentials() {
    try {
      credentialsRWLock.readLock().lock();
      // If SessionCredentials expired, LOG the event, else if
      // credentials were cached use them.
      if (isAWSSessionCredentialsExpired()) {
        LOG.info("AWS Session Credentials about to expire in {} or at [ {} ]."
                + " Renewing Credentials",
            awsSessionCredentialsExpiryTime,
            awsSessionCredentialsExpiryDate);
      } else {
        return credentials;
      }
    } finally {
      credentialsRWLock.readLock().unlock();
    }

    // Extract credentials and cache them
    try {
      credentialsRWLock.writeLock().lock();
      LOG.trace("Credentials write lock initiated");
      // Check if credentials have expired and refresh if needed.
      if (isAWSSessionCredentialsExpired()) {
        // Refreshing the AssumeRoleResult and fetching new credentials.
        refresh();
        LOG.info("Extracting credentials and caching them; Time left on the "
                + "session credentials in Seconds: {}; expiry time: {} or [ "
                + "{} ]",
            getTimeLeftOnSessionCredentials(),
            awsSessionCredentialsExpiryTime,
            awsSessionCredentialsExpiryDate);
      }
    } catch (AWSSecurityTokenServiceException e) {
      LOG.error("Failed to get credentials for role {}",
          arn, e);
      throw e;
    } finally {
      LOG.trace("Credentials write unlock");
      credentialsRWLock.writeLock().unlock();
    }
    return credentials;
  }

  /**
   * Method to return true if the AWS Session credentials have expired else
   * false.
   *
   * @return boolean signifying if Session Credentials have expired or not.
   */
  @VisibleForTesting
  boolean isAWSSessionCredentialsExpired() {
    return getTimeLeftOnSessionCredentials() <= 0;
  }

  private long getTimeLeftOnSessionCredentials() {
    return ((awsSessionCredentialsExpiryTime - awsCredentialExpiryOffsetSeconds) - Instant.now().getEpochSecond());
  }

  /**
   * Method used to do the AssumeRole call. The arn provided as a config
   * would be assumed and a result containing the session credentials and the
   * expiry time would be returned as a response.
   *
   */
  private void getAssumedRoleResult() {
    AssumeRoleRequest request = createAssumeRoleRequest();
    try {
      LOG.info("Assuming role with arn:{}, with SessionName : {} and "
              + "duration: {}",
          arn, roleSessionName, duration);
      assumeRoleResult = getSTSClient().assumeRole(request);
      // Converting AssumeRoleResult credentials into AWSSessionCredential.
      credentials = getSessionCredentials(assumeRoleResult);
    } catch (MalformedPolicyDocumentException | PackedPolicyTooLargeException | RegionDisabledException e) {
      LOG.error("Could not assume the role");
      throw e;
    } catch (AWSSecurityTokenServiceException e) {
      LOG.error("Not permitted to assume role");
      throw e;
    } catch (RuntimeException e) {
      LOG.error("Some other error happened while assuming");
      throw e;
    }
    // Expiry time in Date and seconds.
    awsSessionCredentialsExpiryDate =
        assumeRoleResult.getCredentials().getExpiration();
    awsSessionCredentialsExpiryTime = TimeUnit.MILLISECONDS.toSeconds(
        awsSessionCredentialsExpiryDate.getTime());
  }

  /**
   * Method to populate the AssumeRoleRequest to be sent to acquire a new role.
   * @return Instance of AssumeRoleRequest.
   */
  private AssumeRoleRequest createAssumeRoleRequest() {
    return new AssumeRoleRequest().withRoleArn(arn)
        .withRoleSessionName(roleSessionName)
        .withDurationSeconds(duration);
  }

  /**
   * Method to convert Credentials from AssumeRoleResult into
   * AWSSessionCredentials.
   * @param assumeRoleResult the AssumeRole call result.
   * @return AWSSessionCredentials.
   */
  private AWSSessionCredentials getSessionCredentials(
      AssumeRoleResult assumeRoleResult) {
    return new BasicSessionCredentials(
            assumeRoleResult.getCredentials().getAccessKeyId(),
            assumeRoleResult.getCredentials().getSecretAccessKey(),
            assumeRoleResult.getCredentials().getSessionToken());
  }

  @Override
  public void refresh() {
    // Forcing the AssumeRole call and fetching the credentials.
    getAssumedRoleResult();
  }

  /**
   * STS client.
   *
   * @return AWSSecurityTokenService instance.
   */
  @VisibleForTesting
  AWSSecurityTokenService getSTSClient() {
    if (stsClient == null) {
      LOG.info("Building STS client...");
      RazAWSCredentialsProviderList credentialsProviderList =
          new RazAWSCredentialsProviderList();
      credentialsProviderList.initRazAWSCredentialsProviderList();
      AWSCredentialsProvider sourceAwsCredentialProvider =
          credentialsProviderList.getSourceAwsCredentialProvider();
      LOG.info("AWSCredential Provider used: {}", sourceAwsCredentialProvider);

      AWSSecurityTokenServiceClientBuilder awsSTSClientBuilder =
          AWSSecurityTokenServiceClientBuilder.standard()
              .withCredentials(sourceAwsCredentialProvider);
      String region = getRegion();
      LOG.info("Region used for the STS client: {}", region);
      if (StringUtils.isNotEmpty(region)) {
        AwsClientBuilder.EndpointConfiguration endpointConfiguration =
            new AwsClientBuilder.EndpointConfiguration(
                String.format(Locale.ROOT, "https://sts.%s.amazonaws.com", region),
                region);
        awsSTSClientBuilder.withEndpointConfiguration(endpointConfiguration);
      }
      stsClient = awsSTSClientBuilder.build();
      LOG.info("STS client initiated.");
    }
    return stsClient;
  }

  /**
   * Method to get the region configured, else keep empty as default and let
   * the AWS SDK figure it out.
   *
   * @return Region name as a String.
   */
  private String getRegion() {
    return configuration.getTrimmed(STS_REGION, "");
  }

  /**
   * A Class used to get the Source credentials from the container or EC2
   * instance that would be used by the STS client to do the AssumeRole call.
   */
  private class RazAWSCredentialsProviderList {

    /** Credential provider chain to be used by STS client. */
    private AWSCredentialsProviderChain sourceCredentialsProviderChain;

    /**
     * Initializing the RazAWSCredentialsProviderList class.
     */
    public void initRazAWSCredentialsProviderList() {

      LOG.info("Initializing RazAWSCredentialsProviderList");

      // Custom Credential provider list passed as a config.
      String[] sourceCredentialsProviderListNames = configuration.getStrings(
          CUSTOM_SOURCE_CREDENTIAL_PROVIDER_LIST,
          "com.amazonaws.auth.WebIdentityTokenCredentialsProvider");

      try {
        LOG.info("Value of param :{} : {}",
            CUSTOM_SOURCE_CREDENTIAL_PROVIDER_LIST, sourceCredentialsProviderListNames);

        LOG.info("Initializing source credential provider chain...");
        List<AWSCredentialsProvider> sourceCredentialProviderList =
            new ArrayList<>();

        for (String sourceCredentialProviderClass :
            sourceCredentialsProviderListNames) {
          LOG.info("Initializing AWSCredentialsProvider instance of class: "
              + "{}", sourceCredentialProviderClass);

          AWSCredentialsProvider providerInstance =
              Class.forName(sourceCredentialProviderClass.trim())
                  .asSubclass(AWSCredentialsProvider.class).getConstructor()
                  .newInstance();

          sourceCredentialProviderList.add(providerInstance);
        }
        sourceCredentialsProviderChain = new AWSCredentialsProviderChain(sourceCredentialProviderList);
        LOG.info("Initialized AWS source credential provider chain");
      }  catch (Exception ex) {
        throw new RCSUncheckedException(RCSUncheckedException.Code.GENERAL_SERVER_ERROR, String
            .format("Could not initialize credentials provider for chain "
                + "[%s]", Arrays.toString(sourceCredentialsProviderListNames)),
            ex);
      }
    }

    /**
     * Get the Source AWS Credential Provider chain.
     * @return Instance of AWSCredentialsProvider.
     */
    public AWSCredentialsProvider getSourceAwsCredentialProvider() {
      return sourceCredentialsProviderChain;
    }
  }

  @VisibleForTesting
  String getArn() {
    return arn;
  }

  @VisibleForTesting
  int getDuration() {
    return duration;
  }

  @VisibleForTesting
  String getRoleSessionName() {
    return roleSessionName;
  }

  @VisibleForTesting
  long getAwsCredentialExpiryOffsetSeconds() {
    return awsCredentialExpiryOffsetSeconds;
  }
}
