/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.processor.s3;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.commons.compress.utils.Sets;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.plugin.policyengine.RangerAccessRequest;
import org.apache.ranger.plugin.policyengine.RangerAccessResource;
import org.apache.ranger.plugin.policyengine.RangerAccessResult;
import org.apache.ranger.raz.intg.client.RangerRazClient;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.raz.model.RangerRazResultBase;
import org.apache.ranger.raz.model.RangerRazResultBase.ResourceAccessResult;
import org.apache.ranger.raz.processor.RangerDefaultRazProcessor;
import org.apache.ranger.raz.processor.RangerRazContext;
import org.apache.ranger.raz.processor.RangerRazRequestContext;
import org.apache.ranger.raz.processor.s3.signer.S3RemoteSigner;
import org.apache.ranger.raz.processor.s3.signer.S3RemoteSignerContext;
import org.apache.ranger.raz.s3.lib.HttpMethod;
import org.apache.ranger.raz.s3.lib.RazS3Constants.AccessType;
import org.apache.ranger.raz.s3.lib.authorizer.AuthorizationRequest;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsOperationType;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsRequest;
import org.apache.ranger.raz.s3.lib.signer.Constants;
import org.apache.ranger.raz.s3.lib.signer.SignRequest;
import org.apache.ranger.raz.s3.lib.signer.SignResponse;
import org.apache.ranger.raz.s3.lib.signer.util.ProtoConverters;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.RANGER_RAZ_SERVICE_TYPE_S3_DEFAULT_SERVICE_NAME;
import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.RANGER_RAZ_VALID_SAS_API_VERSION;
import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.S3_SERVICE_NAME_DEFAULT;

public class S3RazProcessor extends RangerDefaultRazProcessor {
	private static final Logger LOG = LoggerFactory.getLogger(S3RazProcessor.class);


	private static String valid_sas_api_version = null;
	private String s3ServiceName = S3_SERVICE_NAME_DEFAULT;
	public String API_VERSION = "1";

	private static S3RemoteSignerContext remoteSignerContext;
	private S3RemoteSigner s3RemoteSigner;
	private Configuration configuration;

	@Override
	public void init(RangerRazContext context) throws RangerRazException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("==> S3RazProcessor.init(context={})", context);
		}

		Iterator<Map.Entry<String, String>> confIterator = context.getRazConfig().iterator();

		if (LOG.isDebugEnabled()) {
			while (confIterator.hasNext()) {
				Map.Entry<String, String> entry = confIterator.next();
				LOG.debug("S3RazProcessor: " + "Key = " + entry.getKey() + " value = " + entry.getValue());
			}
		}

		s3ServiceName = context.getRazConfig().get(RANGER_RAZ_SERVICE_TYPE_S3_DEFAULT_SERVICE_NAME, S3_SERVICE_NAME_DEFAULT);
		Configuration conf = context.getRazConfig();
		valid_sas_api_version = conf.get(RANGER_RAZ_VALID_SAS_API_VERSION, API_VERSION);
		LOG.debug("valid_sas_api_version: {}", valid_sas_api_version);

		configuration = context.getRazConfig();
		// initialize aws credentials chain for signing. this have to be done
		// only at when the service starts
		if (remoteSignerContext == null) {
			remoteSignerContext = new S3RemoteSignerContext(configuration);
		}
		remoteSignerContext.initAWSCredentialsChain();
		s3RemoteSigner = new S3RemoteSigner(remoteSignerContext, configuration);
	}

	@Override
	public List<RangerAccessRequest> preProcess(RangerRazRequest request, RangerRazRequestContext context) throws RangerRazException {
		if (LOG.isTraceEnabled()) {
			LOG.trace("==> S3RazProcessor.preProcess(request={}, context={})", request, context);
		}
		request.setServiceName(s3ServiceName);                  	// ignore the value in request.serviceName
		request.setUser(context.getRemoteUser());                 // ignore the value in request.user
		request.setUserGroups(context.getRemoteUserGroups());     // ignore the value in request.userGroups
		request.setClientIpAddress(context.getRemoteIpAddress()); // ignore the value in request.clientIpAddress
		request.setAccessTime(new Date());

		SignRequest signRequest = getSignRequest(request);

		final List<RangerAccessRequest> ret = getRangerAccessRequests(request, signRequest);
		if (LOG.isTraceEnabled()) {
			LOG.trace("<== S3RazProcessor.preProcess(request={}, context={}): ret={}", request, context, ret);
		}
		return ret;
	}

	@Override
	public RangerRazResult postProcess(RangerRazRequest request, List<RangerAccessResult> accessResults, RangerRazRequestContext context) throws RangerRazException {
		if (LOG.isTraceEnabled()) {
			LOG.trace("==> S3RazProcessor.postProcess(request={}, accessResults={}, context={})", request, accessResults, context);
		}

		RangerRazResult ret = getRangerRazResult(request, accessResults);
		ResourceAccessResult rar = ret != null ? ret.getOperResult() : null;
		if (LOG.isTraceEnabled()) {
			LOG.trace("ResourceAccessResult: {}", rar);
			LOG.trace("<== S3RazProcessor.postProcess(request={}, accessResults={}, context={}): ret={}", request, accessResults, context, ret);
		}

		return ret;
	}

	private SignRequest getSignRequest(RangerRazRequest request) throws RangerRazException {
		// decode signRequest
		SignRequest signRequest = null;
		try {
			final Map<String, String> rContext = request.getContext();
			final String signReqString = rContext.get(Constants.S3_SIGN_REQUEST_FIELD);
			signRequest = ProtoConverters.reqFromBase64(signReqString);
			if (LOG.isTraceEnabled()) {
				LOG.trace("SignRequest: {}", signRequest.toString());
			}
			if (signRequest.getHttpMethod() == HttpMethod.DEFAULT) {
				throw new RangerRazException("HTTPMethod is not set, or lost in translation:" + signRequest.getHttpMethod().toString());
			}
		} catch (IOException e) {
			throw new RangerRazException("Can't reconstruct the S3 Sign Request.", e);
		}
		return signRequest;
	}

	@Override
	public RangerAccessResource getRangerAccessResource(RangerRazRequest.ResourceAccess razResourceAccess) {
		// I use a different mechanism to create the resource than what's done here.
		// It's easier not to use this method here and the pre and post methods of
		// the parent class.
		return null;
	}

	private List<RangerAccessRequest> getRangerAccessRequests(RangerRazRequest request, SignRequest signRequest) throws RangerRazException {
		if (LOG.isTraceEnabled()) {
			LOG.trace("==> S3RazProcessor.getRangerAccessRequests(request={}, signRequest: {})", request, signRequest);
		}

		List<RangerAccessRequest> ret = new ArrayList<>();

		String invokingUser = request.getUser();
		String clientIp = request.getClientIpAddress();
		Set<String> invokingGroups = request.getUserGroups();
		final String clusterName = request.getClusterName();

		final AuthorizationRequest authzRequest =
				s3RemoteSigner.getAuthorizationRequest(signRequest);

		// If the sign request is HEAD on the bucket root, then it does not need to be authorized.
		if (isSignRequestHeadOnRoot(authzRequest)) {
			LOG.debug("The sign request is a HEAD on the root. It won't be added to requests for authorization.");
			return ret;
		}

		for (String path : authzRequest.getPathsToAuthorize()) {
			// CDPD-18654. remove trailing slash for the path as the ranger policy
			// uses exact matching in the path parameter for the rule.
			path = trimTrailingSlashExceptRoot(path);
			AccessType accessType = authzRequest.getRequiredAuthorization();
			RangerCloudStorageAuthorizer.RangerCloudStorageAccessRequest accessRequest
					= new RangerCloudStorageAuthorizer.RangerCloudStorageAccessRequest(
							authzRequest.getBucket(),
							path, invokingUser, invokingGroups, accessType,
							clusterName,
							clientIp,
							request.getOperation().getAction());

			// CDPD-21279. For all HEAD and LIST request if there is a matching policy for the child, then
			// use the same policy for the parent. Adding SELF_OR_CHILD to the matching scope
			// will check for child policy, and apply it to the parent.
			if (isSignRequestHeadOnObject(authzRequest) || isSignRequestListOnObject(authzRequest)) {
				accessRequest.setResourceMatchingScope(RangerAccessRequest.ResourceMatchingScope.SELF_OR_CHILD);
			}

			ret.add(accessRequest);
			if (LOG.isDebugEnabled()) {
				LOG.debug("Bucket: {}, Path: {}, InvokingUser: {}, InvokingGroups: {}, " +
								"AccessType: {}, ClusterName: {}, ClientIp: {}",
						authzRequest.getBucket(), path, invokingUser, invokingGroups,
						accessType, clusterName, clientIp);
				LOG.debug("Access request: {}", accessRequest.toString());
			}
		}

		if (LOG.isTraceEnabled()) {
			LOG.trace("<== S3RazProcessor.getRangerAccessRequests(request={}, context={}): ret={}", request, ret);
		}

		return ret;
	}

	static protected String trimTrailingSlashExceptRoot(String path) {
		// if the length is 1 then the path should be returned as is
		// because that should be the root (otherwise pointless and then denied anyways)
		if (path.length() > 1 && path.endsWith("/")) {
			return path.substring(0, path.length() - 1);
		} else {
			return path;
		}
	}

	private RangerRazResult getRangerRazResult(RangerRazRequest request, List<RangerAccessResult> accessResults) throws RangerRazException {
		if (LOG.isTraceEnabled()) {
			LOG.trace("==> S3RazProcessor.getRangerRazResult(request={}, accessResults={})", request, accessResults);
		}

		SignRequest signRequest = getSignRequest(request);
		String user = request.getUser();

		final RangerRazRequestContext context
				= new RangerRazRequestContext(request.getUser(),
				request.getUserGroups(), request.getClientIpAddress(),
				new String[]{});
		context.setKeysToRedact(RangerRazClient.RAZ_CLIENT_KEYS_TO_REDACT);

		RangerRazResult ret = super.postProcess(request, accessResults, context);

		maybeForceAllowedAccessResult(signRequest, ret, s3RemoteSigner);

		// Sign the request if the result is allowed, and patch the signature in the response
		if (ret.getOperResult().getResult().equals(RangerRazResultBase.AccessResult.ALLOWED)) {
			final SignResponse signResponse = s3RemoteSigner.signRequest(
					signRequest, request.getClientIpAddress(), user, s3RemoteSigner.getGroups(user));
			// encode signResponse
			try {
				final String signResponseStr = ProtoConverters.toBase64(signResponse);
				Map<String, String> rContext = new HashMap<>();
				rContext.put(Constants.S3_SIGN_RESPONSE_FIELD, signResponseStr);
				ret.getOperResult().setAdditionalInfo(rContext);
			} catch (IOException e) {
				throw new RangerRazException("Can't construct the S3 Sign Response.", e);
			}
		} else if (ret.getOperResult().getResult().equals(RangerRazResultBase.AccessResult.NOT_DETERMINED)) {
			// Set result to DENIED in case it's NOT_DETERMINED.
			ret.getOperResult().setResult(RangerRazResultBase.AccessResult.DENIED);
		}

		if (LOG.isTraceEnabled()) {
			LOG.trace("<== S3RazProcessor.getRangerRazResult(request={}, accessResults={}): ret={}",
					request, accessResults, ret);
		}
		return ret;
	}

	/**
	 * For the following request(s) the access result of the authorization check should be ALLOWED:
	 *
	 * - [HTTP method: HEAD, Resource path: ""] -> aws-sdk bucket region lookup
	 *   This request should be signed every time. The AWS SDK triggers a HEAD / when it doesn't know which S3 endpoint
	 *   to use for signing. This is only an existence check, not a probe for any values. HEAD / is always signed. This
	 *   is not a LIST and does not leak any information other than the existence and location of the bucket.
	 *
	 * @param signRequest The sign request from the S3 signer (form hadoop-aws)
	 * @param ret Forced allowed if needed.
	 */
	protected static void maybeForceAllowedAccessResult(SignRequest signRequest, RangerRazResult ret, S3RemoteSigner signer)
			throws RangerRazException {
		LOG.debug("Request method: {}, Resource path: '{}'", signRequest.getHttpMethod(), signRequest.getResourcePath());
		final AuthorizationRequest authRequest = signer.getAuthorizationRequest(signRequest);
		if (isSignRequestHeadOnRoot(authRequest)) {
			LOG.debug("The request should be allowed - bucket root STAT request: aws-sdk request for bucket region.");
			ret.getOperResult().setResult(RangerRazResultBase.AccessResult.ALLOWED);
		}
	}

	/**
	 * HEAD on root will be enabled at all times. This check is for aws-sdk to find out the bucket region.
	 *
	 * @param authRequest the request to use for the decision
	 * @return is the request operation type is HEAD on the root
	 */
	protected static boolean isSignRequestHeadOnRoot(AuthorizationRequest authRequest) {
		return (authRequest instanceof AwsRequest
				&& ((AwsRequest) authRequest).getOperationType().equals(AwsOperationType.HEAD_BUCKET_METADATA));
	}

	/**
	 * HEAD on object is where we add SELF_OR_CHILD to the RANGER request.
	 *
	 * @param authRequest the request to use for the decision
	 * @return is the request operation type is HEAD on object
	 */
	protected static boolean isSignRequestHeadOnObject(AuthorizationRequest authRequest) {
		return (authRequest instanceof AwsRequest
								&& ((AwsRequest) authRequest).getOperationType().equals(AwsOperationType.HEAD_OBJECT_METADATA));
	}

	protected final static HashSet<AwsOperationType> listOperations = Sets.newHashSet(
			AwsOperationType.LIST_V1,
			AwsOperationType.LIST_V2,
			AwsOperationType.MULTIPART_LIST
	);

	/**
	 * LIST on object is where we add SELF_OR_CHILD to the RANGER request.
	 *
	 * @param authRequest the request to use for the decision
	 * @return is the request operation type is LIST on object
	 */
	protected static boolean isSignRequestListOnObject(AuthorizationRequest authRequest) {
		return (authRequest instanceof AwsRequest
								&& listOperations.contains(((AwsRequest) authRequest).getOperationType()));
	}

}
