/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.ranger.raz.processor.s3.dtutil;

import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.auth.AWSSessionCredentials;
import com.google.common.annotations.VisibleForTesting;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.hadoop.classification.InterfaceAudience;
import org.apache.hadoop.classification.InterfaceStability;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.s3a.auth.MarshalledCredentialBinding;
import org.apache.hadoop.fs.s3a.auth.MarshalledCredentials;
import org.apache.hadoop.fs.s3a.auth.NoAwsCredentialsException;
import org.apache.hadoop.fs.s3a.auth.delegation.DelegationTokenIOException;
import org.apache.hadoop.fs.s3native.S3xLoginHelper;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.util.DurationInfo;
import org.apache.knox.gateway.cloud.idbroker.common.KnoxToken;
import org.apache.knox.gateway.cloud.idbroker.messages.RequestDTResponseMessage;
import org.apache.knox.gateway.cloud.idbroker.s3a.S3AIDBClient;
import org.apache.knox.gateway.shell.CloudAccessBrokerSession;
import org.apache.knox.gateway.shell.KnoxSession;
import org.apache.knox.gateway.util.Tokens;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.time.Instant;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import static org.apache.knox.gateway.cloud.idbroker.IDBConstants.MESSAGE_FAILURE_TO_AUTHENTICATE_TO_IDB_DT;
import static org.apache.knox.gateway.cloud.idbroker.IDBConstants.MESSAGE_FAILURE_TO_AUTHENTICATE_TO_IDB_KERBEROS;
import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS;
import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS_DEFAULT;
import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.RANGER_RAZ_IDB_CREDENTIAL_EXPIRY_OFFSET_SECONDS;
import static org.apache.ranger.raz.processor.s3.signer.S3RazConstants.RANGER_RAZ_IDB_CREDENTIAL_EXPIRY_OFFSET_SECONDS_DEFAULT;

/**
 * Support for DT aws credentials provider to be used with credentials
 * got from idbroker
 *
 * We only have one expiry and one token here, because this provider will
 * be used exclusively with one user (rangerraz) who is running the raz service.
 * That user has access for every bucket under the account.
 *
 */
@InterfaceAudience.Public
@InterfaceStability.Stable
public class DTAwsCredentialsProvider implements AWSCredentialsProvider {

  AWSSessionCredentials credentials;

  private static final Logger LOG = LoggerFactory.getLogger(
      DTAwsCredentialsProvider.class);

  private S3AIDBClient idbClient;
  private KnoxToken knoxToken;
  private Configuration configuration;

  private long dtExpiryTime;
  private long awsCredentialsExpiryTime;

  /**
   * Number of seconds prior to aws credential expiry to
   * trigger a new aws credential request.
   */
  private final int idbCredentialExpiryOffsetSeconds;

  /**
   * Number of seconds prior to idBroker token expiry to
   * trigger renewal of the idBroker token.
   */
  private final int awsCredentialExpiryOffsetSeconds;

  String token = "";

  /**
   * Wire name of this binding: {@value}.
   */
  private static final String NAME = "IDBDelegationTokenForRaz";
  private static final String COMPONENT_NAME = NAME;

  /**
   * Session credentials: initially empty.
   */
  private MarshalledCredentials marshalledCredentials;

  private static final String E_NO_KNOX_DELEGATION_TOKEN
      = "No Knox delegation token";

  private static final String E_NO_SESSION_TO_KNOX_AWS
      = "No session to knox AWS credential endpoint";

  /**
   * Build the credentials from a filesystem URI and configuration.
   * @param conf configuration containing secrets/references to.
   * @throws IOException failure
   */
  public DTAwsCredentialsProvider(final Configuration conf) {
    this.configuration = conf;
    idbCredentialExpiryOffsetSeconds = conf.getInt(RANGER_RAZ_IDB_CREDENTIAL_EXPIRY_OFFSET_SECONDS,
            RANGER_RAZ_IDB_CREDENTIAL_EXPIRY_OFFSET_SECONDS_DEFAULT);
    awsCredentialExpiryOffsetSeconds = conf.getInt(RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS,
            RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS_DEFAULT);
    LOG.info("Configured value of param {} is {} ", RANGER_RAZ_IDB_CREDENTIAL_EXPIRY_OFFSET_SECONDS,
            idbCredentialExpiryOffsetSeconds);
    LOG.info("Configured value of param {} is {} ", RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS,
            awsCredentialExpiryOffsetSeconds);
    try {
      getCredentialsFromIDB();
      getAwsSessionCredentialsFromIDB();
    } catch (IOException | URISyntaxException e) {
      LOG.error("Failed while trying to get token from idbroker.", e);
      return;
    }
    LOG.info("DTAwsCredentialsProvider initialized");
  }

  private void getCredentialsFromIDB() throws IOException, URISyntaxException {
    // bucket can be empty, only needed for the client.
    String bucket = "";

    UserGroupInformation currentUser = UserGroupInformation.getCurrentUser();
    LOG.info("CurrentUser: {}", currentUser);

    idbClient = S3AIDBClient.createFullIDBClient(configuration, currentUser, bucket);
    String endpoint = idbClient.getCredentialsURL();
    LOG.info("idb endpoint: {}", endpoint);

    // get knox token
    final Pair<KnoxSession, String> sessionPair = getNewKnoxDelegationTokenSession(configuration);
    KnoxSession session = sessionPair.getLeft();
    String origin = sessionPair.getRight();
    final RequestDTResponseMessage message = idbClient.requestKnoxDelegationToken(session, origin, getCanonicalUri(bucket));

    token = extractTokenFromResponse(message);
    // print a small bit of the secret
    LOG.info("Bonded to Knox token {}", Tokens.getTokenDisplayText(token));
    String gatewayCertificate = extractGatewayCertificate(message);

    if (gatewayCertificate.isEmpty()) {
      LOG.warn("No certificate provided by gateway: renewals will not work");
    }

    this.dtExpiryTime = message.expiryTimeSeconds();
    knoxToken = new KnoxToken("", token, message.token_type,
        this.dtExpiryTime, gatewayCertificate, message.managed);

    LOG.info("Knox DT retrieved at {} going to expiry at {}",
        Instant.now().getEpochSecond(), this.dtExpiryTime);
  }

  private void getAwsSessionCredentialsFromIDB() throws IOException {
    // get access credentials
    CloudAccessBrokerSession knoxCABSession = idbClient.createKnoxCABSession(knoxToken);
    if (knoxCABSession == null) {
      throw new DelegationTokenIOException(E_NO_SESSION_TO_KNOX_AWS);
    }
    marshalledCredentials = idbClient.fetchCloudCredentials(knoxCABSession);

    credentials = (AWSSessionCredentials) MarshalledCredentialBinding.toAWSCredentials(marshalledCredentials,
        MarshalledCredentials.CredentialTypeRequired.SessionOnly,
        COMPONENT_NAME);

    awsCredentialsExpiryTime = TimeUnit.MILLISECONDS.toSeconds(marshalledCredentials.getExpiration());

    LOG.info("AWS credential retrieved from idbroker: Key: {}, Id: {}," +
                    "sessiontoken: {}, expiry time: {}, now: {}",
        credentials.getAWSSecretKey().substring(0, 8),
        credentials.getAWSAccessKeyId().substring(0, 8),
        Tokens.getTokenDisplayText(credentials.getSessionToken()),
        awsCredentialsExpiryTime,
        Instant.now().getEpochSecond());
  }

  /**
   * Get the token from the response -includes a check for a null/empty
   * token.
   *
   * @param response response to a token request.
   * @return the token
   * @throws DelegationTokenIOException if invalid.
   */
  private String extractTokenFromResponse(final RequestDTResponseMessage response)
      throws DelegationTokenIOException {
    final String token = response.access_token;
    if (StringUtils.isEmpty(token)) {
      throw new DelegationTokenIOException(E_NO_KNOX_DELEGATION_TOKEN);
    }
    return token;
  }

  /**
   * Extract the gateway certificate, or "" if there was none in the
   * response.
   *
   * @param response response from the DT request.
   * @return a certificate string.
   */
  private String extractGatewayCertificate(final RequestDTResponseMessage response) {
    String cert = response.endpoint_public_cert;
    if (cert == null) {
      cert = "";
    }
    return cert;
  }

  private URI getCanonicalUri(String bucket) throws URISyntaxException {
    return S3xLoginHelper.canonicalizeUri(new URI(bucket), 0);
  }

  private Pair<KnoxSession, String> getNewKnoxDelegationTokenSession(Configuration config)
      throws IOException {
    LOG.debug(
        "Attempting to create a Knox delegation token session using local credentials (kerberos, simple)");
    Pair<KnoxSession, String> sessionDetails = idbClient.createKnoxDTSession(config);
    if (sessionDetails.getLeft() != null) {
      LOG.debug("Created a Knox delegation token session using local credentials (kerberos, simple)");
    }

    if (sessionDetails.getLeft() == null) {
      String message;

      if (knoxToken == null) {
        message = MESSAGE_FAILURE_TO_AUTHENTICATE_TO_IDB_KERBEROS;
      } else {
        message = MESSAGE_FAILURE_TO_AUTHENTICATE_TO_IDB_DT;
      }

      throw new IllegalStateException(message);
    }

    return sessionDetails;
  }

  ReentrantReadWriteLock credentialsRWLock = new ReentrantReadWriteLock();

  @Override
  public AWSCredentials getCredentials() {
    try {
      credentialsRWLock.readLock().lock();

      boolean isDTExpired = this.isDTExpired();
      boolean isAwsSessionCredentialsExpired = this.isAwsSessionCredentialsExpired();
      LOG.debug("Credentials with expiry. " +
              "Now: {}, " +
              "Aws credentials expiry with offset: {}, " +
              "DT expiry with offset: {}, " +
              "isDTExpired: {}, " +
              "isAwsSessionCredentialsExpired: {}",
          Instant.now().getEpochSecond(),
          awsCredentialsExpiryTime - awsCredentialExpiryOffsetSeconds,
          dtExpiryTime - idbCredentialExpiryOffsetSeconds,
          isDTExpired(),
          isAwsSessionCredentialsExpired);

      if (isAwsSessionCredentialsExpired || isDTExpired) {
        LOG.info("Expired: aws credential: {}, idb credential: {}; needs renewal.",
                isAwsSessionCredentialsExpired, isDTExpired);
      } else {
        // returning credentials here if nothing is expired
        return credentials;
      }
    } finally {
      credentialsRWLock.readLock().unlock();
    }

    try (DurationInfo d = new DurationInfo(LOG, true,
        "Getting credentials writeLock for thread id: {}", Thread.currentThread().getId())) {
      credentialsRWLock.writeLock().lock();
    }
    try {
      if (isDTExpired()) {
        try {
          LOG.info("Getting new idb token and credentials because the idb token got expired.");
          getCredentialsFromIDB();
          getAwsSessionCredentialsFromIDB();
        } catch (IOException | URISyntaxException e) {
          throw new NoAwsCredentialsException(
              DTAwsCredentialsProvider.class.getName(),
              "Exception while renewing DT and aws session credentials: " + e, e);
        }
      } else if (isAwsSessionCredentialsExpired()) {
        LOG.info("Getting new aws session credentials because it got expired.");
        try {
          getAwsSessionCredentialsFromIDB();
        } catch (IOException e) {
          throw new NoAwsCredentialsException(
              DTAwsCredentialsProvider.class.getName(),
              "Exception while renewing aws session credentials: " + e, e);
        }
      } else {
        LOG.debug("DT and AWS credentials were refreshed by another thread.");
      }
    } finally {
      credentialsRWLock.writeLock().unlock();
    }
    return credentials;
  }

  private boolean isDTExpired() {
    return Instant.now().getEpochSecond() > (dtExpiryTime - idbCredentialExpiryOffsetSeconds);
  }

  private boolean isAwsSessionCredentialsExpired() {
    return Instant.now().getEpochSecond() > (awsCredentialsExpiryTime - awsCredentialExpiryOffsetSeconds);
  }

  @VisibleForTesting
  public int getIdbCredentialExpiryOffsetSeconds() {
    return idbCredentialExpiryOffsetSeconds;
  }

  @VisibleForTesting
  public int getAwsCredentialExpiryOffsetSeconds() {
    return awsCredentialExpiryOffsetSeconds;
  }

  @Override
  public void refresh() {
  }

  @Override
  public String toString() {
    return getClass().getSimpleName();
  }

}
