/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.processor.s3.signer;

import com.amazonaws.AmazonServiceException;
import com.amazonaws.AmazonWebServiceRequest;
import com.amazonaws.ClientConfiguration;
import com.amazonaws.DefaultRequest;
import com.amazonaws.Request;
import com.amazonaws.SdkClientException;
import com.amazonaws.auth.AWSCredentials;
import com.amazonaws.auth.AWSSessionCredentials;
import com.amazonaws.auth.Signer;
import com.amazonaws.auth.internal.SignerConstants;
import com.amazonaws.handlers.HandlerContextKey;
import com.amazonaws.http.HttpMethodName;
import com.amazonaws.services.s3.Headers;
import com.amazonaws.services.s3.request.S3HandlerContextKeys;

import org.apache.hadoop.security.Groups;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.s3.lib.RangerRazExceptionS3Headers;
import org.apache.ranger.raz.s3.lib.RazS3Constants;
import org.apache.ranger.raz.s3.lib.authorizer.AuthorizationRequest;
import org.apache.ranger.raz.s3.lib.aws.AWSS3V4SignerWithoutRange;
import org.apache.ranger.raz.s3.lib.aws.parsers.AwsRequestParser;
import org.apache.ranger.raz.s3.lib.aws.GenericAmazonWebServiceRequest;
import org.apache.ranger.raz.s3.lib.exceptions.RCSUncheckedException;
import org.apache.ranger.raz.s3.lib.exceptions.RCSUncheckedException.Code;
import org.apache.ranger.raz.s3.lib.perflogger.PerfLogger;
import org.apache.ranger.raz.s3.lib.perflogger.PerfLoggerConstants;
import org.apache.ranger.raz.s3.lib.signer.SignRequest;
import org.apache.ranger.raz.s3.lib.signer.SignResponse;
import org.apache.ranger.raz.s3.lib.signer.impl.SignResponseImpl;
import org.apache.ranger.raz.s3.lib.signer.util.ExceptionUtils;
import org.apache.ranger.raz.s3.lib.utils.S3Utils;
import com.google.common.annotations.VisibleForTesting;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;

import org.apache.hadoop.conf.Configuration;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.apache.hadoop.fs.s3a.S3AUtils.initProxySupport;
import static org.apache.hadoop.fs.s3a.S3AUtils.propagateBucketOptions;
import static org.apache.ranger.raz.s3.lib.RazS3Constants.HEADER_USER_AGENT;

/**
 * Does not handle content signing. Skips a lot of client configuration options. TODO: Enhance this
 * javadocs
 */
public class S3RemoteSigner {

  private static final Logger LOG = LoggerFactory
      .getLogger(S3RemoteSigner.class);

  private static final ByteArrayInputStream EMPTY_BAIS = new ByteArrayInputStream(
      new byte[0]);

  private final AwsRequestParser awsRequestParser;
  // Need to modify configuration for per-bucket properties propagation.
  private Configuration conf;
  private final S3RemoteSignerContext context;


  public S3RemoteSigner(S3RemoteSignerContext context, Configuration conf) {
    this.conf = conf;
    this.awsRequestParser = new AwsRequestParser();
    this.context = context;
  }

  public SignResponse signRequest(SignRequest signRequest,
      String clientIp, String user,
      Set<String> groups) throws RCSUncheckedException, RangerRazException {
    String contextString = MDC.get(RazS3Constants.MDC_REQUEST_ID);
    try {
      PerfLogger.get().opBegin(contextString, PerfLoggerConstants.REMOTE_SIGNER_SIGN);
      return signRequestInternal(signRequest, clientIp, user, groups);
    } finally {
      PerfLogger.get().opEnd(contextString, PerfLoggerConstants.REMOTE_SIGNER_SIGN);
    }
  }

  public SignResponse signRequestInternal(SignRequest signRequest,
      String clientIp, String user,
      Set<String> groups) throws RCSUncheckedException, RangerRazException {
    String contextString = MDC.get(RazS3Constants.MDC_REQUEST_ID);

    LOG.debug("DEBUG: signRequest: S3SignRequest=[{}], User=[{}], Groups=[{}]",
        signRequest, user, groups);

    String host = signRequest.getEndpoint().getHost();

    String bucketName = getBucketName(host);
    // Propagating per-bucket configs.
    conf = propagateBucketOptions(conf, bucketName);
    String region = null;
    try {
      PerfLogger.get().opBegin(contextString, PerfLoggerConstants.REMOTE_SIGNER_GET_REGION);
      region = getRegion(bucketName);
    } finally {
      PerfLogger.get().opEnd(contextString, PerfLoggerConstants.REMOTE_SIGNER_GET_REGION);
    }
    LOG.debug("Inferred: BucketName: {}, Region: {}, EndPoint: {}", bucketName,
        region, signRequest.getEndpoint());

    AmazonWebServiceRequest innerRequest = new GenericAmazonWebServiceRequest();
    DefaultRequest<?> toSignRequest = constructRequestForSigning(signRequest,
        innerRequest, region);

    if (awsRequestParser.isContentSigningRequired(toSignRequest)) {
      augmentRequestForContentSigning(toSignRequest, innerRequest,
          signRequest.getContentToSign());
    } else {
      maybeAugmentRequestWithEmptyContent(toSignRequest);
    }

    updateUserAgentInRequest(toSignRequest, user);

    if(LOG.isDebugEnabled()) {
      AuthorizationRequest authzRequest = awsRequestParser.parseAwsRequest(toSignRequest);

      LOG.debug(
          "DEBUG: Signing the request: "
              + " [bucket={}, paths={}, requiredAuthz={}, user={}, groups={}]",
          bucketName, Arrays.toString(authzRequest.getPathsToAuthorize()),
          authzRequest.getRequiredAuthorization(), user, groups);
    }

    Signer s3Signer = null;
    try {
      PerfLogger.get().opBegin(contextString, PerfLoggerConstants.REMOTE_SIGNER_ACQUIRE_SIGNER);
      s3Signer = getSignerInstance(signRequest.getEndpointPrefix(),
          signRequest.getServiceName(), region);
    } finally {
      PerfLogger.get().opEnd(contextString, PerfLoggerConstants.REMOTE_SIGNER_ACQUIRE_SIGNER);
    }
    final AWSCredentials credentials = context.getAwsCredentialProvider()
        .getCredentials();
    try {
      PerfLogger.get().opBegin(contextString, PerfLoggerConstants.REMOTE_SIGNER_REAL_SIGN);
      s3Signer.sign(toSignRequest, credentials);
      // The S3Signer will modify the request object itself with the signed headers.
    } finally {
      PerfLogger.get().opEnd(contextString, PerfLoggerConstants.REMOTE_SIGNER_REAL_SIGN);
    }

    Map<String, String> signerGeneratedHeaders = populateSignerGeneratedHeaders(
        toSignRequest, credentials);

    return new SignResponseImpl(
        SignResponse.AuthorizationResponse.getAuthorizedInstance(),
        signerGeneratedHeaders, region);
  }

  public String getBucketName(String host) {
    String bucketName = S3Utils.getBucketNameFromHost(host);
    if (bucketName == null) {
      throw new RCSUncheckedException(Code.GENERAL_SERVER_ERROR,
          "Could not extract bucket name from host: [" + host + "]");
    }
    return bucketName;
  }

  public String getRegion(String bucketName) throws RangerRazException {
    // TODO: ProfileCredentialProvider usage here needs to be a CredentialProviderChain,
    //  one of which will get credentials from Knox.
    try {
      String regionFromCache = S3Utils.lookupRegionFromCache(bucketName);
      if (regionFromCache != null) {
        return regionFromCache;
      }
      final ClientConfiguration awsConf = new ClientConfiguration();
      // Initializing proxy support and wiring the parameters with aws sdk.
      initProxySupport(conf, bucketName, awsConf);
      return S3Utils
          .lookupRegion(bucketName, context.getAwsCredentialProvider(), awsConf);
    } catch (SdkClientException | IOException e) {
      int statusCode = -1;
      checkASEtoRangerRazExceptionAndThrow(e);
      throw ExceptionUtils.createUnableToFindRegionException(statusCode, bucketName, e);
    }
  }

  /**
   * A method to check if we caught an AmazonServiceException and to return
   * proper params to translate this exception.
   *
   * @param excp the Throwable caught.
   */
  private void checkASEtoRangerRazExceptionAndThrow(Throwable excp)
      throws RangerRazException {
    // If the exception is from Amazon Service, raise a
    // WebApplicationException with params containing amazon
    // exception details.
    AmazonServiceException amazonServiceException = null;
    if (excp instanceof AmazonServiceException) {
      amazonServiceException = (AmazonServiceException) excp;
    } else if (excp.getCause() instanceof AmazonServiceException) {
      amazonServiceException = (AmazonServiceException) excp.getCause();
    }

    if (amazonServiceException != null) {
      LOG.debug("Translating Throwable to AmazonServiceException..");
      // Construct an AWS header map to set in RangerRazException.
      Map<String, String> awsHeaders = new HashMap<>();
      awsHeaders.put(RangerRazExceptionS3Headers.AWS_ERROR_CODE,
          amazonServiceException.getErrorCode());
      awsHeaders.put(RangerRazExceptionS3Headers.AWS_STATUS_CODE,
          String.valueOf(amazonServiceException.getStatusCode()));
      awsHeaders.put(RangerRazExceptionS3Headers.AWS_REQUEST_ID,
          amazonServiceException.getRequestId());
      awsHeaders.put(RangerRazExceptionS3Headers.AWS_EXTENDED_REQUEST_ID,
          amazonServiceException.getHttpHeaders().get(
              Headers.EXTENDED_REQUEST_ID));

      // Throw a RangerRazException with AWS headers.
      throw new RangerRazException(amazonServiceException, awsHeaders);
    }
    LOG.debug("Could not translate Throwable to AmazonServiceException.");
  }

  private Map<String, String> populateSignerGeneratedHeaders(
      Request<?> toSignRequest, AWSCredentials credentials) {
    Map<String, String> signerGeneratedHeaders = new HashMap<>();
    addSignerGeneratedHeader(toSignRequest.getHeaders(), SignerConstants.HOST,
        signerGeneratedHeaders);
    addSignerGeneratedHeader(toSignRequest.getHeaders(),
        SignerConstants.X_AMZ_DATE,
        signerGeneratedHeaders);
    addSignerGeneratedHeader(toSignRequest.getHeaders(),
        SignerConstants.X_AMZ_CONTENT_SHA256,
        signerGeneratedHeaders);
    addSignerGeneratedHeader(toSignRequest.getHeaders(),
        SignerConstants.AUTHORIZATION,
        signerGeneratedHeaders);
    addSignerGeneratedHeader(toSignRequest.getHeaders(),
        HEADER_USER_AGENT,
        signerGeneratedHeaders);

    if (credentials instanceof AWSSessionCredentials) {
      signerGeneratedHeaders.put(SignerConstants.X_AMZ_SECURITY_TOKEN,
          ((AWSSessionCredentials) credentials).getSessionToken());
    }

    return signerGeneratedHeaders;
  }

  private void addSignerGeneratedHeader(Map<String, String> allHeaders,
      String key,
      Map<String, String> map) {
    String value = allHeaders.get(key);
    if (value != null) {
      map.put(key, value);
    } else {
      // TODO Monitor this. Ideally all headers should exist.
      //  Otherwise throw an error, rather than WARN messages.
      LOG.warn("Could not find header key: {} after signing in headers: {}",
          key, allHeaders);
    }
  }

  public DefaultRequest<?> constructRequestForSigning(SignRequest request,
      AmazonWebServiceRequest innerRequest, String region) {

    DefaultRequest toSignRequest = new DefaultRequest(innerRequest,
        request.getServiceName());
    toSignRequest.setEndpoint(request.getEndpoint());
    toSignRequest.setHttpMethod(
        HttpMethodName.fromValue(request.getHttpMethod().name()));
    toSignRequest.setHeaders(request.getHeaders());
    toSignRequest.setParameters(request.getParameters());
    toSignRequest.setResourcePath(request.getResourcePath());
    toSignRequest.setTimeOffset(request.getTimeOffset());

    toSignRequest.addHandlerContext(HandlerContextKey.SIGNING_REGION, region);

    LOG.debug("ToSignRequest: {}, RequestForSigning: {}", request, toSignRequest);

    return toSignRequest;
  }

  @VisibleForTesting
  void updateUserAgentInRequest(DefaultRequest toSignRequest, String user) {
    String userAgentFromRequest = ((Map<String, String>) toSignRequest.getHeaders())
        .get(HEADER_USER_AGENT);

    String ugiUserAgent = "user:" + user;
    String userAgent = Stream.of(ugiUserAgent, userAgentFromRequest)
        .filter(s -> s != null && !s.isEmpty())
        .map(String::trim)
        .collect(Collectors.joining(","));

    LOG.debug("Using User-Agent: {}", userAgent);
    toSignRequest.addHeader(HEADER_USER_AGENT, userAgent);
  }

  private void augmentRequestForContentSigning(Request<?> request,
      AmazonWebServiceRequest innerRequest, byte[] contentBytes) {
    if (contentBytes == null) {
      return;
    }
    ByteArrayInputStream bais = new ByteArrayInputStream(contentBytes);

    request.setContent(bais);
    request.addHandlerContext(S3HandlerContextKeys.IS_PAYLOAD_SIGNING_ENABLED,
        true);

    // TODO: Maybe respect client side readLimitInfo if other requests need content signing.
    // All of the payload should be signed
    innerRequest.getRequestClientOptions().setReadLimit(Integer.MAX_VALUE);
  }

  // S3 ImplementationSpecific: S3Signer does a lookup on the contents being null to determine
  //  whether or not to add request parameters to the signer payload. null contents == no parameters
  //  in case of POST requests. See AbstractAwsSigner.getCanonicalizedQueryString
  private void maybeAugmentRequestWithEmptyContent(Request<?> request) {
    if (request.getHttpMethod() == HttpMethodName.POST
        && request.getContent() == null) {
      request.setContent(EMPTY_BAIS);
    }
  }

  private Signer getSignerInstance(String endpointPrefix, String serviceName,
      String region) {
    // TODO: Maybe maintain a signer cache. The AWS library creates an instance each time.
    // HMac/signing keys are cached statically by the AWS4Signer so this isn't critical.
    AWSS3V4SignerWithoutRange s3Signer = new AWSS3V4SignerWithoutRange();
    s3Signer.setEndpointPrefix(endpointPrefix);
    s3Signer.setServiceName(serviceName);
    s3Signer.setRegionName(region);
    return s3Signer;
  }

  public AuthorizationRequest getAuthorizationRequest(SignRequest signRequest)
      throws RangerRazException {
    AwsRequestParser awsRequestParser = new AwsRequestParser();
    AmazonWebServiceRequest innerRequest = new GenericAmazonWebServiceRequest();
    String host = signRequest.getEndpoint().getHost();
    String bucketName = this.getBucketName(host);
    final String region = this.getRegion(bucketName);
    final DefaultRequest<?> toSignRequest
        = this.constructRequestForSigning(signRequest, innerRequest,
        region);

    // inject content to the request
    if (awsRequestParser.isContentSigningRequired(toSignRequest)) {
      augmentRequestForContentSigning(toSignRequest, innerRequest,
          signRequest.getContentToSign());
    } else {
      maybeAugmentRequestWithEmptyContent(toSignRequest);
    }

    return awsRequestParser.parseAwsRequest(toSignRequest);
  }

  private Groups getGroupMapping() {
    // This works off of core-site.xml
    return Groups.getUserToGroupsMappingService();
  }

  public Set<String> getGroups(String user) {
    List<String> groups = null;
    try {
      groups = getGroupMapping().getGroups(user);
    } catch (IOException e) {
      LOG.debug(
          "Failed to get groups for user: [" + user + "]. ExceptionShort=" + e);
      LOG.trace("TRACE for group lookup failure for user: [" + user + "]", e);
    }
    if (groups != null) {
      return new HashSet<>(groups);
    } else {
      return Collections.emptySet();
    }
  }
}