/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.ranger.raz.processor.s3.signer;

import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.auth.AWSCredentialsProviderChain;
import org.apache.ranger.raz.s3.lib.exceptions.RCSUncheckedException;
import org.apache.ranger.raz.s3.lib.exceptions.RCSUncheckedException.Code;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.hadoop.conf.Configuration;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the Context
 */
public class S3RemoteSignerContext {

  private final Configuration conf;
  private AWSCredentialsProviderChain credentialsProviderChain;
  private static final Logger LOG = LoggerFactory
      .getLogger(S3RemoteSignerContext.class);

  public static final String SERVER_AWS_CREDENTIALS_CHAIN
      = "rcs.server.signer.aws.credentials.chain";
  public static final String[] SERVER_AWS_CREDENTIALS_CHAIN_DEFAULT
      = {"org.apache.ranger.raz.processor.s3.dtutil.DTAwsCredentialsProvider",
      "org.apache.ranger.raz.processor.s3.signer.SimpleAWSCredentialsProvider"};

  public S3RemoteSignerContext(Configuration conf) {
    this.conf = conf;
  }

  public void initAWSCredentialsChain() {
    String[] providerChain = conf.getStrings(
        SERVER_AWS_CREDENTIALS_CHAIN,
        SERVER_AWS_CREDENTIALS_CHAIN_DEFAULT);

    try {
      LOG.info("Initializing AWS Credentials Provider for chain {}", providerChain);
      List<AWSCredentialsProvider> providers = new ArrayList(providerChain.length);
      for (String providerClass : providerChain) {
        AWSCredentialsProvider providerInstance =
            Class.forName(providerClass.trim())
                .asSubclass(AWSCredentialsProvider.class).getConstructor(Configuration.class).newInstance(conf);
        providers.add(providerInstance);
      }
      credentialsProviderChain = new AWSCredentialsProviderChain(providers);
      LOG.info("Initialized AWS Credentials Provider for chain {}", providerChain);
    } catch (Exception ex) {
      throw new RCSUncheckedException(Code.GENERAL_SERVER_ERROR, String
          .format("Could not initialize credentials provider for chain [%s]", providerChain),
          ex);
    }
  }

  public AWSCredentialsProvider getAwsCredentialProvider() {
    return credentialsProviderChain;
  }
}
