/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.ranger.raz.processor.s3;

import java.io.IOException;
import java.time.Instant;
import java.util.Date;

import com.amazonaws.auth.AWSSessionCredentials;
import com.amazonaws.services.securitytoken.AWSSecurityTokenService;
import com.amazonaws.services.securitytoken.model.AssumeRoleResult;
import com.amazonaws.services.securitytoken.model.Credentials;
import org.assertj.core.api.Assertions;
import org.junit.Before;
import org.junit.Test;

import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.raz.processor.s3.signer.S3RazConstants;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;

/**
 * Test functionality of {@link RazAssumeRoleCredentialProvider}.
 */
public class TestRazAssumeRoleCredentialProvider {

  private Configuration conf;
  private static final String ACCESS_KEY_ID = "access_key_id";
  private static final String SECRET_KEY_ID = "secret_key_id";
  private static final String SESSION_TOKEN = "session_token";

  @Before
  public void setUp() throws Exception {
    conf = new Configuration();
    conf.set("raz.s3.ext.assumed.role.arn", "testArn");
    conf.setInt("raz.s3.ext.assumed.role.session.duration", 1000);
    conf.set("raz.s3.ext.assumed.role.session.name", "testSession");
  }

  @Test
  public void testAssumeRoleConfigurations() throws IOException {
    RazAssumeRoleCredentialProvider razAssumeRoleCredentialProvider =
        new RazAssumeRoleCredentialProvider(conf);

    Assertions.assertThat(razAssumeRoleCredentialProvider.getArn())
        .describedAs("Role Arn should be set via config")
        .isEqualTo("testArn");
    Assertions.assertThat(razAssumeRoleCredentialProvider.getDuration())
        .describedAs("The Duration for the Session credentials should be set "
            + "via config")
        .isEqualTo(1000);
    Assertions.assertThat(razAssumeRoleCredentialProvider.getRoleSessionName())
        .describedAs("Role Session Name should be set via config")
        .isEqualTo("testSession");

    Assertions.assertThat(razAssumeRoleCredentialProvider.getAwsCredentialExpiryOffsetSeconds())
        .describedAs("Aws Credential Expiry offset should be set to default "
            + "if not passed through the config")
        .isEqualTo(S3RazConstants.RANGER_RAZ_AWS_CREDENTIAL_EXPIRY_OFFSET_SECONDS_DEFAULT);

  }

  @Test
  public void testFetchingCredentials() throws IOException {
    AWSSecurityTokenService mockStsClient =
        mock(AWSSecurityTokenService.class);

    Credentials testCredentials = new Credentials(ACCESS_KEY_ID,
        SECRET_KEY_ID, SESSION_TOKEN,
        new Date(Instant.now().toEpochMilli() + 60 * 60 * 1000));

    AssumeRoleResult mockAssumeRoleResult = mock(AssumeRoleResult.class);

    doReturn(testCredentials).when(mockAssumeRoleResult).getCredentials();
    doReturn(mockAssumeRoleResult).when(mockStsClient).assumeRole(any());

    RazAssumeRoleCredentialProvider razAssumeRoleCredentialProvider =
        new RazAssumeRoleCredentialProvider(conf) {
          public AWSSecurityTokenService getSTSClient() {
            return mockStsClient;
          }
        };

    Assertions.assertThat(razAssumeRoleCredentialProvider.getSTSClient())
        .describedAs("MockSTSClient should be returned when called "
            + "getSTSClient()")
        .isEqualTo(mockStsClient);

    verify(mockStsClient, times(1)).assumeRole(any());

    AWSSessionCredentials fetchedCredential =
        (AWSSessionCredentials) razAssumeRoleCredentialProvider.getCredentials();
    verifyNoMoreInteractions(mockStsClient);

    // Verify fetched credential.
    assertFetchedCredentials(fetchedCredential);

    // Fetch credentials again.
    AWSSessionCredentials fetchedCredential2 =
        (AWSSessionCredentials) razAssumeRoleCredentialProvider.getCredentials();
    verifyNoMoreInteractions(mockStsClient);
    verifyNoMoreInteractions(mockStsClient);

    // Verify fetched credential.
    assertFetchedCredentials(fetchedCredential2);
  }

  @Test
  public void testExpiredCredentialsRenewal() throws IOException {
    AWSSecurityTokenService mockStsClient =
        mock(AWSSecurityTokenService.class);

    Credentials testCredentials1 = new Credentials(ACCESS_KEY_ID,
        SECRET_KEY_ID, SESSION_TOKEN,
        new Date(Instant.now().toEpochMilli() + 4 * 60 * 1000));

    Credentials testCredentials2 = new Credentials(ACCESS_KEY_ID,
        SECRET_KEY_ID, SESSION_TOKEN,
        new Date(Instant.now().toEpochMilli() + 60 * 60 * 1000));

    AssumeRoleResult mockAssumeRoleResult1 = mock(AssumeRoleResult.class);
    doReturn(testCredentials1).when(mockAssumeRoleResult1).getCredentials();

    AssumeRoleResult mockAssumeRoleResult2 = mock(AssumeRoleResult.class);
    doReturn(testCredentials2).when(mockAssumeRoleResult2).getCredentials();

    // Return mockAssumeRoleResult1 with first assumeRole call.
    doReturn(mockAssumeRoleResult1).when(mockStsClient).assumeRole(any());

    RazAssumeRoleCredentialProvider razAssumeRoleCredentialProvider =
        new RazAssumeRoleCredentialProvider(conf) {
          public AWSSecurityTokenService getSTSClient() {
            return mockStsClient;
          }
        };

    // Verify assumeRole call after RazAssumeRoleCredentialProvider constructor.
    verify(mockStsClient, times(1)).assumeRole(any());

    // Return mockAssumeRoleResult2 with the next assumeRole call.
    doReturn(mockAssumeRoleResult2).when(mockStsClient).assumeRole(any());

    // Assert before acquiring new credentials that previous ones have expired.
    Assertions.assertThat(razAssumeRoleCredentialProvider.isAWSSessionCredentialsExpired())
        .describedAs("Credentials fetched the first time, should be expired "
            + "by now").isTrue();
    AWSSessionCredentials fetchedCredential =
        (AWSSessionCredentials) razAssumeRoleCredentialProvider.getCredentials();

    // Since, credentials would be expired, another assumeRole call needs to
    // happen.
    verify(mockStsClient, times(2)).assumeRole(any());
    assertFetchedCredentials(fetchedCredential);

    // Assert newly acquired credentials haven't expired yet.
    Assertions.assertThat(razAssumeRoleCredentialProvider.isAWSSessionCredentialsExpired())
        .describedAs("Credentials fetched the second time, shouldn't be "
            + "expired by now").isFalse();

    // No more interactions expected after another invocation.
    fetchedCredential = (AWSSessionCredentials) razAssumeRoleCredentialProvider.getCredentials();
    verify(mockStsClient, times(2)).assumeRole(any());
    verifyNoMoreInteractions(mockStsClient);
    assertFetchedCredentials(fetchedCredential);
  }

  /**
   * Method to verify the fetched credentials have the same fields as
   * acquired from assumeRoleResult credentials.
   * @param fetchedCredentials fetched credentials.
   */
  private void assertFetchedCredentials(AWSSessionCredentials fetchedCredentials) {

    // Verifying Session credentials fields.
    Assertions.assertThat(fetchedCredentials.getAWSAccessKeyId())
        .describedAs("AccessKey of AWSSessionCredentials should be same as "
            + "AssumeRoleResult")
        .isEqualTo(ACCESS_KEY_ID);
    Assertions.assertThat(fetchedCredentials.getAWSSecretKey())
        .describedAs("SecretKeyID of AWSSessionCredentials should be same as "
            + "AssumeRoleResult")
        .isEqualTo(SECRET_KEY_ID);
    Assertions.assertThat(fetchedCredentials.getSessionToken())
        .describedAs("SessionToken of AWSSessionCredentials should be same as "
            + "AssumeRoleResult")
        .isEqualTo(SESSION_TOKEN);
  }
}
