/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.processor.s3;

import org.junit.Before;
import org.junit.Test;

import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.raz.model.RangerRazResultBase;
import org.apache.ranger.raz.processor.s3.signer.S3RemoteSigner;
import org.apache.ranger.raz.processor.s3.signer.S3RemoteSignerContext;
import org.apache.ranger.raz.s3.lib.HttpMethod;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsOperationType;
import org.apache.ranger.raz.s3.lib.aws.fsrequests.AwsRequest;
import org.apache.ranger.raz.s3.lib.signer.SignRequest;
import org.apache.ranger.raz.s3.lib.utils.S3Utils;

import java.net.URI;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

public class TestS3RazProcessor {

  private Configuration config;
  private S3RemoteSignerContext s3RemoteSignerContext;
  private S3RemoteSigner s3RemoteSigner;
  private URI endpointUri;
  private static final String FAKE_BUCKET_NAME = "fakebucketname";
  private static final String REGION_NAME = "us-west-2";

  @Before
  public void setup() throws Exception {
    config = new Configuration();
    s3RemoteSignerContext = new S3RemoteSignerContext(config);
    s3RemoteSigner = new S3RemoteSigner(s3RemoteSignerContext, config);
    S3Utils.updateBucketToRegionCache(FAKE_BUCKET_NAME, REGION_NAME);
    endpointUri = new URI("https://" + FAKE_BUCKET_NAME + ".s3.amazonaws.com");
  }

  @Test
  public void testTrimTrailingSlashExceptRoot() {
    String root = "/";
    String rootExpectedResult = "/";

    String path1 = "/myPath/";
    String path1ExpectedResult = "/myPath";

    String path2 = "/myPath/isLonger/";
    String path2ExpectedResult = "/myPath/isLonger";

    String path3 = "/myPath/isWithoutTrailingSlash";
    String path3ExpectedResult = path3;

    assertEquals(rootExpectedResult, S3RazProcessor.trimTrailingSlashExceptRoot(root));
    assertEquals(path1ExpectedResult, S3RazProcessor.trimTrailingSlashExceptRoot(path1));
    assertEquals(path2ExpectedResult, S3RazProcessor.trimTrailingSlashExceptRoot(path2));
    assertEquals(path3ExpectedResult, S3RazProcessor.trimTrailingSlashExceptRoot(path3));
  }

  @Test
  public void testMaybeForceAllowedAccessResultForces() throws Exception {
    SignRequest signRequest = mock(SignRequest.class);
    when(signRequest.getEndpoint()).thenReturn(endpointUri);
    when(signRequest.getHttpMethod()).thenReturn(HttpMethod.HEAD);
    when(signRequest.getResourcePath()).thenReturn("");

    RangerRazResult rangerRazResult = mock(RangerRazResult.class);
    RangerRazResultBase.ResourceAccessResult accessResult = mock(RangerRazResultBase.ResourceAccessResult.class);
    when(rangerRazResult.getOperResult()).thenReturn(accessResult);

    S3RazProcessor.maybeForceAllowedAccessResult(signRequest, rangerRazResult, s3RemoteSigner);

    verify(rangerRazResult, times(1)).getOperResult();
    verifyNoMoreInteractions(rangerRazResult);
    verify(accessResult, times(1)).setResult(RangerRazResultBase.AccessResult.ALLOWED);
  }

  @Test
  public void testMaybeForceAllowedAccessResultSkips()
      throws RangerRazException {
    SignRequest signRequest = mock(SignRequest.class);
    when(signRequest.getEndpoint()).thenReturn(endpointUri);
    when(signRequest.getHttpMethod()).thenReturn(HttpMethod.HEAD);
    when(signRequest.getResourcePath()).thenReturn("/some/path");

    RangerRazResult rangerRazResult = mock(RangerRazResult.class);
    RangerRazResultBase.ResourceAccessResult accessResult = mock(RangerRazResultBase.ResourceAccessResult.class);
    when(rangerRazResult.getOperResult()).thenReturn(accessResult);

    S3RazProcessor.maybeForceAllowedAccessResult(signRequest, rangerRazResult, s3RemoteSigner);

    // we want no interactions in the HEAD and path case
    verifyNoMoreInteractions(accessResult);
    verifyNoMoreInteractions(rangerRazResult);
  }

  @Test
  public void testIsSignRequestHeadOnRoot() {
    AwsRequest awsRequest = mock(AwsRequest.class);
    when(awsRequest.getOperationType()).thenReturn(AwsOperationType.HEAD_BUCKET_METADATA);
    assertTrue("The sign request is HEAD_BUCKET_METADATA, so the returned value should be true.",
        S3RazProcessor.isSignRequestHeadOnRoot(awsRequest));

    when(awsRequest.getOperationType()).thenReturn(AwsOperationType.HEAD_OBJECT_METADATA);
    assertFalse("The sign request is HEAD_OBJECT_METADATA, so the returned value should be false.",
        S3RazProcessor.isSignRequestHeadOnRoot(awsRequest));
  }

  @Test
  public void testIsSignRequestHeadOnObject() {
    AwsRequest awsRequest = mock(AwsRequest.class);
    when(awsRequest.getOperationType()).thenReturn(AwsOperationType.HEAD_OBJECT_METADATA);
    assertTrue("The sign request is HEAD_OBJECT_METADATA, so the returned value should be true.",
        S3RazProcessor.isSignRequestHeadOnObject(awsRequest));

    when(awsRequest.getOperationType()).thenReturn(AwsOperationType.HEAD_BUCKET_METADATA);
    assertFalse("The sign request is HEAD_BUCKET_METADATA, so the returned value should be false.",
        S3RazProcessor.isSignRequestHeadOnObject(awsRequest));
  }

  @Test
  public void testIsSignRequestListOnObject() {
    AwsRequest awsRequest = mock(AwsRequest.class);
    when(awsRequest.getOperationType()).thenReturn(AwsOperationType.LIST_V1);
    assertTrue("The sign request is LIST_V1, so the returned value should be true.",
        S3RazProcessor.isSignRequestListOnObject(awsRequest));

    awsRequest = mock(AwsRequest.class);
    when(awsRequest.getOperationType()).thenReturn(AwsOperationType.LIST_V2);
    assertTrue("The sign request is LIST_V2, so the returned value should be true.",
        S3RazProcessor.isSignRequestListOnObject(awsRequest));

    awsRequest = mock(AwsRequest.class);
    when(awsRequest.getOperationType()).thenReturn(AwsOperationType.MULTIPART_LIST);
    assertTrue("The sign request is MULTIPART_LIST, so the returned value should be true.",
        S3RazProcessor.isSignRequestListOnObject(awsRequest));

    awsRequest = mock(AwsRequest.class);
    when(awsRequest.getOperationType()).thenReturn(AwsOperationType.HEAD_OBJECT_METADATA);
    assertFalse("The sign request is HEAD_OBJECT_METADATA, so the returned value should be false.",
        S3RazProcessor.isSignRequestListOnObject(awsRequest));
  }
}