/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.ranger.raz.processor.s3.signer;

import static org.apache.hadoop.fs.s3a.Constants.PROXY_HOST;
import static org.apache.hadoop.fs.s3a.Constants.PROXY_PORT;
import static org.apache.hadoop.fs.s3a.S3AUtils.initProxySupport;
import static org.apache.hadoop.fs.s3a.S3AUtils.propagateBucketOptions;
import static org.apache.ranger.raz.s3.lib.RazS3Constants.HEADER_USER_AGENT;
import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.mock;

import com.amazonaws.ClientConfiguration;
import com.amazonaws.DefaultRequest;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import org.junit.Before;
import org.junit.Test;
import org.powermock.core.classloader.annotations.PrepareForTest;

import org.apache.hadoop.conf.Configuration;

@PrepareForTest(TestS3RemoteSignerImpl.class)
public class TestS3RemoteSignerImpl {

  DefaultRequest toSignReq;
  S3RemoteSigner underTest;

  @Before
  public void setup() {
    S3RemoteSignerContext context = mock(S3RemoteSignerContext.class);

    Configuration configuration = new Configuration();

    underTest = new S3RemoteSigner(context, configuration);
    toSignReq = new DefaultRequest("test");
  }

  @Test
  public void testProxyConfig() throws Exception {
    Configuration configuration = new Configuration();
    String proxyHostName = "test_proxy";
    int proxyPort = 9080;
    configuration.set(PROXY_HOST, proxyHostName);
    configuration.setInt(PROXY_PORT, proxyPort);
    final ClientConfiguration awsConf = new ClientConfiguration();
    initProxySupport(configuration, "test_bucket", awsConf);
    assertEquals("Proxy host should be propagated properly",
            proxyHostName, awsConf.getProxyHost());
    assertEquals("Proxy port should be propagated properly",
            proxyPort, awsConf.getProxyPort());
  }

  @Test
  public void testPerBucketProxyConfig() throws IOException {
    Configuration configuration = new Configuration();
    String proxyHostName = "test_proxy";
    String proxyPort = "9080";
    configuration.set("fs.s3a.bucket.test_bucket.proxy.host", proxyHostName);
    configuration.set("fs.s3a.bucket.test_bucket.proxy.port", proxyPort);
    final ClientConfiguration awsConf = new ClientConfiguration();
    configuration = propagateBucketOptions(configuration, "test_bucket");
    initProxySupport(configuration, "test_bucket", awsConf);
    assertEquals("Proxy host should be propagated properly",
        proxyHostName, awsConf.getProxyHost());
    assertEquals("Proxy port should be propagated properly",
        Integer.parseInt(proxyPort), awsConf.getProxyPort());
  }

  @Test(timeout = 10_000L)
  public void testUpdateExistingUserAgent() {
    Map<String, String> headers = new HashMap<>();
    String userAgentInRequest = "Hadoop 3.0.0-SNAPSHOT, "
        + "aws-sdk-java/1.10.6 Mac_OS_X/10.9.5 "
        + "Java_HotSpot(TM)_64-Bit_Server_VM/24.65-b04/1.7.0_67";
    headers.put(HEADER_USER_AGENT, userAgentInRequest);
    toSignReq.setHeaders(headers);

    underTest.updateUserAgentInRequest(toSignReq, "kerberosuser");
    assertEquals("user:kerberosuser," + userAgentInRequest,
        toSignReq.getHeaders().get(HEADER_USER_AGENT));
  }

  @Test(timeout = 10_000L)
  public void testUpdateEmptyUserAgent() {
    Map<String, String> headers = new HashMap<>();
    headers.put(HEADER_USER_AGENT, "");
    toSignReq.setHeaders(headers);

    underTest.updateUserAgentInRequest(toSignReq, "kerberosuser");
    assertEquals("user:kerberosuser", toSignReq.getHeaders().get(HEADER_USER_AGENT));
  }

  @Test(timeout = 10_000L)
  public void testUpdateNullUserAgent() {
    Map<String, String> headers = new HashMap<>();
    toSignReq.setHeaders(headers);

    underTest.updateUserAgentInRequest(toSignReq, "kerberosuser");
    assertEquals("user:kerberosuser", toSignReq.getHeaders().get(HEADER_USER_AGENT));
  }
}