#!/usr/bin/env python

#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import os
import re
import platform
import subprocess
import sys
import time
import errno
import socket
from re import split
from time import sleep
from xml.dom import minidom


DIR_BIN                = "bin"
DIR_LIB                = "lib"
DIR_LIB_EXT            = "libext"
DIR_CONF               = "conf.dist"
DIR_LOG                = "logs"
ENV_JAVA_HOME              = "JAVA_HOME"
ENV_RANGER_RAZ_CONF        = "RANGER_RAZ_CONF_DIR"
ENV_RANGER_RAZ_HOME        = "RANGER_RAZ_HOME_DIR"
ENV_RANGER_RAZ_LOG         = "RANGER_RAZ_LOG_DIR"
ENV_RANGER_RAZ_PID         = "RANGER_RAZ_PID_DIR"
ENV_RANGER_RAZ_WEBAPP      = "RANGER_RAZ_EXPANDED_WEBAPP_DIR"
ENV_RANGER_RAZ_SERVER_OPTS = "RANGER_RAZ_SERVER_OPTS"
ENV_RANGER_RAZ_OPTS        = "RANGER_RAZ_OPTS"
ENV_RANGER_RAZ_SERVER_HEAP = "RANGER_RAZ_SERVER_HEAP"

ENV_KEYS = [ ENV_JAVA_HOME, ENV_RANGER_RAZ_CONF, ENV_RANGER_RAZ_LOG, ENV_RANGER_RAZ_PID, ENV_RANGER_RAZ_WEBAPP,
             ENV_RANGER_RAZ_SERVER_OPTS, ENV_RANGER_RAZ_OPTS, ENV_RANGER_RAZ_SERVER_HEAP, ENV_RANGER_RAZ_HOME ]

IS_WINDOWS                          = platform.system() == "Windows"
ON_POSIX                            = 'posix' in sys.builtin_module_names
CONF_FILE                           = "ranger-raz-site.xml"
CONF_RANGER_RAZ_HTTP_PORT           = "ranger.raz.service.http.port"
CONF_RANGER_RAZ_HTTPS_PORT          = "ranger.raz.service.https.port"
CONF_RANGER_RAZ_ENABLE_TLS          = "ranger.raz.service.enableTLS"
CONF_RANGER_RAZ_SERVER_BIND_ADDRESS = "ranger.raz.service.bind.address"
DEFAULT_RANGER_RAZ_HTTP_PORT        = "6181"
DEFAULT_RANGER_RAZ_HTTPS_PORT       = "-1"
DEFAULT_RANGER_RAZ_SERVER_HOST      = "localhost"

DEBUG = False

def scriptDir():
    return os.path.dirname(os.path.realpath(__file__))

def razHome():
    return os.environ.get(ENV_RANGER_RAZ_HOME, os.path.dirname(scriptDir()))

def libDir(razHome) :
    return os.path.join(razHome, DIR_LIB)

def libExtDir(razHome) :
    return os.path.join(razHome, DIR_LIB_EXT)

def confDir(razHome):
    return os.environ.get(ENV_RANGER_RAZ_CONF, os.path.join(razHome, DIR_CONF))

def logDir(razHome):
    return os.environ.get(ENV_RANGER_RAZ_LOG, os.path.join(razHome, DIR_LOG))

def pidFile(dir):
    return os.path.join(os.environ.get(ENV_RANGER_RAZ_PID, os.path.join(dir, DIR_LOG)), 'ranger-raz.pid')

def webAppDir(razHome):
    return os.environ.get(ENV_RANGER_RAZ_WEBAPP, os.path.join(razHome, "webapp"))

def expandWebApp(razHome):
    webappDir   = webAppDir(razHome)
    expandedDir = os.path.join(webappDir, "ranger-raz")
    d           = os.sep

    if not os.path.exists(os.path.join(expandedDir, "WEB-INF")):
        try:
            os.makedirs(expandedDir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise e
            pass

        razWarPath = os.path.join(webappDir, "ranger-raz.war")

        if (isCygwin()):
            razWarPath = convertCygwinPath(razWarPath)

        os.chdir(expandedDir)

        jar(razWarPath)

def dirMustExist(dirname):
    if not os.path.exists(dirname):
        os.mkdir(dirname)

    return dirname

def executeEnvSh(confDir):
    envscript = '%s/ranger-raz-env.sh' % confDir

    if not IS_WINDOWS and os.path.exists(envscript):
        envCmd  = 'source %s && env' % envscript
        command = ['bash', '-c', envCmd]

        proc = subprocess.Popen(command, stdout = subprocess.PIPE)

        for line in proc.stdout:
            (key, _, value) = line.strip().partition("=")

            if key in ENV_KEYS:
                os.environ[key] = value

        proc.communicate()

def java(classname, args, classpath, jvm_opts_list, logdir=None):
    java_home = os.environ.get(ENV_JAVA_HOME, None)

    if java_home:
        prg = os.path.join(java_home, "bin", "java")
    else:
        prg = which("java")

    if prg is None:
        raise EnvironmentError('The java binary could not be found in your path or JAVA_HOME')

    commandline = [prg]
    commandline.extend(jvm_opts_list)
    commandline.append("-classpath")
    commandline.append(classpath)
    commandline.append(classname)
    commandline.extend(args)

    return runProcess(commandline, logdir)

def jar(path):
    java_home = os.environ.get(ENV_JAVA_HOME, None)

    if java_home:
        prg = os.path.join(java_home, "bin", "jar")
    else:
        prg = which("jar")

    if prg is None:
        raise EnvironmentError('The jar binary could not be found in your path or JAVA_HOME')

    commandline = [prg]
    commandline.append("-xf")
    commandline.append(path)

    process = runProcess(commandline)
    process.wait()

def is_exe(fpath):
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

def which(program):
    fpath, fname = os.path.split(program)

    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path     = path.strip('"')
            exe_file = os.path.join(path, program)

            if is_exe(exe_file):
                return exe_file

    return None

def runProcess(commandline, logdir=None, shell=False, wait=False):
    """
    Run a process
    :param commandline: command line
    :return:the return code
    """
    global finished
    debug ("Executing : %s" % str(commandline))
    timestr = time.strftime("ranger-raz.%Y%m%d-%H%M%S")
    stdoutFile = None
    stderrFile = None
    if logdir:
        stdoutFile = open(os.path.join(logdir, timestr + ".out"), "w")
        stderrFile = open(os.path.join(logdir,timestr + ".err"), "w")

    p = subprocess.Popen(commandline, stdout=stdoutFile, stderr=stderrFile, shell=shell)

    if wait:
        p.communicate()

    return p

def print_output(name, src, toStdErr):
    """
    Relay the output stream to stdout line by line
    :param name:
    :param src: source stream
    :param toStdErr: flag set if stderr is to be the dest
    :return:
    """

    global needPassword
    debug ("starting printer for %s" % name )
    line = ""
    while not finished:
        (line, done) = read(src, line)
        if done:
            out(toStdErr, line + "\n")
            flush(toStdErr)
            if line.find("Enter password for") >= 0:
                needPassword = True
            line = ""
    out(toStdErr, line)
    # closedown: read remainder of stream
    c = src.read(1)
    while c!="" :
        c = c.decode('utf-8')
        out(toStdErr, c)
        if c == "\n":
            flush(toStdErr)
        c = src.read(1)
    flush(toStdErr)
    src.close()

def read_input(name, exe):
    """
    Read input from stdin and send to process
    :param name:
    :param process: process to send input to
    :return:
    """
    global needPassword
    debug ("starting reader for %s" % name )
    while not finished:
        if needPassword:
            needPassword = False
            if sys.stdin.isatty():
                cred = getpass.getpass()
            else:
                cred = sys.stdin.readline().rstrip()
            exe.stdin.write(cred + "\n")

def debug(text):
    if DEBUG: print('[DEBUG] ' + text)
    print('[DEBUG] ' + text)


def error(text):
    print('[ERROR] ' + text)
    sys.stdout.flush()

def info(text):
    print(text)
    sys.stdout.flush()


def out(toStdErr, text) :
    """
    Write to one of the system output channels.
    This action does not add newlines. If you want that: write them yourself
    :param toStdErr: flag set if stderr is to be the dest
    :param text: text to write.
    :return:
    """
    if toStdErr:
        sys.stderr.write(text)
    else:
        sys.stdout.write(text)

def flush(toStdErr) :
    """
    Flush the output stream
    :param toStdErr: flag set if stderr is to be the dest
    :return:
    """
    if toStdErr:
        sys.stderr.flush()
    else:
        sys.stdout.flush()

def read(pipe, line):
    """
    read a char, append to the listing if there is a char that is not \n
    :param pipe: pipe to read from
    :param line: line being built up
    :return: (the potentially updated line, flag indicating newline reached)
    """

    c = pipe.read(1)
    if c != "":
        o = c.decode('utf-8')
        if o != '\n':
            line += o
            return line, False
        else:
            return line, True
    else:
        return line, False

def writePid(atlas_pid_file, process):
    f = open(atlas_pid_file, 'w')
    f.write(str(process.pid))
    f.close()

def exist_pid(pid):
    if  ON_POSIX:
        #check if process id exist in the current process table
        #See man 2 kill - Linux man page for info about the kill(pid,0) system function
        try:
            os.kill(pid, 0)
        except OSError as e :
            return e.errno == errno.EPERM
        else:
            return True

    elif IS_WINDOWS:
        #The os.kill approach does not work on Windows with python 2.7
        #the output from tasklist command is searched for the process id
        pidStr = str(pid)
        command='tasklist /fi  "pid eq %s"' % pidStr
        sub_process=subprocess.Popen(command, stdout = subprocess.PIPE, shell=False)
        sub_process.communicate()
        output = subprocess.check_output(command)
        output=split(" *",output)
        for line in output:
            if pidStr in line:
                return True
        return False
    #os other than nt or posix - not supported - need to delete the file to restart server if pid no longer exist
    return True

def wait_for_shutdown(pid, msg, wait):
    count = 0
    sys.stdout.write(msg)
    while exist_pid(pid):
        sys.stdout.write('.')
        sys.stdout.flush()
        sleep(1)
        if count > wait:
            break
        count = count + 1

    sys.stdout.write('\n')

def get_raz_url_port(confdir):
    port = None

    if '-port' in sys.argv:
        port = sys.argv[sys.argv.index('-port') + 1]

    if port is None:
        confdir    = os.path.join(confdir, CONF_FILE)
        enable_tls = getConfig(confdir, CONF_RANGER_RAZ_ENABLE_TLS)

        if enable_tls is not None and enable_tls.lower() == 'true':
            port = getConfigWithDefault(confdir, CONF_RANGER_RAZ_HTTPS_PORT, DEFAULT_RANGER_RAZ_HTTPS_PORT)
        else:
            port = getConfigWithDefault(confdir, CONF_RANGER_RAZ_HTTP_PORT, DEFAULT_RANGER_RAZ_HTTP_PORT)

    return port

def get_raz_url_host(confdir):
    confdir = os.path.join(confdir, CONF_FILE)
    host    = getConfigWithDefault(confdir, CONF_RANGER_RAZ_SERVER_BIND_ADDRESS, DEFAULT_RANGER_RAZ_SERVER_HOST)

    if (host == '0.0.0.0'):
        host = DEFAULT_RANGER_RAZ_SERVER_HOST

    return host

def wait_for_startup(confdir, wait):
    count = 0
    host  = get_raz_url_host(confdir)
    port  = get_raz_url_port(confdir)

    print("Waiting for Ranger Raz to listen on host %s:%s" % (host, port))

    while True:
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            s.settimeout(1)
            s.connect((host, int(port)))
            s.close()

            break
        except Exception as e:
            # Wait for 1 sec before next ping
            sys.stdout.write('.')
            sys.stdout.flush()

            sleep(1)

        if count > wait:
            s.close()

            break

        count = count + 1

    sys.stdout.write('\n')

def server_already_running(pid):
    print("Ranger Raz server is already running under process %s" % pid)
    sys.exit()

def server_pid_not_running(pid):
    print("Ranger Raz Server is no longer running with pid %s" %pid)

def getConfig(file, key):
    ret  = None
    conf = minidom.parse(file)

    for property in conf.getElementsByTagName('property'):
        nameNode = property.getElementsByTagName('name').item(0).firstChild

        if nameNode == None:
            continue

        name = nameNode.data.strip()

        if name == key:
            valueNode = property.getElementsByTagName('value').item(0).firstChild
            if valueNode != None:
                ret = valueNode.data.strip()
                break

    print("%s=%s" % (key, ret))

    return ret

def getConfigWithDefault(file, key, defaultValue):
    value = getConfig(file, key)
    if value is None:
        value = defaultValue
    return value

def isCygwin():
    return platform.system().startswith("CYGWIN")

# Convert the specified cygwin-style pathname to Windows format,
# using the cygpath utility.  By default, path is assumed
# to be a file system pathname.  If isClasspath is True,
# then path is treated as a Java classpath string.
def convertCygwinPath(path, isClasspath = False):
    if isClasspath:
        cygpathArgs = ["cygpath", "-w", "-p", path]
    else:
        cygpathArgs = ["cygpath", "-w", path]

    windowsPath = subprocess.Popen(cygpathArgs, stdout=subprocess.PIPE).communicate()[0]
    windowsPath = windowsPath.strip()

    return windowsPath
