#!/usr/bin/env python

#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import traceback
import raz_config as conf


RANGER_RAZ_LOG_OPTS     = "-Dranger.raz.log.dir=%s -Dranger.raz.log.file=%s.log"
RANGER_RAZ_COMMAND_OPTS = "-Dranger.raz.home=%s -Dranger.raz.webapp.dir=%s"
RANGER_RAZ_CONFIG_OPTS  = "-Dranger.raz.conf=%s"
DEFAULT_JVM_HEAP_OPTS   = "-Xmx1024m"
DEFAULT_JVM_OPTS        = "-Dlog4j.configuration=ranger-raz-log4j.properties -Djava.net.preferIPv4Stack=true -server"


def main():
    razHome = conf.razHome()
    confDir = conf.dirMustExist(conf.confDir(razHome))

    conf.executeEnvSh(confDir)

    logDir     = conf.dirMustExist(conf.logDir(razHome))
    webappDir  = conf.webAppDir(razHome)
    webappPath = os.path.join(webappDir, "ranger-raz")

    #expand web app dir
    conf.expandWebApp(razHome)

    if conf.isCygwin():
        # Pathnames that are passed to JVM must be converted to Windows format.
        jvmRazHome = conf.convertCygwinPath(razHome)
        jvmConfDir = conf.convertCygwinPath(confDir)
        jvmLogDir  = conf.convertCygwinPath(logDir)
    else:
        jvmRazHome = razHome
        jvmConfDir = confDir
        jvmLogDir  = logDir

    cmdOpts          = (RANGER_RAZ_COMMAND_OPTS % (jvmRazHome, webappPath))
    confgOpts        = (RANGER_RAZ_CONFIG_OPTS % jvmConfDir)
    razHeapOpts      = os.environ.get(conf.ENV_RANGER_RAZ_SERVER_HEAP, DEFAULT_JVM_HEAP_OPTS)
    razServerJvmOpts = os.environ.get(conf.ENV_RANGER_RAZ_SERVER_OPTS)
    razJvmOpts       = os.environ.get(conf.ENV_RANGER_RAZ_OPTS, DEFAULT_JVM_OPTS)

    jvmOptsList = (RANGER_RAZ_LOG_OPTS % (jvmLogDir, "ranger-raz")).split()

    jvmOptsList.extend(cmdOpts.split())
    jvmOptsList.extend(confgOpts.split())
    jvmOptsList.extend(razHeapOpts.split())

    if razServerJvmOpts:
        jvmOptsList.extend(razServerJvmOpts.split())

    jvmOptsList.extend(razJvmOpts.split())

    p = os.pathsep
    razClasspath = confDir + p \
                       + os.path.join(webappDir, "ranger-raz", "WEB-INF", "classes" ) + p \
                       + os.path.join(webappDir, "ranger-raz", "WEB-INF", "lib", "*" ) 

    if conf.isCygwin():
        razClasspath = conf.convertCygwinPath(razClasspath, True)

    razPidFile = conf.pidFile(razHome)

    if os.path.isfile(razPidFile):
       # Check if process listed in ranger.raz.pid file is still running
       pf  = file(razPidFile, 'r')
       pid = pf.read().strip()

       pf.close()

       if pid != "":
           if conf.exist_pid((int)(pid)):
               conf.server_pid_not_running(pid)

    if (conf.isCygwin()):
        webappPath = conf.convertCygwinPath(webappPath)

    start_raz_server(razClasspath, razPidFile, jvmLogDir, jvmOptsList, webappPath)

    conf.wait_for_startup(confDir, 300)

    print("Ranger RAZ Server started!!!\n")


def start_raz_server(razClasspath, razPidFile, jvmLogDir, jvmOptsList, webappPath):
    args = ["-app", webappPath]

    args.extend(sys.argv[1:])

    process = conf.java("org.apache.ranger.raz.server.RangerRazServer", args, razClasspath, jvmOptsList, jvmLogDir)

    conf.writePid(razPidFile, process)

if __name__ == '__main__':
    try:
        returncode = main()
    except Exception as e:
        print("Exception: %s " % str(e))
        print(traceback.format_exc())

        returncode = -1

    sys.exit(returncode)
