/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz;

import org.apache.ranger.audit.model.AuthzAuditEvent;
import org.apache.ranger.plugin.audit.RangerDefaultAuditHandler;
import org.apache.ranger.plugin.policyengine.RangerAccessResult;
import org.apache.ranger.plugin.policyengine.RangerAccessResultProcessor;
import org.apache.ranger.raz.util.RangerRazConstants;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class RangerRazAuditHandler implements RangerAccessResultProcessor {
    private final RangerDefaultAuditHandler defaultHandler;
    private final String                    fallbackAclEnforcer;
    private final List<AuthzAuditEvent>     auditEvents  = new ArrayList<>();
    private       boolean                   deniedExists = false;
    private final boolean                   generateSingleAuditLog = true;


    public RangerRazAuditHandler(RangerDefaultAuditHandler defaultHandler, String fallbackAclEnforcer) {
        this.defaultHandler = defaultHandler;
        this.fallbackAclEnforcer    = fallbackAclEnforcer;
    }

    @Override
    public void processResult(RangerAccessResult result) {
        generateAuditEvent(result);
    }

    @Override
    public void processResults(Collection<RangerAccessResult> results) {
        for (RangerAccessResult result : results) {
            generateAuditEvent(result);
        }
    }

    private void generateAuditEvent(RangerAccessResult result) {
        AuthzAuditEvent auditEvent = defaultHandler.getAuthzEvents(result);
        if(auditEvent != null) {
            if (!result.getIsAccessDetermined()) {
                auditEvent.setAccessResult(RangerRazConstants.ACCESS_NOT_DETERMINED); //To record result is not determined by RangerRaz and falling back to native ACL.
            }
            addAuthzAuditEvent(auditEvent);
        }
    }

    public void setAccessDenied() {
        deniedExists = true;

        for (AuthzAuditEvent auditEvent : auditEvents) {
            if (auditEvent.getAccessResult() != RangerRazConstants.ACCESS_DENIED) {
                auditEvent.setAccessResult(RangerRazConstants.ACCESS_DENIED);
                auditEvent.setPolicyId(-1);
            }
        }
    }

    public void flushAudit() {
        for (AuthzAuditEvent auditEvent : auditEvents) {
            if (deniedExists && auditEvent.getAccessResult() == RangerRazConstants.ACCESS_ALLOWED) { // if deny exists, skip logging for allowed results
                continue;
            }

            if (auditEvent.getAccessResult() == RangerRazConstants.ACCESS_NOT_DETERMINED) {
                auditEvent.setAclEnforcer(fallbackAclEnforcer);
            }

            defaultHandler.logAuthzAudit(auditEvent);

            if (generateSingleAuditLog) {
                break;
            }
        }
    }

    private void addAuthzAuditEvent(AuthzAuditEvent auditEvent) {
        if(auditEvent != null) {
            if(auditEvent.getAccessResult() == RangerRazConstants.ACCESS_DENIED) {
                deniedExists = true;
            }

            auditEvents.add(auditEvent);
        }
    }
}
