/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.raz;

import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.authorization.hadoop.config.RangerPluginConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Iterator;
import java.util.Map;

public class RangerRazPluginConfig extends RangerPluginConfig {
    private static final Logger LOG = LoggerFactory.getLogger(RangerRazPluginConfig.class);

    public RangerRazPluginConfig(String serviceType, String serviceName, Configuration razConfig) {
        super(serviceType, serviceName, "raz", null, null, null);

        // add necessary config "overrides", so that RangerAdminClient implementations (like RangerAdminRESTClient)
        // will use configurations from ranger-raz-site.xml (razConfig) to connect to Ranger Admin

        set(getPropertyPrefix() + ".service.name", serviceName);
        copyProperty(razConfig, ".policy.source.impl");
        copyProperty(razConfig, ".policy.cache.dir");
        copyProperty(razConfig, ".policy.rest.url");
        copyProperty(razConfig, ".policy.rest.ssl.config.file");
        copyProperty(razConfig, ".policy.pollIntervalMs", 30 * 1000);
        copyProperty(razConfig, ".policy.rest.client.connection.timeoutMs", 120 * 1000);
        copyProperty(razConfig, ".policy.rest.read.timeoutMs", 30 * 1000);
        copyProperty(razConfig, ".policy.rest.supports.policy.deltas", false);
        copyProperty(razConfig, ".tag.rest.supports.tag.deltas", false);
        copyProperty(razConfig, ".super.users");
        copyProperty(razConfig, ".super.groups");
        copyProperty(razConfig, ".audit.exclude.users");
        copyProperty(razConfig, ".audit.exclude.groups");
        copyProperty(razConfig, ".audit.exclude.roles");

        copyServiceTypeConfig(razConfig, serviceType);
        copyServiceConfig(razConfig, serviceName);
        copyPluginConfig(razConfig, "azure");
        copyPluginConfig(razConfig, "s3a");
    }

    private void copyProperty(Configuration razConfig, String propertySuffix) {
        String val = razConfig.get("ranger.raz" + propertySuffix);

        if (val != null) {
            set(getPropertyPrefix() + propertySuffix, val);
        }
    }

    private void copyProperty(Configuration razConfig, String propertySuffix, int defaultValue) {
        setInt(getPropertyPrefix() + propertySuffix, razConfig.getInt("ranger.raz" + propertySuffix, defaultValue));
    }

    private void copyProperty(Configuration razConfig, String propertySuffix, boolean defaultValue) {
        setBoolean(getPropertyPrefix() + propertySuffix, razConfig.getBoolean("ranger.raz" + propertySuffix, defaultValue));
    }

    private void copyServiceTypeConfig(Configuration razConfig, String serviceType) {
        String svcPrefix = "ranger.raz.service-type." + serviceType + ".";

        for (Iterator<Map.Entry<String, String>> iter = razConfig.iterator(); iter.hasNext(); ) {
            Map.Entry<String, String> entry = iter.next();
            String                    key   = entry.getKey();

            if (key.startsWith(svcPrefix)) {
                String suffix   = key.substring(svcPrefix.length());
                String cfgName  = getPropertyPrefix() + "." + suffix;
                String cfgValue = entry.getValue();

                if (LOG.isDebugEnabled()) {
                    LOG.debug("{}: adding service-type config: {}={}", serviceType, cfgName, cfgValue);
                }

                set(cfgName, cfgValue);
            }
        }
    }

    private void copyServiceConfig(Configuration razConfig, String serviceName) {
        String svcPrefix = "ranger.raz.service." + serviceName + ".";

        for (Iterator<Map.Entry<String, String>> iter = razConfig.iterator(); iter.hasNext(); ) {
            Map.Entry<String, String> entry = iter.next();
            String                    key   = entry.getKey();

            if (key.startsWith(svcPrefix)) {
                String suffix   = key.substring(svcPrefix.length());
                String cfgName  = getPropertyPrefix() + "." + suffix;
                String cfgValue = entry.getValue();

                if (LOG.isDebugEnabled()) {
                    LOG.debug("{}: adding service config: {}={}", serviceName, cfgName, cfgValue);
                }

                set(cfgName, cfgValue);
            }
        }
    }

	private void copyPluginConfig(final Configuration razConfig,
      String pluginType) {
		String svcPrefix = "fs." + pluginType + ".ext.raz." + getServiceType();
		for (Iterator<Map.Entry<String, String>> iter = razConfig.iterator(); iter.hasNext();) {
			Map.Entry<String, String> entry = iter.next();
			String key = entry.getKey();
			if (key.startsWith(svcPrefix)) {
				String suffix = key.substring(svcPrefix.length());
				String cfgName = getPropertyPrefix() + suffix;
				String cfgValue = entry.getValue();

				if (LOG.isDebugEnabled()) {
					LOG.debug("RAZ - adding plugin config: {}={}", cfgName, cfgValue);
				}

				set(cfgName, cfgValue);
			}
		}
	}
}
