/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz;

import static org.apache.ranger.raz.server.EmbeddedServer.DEFAULT_NAME_RULE;
import static org.apache.ranger.raz.server.EmbeddedServer.HADOOP_AUTHENTICATION_TYPE;
import static org.apache.ranger.raz.server.EmbeddedServer.HADOOP_AUTH_TYPE_KERBEROS;
import static org.apache.ranger.raz.server.EmbeddedServer.HADOOP_KERBEROS_NAME_RULES;
import static org.apache.ranger.raz.server.EmbeddedServer.RANGER_KERBEROS_KEYTAB;
import static org.apache.ranger.raz.server.EmbeddedServer.RANGER_KERBEROS_PRINCIPAL;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;

import javax.annotation.PostConstruct;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.SecureClientLogin;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.audit.provider.AuditProviderFactory;
import org.apache.ranger.audit.provider.MiscUtil;
import org.apache.ranger.authorization.hadoop.config.RangerAuditConfig;
import org.apache.ranger.authorization.hadoop.config.RangerPluginConfig;
import org.apache.ranger.plugin.audit.RangerDefaultAuditHandler;
import org.apache.ranger.plugin.contextenricher.RangerRESTClientUserStoreRetriever;
import org.apache.ranger.plugin.contextenricher.RangerUserStoreRefresher;
import org.apache.ranger.plugin.contextenricher.RangerUserStoreRetriever;
import org.apache.ranger.plugin.policyengine.RangerAccessRequest;
import org.apache.ranger.plugin.policyengine.RangerAccessResult;
import org.apache.ranger.plugin.service.RangerBasePlugin;
import org.apache.ranger.plugin.store.RangerUserStoreListener;
import org.apache.ranger.plugin.util.DownloadTrigger;
import org.apache.ranger.plugin.util.DownloaderTask;
import org.apache.ranger.plugin.util.RangerRESTClient;
import org.apache.ranger.plugin.util.RangerUserStore;
import org.apache.ranger.plugin.util.RangerUserStoreUtil;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.model.RangerRazMultiOperationRequest;
import org.apache.ranger.raz.model.RangerRazMultiOperationResult;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.apache.ranger.raz.model.RangerRazRequestBase.ResourceAccess;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.raz.model.RangerRazIdentityMappings;
import org.apache.ranger.raz.model.RangerRazResultBase.ResourceAccessResult;
import org.apache.ranger.raz.processor.RangerDefaultRazProcessor;
import org.apache.ranger.raz.processor.RangerRazContext;
import org.apache.ranger.raz.processor.RangerRazProcessor;
import org.apache.ranger.raz.processor.RangerRazRequestContext;
import org.apache.ranger.raz.server.RangerRazConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

@Component
public class RangerRemoteAuthorizer implements RangerUserStoreListener {
    private static final Logger LOG = LoggerFactory.getLogger(RangerRemoteAuthorizer.class);

    private static final String RAZ_CONFIG_PREFIX = "ranger.raz.";
    private static final String USERSTORE_REFRESHER_POLLINGINTERVAL_OPTION = "userStoreRefresherPollingInterval";

    private static final long   RAZ_TGT_RENEW_INTERVAL_DEFAULT = 4;
    private static final String RAZ_CONF_TGT_RENEW_INTERVAL    = RAZ_CONFIG_PREFIX+"tgt.renew.interval.in-mins";
    private static final String CONF_AUDIT_ID_STRICT_UUID      = "xasecure.audit.auditid.strict.uuid";
    private static final Object RANGER_USER_STORE_LOCK         = new Object();

    private final RangerRazContext                context;
    private final Map<String, RangerRazProcessor> razProcessors       = new HashMap<>();
    private final RangerRazProcessor              defaultRazProcessor = new RangerDefaultRazProcessor();
    private final Map<String, RangerBasePlugin>   servicePlugins      = new HashMap<>();
    private final RangerRazConfig                 rangerRazConfig     = new RangerRazConfig();
    private final Set<String>                     defaultKeysToRedact = new HashSet<>(Arrays.asList("ADLS_DSAS", "S3_SIGN_RESPONSE"));
    private       Set<String>                     keysToRedact        = new HashSet<>(defaultKeysToRedact);
    private       RangerDefaultAuditHandler       defaultAuditHandler;
    private       RangerUserStore                 rangerUserStore;
    private       RangerUserStoreRefresher        userStoreRefresher;
    private       Timer                           userStoreDownloadTimer;
    private final RangerUserStoreUtil             rangerUserStoreUtil = new RangerUserStoreUtil();
    private RangerRazIdentityMappings             cachedRazIdMappings = new RangerRazIdentityMappings();

    public RangerRemoteAuthorizer() {
        this.context = new RangerRazContext(rangerRazConfig);
    }

    @PostConstruct
    public void init() {
        LOG.info("==> RangerRemoteAuthorizer.init()");

        initKerberos();

        this.defaultAuditHandler = initAudit();
        this.rangerUserStore    = new RangerUserStore();
        initRangerUserStore();

        initProcessors();

        startTgtRenewer();

        if (this.cachedRazIdMappings.getVersion() == null) {
            this.cachedRazIdMappings.setVersion(Long.valueOf(-1L));
        }

        LOG.info("<== RangerRemoteAuthorizer.init()");
    }

    public RangerRazResult authorize(String serviceType, RangerRazRequest request, RangerRazRequestContext context) throws RangerRazException {
        if (LOG.isTraceEnabled()) {
            LOG.trace("==> RangerRemoteAuthorizer.authorize({}, {}, {})", serviceType, request, context);
        }

        context.setRangerUserStore(rangerUserStore);

        RangerRazProcessor          razProcessor        = getOrCreateRazProcessor(serviceType);
        List<RangerAccessRequest>   accessRequests      = razProcessor.preProcess(request, context);
        RangerBasePlugin            servicePlugin       = getOrCreateServicePlugin(serviceType, request.getServiceName());
        List<RangerAccessResult>    accessResults       = new ArrayList<>();
        String                      fallbackAclEnforcer = razProcessor.getFallbackAclEnforcerName();
        RangerRazAuditHandler       auditHandler        = new RangerRazAuditHandler(defaultAuditHandler, fallbackAclEnforcer);

        for (RangerAccessRequest accessRequest : accessRequests) {
            RangerAccessResult accessResult = servicePlugin.isAccessAllowed(accessRequest, auditHandler);

            accessResults.add(accessResult);
        }

        context.setKeysToRedact(keysToRedact);

        try {
            RangerRazResult ret = razProcessor.postProcess(request, accessResults, context);

            if (LOG.isTraceEnabled()) {
                LOG.trace("<== RangerRemoteAuthorizer.authorize({}, {}, {}): ret={}", serviceType, request, context, ret);
            }

            return ret;
        } catch (Exception excp) {
            auditHandler.setAccessDenied();

            throw excp;
        } finally {
            auditHandler.flushAudit();
        }
    }

    public RangerRazMultiOperationResult authorize(String serviceType, RangerRazMultiOperationRequest requests, RangerRazRequestContext context) throws RangerRazException {
        if (LOG.isTraceEnabled()) {
            LOG.trace("==> RangerRemoteAuthorizer.authorize({}, {}, {})", serviceType, requests, context);
        }

        context.setRangerUserStore(rangerUserStore);

        RangerRazMultiOperationResult ret = new RangerRazMultiOperationResult(requests.getRequestId());

        for (ResourceAccess resourceAccess : requests.getOperations()) {
            RangerRazRequest           request             = new RangerRazRequest(requests, resourceAccess);
            RangerRazProcessor         razProcessor        = getOrCreateRazProcessor(serviceType);
            List<RangerAccessRequest>  accessRequests      = razProcessor.preProcess(request, context);
            RangerBasePlugin           servicePlugin       = getOrCreateServicePlugin(serviceType, requests.getServiceName());
            List<RangerAccessResult>   accessResults       = new ArrayList<>();
            String                     fallbackAclEnforcer = razProcessor.getFallbackAclEnforcerName();
            RangerRazAuditHandler      auditHandler        = new RangerRazAuditHandler(defaultAuditHandler, fallbackAclEnforcer);

            for (RangerAccessRequest accessRequest : accessRequests) {
                RangerAccessResult accessResult = servicePlugin.isAccessAllowed(accessRequest, auditHandler);

                accessResults.add(accessResult);
            }

            context.setKeysToRedact(keysToRedact);

            try {
                RangerRazResult      result = razProcessor.postProcess(request, accessResults, context);
                ResourceAccessResult rar    = result != null ? result.getOperResult() : null;

                ret.addOperResult(rar);
            } catch (Exception excp) {
                auditHandler.setAccessDenied();

                throw excp;
            } finally {
                auditHandler.flushAudit();
            }
        }

        if (LOG.isTraceEnabled()) {
            LOG.trace("<== RangerRemoteAuthorizer.authorize({}, {}, {}): ret={}", serviceType, requests, context, ret);
        }

        return ret;
    }

    private void initKerberos() {
        LOG.info("==> RangerRemoteAuthorizer.initKerberos()");

        Configuration config = context.getRazConfig();

        String  authenticationType = config.get(HADOOP_AUTHENTICATION_TYPE);
        boolean isKerberosAuth     = StringUtils.equals(authenticationType, HADOOP_AUTH_TYPE_KERBEROS);

        if (isKerberosAuth) {
            String hostName      = config.get(RAZ_CONFIG_PREFIX + "service.host");
            String keyTab        = config.get(RAZ_CONFIG_PREFIX + RANGER_KERBEROS_KEYTAB);
            String kerbPrincipal = config.get(RAZ_CONFIG_PREFIX + RANGER_KERBEROS_PRINCIPAL);
            String nameRules     = config.get(HADOOP_KERBEROS_NAME_RULES, DEFAULT_NAME_RULE);

            LOG.info("RangerRemoteAuthorizer.initKerberos(): kerberos is enabled");

            UserGroupInformation.setConfiguration(config);

            try {
                String  principal = SecureClientLogin.getPrincipal(kerbPrincipal, hostName);
                MiscUtil.loginWithKeyTab(keyTab, principal, nameRules);
            } catch (IOException excp) {
                LOG.warn("RangerRemoteAuthorizer.initKerberos(): failed to get principal" + excp);
            }
        } else {
            LOG.info("RangerRemoteAuthorizer.initKerberos(): kerberos is not enabled");

        }

        LOG.info("UserGroupInformation.isSecurityEnabled()={}, MiscUtil.getUGILoginUser()={}", UserGroupInformation.isSecurityEnabled(), MiscUtil.getUGILoginUser());

        LOG.info("<== RangerRemoteAuthorizer.initKerberos()");
    }

    private RangerDefaultAuditHandler initAudit() {
        LOG.info("==> RangerRemoteAuthorizer.initAudit()");

        RangerAuditConfig   auditConfig = new RangerAuditConfig("raz");
        if (!auditConfig.getProperties().containsKey(CONF_AUDIT_ID_STRICT_UUID)) {
            auditConfig.setBoolean(CONF_AUDIT_ID_STRICT_UUID, true);
        }
        LOG.info("Using {}={}", CONF_AUDIT_ID_STRICT_UUID, auditConfig.getBoolean(CONF_AUDIT_ID_STRICT_UUID,false));

        auditConfig.addResource("ranger-raz-policymgr-ssl.xml");

        AuditProviderFactory auditProviderFactory = AuditProviderFactory.getInstance();
        auditProviderFactory.init(auditConfig.getProperties(), "raz");

        RangerDefaultAuditHandler ret = new RangerDefaultAuditHandler(auditConfig);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRemoteAuthorizer.initAudit()");
        }

        LOG.info("<== RangerRemoteAuthorizer.initAudit()");

        return ret;
    }

    private void initRangerUserStore() {
        LOG.info("==> RangerRemoteAuthorizer.initRangerUserStore()");

       Configuration   config                   = context.getRazConfig();
        boolean         downloadUserStoreEnabled = config.getBoolean(RAZ_CONFIG_PREFIX + "userstore.download.enabled", true);

        if (downloadUserStoreEnabled) {
            String url               = "";
            String tmpUrl            = config.get(RAZ_CONFIG_PREFIX + "policy.rest.url");
            String sslConfigFileName = config.get(RAZ_CONFIG_PREFIX + "policy.rest.ssl.config.file");

            if (!StringUtils.isEmpty(tmpUrl)) {
                url = tmpUrl.trim();
            }

            if (url.endsWith("/")) {
                url = url.substring(0, url.length() - 1);
            }

            RangerRESTClient restClient    = new RangerRESTClient(url, sslConfigFileName, config);
            String           cacheDir      = config.get(RAZ_CONFIG_PREFIX + "policy.cache.dir", null);
            String           cacheFilename = String.format("%s_userstore.json", "raz");

            cacheFilename = cacheFilename.replace(File.separatorChar,  '_');
            cacheFilename = cacheFilename.replace(File.pathSeparatorChar,  '_');

            String                         cacheFile              = cacheDir == null ? null : (cacheDir + File.separator + cacheFilename);
            BlockingQueue<DownloadTrigger> userStoreDownloadQueue = new LinkedBlockingQueue<>();
            long                           pollingIntervalMs      = config.getLong(RAZ_CONFIG_PREFIX + USERSTORE_REFRESHER_POLLINGINTERVAL_OPTION, 3600 * 1000);

            LOG.info("RangerRemoteAuthorizer.initRangerUserStore(): cacheFile={}; pollingInterval={}ms", cacheFile, pollingIntervalMs);

            RangerUserStoreRetriever rangerUserStoreRetriever = new RangerRESTClientUserStoreRetriever(restClient);

            userStoreRefresher = new RangerUserStoreRefresher(rangerUserStoreRetriever, null, this, -1L, userStoreDownloadQueue, cacheFile);

            try {
                userStoreRefresher.populateUserStoreInfo();
            } catch (Throwable exception) {
                LOG.error("Exception when retrieving userstore information for this enricher", exception);
            }

            userStoreRefresher.setDaemon(true);
            userStoreRefresher.startRefresher();

            userStoreDownloadTimer = new Timer("userStoreDownloadTimer", true);

            try {
                userStoreDownloadTimer.schedule(new DownloaderTask(userStoreDownloadQueue), pollingIntervalMs, pollingIntervalMs);

                if (LOG.isDebugEnabled()) {
                    LOG.debug("Scheduled userStoreDownloadRefresher to download userstore every {}ms", pollingIntervalMs);
                }
            } catch (IllegalStateException exception) {
                LOG.error("Error scheduling userStoreDownloadTimer:", exception);
                LOG.error("*** UserStore information will NOT be downloaded every {}ms ***", pollingIntervalMs);

                userStoreDownloadTimer = null;
            }
        }

        LOG.info("<== RangerRemoteAuthorizer.initRangerUserStore()");
    }

    private void initProcessors() {
        LOG.info("==> RangerRemoteAuthorizer.initProcessors()");

        Configuration config       = context.getRazConfig();
        String        serviceTypes = config.get("ranger.raz.bootstrap.servicetypes");
        String        redactKeys   = config.get("ranger.raz.result.keys.to.redact");

        if (StringUtils.isNotBlank(redactKeys)) {
            String[] keys = redactKeys.split(",");
            for (String keyToRedact : keys) {
                if (StringUtils.isNotBlank(keyToRedact)) {
                    keysToRedact.addAll(new HashSet<>(Arrays.asList(keys)));
                }
            }
        }

        if (StringUtils.isNotBlank(serviceTypes)) {
            for (String serviceType : serviceTypes.split(",")) {
                LOG.info("RangerRemoteAuthorizer.initProcessors(): initializing for serviceType={}", serviceType);

                try {
                    RangerRazProcessor processor = createRazProcessor(serviceType);

                    if (processor != null) {
                        razProcessors.put(serviceType, processor);
                    }
                } catch (RangerRazException excp) {
                    LOG.error("RangerRemoteAuthorizer.initProcessors(): failed to initialize processor for serviceType={}", serviceType, excp);
                }
            }
        }

        LOG.info("<== RangerRemoteAuthorizer.initProcessors()");
    }

    private RangerRazProcessor getOrCreateRazProcessor(String serviceType) {
        RangerRazProcessor ret = razProcessors.get(serviceType);

        if (ret == null) {
            synchronized (this) {
                ret = razProcessors.get(serviceType);

                if (ret == null) {
                    try {
                        ret = createRazProcessor(serviceType);
                    } catch (Exception excp) {
                        LOG.error("failed to instantiate processor for service-type {}", serviceType, excp);

                        ret = defaultRazProcessor;
                    }

                    razProcessors.put(serviceType, ret);
                }
            }
        }

        return ret;
    }

    private RangerRazProcessor createRazProcessor(String serviceType) throws RangerRazException {
        LOG.info("==> createRazProcessor(serviceType={})", serviceType);

        final RangerRazProcessor ret;

        try {
            String clsName = context.getRazConfig().get("ranger.raz.processor." + serviceType);

            if (StringUtils.isNotBlank(clsName)) {
                String      libDir = context.getRazProcessorLibDir(serviceType);
                File        dir    = new File(libDir);
                ClassLoader clsLoader;

                if (dir.exists() && dir.isDirectory()) {
                    LOG.info("Searching processor {} in directory={}", clsName, libDir);

                    clsLoader = new RangerRazProcessorClassLoader(libDir);
                } else {
                    LOG.info("Directory {} not found. Searching processor {} in webapp libraries", libDir, clsName);

                    clsLoader = Thread.currentThread().getContextClassLoader();
                }

                Class<RangerRazProcessor> cls = (Class<RangerRazProcessor>) clsLoader.loadClass(clsName.trim());

                ret = cls.newInstance();

                ret.init(context);
            } else {
                LOG.info("using default processor for serviceType={}", serviceType);

                ret = defaultRazProcessor;
            }
        } catch (Exception excp) {
            LOG.error("failed to instantiate processor for service-type {}", serviceType, excp);

            throw new RangerRazException(excp);
        }

        LOG.info("<== createRazProcessor(serviceType={}): ret={}", serviceType, ret);

        return ret;
    }

    private RangerBasePlugin getOrCreateServicePlugin(String serviceType, String serviceName)
        throws RangerRazException {
        RangerBasePlugin ret = servicePlugins.get(serviceName);

        if (ret == null) {
            synchronized (servicePlugins) {
                ret = servicePlugins.get(serviceName);

                try {
                    if (ret == null) {
                        LOG.info("Instantiating plugin for serviceType={}, serviceName={}", serviceType, serviceName);

                        RangerRazPluginConfig pluginConfig = new RangerRazPluginConfig(serviceType, serviceName, context.getRazConfig());
                        // Getting class name of ranger-plugin for a serviceType.
                        String className = context.getRazConfig().get("ranger.raz.plugin." + serviceType);
                        if (StringUtils.isNotBlank(className)) {
                            String libDir = context.getRazProcessorLibDir(serviceType);
                            File dirPath = new File(libDir);
                            final ClassLoader basePluginClassLoader;

                            if (dirPath.exists() && dirPath.isDirectory()) {
                                LOG.info("Searching ranger plugin {} in directory={}", className, libDir);

                                basePluginClassLoader = new RangerBasePluginLoader(libDir);
                            } else {
                                LOG.info("Directory {} not found. Searching ranger plugin {} in webapp libraries", libDir, className);

                                basePluginClassLoader = Thread.currentThread().getContextClassLoader();
                            }

                            Class<RangerBasePlugin> cls =
                                (Class<RangerBasePlugin>) basePluginClassLoader.loadClass(className.trim());

                            ret = cls.getConstructor(RangerPluginConfig.class).newInstance(pluginConfig);
                        } else {
                            if(LOG.isDebugEnabled()) {
                                LOG.debug("No class name found with the serviceType {}, using Default RangerBasePlugin", serviceType);
                            }
                            ret = new RangerBasePlugin(pluginConfig);
                        }

                        ret.init();

                        servicePlugins.put(serviceName, ret);
                    }
                } catch (Exception e) {
                    throw new RangerRazException("Failed to instantiate ranger plugin", e);
                }
            }
        }

        return ret;
    }

    private void startTgtRenewer() {
        long tgtRenewIntervalInMins = rangerRazConfig.getLong(RAZ_CONF_TGT_RENEW_INTERVAL, RAZ_TGT_RENEW_INTERVAL_DEFAULT);
        ScheduledExecutorService executorService = Executors.newSingleThreadScheduledExecutor();
        executorService.scheduleAtFixedRate(new KrbTicketRenewer(), tgtRenewIntervalInMins, tgtRenewIntervalInMins, TimeUnit.MINUTES);
    }

    @Override
    public void updateRangerUserStore(RangerUserStore rangerUserStore) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Updating userstore for raz from version = " + this.rangerUserStore.getUserStoreVersion() + " to " + rangerUserStore.getUserStoreVersion());
        }

        RangerRazIdentityMappings ret = new RangerRazIdentityMappings();
        Map<String, String> userClusterIdToCloudIdMap  = generateIdentityMap(rangerUserStore, rangerUserStore::getUserAttrMapping);  //User
        Map<String, String> groupClusterIdToCloudIdMap = generateIdentityMap(rangerUserStore, rangerUserStore::getGroupAttrMapping); //Group

        ret.setUserClusterIdToCloudId(userClusterIdToCloudIdMap);
        ret.setGroupClusterIdToCloudId(groupClusterIdToCloudIdMap);
        ret.setVersion(rangerUserStore.getUserStoreVersion());

        synchronized (RANGER_USER_STORE_LOCK) {
            this.cachedRazIdMappings = ret;
            this.rangerUserStore = rangerUserStore;
        }
    }

    private class KrbTicketRenewer implements Runnable {
        @Override
        public void run() {
            UserGroupInformation userGroupInformation = MiscUtil.getUGILoginUser();
            if (LOG.isDebugEnabled()) {
                LOG.debug("===>> KrbTicketRenewer.run(): TGT renewed for {}", userGroupInformation);
            }
        }
    }

    public RangerRazIdentityMappings getRangerRazUserIdentityMappings() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerRemoteAuthorizer.getRangerRazUserIdentityMappings()");
        }

        RangerRazIdentityMappings ret = null;
        synchronized (RANGER_USER_STORE_LOCK) {
            ret = this.cachedRazIdMappings;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerRemoteAuthorizer.getRangerRazUserIdentityMappings(): ret={}", ret);
        }
        return ret;
    }

    private Map<String, String> generateIdentityMap(RangerUserStore rangerUserStore, Supplier<Map<String, Map<String, String>>> attrMapSupplier) {
        Map<String, String> clusterIdToCloudIdMap    = new HashMap<>();
        Map<String, Map<String, String>> attrMapping = attrMapSupplier.get();

        if (MapUtils.isNotEmpty(attrMapping)) {
            for (Map.Entry<String, Map<String,String>> entry : attrMapping.entrySet()) {
                String clusterIdentity = entry.getKey();
                String cloudIdentity   = rangerUserStoreUtil.getCloudId(attrMapping, clusterIdentity);

                if (StringUtils.isNotBlank(clusterIdentity) && StringUtils.isNotBlank(cloudIdentity)) {
                    clusterIdToCloudIdMap.put(clusterIdentity, cloudIdentity);
                }
            }
        }

        return clusterIdToCloudIdMap;
    }
}
