/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.raz.rest;

import org.apache.commons.logging.Log;
import org.apache.hadoop.util.StopWatch;
import org.apache.ranger.plugin.util.RangerPerfTracer;
import org.apache.ranger.raz.RangerRemoteAuthorizer;
import org.apache.ranger.raz.intg.RangerRazException;
import org.apache.ranger.raz.metrics.RazMetricsWrapper;
import org.apache.ranger.raz.model.RangerRazMultiOperationRequest;
import org.apache.ranger.raz.model.RangerRazMultiOperationResult;
import org.apache.ranger.raz.model.RangerRazRequest;
import org.apache.ranger.raz.model.RangerRazResult;
import org.apache.ranger.raz.model.RangerRazResultBase.AccessResult;
import org.apache.ranger.raz.model.RangerRazResultBase.ResourceAccessResult;
import org.apache.ranger.raz.model.RangerRazIdentityMappings;
import org.apache.ranger.raz.processor.RangerRazRequestContext;
import org.apache.ranger.raz.util.RequestContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.Context;

import java.util.Set;
import java.util.concurrent.TimeUnit;

@Path("authz")
@Component
@Scope("request")
public class AuthzREST {
    private static final Logger LOG      = LoggerFactory.getLogger(AuthzREST.class);
    private static final Log    PERF_LOG = RangerPerfTracer.getPerfLogger("rest.AuthzREST");

    @Autowired
    RangerRemoteAuthorizer authorizer;

    @Autowired
    private RazMetricsWrapper metrics;

    @POST
    @Path("/{serviceType}/access")
    @Produces({ "application/json", "application/xml" })
    public RangerRazResult authorizeAccess(@PathParam("serviceType") String serviceType, RangerRazRequest request, @Context HttpServletRequest httpRequest) {
        if(LOG.isTraceEnabled()) {
            LOG.trace("==> AuthzREST.authorizeAccess({})", request);
        }

        RangerRazResult  ret         = null;
        RangerPerfTracer perf        = null;
        StopWatch        stopWatch   = new StopWatch();

        stopWatch.start();

        try {
            if (RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
                perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "AuthzREST.authorizeAccess(request=" + request + ")");
            }

            RequestContext requestContext        = RequestContext.get();
            String         userName              = requestContext.getUserName();
            Set<String>    userGroups            = requestContext.getUserGroups();
            String         remoteIpAddress       = requestContext.getClientIPAddress();
            String[]       forwardedForAddresses = requestContext.getForwardedForAddresses();

            RangerRazRequestContext context = new RangerRazRequestContext(userName, userGroups, remoteIpAddress, forwardedForAddresses);

            if(LOG.isTraceEnabled()) {
                LOG.trace("==> AuthzREST.authorizeAccess({}, {}, {})", serviceType, request, context);
            }

            ret = authorizer.authorize(serviceType, request, context);
        } catch (RangerRazException excp) {
            LOG.error("AuthzREST.authorizeAccess()", excp);

            rangerRazExceptionToWebApplication(excp);
        } catch (Throwable excp) {
            LOG.error("AuthzREST.authorizeAccess()", excp);

            // Just throw the exception here, since, we intend that the
            // exception should stay in the server and not to be seen by client.
            throw excp;
        } finally {
            RangerPerfTracer.log(perf);
            stopWatch.stop();
            AccessResult accessResult = (ret != null && ret.getOperResult() != null) ? ret.getOperResult().getResult() : null;
            metrics.update(accessResult, TimeUnit.MICROSECONDS.convert(stopWatch.now(), TimeUnit.MILLISECONDS));
            stopWatch.close();
        }

        if(LOG.isTraceEnabled()) {
            LOG.trace("<== AuthzREST.authorizeAccess({}, {}, {})", serviceType, request, ret);
        }

        if(LOG.isTraceEnabled()) {
            LOG.trace("<== AuthzREST.authorizeAccess({}): ret={}", request, ret);
        }

        return ret;
    }

    @POST
    @Path("/{serviceType}/accesses")
    @Produces({ "application/json", "application/xml" })
    public RangerRazMultiOperationResult authorizeAccesses(@PathParam("serviceType") String serviceType, RangerRazMultiOperationRequest requests, @Context HttpServletRequest httpRequest) {
        if(LOG.isTraceEnabled()) {
            LOG.trace("==> AuthzREST.authorizeAccesses({})", requests);
        }

        RangerRazMultiOperationResult ret            = null;
        RangerPerfTracer              perf           = null;
        StopWatch                     stopWatch      = new StopWatch();
        long                          processingTime = 0;

        try {
            if (RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
                perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "AuthzREST.authorizeAccesses(request=" + requests + ")");
            }

            stopWatch.start();
            RequestContext requestContext        = RequestContext.get();
            String         userName              = requestContext.getUserName();
            Set<String>    userGroups            = requestContext.getUserGroups();
            String         remoteIpAddress       = requestContext.getClientIPAddress();
            String[]       forwardedForAddresses = requestContext.getForwardedForAddresses();

            RangerRazRequestContext context = new RangerRazRequestContext(userName, userGroups, remoteIpAddress, forwardedForAddresses);

            if(LOG.isDebugEnabled()) {
                LOG.debug("==> AuthzREST.authorizeAccesses({}, {}, {})", serviceType, requests, context);
            }

            ret = authorizer.authorize(serviceType, requests, context);
        } catch (RangerRazException excp) {
            LOG.error("AuthzREST.authorizeAccesses()", excp);

            rangerRazExceptionToWebApplication(excp);
        }  catch (Throwable excp) {
            LOG.error("AuthzREST.authorizeAccess()", excp);

            // Just throw the exception here, since, we intend that the
            // exception should stay in the server and not to be seen by client.
            throw excp;
        } finally {
            RangerPerfTracer.log(perf);
            stopWatch.stop();
            processingTime = TimeUnit.MICROSECONDS.convert(stopWatch.now(), TimeUnit.MILLISECONDS);
            stopWatch.close();
            if (ret != null) {
                for (ResourceAccessResult resourceAccessResult : ret.getOperResults()) {
                    AccessResult accessResult = resourceAccessResult != null ? resourceAccessResult.getResult() : null;
                    metrics.update(accessResult, processingTime);
                }
            }
        }

        if(LOG.isTraceEnabled()) {
            LOG.trace("<== AuthzREST.authorizeAccesses({}, {}, {})", serviceType, requests, ret);
        }

        if(LOG.isTraceEnabled()) {
            LOG.trace("<== AuthzREST.authorizeAccesses({}): ret={}", requests, ret);
        }

        return ret;
    }

    /**
     * Gets latest azure identity mappings.
     * @param lastKnownVersion
     * @param request
     * @return
     */
    @GET
    @Path("/identity-mappings")
    @Produces({ "application/json", "application/xml" })
    public RangerRazIdentityMappings getIdentiyMappings() {
        if(LOG.isDebugEnabled()) {
            LOG.debug("==> AuthzREST.getIdentiyMappings()");
        }

        RangerRazIdentityMappings ret = authorizer.getRangerRazUserIdentityMappings();

        if(LOG.isDebugEnabled()) {
            LOG.debug("<== AuthzREST.getIdentiyMappings(): ret={}", ret);
        }
        return ret;
    }

    /**
     * Method to convert a RangerRazException to WebApplicationException.
     * @param excp RangerRazException to convert.
     */
    private void rangerRazExceptionToWebApplication(RangerRazException excp) {
        if(!excp.getRangerRazExceptionHeadersMap().isEmpty()) {
            throw excp.toWebApplicationExceptionWithHeaders();
        } else {
            throw excp.toWebApplicationException();
        }
    }
}
