/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.raz.security;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Properties;

import javax.annotation.PostConstruct;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apache.commons.collections.iterators.IteratorEnumeration;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.SecurityUtil;
import org.apache.hadoop.security.authentication.server.KerberosAuthenticationHandler;
import org.apache.hadoop.security.authentication.server.PseudoAuthenticationHandler;
import org.apache.hadoop.security.authorize.ProxyUsers;
import org.apache.hadoop.security.token.delegation.web.DelegationTokenAuthenticationFilter;
import org.apache.hadoop.security.token.delegation.web.DelegationTokenAuthenticationHandler;
import org.apache.hadoop.security.token.delegation.web.KerberosDelegationTokenAuthenticationHandler;
import org.apache.hadoop.security.token.delegation.web.PseudoDelegationTokenAuthenticationHandler;
import org.apache.ranger.raz.server.RangerRazConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.annotation.Lazy;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Component;

/**
 * Responsibility of this class is to authenticate request.
 * API requests can authenticate using Kerberos ONLY, or Delegation Token.
 * Delegation Token can be achieved, after Kerberos authentication.
 * Once Delegation Token is achieved, it can be used independently to access APIs.
 */
@Lazy(true)
@Component
public class RazDelegationTokenFilter extends DelegationTokenAuthenticationFilter {
    private static final Logger LOG = LoggerFactory.getLogger(RazDelegationTokenFilter.class);

    private static final String          CONF_RAZ_SERVER_BIND_ADDRESS_PARAM                    = "ranger.raz.service.host";
    private static final String          CONF_RAZ_FS_AZURE_EXT_RAZ_DELEGATION_TOKEN_KIND_PARAM = "fs.azure.ext.raz.delegation-token.token-kind";
    private static final String          CONF_RAZ_FS_S3_EXT_RAZ_DELEGATION_TOKEN_KIND_PARAM    = "fs.s3a.ext.raz.delegation-token.token-kind";
    private static final String          CONF_RAZ_PROXY_USER_PREFIX                            = "ranger.raz.proxyuser.";
    private static final String          CONF_RAZ_SUPPORTED_SERVICE_TYPES                      = "ranger.raz.bootstrap.servicetypes";
    private static final String          RAZ_DELEGATION_TOKEN_KIND_DEFAULT                     = "raz-dt";
    private static final RangerRazConfig razConfig                                             = new RangerRazConfig();

    private static final String MULTIPLE_SERVICE_NOT_SUPPORTED_STR = "Multiple service types are not supported.";
    private static final String INVALID_SERVICE_TYPE_STR           = "In-valid service-type (ranger.raz.bootstrap.servicetypes) provided.";
    private static final String SUPPORTED_CLOUD_SERVICE_TYPE_AZURE = "adls";
    private static final String SUPPORTED_CLOUD_SERVICE_TYPE_S3    = "s3";

    private final ServletContext nullContext = new NullServletContext();

    private String tokenKindStr = null;

    @PostConstruct
    public void initialize() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("===>>> RazDelegationTokenFilter.initialize()");
        }

        String serviceType = razConfig.get(CONF_RAZ_SUPPORTED_SERVICE_TYPES);
        if (serviceType.contains(",")) {
            LOG.error(MULTIPLE_SERVICE_NOT_SUPPORTED_STR);
            throw new RuntimeException(MULTIPLE_SERVICE_NOT_SUPPORTED_STR);
        }

        switch(serviceType)
        {
            case SUPPORTED_CLOUD_SERVICE_TYPE_AZURE:
                tokenKindStr = razConfig.get(CONF_RAZ_FS_AZURE_EXT_RAZ_DELEGATION_TOKEN_KIND_PARAM, RAZ_DELEGATION_TOKEN_KIND_DEFAULT);
                break;

            case SUPPORTED_CLOUD_SERVICE_TYPE_S3:
                tokenKindStr = razConfig.get(CONF_RAZ_FS_S3_EXT_RAZ_DELEGATION_TOKEN_KIND_PARAM, RAZ_DELEGATION_TOKEN_KIND_DEFAULT);
                break;

            default:
                LOG.error(INVALID_SERVICE_TYPE_STR + " service-type=" + serviceType);
                throw new RuntimeException(INVALID_SERVICE_TYPE_STR);
        }

        try {
            /**
             * If this filter is configured in spring security. The
             * {@link org.springframework.web.filter.DelegatingFilterProxy
             * DelegatingFilterProxy} does not invoke init method (like Servlet container).
             */
            init(null);
        } catch (ServletException e) {
            LOG.error("RazDelegationTokenFilter(): initialization failure", e);
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("<<<=== RazDelegationTokenFilter.initialize()");
        }
    }

	/**
	 * Initialized this filter. Only CONFIG_PREFIX needs to be set.
	 * Rest other configurations will be fetched from {@link java.util.Properties config}
	 * return by {@link #getConfiguration(String, FilterConfig)}.
	 */
	@Override
	public void init(final FilterConfig filterConfig) throws ServletException {
		final Map<String, String> params = new HashMap<>();
		params.put(CONFIG_PREFIX, razConfig.get(CONFIG_PREFIX));
		if (LOG.isDebugEnabled()) {
			LOG.debug("Config Prefix used for RazDelegationTokenFilter is [{}].", params.get(CONFIG_PREFIX));
		}

		FilterConfig filterConfigNew = new FilterConfig() {
			@Override
			public ServletContext getServletContext() {
				if (Objects.nonNull(filterConfig)) {
					return filterConfig.getServletContext();
				} else {
					return nullContext;
				}
			}

			@SuppressWarnings("unchecked")
			@Override
			public Enumeration<String> getInitParameterNames() {
				return new IteratorEnumeration(params.keySet().iterator());
			}

			@Override
			public String getInitParameter(String param) {
				return params.get(param);
			}

			@Override
			public String getFilterName() {
				return "RazDelegationTokenFilter";
			}
		};
		super.init(filterConfigNew);
		Configuration conf1 = this.getProxyuserConfiguration(filterConfigNew);
		ProxyUsers.refreshSuperUserGroupsConfiguration(conf1, CONF_RAZ_PROXY_USER_PREFIX);
	}

	/**
	 * Return {@link java.util.Properties configuration} required for this filter to work properly.
	 */
	@Override
	protected Properties getConfiguration(String configPrefix, FilterConfig filterConfig) throws ServletException {
		Properties props = new Properties();
		props.putAll(razConfig.getPropsWithPrefix(configPrefix));

		String authType = props.getProperty(AUTH_TYPE, "simple");
		if (authType.equals(PseudoAuthenticationHandler.TYPE)) {
			props.setProperty(AUTH_TYPE, PseudoDelegationTokenAuthenticationHandler.class.getName());
		} else if (authType.equals(KerberosAuthenticationHandler.TYPE)) {
			props.setProperty(AUTH_TYPE, KerberosDelegationTokenAuthenticationHandler.class.getName());
		}

		props.setProperty(DelegationTokenAuthenticationHandler.TOKEN_KIND, tokenKindStr);

		// Resolve _HOST into bind address
		String bindAddress = razConfig.get(CONF_RAZ_SERVER_BIND_ADDRESS_PARAM);

		if (Objects.isNull(bindAddress)) {
			LOG.warn("No host name configured.  Defaulting to local host name.");

			try {
				bindAddress = InetAddress.getLocalHost().getHostName();
			} catch (UnknownHostException e) {
				throw new ServletException("Unable to obtain host name", e);
			}
		}

		// Resolve principle using latest bindAddress
		String principal = props.getProperty(KerberosAuthenticationHandler.PRINCIPAL);

		if (Objects.nonNull(principal)) {
			try {
				principal = SecurityUtil.getServerPrincipal(principal, bindAddress);
			} catch (IOException ex) {
				throw new RuntimeException("Could not resolve Kerberos principal name: " + ex.toString(), ex);
			}
			props.put(KerberosAuthenticationHandler.PRINCIPAL, principal);
		}

		return props;
	}

	@Override
	protected Configuration getProxyuserConfiguration(FilterConfig filterConfig) {
		Configuration conf = new Configuration(false);
		Map<String, String> proxyuserConf = razConfig.getPropsWithPrefix(CONF_RAZ_PROXY_USER_PREFIX);
		for (Map.Entry<String, String> entry : proxyuserConf.entrySet()) {
			conf.set(CONF_RAZ_PROXY_USER_PREFIX+entry.getKey(), entry.getValue());
			if (LOG.isDebugEnabled()) {
				LOG.debug("Proxyuser config = " + CONF_RAZ_PROXY_USER_PREFIX+entry.getKey() + " = " + entry.getValue());
			}
		}
		return conf;
	}

	@Override
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain filterChain) throws IOException, ServletException {
		Authentication previousAuth = SecurityContextHolder.getContext().getAuthentication();
		if (previousAuth == null || !previousAuth.isAuthenticated()) {
			FilterChainWrapper filterChainWrapper = new FilterChainWrapper(filterChain);

			if (LOG.isDebugEnabled()) {
				LOG.debug("Trying to authenticate user via RazDelegationTokenFilter.");
			}
			super.doFilter(request, response, filterChainWrapper);
		} else {
			// Already authenticated
			if (LOG.isDebugEnabled()) {
				LOG.debug("User [{}] is already authenticated, proceeding with filter chain.", previousAuth.getPrincipal());
			}

			filterChain.doFilter(request, response);
		}

		Authentication authentication = SecurityContextHolder.getContext().getAuthentication();

		if (LOG.isDebugEnabled()) {
			LOG.debug("Leaving RazDelegationTokenFilter, isUserAuthenticated? [{}]", (authentication != null && authentication.isAuthenticated()));
		}
	}
}
