/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.raz.security;

import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.Properties;

import javax.annotation.PostConstruct;
import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;

import org.apache.ranger.authz.handler.RangerAuth;
import org.apache.ranger.authz.handler.jwt.RangerDefaultJwtAuthHandler;
import org.apache.ranger.raz.server.RangerRazConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.annotation.Lazy;
import org.springframework.security.authentication.AbstractAuthenticationToken;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.User;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.web.authentication.WebAuthenticationDetails;
import org.springframework.stereotype.Component;

@Lazy(true)
@Component
public class RazJwtAuthFilter extends RangerDefaultJwtAuthHandler implements Filter {
    private static final Logger          LOG              = LoggerFactory.getLogger(RazJwtAuthFilter.class);
    private static final RangerRazConfig razConfig        = new RangerRazConfig();
    private static final String          CONFIG_PREFIX    = "config.prefix";
    private static final String          DEFAULT_RAZ_ROLE = "ROLE_USER";

    @PostConstruct
    public void initialize() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("===>>> RazJwtAuthFilter.initialize()");
        }

        /**
         * If this filter is configured in spring security. The
         * {@link org.springframework.web.filter.DelegatingFilterProxy
         * DelegatingFilterProxy} does not invoke init method (like Servlet container).
         */
        try {
            String     configPrefix = razConfig.get(CONFIG_PREFIX);
            Properties config       = new Properties();

            Iterator<Entry<String, String>> iterator = razConfig.iterator();
            while (iterator.hasNext()) {
                Entry<String, String> entry = iterator.next();
                if (entry.getKey().startsWith(configPrefix)) {
                    config.put(entry.getKey().substring(configPrefix.length() + 1), entry.getValue());
                }
            }

            super.initialize(config);
        } catch (Exception e) {
            LOG.error("Failed to initialize Ranger RAZ JWT Auth Filter.", e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<<<=== RazJwtAuthFilter.initialize()");
        }
    }

    @Override
    public void init(FilterConfig filterConfig) throws ServletException {
        // Auto-generated method stub

    }

    @Override
    public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("===>>> RazJwtAuthFilter.doFilter()");
        }

        Authentication previousAuth = SecurityContextHolder.getContext().getAuthentication();

        // check if not already authenticated
        if (previousAuth == null || !previousAuth.isAuthenticated()) {
            HttpServletRequest httpServletRequest = (HttpServletRequest) request;

            RangerAuth rangerAuth = authenticate(httpServletRequest);

            if (rangerAuth != null) {
                final List<GrantedAuthority>   grantedAuths        = Arrays.asList(new SimpleGrantedAuthority(DEFAULT_RAZ_ROLE));
                final UserDetails              principal           = new User(rangerAuth.getUserName(), "", grantedAuths);
                final Authentication           finalAuthentication = new UsernamePasswordAuthenticationToken(principal, "", grantedAuths);
                final WebAuthenticationDetails webDetails          = new WebAuthenticationDetails(httpServletRequest);
                ((AbstractAuthenticationToken) finalAuthentication).setDetails(webDetails);
                SecurityContextHolder.getContext().setAuthentication(finalAuthentication);
            }
        } else {
            // Already authenticated
            if (LOG.isDebugEnabled()) {
                LOG.debug("User [{}] is already authenticated, proceeding with filter chain.", previousAuth.getPrincipal());
            }
        }

        //Log final status of request.
        Authentication finalAuth = SecurityContextHolder.getContext().getAuthentication();
        if (finalAuth != null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("<<<=== RazJwtAuthFilter.doFilter() - user=[{}], isUserAuthenticated? [{}]", finalAuth.getPrincipal(), finalAuth.isAuthenticated());
            }
        } else {
            LOG.warn("<<<=== RazJwtAuthFilter.doFilter() - Failed to authenticate request using RAZ JWT authentication framework.");
        }
    }

    @Override
    public void destroy() {
        // Auto-generated method stub

    }
}
