/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.rms.security;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Properties;

import javax.annotation.PostConstruct;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apache.commons.collections.iterators.IteratorEnumeration;
import org.apache.hadoop.security.SecurityUtil;
import org.apache.hadoop.security.authentication.server.KerberosAuthenticationHandler;
import org.apache.hadoop.security.authentication.server.PseudoAuthenticationHandler;
import org.apache.hadoop.security.token.delegation.web.DelegationTokenAuthenticationFilter;
import org.apache.hadoop.security.token.delegation.web.DelegationTokenAuthenticationHandler;
import org.apache.hadoop.security.token.delegation.web.KerberosDelegationTokenAuthenticationHandler;
import org.apache.hadoop.security.token.delegation.web.PseudoDelegationTokenAuthenticationHandler;
import org.apache.ranger.rms.server.RMSConfig;
import org.apache.ranger.rms.util.AppConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;

public class RMSDelegationTokenFilter extends DelegationTokenAuthenticationFilter {

	private static final Logger LOG = LoggerFactory.getLogger(RMSDelegationTokenFilter.class);

	private final ServletContext nullContext = new NullServletContext();

	protected RMSConfig rmsConfig;

	private String tokenKindStr = null;

	/**
	 * Initialized this filter. Only CONFIG_PREFIX needs to be set.
	 * Rest other configurations will be fetched from {@link java.util.Properties config}
	 * return by {@link #getConfiguration(String, FilterConfig)}.
	 */
	@PostConstruct
	public void initialize() {
		if (LOG.isDebugEnabled()) {
            LOG.debug("===>>> RMSDelegationTokenFilter.initialize()");
        }

		this.rmsConfig = RMSConfig.getInstance();

		try {
			// Currently setting tokenKindStr to default value. Later we can configure as per requirement.
			// tokenKindStr = rmsConfig.get(AppConstants.PROP_RMS_FS_S3_EXT_RMS_DELEGATION_TOKEN_KIND_PARAM, AppConstants.RMS_DELEGATION_TOKEN_KIND_DEFAULT);
			tokenKindStr = AppConstants.RMS_DELEGATION_TOKEN_KIND_DEFAULT;
			final Map<String, String> params = new HashMap<>();
			params.put(CONFIG_PREFIX, rmsConfig.get(CONFIG_PREFIX));
			if (LOG.isDebugEnabled()) {
				LOG.debug("Config Prefix used for RMSDelegationTokenFilter is [{}].", params.get(CONFIG_PREFIX));
			}

			FilterConfig filterConfigNew = new FilterConfig() {
				@Override
				public ServletContext getServletContext() {
					return nullContext;
				}

				@SuppressWarnings("unchecked")
				@Override
				public Enumeration<String> getInitParameterNames() {
					return new IteratorEnumeration(params.keySet().iterator());
				}

				@Override
				public String getInitParameter(String param) {
					return params.get(param);
				}

				@Override
				public String getFilterName() {
					return "RMSDelegationTokenFilter";
				}
			};
			super.init(filterConfigNew);

		} catch (ServletException e) {
			LOG.error("RMSDelegationTokenFilter(): initialization failure", e);
		}

		if (LOG.isDebugEnabled()) {
            LOG.debug("<<<=== RMSDelegationTokenFilter.initialize()");
        }
	}

	/**
	 * Return {@link java.util.Properties configuration} required for this filter to work properly.
	 */
	@Override
	protected Properties getConfiguration(String configPrefix, FilterConfig filterConfig) throws ServletException {
		Properties props = new Properties();
		props.putAll(rmsConfig.getPropsWithPrefix(configPrefix));

		String authType = props.getProperty(AUTH_TYPE, "simple");
		if (authType.equals(PseudoAuthenticationHandler.TYPE)) {
			props.setProperty(AUTH_TYPE, PseudoDelegationTokenAuthenticationHandler.class.getName());
		} else if (authType.equals(KerberosAuthenticationHandler.TYPE)) {
			props.setProperty(AUTH_TYPE, KerberosDelegationTokenAuthenticationHandler.class.getName());
		}

		props.setProperty(DelegationTokenAuthenticationHandler.TOKEN_KIND, tokenKindStr);

		// Resolve _HOST into bind address
		String bindAddress = rmsConfig.get(AppConstants.PROP_RMS_SERVER_BIND_ADDRESS);

		if (Objects.isNull(bindAddress)) {
			LOG.warn("No host name configured.  Defaulting to local host name.");

			try {
				bindAddress = InetAddress.getLocalHost().getHostName();
			} catch (UnknownHostException e) {
				throw new ServletException("Unable to obtain host name", e);
			}
		}

		// Resolve principle using latest bindAddress
		String principal = props.getProperty(KerberosAuthenticationHandler.PRINCIPAL);

		if (Objects.nonNull(principal)) {
			try {
				principal = SecurityUtil.getServerPrincipal(principal, bindAddress);
			} catch (IOException ex) {
				throw new RuntimeException("Could not resolve Kerberos principal name: " + ex.toString(), ex);
			}
			props.put(KerberosAuthenticationHandler.PRINCIPAL, principal);
		}

		return props;
	}

	@Override
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain filterChain)
			throws IOException, ServletException {
		Authentication previousAuth = SecurityContextHolder.getContext().getAuthentication();
		if (previousAuth == null || !previousAuth.isAuthenticated()) {
			FilterChainWrapper filterChainWrapper = new FilterChainWrapper(filterChain);

			if (LOG.isDebugEnabled()) {
				LOG.debug("Trying to authenticate user via RMSDelegationTokenFilter.");
			}
			super.doFilter(request, response, filterChainWrapper);
		} else {
			// Already authenticated
			if (LOG.isDebugEnabled()) {
				LOG.debug("User [{}] is already authenticated, proceeding with filter chain.", previousAuth.getPrincipal());
			}

			filterChain.doFilter(request, response);
		}

		Authentication authentication = SecurityContextHolder.getContext().getAuthentication();

		if (LOG.isDebugEnabled()) {
			LOG.debug("Leaving RMSDelegationTokenFilter, isUserAuthenticated? [{}]", (authentication != null && authentication.isAuthenticated()));
		}
	}
}
