/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.rms.security;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;

import javax.annotation.PostConstruct;
import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;

import org.apache.ranger.authz.handler.RangerAuth;
import org.apache.ranger.authz.handler.jwt.RangerDefaultJwtAuthHandler;
import org.apache.ranger.rms.server.RMSConfig;
import org.apache.ranger.rms.util.AppConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.authentication.AbstractAuthenticationToken;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.User;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.web.authentication.WebAuthenticationDetails;

public class RMSJwtAuthFilter extends RangerDefaultJwtAuthHandler implements Filter {
    private static final Logger          LOG              = LoggerFactory.getLogger(RMSJwtAuthFilter.class);
    private static final String          DEFAULT_RMS_ROLE = "ROLE_USER";
    private static final String CONFIG_PREFIX = "config.prefix";

    @PostConstruct
    public void initialize() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("===>>> RMSJwtAuthFilter.initialize()");
        }
        RMSConfig rmsConfig = RMSConfig.getInstance();

        /**
         * If this filter is configured in spring security. The
         * {@link org.springframework.web.filter.DelegatingFilterProxy
         * DelegatingFilterProxy} does not invoke init method (like Servlet container).
         */
        try {
			String configPrefix = rmsConfig.get(CONFIG_PREFIX, AppConstants.CONFIG_PREFIX + "jwt.auth").concat(".");
            Properties config       = new Properties();
            config.putAll(rmsConfig.getPropsWithPrefix(configPrefix));
            if (LOG.isDebugEnabled()) {
				LOG.debug("JWT Auth configs : " + config.toString());
            }

            super.initialize(config);
        } catch (Exception e) {
            LOG.error("Failed to initialize Ranger RMS JWT Auth Filter.", e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<<<=== RMSJwtAuthFilter.initialize()");
        }
    }

    @Override
    public void init(FilterConfig filterConfig) throws ServletException {
        // Auto-generated method stub

    }

    @Override
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain)
			throws IOException, ServletException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("===>>> RMSJwtAuthFilter.doFilter({}, {}, {})", request, response, chain);
		}

		if (request != null) {
			if (canAuthenticateRequest(request)) {
				Authentication previousAuth = SecurityContextHolder.getContext().getAuthentication();

				// check if not already authenticated
				if (previousAuth == null || !previousAuth.isAuthenticated()) {
					HttpServletRequest httpServletRequest = (HttpServletRequest) request;

					RangerAuth rangerAuth = authenticate(httpServletRequest);

					if (rangerAuth != null) {
						final List<GrantedAuthority> grantedAuths = Arrays.asList(new SimpleGrantedAuthority(DEFAULT_RMS_ROLE));
						final UserDetails principal = new User(rangerAuth.getUserName(), "", grantedAuths);
						final Authentication finalAuthentication = new UsernamePasswordAuthenticationToken(principal,"", grantedAuths);
						final WebAuthenticationDetails webDetails = new WebAuthenticationDetails(httpServletRequest);
						((AbstractAuthenticationToken) finalAuthentication).setDetails(webDetails);
						SecurityContextHolder.getContext().setAuthentication(finalAuthentication);
					}
				} else {
					// Already authenticated
					if (LOG.isDebugEnabled()) {
						LOG.debug("User [{}] is already authenticated, proceeding with filter chain.", previousAuth.getPrincipal());
					}
				}

				// Log final status of request.
				Authentication finalAuth = SecurityContextHolder.getContext().getAuthentication();
				if (finalAuth != null) {
					if (LOG.isDebugEnabled()) {
						LOG.debug("RMSJwtAuthFilter.doFilter() - user=[{}], isUserAuthenticated? [{}]", finalAuth.getPrincipal(), finalAuth.isAuthenticated());
					}
				} else {
					LOG.warn("RMSJwtAuthFilter.doFilter() - Failed to authenticate request using RMS JWT authentication framework.");
				}
			} else {
				if (LOG.isDebugEnabled()) {
					LOG.debug("Skipping JWT RMS auth for request.");
				}
			}
		}

		chain.doFilter(request, response); // proceed with filter chain

		if (LOG.isDebugEnabled()) {
			LOG.debug("<<<=== RMSJwtAuthFilter.doFilter()");
		}

	}

    @Override
    public void destroy() {
        // Auto-generated method stub

    }
}
