/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.ranger.rms.util;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.http.client.utils.URIBuilder;
import org.apache.ranger.rms.server.RMSConfig;
import org.apache.ranger.rms.util.AppConstants.FILE_SCHEME_TYPES;

public class PathUtils {
  private static final Configuration CONF = new Configuration();
  private static final String DEFAULT_SCHEME = FileSystem.getDefaultUri(CONF).getScheme();
  private static final String SUPPORTED_SCHEME = RMSConfig.getInstance().get(AppConstants.PROP_RMS_SUPPORTED_URI_SCHEME, "hdfs");

  public static Map<String, String> parsePath(String path) throws Exception {
	    Map<String, String> ret = new HashMap<>();
	    if (StringUtils.isEmpty(path)) {
	      return null;
	    }

	    URI uri;
	    try {
	      // Below converts unescaped path to escaped string and the constructs a URI from it.
	      uri = new URI(StringUtils.stripStart(new URIBuilder().setPath(path).toString(), "/"));
	    } catch (URISyntaxException e) {
	      throw new Exception("Incomprehensible path [" + path + "]", e);
	    }

	    String scheme = uri.getScheme();
	    if (scheme == null) {
	      scheme = DEFAULT_SCHEME;
	      if(scheme == null) {
	        throw new Exception(
	                "Scheme is missing and could not be constructed from configuration");
	      }
	    }

	    // Non-HDFS paths are skipped.
	    if(!scheme.equalsIgnoreCase(SUPPORTED_SCHEME)) {
	      return null;
	    }

	    String uriPath = uri.getPath();
	    if(uriPath == null) {
	      throw new Exception("Path is empty. uri=" + uri);
	    }
	    if (!uriPath.startsWith("/")) {
	      throw new Exception("Path part of uri does not seem right, was expecting a non empty path" +
	              ": path = " + uriPath + ", uri=" + uri);
	    }
	    // Reduce multiple consecutive forward slashes to one.
	    // It's probably a rare case, so use indexOf() before expensive regex.
	    if (uriPath.contains("//")) {
	      uriPath = uriPath.replaceAll("//*", "/");
	    }
	    // Remove leading and trailing slashes
	    String str = StringUtils.stripEnd(uriPath, "/");
	    ret.put(AppConstants.RESOURCE_PATH, str);

		if (FILE_SCHEME_TYPES.s3a.name().equals(SUPPORTED_SCHEME)) {
			ret.put(AppConstants.RESOURCE_BUCKET, uri.getHost());
		}
	    return ret;
	  }
}
