/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.rms.util;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.KeyStore;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.apache.ranger.credentialapi.CredentialReader;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.beans.factory.config.PropertyPlaceholderConfigurer;

public class PropertiesUtil extends PropertyPlaceholderConfigurer {
    private static final Map<String, String> propertiesMap = new HashMap<>();
    private static final Logger logger = Logger.getLogger(PropertiesUtil.class);

    private PropertiesUtil() {
    }

    @Override
    protected void processProperties(ConfigurableListableBeanFactory beanFactory, Properties props) throws BeansException {

    // First let's add the system properties
	Set<Object> keySet = System.getProperties().keySet();
	for (Object key : keySet) {
	    String keyStr = key.toString();
	    propertiesMap.put(keyStr, System.getProperties().getProperty(keyStr).trim());
	}

	// Let's add our properties now
	keySet = props.keySet();
	for (Object key : keySet) {
	    String keyStr = key.toString();
	    propertiesMap.put(keyStr, props.getProperty(keyStr).trim());
	}

	String storeType = propertiesMap.get(AppConstants.PROP_KEYSTORE_TYPE);
	// update system trust store path with custom trust store.
	if (propertiesMap.containsKey(AppConstants.PROP_TRUSTSTORE)) {
		if(!StringUtils.isEmpty(propertiesMap.get(AppConstants.PROP_TRUSTSTORE))){
			System.setProperty("javax.net.ssl.trustStore", propertiesMap.get(AppConstants.PROP_TRUSTSTORE));
			System.setProperty("javax.net.ssl.trustStoreType", KeyStore.getDefaultType());
			Path trustStoreFile = Paths.get(propertiesMap.get(AppConstants.PROP_TRUSTSTORE));
			if (!Files.exists(trustStoreFile) || !Files.isReadable(trustStoreFile)) {
				logger.debug("Could not find or read truststore file '"+propertiesMap.get(AppConstants.PROP_TRUSTSTORE)+"'");
			}else{
				if(propertiesMap.containsKey(AppConstants.PROP_CREDENTIAL_PROVIDER_PATH)){
					String path=propertiesMap.get(AppConstants.PROP_CREDENTIAL_PROVIDER_PATH);
					String trustStoreAlias=getProperty(AppConstants.PROP_TRUSTSTORE_ALIAS,"trustStoreAlias");
					if(path!=null && trustStoreAlias!=null){
						String trustStorePassword=CredentialReader.getDecryptedString(path.trim(), trustStoreAlias.trim(), storeType);
						if(trustStorePassword!=null&& !trustStorePassword.trim().isEmpty() && !trustStorePassword.trim().equalsIgnoreCase("none")){
							propertiesMap.put(AppConstants.PROP_TRUSTSTORE_PASSWORD, trustStorePassword);
							props.put(AppConstants.PROP_TRUSTSTORE_PASSWORD, trustStorePassword);
						}else{
							logger.info("trustStorePassword password not applied; clear text password shall be applicable");
						}
					}
				}
			}
		}
		System.setProperty("javax.net.ssl.trustStorePassword", propertiesMap.get(AppConstants.PROP_TRUSTSTORE_PASSWORD));
	}

	// update system key store path with custom key store.
	if (propertiesMap.containsKey(AppConstants.PROP_KEYTSTORE)) {
		if(!StringUtils.isEmpty(propertiesMap.get(AppConstants.PROP_KEYTSTORE))){
			System.setProperty("javax.net.ssl.keyStore", propertiesMap.get(AppConstants.PROP_KEYTSTORE));
			System.setProperty("javax.net.ssl.keyStoreType", KeyStore.getDefaultType());
			Path keyStoreFile = Paths.get(propertiesMap.get(AppConstants.PROP_KEYTSTORE));
			if (!Files.exists(keyStoreFile) || !Files.isReadable(keyStoreFile)) {
				logger.debug("Could not find or read keystore file '"+propertiesMap.get(AppConstants.PROP_KEYTSTORE)+"'");
			}else{
				if(propertiesMap.containsKey(AppConstants.PROP_CREDENTIAL_PROVIDER_PATH)){
					String path=propertiesMap.get(AppConstants.PROP_CREDENTIAL_PROVIDER_PATH);
					String keyStoreAlias=getProperty(AppConstants.PROP_KEYSTORE_ALIAS,"keyStoreAlias");
					if(path!=null && keyStoreAlias!=null){
						String keyStorePassword=CredentialReader.getDecryptedString(path.trim(), keyStoreAlias.trim(), storeType);
						if(keyStorePassword!=null&& !keyStorePassword.trim().isEmpty() && !keyStorePassword.trim().equalsIgnoreCase("none")){
							propertiesMap.put(AppConstants.PROP_KEYSTORE_PASSWORD, keyStorePassword);
							props.put(AppConstants.PROP_KEYSTORE_PASSWORD, keyStorePassword);
						}else{
							logger.info("keyStorePassword password not applied; clear text password shall be applicable");
						}
					}
				}
			}
			if (StringUtils.isNotBlank(propertiesMap.get(AppConstants.PROP_KEYSTORE_PASSWORD))) {
				System.setProperty("javax.net.ssl.keyStorePassword", propertiesMap.get(AppConstants.PROP_KEYSTORE_PASSWORD));
			} else {
				logger.warn("keyStorePassword password is not available; therefore javax.net.ssl.keyStorePassword system property can't be set");
			}
		}
	}

	//update credential from keystore
	if(propertiesMap.containsKey(AppConstants.PROP_JDBC_CREDENTIAL_PROVIDER_PATH) && propertiesMap.containsKey(AppConstants.PROP_JDBC_CREDENTIAL_ALIAS)){
		String path=propertiesMap.get(AppConstants.PROP_JDBC_CREDENTIAL_PROVIDER_PATH);
		String alias=propertiesMap.get(AppConstants.PROP_JDBC_CREDENTIAL_ALIAS);
		if(path!=null && alias!=null){
			String xaDBPassword=CredentialReader.getDecryptedString(path.trim(),alias.trim(), storeType);
			if(xaDBPassword!=null&& !xaDBPassword.trim().isEmpty() &&
					!"none".equalsIgnoreCase(xaDBPassword.trim())){
				propertiesMap.put(AppConstants.PROP_JDBC_PASSWORD, xaDBPassword);
				props.put(AppConstants.PROP_JDBC_PASSWORD, xaDBPassword);
			}else{
				logger.info("Credential keystore password not applied for ranger-rms DB; clear text password shall be applicable");
			}
		}
	}

	if(propertiesMap.containsKey("ranger.authentication.method")){
		String authenticationMethod=propertiesMap.get("ranger.authentication.method");
		if(("ACTIVE_DIRECTORY".equalsIgnoreCase(authenticationMethod) || "AD".equalsIgnoreCase(authenticationMethod))){
			if(propertiesMap.containsKey(AppConstants.PROP_CREDENTIAL_PROVIDER_PATH) && propertiesMap.containsKey("ranger.ldap.ad.binddn.credential.alias")){
				String path=propertiesMap.get(AppConstants.PROP_CREDENTIAL_PROVIDER_PATH);
				String alias=propertiesMap.get("ranger.ldap.ad.binddn.credential.alias");
				if(path!=null && alias!=null){
					String bindDNPassword=CredentialReader.getDecryptedString(path.trim(), alias.trim(), storeType);
					if(bindDNPassword!=null&& !bindDNPassword.trim().isEmpty() &&
							!"none".equalsIgnoreCase(bindDNPassword.trim())){
						propertiesMap.put("ranger.ldap.ad.bind.password", bindDNPassword);
						props.put("ranger.ldap.ad.bind.password", bindDNPassword);
					}else{
						logger.info("Credential keystore password not applied for AD Bind DN; clear text password shall be applicable");
					}
				}
			}
		}
	}
	if(propertiesMap.containsKey("ranger.authentication.method")){
		String authenticationMethod=propertiesMap.get("ranger.authentication.method");
		if("LDAP".equalsIgnoreCase(authenticationMethod)) {
			if(propertiesMap.containsKey(AppConstants.PROP_CREDENTIAL_PROVIDER_PATH) && propertiesMap.containsKey("ranger.ldap.binddn.credential.alias")){
				String path=propertiesMap.get(AppConstants.PROP_CREDENTIAL_PROVIDER_PATH);
				String alias=propertiesMap.get("ranger.ldap.binddn.credential.alias");
				if(path!=null && alias!=null){
					String bindDNPassword=CredentialReader.getDecryptedString(path.trim(), alias.trim(), storeType);
					if(bindDNPassword!=null&& !bindDNPassword.trim().isEmpty() &&
							!"none".equalsIgnoreCase(bindDNPassword.trim())){
						propertiesMap.put("ranger.ldap.bind.password", bindDNPassword);
						props.put("ranger.ldap.bind.password", bindDNPassword);
					}else{
						logger.info("Credential keystore password not applied for LDAP Bind DN; clear text password shall be applicable");
					}
				}
			}
		}
	}

	if(DBUtil.getDBFlavor()==AppConstants.DB_FLAVOR_MYSQL || DBUtil.getDBFlavor()==AppConstants.DB_FLAVOR_POSTGRES){
		if(propertiesMap.containsKey(AppConstants.PROP_DB_SSL_ENABLED)){
			String db_ssl_enabled=propertiesMap.get(AppConstants.PROP_DB_SSL_ENABLED);
			if(StringUtils.isEmpty(db_ssl_enabled)|| !"true".equalsIgnoreCase(db_ssl_enabled)){
				db_ssl_enabled="false";
			}
			db_ssl_enabled=db_ssl_enabled.toLowerCase();
			if("true".equalsIgnoreCase(db_ssl_enabled)){
				String db_ssl_required=propertiesMap.get("ranger-rms.db.ssl.required");
				if(StringUtils.isEmpty(db_ssl_required)|| !"true".equalsIgnoreCase(db_ssl_required)){
					db_ssl_required="false";
				}
				db_ssl_required=db_ssl_required.toLowerCase();
				String db_ssl_verifyServerCertificate=propertiesMap.get("ranger-rms.db.ssl.verifyServerCertificate");
				if(StringUtils.isEmpty(db_ssl_verifyServerCertificate)|| !"true".equalsIgnoreCase(db_ssl_verifyServerCertificate)){
					db_ssl_verifyServerCertificate="false";
				}
				db_ssl_verifyServerCertificate=db_ssl_verifyServerCertificate.toLowerCase();
				String db_ssl_auth_type=propertiesMap.get("ranger-rms.db.ssl.auth.type");
				if(StringUtils.isEmpty(db_ssl_auth_type)|| !"1-way".equalsIgnoreCase(db_ssl_auth_type)){
					db_ssl_auth_type="2-way";
				}
				propertiesMap.put("ranger-rms.db.ssl.enabled", db_ssl_enabled);
				props.put("ranger-rms.db.ssl.enabled", db_ssl_enabled);
				propertiesMap.put("ranger-rms.db.ssl.required", db_ssl_required);
				props.put("ranger-rms.db.ssl.required", db_ssl_required);
				propertiesMap.put("ranger-rms.db.ssl.verifyServerCertificate", db_ssl_verifyServerCertificate);
				props.put("ranger-rms.db.ssl.verifyServerCertificate", db_ssl_verifyServerCertificate);
				propertiesMap.put("ranger-rms.db.ssl.auth.type", db_ssl_auth_type);
				props.put("ranger-rms.db.ssl.auth.type", db_ssl_auth_type);
				String ranger_jpa_jdbc_url=propertiesMap.get("ranger-rms.jpa.jdbc.url");
				if (StringUtils.isNotEmpty(ranger_jpa_jdbc_url) && !ranger_jpa_jdbc_url.contains("?")) {
					StringBuilder ranger_jpa_jdbc_url_ssl=new StringBuilder(ranger_jpa_jdbc_url);
					if (DBUtil.getDBFlavor()==AppConstants.DB_FLAVOR_MYSQL) {
						ranger_jpa_jdbc_url_ssl.append("?useSSL=").append(db_ssl_enabled).append("&requireSSL=").append(db_ssl_required).append("&verifyServerCertificate=").append(db_ssl_verifyServerCertificate);
					} else if (DBUtil.getDBFlavor()==AppConstants.DB_FLAVOR_POSTGRES) {
						String db_ssl_certificate_file = propertiesMap.get("ranger-rms.db.ssl.certificateFile");
						if (StringUtils.isNotEmpty(db_ssl_certificate_file)) {
							ranger_jpa_jdbc_url_ssl.append("?ssl=" + db_ssl_enabled + "&sslmode=verify-full" + "&sslrootcert=" + db_ssl_certificate_file);
						} else if ("true".equalsIgnoreCase(db_ssl_verifyServerCertificate) || "true".equalsIgnoreCase(db_ssl_required)) {
							ranger_jpa_jdbc_url_ssl.append("?ssl=" + db_ssl_enabled + "&sslmode=verify-full" + "&sslfactory=org.postgresql.ssl.DefaultJavaSSLFactory");
						} else {
							ranger_jpa_jdbc_url_ssl.append("?ssl=" + db_ssl_enabled);
						}
					}
					propertiesMap.put("ranger-rms.jpa.jdbc.url", ranger_jpa_jdbc_url_ssl.toString());
				}
				ranger_jpa_jdbc_url = propertiesMap.get("ranger-rms.jpa.jdbc.url");
				if (StringUtils.isNotEmpty(ranger_jpa_jdbc_url)) {
					props.put("ranger-rms.jpa.jdbc.url", ranger_jpa_jdbc_url);
				}
				logger.info("ranger-rms.jpa.jdbc.url=" + ranger_jpa_jdbc_url);
			} else {
				String ranger_jpa_jdbc_url = propertiesMap.get("ranger-rms.jpa.jdbc.url");
				String ranger_jpa_jdbc_url_extra_args="";
				if (!StringUtils.isEmpty(ranger_jpa_jdbc_url)) {
					if (ranger_jpa_jdbc_url.contains("?")) {
                        ranger_jpa_jdbc_url_extra_args = ranger_jpa_jdbc_url.substring(ranger_jpa_jdbc_url.indexOf("?")+1);
						ranger_jpa_jdbc_url = ranger_jpa_jdbc_url.substring(0, ranger_jpa_jdbc_url.indexOf("?"));
					}
					if (DBUtil.getDBFlavor() == AppConstants.DB_FLAVOR_MYSQL) {
						StringBuilder ranger_jpa_jdbc_url_no_ssl = new StringBuilder(ranger_jpa_jdbc_url);
						if ( !ranger_jpa_jdbc_url_extra_args.contains("useSSL")) {
							ranger_jpa_jdbc_url_no_ssl.append("?useSSL=false");
						}
						if(!StringUtils.isEmpty(ranger_jpa_jdbc_url_extra_args) && ranger_jpa_jdbc_url_no_ssl.toString().contains("useSSL")) {
							ranger_jpa_jdbc_url_no_ssl.append("&").append(ranger_jpa_jdbc_url_extra_args);
						} else if (!StringUtils.isEmpty(ranger_jpa_jdbc_url_extra_args) && !ranger_jpa_jdbc_url_no_ssl.toString().contains("useSSL")) {
							ranger_jpa_jdbc_url_no_ssl.append("?").append(ranger_jpa_jdbc_url_extra_args);
						}
						propertiesMap.put("ranger-rms.jpa.jdbc.url", ranger_jpa_jdbc_url_no_ssl.toString());
					}
					ranger_jpa_jdbc_url=propertiesMap.get("ranger-rms.jpa.jdbc.url");
					if(StringUtils.isNotEmpty(ranger_jpa_jdbc_url)) {
						props.put("ranger-rms.jpa.jdbc.url", ranger_jpa_jdbc_url);
					}
					logger.info("ranger-rms.jpa.jdbc.url="+ranger_jpa_jdbc_url);
				}
			}
		}
	}

	keySet = props.keySet();
	for (Object key : keySet) {
		String keyStr = key.toString();
		if (logger.isDebugEnabled()) {
			logger.debug("PropertiesUtil:[" + keyStr + "][" + (keyStr.toLowerCase().contains("pass") ? "********]" : props.get(keyStr)) + "]");
		}
	}

	super.processProperties(beanFactory, props);
    }

    public static String getProperty(String key, String defaultValue) {
	if (key == null) {
	    return null;
	}
	String rtrnVal = propertiesMap.get(key);
	if (rtrnVal == null) {
	    rtrnVal = defaultValue;
	}
	return rtrnVal;
    }

    public static String getProperty(String key) {
	if (key == null) {
	    return null;
	}
	return propertiesMap.get(key);
    }

    public static String[] getPropertyStringList(String key) {
	if (key == null) {
	    return null;
	}
	String value = propertiesMap.get(key);
	if (value != null) {
	    String[] splitValues = value.split(",");
	    String[] returnValues = new String[splitValues.length];
	    for (int i = 0; i < splitValues.length; i++) {
		returnValues[i] = splitValues[i].trim();
	    }
	    return returnValues;
	} else {
	    return new String[0];
	}
    }

    public static Integer getIntProperty(String key, int defaultValue) {
	if (key == null) {
	    return defaultValue;
	}
	String rtrnVal = propertiesMap.get(key);
	if (rtrnVal == null) {
	    return defaultValue;
	}
	return Integer.valueOf(rtrnVal);
    }

    public static long getLongProperty(String key, long defaultValue) {
		if (key == null) {
		    return defaultValue;
		}
		String rtrnVal = propertiesMap.get(key);
		if (rtrnVal == null) {
		    return defaultValue;
		}
		return Long.parseLong(rtrnVal);
    }

    public static Integer getIntProperty(String key) {
	if (key == null) {
	    return null;
	}
	String rtrnVal = propertiesMap.get(key);
	if (rtrnVal == null) {
	    return null;
	}
	return Integer.valueOf(rtrnVal);
    }

    public static boolean getBooleanProperty(String key, boolean defaultValue) {
	if (key == null) {
	    return defaultValue;
	}
	String value = getProperty(key);
	if (value == null) {
	    return defaultValue;
	}
	return Boolean.parseBoolean(value);
    }
	public static Map<String, String> getPropertiesMap() {
		return propertiesMap;
	}
	public static Properties getProps() {
		Properties ret = new Properties();
		ret.putAll(propertiesMap);
		return ret;
	}
}
