/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.ranger.hms;

import org.apache.commons.lang.StringUtils;
import org.apache.ranger.rms.server.RMSConfig;
import org.apache.ranger.rms.util.AppConstants;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/*
 * Assumption: There is exactly one HMS for a combination of  higher-level service (such as a Hive service) and lower-level service (such as HDFS)  in Ranger.
 *             This assumption may not be true, as the security-domain (represented by Ranger) and storage-domain (represented by HMS) may not have exactly one-to-one relationship for all potential resources that need authoriation.
 */

public class HMSConfiguration {
    public static final String TheHmsName = "HMS";
    private Map<String, HMSSpecification> hmsConfiguration = new HashMap<>();
    private static volatile HMSConfiguration me;

    public static HMSConfiguration getHMSConfiguration() {
        if (me == null) {
            synchronized (HMSConfiguration.class) {
                if (me == null) {
                    me = new HMSConfiguration();
                    /* Only one entry at this time */
                    Set<Map<String, List<String>>> resourceRoots = new HashSet<>();

                    /* No resource-filtering for now */
                    /*
                    Map<String, List<String>> resourceRoot = new HashMap<>();

                    List<String> values = new ArrayList<>();
                    values.add("default");
                    resourceRoot.put(AppConstants.RESOURCE_DATABASE, values);
                    resourceRoots.add(resourceRoot);
                     */
                    RMSConfig conf = RMSConfig.getInstance();

                    String sourceServiceType = conf.get(AppConstants.CONFIG_PREFIX + TheHmsName + AppConstants.PROP_RMS_SOURCE_SERVICE_TYPE, "hdfs");
                    String sourceServiceName = conf.get(AppConstants.CONFIG_PREFIX + TheHmsName + AppConstants.PROP_RMS_SOURCE_SERVICE_NAME, "cm_hdfs");

                    String targetServiceType = conf.get(AppConstants.CONFIG_PREFIX + TheHmsName + AppConstants.PROP_RMS_TARGET_SERVICE_TYPE, "hive");
                    String targetServiceName = conf.get(AppConstants.CONFIG_PREFIX + TheHmsName + AppConstants.PROP_RMS_TARGET_SERVICE_NAME, "cm_hive");

                    String mappingSourceURL = conf.get(AppConstants.PROP_HiveMetaStore_URIs);

                    boolean isMapManagedTables = conf.getBoolean(AppConstants.CONFIG_PREFIX + TheHmsName + AppConstants.PROP_RMS_MAP_MANAGED_TABLES, false);

                    HLServiceSpecification serviceSpecification = new HLServiceSpecification(targetServiceName, resourceRoots);
                    HMSType hmsType = new HMSType("HiveToHdfs", targetServiceType, sourceServiceType);
                    HMSSpecification hmsSpecification = new HMSSpecification(TheHmsName, 1, hmsType, mappingSourceURL, sourceServiceName, serviceSpecification, isMapManagedTables);
                    me.hmsConfiguration.put(TheHmsName, hmsSpecification);
                }
            }
        }
        return me;
    }

    public static Set<String> getHMSNames() {
        return getHMSConfiguration().hmsConfiguration.keySet();
    }

    public static Set<String> getAllLlServiceNames() {
        Set<String> ret = new HashSet<>();
        for (String hmsName : getHMSNames()) {
            ret.add(getLlServiceName(hmsName));
        }
        return ret;
    }
    public static Set<String> getAllHlServiceNames() {
        Set<String> ret = new HashSet<>();
        for (String hmsName : getHMSNames()) {
            ret.add(getHlServiceName(hmsName));
        }
        return ret;
    }
    public static String getHlServiceType(String name) {
        HMSSpecification hmsSpecification = getHMSConfiguration().hmsConfiguration.get(name);
        return hmsSpecification != null ? hmsSpecification.type.hlServiceType : StringUtils.EMPTY;
    }
    public static String getLlServiceType(String name) {
        HMSSpecification hmsSpecification = getHMSConfiguration().hmsConfiguration.get(name);
        return hmsSpecification != null ? hmsSpecification.type.llServiceType : StringUtils.EMPTY;
    }

    public String getHMSConnectionURL(String name) {
        HMSSpecification hmsSpecification = getHMSConfiguration().hmsConfiguration.get(name);
        return hmsSpecification != null ? hmsSpecification.connectionURL : StringUtils.EMPTY;
    }

    public static String getLlServiceName(String name) {
        HMSSpecification hmsSpecification = getHMSConfiguration().hmsConfiguration.get(name);
        return hmsSpecification != null ? hmsSpecification.llServiceName : StringUtils.EMPTY;
    }

    public static String getHlServiceName(String name) {
        String ret = null;
        HMSSpecification hmsSpecification = getHMSConfiguration().hmsConfiguration.get(name);
        if (hmsSpecification != null && hmsSpecification.hlServiceSpecification != null) {
            ret = hmsSpecification.hlServiceSpecification.hlServiceName;
        }
        return ret;
    }

    public static Set<Map<String, List<String>>> getHlServiceRoots(String name) {
        Set<Map<String, List<String>>> ret = null;
        HMSSpecification hmsSpecification = getHMSConfiguration().hmsConfiguration.get(name);
        if (hmsSpecification != null) {
            HLServiceSpecification hlServiceSpecification = hmsSpecification.hlServiceSpecification;
            if (hlServiceSpecification != null) {
                ret = hlServiceSpecification.resourceRoots;
            }
        } else {
            ret = Collections.emptySet();
        }
        return ret;
    }

    public static Map<String, Set<Map<String, List<String>>>> getAllHlServiceRoots() {
        Map<String, Set<Map<String, List<String>>>> ret = null;
        for (String hmsName : getHMSNames()) {
            HMSSpecification hmsSpecification = getHMSConfiguration().hmsConfiguration.get(hmsName);
            if (hmsSpecification != null) {
                HLServiceSpecification hlServiceSpecification = hmsSpecification.hlServiceSpecification;
                if (hlServiceSpecification != null) {
                    if (ret == null) {
                        ret = new HashMap<>();
                    }
                    ret.put(hlServiceSpecification.hlServiceName, hlServiceSpecification.resourceRoots);
                }
            }
        }
        return ret;
    }

    public static boolean getIsMapManagedTables(String name) {
        HMSSpecification hmsSpecification = getHMSConfiguration().hmsConfiguration.get(name);
        return hmsSpecification != null && hmsSpecification.isMapManagedTables;
    }

    private HMSConfiguration() {}

    private static class HMSSpecification {
        String   name;
        int      id;
        HMSType  type;
        String   connectionURL;
        String   llServiceName;
        HLServiceSpecification hlServiceSpecification;
        boolean  isMapManagedTables;

        HMSSpecification(String name, int id, HMSType type, String connectionURL, String llServiceName, HLServiceSpecification hlServiceSpecification, boolean isMapManagedTables) {
            this.name = name;
            this.id = id;
            this.type = type;
            this.connectionURL = connectionURL;
            this.llServiceName = llServiceName;
            this.hlServiceSpecification = hlServiceSpecification;
            this.isMapManagedTables = isMapManagedTables;
        }

    }
    private static class HLServiceSpecification {
        String                     hlServiceName;
        Set<Map<String, List<String>>> resourceRoots;

        HLServiceSpecification(String hlServiceName, Set<Map<String, List<String>>> resourceRoots) {
            this.hlServiceName = hlServiceName;
            this.resourceRoots = resourceRoots;
        }
    }

    private static class HMSType {
        String name;
        String hlServiceType;
        String llServiceType;

        HMSType(String name, String hlServiceType, String llServiceType) {
            this.name = name;
            this.hlServiceType = hlServiceType;
            this.llServiceType = llServiceType;
        }
    }
}
