/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.hms;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.PostConstruct;
import javax.security.auth.login.LoginException;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.hadoop.hive.metastore.HiveMetaStoreClient;
import org.apache.hadoop.hive.metastore.TableType;
import org.apache.hadoop.hive.metastore.api.Database;
import org.apache.hadoop.hive.metastore.api.GetProjectionsSpec;
import org.apache.hadoop.hive.metastore.api.NoSuchObjectException;
import org.apache.hadoop.hive.metastore.api.NotificationEvent;
import org.apache.hadoop.hive.metastore.api.PrincipalType;
import org.apache.hadoop.hive.metastore.api.Table;
import org.apache.ranger.plugin.model.RangerPolicy.RangerPolicyResource;
import org.apache.ranger.plugin.policyengine.RangerAccessResource;
import org.apache.ranger.plugin.policyengine.RangerAccessResourceImpl;
import org.apache.ranger.plugin.policyresourcematcher.RangerDefaultPolicyResourceMatcher;
import org.apache.ranger.plugin.policyresourcematcher.RangerPolicyResourceMatcher;
import org.apache.ranger.plugin.util.RangerPerfTracer;
import org.apache.ranger.rms.server.RMSConfig;
import org.apache.ranger.rms.util.AppConstants;
import org.apache.ranger.rms.util.PathUtils;
import org.apache.ranger.rms.util.RMSTransactionHook;
import org.apache.ranger.rms.util.ResourceManagementHook;
import org.apache.ranger.rms.util.AppConstants.FILE_SCHEME_TYPES;
import org.apache.ranger.rms.view.RMSNotification;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

@Component
public class HMSFollower {

	private Map<String, HiveSimpleConnectionFactory> hiveConnectionFactories = new HashMap<>();

	private static final Logger LOG      = LoggerFactory.getLogger(HMSFollower.class);
	private static final Log    PERF_LOG = RangerPerfTracer.getPerfLogger("HMSFollower");

	private static final int    DEFAULT_RMS_TRANSACTION_SIZE_FOR_PERSISTING_MAPPINGS = 200;
	private static final int    DEFAULT_RMS_BATCH_SIZE_FOR_DOWNLOADING_MAPPINGS      = 400;
	private static final boolean DEFAULT_RMS_USE_THREAD_FOR_PERSISTING_MAPPINGS  = false;
	private static final boolean DEFAULT_RMS_ENABLE_DATABASE_SYNC                = true;

	private RMSTransactionHook rmsTransactionHook;
	private int                maxTransactionSize;
	private int                maxBatchSize;
	private int                maxRequestedNotifications;
	private boolean            isUseCommitThread;
	private String             supportedScheme;
	private boolean            isDbSyncEnbled;

	private ResourceManagementHook resourceManagementHook;

	@PostConstruct
	public void init() throws LoginException, IOException {
		RMSConfig conf = RMSConfig.getInstance();
		maxTransactionSize = conf.getInt(AppConstants.PROP_RMS_TRANSACTION_SIZE_FOR_PERSISTING_MAPPINGS, DEFAULT_RMS_TRANSACTION_SIZE_FOR_PERSISTING_MAPPINGS);
		maxBatchSize = conf.getInt(AppConstants.PROP_RMS_BATCH_SIZE_FOR_DOWNLOADING_MAPPINGS, DEFAULT_RMS_BATCH_SIZE_FOR_DOWNLOADING_MAPPINGS);
		isUseCommitThread = conf.getBoolean(AppConstants.PROP_RMS_USE_THREAD_FOR_PERSISTING_MAPPINGS, DEFAULT_RMS_USE_THREAD_FOR_PERSISTING_MAPPINGS);
		maxRequestedNotifications = conf.getInt(AppConstants.PROP_RMS_MAX_REQUESTED_NOTIFICATIONS, HiveNotificationFetcher.getMaxRequestedNotifications());
		if (maxRequestedNotifications < 1 || maxRequestedNotifications > HiveNotificationFetcher.getMaxRequestedNotifications()) {
			maxRequestedNotifications = HiveNotificationFetcher.getMaxRequestedNotifications();
		}
		supportedScheme = conf.get(AppConstants.PROP_RMS_SUPPORTED_URI_SCHEME, "hdfs");
		isDbSyncEnbled = conf.getBoolean(AppConstants.PROP_RMS_ENABLE_DATABASE_SYNC, DEFAULT_RMS_ENABLE_DATABASE_SYNC);

		for (String hmsName : HMSConfiguration.getHMSNames()) {
			//HiveConf hiveConf = conf.getHiveConf();
			HiveSimpleConnectionFactory hiveConnectionFactory = new HiveSimpleConnectionFactory(conf);
			hiveConnectionFactory.init(hmsName);
			hiveConnectionFactories.put(hmsName, hiveConnectionFactory);
		}
	}

	public void setTransactionHook(RMSTransactionHook rmsTransactionHook) {
		LOG.info("setTransactionHook called with " + rmsTransactionHook);
		this.rmsTransactionHook = rmsTransactionHook;
	}

	public void setResourceManagementHook(ResourceManagementHook resourceManagementHook) {
		LOG.info("setResourceManagementHook called with " + resourceManagementHook);
		this.resourceManagementHook = resourceManagementHook;
	}

	public long getCurrentNotificationId(String hmsName) throws LoginException {
		HiveSimpleConnectionFactory hiveConnectionFactory = hiveConnectionFactories.get(hmsName);
		if (hiveConnectionFactory == null) {
			throw new LoginException("No hive-connection factory for hms:[" + hmsName + "]");
		}

		HiveNotificationFetcher hiveNotificationFetcher = new HiveNotificationFetcher(hiveConnectionFactory);
		long currentNotificationID = -1;
		try {
			currentNotificationID = hiveNotificationFetcher.getCurrentNotificationId();
		} catch (Exception e) {
			LOG.error("Error in reading Current Notification ID", e);
		} finally {
			hiveNotificationFetcher.close();
		}
		return currentNotificationID;
	}

	private Table getTable(HiveMetaStoreClient hiveMetaStoreClient, NotificationEvent event, boolean isMapManagedTables) {
		Table table = null;
		try {
			table = hiveMetaStoreClient.getTable(event.getDbName(), event.getTableName());
			if (!isMapManagedTables && !TableType.EXTERNAL_TABLE.name().equals(table.getTableType())) {
				table = null;
			}
		} catch (NoSuchObjectException ex) {
			LOG.warn("Table doesn't exist:" + event.getDbName() + AppConstants.DBNAME_TABLENAME_SEPARATOR + event.getTableName());
		} catch (Exception ex) {
			LOG.error("Exception thrown when getting table from event:[" + event.getEventId() + "], message:[" +  ex.getMessage() + "]");
		}
		return table;
	}

	private Database getDatabase(HiveMetaStoreClient hiveMetaStoreClient, NotificationEvent event) {
		Database database = null;
		try {
			database = hiveMetaStoreClient.getDatabase(event.getDbName());
		} catch (NoSuchObjectException ex) {
			LOG.warn("Database doesn't exist:" + event.getDbName());
		} catch (Exception ex) {
			LOG.error("Exception thrown when getting database from event:[" + event.getEventId() + "], message:[" +  ex.getMessage() + "]");
		}
		return database;
	}

	private List<String> getTableNames(HiveMetaStoreClient hiveMetaStoreClient, String dbName, boolean isMapManagedTables) throws Exception {
		List<String> ret;

		ret = hiveMetaStoreClient.getTables(dbName, AppConstants.TABLE_PATTERN, TableType.EXTERNAL_TABLE);

		if (isMapManagedTables) {
			List<String> managedTableNames = hiveMetaStoreClient.getTables(dbName, AppConstants.TABLE_PATTERN, TableType.MANAGED_TABLE);
			ret.addAll(managedTableNames);
		}


		return ret;
	}
	private Map<String, RangerPolicyResource> getElements(Table table) {
		return getElements(table.getDbName(), table.getTableName());
	}
	private Map<String, RangerPolicyResource> getElements(Database database) {
		return getElements(database.getName(), null);
	}

	private Map<String, RangerPolicyResource> getElements(NotificationEvent event) {
		return getElements(event.getDbName(), event.getTableName());
	}

	private Map<String, RangerPolicyResource> getElements(String dbName, String tableName) {
		Map<String, RangerPolicyResource> elements = new HashMap<>();
		elements.put(AppConstants.RESOURCE_DATABASE, new RangerPolicyResource(dbName));
		if (StringUtils.isNotBlank(tableName)) {
			elements.put(AppConstants.RESOURCE_TABLE, new RangerPolicyResource(tableName));
		}
		return elements;
	}

	private Map<String, RangerPolicyResource> getSourceElements(Table table) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("Location for Table:[" + table.getDbName() + ":" + table.getTableName() + "] is [" + table.getSd().getLocation() + "]");
		}
		Map<String, String> locationMap = PathUtils.parsePath(table.getSd().getLocation());
		if (MapUtils.isEmpty(locationMap)) {
			LOG.error("location is null for table:[" + table.getDbName() + ":" + table.getTableName() + "]. Will not map this table");
			return null;
		}

		Map<String, RangerPolicyResource> ret = new HashMap<>();
		ret.put(AppConstants.RESOURCE_PATH, new RangerPolicyResource(locationMap.get(AppConstants.RESOURCE_PATH), false, true));
		if(FILE_SCHEME_TYPES.s3a.name().equals(supportedScheme)) {
			ret.put(AppConstants.RESOURCE_BUCKET, new RangerPolicyResource(locationMap.get(AppConstants.RESOURCE_BUCKET), false, false));
		}
		return ret;
	}

	private Map<String, RangerPolicyResource> getSourceElements(Database database) throws Exception {
		if (LOG.isDebugEnabled()) {
			LOG.debug("Location for Database:[" + database.getName() + "] is [" + database.getLocationUri() + "]");
		}
		Map<String, String> locationMap =  PathUtils.parsePath(database.getLocationUri());
		if (MapUtils.isEmpty(locationMap)) {
			LOG.error("location is null for table:[" + database.getName() + "]. Will not map this database");
			return null;
		}

		Map<String, RangerPolicyResource> ret = new HashMap<>();
		ret.put(AppConstants.RESOURCE_PATH, new RangerPolicyResource(locationMap.get(AppConstants.RESOURCE_PATH), false, true));
		if(FILE_SCHEME_TYPES.s3a.name().equals(supportedScheme)) {
			ret.put(AppConstants.RESOURCE_BUCKET, new RangerPolicyResource(locationMap.get(AppConstants.RESOURCE_BUCKET), false, false));
		}
		return ret;
	}

	private RangerAccessResource getRangerAccessResource(Map<String, RangerPolicyResource> elements) {
		RangerAccessResourceImpl ret = new RangerAccessResourceImpl();
		for (Map.Entry<String, RangerPolicyResource> element : elements.entrySet()) {
			ret.setValue(element.getKey(), element.getValue().getValues());
		}
		return ret;
	}

	private boolean isMatch(Set<RangerDefaultPolicyResourceMatcher> matchers, Map<String, RangerPolicyResource> elements) {

		boolean ret = true;

		if (CollectionUtils.isNotEmpty(matchers)) {
			RangerAccessResource accessResource = getRangerAccessResource(elements);
			for (RangerDefaultPolicyResourceMatcher matcher : matchers) {
				ret = matcher.isMatch(accessResource, RangerPolicyResourceMatcher.MatchScope.SELF_OR_ANCESTOR, null);
				if (!ret) {
					break;
				}
			}
		} else {
			LOG.debug("No matchers specified. By default, all resources match!");
			ret = true;
		}
		return ret;
	}

	/*
	 * Return value : null - Filter did not match at the database level
	 *              : TRUE - Filter matched with SELF or SELF_WITH_ALL_DESCENDANTS (no need to check at table level)
	 *              : FALSE - Filter matched with DESCENDANT (need to check at table level)
	 */
	private Boolean getMatchType(Set<RangerDefaultPolicyResourceMatcher> matchers, String dbName) {
		Boolean ret = null;

		RangerAccessResourceImpl accessResource = new RangerAccessResourceImpl();
		accessResource.setValue(AppConstants.RESOURCE_DATABASE, dbName);

		if (CollectionUtils.isNotEmpty(matchers)) {
			for (RangerDefaultPolicyResourceMatcher matcher : matchers) {

				RangerPolicyResourceMatcher.MatchType matchType = matcher.getMatchType(accessResource, null);

				if (matchType == RangerPolicyResourceMatcher.MatchType.SELF || matchType == RangerPolicyResourceMatcher.MatchType.SELF_AND_ALL_DESCENDANTS) {
					ret = Boolean.TRUE;
					break;
				} else if (matchType != RangerPolicyResourceMatcher.MatchType.NONE) {
					ret = Boolean.FALSE;
				}
			}
		} else {
			LOG.debug("No matchers specified. By default, all resources match!");
			ret = true;
		}

		return ret;
	}

	public boolean purgeIfInvalidServices(String hmsName) {
		boolean ret = isServiceDataValid(hmsName);

		if (!ret) {
			LOG.error("Service data for HMS:[" + hmsName + "] is not available! Will not be able to convert HMS resources to RMS resources!");

			String llServiceName = HMSConfiguration.getLlServiceName(hmsName);
			String hlServiceName = HMSConfiguration.getHlServiceName(hmsName);

			rmsTransactionHook.purgeServiceResourcesAndMappings(llServiceName, hlServiceName);
			commitWork(hmsName, Collections.emptyList(), -1L, true);
		}

		return ret;
	}

	public void syncFromHMS(String hmsName, Set<RangerDefaultPolicyResourceMatcher> matchers, long lastSyncedNotificationId, long currentNotificationId) {
		LOG.debug("==> syncFromHMS(hmsName=" + hmsName + ",lastSyncedNotificationId=" + lastSyncedNotificationId + ", currentNotificationId=" + currentNotificationId);

		if (lastSyncedNotificationId < currentNotificationId) {

			HiveSimpleConnectionFactory hiveConnectionFactory = hiveConnectionFactories.get(hmsName);

			if (hiveConnectionFactory != null) {

				long syncedNotificationId = lastSyncedNotificationId;

				if (lastSyncedNotificationId > 0L) {

					for (syncedNotificationId = lastSyncedNotificationId; syncedNotificationId < currentNotificationId; ) {

						try {
							syncedNotificationId = handleDeltaSync(hmsName, hiveConnectionFactory, matchers, syncedNotificationId, currentNotificationId);
						} catch (Exception e) {
							LOG.error("Cannot do delta-sync. Breaking out of delta-sync loop. Will try full sync!", e);
							break;
						}
					}
				}

				if (syncedNotificationId < currentNotificationId) {
					try {
						syncedNotificationId = handleFullSync(hmsName, hiveConnectionFactory, matchers, -1L, currentNotificationId);
						if (syncedNotificationId == -1L) {
							LOG.info("Full-sync returned no databases!! Unusual condition, will need to retry full-sync in next cycle!");
						}
					} catch (Exception e) {
						LOG.error("Cannot do full-sync!", e);
					}
				}

			} else {
				LOG.error("Failed to get HiveConnectionFactory for hms:[" + hmsName + "]");
			}
		} else {
			LOG.debug("No need to do anything, currentNotificationId:[" + currentNotificationId + "], lastSyncedNotificationId:[" + lastSyncedNotificationId + "]");
		}
		LOG.debug("<== syncFromHMS(hmsName=" + hmsName + ",lastSyncedNotificationId=" + lastSyncedNotificationId + ", currentNotificationId=" + currentNotificationId);

	}

	public long handleDeltaSync(String hmsName, HiveSimpleConnectionFactory hiveConnectionFactory, final Set<RangerDefaultPolicyResourceMatcher> matchers, final long lastSyncedNotificationId, final long currentNotificationId) throws Exception {

		LOG.debug("==> HMSFollower.handleDeltaSync(hmsName=" + hmsName + ", lastSyncedNotificationId=" + lastSyncedNotificationId + ", currentNotificationId=" + currentNotificationId + ")");

		long ret = lastSyncedNotificationId;

		RangerPerfTracer perf = null;

		if (RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
			perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "HMSFollower.getResourceMappingsFromHMS");
		}

		List<RMSNotification> notifications = new ArrayList<>();
		HMSClient            hmsClient     = null;
		boolean isMapManagedTables = HMSConfiguration.getIsMapManagedTables(hmsName);
		HiveNotificationFetcher hiveNotificationFetcher = null;

		try {
			hmsClient = hiveConnectionFactory.connect();
			hiveNotificationFetcher = new HiveNotificationFetcher(hiveConnectionFactory);
			if (LOG.isDebugEnabled()) {
				LOG.debug("Maximum number of requested notification during delta-sync is maxRequestedNotifications = "+ maxRequestedNotifications);
			}
			List<NotificationEvent> events = hiveNotificationFetcher.fetchNotifications(lastSyncedNotificationId, maxRequestedNotifications);

			for (NotificationEvent event : events) {

				LOG.debug("Processing NotificationEvent:[" + event + "]");

				ret = event.getEventId();

				if (AppConstants.NF_CREATE_TABLE.equalsIgnoreCase(event.getEventType())) {
					Table table = getTable(hmsClient.getClient(), event, isMapManagedTables);
					if (table != null) {
						Map<String, RangerPolicyResource> elements = getElements(table);
						if (!isMatch(matchers, elements)) {
							LOG.debug("Event with id:[" + event.getEventId() + "] did not match specified filter. Dropping it!");
							continue;
						}
						Map<String, RangerPolicyResource> sourceElements = getSourceElements(table);
						if (MapUtils.isNotEmpty(sourceElements)) {
							String ownerName = table.getOwnerType() == PrincipalType.USER ? table.getOwner() : null;
							String tableType = table.getTableType();
							Map<String, String> additionalInfo = new HashMap<>();
							additionalInfo.put("tableType", tableType);
							RMSNotification vXResourceMapping = new RMSNotification(event.getEventId(), new Date(), hmsClient.getHmsName(), AppConstants.OP_ADD, elements, sourceElements, ownerName, additionalInfo);
							notifications.add(vXResourceMapping);
						}
					}
				} else if (AppConstants.NF_ALTER_TABLE.equalsIgnoreCase(event.getEventType())) {
					Table table = getTable(hmsClient.getClient(), event, isMapManagedTables);
					if (table != null) {
						Map<String, RangerPolicyResource> elements = getElements(table);

						if (!isMatch(matchers, elements)) {
							continue;
						}
						Map<String, RangerPolicyResource> sourceElements = getSourceElements(table);
						if (MapUtils.isNotEmpty(sourceElements)) {
							String ownerName = table.getOwnerType() == PrincipalType.USER ? table.getOwner() : null;
							String tableType = table.getTableType();
							Map<String, String> additionalInfo = new HashMap<>();
							additionalInfo.put("tableType", tableType);
							RMSNotification vXResourceMapping = new RMSNotification(event.getEventId(), new Date(), hmsClient.getHmsName(), AppConstants.OP_UPDATE, elements, sourceElements, ownerName, additionalInfo);
							notifications.add(vXResourceMapping);
						}
					}
				} else if (AppConstants.NF_DROP_TABLE.equalsIgnoreCase(event.getEventType())) {
					Map<String, RangerPolicyResource> elements = getElements(event);
					if (!isMatch(matchers, elements)) {
						LOG.debug("Event with id:[" + event.getEventId() + "] did not match specified filter. Dropping it!");

						continue;
					}
					RMSNotification vXResourceMapping = new RMSNotification(event.getEventId(), new Date(), hmsClient.getHmsName(), AppConstants.OP_DELETE, elements, null);
					notifications.add(vXResourceMapping);
				} else if (AppConstants.NF_CREATE_DATABASE.equalsIgnoreCase(event.getEventType())) {
					Database database = getDatabase(hmsClient.getClient(), event);
					if (database != null) {
						Map<String, RangerPolicyResource> elements = getElements(database);
						if (!isMatch(matchers, elements)) {
							LOG.debug("Event with id:[" + event.getEventId() + "] did not match specified filter. Dropping it!");
							continue;
						}
						if (isDbSyncEnbled) {
							Map<String, RangerPolicyResource> sourceElements = getSourceElements(database);
							if (MapUtils.isNotEmpty(sourceElements)) {
								String ownerName = database.getOwnerType() == PrincipalType.USER ? database.getOwnerName() : null;
								Map<String, String> additionalInfo = new HashMap<>();
								RMSNotification vXResourceMapping = new RMSNotification(event.getEventId(), new Date(), hmsClient.getHmsName(), AppConstants.OP_ADD, elements, sourceElements, ownerName, additionalInfo);
								notifications.add(vXResourceMapping);
							}
						}
					}
				} else if (AppConstants.NF_DROP_DATABASE.equalsIgnoreCase(event.getEventType())) {
					Map<String, RangerPolicyResource> elements = getElements(event);
					if (!isMatch(matchers, elements)) {
						LOG.debug("Event with id:[" + event.getEventId() + "] did not match specified filter. Dropping it!");

						continue;
					}
					if (isDbSyncEnbled) {
						RMSNotification vXResourceMapping = new RMSNotification(event.getEventId(), new Date(), hmsClient.getHmsName(), AppConstants.OP_DELETE, elements, null);
						notifications.add(vXResourceMapping);
					}
				} else {
					LOG.debug("Notification-Event-Type:[" + event.getEventType() + "] is not handled currently. Ignoring Notification-Event with id:[" + event.getEventId() + "]");
				}

				if (notifications.size() >= maxTransactionSize) {
					commitWork(hmsName, notifications, ret, true);
					notifications.clear();
				}
			}
			LOG.debug("Updating currentNotificationId for hmsName:[" + hmsName + "], currentNotificationId:[" + currentNotificationId + "]");

			commitWork(hmsName, notifications, ret, true);
			notifications.clear();

		} catch (Exception ex) {
			LOG.error("HMS Notification fetch failed", ex);
			throw ex;
		} finally {
			if (hiveNotificationFetcher != null) {
				hiveNotificationFetcher.close();
			}
			if (hmsClient != null) {
				hmsClient.close();
			}
			RangerPerfTracer.log(perf);
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("<== HMSFollower.handleDeltaSync(hmsName=" + hmsName + ", lastSyncedNotificationId=" + ret + ", currentNotificationId=" + currentNotificationId + ")");
		}

		return ret;
	}

	private long handleFullSync(String hmsName, HiveSimpleConnectionFactory hiveConnectionFactory, Set<RangerDefaultPolicyResourceMatcher> matchers, final long lastSyncedNotificationId, final long currentNotificationId) throws Exception {

		LOG.info("==> HMSFollower.handleFullSync(hmsName=" + hmsName + ", lastSyncedNotificationId=" + lastSyncedNotificationId + ")");

		long ret;
		List<RMSNotification> notifications = new ArrayList<>();
		HMSClient            hmsClient     = null;

		try {
			String llServiceName = HMSConfiguration.getLlServiceName(hmsName);
			String hlServiceName = HMSConfiguration.getHlServiceName(hmsName);
			boolean isMapManagedTables = HMSConfiguration.getIsMapManagedTables(hmsName);

			rmsTransactionHook.purgeServiceResourcesAndMappings(llServiceName, hlServiceName);
			commitWork(hmsName, Collections.emptyList(), -1L, true);

			hmsClient = hiveConnectionFactory.connect();
			HiveMetaStoreClient hiveMetaStoreClient = hmsClient.getClient();
			List<String> dbList = hiveMetaStoreClient.getAllDatabases();

			if (CollectionUtils.isNotEmpty(dbList)) {
				for (String dbName : dbList) {
					Boolean matchType = getMatchType(matchers, dbName);
					if (matchType == null) {
						LOG.info("Database:[" + dbName + "] did not match specified filter. Will not sync its data to RMS!");

						continue;
					}

					Database database = hiveMetaStoreClient.getDatabase(dbName);

					if (database == null) {
						continue;
					}

					Map<String, RangerPolicyResource> sourceElements;

					if (isDbSyncEnbled) {
						sourceElements = getSourceElements(database);

						if (MapUtils.isNotEmpty(sourceElements)) {
							String ownerName = database.getOwnerType() == PrincipalType.USER ? database.getOwnerName() : null;
							Map<String, String> additionalInfo = new HashMap<>();
							RMSNotification vXResourceMapping = new RMSNotification(currentNotificationId, new Date(), hmsClient.getHmsName(), AppConstants.OP_ADD, getElements(database), sourceElements, ownerName, additionalInfo);
							notifications.add(vXResourceMapping);
						}
					}

					List<String> tableNames = getTableNames(hiveMetaStoreClient, dbName, isMapManagedTables);

					if (CollectionUtils.isNotEmpty(tableNames)) {
						checkHMSTablePermissions(hiveMetaStoreClient, dbName, tableNames.get(0));

						LOG.info("Database:[" + dbName + "]: There are [" + tableNames.size() + "] tables to be retrieved for full sync. Will do it in batches of size [" + maxBatchSize + "] or less");

						// Split tableNames into batches of maxBatchSize or smaller to accommodate HMS limitation

						int currentListIndex = 0;

						while (currentListIndex < tableNames.size()) {
							final int batchSize = Math.min(tableNames.size() - currentListIndex,  maxBatchSize);

							List<String> currentTableNames = tableNames.subList(currentListIndex, currentListIndex + batchSize);
							currentListIndex += batchSize;

							RangerPerfTracer perf = null;

							if (RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
								perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "HMSFollower.getTableObjectsByName(size=" + batchSize + ")");
							}

							GetProjectionsSpec projectionsSpec = new GetProjectionsSpec();
							List<String> projectedFields = Arrays.asList("dbName", "tableName", "owner", "ownerType", "tableType", "sd.location");
							projectionsSpec.setFieldList(projectedFields);

							List<Table> tables = hiveMetaStoreClient.getTables(null, dbName, currentTableNames, projectionsSpec);

							RangerPerfTracer.logAlways(perf);

							LOG.info("Database:[" + dbName + "]: Processing batch of tables from [" + (currentListIndex - batchSize) + "] to [" + currentListIndex + "]");

							for (Table table : tables) {
								Map<String, RangerPolicyResource> elements = getElements(table);

								if (matchType.equals(Boolean.FALSE)) {
									// Check at table level
									if (!isMatch(matchers, elements)) {
										LOG.info("Database:[" + dbName + "], Table:[" + table.getTableName() + "] did not match specified filter. Will not sync it with RMS!");
										continue;
									}
								}
								sourceElements = getSourceElements(table);
								if (MapUtils.isNotEmpty(sourceElements)) {
									String ownerName = table.getOwnerType() == PrincipalType.USER ? table.getOwner() : null;
									String tableType = table.getTableType();
									Map<String, String> additionalInfo = new HashMap<>();
									additionalInfo.put("tableType", tableType);
									RMSNotification vXResourceMapping = new RMSNotification(currentNotificationId, new Date(), hmsName, AppConstants.OP_ADD, elements, sourceElements, ownerName, additionalInfo);
									notifications.add(vXResourceMapping);
								}
								if (notifications.size() >= maxTransactionSize) {
									commitWork(hmsName, notifications, -1L, false);
									notifications.clear();
								}
							}
						}
					}
				}

				LOG.debug("Adding final RMSNotification to the list for marking that full-sync was performed");

				commitWork(hmsName, notifications, currentNotificationId, false);
				notifications.clear();

				Map<String, RangerPolicyResource> dontCare = new HashMap<>();
				RMSNotification sentinel = new RMSNotification(currentNotificationId, new Date(), hmsName, AppConstants.OP_INVALID, dontCare, dontCare);
				notifications.add(sentinel);
				commitWork(hmsName, notifications, currentNotificationId, true);
				notifications.clear();

				ret = currentNotificationId;
			} else {
				LOG.info("No databases returned!! Should retry full-sync later!");
				ret = lastSyncedNotificationId;
			}

		} catch (Exception exception) {
			LOG.error("Error doing full-sync", exception);
			throw exception;
		} finally {
			if (hmsClient != null) {
				hmsClient.close();
			}
		}

		LOG.info("<== HMSFollower.handleFullSync(hmsName=" + hmsName + ", lastSyncedNotificationId=" + ret + ")");

		return ret;
	}

	private boolean isServiceDataValid(String hmsName) {
		return resourceManagementHook != null && resourceManagementHook.getServiceInfo(hmsName);
	}

	private void checkHMSTablePermissions(HiveMetaStoreClient hiveMetaStoreClient, String dbName, String tableName) throws Exception {
		hiveMetaStoreClient.getTable(dbName, tableName);
	}

	private void commitWork(String hmsName, List<RMSNotification> notifications, Long currentNotificationId, boolean isPersistNotification) {

		if (isUseCommitThread) {
			CommitThread t = new CommitThread(hmsName, notifications, currentNotificationId, isPersistNotification);
			t.setName("CommitWork-Thread-for" + hmsName + "-" + t.getId());
			t.setDaemon(true);
			t.start();

			try {
				t.join();
			} catch (InterruptedException ie) {
				LOG.error("Failed to commit mappings in a separate thread and using a new transaction", ie);
			}
		} else {
			rmsTransactionHook.commitWork(hmsName, notifications, currentNotificationId, isPersistNotification);
		}
	}

	private class CommitThread extends Thread {
		final String                  hmsName;
		final List<RMSNotification>    notifications;
		final Long                    currentNotificationId;
		final boolean                 isPersistNotification;

		CommitThread(String hmsName, List<RMSNotification> notifications, Long currentNotificationId, boolean isPersistNotification) {
			this.hmsName               = hmsName;
			this.notifications         = notifications;
			this.currentNotificationId = currentNotificationId;
			this.isPersistNotification = isPersistNotification;
		}

		@Override
		public void run() {
			try {
				rmsTransactionHook.commitWork(hmsName, notifications, currentNotificationId, isPersistNotification);
			} catch (Throwable ex) {
				LOG.error("Failed to commit " + notifications.size() + " mappings to persistence hmsName:[" + hmsName + "], ending with notificationId:[" + currentNotificationId + "] in a new transaction", ex);
			}
		}
	}

}
