/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.hms;

import java.io.IOException;
import java.security.PrivilegedExceptionAction;

import javax.security.auth.login.LoginException;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hive.metastore.HiveMetaStoreClient;
import org.apache.hadoop.security.SecureClientLogin;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.rms.util.AppConstants;
import org.apache.ranger.rms.util.PropertiesUtil;
import org.apache.ranger.rms.util.RMSToHMSKerberosContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Factory used to generate Hive connections. Supports insecure and Kerberos
 * connections. For Kerberos connections starts the ticket renewal thread and
 * sets up Kerberos credentials.
 */
public final class HiveSimpleConnectionFactory implements HiveConnectionFactory {
	private static final Logger LOGGER = LoggerFactory.getLogger(HiveSimpleConnectionFactory.class);

	/** site.xml configuration */
	private final Configuration conf;
	/** Hive configuration */
	//private final HiveConf hiveConf;
	private final boolean insecure;
	private String hmsName;
	private RMSToHMSKerberosContext kerberosContext = null;

	public HiveSimpleConnectionFactory(Configuration conf) {
		this.conf = conf;
		//this.hiveConf = hiveConf;
		insecure = !conf.getBoolean(AppConstants.PROP_HiveMetaStore_Auth_Type, true);
	}

	/**
	 * Initialize the Factory. For insecure connections there is nothing to
	 * initialize. For Kerberos connections sets up ticket renewal thread.
	 *
	 * @throws IOException          if could not establish connection
	 * @throws LoginException		if could not log in
	 */
	public void init(String hmsName) throws IOException, LoginException {
		this.hmsName = hmsName;
		if (insecure) {
			LOGGER.info("Using insecure connection to HMS");
		} else {
			String principal = SecureClientLogin.getPrincipal(conf.get(AppConstants.PROP_RMS_SERVICE_PRINCIPAL), conf.get(AppConstants.PROP_RMS_SERVER_BIND_ADDRESS));
			String keytab    = conf.get(AppConstants.PROP_RMS_SERVICE_KEYTAB);
			String nameRules = conf.get(AppConstants.PROP_NAME_RULES, PropertiesUtil.getProperty(AppConstants.PROP_NAME_RULES, AppConstants.DEFAULT_NAME_RULES) );
			LOGGER.info("principal:" + principal);
			if (LOGGER.isDebugEnabled()){
				LOGGER.debug("keytab:" + keytab);
			}
			if(SecureClientLogin.isKerberosCredentialExists(principal, keytab)) {
				LOGGER.info("nameRules:"+nameRules);
				kerberosContext = new RMSToHMSKerberosContext(principal, keytab, false);
			}
		}
	}
	/**
	 * Connect to HMS in unsecure mode or in Kerberos mode according to config.
	 *
	 * @return HMS connection
	 * @throws IOException          if could not establish connection
	 * @throws InterruptedException if connection was interrupted
	 */
	public HMSClient connect() throws IOException, InterruptedException {
		UserGroupInformation clientUGI;
		if (insecure) {
			clientUGI = UserGroupInformation.getCurrentUser();
		} else {
			clientUGI = UserGroupInformation.getUGIFromSubject(kerberosContext.getSubject());
		}
		return new HMSClient(clientUGI.doAs((PrivilegedExceptionAction<HiveMetaStoreClient>) () -> new HiveMetaStoreClient(conf)), hmsName);
	}

	@Override
	public void close() {
		if (kerberosContext != null) {
			kerberosContext.shutDown();
			kerberosContext = null;
		}
	}

}
