/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.plugin.service;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.authorization.hadoop.config.RangerPluginConfig;
import org.apache.ranger.authorization.hadoop.config.RangerRmsPluginConfig;
import org.apache.ranger.plugin.policyengine.RangerAccessRequest;
import org.apache.ranger.plugin.util.DownloadTrigger;
import org.apache.ranger.plugin.util.DownloaderTask;
import org.apache.ranger.plugin.util.RangerMappingProvider;
import org.apache.ranger.plugin.util.RangerMappingRefresher;
import org.apache.ranger.plugin.util.RangerMappingRetriever;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.Timer;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

public abstract class RangerChainedMapperPlugin extends RangerChainedPlugin {
    private static final Log LOG = LogFactory.getLog(RangerChainedMapperPlugin.class);

    public static final  String ACL_MAPPING_SOURCE_TYPE_PARAMETER                    = ".mapping.source.type";
    public static final  String ACL_MAPPING_SOURCE_CLASS_PARAMETER                   = ".mapping.source.impl";
    public static final  String ACL_MAPPING_SOURCE_URL_PARAMETER                     = ".mapping.source.url";
    public static final  String ACL_MAPPING_SOURCE_FILE_PARAMETER                    = ".mapping.source.file.name";
    public static final  String ACL_MAPPING_DOWNLOAD_INTERVAL_PARAMETER              = ".mapping.source.download.interval";

    private static final String ACL_MAPPING_SERVICE_WHITELISTED_USERS_PARAMETER      = ".whitelisted.users";
    private static final String ACL_MAPPING_SERVICE_WHITELISTED_GROUPS_PARAMETER     = ".whitelisted.groups";

    protected RangerMappingProvider mappingProvider;
    protected RangerMappingRefresher mappingRefresher;
    private   Set<String>           whitelistedUsers;
    private   Set<String>           whitelistedGroups;
    private   boolean               isInited;

    protected RangerChainedMapperPlugin(RangerBasePlugin rootPlugin, String serviceType, String serviceName) {
        super(rootPlugin, serviceType, serviceName);
    }

    @Override
    protected RangerBasePlugin buildChainedPlugin(String serviceType, String serviceName, String appId) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerChainedMapperPlugin.buildChainedPlugin(serviceName=" + serviceName + ")");
        }

        RangerPluginConfig pluginConfig = new RangerRmsPluginConfig(serviceType, serviceName, appId, rootPlugin.getConfig());

        final RangerBasePlugin ret = new RangerBasePlugin(pluginConfig);
        isInited = initChainedPlugin(ret);

        if (!isInited) {
            LOG.error("Could not initialize RangerChainedMapperPlugin object. The plugin for service:[" + serviceName + "] will not be operational!");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerChainedMapperPlugin.buildChainedPlugin(serviceName=" + serviceName + "): isInited:[" + isInited + "]");
        }

        return ret;
    }

    protected boolean initChainedPlugin(RangerBasePlugin plugin) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerChainedMapperPlugin.initChainedPlugin(serviceName=" + serviceName + ")");
        }

        boolean ret = false;

        RangerPluginConfig pluginConfig = plugin.getConfig();

        this.mappingProvider = new RangerMappingProvider(rootPlugin.getServiceType(), rootPlugin.getServiceName(), pluginConfig);

        // This should contain all required configurations
        String mappingRetrieverType      = pluginConfig.get(pluginConfig.getPropertyPrefix()    + ACL_MAPPING_SOURCE_TYPE_PARAMETER, "rest");
        String mappingRetrieverClass     = pluginConfig.get(pluginConfig.getPropertyPrefix()    + ACL_MAPPING_SOURCE_CLASS_PARAMETER);
        String mappingProviderUrl        = pluginConfig.get(pluginConfig.getPropertyPrefix()    + ACL_MAPPING_SOURCE_URL_PARAMETER);
        String mappingProviderFileName   = pluginConfig.get(pluginConfig.getPropertyPrefix()    + ACL_MAPPING_SOURCE_FILE_PARAMETER);
        int    downloadInterval          = pluginConfig.getInt(pluginConfig.getPropertyPrefix() + ACL_MAPPING_DOWNLOAD_INTERVAL_PARAMETER, 30*1000);

        String whitelistedUsers          = pluginConfig.get(pluginConfig.getPropertyPrefix()    + ACL_MAPPING_SERVICE_WHITELISTED_USERS_PARAMETER);
        String whitelistedGroups         = pluginConfig.get(pluginConfig.getPropertyPrefix()    + ACL_MAPPING_SERVICE_WHITELISTED_GROUPS_PARAMETER);

        if (LOG.isDebugEnabled()) {
            LOG.debug("Data for service-name = " + serviceName + ": [targetServiceType=" + serviceType + ", mappingRetrieverType=" + mappingRetrieverType
                    + ", downloadInterval=" + downloadInterval + ", mappingProviderUrl=" + mappingProviderUrl + ", mappingProviderFileName=" + mappingProviderFileName
                    + ", whitelistedUsers=" + whitelistedUsers + ", whitelistedGroups=" + whitelistedGroups + "]");
        }
        RangerMappingRetriever mappingRetriever = getMappingRetriever(mappingRetrieverType, mappingRetrieverClass, mappingProviderUrl, mappingProviderFileName);

        if (mappingRetriever != null) {
            mappingRetriever.init(rootPlugin.getServiceName(), serviceName, pluginConfig);

            this.whitelistedUsers = buildSet(whitelistedUsers);
            this.whitelistedGroups = buildSet(whitelistedGroups);

            if (LOG.isDebugEnabled()) {
                LOG.debug("Initialized chainedPlugin for service:[" + serviceName + "]");
            }

            final BlockingQueue<DownloadTrigger> mappingDownloadQueue = new LinkedBlockingQueue<>();

            this.mappingRefresher = new RangerMappingRefresher(plugin, mappingRetriever, mappingProvider, -1L, mappingDownloadQueue);
            LOG.info("Created RangerMappingRefresher Thread(" + mappingRefresher.getName() + ")");

            try {
                mappingRefresher.populateMappings();

                Timer mappingDownloadTimer = new Timer("mappingDownloadTimer", true);

                mappingDownloadTimer.schedule(new DownloaderTask(mappingDownloadQueue), downloadInterval, downloadInterval);

                if (LOG.isDebugEnabled()) {
                    LOG.debug("Scheduled mappingRefresher to download mappings every " + downloadInterval + " milliseconds");
                }

                mappingRefresher.setDaemon(true);
                mappingRefresher.startRefresher();

                ret = true;

            } catch (InterruptedException exception) {
                LOG.error("Caught InterruptedException. Cleaning up before interrupting this thread: [" + Thread.currentThread().getName() + "]");
                LOG.error("*** Mappings will NOT be downloaded every " + downloadInterval + " milliseconds ***");
                Thread.currentThread().interrupt();
            } catch (IllegalStateException exception) {
                LOG.error("Error scheduling mappingRefresher:", exception);
                LOG.error("*** Mappings will NOT be downloaded every " + downloadInterval + " milliseconds ***");
            }
        } else {
            LOG.error("Cannot initialize mapping-retriever");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerChainedMapperPlugin.initChainedPlugin(serviceName=" + serviceName + ") : ret:[" + ret + "]");
        }
        return ret;

    }

    protected boolean isNotWhitelisted(RangerAccessRequest request) {
        if (whitelistedUsers.contains(request.getUser())) {
            return false;
        }

        for (String group : request.getUserGroups()) {
            if (whitelistedGroups.contains(group)) {
                return false;
            }
        }

        return true;
    }

    protected boolean getIsInited() {
        return isInited;
    }

    private RangerMappingRetriever getMappingRetriever(String mappingRetrieverType, String mappingRetrieverClassName, String mappingProviderUrl, String mappingProviderFileName) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerChainedMapperPlugin.getMappingRetriever(mappingRetrieverType=" + mappingRetrieverType + ", mappingRetrieverClassName=" + mappingRetrieverClassName + ", mappingProviderUrl=" + mappingProviderUrl + ", mappingProviderFileName=" + mappingProviderFileName + ")");
        }

        final RangerMappingRetriever ret;

        if (StringUtils.equals(mappingRetrieverType, "rest")) {
            if (StringUtils.isEmpty(mappingProviderUrl)) {
                LOG.error("No Mapping provider URL");
            } else {
                if (StringUtils.isBlank(mappingRetrieverClassName)) {
                    mappingRetrieverClassName = "org.apache.ranger.plugin.util.RangerMappingRESTRetriever";
                }
            }
        } else if (StringUtils.equals(mappingRetrieverType, "file")) {
            if (StringUtils.isEmpty(mappingProviderFileName)) {
                LOG.error("No Mapping provider File");
            } else {
                if (StringUtils.isBlank(mappingRetrieverClassName)) {
                    LOG.error("No class provided for retrieving resource mappings");
                }
            }
        } else {
            LOG.error("Unknown mappingRetrieverType:[" + mappingRetrieverType + "]");
        }
        if (StringUtils.isNotEmpty(mappingRetrieverClassName)) {
            ret = (RangerMappingRetriever) getMapper(mappingRetrieverClassName, RangerMappingRetriever.class);
        } else {
            ret = null;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerChainedMapperPlugin.getMappingRetriever(mappingRetrieverType=" + mappingRetrieverType + ", mappingRetrieverClassName=" + mappingRetrieverClassName + ", mappingProviderUrl=" + mappingProviderUrl + ", mappingProviderFileName=" + mappingProviderFileName + ") : ret:[" + ret + "]");
        }

        return ret;
    }

    private Object getMapper(String className, Class<?> clazz) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerChainedMapperPlugin.getMapper(className=" + className + ", clazz=" + clazz.getCanonicalName() + ")");
        }
        Object mapper = null;

        try {
            Class<?> mapperClass = Class.forName(className);

            if (clazz.isAssignableFrom(mapperClass)) {
                mapper = mapperClass.newInstance();
            } else {
                LOG.error("Class " + clazz.getCanonicalName() + " is not assignable from " + mapperClass.getCanonicalName());
            }
        } catch (ClassNotFoundException exception) {
            LOG.error("Class " + className + " not found, exception=" + exception);
        } catch (IllegalAccessException exception) {
            LOG.error("Class " + className + " illegally accessed, exception=" + exception);
        } catch (InstantiationException exception) {
            LOG.error("Class " + className + " could not be instantiated, exception=" + exception);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerChainedMapperPlugin.getMapper(className=" + className + ", clazz=" + clazz.getCanonicalName() + ") : mapper:[" + mapper + "]");
        }

        return mapper;
    }

    private Set<String> buildSet(String str) {
        Set<String> ret = new HashSet<>();

        if (StringUtils.isNotBlank(str)) {
            String[] substrs = StringUtils.split(str, ", ");
            Collections.addAll(ret, substrs);
        }
        return ret;
    }

}
