/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.plugin.util;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.authorization.hadoop.config.RangerPluginConfig;
import org.apache.ranger.plugin.contextenricher.RangerServiceResourceMatcher;
import org.apache.ranger.plugin.contextenricher.RangerTagEnricher;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerServiceDef;
import org.apache.ranger.plugin.model.RangerServiceResource;
import org.apache.ranger.plugin.model.validation.RangerServiceDefHelper;
import org.apache.ranger.plugin.policyengine.RangerAccessRequest;
import org.apache.ranger.plugin.policyengine.RangerAccessResource;
import org.apache.ranger.plugin.policyengine.RangerAccessResourceImpl;
import org.apache.ranger.plugin.policyengine.RangerResourceTrie;
import org.apache.ranger.plugin.policyresourcematcher.RangerDefaultPolicyResourceMatcher;
import org.apache.ranger.plugin.policyresourcematcher.RangerPolicyResourceMatcher;
import org.apache.ranger.plugin.store.EmbeddedServiceDefsUtil;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class RangerMappingProvider {
    private static final Log LOG                           = LogFactory.getLog(RangerMappingProvider.class);
    private static final Log PERF_MAPPINGS_RETRIEVAL_LOG   = RangerPerfTracer.getPerfLogger("mappings.retrieval");
    private static final Log PERF_MAPPINGS_UPDATE_LOG      = RangerPerfTracer.getPerfLogger("mappings.update");

    private static final String TRIE_CONCURRENCY_CONTROLLER = ".use.readwrite.lock";

    private static final String MAPPING_SAVE_INTERVAL_PROP = ".mappings.save.interval";
    private static final int    DEFAULT_MAPPINGS_SAVE_INTERVAL = 1;
    private static final int    MAX_MAPPINGS_SAVE_INTERVAL     = 100;

    private final RangerServiceDef  sourceServiceDef;
    private final String            sourceServiceName;
    private final RangerServiceDef  targetServiceDef;
    private final String            targetServiceName;

    private       IndexedMappings   resourceMappings;
    private       RangerMappingRefresher mappingRefresher;

    private final int               cacheFlushInterval;
    private       int               cacheFlushCount;

    private final RangerReadWriteLock lock;

    final private RangerServiceDefHelper serviceDefHelper;

    static private class IndexedMappings {
        private final ServiceResourceMappings                                       serviceResourceMappings;
        private final Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> indexedServiceResourceMappings;
        
        IndexedMappings() {
            this(null, null);
        }

        IndexedMappings(ServiceResourceMappings serviceResourceMappings, Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> indexedServiceResourceMappings) {
            this.serviceResourceMappings        = serviceResourceMappings == null ? new ServiceResourceMappings() : serviceResourceMappings;
            this.indexedServiceResourceMappings = indexedServiceResourceMappings == null ? new HashMap<>() : indexedServiceResourceMappings;
        }

        ServiceResourceMappings getServiceResourceMappings() { return serviceResourceMappings; }

        Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> getIndexedServiceResourceMappings() { return indexedServiceResourceMappings; }
    }

    public RangerMappingProvider(String sourceServiceType, String sourceServiceName, RangerPluginConfig pluginConfig) {
        this.sourceServiceDef  = getValidatedServiceDef(sourceServiceType);
        this.sourceServiceName = sourceServiceName;
        this.targetServiceName = pluginConfig.getServiceName();
        this.targetServiceDef  = getValidatedServiceDef(pluginConfig.getServiceType());
        this.resourceMappings  = new IndexedMappings();
        boolean isUseReadWriteLock = pluginConfig.getBoolean(pluginConfig.getPropertyPrefix() + TRIE_CONCURRENCY_CONTROLLER, false);
        this.lock              = new RangerReadWriteLock(isUseReadWriteLock);
        this.cacheFlushCount   = 0;
        int configuredCacheFlushInterval    = pluginConfig.getInt(pluginConfig.getPropertyPrefix() + MAPPING_SAVE_INTERVAL_PROP, DEFAULT_MAPPINGS_SAVE_INTERVAL);
        configuredCacheFlushInterval = Math.min(configuredCacheFlushInterval, MAX_MAPPINGS_SAVE_INTERVAL);
        this.cacheFlushInterval   = Math.max(configuredCacheFlushInterval, DEFAULT_MAPPINGS_SAVE_INTERVAL);

        serviceDefHelper = new RangerServiceDefHelper(sourceServiceDef, false);

        LOG.info("Resource-Mappings are" + (isUseReadWriteLock ? " " : " not ") + "managed by locking mechanism");
    }

    public void setMappingRefresher(RangerMappingRefresher mappingRefresher) {
        this.mappingRefresher = mappingRefresher;
    }

    public List<RangerAccessResource> getMappedResources(RangerAccessRequest request) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.getMappedResources(request=" + request + ")");
        }

        final List<RangerAccessResource> ret;

        IndexedMappings currentMappings = resourceMappings;

        if (MapUtils.isNotEmpty(currentMappings.getIndexedServiceResourceMappings())) {
            // Search in serviceResourceTrie
            List<RangerServiceResourceMatcher> evaluators = getEvaluators(request.getResource(), currentMappings.getIndexedServiceResourceMappings());

            if (LOG.isDebugEnabled()) {
                LOG.debug("Found [" + (CollectionUtils.isEmpty(evaluators) ? 0 : evaluators.size()) + "] evaluators");
            }

            Set<Long> serviceResourceIds = new HashSet<>();

            if (CollectionUtils.isNotEmpty(evaluators)) {
                for (RangerServiceResourceMatcher evaluator : evaluators) {
                    RangerPolicyResourceMatcher.MatchType matchType = evaluator.getMatchType(request.getResource(), request.getContext());
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("matchType for evaluator id:[" + evaluator.getId() + "] for resource:[" + request.getResource().getAsString() + "] is [" + matchType + "]");
                    }

                    if (matchType == RangerPolicyResourceMatcher.MatchType.SELF || matchType == RangerPolicyResourceMatcher.MatchType.ANCESTOR) {

                        List<Long> tmp = currentMappings.getServiceResourceMappings().getMappings().get(evaluator.getId());

                        if (CollectionUtils.isNotEmpty(tmp)) {
                            serviceResourceIds.addAll(tmp);
                        } else {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("No service-resource found with id:[" + evaluator.getId() + "]");
                            }
                        }
                    } else {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Ignoring evaluator:[" + evaluator.getId() + "] as match-type is neither SELF nor ANCESTOR");
                        }
                    }
                }
            } else {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("No evaluators matched for resource :[" + request.getResource() + "]");
                }
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("Found serviceResourceIds=[" + serviceResourceIds + "]");
            }

            if (serviceResourceIds.size() > 1) {
                Map<Long, RangerDefaultPolicyResourceMatcher> matchers = new HashMap<>();
                Map<Long, RangerAccessResource> resources = new HashMap<>();

                for (Long serviceResourceId : serviceResourceIds) {
                    RangerServiceResource serviceResource = currentMappings.getServiceResourceMappings().getServiceResources().get(serviceResourceId);

                    if (serviceResource != null) {
                        RangerDefaultPolicyResourceMatcher matcher = new RangerDefaultPolicyResourceMatcher();
                        matcher.setPolicyResources(serviceResource.getResourceElements());
                        matcher.setServiceDef(targetServiceDef);
                        matcher.init();

                        matchers.put(serviceResourceId, matcher);

                        RangerRMSAccessResourceImpl accessResource = new RangerRMSAccessResourceImpl();

                        accessResource.setOwnerUser(serviceResource.getOwnerUser());
                        accessResource.setOwnerUser(serviceResource.getOwnerUser());
                        accessResource.setAdditionalInfo(serviceResource.getAdditionalInfo());

                        for (Map.Entry<String, RangerPolicy.RangerPolicyResource> elem : serviceResource.getResourceElements().entrySet()) {
                            accessResource.setValue(elem.getKey(), elem.getValue().getValues());
                        }

                        resources.put(serviceResourceId, accessResource);
                    } else {
                        LOG.error("No mapped service-resource in the mappings with id :[" + serviceResourceId + "]");
                    }
                }

                if (matchers.size() > 1) {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Filtering out ANCESTOR level matches");
                    }

                    Long resourceIdToDiscard = -1L;

                    for (Map.Entry<Long, RangerDefaultPolicyResourceMatcher> matcherElement : matchers.entrySet()) {

                        if (resourceIdToDiscard != -1L) {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Removing resource with id:[" + resourceIdToDiscard + "] as it is an ancestor");
                            }
                            resources.remove(resourceIdToDiscard);
                            resourceIdToDiscard = -1L;
                        }

                        for (Map.Entry<Long, RangerAccessResource> resourceElement : resources.entrySet()) {
                            if (matcherElement.getKey().equals(resourceElement.getKey())) {
                                continue;
                            }
                            RangerPolicyResourceMatcher.MatchType matchType = matcherElement.getValue().getMatchType(resourceElement.getValue(), null);
                            if (matchType == RangerPolicyResourceMatcher.MatchType.ANCESTOR) {
                                if (LOG.isDebugEnabled()) {
                                    LOG.debug("Will discard matched resourceId:[" + matcherElement.getKey() + "] as it is ancestor of resourceId:[" + resourceElement.getKey() + "]");
                                }

                                resourceIdToDiscard = matcherElement.getKey();
                                break;
                            }
                        }
                    }

                    if (resourceIdToDiscard != -1L) {
                        LOG.debug("After the loops are done.. Removing resource with id:[" + resourceIdToDiscard + "] as it is an ancestor");
                        resources.remove(resourceIdToDiscard);
                    }
                }

                ret = new ArrayList<>(resources.values());
            } else {
                ret = new ArrayList<>();

                if (serviceResourceIds.size() == 1) {
                    Long[] ids = new Long[1];

                    serviceResourceIds.toArray(ids);
                    RangerServiceResource serviceResource = currentMappings.getServiceResourceMappings().getServiceResources().get(ids[0]);

                    if (serviceResource != null) {
                        RangerRMSAccessResourceImpl accessResource = new RangerRMSAccessResourceImpl();

                        accessResource.setOwnerUser(serviceResource.getOwnerUser());
                        accessResource.setOwnerUser(serviceResource.getOwnerUser());
                        accessResource.setAdditionalInfo(serviceResource.getAdditionalInfo());

                        for (Map.Entry<String, RangerPolicy.RangerPolicyResource> elem : serviceResource.getResourceElements().entrySet()) {
                            accessResource.setValue(elem.getKey(), elem.getValue().getValues());
                        }
                        ret.add(accessResource);
                    }
                }
            }
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Cannot find serviceResourceTrie for service :[" + sourceServiceName + "]");
            }
            ret = new ArrayList<>();
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.getMappedResources(request=" + request + ")");
        }

        return ret;
    }

    public void setMappings(@Nonnull final ServiceResourceMappings mappings) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.setMappings(serviceName=" + mappings.getTargetServiceName() + ")");
        }

        if (!StringUtils.equals(targetServiceName, mappings.getTargetServiceName())) {
            LOG.error("setMappings called for incorrect target service, expected:[" + targetServiceName + "], received:[" + mappings.getTargetServiceName() + "]");
        } else {
            LOG.info("Mappings changed for target-service [" + targetServiceName + "]: prev-notification-id:[" + mappings.getPreviousNotificationId() + "], current-notification-id:[" + mappings.getCurrentNotificationId() + "]");

            RangerPerfTracer perf = null;

            if (RangerPerfTracer.isPerfTraceEnabled(PERF_MAPPINGS_UPDATE_LOG)) {
                perf = RangerPerfTracer.getPerfTracer(PERF_MAPPINGS_UPDATE_LOG, "RangerMappingProvider.setMappings(version=" + mappings.getCurrentNotificationId() + ")");
            }

            this.resourceMappings = mappings.getIsDelta() ? processMappingDeltas(mappings) : processAllMappings(mappings);

            RangerPerfTracer.logAlways(perf);

            if (mappingRefresher != null) {
                cacheFlushCount++;
                if (!mappings.getIsDelta() || (cacheFlushCount % cacheFlushInterval) == 0) {
                    mappingRefresher.saveToCache(resourceMappings.getServiceResourceMappings());
                    cacheFlushCount = 0;
                } else {
                    mappingRefresher.saveToCache(mappings);
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.setMappings(serviceName=" + mappings.getTargetServiceName() + ")");
        }
    }

    private IndexedMappings processAllMappings(final ServiceResourceMappings mappings) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.processAllMappings(serviceName=" + targetServiceName + ")");
        }
        final IndexedMappings ret;

        IndexedMappings currentMappings = this.resourceMappings;

        // Filter mappings to get only those service resource which map to other resources
        if (MapUtils.isNotEmpty(mappings.getMappings())) {
            Map<Long, RangerServiceResource> sourceServiceResources = new HashMap<>();

            Set<Long> targetServiceResourceIds = mappings.getMappings().keySet();

            for (Long targetServiceResourceId : targetServiceResourceIds) {
                RangerServiceResource targetServiceResource = mappings.getServiceResources().get(targetServiceResourceId);
                if (targetServiceResource != null) {
                    sourceServiceResources.put(targetServiceResourceId, targetServiceResource);
                }
            }

            if (MapUtils.isNotEmpty(sourceServiceResources)) {
                RangerTagEnricher.ResourceHierarchies hierarchies = new RangerTagEnricher.ResourceHierarchies();
                RangerServiceDefHelper sourceServiceDefHelper     = new RangerServiceDefHelper(sourceServiceDef, false);

                // matcher contains list of ids of other resources
                List<RangerServiceResourceMatcher> matchers = null;

                for (RangerServiceResource serviceResource : sourceServiceResources.values()) {
                    RangerServiceResourceMatcher matcher = RangerTagEnricher.createRangerServiceResourceMatcher(serviceResource, sourceServiceDefHelper, hierarchies);

                    if (matcher != null) {
                        if (matchers == null) {
                            matchers = new ArrayList<>();
                        }
                        matchers.add(matcher);
                    }
                }

                if (CollectionUtils.isNotEmpty(matchers)) {
                    Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> serviceResourceTrie = new HashMap<>();

                    for (RangerServiceDef.RangerResourceDef resourceDef : sourceServiceDef.getResources()) {
                        serviceResourceTrie.put(resourceDef.getName(), new RangerResourceTrie<>(resourceDef, matchers, false, null));
                    }

                    ret = new IndexedMappings(mappings, serviceResourceTrie);
                } else {
                    LOG.error("No valid RangerServiceResourceMatcher created for service-type=[" + sourceServiceDef.getName() + "]");
                    ret = currentMappings;
                }
            } else {
                LOG.warn("No service-resources received. Ignoring this change!!");
                ret = currentMappings;
            }
        } else {
            ret = emptyMappings();
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.processAllMappings(serviceName=" + targetServiceName + ")");
        }
        return ret;
    }

    private IndexedMappings processMappingDeltas(ServiceResourceMappings deltas) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.processMappingDeltas(serviceName=" + targetServiceName + ")");
        }
        final IndexedMappings ret;

        IndexedMappings currentMappings = this.resourceMappings;

        // For each service-resource, if signature is null, remove it from original mappings (and from trie) else add it
        if (MapUtils.isNotEmpty(deltas.getServiceResources())) {

            if (LOG.isDebugEnabled()) {
                LOG.debug("Current mappings=[" + currentMappings + "]");
            }

            if (currentMappings != null) {

                try (RangerReadWriteLock.RangerLock writeLock = lock.getWriteLock()) {
                    if (LOG.isDebugEnabled()) {
                        if (writeLock.isLockingEnabled()) {
                            LOG.debug("Acquired lock - " + writeLock);
                        }
                    }

                    Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> newTrieMap = removeUnmappedResources(deltas, currentMappings);

                    addNewResources(deltas, newTrieMap);

                    // Update currentMappings by applying deltas provided by mappings
                    ServiceResourceMappingsDeltaUtil.applyDelta(currentMappings.getServiceResourceMappings(), deltas);

                    if (LOG.isDebugEnabled()) {
                        LOG.debug("After applying deltas=[" + deltas + "], new mappings=[" + currentMappings + "]");
                    }

                    if (MapUtils.isEmpty(currentMappings.getServiceResourceMappings().getMappings())) {
                        ret = emptyMappings();
                    } else {
                        for (Map.Entry<String, RangerResourceTrie<RangerServiceResourceMatcher>> entry : newTrieMap.entrySet()) {
                            entry.getValue().wrapUpUpdate();
                        }
                        ret = new IndexedMappings(currentMappings.getServiceResourceMappings(), newTrieMap);
                    }
                }
            } else {
                LOG.error("There are no existing mappings for the service:[" + targetServiceName + "]. Cannot apply deltas!");
                ret = null;
            }
        } else if (MapUtils.isEmpty(deltas.getMappings())) {
            ret = emptyMappings();
        } else {
            LOG.error("Non-empty mappings and empty resources received! Ignoring!");
            ret = currentMappings;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.processMappingDeltas(serviceName=" + deltas.getTargetServiceName() + ")");
        }
        return ret;
    }

    private Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> removeUnmappedResources(final ServiceResourceMappings deltas, final IndexedMappings currentMappings) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.removeUnmappedResources(serviceName=" + deltas.getTargetServiceName() + ")");
        }

        final Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> ret;

        try (RangerReadWriteLock.RangerLock writeLock = lock.getWriteLock()) {
            if (writeLock.isLockingEnabled()) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Acquired lock - " + writeLock);
                }
                ret = currentMappings.getIndexedServiceResourceMappings();
            } else {
                ret = new HashMap<>();
                for (Map.Entry<String, RangerResourceTrie<RangerServiceResourceMatcher>> entry : currentMappings.getIndexedServiceResourceMappings().entrySet()) {
                    RangerResourceTrie<RangerServiceResourceMatcher> resourceTrie = new RangerResourceTrie<>(entry.getValue());
                    ret.put(entry.getKey(), resourceTrie);
                }
            }
        }

        for (Map.Entry<Long, RangerServiceResource> entry : deltas.getServiceResources().entrySet()) {
            RangerServiceResource serviceResource = entry.getValue();

            if (isSourceResourceDeleted(serviceResource)) {
                RangerServiceResource oldResource = currentMappings.getServiceResourceMappings().getServiceResources().get(serviceResource.getId());

                if (oldResource != null) {
                    RangerAccessResourceImpl accessResource = new RangerAccessResourceImpl();

                    for (Map.Entry<String, RangerPolicy.RangerPolicyResource> policyResourceEntry : oldResource.getResourceElements().entrySet()) {
                        accessResource.setValue(policyResourceEntry.getKey(), entry.getValue());
                    }
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("RangerAccessResource:[" + accessResource + "] created to represent service-resource[" + oldResource + "] to find evaluators from trie-map");
                    }

                    List<RangerServiceResourceMatcher> oldMatchers = getEvaluators(accessResource, currentMappings.getIndexedServiceResourceMappings());

                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Found [" + oldMatchers.size() + "] matchers for service-resource[" + serviceResource + "]");
                    }

                    for (RangerServiceResourceMatcher matcher : oldMatchers) {
                        for (RangerServiceDef.RangerResourceDef resourceDef : sourceServiceDef.getResources()) {
                            RangerResourceTrie<RangerServiceResourceMatcher> trie = ret.get(resourceDef.getName());

                            if (trie != null) {
                                if (LOG.isDebugEnabled()) {
                                    LOG.debug("Found trie for resourceDef:[" + resourceDef.getName() + "], deleting matcher from it");
                                }
                                trie.delete(entry.getValue().getResourceElements().get(resourceDef.getName()), matcher);
                            } else {
                                LOG.warn("Did not find trie for resourceDef:[" + resourceDef.getName() + "]. Should not have come here");
                            }
                        }
                    }
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.removeUnmappedResources(serviceName=" + deltas.getTargetServiceName() + ")");
        }

        return ret;
    }

    private void addNewResources(final ServiceResourceMappings deltas, final Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> serviceResourceTrieMap) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.addNewResources(serviceName=" + deltas.getTargetServiceName() + ")");
        }

        RangerTagEnricher.ResourceHierarchies hierarchies            = new RangerTagEnricher.ResourceHierarchies();
        RangerServiceDefHelper                sourceServiceDefHelper = new RangerServiceDefHelper(sourceServiceDef, false);

        for (Map.Entry<Long, RangerServiceResource> entry : deltas.getServiceResources().entrySet()) {
            RangerServiceResource serviceResource = entry.getValue();

            if (isSourceResourceAdded(serviceResource)) {
                RangerServiceResourceMatcher matcher = RangerTagEnricher.createRangerServiceResourceMatcher(serviceResource, sourceServiceDefHelper, hierarchies);

                if (LOG.isDebugEnabled()) {
                    LOG.debug("Created matcher. serviceResource:[" + matcher.getServiceResource() + "], policyResourceMatcher:[" + matcher.getPolicyResourceMatcher() + "]");
                }

                for (RangerServiceDef.RangerResourceDef resourceDef : sourceServiceDef.getResources()) {
                    RangerResourceTrie<RangerServiceResourceMatcher> trie = serviceResourceTrieMap.get(resourceDef.getName());

                    if (trie != null) {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Found trie for resourceDef:[" + resourceDef.getName() + "], adding matcher to it");
                        }
                        trie.add(entry.getValue().getResourceElements().get(resourceDef.getName()), matcher);
                    } else {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Did not find trie for resourceDef:[" + resourceDef.getName() + "]");
                        }
                        trie = new RangerResourceTrie<>(resourceDef, Collections.singletonList(matcher), false, null);
                        serviceResourceTrieMap.put(resourceDef.getName(), trie);
                    }
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.addNewResources(serviceName=" + deltas.getTargetServiceName() + ")");
        }
    }

    private IndexedMappings emptyMappings() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.emptyMappings(source-serviceName=" + sourceServiceName + ", target=serviceName=" + targetServiceName + ")");
        }
        ServiceResourceMappings serviceResourceMappings = new ServiceResourceMappings(ServiceResourceMappings.OP_ADD, sourceServiceName, targetServiceName, -1L, -1L, null, null, null);
        IndexedMappings ret = new IndexedMappings(serviceResourceMappings, null);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.emptyMappings(serviceName=" + sourceServiceName + ", target=serviceName=" + targetServiceName+ ")");
        }
        return ret;
    }

    private boolean isResourceDeleted(final RangerServiceResource resource) {
        return StringUtils.isEmpty(resource.getResourceSignature());
    }

    private boolean isSourceResource(final RangerServiceResource resource) {
        return StringUtils.equals(resource.getServiceName(), sourceServiceName);
    }

    private boolean isSourceResourceDeleted(final RangerServiceResource resource) {
        return isResourceDeleted(resource) && isSourceResource(resource);
    }

    private boolean isSourceResourceAdded(final RangerServiceResource resource) {
        return !isResourceDeleted(resource) && isSourceResource(resource);
    }


    @SuppressWarnings("unchecked")
    private List<RangerServiceResourceMatcher> getEvaluators(RangerAccessResource resource, Map<String, RangerResourceTrie<RangerServiceResourceMatcher>> serviceResourceTrie) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.getEvaluators(" + (resource != null ? resource.getAsString() : null) + ")");
        }

        List<RangerServiceResourceMatcher> ret = Collections.EMPTY_LIST;

        if (resource != null && MapUtils.isNotEmpty(serviceResourceTrie)) {

            RangerPerfTracer perf = null;

            if (RangerPerfTracer.isPerfTraceEnabled(PERF_MAPPINGS_RETRIEVAL_LOG)) {
                perf = RangerPerfTracer.getPerfTracer(PERF_MAPPINGS_RETRIEVAL_LOG, "RangerMappingProvider.getEvaluators(resource=" + resource.getAsString() + ")");
            }

            List<String> resourceKeys = serviceDefHelper.getOrderedResourceNames(resource.getKeys());
            Set<RangerServiceResourceMatcher> smallestList = null;

            if (CollectionUtils.isNotEmpty(resourceKeys)) {

                try (RangerReadWriteLock.RangerLock readLock = lock.getReadLock()) {
                    if (LOG.isDebugEnabled()) {
                        if (readLock.isLockingEnabled()) {
                            LOG.debug("Acquired lock - " + readLock);
                        }
                    }
                    for (String resourceName : resourceKeys) {
                        RangerResourceTrie<RangerServiceResourceMatcher> trie = serviceResourceTrie.get(resourceName);

                        if (trie == null) { // if no trie exists for this resource level, ignore and continue to next level
                            continue;
                        }

                        Set<RangerServiceResourceMatcher> serviceResourceMatchersForResource = trie.getEvaluatorsForResource(resource.getValue(resourceName));
                        Set<RangerServiceResourceMatcher> inheritedResourceMatchers = trie.getInheritedEvaluators();

                        if (smallestList != null) {
                            if (CollectionUtils.isEmpty(inheritedResourceMatchers) && CollectionUtils.isEmpty(serviceResourceMatchersForResource)) {
                                smallestList = null;
                            } else if (CollectionUtils.isEmpty(inheritedResourceMatchers)) {
                                smallestList.retainAll(serviceResourceMatchersForResource);
                            } else if (CollectionUtils.isEmpty(serviceResourceMatchersForResource)) {
                                smallestList.retainAll(inheritedResourceMatchers);
                            } else {
                                Set<RangerServiceResourceMatcher> smaller, bigger;
                                if (serviceResourceMatchersForResource.size() < inheritedResourceMatchers.size()) {
                                    smaller = serviceResourceMatchersForResource;
                                    bigger = inheritedResourceMatchers;
                                } else {
                                    smaller = inheritedResourceMatchers;
                                    bigger = serviceResourceMatchersForResource;
                                }
                                Set<RangerServiceResourceMatcher> tmp = new HashSet<>();
                                if (smallestList.size() < smaller.size()) {
                                    smallestList.stream().filter(smaller::contains).forEach(tmp::add);
                                    smallestList.stream().filter(bigger::contains).forEach(tmp::add);
                                } else {
                                    smaller.stream().filter(smallestList::contains).forEach(tmp::add);
                                    if (smallestList.size() < bigger.size()) {
                                        smallestList.stream().filter(bigger::contains).forEach(tmp::add);
                                    } else {
                                        bigger.stream().filter(smallestList::contains).forEach(tmp::add);
                                    }
                                }
                                smallestList = tmp;
                            }
                        } else {
                            if (CollectionUtils.isEmpty(inheritedResourceMatchers) || CollectionUtils.isEmpty(serviceResourceMatchersForResource)) {
                                Set<RangerServiceResourceMatcher> tmp = CollectionUtils.isEmpty(inheritedResourceMatchers) ? serviceResourceMatchersForResource : inheritedResourceMatchers;
                                smallestList = resourceKeys.size() == 1 || CollectionUtils.isEmpty(tmp) ? tmp : new HashSet<>(tmp);
                            } else {
                                smallestList = new HashSet<>(serviceResourceMatchersForResource);
                                smallestList.addAll(inheritedResourceMatchers);
                            }
                        }

                        if (CollectionUtils.isEmpty(smallestList)) {
                            smallestList = null;
                            break;
                        }
                    }
                    if (smallestList != null) {
                        ret = new ArrayList<>(smallestList);
                    }
                }
            }

            RangerPerfTracer.logAlways(perf);
        } else {
            LOG.error("Resource was null in the access-request!");
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.getEvaluators(" + (resource != null ? resource.getAsString() : null) + ") :" + Arrays.deepToString(ret.toArray()));
        }

        return ret;
    }

    private RangerServiceDef getValidatedServiceDef(String serviceType) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingProvider.getValidatedServiceDef(serviceType=" + serviceType + ")");
        }

        RangerServiceDef ret = null;

        try {
            ret = EmbeddedServiceDefsUtil.instance().getEmbeddedServiceDef(serviceType);
            if (ret == null) {
                LOG.error("There is no service-definition for specified service-type:[" + serviceType + "]");
            }
        } catch (Exception e) {
            LOG.error("Cannot get the service-definition corresponding to service-type:[" + serviceType + "]", e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingProvider.getValidatedServiceDef(serviceType=" + serviceType + ") : ret:[" + ret + "]");
        }

        return ret;
    }
}
