/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.plugin.util;

import com.sun.jersey.api.client.ClientResponse;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.authorization.hadoop.config.RangerPluginConfig;
import org.apache.ranger.plugin.service.RangerChainedMapperPlugin;

import java.security.PrivilegedAction;
import java.util.HashMap;
import java.util.Map;

public class RangerMappingRESTRetriever extends RangerMappingRetriever {
    private static final Log LOG = LogFactory.getLog(RangerMappingRESTRetriever.class);

    public static final String HTTP_URL_SEPARATOR = "/";

    public static final Object DUMMY_POST_PARAM = new Object();
    public  enum HTTP_METHOD {
        GET, POST, PUT
    }

    private static final String RMS_MAPPINGS_DOWNLOAD_REST_ENDPOINT = "/api/plugins/downloadmappings/from/_SOURCE_SERVICE_NAME_/to/_MAPPED_SERVICE_NAME_";

    private UserGroupInformation user;
    private RangerRESTClient     client;
    private String               supportsMappingDeltas;

    @Override
    public boolean init(String sourceServiceName, String targetServiceName, RangerPluginConfig pluginConfig) {

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> init(sourceServiceName:[" + sourceServiceName + "], targetServiceName:[" + targetServiceName + "], pluginConfig:[" + pluginConfig + "]");
        }

        try {
            super.init(sourceServiceName, targetServiceName, pluginConfig);

            String providerURL = pluginConfig.get(pluginConfig.getPropertyPrefix() + RangerChainedMapperPlugin.ACL_MAPPING_SOURCE_URL_PARAMETER);
            String sslConfigFileName = pluginConfig.get(pluginConfig.getPropertyPrefix() + ".policy.rest.ssl.config.file");
            int restClientConnTimeOutMs = pluginConfig.getInt(pluginConfig.getPropertyPrefix() + ".policy.rest.client.connection.timeoutMs", 120*1000);
            int restClientReadTimeOutMs = pluginConfig.getInt("ranger.rest.client.read.timeoutMs", 30 * 1000);
            supportsMappingDeltas       = pluginConfig.get("ranger.rms.mapping.rest.supports.mapping.deltas", "true");

            if (!"true".equalsIgnoreCase(supportsMappingDeltas)) {
                supportsMappingDeltas = "false";
            }

            if (StringUtils.isNotEmpty(providerURL)) {
                providerURL = providerURL.trim();
                if (StringUtils.isNotEmpty(providerURL)) {
                    if (providerURL.endsWith(HTTP_URL_SEPARATOR)) {
                        providerURL = providerURL.substring(0, providerURL.length() - 1);
                    }
                }
            }

            if (StringUtils.isNotEmpty(providerURL)) {
                client = new RangerRESTClient(providerURL, sslConfigFileName, pluginConfig);
                client.setRestClientConnTimeOutMs(restClientConnTimeOutMs);
                client.setRestClientReadTimeOutMs(restClientReadTimeOutMs);

                user = UserGroupInformation.getCurrentUser();

                return true;
            }
        } catch (Exception e) {
            LOG.error("Encountered exception: ", e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== init(sourceServiceName:[" + sourceServiceName + "], targetServiceName:[" + targetServiceName + "], pluginConfig:[" + pluginConfig + "]");
        }

        return false;
    }


    public boolean init(String providerURL, String sslConfigFileName, Configuration configuration, UserGroupInformation user) {

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> init(providerURL:[" + providerURL + "], sslConfigFileName:[" + sslConfigFileName + "], configuration:[" + configuration + "], user:[" + (user != null ? user.getShortUserName() : null) + "]");
        }

        boolean ret;
        try {
            if (StringUtils.isNotEmpty(providerURL)) {
                if (!StringUtils.isEmpty(providerURL)) {
                    providerURL = providerURL.trim();
                }

                if (providerURL.endsWith(HTTP_URL_SEPARATOR)) {
                    providerURL = providerURL.substring(0, providerURL.length() - 1);
                }

                client = new RangerRESTClient(providerURL, sslConfigFileName, configuration);
                client.setRestClientConnTimeOutMs(120 * 1000);
                client.setRestClientReadTimeOutMs(30 * 1000);

                if (user == null) {
                    this.user = UserGroupInformation.getCurrentUser();
                } else {
                    this.user = user;
                }
            }
            ret = true;
        } catch (Exception exception) {
            LOG.error("Failed with exception, URL:", exception);
            ret = false;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== init(providerURL:[" + providerURL + "], sslConfigFileName:[" + sslConfigFileName + "], configuration:[" + configuration + "], user:[" + (user != null ? user.getShortUserName() : null) + "]");
        }

        return ret;
    }

    @Override
    public ServiceResourceMappings retrieve(long lastKnownVersion, long lastActivationTimeInMillis) throws Exception {
        ServiceResourceMappings ret = null;
        if (client != null) {
            ret = getMappings(lastKnownVersion, lastActivationTimeInMillis);
        }
        return ret;
    }

    public PrivilegedAction<ClientResponse> getPrivilegedAction(final HTTP_METHOD method, final String relativeUrl, final Map<String, String> queryParams, final Object postParam) {
        // Create PrivilegedAction
        return () -> {
            ClientResponse clientResponse = null;
            try {
                switch (method) {
                    case GET:	clientResponse = client.get(relativeUrl, queryParams);
                        break;

                    case POST:	clientResponse = client.post(relativeUrl, queryParams, postParam);
                        break;

                    case PUT:	clientResponse = client.put(relativeUrl, queryParams, postParam);
                        break;

                    default:
                        LOG.error("Unsupported HTTP method : "+ method);
                        break;
                }

            } catch (Exception exception) {
                LOG.error("RangerMappingRESTRetriever.getPrivilegedAction() - Exception:[" +  exception.getMessage() + "]");
            }
            return clientResponse;
        };
    }

    public <T> ImmutablePair<ClientResponse, T> executeAction(final PrivilegedAction<ClientResponse> action, final Class<T> clazz) throws Exception {
        ClientResponse clientResponse = user.doAs(action);

        if (Thread.interrupted()) { // Clears interrupted status!
            throw new InterruptedException();
        }
        if (clientResponse != null) {
            return parseResponse(clientResponse, clazz);
        } else {
            throw new Exception("Server REST call failed with exception!");
        }
    }

    public <T> ImmutablePair<ClientResponse, T> parseResponse(final ClientResponse clientResponse, final Class<T> clazz) {
        T t = null;

        if (clientResponse.getStatus() == 200) {
            t = clientResponse.getEntity(clazz);
        } else if (clientResponse.getStatus() != 204){
            LOG.error("Request failed : clientResponse = ["+clientResponse+"]");
        }

        return ImmutablePair.of(clientResponse, t);
    }

    public ServiceResourceMappings getMappings(long lastKnownVersion, long lastActivationTimeInMillis) throws Exception {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRESTRetriever.getMappings() - Getting mappings ");
        }

        ServiceResourceMappings ret = null;

        // Prepare query parameters
        final Map<String, String> queryParams = new HashMap<>();
        queryParams.put("lastKnownVersion", String.valueOf(lastKnownVersion));
        queryParams.put("lastActivationTimeInMillis", String.valueOf(lastActivationTimeInMillis));
        queryParams.put("supportsMappingDeltas", supportsMappingDeltas);

        String relativeUrl = RMS_MAPPINGS_DOWNLOAD_REST_ENDPOINT.replaceFirst("_SOURCE_SERVICE_NAME_", getSourceServiceName());
        relativeUrl = relativeUrl.replaceFirst("_MAPPED_SERVICE_NAME_", getTargetServiceName());

        // Create PrivilegedAction
        PrivilegedAction<ClientResponse> action = getPrivilegedAction(HTTP_METHOD.GET, relativeUrl, queryParams, DUMMY_POST_PARAM);

        // Perform action and return result
        ImmutablePair<ClientResponse, ServiceResourceMappings> responsePair = executeAction(action, ServiceResourceMappings.class);

        if (responsePair.right != null) {
            ret = responsePair.right;
        }

        if(LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRESTRetriever.getMappings()");
        }
        return ret;
    }

}
