/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.plugin.util;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.plugin.service.RangerBasePlugin;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.Date;
import java.util.concurrent.BlockingQueue;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

public class RangerMappingRefresher extends Thread {
    private static final Log LOG = LogFactory.getLog(RangerMappingRefresher.class);

    private static final Log PERF_RESOURCEMAPPING_INIT_LOG = RangerPerfTracer.getPerfLogger("resourcemapping.init");

    private final RangerBasePlugin          plugIn;
    private final RangerMappingRetriever    mappingRetriever;
    private final RangerMappingProvider     mappingProvider;
    private long                            lastKnownVersion;
    private final BlockingQueue<DownloadTrigger> mappingDownloadQueue;
    private long                            lastActivationTimeInMillis;
    private boolean                         mappingsSetInPlugin;
    private final String                    serviceType;
    private final String                    serviceName;
    private final String                    cacheFileName;
    private final String                    cacheDir;
    private final Gson                      gson;

    public RangerMappingRefresher(RangerBasePlugin plugIn, RangerMappingRetriever mappingRetriever, RangerMappingProvider mappingProvider, long lastKnownVersion, BlockingQueue<DownloadTrigger> mappingDownloadQueue) {
        this.plugIn                 = plugIn;
        this.mappingRetriever       = mappingRetriever;
        this.mappingProvider        = mappingProvider;
        this.lastKnownVersion       = lastKnownVersion;
        this.mappingDownloadQueue   = mappingDownloadQueue;
        this.lastActivationTimeInMillis = -1L;
        this.serviceType            = plugIn.getServiceType();
        this.serviceName            = plugIn.getServiceName();
        String configuredCacheDir   = plugIn.getConfig().get(plugIn.getConfig().getPropertyPrefix() + ".mapping.cache.dir");
        if (StringUtils.isBlank(configuredCacheDir)) {
            //this.cacheDir = "/var/lib/ranger/rms/" + serviceType +"/mapping_cache";
            this.cacheDir = plugIn.getConfig().get(plugIn.getConfig().getPropertyPrefix() + ".policy.cache.dir");
        } else {
            this.cacheDir = configuredCacheDir;
        }

        String appId = StringUtils.isEmpty(plugIn.getAppId()) ? serviceType : plugIn.getAppId();

        Gson gson = null;
        try {
            gson = new GsonBuilder().setDateFormat("yyyyMMdd-HH:mm:ss.SSS-Z").create();
        } catch(Throwable excp) {
            LOG.fatal("PolicyRefresher(): failed to create GsonBuilder object", excp);
        }
        this.gson                   = gson;

        String cacheFilename = String.format("%s_%s_resource_mapping.json", appId, serviceName);

        cacheFilename = cacheFilename.replace(File.separatorChar,  '_');
        this.cacheFileName          = cacheFilename.replace(File.pathSeparatorChar,  '_');

        if (mappingProvider != null) {
            mappingProvider.setMappingRefresher(this);
        }

        setName("RangerMappingRefresher(serviceName=" + mappingRetriever.getTargetServiceName() + ")-" + getId());
    }

    public RangerBasePlugin getPlugin() {
        return plugIn;
    }

    public String getServiceType() {
        return serviceType;
    }

    public String getServiceName() {
        return serviceName;
    }

    @Override
    public void run() {

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher().run()");
        }

        while (true) {
            DownloadTrigger trigger = null;
            try {
                trigger = mappingDownloadQueue.take();

                RangerPerfTracer perf = null;

                if(RangerPerfTracer.isPerfTraceEnabled(PERF_RESOURCEMAPPING_INIT_LOG)) {
                    perf = RangerPerfTracer.getPerfTracer(PERF_RESOURCEMAPPING_INIT_LOG, "RangerMappingRefresher.populateMappings(serviceName=" + mappingRetriever.getTargetServiceName() + ",lastKnownVersion=" + lastKnownVersion + ")");
                }

                populateMappings();

                RangerPerfTracer.log(perf);

            } catch (InterruptedException excp) {
                LOG.info("RangerMappingRefresher(" + getName() + ").run(): Interrupted! Exiting thread ", excp);

                break;
            } finally {
                if (trigger != null) {
                    trigger.signalCompletion();
                }
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher().run()");
        }
    }

    public void populateMappings() throws InterruptedException {

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher().populateMappings()");
        }
        if (mappingProvider != null) {
            ServiceResourceMappings mappings;

            try {
                mappings = mappingRetriever.retrieve(lastKnownVersion, lastActivationTimeInMillis);
            } catch (InterruptedException ie) {
                throw ie;
            } catch (Exception e) {
                LOG.error("RangerMappingRefresher(sourceService=" + mappingRetriever.getSourceServiceName() + ", targetService=" + mappingRetriever.getTargetServiceName() + ").populateMappings(): Encountered unexpected exception. Ignoring", e);
                // TODO: Handle loading from cache here
                if (!mappingsSetInPlugin) {
                    mappings = loadFromCache();
                } else {
                    mappings = null;
                }
            }
            if (mappings != null) {
                try {
                    if (!mappingRetriever.getTargetServiceName().equals(mappings.getTargetServiceName())) {
                        LOG.error("Received incorrect mappings. service-names do not match! retriever-service-name:[" + mappingRetriever.getTargetServiceName() + ", targetServiceName in mappings:[" + mappings.getTargetServiceName() + "]");
                        LOG.error("Will not call setMappings()");
                    } else {
                        mappings.setPreviousNotificationId(lastKnownVersion); // Assume that server is providing correct mappings. This helps with logging on plugin side
                        mappingProvider.setMappings(mappings);
                        mappingsSetInPlugin = true;
                        lastKnownVersion = mappings.getCurrentNotificationId();
                        lastActivationTimeInMillis = new Date().getTime();
                    }
                } catch (Exception e) {
                    LOG.error("RangerMappingRefresher(sourceService=" + mappingRetriever.getSourceServiceName() + ", targetService=" + mappingRetriever.getTargetServiceName() + ").populateMappings(): Encountered unexpected exception. Ignoring", e);
                }
            } else {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("No change in mappings!");
                }
            }

        } else {
            LOG.error("RangerMappingRefresher(sourceService=" + mappingRetriever.getSourceServiceName() + ", targetService=" + mappingRetriever.getTargetServiceName() + ").populateMappings(): No receiver to update mapping cache");
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher().populateMappings()");
        }
    }

    final public void startRefresher() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher().startRefresher()");
        }
        try {
            super.start();
        } catch (Exception excp) {
            LOG.error("RangerMappingRefresher(" + getName() + ").startRetriever(): Failed to start", excp);
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher().startRefresher()");
        }
    }

    private void stopRefresher() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher().stopRefresher()");
        }
        if (super.isAlive()) {
            super.interrupt();

            boolean setInterrupted = false;
            boolean isJoined = false;

            while (!isJoined) {
                try {
                    super.join();
                    isJoined = true;
                } catch (InterruptedException excp) {
                    LOG.warn("RangerMappingRefresher(" + getName() + ").stopRefresher(): Error while waiting for thread to exit", excp);
                    LOG.warn("Retrying Thread.join(), interrupted flag will be set after Thread.join() succeeds");
                    setInterrupted = true;
                }
            }
            if (setInterrupted) {
                Thread.currentThread().interrupt();
            }
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher().stopRefresher()");
        }
    }

    public void cleanup() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher.cleanup()");
        }

        stopRefresher();

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher.cleanup()");
        }
    }

    private ServiceResourceMappings loadFromCache() {

        ServiceResourceMappings ret = null;

        if(LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher(serviceName=" + serviceName + ").loadFromCache()");
        }

        File cacheFile = cacheDir == null ? null : new File(cacheDir + File.separator + cacheFileName);

        if(cacheFile != null && cacheFile.isFile() && cacheFile.canRead()) {
            Reader reader = null;

            RangerPerfTracer perf = null;

            if(RangerPerfTracer.isPerfTraceEnabled(PERF_RESOURCEMAPPING_INIT_LOG)) {
                perf = RangerPerfTracer.getPerfTracer(PERF_RESOURCEMAPPING_INIT_LOG, "RangerMappingRefresher.loadFromCache(serviceName=" + serviceName + ")");
            }

            try {
                reader = new FileReader(cacheFile);

                ServiceResourceMappings mappings = gson.fromJson(reader, ServiceResourceMappings.class);

                if(mappings != null) {
                    if(!StringUtils.equals(serviceName, mappings.getTargetServiceName())) {
                        LOG.warn("ignoring unexpected serviceName '" + mappings.getTargetServiceName() + "' in cache file '" + cacheFile.getAbsolutePath() + "'");

                        mappings.setTargetServiceName(serviceName);
                    }

                    lastKnownVersion = mappings.getCurrentNotificationId() == null ? -1 : mappings.getCurrentNotificationId();

                    // Load deltas if any
                    ret = loadAndApplyDeltas(mappings);
                }
            } catch (Exception excp) {
                LOG.error("failed to load mappings from cache file " + cacheFile.getAbsolutePath(), excp);
            } finally {
                RangerPerfTracer.logAlways(perf);

                if(reader != null) {
                    try {
                        reader.close();
                    } catch(Exception excp) {
                        LOG.error("error while closing opened cache file " + cacheFile.getAbsolutePath(), excp);
                    }
                }
            }
        } else {
            LOG.warn("cache file does not exist or not readable '" + (cacheFile == null ? null : cacheFile.getAbsolutePath()) + "'");
        }

        if(LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher(serviceName=" + serviceName + ").loadFromCache()");
        }

        return ret;
    }

    private ServiceResourceMappings loadAndApplyDeltas(ServiceResourceMappings mappings) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher(serviceName=" + serviceName + ").loadAndApplyDeltas()");
        }

        String deltaCacheDirName = cacheDir + File.separator + "deltas";
        File   deltaCacheDir     = new File(deltaCacheDirName);

        if (deltaCacheDir.exists() && deltaCacheDir.isDirectory() && deltaCacheDir.canRead()) {
            RangerPerfTracer perf = null;

            if(RangerPerfTracer.isPerfTraceEnabled(PERF_RESOURCEMAPPING_INIT_LOG)) {
                perf = RangerPerfTracer.getPerfTracer(PERF_RESOURCEMAPPING_INIT_LOG, "RangerMappingRefresher.loadAndApplyDeltas(serviceName=" + serviceName + ")");
            }

            File deltaFile;

            do {
                String deltaFileName = deltaCacheDirName + File.separator + cacheFileName + "_" + mappings.getCurrentNotificationId();
                deltaFile            = new File(deltaFileName);
                Reader reader        = null;

                if (deltaFile.exists() && deltaFile.canRead()) {

                    try {
                        reader = new FileReader(deltaFile);
                        ServiceResourceMappings deltas = gson.fromJson(reader, ServiceResourceMappings.class);

                        if(deltas != null) {
                            if(!StringUtils.equals(serviceName, deltas.getTargetServiceName())) {
                                LOG.warn("ignoring unexpected serviceName '" + deltas.getTargetServiceName() + "' in cache file '" + deltaFile.getAbsolutePath() + "'");

                                deltas.setTargetServiceName(serviceName);
                            }
                            ServiceResourceMappingsDeltaUtil.applyDelta(mappings, deltas);
                        }
                    } catch (Exception excp) {
                        LOG.error("failed to load mappings from cache file " + deltaFile.getAbsolutePath(), excp);
                        deltaFile = null;
                    }
                } else {
                    LOG.info("File [" + deltaFile.getAbsolutePath() + "] does not exist or is not readable");
                    deltaFile = null;
                }

                if(reader != null) {
                    try {
                        reader.close();
                    } catch(Exception e) {
                        LOG.error("error while closing opened cache file " + deltaFileName, e);
                    }
                }

            } while (deltaFile != null);

            RangerPerfTracer.logAlways(perf);

        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher(serviceName=" + serviceName + ").loadAndApplyDeltas()");
        }

        return mappings;
    }

    public void saveToCache(ServiceResourceMappings mappings) {
        if(LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher(serviceName=" + serviceName + ").saveToCache(isDelta=" + (mappings != null && mappings.getIsDelta())  + ")");
        }

        if(mappings != null) {
            File cacheFile = null;

            if (cacheDir != null) {
                String realCacheDirName = mappings.getIsDelta() ? cacheDir + File.separator + "deltas" : cacheDir;
                String realCacheFileName = mappings.getIsDelta() ? cacheFileName + "_" + mappings.getPreviousNotificationId() : cacheFileName;

                try {
                    File cacheDirTmp = new File(realCacheDirName);

                    if (!cacheDirTmp.exists()) {
                        cacheDirTmp.mkdirs();
                    }
                    cacheFile = new File(realCacheDirName + File.separator + realCacheFileName);

                } catch (SecurityException ex) {
                    LOG.error("Cannot create cache directory:[" + realCacheDirName + "]", ex);
                }
            }

            if (cacheFile != null) {

                RangerPerfTracer perf = null;
                if (RangerPerfTracer.isPerfTraceEnabled(PERF_RESOURCEMAPPING_INIT_LOG)) {
                    perf = RangerPerfTracer.getPerfTracer(PERF_RESOURCEMAPPING_INIT_LOG, "RangerMappingRefresher.saveToCache(serviceName=" + serviceName + ")");
                }

                Writer writer = null;
                try {
                    writer = new FileWriter(cacheFile);
                    gson.toJson(mappings, writer);
                } catch (Exception excp) {
                    LOG.error("failed to save mappings to cache file '" + cacheFile.getAbsolutePath() + "'", excp);
                } finally {
                    if (writer != null) {
                        try {
                            writer.close();
                        } catch (Exception excp) {
                            LOG.error("error while closing opened cache file '" + cacheFile.getAbsolutePath() + "'", excp);
                        }
                    }
                }

                if (!mappings.getIsDelta()) {
                    removeDeltaFiles();
                }

                RangerPerfTracer.logAlways(perf);
            }

        } else {
            LOG.info("mappings is null or mappings are loaded from cache. No need to save mappings to cache");
        }

        if(LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher(serviceName=" + serviceName + ").saveToCache(isDeltas=" + (mappings != null && mappings.getIsDelta())  + ")");
        }
    }

    private void removeDeltaFiles() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> RangerMappingRefresher(serviceName=" + serviceName + ").removeDeltaFiles()");
        }

        String deltaCacheDirName = cacheDir + File.separator + "deltas";
        String deltaCacheFileNamePrefix = cacheFileName + "_";

        try {
            File deltaCacheDir = new File(deltaCacheDirName);

            if (deltaCacheDir.exists() && deltaCacheDir.isDirectory() && deltaCacheDir.canRead() && deltaCacheDir.canExecute() && deltaCacheDir.canWrite()) {
                String[] fileNames = deltaCacheDir.list();
                if (fileNames != null) {
                    for (String fileName : fileNames) {
                        if (StringUtils.startsWith(fileName, deltaCacheFileNamePrefix)) {
                            File deltaFile = new File(deltaCacheDirName + File.separator + fileName);
                            if (!deltaFile.delete()) {
                                LOG.error("Could not delete [" + fileName + "]");
                            }
                        }
                    }
                }
            }
        } catch (SecurityException ex) {
            LOG.error("Cannot remove delta files from directory:[" + deltaCacheDirName + "]", ex);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== RangerMappingRefresher(serviceName=" + serviceName + ").removeDeltaFiles()");
        }
    }
}
