/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.plugin.util;

import org.apache.ranger.plugin.model.RangerServiceResource;
import org.codehaus.jackson.annotate.JsonAutoDetect;
import org.codehaus.jackson.annotate.JsonIgnoreProperties;
import org.codehaus.jackson.map.annotate.JsonSerialize;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlRootElement;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@JsonAutoDetect(fieldVisibility= JsonAutoDetect.Visibility.ANY)
@JsonSerialize(include=JsonSerialize.Inclusion.NON_NULL)
@JsonIgnoreProperties(ignoreUnknown=true)
@XmlRootElement
@XmlAccessorType(XmlAccessType.FIELD)
public class ServiceResourceMappings implements java.io.Serializable {
    public static final String OP_ADD     = "add";
    public static final String OP_DELETE  = "delete";
    public static final String OP_UPDATE  = "update";

    private String                      op = OP_ADD;
    private String                      sourceServiceName;
    private String                      targetServiceName;
    private Long                        currentNotificationId;
    private Long                        previousNotificationId;
    private Date                        mappingUpdateTime;
    private Map<Long, RangerServiceResource> serviceResources;
    private Map<Long, List<Long>>       mappings;
    private Boolean                     isDelta;

    public ServiceResourceMappings() {
        this(null, null, null, null, null, null, null, null);
    }

    public ServiceResourceMappings(String op, String sourceServiceName, String targetServiceName, Long currentNotificationId, Long previousNotificationId,
                                   Date mappingUpdateTime, Map<Long, RangerServiceResource> serviceResources, Map<Long, List<Long>> mappings) {
        this(op, sourceServiceName, targetServiceName, currentNotificationId, previousNotificationId, mappingUpdateTime, serviceResources, mappings, false);
    }

    public ServiceResourceMappings(String op, String sourceServiceName, String targetServiceName, Long currentNotificationId, Long previousNotificationId,
                                   Date mappingUpdateTime, Map<Long, RangerServiceResource> serviceResources, Map<Long, List<Long>> mappings, Boolean isDelta) {

        setOp(op);
        setSourceServiceName(sourceServiceName);
        setTargetServiceName(targetServiceName);
        setCurrentNotificationId(currentNotificationId);
        setPreviousNotificationId(previousNotificationId);
        setMappingUpdateTime(mappingUpdateTime);
        setServiceResources(serviceResources);
        setMappings(mappings);
        setIsDelta(isDelta);
    }

    public void setOp(String op) {
        this.op = op;
    }

    public void setSourceServiceName(String sourceServiceName) {
        this.sourceServiceName = sourceServiceName;
    }

    public void setTargetServiceName(String targetServiceName) {
        this.targetServiceName = targetServiceName;
    }

    public void setCurrentNotificationId(Long currentNotificationId) {
        this.currentNotificationId = currentNotificationId;
    }

    public void setPreviousNotificationId(Long previousNotificationId) {
        this.previousNotificationId = previousNotificationId;
    }

    public void setMappingUpdateTime(Date mappingUpdateTime) {
        this.mappingUpdateTime = mappingUpdateTime == null ? new Date() : mappingUpdateTime;
    }

    public void setServiceResources(Map<Long, RangerServiceResource> serviceResources) {
        this.serviceResources = serviceResources != null ? serviceResources : new HashMap<>();
    }

    public void setMappings(Map<Long, List<Long>> mappings) {
        this.mappings = mappings != null ? mappings : new HashMap<>();
    }

    public void setIsDelta(Boolean isDelta) {
        this.isDelta = isDelta;
    }

    public String getOp() {
        return op;
    }

    public String getSourceServiceName() {
        return sourceServiceName;
    }

    public String getTargetServiceName() {
        return targetServiceName;
    }

    public Long getCurrentNotificationId() {
        return currentNotificationId;
    }

    public Long getPreviousNotificationId() {
        return previousNotificationId;
    }

    public Date getMappingUpdateTime() {
        return mappingUpdateTime;
    }

    public Map<Long, RangerServiceResource> getServiceResources() {
        return serviceResources;
    }

    public Map<Long, List<Long>> getMappings() {
        return mappings;
    }

    public boolean getIsDelta() { return isDelta == null ? Boolean.FALSE : isDelta; }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();

        toString(sb);

        return sb.toString();
    }

    public StringBuilder toString(StringBuilder sb) {
        sb.append("ServiceResourceMappings={")
                .append("op=").append(op).append(", ")
                .append("sourceServiceName=").append(sourceServiceName).append(", ")
                .append("targetServiceName=").append(targetServiceName).append(", ")
                .append("currentNotificationId=").append(currentNotificationId).append(", ")
                .append("previousNotificationId=").append(previousNotificationId).append(", ")
                .append("mappingUpdateTime={").append(mappingUpdateTime).append("}, ")
                .append("isDelta={").append(isDelta).append("}")
                .append(printableServiceResources())
                .append(printableMappings())
                .append("}");

        return sb;
    }
    private StringBuilder printableServiceResources() {
        StringBuilder sb = new StringBuilder();

        if (serviceResources != null) {
            sb.append(", serviceResources= {");
            for (Map.Entry<Long, RangerServiceResource> entry : serviceResources.entrySet()) {
                sb.append("{").append(entry.getKey()).append(",").append(entry.getValue()).append("}");
            }
            sb.append("}");
        }
        return sb;
    }

    private StringBuilder printableMappings() {
        StringBuilder sb = new StringBuilder();
        if (mappings != null) {
            sb.append(", mappings= {");
            for (Map.Entry<Long, List<Long>> entry : mappings.entrySet()) {
                sb.append("{").append(entry.getKey()).append(",").append(Arrays.toString(entry.getValue().toArray())).append("}");
            }
            sb.append("}");
        }
        return sb;
    }

}
