/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.plugin.util;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.plugin.model.RangerServiceResource;

import java.util.List;
import java.util.Map;

public class ServiceResourceMappingsDeltaUtil {

    private static final Log LOG = LogFactory.getLog(ServiceResourceMappingsDeltaUtil.class);

    private static final Log PERF_RESOURCE_MAPPINGS_DELTA_LOG = RangerPerfTracer.getPerfLogger("serviceresourcemappings.delta");

    /*
    It should be possible to call applyDelta() multiple times with ServiceResourceMappings and delta resulting from previous call to applyDelta()
    The end result should be same if called once or multiple times.

    This results in currentServiceResourceMappings being changed in place. deltas remain unchanged.
     */
    static public void applyDelta(ServiceResourceMappings currentServiceResourceMappings, ServiceResourceMappings deltas) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("==> ServiceResourceMappingsDeltaUtil.applyDelta(serviceResourceMappings:[" + currentServiceResourceMappings + "], delta:[" + deltas + "]");
        }
        RangerPerfTracer perf = null;

        if(RangerPerfTracer.isPerfTraceEnabled(PERF_RESOURCE_MAPPINGS_DELTA_LOG)) {
            perf = RangerPerfTracer.getPerfTracer(PERF_RESOURCE_MAPPINGS_DELTA_LOG, "ServiceResourceMappingsDeltaUtil.applyDelta()");
        }
        if (currentServiceResourceMappings != null && !currentServiceResourceMappings.getIsDelta() && deltas != null && deltas.getIsDelta()) {
            currentServiceResourceMappings.setSourceServiceName(deltas.getSourceServiceName());
            currentServiceResourceMappings.setTargetServiceName(deltas.getTargetServiceName());
            currentServiceResourceMappings.setPreviousNotificationId(deltas.getPreviousNotificationId());
            currentServiceResourceMappings.setCurrentNotificationId(deltas.getCurrentNotificationId());
            currentServiceResourceMappings.setMappingUpdateTime(deltas.getMappingUpdateTime());

            // merge
            Map<Long, RangerServiceResource> currentServiceResources = currentServiceResourceMappings.getServiceResources();
            Map<Long, List<Long>>            currentMappings         = currentServiceResourceMappings.getMappings();

            for (Map.Entry<Long, RangerServiceResource> entry : deltas.getServiceResources().entrySet()) {

                RangerServiceResource deltaServiceResource = entry.getValue();

                if (StringUtils.isEmpty(deltaServiceResource.getResourceSignature())) {

                    // This resource needs to be deleted
                    currentServiceResources.remove(deltaServiceResource.getId());
                    currentMappings.remove(deltaServiceResource.getId());

                } else {
                    currentServiceResources.put(deltaServiceResource.getId(), deltaServiceResource);
                }
            }

            for (Map.Entry<Long, List<Long>> entry : deltas.getMappings().entrySet()) {
                List<Long> value = entry.getValue();
                Long key = entry.getKey();

                List<Long> existing = currentMappings.get(key);
                if (CollectionUtils.isNotEmpty(existing)) {
                    existing.addAll(value);
                } else {
                    existing = value;
                }
                currentMappings.put(key, existing);
            }

        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Cannot apply deltas to service-tags as one of preconditions is violated. Returning received ServiceResourceMappings without applying delta!!");
            }
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== ServiceResourceMappingsDeltaUtil.applyDelta() : " + currentServiceResourceMappings);
        }

        RangerPerfTracer.logAlways(perf);

    }

    public static void pruneUnusedAttributes(ServiceResourceMappings serviceResourceMappings) {
        if (serviceResourceMappings != null) {
            serviceResourceMappings.setMappingUpdateTime(null);

            for (Map.Entry<Long, RangerServiceResource>  entry : serviceResourceMappings.getServiceResources().entrySet()) {
                RangerServiceResource serviceResource = entry.getValue();

                serviceResource.setCreatedBy(null);
                serviceResource.setCreateTime(null);
                serviceResource.setUpdatedBy(null);
                serviceResource.setUpdateTime(null);
                serviceResource.setGuid(null);

                serviceResource.setServiceName(null);
            }
        }
    }
}

