# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License. See accompanying LICENSE file.

import logging
import os
import shlex
import subprocess
import sys

from optparse import OptionParser
from xml.etree import ElementTree as ET
from xml.parsers.expat import ExpatError

def main():

  parser = OptionParser()

  parser.add_option("-d", "--database", dest="db_type", help="Ranger RMS database type.")
  parser.add_option("-p", "--password", dest="db_password", help="Ranger RMS database user password.")
  parser.add_option("-f", "--file", dest="rms_config_file", help="Ranger RMS configuration file.")

  (options, args) = parser.parse_args()

  if options.db_type is None:
    print("ERROR: Option -d not provided to specify database type")
    sys.exit(1)
  else:
    db_type = options.db_type

  if options.db_password is None:
    print("ERROR: Option -p not provided to specify database user password")
    sys.exit(1)
  else:
    db_password = options.db_password

  if options.rms_config_file is None:
    print("ERROR: Option -f not provided to specify ranger-rms-site.xml config file")
    sys.exit(1)
  else:
    rms_config_file = options.rms_config_file

  if not (os.path.isfile(rms_config_file)):
    print("ERROR: Config file " + str(rms_config_file) + " does not exists")
    sys.exit(1)

  FORMAT = '%(asctime)-15s %(message)s'
  logging.basicConfig(format=FORMAT, level=logging.DEBUG)

  JAVA_HOME = os.getenv('JAVA_HOME')
  if JAVA_HOME is None:
    print("ERROR: JAVA_HOME environment property was not defined, exit.")
    sys.exit(1)

  SQL_CONNECTOR_JAR = os.getenv('SQL_CONNECTOR_JAR')
  if SQL_CONNECTOR_JAR is None:
    print("ERROR: SQL_CONNECTOR_JAR environment property was not defined, exit.")
    sys.exit(1)

  RANGER_RMS_HOME = os.getenv('RANGER_RMS_HOME')
  if RANGER_RMS_HOME is None:
    print("ERROR: RANGER_RMS_HOME environment property was not defined, exit.")
    sys.exit(1)

  db_user = get_config_xml(rms_config_file, "ranger-rms.jpa.jdbc.user")
  db_jdbc_url = get_config_xml(rms_config_file, "ranger-rms.jpa.jdbc.url")
  db_ssl_enabled = get_config_xml(rms_config_file, "ranger-rms.db.ssl.enabled")
  db_ssl_required = get_config_xml(rms_config_file, "ranger-rms.db.ssl.required")
  db_ssl_verifyServerCertificate = get_config_xml(rms_config_file, "ranger-rms.db.ssl.verifyServerCertificate")
  db_ssl_auth_type = get_config_xml(rms_config_file, "ranger-rms.db.ssl.auth.type")
  db_ssl_certificate_file = get_config_xml(rms_config_file, "ranger-rms.db.ssl.certificateFile")
  javax_net_ssl_trustStore = get_config_xml(rms_config_file, "ranger-rms.truststore.file")
  javax_net_ssl_keyStore = get_config_xml(rms_config_file, "ranger-rms.keystore.file")

  javax_net_ssl_trustStorePassword = os.getenv('RANGER_RMS_TRUSTSTORE_PASSWORD')
  if db_ssl_enabled == "true" and db_ssl_auth_type == "1-way" and javax_net_ssl_trustStorePassword is None:
    print("ERROR: RANGER_RMS_TRUSTSTORE_PASSWORD environment property was not defined, exit.")
    sys.exit(1)

  javax_net_ssl_keyStorePassword = os.getenv('RANGER_RMS_DB_KEYSTORE_PASSWORD')
  if db_ssl_enabled == "true" and db_ssl_auth_type == "2-way" and javax_net_ssl_keyStorePassword is None:
    print("ERROR: RANGER_RMS_TRUSTSTORE_PASSWORD environment property was not defined, exit.")
    sys.exit(1)

  db_ssl_param = ""
  db_ssl_cert_param = ""

  if db_ssl_enabled == "true":
    if db_type.lower() == "mysql":
      db_ssl_param = "?useSSL={0}&requireSSL={1}&verifyServerCertificate={2}".format(db_ssl_enabled, db_ssl_required, db_ssl_verifyServerCertificate)
      if db_ssl_verifyServerCertificate == "true":
        if db_ssl_auth_type == "1-way":
          db_ssl_cert_param = " -Djavax.net.ssl.trustStore={0} -Djavax.net.ssl.trustStorePassword={1} ".format(javax_net_ssl_trustStore, javax_net_ssl_trustStorePassword)
        else:
          db_ssl_cert_param = " -Djavax.net.ssl.keyStore={0} -Djavax.net.ssl.keyStorePassword={1} -Djavax.net.ssl.trustStore={2} -Djavax.net.ssl.trustStorePassword={3} ".format(javax_net_ssl_keyStore, javax_net_ssl_keyStorePassword, javax_net_ssl_trustStore, javax_net_ssl_trustStorePassword)
    elif db_type.lower() == "postgresql":
      if db_ssl_certificate_file is not None:
        db_ssl_param = "?ssl={0}&sslmode=verify-full&sslrootcert={1}".format(db_ssl_enabled, db_ssl_certificate_file)
      elif db_ssl_verifyServerCertificate == "true" or db_ssl_required == "true":
        db_ssl_param = "?ssl={0}&sslmode=verify-full&sslfactory=org.postgresql.ssl.DefaultJavaSSLFactory".format(db_ssl_enabled)
        if db_ssl_auth_type == "1-way":
          db_ssl_cert_param = " -Djavax.net.ssl.trustStore={0} -Djavax.net.ssl.trustStorePassword={1} ".format(javax_net_ssl_trustStore, javax_net_ssl_trustStorePassword)
        else:
          db_ssl_cert_param = " -Djavax.net.ssl.keyStore={0} -Djavax.net.ssl.keyStorePassword={1} -Djavax.net.ssl.trustStore={2} -Djavax.net.ssl.trustStorePassword={3} ".format(javax_net_ssl_keyStore, javax_net_ssl_keyStorePassword, javax_net_ssl_trustStore, javax_net_ssl_trustStorePassword)
      else:
        db_ssl_param = "?ssl={0}".format(db_ssl_enabled)
    else:
      print("ERROR: Unsupported database type, exit")
      sys.exit(1)

  if db_type.lower() == "mysql":
    db_sql_file = "{0}/db/mysql/ranger_rms_truncate_db_mysql.sql".format(RANGER_RMS_HOME)
    jisql_cmd = "{0}/bin/java {9} -cp {1}:{2}/jisql/lib/* org.apache.util.sql.Jisql -driver {3} -cstring {4}{8} -u '{5}' -p '{6}' -noheader -trim -c \; -input {7}".format(JAVA_HOME, SQL_CONNECTOR_JAR, RANGER_RMS_HOME, "mysqlconj", db_jdbc_url, db_user, db_password, db_sql_file, db_ssl_param, db_ssl_cert_param)
  elif db_type.lower() == "postgresql":
    db_sql_file = "{0}/db/postgres/ranger_rms_truncate_db_postgres.sql".format(RANGER_RMS_HOME)
    jisql_cmd = "{0}/bin/java {9} -cp {1}:{2}/jisql/lib/* org.apache.util.sql.Jisql -driver {3} -cstring {4}{8} -u '{5}' -p '{6}' -noheader -trim -c \; -input {7}".format(JAVA_HOME, SQL_CONNECTOR_JAR, RANGER_RMS_HOME, "postgresql", db_jdbc_url, db_user, db_password, db_sql_file, db_ssl_param, db_ssl_cert_param)
  elif db_type.lower() == "oracle":
    db_sql_file = "{0}/db/oracle/ranger_rms_truncate_db_oracle.sql".format(RANGER_RMS_HOME)
    jisql_cmd = "{0}/bin/java -cp {1}:{2}/jisql/lib/* org.apache.util.sql.Jisql -driver {3} -cstring {4} -u '{5}' -p '{6}' -noheader -trim -c \; -input {7}".format(JAVA_HOME, SQL_CONNECTOR_JAR, RANGER_RMS_HOME, "oraclethin", db_jdbc_url, db_user, db_password, db_sql_file)
  else:
    print("ERROR: Unsupported database type, exit")
    sys.exit(1)

  log("[I] Executing truncate operation on Ranger RMS tables", "info")
  jisql_log(jisql_cmd, db_password)
  jisql_cmd_output = subprocess.call(shlex.split(jisql_cmd))

  if jisql_cmd_output == 0:
    log("[I] Truncate operation executed successfully on Ranger RMS tables", "info")
  else:
    log("[E] Truncate operation failed on Ranger RMS tables", "error")
    sys.exit(1)

def log(msg, type):
  if type == "info":
    logging.info("{0}".format(msg))
  if type == "debug":
    logging.debug("{0}".format(msg))
  if type == "warning":
    logging.warning("{0}".format(msg))
  if type == "exception":
    logging.exception("{0}".format(msg))
  if type == "error":
    logging.error("{0}".format(msg))

def get_config_xml(config_file, config_name):
  value = ""
  try:
    tree = ET.parse(config_file)
  except ExpatError:
    print("ERROR: While parsing file configuration file : " + config_file)
    sys.exit(1)
  root = tree.getroot()
  for child in root.findall('property'):
    name = child.find("name").text.strip()
    if name == config_name:
      value = child.find("value").text
      break
  return value

def jisql_log(jisql_cmd, db_password):
  jisql_cmd = jisql_cmd.replace(" -p '{0}'".format(db_password), " -p '********'")
  log("[JISQL] {0}".format(jisql_cmd), "info")

if __name__ == '__main__':
  main()