/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.common;

import org.apache.commons.collections.CollectionUtils;
import org.apache.ranger.entity.XXService;
import org.apache.ranger.plugin.model.RangerServiceResource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ranger.plugin.util.ServiceResourceMappings;
import org.apache.ranger.plugin.util.ServiceResourceMappingsDeltaUtil;
import org.apache.ranger.rms.biz.ResourceMgr;
import org.apache.ranger.rms.server.RMSConfig;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;

@Service
public class ServiceResourceMappingsCache {
    private static final Log LOG = LogFactory.getLog(ServiceResourceMappingsCache.class);

    private static final int MAX_WAIT_TIME_FOR_UPDATE = 10;

    @Autowired
    ResourceMgr resourceMgr;

    private       int                                              waitTimeInSeconds;
    private final Map<ServicePair, ServiceResourceMappingsWrapper> ServiceResourceMappingsMap = new HashMap<>();

    static class ServicePair {
        XXService llService;
        XXService hlService;

        ServicePair(XXService llService, XXService hlService) {
            this.llService = llService;
            this.hlService = hlService;
        }
        @Override
        public boolean equals(Object other) {
            if (other == null) return false;
            if (this == other) return true;
            if (!other.getClass().equals(this.getClass())) return false;
            ServicePair otherPair = (ServicePair)other;
            return this.llService.getId().equals(otherPair.llService.getId()) && this.hlService.getId().equals(otherPair.hlService.getId());
        }
        @Override
        public int hashCode() {
            return llService.getId().hashCode() + hlService.getId().hashCode();
        }

        @Override
        public String toString() {
            return "{llService=" + llService.getName() + ", hlService=" + hlService.getName() + "}";
        }
    }

    @PostConstruct
    public void init() {
        RMSConfig config = RMSConfig.getInstance();

        waitTimeInSeconds = config.getInt("ranger-rms.mappings.download.cache.max.waittime.for.update", MAX_WAIT_TIME_FOR_UPDATE);
    }

    public void dump() {
        final Set<ServicePair> servicePairs;

        synchronized (this) {
            servicePairs = ServiceResourceMappingsMap.keySet();
        }

        if (CollectionUtils.isNotEmpty(servicePairs)) {

            for (ServicePair servicePair : servicePairs) {
                final ServiceResourceMappingsWrapper cachedServiceResourceMappingsWrapper;

                synchronized (this) {
                    cachedServiceResourceMappingsWrapper = ServiceResourceMappingsMap.get(servicePair);
                }
                LOG.debug("servicePair:" + servicePair + ", Cached-MetaData:" + cachedServiceResourceMappingsWrapper);

            }
        }
    }

    public ServiceResourceMappings getServiceResourceMappings(XXService llService, XXService hlService, Long lastKnownVersion, boolean needsBackwardCompatibility) throws Exception {

        if (LOG.isDebugEnabled()) {
            LOG.debug("==> ServiceResourceMappingsCache.getServiceResourceMappings(" + llService.getName() + ", " + hlService.getName() + ", " + lastKnownVersion + ", " + needsBackwardCompatibility + ")");
        }

        ServiceResourceMappings        ret;

        ServicePair                    servicePair = new ServicePair(llService, hlService);
        ServiceResourceMappingsWrapper serviceResourceMappingsWrapper;

        synchronized (this) {
            serviceResourceMappingsWrapper = ServiceResourceMappingsMap.get(servicePair);

            if (serviceResourceMappingsWrapper == null) {
                serviceResourceMappingsWrapper = new ServiceResourceMappingsWrapper(servicePair);
                ServiceResourceMappingsMap.put(servicePair, serviceResourceMappingsWrapper);
            }
        }

        ret = serviceResourceMappingsWrapper.getLatestOrCached(servicePair, lastKnownVersion, needsBackwardCompatibility);

        if (LOG.isDebugEnabled()) {
            LOG.debug("<== ServiceResourceMappingsCache.getServiceResourceMappings(" + llService.getName() + ", " + hlService.getName() + ", " + lastKnownVersion + ", " + needsBackwardCompatibility + "): count=" + ((ret == null || ret.getMappings() == null) ? 0 : ret.getMappings().size()));
        }

        return ret;
    }

    static class MappingDeltasFromVersion {
        Long                    fromVersion;
        ServiceResourceMappings deltas;

        MappingDeltasFromVersion(Long fromVersion, ServiceResourceMappings deltas) {
            this.fromVersion = fromVersion;
            this.deltas      = deltas;
        }

        Long getFromVersion() { return fromVersion; }
        ServiceResourceMappings getDeltas() { return deltas; }

    }

    class ServiceResourceMappingsWrapper {
        final ServiceResourceMappingsCache.ServicePair servicePair;
        ServiceResourceMappings                        serviceResourceMappings;
        final ReentrantLock                            lock                      = new ReentrantLock();
        Map<Long, MappingDeltasFromVersion>            deltaCache                = new HashMap<>();

        ServiceResourceMappingsWrapper(ServiceResourceMappingsCache.ServicePair servicePair) {
            this.servicePair = servicePair;
            serviceResourceMappings = null;
        }

        ServiceResourceMappings getServiceResourceMappings() {
            return serviceResourceMappings;
        }

        ServiceResourceMappings getLatestOrCached(ServicePair servicePair, Long lastKnownVersion, boolean needsBackwardCompatibility) throws Exception {
            if (LOG.isDebugEnabled()) {
                LOG.debug("==> ServiceResourceMappingsCache.getLatestOrCached(lastKnownVersion=" + lastKnownVersion + ", " + needsBackwardCompatibility + ")");
            }
            ServiceResourceMappings ret        = null;
            boolean                 lockResult = false;

            try {
                final boolean isCacheReloadedByDQEvent;

                lockResult = lock.tryLock(waitTimeInSeconds, TimeUnit.SECONDS);

                if (lockResult) {
                    isCacheReloadedByDQEvent = getLatest(servicePair, lastKnownVersion, needsBackwardCompatibility);

                    if (isCacheReloadedByDQEvent) {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("ServiceResourceMappings cache was completely loaded from database because of a disqualifying event - such as full-sync!");
                        }
                    }

                    if (needsBackwardCompatibility || isCacheReloadedByDQEvent || lastKnownVersion == -1L || serviceResourceMappings == null || lastKnownVersion.equals(serviceResourceMappings.getCurrentNotificationId())) {
                        // Looking for all mappings, or Some disqualifying change encountered
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("All mappings were requested, returning cached ServiceResourceMappings");
                        }
                        ret = this.serviceResourceMappings;
                    } else {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Deltas are requested, computing deltas!");
                        }

                        ServiceResourceMappings serviceResourceMappingsDeltas = null;

                        MappingDeltasFromVersion cachedDelta = deltaCache.get(serviceResourceMappings.getCurrentNotificationId());

                        if (cachedDelta != null && cachedDelta.getFromVersion().equals(lastKnownVersion)) {
                            ret = cachedDelta.getDeltas();
                        } else {
                            //  Get the deltas all-over again
                            try {
                                serviceResourceMappingsDeltas = resourceMgr.getResourceMappingDeltas(servicePair.llService, servicePair.hlService, lastKnownVersion, serviceResourceMappings.getCurrentNotificationId());
                                if (serviceResourceMappingsDeltas == null) {
                                    LOG.info("Deltas were requested, but perhaps no significant changes between notification-id:[" + lastKnownVersion + "] and [" + serviceResourceMappings.getCurrentNotificationId() + "]");
                                } else {
                                    if (LOG.isDebugEnabled()) {
                                        LOG.debug("Returning deltas!");
                                    }
                                }

                                deltaCache.clear();
                                deltaCache.put(serviceResourceMappings.getCurrentNotificationId(), new MappingDeltasFromVersion(lastKnownVersion, serviceResourceMappingsDeltas));

                                ret = serviceResourceMappingsDeltas;
                            } catch (Exception e) {
                                LOG.error("Could not get deltas!", e);
                                LOG.warn("Deltas were requested, but could not get deltas. Returning all mappings!");
                                ret = this.serviceResourceMappings;

                            }
                        }
                    }
                } else {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Could not get lock in [" + waitTimeInSeconds + "] seconds, returning cached ServiceResourceMappings");
                    }
                    ret = this.serviceResourceMappings;
                }
            } catch (InterruptedException exception) {
                LOG.error("getLatestOrCached:lock got interrupted..", exception);
            } finally {
                if (lockResult) {
                    lock.unlock();
                }
            }
            if (LOG.isTraceEnabled()) {
                LOG.trace("ServiceResourceMappingsCache.getLatestOrCached - Returns ServiceResourceMappings:[" + ret +"]");
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("<== ServiceResourceMappingsCache.getLatestOrCached(lastKnownVersion=" + lastKnownVersion + ", " + needsBackwardCompatibility + ") : " + ret);
            }
            return ret;
        }

        boolean getLatest(ServicePair servicePair, Long lastKnownVersion, boolean needsBackwardCompatibility) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("==> ServiceResourceMappingsWrapper.getLatest(servicePair=" + servicePair + ", lastKnownVersion=" + lastKnownVersion + ", needsBackwardCompatibility=" + needsBackwardCompatibility + ")");
            }

            boolean isCacheReloadedByDQEvent = false;

            final long serviceResourcesMappingVersionInDb   = resourceMgr.getLatestKnownDownloadableVersion(servicePair.llService.getId(), servicePair.hlService.getId());

            if (serviceResourcesMappingVersionInDb > 0L && serviceResourceMappings == null) {
                // load all service-mappings
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Initializing ServiceResourceMappings cache for the first time");
                }
                deltaCache.clear();
                serviceResourceMappings = resourceMgr.buildServiceResourceMappings(servicePair.llService, servicePair.hlService, lastKnownVersion, serviceResourcesMappingVersionInDb);
                pruneUnusedAttributes();

                isCacheReloadedByDQEvent = true;

                if (serviceResourceMappings == null) {
                    LOG.error("Could not get mappings from database, to-version:[" + serviceResourcesMappingVersionInDb + ")");
                    LOG.error("Cannot initialize in-memory cache of mappings! CANNOT SERVE MAPPING DOWNLOADS....");
                }
            }

            if (serviceResourceMappings != null) {
                final long cachedServiceResourceMappingsVersion = serviceResourceMappings.getCurrentNotificationId();

                if (LOG.isDebugEnabled()) {
                    LOG.debug("ServiceResourceMappings version in cache[" + cachedServiceResourceMappingsVersion + "], ServiceResourceMappings version in database[" + serviceResourcesMappingVersionInDb + "]");
                }

                if (serviceResourcesMappingVersionInDb != 0L) {
                    if (serviceResourcesMappingVersionInDb != cachedServiceResourceMappingsVersion) {

                        if (LOG.isDebugEnabled()) {
                            LOG.debug("loading serviceResourceMappings from database");
                        }

                        deltaCache.clear();

                        if (!needsBackwardCompatibility) {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Attempting to get service-mappings deltas");
                            }

                            try {
                                ServiceResourceMappings serviceResourceMappingsDeltas = resourceMgr.getResourceMappingDeltas(servicePair.llService, servicePair.hlService, cachedServiceResourceMappingsVersion, serviceResourcesMappingVersionInDb);
                                if (serviceResourceMappingsDeltas == null) {
                                    LOG.info("Deltas are null! Probably spurious notifications were received from notification-id:[" + cachedServiceResourceMappingsVersion + "] and [" + serviceResourcesMappingVersionInDb + "]");
                                    serviceResourceMappings.setCurrentNotificationId(serviceResourcesMappingVersionInDb);
                                } else {
                                    ServiceResourceMappingsDeltaUtil.applyDelta(serviceResourceMappings, serviceResourceMappingsDeltas);
                                }

                                if (lastKnownVersion != -1L) {
                                    deltaCache.put(serviceResourcesMappingVersionInDb, new MappingDeltasFromVersion(lastKnownVersion, serviceResourceMappingsDeltas));
                                }

                                pruneUnusedAttributes();
                            } catch (Exception exception) {
                                LOG.error("Could not get deltas from database:", exception);
                                isCacheReloadedByDQEvent = true;
                            }

                        }
                        if (isCacheReloadedByDQEvent) {
                            deltaCache.clear();
                            serviceResourceMappings = resourceMgr.buildServiceResourceMappings(servicePair.llService, servicePair.hlService, lastKnownVersion, serviceResourcesMappingVersionInDb);

                            if (serviceResourceMappings != null) {
                                if (LOG.isDebugEnabled()) {
                                    LOG.debug("Successfully loaded ServiceResourceMappings from database: ServiceResourceMappings:[" + serviceResourceMappings + "]");
                                }

                                pruneUnusedAttributes();

                            } else {
                                LOG.error("Could not get mappings from database, from-version:[" + cachedServiceResourceMappingsVersion + ")");
                                LOG.error("Cannot initialize in-memory cache of mappings! CANNOT SERVE MAPPING DOWNLOADS....");
                            }
                        }
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("ServiceResourceMappings old-version:[" + cachedServiceResourceMappingsVersion + "], new-version:[" + serviceResourceMappings.getCurrentNotificationId() + "]");
                        }
                    } else {
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("ServiceResourceMappings Cache already has the latest version, version:[" + serviceResourceMappings.getCurrentNotificationId() + "]");
                        }
                    }
                } else {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("No notifications found for given source and target services: [" + servicePair.llService.getName() + ", " + servicePair.hlService.getName() + "]");
                    }
                }

                if (LOG.isTraceEnabled()) {
                    LOG.trace("Latest Cached ServiceResourceMappings:[" + serviceResourceMappings + "]");
                }
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("<== ServiceResourceMappingsWrapper.getLatest(servicePair=" + servicePair + ", lastKnownVersion=" + lastKnownVersion + ", needsBackwardCompatibility=" + needsBackwardCompatibility + ") : " + isCacheReloadedByDQEvent);
            }
            return isCacheReloadedByDQEvent;
        }

        private void pruneUnusedAttributes() {
            if (serviceResourceMappings != null) {
                pruneUnusedPolicyAttributes(serviceResourceMappings.getServiceResources());
            }
        }

        private void pruneUnusedPolicyAttributes(Map<Long, RangerServiceResource> serviceResources) {

            // Null out attributes not required by plug-ins
            for (Map.Entry<Long, RangerServiceResource> entry : serviceResources.entrySet()) {
                RangerServiceResource serviceResource = entry.getValue();
                serviceResource.setCreatedBy(null);
                serviceResource.setCreateTime(null);
                serviceResource.setUpdatedBy(null);
                serviceResource.setUpdateTime(null);
                serviceResource.setGuid(null);
            }

        }

        StringBuilder toString(StringBuilder sb) {
            sb.append("RangerServiceResourceMappingsWrapper={");

            sb.append(", Service-Version:").append(serviceResourceMappings != null ? serviceResourceMappings.getCurrentNotificationId() : "null")
                    .append(", Number-Of-mappings:").append(serviceResourceMappings != null && serviceResourceMappings.getMappings() != null ? serviceResourceMappings.getMappings().size() : 0);

            sb.append("} ");

            return sb;
        }

        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();

            toString(sb);

            return sb.toString();
        }
    }

}


