/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.ranger.rms.biz;

import java.util.Set;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.logging.Log;
import org.apache.ranger.hms.HMSConfiguration;
import org.apache.ranger.hms.HMSFollower;
import org.apache.ranger.plugin.policyresourcematcher.RangerDefaultPolicyResourceMatcher;
import org.apache.ranger.plugin.util.RangerPerfTracer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Component;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.support.TransactionTemplate;

@Component
public class MetadataSyncService {
	
	@Autowired
	ResourceMgr resourceMgr;

	@Autowired
	HMSFollower hmsFollower;

    @Autowired
    @Qualifier(value = "transactionManager")
    PlatformTransactionManager txManager;

	private static final Logger LOG = LoggerFactory.getLogger(MetadataSyncService.class);
	private static final Log PERF_LOG = RangerPerfTracer.getPerfLogger("MetadataSyncService.sync");

	private TransactionTemplate txTemplate;

	public void startMetadataSynchronizerService() {
		LOG.debug("==> startMetadataSynchronizerService()");

		synchronized(MetadataSyncService.class) {

			Set<String> configuredHMSs = HMSConfiguration.getHMSNames();

			if (CollectionUtils.isNotEmpty(configuredHMSs)) {
				txTemplate = new TransactionTemplate(txManager);
				txTemplate.setPropagationBehavior(TransactionDefinition.PROPAGATION_REQUIRES_NEW);

				boolean isMappingProvidersInited = initializeMappingProviders(configuredHMSs);

				if (isMappingProvidersInited) {
					for (String hmsName : configuredHMSs) {
						RangerPerfTracer perf = null;

						if (RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
							perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "MetadataSyncService.startSync " + hmsName);
						}

						try {
							txTemplate.execute(status -> {

								try {
									boolean isValidServices = hmsFollower.purgeIfInvalidServices(hmsName);

									if (isValidServices) {
										long lastSyncedNotificationId = resourceMgr.getLastSyncedNotificationId(hmsName);
										long currentNotificationId    = hmsFollower.getCurrentNotificationId(hmsName);

										if (currentNotificationId != lastSyncedNotificationId) {
											LOG.info("HMS:[" + hmsName + "]: " + "Last Synced Notification Id:[" + lastSyncedNotificationId + "], Current Notification Id:[" + currentNotificationId + "]");

											try {
												Set<RangerDefaultPolicyResourceMatcher> matchers = resourceMgr.getResourceFilter(hmsName);
												hmsFollower.syncFromHMS(hmsName, matchers, lastSyncedNotificationId, currentNotificationId);
											} catch (Throwable e) {
												LOG.error("Cannot get resource-mappings from HMS:[" + hmsName + "] or could not save notifications to Ranger-rms database", e);
											}
										} else {
											LOG.debug("No new notifications in HMS:[" + hmsName + "], currentNotificationId:[" + currentNotificationId + "], lastSyncedNotificationId:[" + lastSyncedNotificationId + "]");
										}
									} else {
										LOG.info("Invalid service-data, not processing this notification!");
									}
								} catch (Exception ex) {
									LOG.error("Cannot get current notification ID from HMS:[" + hmsName + "]", ex);
								}

								return null;
							});
						} catch (Throwable exception) {
							LOG.error("MetadataSyncService.startMetadataSynchronizerService(): Failed to fetch/save notifications for HMS:[" + hmsName + "]", exception);
						} finally {
							RangerPerfTracer.logAlways(perf);
						}
					}
				}
			} else {
				LOG.error("At least one mapping provider was not initialized correctly. Aborting this sync attempt!");
			}
		}

		LOG.debug("<== startMetadataSynchronizerService()");
	}

	private boolean initializeMappingProviders(Set<String> configuredHMSs) {
		LOG.debug("==> initializeMappingProviders()");

		boolean ret = true;

		for (String hmsName : configuredHMSs) {
			try {
				txTemplate.execute(status -> {
					resourceMgr.initMappingProvider(hmsName);
					return null;
				});
			} catch (Throwable exception) {
				LOG.error("MetadataSyncService.startMetadataSynchronizerService(): Failed to initialize mapping provider:[" + hmsName + "]", exception);
				ret = false;
			}
		}

		LOG.debug("<== initializeMappingProviders() : ret:[" + ret + "]");

		return ret;
	}
}
