/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.rms.biz;

import java.io.IOException;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import com.sun.jersey.api.client.ClientResponse;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.logging.Log;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.SecureClientLogin;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.ranger.audit.provider.MiscUtil;
import org.apache.ranger.db.RangerDaoManager;
import org.apache.ranger.entity.XXService;
import org.apache.ranger.hms.HMSConfiguration;
import org.apache.ranger.hms.HMSFollower;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerServiceDef;
import org.apache.ranger.plugin.model.RangerServiceResource;
import org.apache.ranger.plugin.policyresourcematcher.RangerDefaultPolicyResourceMatcher;
import org.apache.ranger.plugin.store.RangerServiceResourceSignature;
import org.apache.ranger.plugin.util.RangerMappingRESTRetriever;
import org.apache.ranger.plugin.util.RangerPerfTracer;
import org.apache.ranger.plugin.util.ServiceResourceMappings;
import org.apache.ranger.db.XXRMSMappingProviderDao;
import org.apache.ranger.db.XXRMSNotificationDao;
import org.apache.ranger.db.XXRMSResourceMappingDao;
import org.apache.ranger.db.XXRMSServiceResourceDao;
import org.apache.ranger.entity.XXRMSMappingProvider;
import org.apache.ranger.entity.XXRMSNotification;
import org.apache.ranger.entity.XXRMSResourceMapping;
import org.apache.ranger.entity.XXRMSServiceResource;
import org.apache.ranger.rms.server.EmbeddedServer;
import org.apache.ranger.rms.server.RMSConfig;
import org.apache.ranger.rms.util.AppConstants;
import org.apache.ranger.rms.util.RMSTransactionHook;
import org.apache.ranger.rms.util.ResourceManagementHook;
import org.apache.ranger.rms.view.RMSNotification;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Component;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.support.TransactionTemplate;

import javax.annotation.PostConstruct;
import javax.security.auth.Subject;

@Component
public class ResourceMgr extends ResourceManagementHook implements RMSTransactionHook {

	private static final String RANGER_ADMIN_REST_URL                 = "ranger.admin.rest.url";
	private static final String RANGER_ADMIN_SSL_CONFIG_FILE          = "ranger.admin.rest.ssl.config.file";
	private static final String RANGER_ADMIN_SERVICEDEF_ENDPOINT_ROOT = "/service/plugins/definitions/";

	private static final Logger LOG = LoggerFactory.getLogger(ResourceMgr.class);
	private static final Log    PERF_LOG = RangerPerfTracer.getPerfLogger("ResourceMgr.db");

	private static final Map<String, XXService>    rangerServices = Collections.synchronizedMap(new HashMap<>());
	private static final Map<Long, RangerServiceDef> rangerServiceDefs = Collections.synchronizedMap(new HashMap<>());

	private        final Map<String, Set<RangerDefaultPolicyResourceMatcher>> mappingResourceFilters = new HashMap<>();

	private static RangerDaoManager _daoManager;
	private RangerMappingRESTRetriever client;

	@Autowired
	RangerDaoManager daoManager;

	@Autowired
	HMSFollower hmsFollower;

	@Autowired
	@Qualifier(value = "transactionManager")
	PlatformTransactionManager txManager;

	@PostConstruct
	public void configure() throws Exception {
		LOG.info("==> ResourceMgr.configure()");

		try {

			initKerberos();

			initRMSDaos();

			RMSConfig conf = RMSConfig.getInstance();

			String providerURL = conf.get(RANGER_ADMIN_REST_URL);
			String sslConfigFileName = conf.get(RANGER_ADMIN_SSL_CONFIG_FILE);

			LOG.debug("ranger-admin URL=[" + providerURL + "], sslConfigFileName:[" + sslConfigFileName + "]");

			RangerMappingRESTRetriever restClient = new RangerMappingRESTRetriever();
			if (restClient.init(providerURL, sslConfigFileName, conf, null)) {

				client = restClient;

				Set<String> hmsNames = HMSConfiguration.getHMSNames();

				for (String hmsName : hmsNames) {
					if (!getServiceInfo(hmsName)) {
						LOG.error("Could not get service information required by HMS:[" + hmsName + "] from Ranger");
					}
				}
			} else {
				String errorMessage = "Could not initialize REST client for communicating with ranger-admin";
				LOG.error(errorMessage);
				throw new Exception(errorMessage);
			}

			hmsFollower.setTransactionHook(this);
			hmsFollower.setResourceManagementHook(this);

		} catch (Exception exception) {
			LOG.error("Error while downloading services and service-defs. Resource-Manager is not constructed correctly", exception);
			throw exception;
		}

		LOG.info("<== ResourceMgr.configure()");
	}

	@Override
	public boolean getServiceInfo(String hmsName) {
		LOG.debug("==> getServiceInfo(" + hmsName + ")");

		boolean ret = true;

		String llServiceName = HMSConfiguration.getLlServiceName(hmsName);
		String hlServiceName = HMSConfiguration.getHlServiceName(hmsName);

		Set<Map<String, List<String>>> serviceRoots = HMSConfiguration.getHlServiceRoots(hmsName);

		boolean isServiceDefNeeded = CollectionUtils.isNotEmpty(serviceRoots);

		ret = ret && getServiceDetails(llServiceName, false);
		ret = ret && getServiceDetails(hlServiceName, isServiceDefNeeded);

		LOG.debug("<== getServiceInfo(" + hmsName + "): ret:[" + ret + "]");

		return ret;
	}

	private boolean getServiceDetails(String serviceName, boolean isServiceDefNeeded) {
		LOG.debug("==> getServiceDetails(" + serviceName + ", " + isServiceDefNeeded + ")");

		boolean ret = true;

		XXService service = daoManager.getXXService().findByName(serviceName);

		if (service != null) {
			rangerServices.put(serviceName, service);

			LOG.debug("Added RangerService for serviceName:[" + serviceName + "] to in-memory cache");
		} else {
			String errorMessage = "Failed to get RangerService for serviceName:[" + serviceName + "]";
			LOG.error(errorMessage);
			ret = false;
		}

		if (ret && isServiceDefNeeded) {  // Some filter is specified

			if (client != null) {
				PrivilegedAction<ClientResponse> action = client.getPrivilegedAction(RangerMappingRESTRetriever.HTTP_METHOD.GET, RANGER_ADMIN_SERVICEDEF_ENDPOINT_ROOT + service.getType(), null, null);

				try {
					ImmutablePair<ClientResponse, RangerServiceDef> serviceDefResponsePair = client.executeAction(action, RangerServiceDef.class);

					if (serviceDefResponsePair.right != null) {
						RangerServiceDef serviceDef = serviceDefResponsePair.right;
						rangerServiceDefs.put(service.getType(), serviceDef);
						LOG.debug("Added RangerServiceDef for serviceType:[" + service.getType() + "] to in-memory cache");
					} else {
						String errorMessage = "Failed to get RangerServiceDef for serviceType:[" + service.getType() + "]";
						LOG.error(errorMessage);
						ret = false;
					}
				} catch (Exception e) {
					LOG.error("Cannot get service-definition for [" + service.getType() + "]", e);
					ret = false;
				}
			} else {
				String errorMessage = "Failed to get RangerServiceDef for serviceType:[" + service.getType() + "]";
				LOG.error(errorMessage);
				ret = false;
			}
		}
		LOG.debug("<== getServiceDetails(" + serviceName + ", " + isServiceDefNeeded + ") : ret:[" + ret + "]");

		return ret;
	}

	@Override
	public void commitWork(String hmsName, final List<RMSNotification> notifications, final long lastSyncedNotification, final boolean isPersistNotification) {
		LOG.debug("==> commitWork(hmsName=" + hmsName + ", number-of-notifications=" + notifications.size() + ", lastSyncedNotification=" + lastSyncedNotification + ", " + isPersistNotification + ")");

		RangerPerfTracer perf = null;

		if(RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
			perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "ResourceMgr.commitWork(lastSyncedNotification=" + lastSyncedNotification + ", numberOfNotifications=" + notifications.size() + ")");
		}

		TransactionTemplate txTemplate = new TransactionTemplate(txManager);
		txTemplate.setPropagationBehavior(TransactionDefinition.PROPAGATION_REQUIRES_NEW);

		txTemplate.execute(status -> {
			createNotifications(notifications, isPersistNotification);
			updateCurrentNotificationId(hmsName, lastSyncedNotification);
			return null;
		});

		RangerPerfTracer.logAlways(perf);

		LOG.debug("<== commitWork(hmsName=" + hmsName + ", number-of-notifications=" + notifications.size() + ", lastSyncedNotification=" + lastSyncedNotification + ", " + isPersistNotification + ")");
	}

	@Override
	public boolean purgeServiceResourcesAndMappings(String llServiceName, String hlServiceName) {
		LOG.debug("==> purgeServiceResourcesAndMappings(llServiceName=" + llServiceName + ", hlServiceName=" + hlServiceName + ")");

		boolean ret;
		XXService llService = rangerServices.get(llServiceName);
		XXService hlService = rangerServices.get(hlServiceName);

		if (llService != null || hlService != null) {
			TransactionTemplate txTemplate = new TransactionTemplate(txManager);
			txTemplate.setPropagationBehavior(TransactionDefinition.PROPAGATION_REQUIRES_NEW);

			txTemplate.execute(status -> {
				if (llService != null) {
					daoManager.getXXRMSServiceResource().purge(llService.getId());
				}
				if (hlService != null) {
					daoManager.getXXRMSServiceResource().purge(hlService.getId());
				}
				return null;
			});

			rangerServices.remove(llServiceName);
			rangerServices.remove(hlServiceName);

			ret = true;
		} else {
			LOG.error("Failed to get service ids for both llServiceName:[" + llServiceName + "] and hlServiceName:[" + hlServiceName + "]");
			ret = false;
		}

		LOG.debug("<== purgeServiceResourcesAndMappings(llServiceName=" + llServiceName + ", hlServiceName=" + hlServiceName + "): ret:[" + ret + "]");

		return ret;
	}

	public static XXService getServiceByName(String serviceName) {
		return _daoManager == null ? null : _daoManager.getXXService().findByName(serviceName);
	}

	public ServiceResourceMappings buildServiceResourceMappings(XXService sourceService, XXService targetService, long lastKnownVersion, long currentNotificationId) {

		LOG.debug("==> buildServiceResourceMappings(sourceService=" + sourceService.getName() + ", targetService=" + targetService.getName() + ", lastKnownVersion=" + lastKnownVersion + ", currentNotificationId=" + currentNotificationId + ")");

		ServiceResourceMappings          ret;

		LOG.debug("Load mappings in a separate thread and using a new read-only transaction");

		TransactionTemplate txTemplate = new TransactionTemplate(txManager);
		txTemplate.setReadOnly(true);

		MappingLoaderThread t = new MappingLoaderThread(txTemplate, sourceService.getId());
		t.setName("Mapping-Loader-Thread-for" + sourceService.getName() + "-" + t.getId());
		t.setDaemon(true);
		t.start();

		try {
			t.join();

			if (t.getIsMappingsLoaded()) {
				Map<Long, RangerServiceResource> serviceResources = t.getServiceResources();
				Map<Long, List<Long>> mappings = t.getMappings();

				ret = new ServiceResourceMappings(AppConstants.OP_ADD, sourceService.getName(), targetService.getName(), currentNotificationId, lastKnownVersion, new Date(), serviceResources, mappings);
			} else {
				ret = null;
			}

		} catch (InterruptedException ie) {
			LOG.error("Failed to get mappings in a separate thread and using a new transaction", ie);
			ret = null;
		}

		LOG.debug("<== buildServiceResourceMappings(sourceService=" + sourceService.getName() + ", targetService=" + targetService.getName() + ", lastKnownVersion=" + lastKnownVersion + ", currentNotificationId=" + currentNotificationId + ")");

		return ret;
	}

	public boolean createNotifications(List<RMSNotification> rmsNotifications, final boolean isPersistNotification) {

		LOG.debug("==> createNotifications()");

		boolean ret = true;

		RangerPerfTracer perf = null;

		if(RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
			perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "ResourceMgr.createNotifications");
		}

		for (RMSNotification rmsNotification : rmsNotifications) {

			if (!isValidNotification(rmsNotification)) {
				LOG.warn("Invalid notification:[" + rmsNotification +"]");
				continue;
			}

			setupNotification(rmsNotification);

			boolean result;

			if (AppConstants.OP_ADD.equalsIgnoreCase(rmsNotification.getOp())
				|| AppConstants.OP_UPDATE.equalsIgnoreCase(rmsNotification.getOp())) {
				result = handleAddOp(rmsNotification);
			}
			/*
			else if (AppConstants.OP_UPDATE.equalsIgnoreCase(rmsNotification.getOp())) {
				result = handleUpdateOp(rmsNotification);
				if (result) {
					rmsNotification.setOp(AppConstants.OP_ADD);
					result = handleAddOp(rmsNotification);
				}
			}
			*/
			else if (AppConstants.OP_DELETE.equalsIgnoreCase(rmsNotification.getOp())) {
				result = handleDeleteOp(rmsNotification);
			} else if (AppConstants.OP_INVALID.equalsIgnoreCase(rmsNotification.getOp())) {
				result = true;
			} else {
				LOG.info("Notification :[" + rmsNotification + "] will not be processed.");
				continue;
			}

			if (result) {
				if (isPersistNotification) {
					if (createXXNotification(rmsNotification) == null) {
						LOG.error("Failed to create notification, failed notification:[" + rmsNotification + "]");
					}
				}
			} else {
				LOG.warn("Did not persist notification:[" + rmsNotification + "] as notification-handler returned false");
				ret = ret && false;
			}
		}
		RangerPerfTracer.log(perf);

		LOG.debug("<== createNotifications(): ret=[" + ret + "]");

		return ret;

	}

	public long getLastSyncedNotificationId(String hmsName) {
		return daoManager.getXXRMSMappingProvider().getLastKnownVersion(hmsName);
	}

	public void updateCurrentNotificationId(String hmsName, long currentNotificationId) {
		daoManager.getXXRMSMappingProvider().updateLastKnownVersion(hmsName, currentNotificationId);
	}

	public long  getLatestKnownDownloadableVersion(long llServiceId, long hlServiceId) {
		return daoManager.getXXRMSNotification().getMaxIdOfNotifications(llServiceId, hlServiceId);
	}

	public Set<RangerDefaultPolicyResourceMatcher> getResourceFilter(String providerName) {
		LOG.debug("==> ResourceMgr.getResourceFilter(" + providerName + ")");

		Set<RangerDefaultPolicyResourceMatcher> ret = mappingResourceFilters.get(providerName);

		if (ret == null) {
			LOG.debug("ResourceMgr.getResourceFilter() - building resourceFilter for provider:[" + providerName + "]");

			ret = new HashSet<>();

			Set<Map<String, List<String>>> rootResources = HMSConfiguration.getHlServiceRoots(providerName);
			String serviceName = HMSConfiguration.getHlServiceName(providerName);
			RangerServiceDef serviceDef = getServiceDefByServiceName(serviceName);

			if (serviceDef != null) {

				LOG.debug("ResourceMgr.getResourceFilter() - processing rootResources for service:[" + serviceName + "]");

				for (Map<String, List<String>> rootResource : rootResources) {
					Map<String, RangerPolicy.RangerPolicyResource> oneResource = new HashMap<>();

					for (Map.Entry<String, List<String>> entry : rootResource.entrySet()) {
						RangerPolicy.RangerPolicyResource policyResource = new RangerPolicy.RangerPolicyResource();
						policyResource.setValues(entry.getValue());
						oneResource.put(entry.getKey(), policyResource);
					}

					RangerDefaultPolicyResourceMatcher matcher = new RangerDefaultPolicyResourceMatcher();
					matcher.setPolicyResources(oneResource);
					matcher.setServiceDef(serviceDef);
					matcher.init();

					ret.add(matcher);

				}
			}
			mappingResourceFilters.put(providerName, ret);
		}

		LOG.debug("<== ResourceMgr.getResourceFilter(" + providerName + ") : matcher set-size:[" + ret.size() + "]");

		return ret;
	}

	public void initMappingProvider(String mappingProviderName) {
		LOG.debug("==> ResourceMgr. initMappingProvider(" + mappingProviderName + ")");

		XXRMSMappingProviderDao mappingProviderDao = daoManager.getXXRMSMappingProvider();

		XXRMSMappingProvider mappingProvider = mappingProviderDao.findByName(mappingProviderName);

		if (mappingProvider == null) {
			LOG.debug("Creating a new mapping provider with name:[" + mappingProviderName + "]");
			mappingProvider = new XXRMSMappingProvider();
			mappingProvider.setName(mappingProviderName);
			mappingProvider.setLastKnownVersion(-1L);
			mappingProvider.setChangeTimestamp(new Date());
			mappingProviderDao.create(mappingProvider);
		} else {
			LOG.debug("Current last-known-version retrieved from mappingProvider:[" + mappingProviderName + "] is [" + mappingProvider.getLastKnownVersion() + "]");
		}

		LOG.debug("<== ResourceMgr. initMappingProvider(" + mappingProviderName + ")");
	}

	private static RangerServiceDef getServiceDefByType(Long serviceType) {
		return rangerServiceDefs.get(serviceType);
	}

	private static RangerServiceDef getServiceDefByServiceName(String serviceName) {
		XXService service = getServiceByName(serviceName);
		return service == null ? null : getServiceDefByType(service.getType());
	}

	private void initKerberos() {
		LOG.info("==> ResourceMgr.initKerberos()");

		Configuration config = RMSConfig.getInstance();

		String  authenticationType = config.get(EmbeddedServer.HADOOP_AUTHENTICATION_TYPE);
		boolean isKerberosAuth     = org.apache.commons.lang.StringUtils.equals(authenticationType, EmbeddedServer.HADOOP_AUTH_TYPE_KERBEROS);

		if (isKerberosAuth) {
			String kerbPrincipal = config.get(AppConstants.CONFIG_PREFIX + EmbeddedServer.RANGER_RMS_KERBEROS_PRINCIPAL);
			String keyTab        = config.get(AppConstants.CONFIG_PREFIX + EmbeddedServer.RANGER_RMS_KERBEROS_KEYTAB);
			String hostName      = config.get(AppConstants.PROP_RMS_SERVER_BIND_ADDRESS);
			String nameRules     = config.get(EmbeddedServer.HADOOP_KERBEROS_NAME_RULES, EmbeddedServer.DEFAULT_NAME_RULE);

			LOG.info("ResourceMgr.initKerberos(): kerberos is enabled");

			UserGroupInformation.setConfiguration(config);

			try {
				String  principal = SecureClientLogin.getPrincipal(kerbPrincipal, hostName);
				Subject sub       = SecureClientLogin.loginUserFromKeytab(principal, keyTab, nameRules);

				UserGroupInformation.loginUserFromSubject(sub);

				MiscUtil.setUGILoginUser(UserGroupInformation.getLoginUser(), sub);
			} catch (IOException excp) {
				LOG.warn("ResourceMgr.initKerberos(): failed to get principal" + excp);
			}
		} else {
			LOG.info("ResourceMgr.initKerberos(): kerberos is not enabled");

		}

		LOG.info("UserGroupInformation.isSecurityEnabled()={}, MiscUtil.getUGILoginUser()={}", UserGroupInformation.isSecurityEnabled(), MiscUtil.getUGILoginUser());

		LOG.info("<== ResourceMgr.initKerberos()");
	}

	private void initRMSDaos() {
		_daoManager = daoManager;

		daoManager.getXXService();
		daoManager.getXXRMSMappingProvider();
		daoManager.getXXRMSNotification();
		daoManager.getXXRMSServiceResource();
		daoManager.getXXRMSResourceMapping();
	}

	private boolean isValidNotification(RMSNotification rmsNotification) {
		boolean validNotification = true;

		if (rmsNotification == null || rmsNotification.getHlResource() == null) {
			validNotification = false;
		}
		return validNotification;
	}

	private void setupNotification(RMSNotification notification) {
		String hmsName       = notification.getHmsName();
		String llServiceName = HMSConfiguration.getLlServiceName(hmsName);
		String hlServiceName = HMSConfiguration.getHlServiceName(hmsName);

		notification.setHlServiceName(hlServiceName);
		notification.setLlServiceName(llServiceName);
	}

	private XXRMSNotification createXXNotification(RMSNotification vObj) {
		XXRMSNotification ret = populateEntityBean(vObj);

		if(ret != null) {
			ret = daoManager.getXXRMSNotification().create(ret);
		}
		return ret;
	}

	private XXRMSNotification populateEntityBean(RMSNotification vObj) {
		XXRMSNotification ret = null;

		if (vObj != null) {
			ret = new XXRMSNotification();
			ret.setHmsName(vObj.getHmsName());
			ret.setNotificationId(vObj.getNotificationID());
			ret.setChangeTimestamp(new Date());
			ret.setChangeType(vObj.getOp());
			ret.setHlResourceId(vObj.getHlResourceId());
			ret.setLlResourceId(vObj.getLlResourceId());

			XXService service = getServiceByName(vObj.getHlServiceName());
			if (service != null) {
				ret.setHlServiceId(service.getId());
			}

			service = getServiceByName(vObj.getLlServiceName());
			if (service != null) {
				ret.setLlServiceId(service.getId());
			}
		} else {
			LOG.error("Cannot populate entity bean from null RMSNotification object!");
		}
		return ret;
	}

	private XXRMSResourceMapping createXXResourceMapping(Long hlResourceId, Long llResourceId) {
		XXRMSResourceMapping newXXResourceMapping = null;

		if(hlResourceId != null && llResourceId != null) {
			newXXResourceMapping = daoManager.getXXRMSResourceMapping().findByHlAndLlResourceId(hlResourceId, llResourceId);
			if (newXXResourceMapping == null) {
				newXXResourceMapping = new XXRMSResourceMapping();
				newXXResourceMapping.setChangeTimestamp(new Date());
				newXXResourceMapping.setHlResourceId(hlResourceId);
				newXXResourceMapping.setLlResourceId(llResourceId);
				newXXResourceMapping = daoManager.getXXRMSResourceMapping().create(newXXResourceMapping);
			}
		}
		return newXXResourceMapping;
	}

	private void deleteXXResourceMapping(Long hlResourceId, Long llResourceId) {
		if (hlResourceId != null && llResourceId != null) {
			XXRMSResourceMappingDao mappingDao         = daoManager.getXXRMSResourceMapping();
			XXRMSServiceResourceDao serviceResourceDao = daoManager.getXXRMSServiceResource();

			List<Long> llResourceIds = mappingDao.findByHlResourceId(hlResourceId);
			List<Long> hlResourceIds = mappingDao.findByLlResourceId(llResourceId);

			mappingDao.deleteByHlAndLlResourceId(hlResourceId, llResourceId);

			if (CollectionUtils.isEmpty(llResourceIds) || llResourceIds.size() == 1) {
				serviceResourceDao.deleteById(hlResourceId);
			}

			if (CollectionUtils.isEmpty(hlResourceIds) || hlResourceIds.size() == 1) {
				serviceResourceDao.deleteById(llResourceId);
			}
		}
	}

	private boolean handleAddOp(RMSNotification rmsNotification) {

		LOG.debug("==> ResourceMgr.handleAddOp(notification:[" + rmsNotification + "])");

		boolean ret = false;

		/* Ideally, this should be a mapper which takes the service-name and stringified value of resource and returns a RangerServiceResource */

		RangerServiceResource llServiceResource = rmsNotification.getLlAsServiceResource();

		RangerServiceResourceSignature serializer;
		serializer = new RangerServiceResourceSignature(llServiceResource);

		String resourceSignature = serializer.getSignature();
		llServiceResource.setResourceSignature(resourceSignature);

		RangerServiceResource existing = null;

		try {
			existing = daoManager.getXXRMSServiceResource().getServiceResourceByServiceAndResourceSignature(llServiceResource.getServiceName(), resourceSignature);
		} catch (Exception e) {
			LOG.error("Cannot find service-resource:", e);
		}

		if (existing == null) {
			llServiceResource = daoManager.getXXRMSServiceResource().createServiceResource(llServiceResource);
		} else {
			llServiceResource = existing;
		}
		rmsNotification.setLlResourceId(llServiceResource.getId());

		/* Ideally, this should be a mapper which takes the service-name and stringified value of resource and returns a RangerServiceResource */

		RangerServiceResource hlServiceResource = rmsNotification.getHlAsServiceResource();
		hlServiceResource.setOwnerUser(rmsNotification.getOwnerName());
		hlServiceResource.setAdditionalInfo(rmsNotification.getAdditionalInfo());

		//new RangerServiceResource("", rmsNotification.getHlServiceName(), rmsNotification.getHlResource());

		if(MapUtils.isNotEmpty(hlServiceResource.getResourceElements())) {
			serializer = new RangerServiceResourceSignature(hlServiceResource);

			resourceSignature = serializer.getSignature();
			hlServiceResource.setResourceSignature(resourceSignature);

			existing = null;

			try {
				existing = daoManager.getXXRMSServiceResource().getServiceResourceByServiceAndResourceSignature(hlServiceResource.getServiceName(), resourceSignature);
			} catch (Exception e) {
				LOG.error("Cannot find service-resource:", e);
			}

			if (existing == null) {
				hlServiceResource = daoManager.getXXRMSServiceResource().createServiceResource(hlServiceResource);
			} else {
				hlServiceResource = existing;
			}

			rmsNotification.setHlResourceId(hlServiceResource.getId());

		}

		XXRMSResourceMapping mapping = createXXResourceMapping(rmsNotification.getHlResourceId(), rmsNotification.getLlResourceId());
		if (mapping != null) {
			LOG.debug("Created mapping between resources: from:[" + rmsNotification.getHlResourceId() + "], to:[" + rmsNotification.getLlResourceId() + "]. mapping-id:[" + mapping.getId() + "]");
			ret = true;
		} else {
			LOG.error("Cannot create mapping between resources: from:[" + rmsNotification.getHlResourceId() + "], to:[" + rmsNotification.getLlResourceId() + "]");
		}

		if (!ret) {
			LOG.info("ResourceMgr.handleAddOp() not processed successfully, notification:[" + rmsNotification + "]");
		}

		LOG.debug("<== ResourceMgr.handleAddOp(notification:[" + rmsNotification + "]) : ret:[" + ret + "]");

		return ret;

	}

	private boolean handleDeleteOp(RMSNotification rmsNotification) {

		LOG.debug("==> ResourceMgr.handleDeleteOp(notification:[" + rmsNotification + "])");

		boolean ret = false;

		RangerServiceResource hlServiceResource = rmsNotification.getHlAsServiceResource();

		RangerServiceResourceSignature serializer = new RangerServiceResourceSignature(hlServiceResource);

		String resourceSignature = serializer.getSignature();
		hlServiceResource.setResourceSignature(resourceSignature);

		RangerServiceResource existing = null;

		try {
			existing = daoManager.getXXRMSServiceResource().getServiceResourceByServiceAndResourceSignature(hlServiceResource.getServiceName(), resourceSignature);
		} catch (Exception e) {
			LOG.warn("Cannot find service-resource:[" + hlServiceResource + "]", e);
		}
		if (existing != null) {
			hlServiceResource = existing;

			// Find those resources which associate with High-Level resource

			List<Long> llResourceIds = daoManager.getXXRMSResourceMapping().findByHlResource(hlServiceResource);

			if (CollectionUtils.isEmpty(llResourceIds)) {
				LOG.warn("Should not come here! Low-level resource cannot be absent from aclsync database for hlResource:[" + hlServiceResource + "]");
			} else {
				rmsNotification.setHlResourceId(hlServiceResource.getId());

				int numOfLlResources = llResourceIds.size();

				rmsNotification.setLlResourceId(llResourceIds.get(numOfLlResources-1));
				deleteXXResourceMapping(hlServiceResource.getId(), llResourceIds.get(numOfLlResources-1));

				for (int i = 0; i < numOfLlResources-1; i++) {
					Long llResourceId = llResourceIds.get(i);

					deleteXXResourceMapping(hlServiceResource.getId(), llResourceId);

					RMSNotification copy = new RMSNotification(rmsNotification);

					copy.setLlResourceId(llResourceId);

					if (createXXNotification(copy) == null) {
						LOG.error("Failed to create notification, failed notification:[" + copy + "]");
					}
				}
				ret = true;

			}

		}
		if (!ret) {
			LOG.info("ResourceMgr.handleDeleteOp() not processed successfully, notification:[" + rmsNotification + "]");
		}

		LOG.debug("<== ResourceMgr.handleDeleteOp(notification:[" + rmsNotification + "]: ret:[" + ret + "]");

		return ret;
	}

	/*
	private boolean handleUpdateOp(RMSNotification rmsNotification) {

		LOG.debug("==> ResourceMgr.handleUpdateOp(notification:[" + rmsNotification + "])");

		boolean ret = false;

		// Ideally, this should be a mapper which takes the service-name and stringified value of resource and returns a RangerServiceResource

		RangerServiceResource llServiceResource = rmsNotification.getLlAsServiceResource();

		RangerServiceResourceSignature serializer;
		serializer = new RangerServiceResourceSignature(llServiceResource);

		String resourceSignature = serializer.getSignature();
		llServiceResource.setResourceSignature(resourceSignature);

		RangerServiceResource existing = null;
		try {
			existing = daoManager.getXXRMSServiceResource().getServiceResourceByServiceAndResourceSignature(llServiceResource.getServiceName(), resourceSignature);
		} catch (Exception e) {
			LOG.error("Cannot find service-resource:[" + llServiceResource + "]", e);
		}

		if (existing != null) {

			llServiceResource = existing;

			// Find those resources which associate with Low-Level resource

			List<Long> hlResourceIds = daoManager.getXXRMSResourceMapping().findByLlResource(llServiceResource);

			if (CollectionUtils.isEmpty(hlResourceIds)) {
				LOG.warn("Should not come here! higher-level resource cannot be absent from aclsync database for llResource:[" + llServiceResource + "]");
			} else {
				rmsNotification.setLlResourceId(llServiceResource.getId());

				for (Long hlResourceId : hlResourceIds) {
					RMSNotification copy = new RMSNotification(rmsNotification);

					copy.setHlResourceId(hlResourceId);
					copy.setOp(AppConstants.OP_DELETE);

					deleteXXResourceMapping(hlResourceId, llServiceResource.getId());

					if (createXXNotification(copy) == null) {
						LOG.error("Failed to create notification, failed notification:[" + copy + "]");
					}
				}
				ret = true;
			}
		} else {
			LOG.warn("Cannot find service-resource:[" + llServiceResource + "]");
		}

		if (!ret) {
			LOG.info("ResourceMgr.handleUpdateOp() not processed successfully, notification:[" + rmsNotification + "]");
		}
		LOG.debug("<== ResourceMgr.handleUpdateOp(notification:[" + rmsNotification + "]): ret:[" + ret + "]");

		return ret;

	}
	*/

	public ServiceResourceMappings getResourceMappingDeltas(XXService llService, XXService hlService, long lastKnownVersion, long serviceResourcesMappingVersionInDb) throws Exception {

		LOG.debug("==> ResourceMgr.getResourceMappingDeltas(llServiceName:[" + llService.getName() + "], hlServiceName:[" + hlService.getName() + "], lastKnownVersion:[" + lastKnownVersion + "], versionInDB:[" + serviceResourcesMappingVersionInDb + "])");

		ServiceResourceMappings ret = null;

		Map<Long, RangerServiceResource> serviceResources;
		Map<Long, List<Long>>            mappings = null;

		if (lastKnownVersion < 0L) {
			throw new Exception("Should not be called for getting all mappings, lastKnownVersion:[" + lastKnownVersion + "]");
		}
		// Find last notificationId which was invalid
		long latestInvalidNotificationId = getLatestInvalidNotificationId(llService.getId(), hlService.getId(), lastKnownVersion);

		if (lastKnownVersion <= latestInvalidNotificationId) {
			throw new Exception("Full sync was done after lastKnownVersion, latestInvalidNotificationId:[" + latestInvalidNotificationId + "], lastKnownVersion:[" + lastKnownVersion + "]");
		}
		// Get Map of added and deleted changes with hlResourceId as key
		Map<Long, XXRMSNotification> added = getAddedNotificationRecords(llService.getId(), hlService.getId(), lastKnownVersion);
		Map<Long, XXRMSNotification> deleted = getDeletedNotificationRecords(llService.getId(), hlService.getId(), lastKnownVersion);

		// Compute the set of net added and net deleted hlResourceIds
		if (LOG.isDebugEnabled()) {
			LOG.debug("Notifications of add type:[" + Arrays.toString(added.keySet().toArray()) + "]");
			LOG.debug("Notifications of add type:[" + Arrays.toString(deleted.keySet().toArray()) + "]");
		}

		Collection<Long> netAdded = CollectionUtils.subtract(added.keySet(), deleted.keySet());
		Collection<Long> netDeleted = CollectionUtils.subtract(deleted.keySet(), added.keySet());

		if (LOG.isDebugEnabled()) {
			LOG.debug("net-added notifications:[" + Arrays.toString(netAdded.toArray()) + "]");
			LOG.debug("net-deleted notifications:[" + Arrays.toString(netDeleted.toArray()) + "]");
		}

		if (CollectionUtils.isNotEmpty(netAdded) || CollectionUtils.isNotEmpty(netDeleted)) {
			serviceResources = new HashMap<>();

			// Add all deleted entries with signature set to null
			for (Long deletedHlResourceId : netDeleted) {

				RangerServiceResource deletedHlResource = new RangerServiceResource();
				deletedHlResource.setId(deletedHlResourceId);

				serviceResources.put(deletedHlResource.getId(), deletedHlResource);

				// This will in general be multiple entries.
				List<XXRMSNotification> deletedNotifications = daoManager.getXXRMSNotification().getDeletedNotificationsByHlResourceId(deletedHlResourceId, lastKnownVersion);
				for (XXRMSNotification deleteNotification : deletedNotifications) {
					RangerServiceResource deletedResource = new RangerServiceResource();
					deletedResource.setId(deleteNotification.getLlResourceId());
					serviceResources.put(deletedResource.getId(), deletedResource);

				}

			}

			// Add all netAdded hl-resources and corresponding mapping with ll-resource
			for (Long addedResourceId : netAdded) {
				XXRMSNotification entry = added.get(addedResourceId);

				// Fetch resources for both resources
				RangerServiceResource hlResource = getResourceById(addedResourceId);
				RangerServiceResource llResource = getResourceById(entry.getLlResourceId());

				if (hlResource == null || llResource == null) {
					throw new Exception("Cannot get resource with hlResourceId:[" + addedResourceId + "] or llResourceId:[" + entry.getLlResourceId() + "]");
				}

				serviceResources.put(llResource.getId(), llResource);
				serviceResources.put(hlResource.getId(), hlResource);

				if (mappings == null) {
					mappings = new HashMap<>();
				}

				List<Long> existing = mappings.get(llResource.getId());

				if (CollectionUtils.isEmpty(existing)) {
					existing = new ArrayList<>();
				}
				existing.add(hlResource.getId());
				mappings.put(llResource.getId(), existing);
			}

			ret = new ServiceResourceMappings();
			ret.setServiceResources(serviceResources);
			ret.setMappings(mappings);
			ret.setCurrentNotificationId(serviceResourcesMappingVersionInDb);
			ret.setPreviousNotificationId(lastKnownVersion);
			ret.setSourceServiceName(llService.getName());
			ret.setTargetServiceName(hlService.getName());
			ret.setMappingUpdateTime(new Date());
			ret.setIsDelta(true);

		}

		LOG.debug("<== ResourceMgr.getResourceMappingDeltas(llServiceName:[" + llService.getName() + "], hlServiceName:[" + hlService.getName() + "], lastKnownVersion:[" + lastKnownVersion + "], versionInDB:[" + serviceResourcesMappingVersionInDb + "]) : " + ret);

		return ret;
	}

	long getLatestInvalidNotificationId(long llServiceId, long hlServiceId, long lastKnownVersion) {
		// Search to get the notificationId which is max() of all records with operation = invalid
		XXRMSNotificationDao notificationDao = daoManager.getXXRMSNotification();
		return notificationDao.findLatestInvalidNotificationId(llServiceId, hlServiceId, lastKnownVersion );
	}

	Map<Long, XXRMSNotification> getAddedNotificationRecords(Long llServiceId, long hlServiceId, long lastKnownVersion) {
		return getNotificationRecords(llServiceId, hlServiceId, lastKnownVersion, "add");
	}

	Map<Long, XXRMSNotification> getDeletedNotificationRecords(Long llServiceId, long hlServiceId, long lastKnownVersion) {
		return getNotificationRecords(llServiceId, hlServiceId, lastKnownVersion, "delete");
	}

	Map<Long, XXRMSNotification> getNotificationRecords(Long llServiceId, long hlServiceId, long lastKnownVersion, String op) {

		Map<Long, XXRMSNotification> ret = new HashMap<>();

		List<XXRMSNotification> notifications = daoManager.getXXRMSNotification().getNotificationWithTypeAfterNotificationId(llServiceId, hlServiceId, op, lastKnownVersion);
		// Search notification table from lastKnownVersion onwards with specified criteria (llServiceId, hlServiceId and operation
		// put the notifications in the hash table with HlResourceId as key
		for (XXRMSNotification notification : notifications) {
			ret.put(notification.getHlResourceId(), notification);
		}

		return ret;
	}

	RangerServiceResource getResourceById(Long resourceId) {
		final RangerServiceResource ret;

		XXRMSServiceResource dbObj = daoManager.getXXRMSServiceResource().getById(resourceId);

		if (dbObj != null) {
			ret = XXRMSServiceResourceDao.populateViewBean(dbObj);
		} else {
			ret = null;
		}

		return ret;
	}

	private class MappingLoaderThread extends Thread {
		final TransactionTemplate txTemplate;
		final Long                sourceServiceId;

		Boolean                          isMappingsLoaded;
		Map<Long, RangerServiceResource> serviceResources = new HashMap<>();
		Map<Long, List<Long>>            mappings         = new HashMap<>();

		MappingLoaderThread(TransactionTemplate txTemplate, final Long sourceServiceId) {
			this.txTemplate      = txTemplate;
			this.sourceServiceId = sourceServiceId;
		}

		boolean getIsMappingsLoaded() {
			return isMappingsLoaded;
		}

		Map<Long, RangerServiceResource> getServiceResources() {
			return serviceResources;
		}

		Map<Long, List<Long>> getMappings() {
			return mappings;
		}

		@Override
		public void run() {
			try {
				txTemplate.setReadOnly(true);
				isMappingsLoaded = txTemplate.execute(status -> {
					boolean ret = initializeMappingCache(sourceServiceId);
					if (!ret) {
						status.setRollbackOnly();
						LOG.error("Failed to get mappings for source service:[" + sourceServiceId + "] in a new transaction");
					}
					return ret;
				});
				if (LOG.isDebugEnabled()) {
					LOG.debug("transaction result:[" + isMappingsLoaded +"]");
				}
			} catch (Throwable ex) {
				LOG.error("Failed to get mappings for source service:[" + sourceServiceId + "] in a new transaction", ex);
				isMappingsLoaded = false;
			}
		}

		private boolean initializeMappingCache(Long sourceServiceId) {
			LOG.debug("==> initializeMappingCache(sourceServiceId=" + sourceServiceId + ")");

			boolean ret;

			RangerPerfTracer perf = null;

			if (RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
				perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "ResourceMgr.createNotifications");
			}
			try {
				List<RangerServiceResource>      llResources      = daoManager.getXXRMSServiceResource().findByLlServiceId(sourceServiceId);

				LOG.debug("Retrieved :[" + llResources.size() + "] ll-resources");

				for (RangerServiceResource llResource : llResources) {

					List<RangerServiceResource> hlResources = daoManager.getXXRMSResourceMapping().getServiceResourcesByLlResourceId(llResource.getId());

					if (CollectionUtils.isNotEmpty(hlResources)) {

						serviceResources.put(llResource.getId(), llResource);

						hlResources.forEach((hlResource) -> serviceResources.putIfAbsent(hlResource.getId(), hlResource));

						mappings.put(llResource.getId(), hlResources.stream().map(RangerServiceResource::getId).collect(Collectors.toList()));
					}
				}

				ret = true;
			} catch (Exception ex) {
				LOG.error("Failed to get mappings for service:[" + sourceServiceId + "]", ex);
				serviceResources = null;
				mappings         = null;
				ret = false;
			}

			RangerPerfTracer.logAlways(perf);

			LOG.debug("<== initializeMappingCache(sourceServiceId=" + sourceServiceId + "): ret:[" + ret + "]");

			return ret;
		}
	}
}
