/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.rms.ha;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.ranger.rms.util.AppConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class HAConfiguration {

	private static final Logger LOG = LoggerFactory.getLogger(HAConfiguration.class);
	public static final String RANGER_RMS_SERVER_ZK_ROOT_DEFAULT = "/apache_ranger_rms_zkroot";

	private HAConfiguration() {
	}

	private static final String RANGER_RMS_SERVER_PREFIX                    = "ranger-rms.server.";
	private static final String RANGER_RMS_SERVER_HA_PREFIX                 = RANGER_RMS_SERVER_PREFIX + "ha.";
	private static final String ZOOKEEPER_PREFIX                            = "zookeeper.";
	public  static final String RANGER_RMS_SERVER_ADDRESS_PREFIX            = RANGER_RMS_SERVER_PREFIX + "address.";

	private static final String RANGER_RMS_SERVER_HA_ZK_ROOT_KEY            = RANGER_RMS_SERVER_HA_PREFIX + ZOOKEEPER_PREFIX + "zkroot";
	private static final String RANGER_RMS_SERVER_HA_ENABLED_KEY            = RANGER_RMS_SERVER_HA_PREFIX + "enabled";
	public  static final String RANGER_RMS_SERVER_IDS                       = RANGER_RMS_SERVER_PREFIX + "ids";
	private static final String HA_ZOOKEEPER_CONNECT                        = RANGER_RMS_SERVER_HA_PREFIX + ZOOKEEPER_PREFIX + "connect";
	private static final int    DEFAULT_ZOOKEEPER_CONNECT_SLEEPTIME_MILLIS  = 1000;
	private static final String HA_ZOOKEEPER_RETRY_SLEEPTIME_MILLIS         = RANGER_RMS_SERVER_HA_PREFIX + ZOOKEEPER_PREFIX + "retry.sleeptime.ms";
	private static final String HA_ZOOKEEPER_NUM_RETRIES                    = RANGER_RMS_SERVER_HA_PREFIX + ZOOKEEPER_PREFIX + "num.retries";
	private static final int    DEFAULT_ZOOKEEPER_CONNECT_NUM_RETRIES       = 3;
	private static final String HA_ZOOKEEPER_SESSION_TIMEOUT_MS             = RANGER_RMS_SERVER_HA_PREFIX + ZOOKEEPER_PREFIX + "session.timeout.ms";
	private static final int    DEFAULT_ZOOKEEPER_SESSION_TIMEOUT_MILLIS    = 20000;
	private static final String HA_ZOOKEEPER_ACL                            = RANGER_RMS_SERVER_HA_PREFIX + ZOOKEEPER_PREFIX + "acl";
	private static final String HA_ZOOKEEPER_AUTH                           = RANGER_RMS_SERVER_HA_PREFIX + ZOOKEEPER_PREFIX + "auth";

	/**
	 * Return whether HA is enabled or not.
	 * 
	 * @param configuration underlying configuration instance
	 * @return true if more than 1 ids, false otherwise
	 */
	public static boolean isHAEnabled(Configuration configuration) {
		LOG.info("RANGER_RMS_SERVER_HA_ENABLED_KEY :  " + configuration.get(HAConfiguration.RANGER_RMS_SERVER_HA_ENABLED_KEY));

		final boolean ret;

		if (configuration.get(HAConfiguration.RANGER_RMS_SERVER_HA_ENABLED_KEY) != null) {
			ret = configuration.getBoolean(RANGER_RMS_SERVER_HA_ENABLED_KEY, false);
		} else {
			String[] ids = configuration.getStrings(HAConfiguration.RANGER_RMS_SERVER_IDS);

			ret = ids != null && ids.length > 1;
		}

		LOG.info("inside HAConfiguration isHAEnabled() ret :" + ret);

		return ret;
	}

	/**
	 * Get the web server address that a server instance with the passed ID is bound
	 * to.
	 *
	 *
	 * @param configuration underlying configuration
	 * @param serverId      serverId whose host:port property is picked to build the
	 *                      web server address.
	 * @return address string
	 */
	public static String getBoundAddressForId(Configuration configuration, String serverId) {

		String  hostPort = configuration.get(RANGER_RMS_SERVER_ADDRESS_PREFIX + serverId);
		boolean isSecure = configuration.getBoolean(AppConstants.PROP_RMS_SERVICE_SSL_ENABLED, false);
		String  protocol = (isSecure) ? "https://" : "http://";

		return protocol + hostPort;
	}

	public static List<String> getServerInstances(Configuration configuration) {
		String[]     serverIds       = configuration.getStrings(RANGER_RMS_SERVER_IDS);
		List<String> serverInstances = new ArrayList<>(serverIds.length);

		for (String serverId : serverIds) {
			serverInstances.add(getBoundAddressForId(configuration, serverId));
		}

		return serverInstances;
	}

	/**
	 * A collection of Zookeeper specific configuration that is used by High
	 * Availability code.
	 */
	public static class ZookeeperProperties {
		private String connectString;
		private String zkRoot;
		private int    retriesSleepTimeMillis;
		private int    numRetries;
		private int    sessionTimeout;
		private String acl;
		private String auth;

		public ZookeeperProperties(String connectString, String zkRoot, int retriesSleepTimeMillis, int numRetries, int sessionTimeout, String acl, String auth) {
			this.connectString          = connectString;
			this.zkRoot                 = zkRoot;
			this.retriesSleepTimeMillis = retriesSleepTimeMillis;
			this.numRetries             = numRetries;
			this.sessionTimeout         = sessionTimeout;
			this.acl                    = acl;
			this.auth                   = auth;
		}

		public String getConnectString() {
			return connectString;
		}

		public int getRetriesSleepTimeMillis() {
			return retriesSleepTimeMillis;
		}

		public int getNumRetries() {
			return numRetries;
		}

		public int getSessionTimeout() {
			return sessionTimeout;
		}

		public String getAcl() {
			return acl;
		}

		public String getAuth() {
			return auth;
		}

		public String getZkRoot() {
			return zkRoot;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (o == null || getClass() != o.getClass())
				return false;

			ZookeeperProperties that = (ZookeeperProperties) o;

			return retriesSleepTimeMillis == that.retriesSleepTimeMillis && numRetries == that.numRetries
					&& sessionTimeout == that.sessionTimeout && Objects.equals(connectString, that.connectString)
					&& Objects.equals(zkRoot, that.zkRoot) && Objects.equals(acl, that.acl)
					&& Objects.equals(auth, that.auth);
		}

		@Override
		public int hashCode() {
			return Objects.hash(connectString, zkRoot, retriesSleepTimeMillis, numRetries, sessionTimeout, acl, auth);
		}

		public boolean hasAcl() {
			return getAcl() != null;
		}

		public boolean hasAuth() {
			return getAuth() != null;
		}
	}

	public static ZookeeperProperties getZookeeperProperties(Configuration configuration) {
		String[] zkServers;

		if (configuration.get(HA_ZOOKEEPER_CONNECT) != null) {
			zkServers = configuration.getStrings(HA_ZOOKEEPER_CONNECT);
		} else {
			zkServers = configuration.getStrings("rms.kafka." + ZOOKEEPER_PREFIX + "connect");
		}

		String zkRoot                 = configuration.get(RANGER_RMS_SERVER_HA_ZK_ROOT_KEY, RANGER_RMS_SERVER_ZK_ROOT_DEFAULT);
		int    retriesSleepTimeMillis = configuration.getInt(HA_ZOOKEEPER_RETRY_SLEEPTIME_MILLIS, DEFAULT_ZOOKEEPER_CONNECT_SLEEPTIME_MILLIS);
		int    numRetries             = configuration.getInt(HA_ZOOKEEPER_NUM_RETRIES, DEFAULT_ZOOKEEPER_CONNECT_NUM_RETRIES);
		int    sessionTimeout         = configuration.getInt(HA_ZOOKEEPER_SESSION_TIMEOUT_MS, DEFAULT_ZOOKEEPER_SESSION_TIMEOUT_MILLIS);
		String acl                    = configuration.get(HA_ZOOKEEPER_ACL);
		String auth                   = configuration.get(HA_ZOOKEEPER_AUTH);

		LOG.info(" ===>ZookeeperProperties zkServers: " + Arrays.toString(zkServers) + " zkRoot: " + zkRoot + " retriesSleepTimeMillis: "
				+ retriesSleepTimeMillis + " numRetries: " + numRetries + " sessionTimeout: " + sessionTimeout
				+ " acl: " + acl + " auth: " + auth);

		return new ZookeeperProperties(StringUtils.join(zkServers, ','), zkRoot, retriesSleepTimeMillis, numRetries, sessionTimeout, acl, auth);
	}
}
