/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.ranger.rms.rest;

import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;

import org.apache.commons.logging.Log;
import org.apache.ranger.common.ServiceResourceMappingsCache;
import org.apache.ranger.entity.XXService;
import org.apache.ranger.plugin.util.RangerPerfTracer;
import org.apache.ranger.plugin.util.ServiceResourceMappings;
import org.apache.ranger.rms.biz.ResourceMgr;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import java.util.Date;

@Path("plugins")
@Component
@Scope("request")
@Transactional(propagation = Propagation.REQUIRES_NEW)
public class SyncREST {

	@Autowired
	ResourceMgr resourceMgr;

	@Autowired
	ServiceResourceMappingsCache serviceResourceMappingsCache;

	public static final Logger LOG = LoggerFactory.getLogger(SyncREST.class);
	private static final Log PERF_LOG = RangerPerfTracer.getPerfLogger("rest.SyncREST");

	@GET
	@Path("/downloadmappings/from/{sourceServiceName}/to/{targetServiceName}")

	@Produces({ "application/json", "application/xml" })
	public ServiceResourceMappings getServiceResourceMappingsIfUpdated(@PathParam("sourceServiceName") String sourceServiceName, @PathParam("targetServiceName") String targetServiceName, @QueryParam("lastKnownVersion") Long lastKnownVersion,
			@DefaultValue("0") @QueryParam("lastActivationTimeInMillis") Long lastActivationTimeInMillis, @DefaultValue("true") @QueryParam("supportsMappingDeltas") Boolean supportsMappingDeltas, @Context HttpServletRequest request) throws Exception {

		LOG.debug("==> SyncREST.getServiceResourceMappingsIfUpdated(sourceServiceName=" + sourceServiceName + ", targetServiceName="+ targetServiceName + ", lastKnownVersion=" + lastKnownVersion + ", lastActivationTimeInMillis=" + lastActivationTimeInMillis + ", supportsMappingDeltas=" + supportsMappingDeltas + ")");

		final ServiceResourceMappings ret;

		RangerPerfTracer perf = null;

		if (RangerPerfTracer.isPerfTraceEnabled(PERF_LOG)) {
			perf = RangerPerfTracer.getPerfTracer(PERF_LOG, "SyncREST.getServiceResourceMappingsIfUpdated(lastKnownVersion=" + lastKnownVersion + ")");
		}

		XXService sourceService = ResourceMgr.getServiceByName(sourceServiceName);
		XXService targetService = ResourceMgr.getServiceByName(targetServiceName);

		if (sourceService == null || targetService == null) {

			LOG.error("Invalid source or resource serviceName! sourceServiceName:[" + sourceServiceName + "], targetServiceName:[" + targetServiceName + "], returning empty mappings");
			ret = new ServiceResourceMappings(ServiceResourceMappings.OP_ADD, sourceServiceName, targetServiceName, -1L, lastKnownVersion, new Date(), null, null);

		} else {

			if (lastKnownVersion == null || lastKnownVersion <= 0) {
				lastKnownVersion = -1L;
			}

			ServiceResourceMappings fromCache = serviceResourceMappingsCache.getServiceResourceMappings(sourceService, targetService, lastKnownVersion, !supportsMappingDeltas);

			if (fromCache != null && fromCache.getCurrentNotificationId().equals(lastKnownVersion)) {
				// No change
				ret = null;
			} else {
				ret = fromCache;
			}
		}

		RangerPerfTracer.log(perf);

		LOG.debug("<== SyncREST.getServiceResourceMappingsIfUpdated(sourceServiceName=" + sourceServiceName + ", targetServiceName=" + targetServiceName + ", lastKnownVersion=" + lastKnownVersion + ", lastActivationTimeInMillis=" + lastActivationTimeInMillis + ", supportsMappingDeltas=" + supportsMappingDeltas + "): ret={}", ret);

		return ret;
	}

}
