/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.rms.server;

import java.io.File;
import java.io.IOException;
import java.security.KeyStore;
import java.security.PrivilegedAction;
import java.util.*;

import org.apache.catalina.Context;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.WebResourceRoot;
import org.apache.catalina.connector.Connector;
import org.apache.catalina.core.StandardContext;
import org.apache.catalina.startup.Tomcat;
import org.apache.catalina.valves.AccessLogValve;
import org.apache.catalina.webresources.StandardRoot;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.security.SecureClientLogin;
import org.apache.hadoop.security.alias.BouncyCastleFipsKeyStoreProvider;
import org.apache.hadoop.security.alias.CredentialProvider;
import org.apache.hadoop.security.alias.CredentialProviderFactory;
import org.apache.hadoop.security.alias.JavaKeyStoreProvider;
import org.apache.hadoop.security.alias.LocalBouncyCastleFipsKeyStoreProvider;
import org.apache.log4j.PropertyConfigurator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.security.auth.Subject;


public class EmbeddedServer {
    private static final Logger LOG = LoggerFactory.getLogger(EmbeddedServer.class);

    public  static final String RANGER_RMS_KERBEROS_KEYTAB              = "kerberos.keytab";
    public  static final String RANGER_RMS_KERBEROS_PRINCIPAL           = "kerberos.principal";
    public  static final String HADOOP_KERBEROS_NAME_RULES              = "hadoop.security.auth_to_local";
    public  static final String HADOOP_AUTHENTICATION_TYPE              = "hadoop.security.authentication";
    public  static final String HADOOP_AUTH_TYPE_KERBEROS               = "kerberos";
    private static final String ACCESS_LOG_PREFIX                       = "accesslog.prefix";
    private static final String ACCESS_LOG_DATE_FORMAT                  = "accesslog.dateformat";
    private static final String ACCESS_LOG_PATTERN                      = "accesslog.pattern";
    private static final String ACCESS_LOG_ROTATE_MAX_DAYS              = "accesslog.rotate.max.days";


    private static final String DEFAULT_LOG_DIR                         = "/tmp";
    private static final int    DEFAULT_HTTP_PORT                       = 8383;
    private static final int    DEFAULT_HTTPS_PORT                      = -1;
    protected static final int  DEFAULT_SHUTDOWN_PORT                   = 8388;
    private static final String DEFAULT_SHUTDOWN_COMMAND                = "SHUTDOWN";
    private static final String DEFAULT_WEBAPPS_ROOT_FOLDER             = "webapps";
    private static final String DEFAULT_ENABLED_PROTOCOLS               = "TLSv1.2";
    public  static final String DEFAULT_NAME_RULE                       = "DEFAULT";
    public static final String KEYSTORE_FILE_TYPE_DEFAULT               = KeyStore.getDefaultType();
    public static final String TRUSTSTORE_FILE_TYPE_DEFAULT             = KeyStore.getDefaultType();

    /* These are the configurations prefixed with ranger-rms. that are defined in ranger-rms-site.xml
    service.host
    service.http.port
    service.shutdown.port
            contextName
    service.https.port
    service.https.attrib.ssl.enabled
    service.https.attrib.client.auth
    credential.provider.path
    service.https.attrib.keystore.file
    service.https.attrib.keystore.keyalias
    service.https.attrib.keystore.pass
    service.https.attrib.keystore.credential.alias
    keystore.file
    keystore.alias
    keystore.password
    truststore.file
    truststore.alias
    truststore.password
    kerberos.principal
    kerberos.keytab
    jdbc.sqlconnectorjar
    db.flavor
    db.ssl.enabled
    db.ssl.required
    db.ssl.verifyServerCertificate
    db.ssl.auth.type
    jpa.jdbc.url
    jpa.jdbc.user
    jpa.jdbc.password
    jpa.jdbc.credential.provider.path
    jpa.jdbc.credential.alias
    jpa.jdbc.dialect
    jpa.jdbc.driver
    jpa.showsql
    jpa.jdbc.maxpoolsize
    jpa.jdbc.minpoolsize
    jpa.jdbc.initialpoolsize
    jpa.jdbc.maxidletime
    jpa.jdbc.maxstatements
    jpa.jdbc.preferredtestquery
    jpa.jdbc.idleconnectiontestperiod
    server.ha.enabled
    server.ha.zookeeper.zkroot
    server.ha.zookeeper.connect
    server.ids
    server.address.id1
    server.address.id2
    server.address.id3
    server.ha.zookeeper.retry.sleeptime.ms
    server.ha.zookeeper.session.timeout.ms
    server.ha.zookeeper.num.retries
    server.ha.zookeeper.acl
    server.ha.zookeeper.acl.auth

     */

    private final Configuration configuration;
    private final String        appName;
    private final String        configPrefix;

    public static void main(String[] args) {
        Configuration config       = RMSConfig.getInstance();
        String        appName      = "ranger-rms";
        String        configPrefix = "ranger-rms.";

        new EmbeddedServer(config, appName, configPrefix).start();
    }

    public EmbeddedServer(Configuration configuration, String appName, String configPrefix) {
        LOG.info("==> EmbeddedServer(appName={}, configPrefix={})", appName, configPrefix);

        try {
			// load log configuration file dynamically if log4j.properties changed
			if (StringUtils.isNotBlank(System.getProperty("log4j.configuration"))) {
				String logPropFile = StringUtils.splitByWholeSeparator(System.getProperty("log4j.configuration"), ":")[1];
				PropertyConfigurator.configureAndWatch(logPropFile, 10000L);
			}
		} catch (Exception ignored) {
			LOG.warn("Failed to get log4j.configuration  Reason: " + ignored.toString());
		}

        this.configuration = configuration;
        this.appName       = appName;
        this.configPrefix  = configPrefix;

        LOG.info("<== EmbeddedServer(appName={}, configPrefix={}", appName, configPrefix);
    }

    public void start() {
        LOG.info("==> EmbeddedServer.start(appName={})", appName);

        final Tomcat server = new Tomcat();

        String  logDir          = getConfig("logdir", DEFAULT_LOG_DIR);
        String  hostName        = getConfig("service.host");
        int     serverPort      = getIntConfig("service.http.port", DEFAULT_HTTP_PORT);
        int     sslPort         = getIntConfig("service.https.port", DEFAULT_HTTPS_PORT);
        int     shutdownPort    = getIntConfig("service.shutdown.port", DEFAULT_SHUTDOWN_PORT);
        String  shutdownCommand = getConfig("service.shutdown.command", DEFAULT_SHUTDOWN_COMMAND);
        boolean isHttpsEnabled  = getBooleanConfig("service.https.attrib.ssl.enabled", false);
        boolean ajpEnabled      = getBooleanConfig("ajp.enabled", false);

        server.setHostname(hostName);
        server.setPort(serverPort);
        server.getServer().setPort(shutdownPort);
        server.getServer().setShutdown(shutdownCommand);

        if (ajpEnabled) {
            Connector ajpConnector = new Connector("org.apache.coyote.ajp.AjpNioProtocol");

            ajpConnector.setPort(serverPort);
            ajpConnector.setProperty("protocol", "AJP/1.3");

            server.getService().addConnector(ajpConnector);

            // Making this as a default connector
            server.setConnector(ajpConnector);

            LOG.info("Created AJP Connector");
        } else if (isHttpsEnabled && sslPort > 0) {
            String clientAuth              = getConfig("service.https.attrib.clientAuth", "false");
            String providerPath            = getConfig("credential.provider.path");
            String keyAlias                = getConfig("service.https.attrib.keystore.credential.alias", "keyStoreCredentialAlias");
            String keystorePass            = null;
            String enabledProtocols        = getConfig("service.https.attrib.ssl.enabled.protocols", DEFAULT_ENABLED_PROTOCOLS);
            String ciphers                 = getConfig("tomcat.ciphers");

            if (StringUtils.equalsIgnoreCase(clientAuth, "false")) {
                clientAuth = getConfig("service.https.attrib.client.auth", "want");
            }

            if (providerPath != null && keyAlias != null) {
                keystorePass = getDecryptedString(providerPath.trim(), keyAlias.trim(), getConfig("ranger-rms.keystore.file.type", KEYSTORE_FILE_TYPE_DEFAULT));

                if (StringUtils.isBlank(keystorePass) || StringUtils.equalsIgnoreCase(keystorePass.trim(), "none")) {
                    keystorePass = getConfig("service.https.attrib.keystore.pass");
                }
            }

            Connector ssl = new Connector();
            String sslKeystoreKeyAlias = getConfig("service.https.attrib.keystore.keyalias","");

            ssl.setPort(sslPort);
            ssl.setSecure(true);
            ssl.setScheme("https");
            ssl.setAttribute("SSLEnabled", "true");
            ssl.setAttribute("sslProtocol", getConfig("service.https.attrib.ssl.protocol", "TLSv1.2"));
            ssl.setAttribute("clientAuth",clientAuth);
            if (StringUtils.isNotBlank(sslKeystoreKeyAlias)) {
                ssl.setAttribute("keyAlias", sslKeystoreKeyAlias);
            }
            ssl.setAttribute("keystorePass", keystorePass);
            ssl.setAttribute("keystoreFile", getKeystoreFile());
            ssl.setAttribute("sslEnabledProtocols", enabledProtocols);
            ssl.setAttribute("keystoreType", getConfig("ranger-rms.keystore.file.type", KEYSTORE_FILE_TYPE_DEFAULT));
            ssl.setAttribute("truststoreType", getConfig("ranger-rms.truststore.file.type", TRUSTSTORE_FILE_TYPE_DEFAULT));

            if (StringUtils.isNotBlank(ciphers)) {
                ssl.setAttribute("ciphers", ciphers);
            }

            server.getService().addConnector(ssl);

            //
            // Making this as a default connector
            //
            server.setConnector(ssl);
        }

        updateHttpConnectorAttribConfig(server);

        File logDirectory = new File(logDir);

        if (!logDirectory.exists()) {
            logDirectory.mkdirs();
        }

        String logPattern = getConfig(ACCESS_LOG_PATTERN, "%h %l %u %t \"%r\" %s %b");

        AccessLogValve valve = new AccessLogValve();

        valve.setRotatable(true);
        valve.setAsyncSupported(true);
        valve.setBuffered(false);
        valve.setEnabled(true);
        valve.setPrefix(getConfig(ACCESS_LOG_PREFIX,"access_log-" + hostName +"-"));
        valve.setFileDateFormat(getConfig(ACCESS_LOG_DATE_FORMAT, "yyyy-MM-dd.HH"));
        valve.setDirectory(logDirectory.getAbsolutePath());
        valve.setSuffix(".log");
        valve.setPattern(logPattern);
        valve.setMaxDays(getIntConfig(ACCESS_LOG_ROTATE_MAX_DAYS, 15));

        server.getHost().getPipeline().addValve(valve);

        try {
            String webappDir = getConfig("webapp.dir");

            if (webappDir == null || webappDir.trim().isEmpty()) {
                // If webapp location property is not set, then let's derive
                // from catalina_base
                String catalinaBase = getConfig("catalina.base");
                if (catalinaBase == null || catalinaBase.trim().isEmpty()) {
                    LOG.error("Tomcat Server failed to start: catalina.base and/or webapp.dir is not set");
                    System.exit(1);
                }
                webappDir = catalinaBase + File.separator + "webapp";
                LOG.info("Deriving webapp folder from catalina.base property. folder=" + webappDir);
            }

            String webContextName = getConfig("contextName", "");

            if (StringUtils.isBlank(webContextName)) {
                webContextName = "";
            } else if (!webContextName.startsWith("/")) {
                LOG.info("Context Name [{}] is being loaded as [ /{}]", webContextName, webContextName);

                webContextName = "/" + webContextName;
            }

            File wad = new File(webappDir);

            if (wad.isDirectory()) {
                LOG.info("Webapp dir={}, webAppName={}", webappDir, webContextName);
            } else if (wad.isFile()) {
                File webAppDir = new File(DEFAULT_WEBAPPS_ROOT_FOLDER);

                if (!webAppDir.exists()) {
                    webAppDir.mkdirs();
                }

                LOG.info("Webapp file={}, webAppName={}", webappDir, webContextName);
            }

            LOG.info("Adding webapp [{}] = path [{}] .....", webContextName, webappDir);

            Context webappCtx = server.addWebapp(webContextName, new File(webappDir).getAbsolutePath());
            if (webappCtx instanceof StandardContext) {
                boolean allowLinking = getBooleanConfig("allow.linking", true);
                StandardContext standardContext = (StandardContext) webappCtx;
                String workDirPath = getConfig("tomcat.work.dir", "");
                if (!workDirPath.isEmpty() && new File(workDirPath).exists()) {
                    standardContext.setWorkDir(workDirPath);
                } else {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Skipping to set tomcat server work directory, '" + workDirPath +
                            "', as it is blank or directory does not exist.");
                    }
                }
                WebResourceRoot resRoot = new StandardRoot(standardContext);
                webappCtx.setResources(resRoot);
                webappCtx.getResources().setAllowLinking(allowLinking);
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Tomcat Configuration - allowLinking=[{}]", allowLinking);
                }
            } else {
                LOG.error("Tomcat Context [{}] is either NULL OR it's NOT instanceof StandardContext", webappCtx);
            }

            webappCtx.init();

            LOG.info("Finished init of webapp [{}] = path [{}].", webContextName, webappDir);
        } catch (LifecycleException lce) {
            LOG.error("Tomcat Server failed to start webapp", lce);

            lce.printStackTrace();
        }

        String authenticationType = getConfig(HADOOP_AUTHENTICATION_TYPE);

        if (StringUtils.equalsIgnoreCase(authenticationType, HADOOP_AUTH_TYPE_KERBEROS)) {
            String keyTab    = getConfig(RANGER_RMS_KERBEROS_KEYTAB);
            String principal = null;
            String nameRules = getConfig(HADOOP_KERBEROS_NAME_RULES, DEFAULT_NAME_RULE);

            try {
                principal = SecureClientLogin.getPrincipal(getConfig(RANGER_RMS_KERBEROS_PRINCIPAL), hostName);
            } catch (IOException excp) {
                LOG.warn("Failed to get principal" + excp);
            }

            if (SecureClientLogin.isKerberosCredentialExists(principal, keyTab)) {
                try {
                    LOG.info("Logging in with Kerberos: principal={}, keytab={}, nameRules={}", principal, keyTab, nameRules);

                    Subject sub = SecureClientLogin.loginUserFromKeytab(principal, keyTab, nameRules);

                    Subject.doAs(sub, new PrivilegedAction<Void>() {
                        @Override
                        public Void run() {
                            LOG.info("Starting Server using kerberos credential");

                            startServer(server);

                            return null;
                        }
                    });
                } catch (Exception excp) {
                    LOG.error("Tomcat Server failed to start", excp);
                }
            } else {
                LOG.warn("Kerberos principal={} not found in keytab={}. Starting server in non-kerberos mode", principal, keyTab);

                startServer(server);
            }
        } else {
            LOG.info("Starting server in non-kerberos mode");

            startServer(server);
        }

        LOG.info("<== EmbeddedServer.start(appName={})", appName);
    }


    private void startServer(final Tomcat server) {
        LOG.info("==> EmbeddedServer.startServer(appName={})", appName);

        try {
            server.start();

            server.getServer().await();

            shutdownServer();
        } catch (LifecycleException e) {
            LOG.error("Tomcat Server failed to start", e);

            e.printStackTrace();
        } catch (Exception e) {
            LOG.error("Tomcat Server failed to start", e);

            e.printStackTrace();
        }

        LOG.info("<== EmbeddedServer.startServer(appName={})", appName);
    }

    private void shutdownServer() {
        LOG.info("==> EmbeddedServer.shutdownServer(appName={})", appName);

        int timeWaitForShutdownInSeconds = getIntConfig("service.waitTimeForForceShutdownInSeconds", 0);

        if (timeWaitForShutdownInSeconds > 0) {
            long endTime = System.currentTimeMillis() + (timeWaitForShutdownInSeconds * 1000L);

            LOG.info("Will wait for all threads to shutdown gracefully. Final shutdown Time: " + new Date(endTime));

            while (System.currentTimeMillis() < endTime) {
                int activeCount = Thread.activeCount();

                if (activeCount == 0) {
                    LOG.info("Number of active threads = " + activeCount + ".");

                    break;
                } else {
                    LOG.info("Number of active threads = {}. Waiting for all threads to shutdown ...", activeCount);

                    try {
                        Thread.sleep(5000L);
                    } catch (InterruptedException e) {
                        LOG.warn("shutdownServer process is interrupted with exception", e);

                        break;
                    }
                }
            }
        }

        LOG.info("Shutting down the Server.");

        LOG.info("<== EmbeddedServer.shutdownServer(appName={})", appName);

        System.exit(0);
    }

    private void updateHttpConnectorAttribConfig(Tomcat server) {
        server.getConnector().setAllowTrace(getBooleanConfig("service.http.connector.attrib.allowTrace", false));
        server.getConnector().setAsyncTimeout(getLongConfig("service.http.connector.attrib.asyncTimeout", 10000));
        server.getConnector().setEnableLookups(getBooleanConfig("service.http.connector.attrib.enableLookups", false));
        server.getConnector().setMaxParameterCount(getIntConfig("service.http.connector.attrib.maxParameterCount", 10000));
        server.getConnector().setMaxPostSize(getIntConfig("service.http.connector.attrib.maxPostSize", 2097152));
        server.getConnector().setMaxSavePostSize(getIntConfig("service.http.connector.attrib.maxSavePostSize", 4096));
        server.getConnector().setParseBodyMethods(getConfig("service.http.connector.attrib.methods", "POST"));
        server.getConnector().setURIEncoding(getConfig("service.http.connector.attrib.URIEncoding", "UTF-8"));
        server.getConnector().setXpoweredBy(false);
        server.getConnector().setAttribute("server", getConfig("servername", "Apache Ranger"));
        server.getConnector().setProperty("sendReasonPhrase",getConfig(configPrefix + "service.http.connector.property.sendReasonPhrase", "true"));

        for (Iterator<Map.Entry<String, String>> iterator = configuration.iterator(); iterator.hasNext(); ) {
            Map.Entry<String, String> entry = iterator.next();
            String                    key   = entry.getKey();
            String                    value = entry.getValue();

            if (key != null && key.startsWith(configPrefix + "service.http.connector.property.")) {
                String property = key.replace(configPrefix + "service.http.connector.property.", "");

                server.getConnector().setProperty(property, value);

                LOG.info("{}:{}", property, server.getConnector().getProperty(property));
            }
        }
    }

    public String getDecryptedString(String crendentialProviderPath, String alias, String storeType) {
        String ret = null;

        if (StringUtils.isNotBlank(crendentialProviderPath) && StringUtils.isNotBlank(alias)) {
            try {
                Configuration conf             = new Configuration();
                String        prefixJceks      = JavaKeyStoreProvider.SCHEME_NAME + "://file";
                String        prefixLocalJceks = "localjceks://file";
                String        prefixBcfks      = BouncyCastleFipsKeyStoreProvider.SCHEME_NAME + "://file";
                String        prefixLocalBcfks = LocalBouncyCastleFipsKeyStoreProvider.SCHEME_NAME + "://file";

                if (StringUtils.startsWithIgnoreCase(crendentialProviderPath, prefixJceks) ||
						StringUtils.startsWithIgnoreCase(crendentialProviderPath, prefixLocalJceks) ||
						StringUtils.startsWithIgnoreCase(crendentialProviderPath, prefixBcfks) ||
						StringUtils.startsWithIgnoreCase(crendentialProviderPath, prefixLocalBcfks)) {
                    conf.set(CredentialProviderFactory.CREDENTIAL_PROVIDER_PATH, crendentialProviderPath);
                } else {
                    if (crendentialProviderPath.startsWith("/")) {
						if (StringUtils.equalsIgnoreCase(storeType, "bcfks")) {
							conf.set(CredentialProviderFactory.CREDENTIAL_PROVIDER_PATH, crendentialProviderPath);
						} else {
							conf.set(CredentialProviderFactory.CREDENTIAL_PROVIDER_PATH, prefixJceks + crendentialProviderPath);
						}
                    } else {
                        conf.set(CredentialProviderFactory.CREDENTIAL_PROVIDER_PATH, prefixJceks + "/" + crendentialProviderPath);
                    }
                }

                List<CredentialProvider> providers = CredentialProviderFactory.getProviders(conf);
                if (LOG.isDebugEnabled()) {
                    LOG.debug("crendentialProviderPath="+ crendentialProviderPath + " alias="+ alias + " storeType="+ storeType);
                    LOG.debug("List of CredentialProvider = " + providers.toString());
                }

                for (CredentialProvider provider : providers) {
                    List<String> aliasesList = provider.getAliases();

                    if (CollectionUtils.isNotEmpty(aliasesList) && aliasesList.contains(alias)) {
                        CredentialProvider.CredentialEntry credEntry = provider.getCredentialEntry(alias);
                        char[]                             pass      = credEntry.getCredential();

                        if (pass != null && pass.length > 0) {
                            ret = String.valueOf(pass);

                            break;
                        }
                    }
                }
            } catch (Exception ex) {
                LOG.error("CredentialReader failed while decrypting provided string. Reason: " + ex.toString());

                ret = null;
            }
        }

        if (LOG.isDebugEnabled()) {
			LOG.debug("getDecryptedString() : ret = " + ret);
        }
        return ret;
    }

    private String getKeystoreFile() {
        String ret = getConfig("service.https.attrib.keystore.file");

        if (StringUtils.isBlank(ret)) {
            // new property not configured, lets use the old property
            ret = getConfig("https.attrib.keystore.file");
        }

        return ret;
    }

    private String getConfig(String key) {
        String propertyWithPrefix = configPrefix + key;
        String value              = configuration.get(propertyWithPrefix);

        if (value == null) { // config-with-prefix not found; look at System properties
            value = System.getProperty(propertyWithPrefix);

            if (value == null) { // config-with-prefix not found in System properties; look for config-without-prefix
                value = configuration.get(key);

                if (value == null) { // config-without-prefix not found; look at System properties
                    value = System.getProperty(key);
                }
            }
        }

        return value;
    }

    protected String getConfig(String key, String defaultValue) {
        String ret = getConfig(key);

        if (ret == null) {
            ret = defaultValue;
        }

        return ret;
    }

    protected int getIntConfig(String key, int defaultValue) {
        int    ret      = defaultValue;
        String strValue = getConfig(key);

        try {
            if (strValue != null) {
                ret = Integer.parseInt(strValue);
            }
        } catch (Exception err) {
            LOG.warn(strValue + " can't be parsed to int. Reason: " + err.toString());
        }

        return ret;
    }

    protected long getLongConfig(String key, long defaultValue) {
        long   ret    = defaultValue;
        String retStr = getConfig(key);

        try {
            if (retStr != null) {
                ret = Long.parseLong(retStr);
            }
        } catch(Exception err){
            LOG.warn(retStr + " can't be parsed to long. Reason: " + err.toString());
        }

        return ret;
    }

    private boolean getBooleanConfig(String key, boolean defaultValue) {
        boolean ret      = defaultValue;
        String  strValue = getConfig(key);

        if (StringUtils.isNotBlank(strValue)) {
            ret = Boolean.valueOf(strValue);
        }

        return ret;
    }
}
