/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.patch.cliutil;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.StringReader;
import java.nio.ByteBuffer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Scanner;
import java.util.Set;

import javax.security.auth.Subject;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.security.SecureClientLogin;
import org.apache.http.HttpEntity;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.util.EntityUtils;
import org.apache.ranger.authorization.utils.StringUtil;
import org.apache.ranger.common.PropertiesUtil;
import org.apache.ranger.patch.BaseLoader;
import org.apache.ranger.util.CLIUtil;
import org.apache.solr.client.solrj.SolrClient;
import org.apache.solr.client.solrj.impl.CloudSolrClient;
import org.apache.solr.client.solrj.impl.HttpClientUtil;
import org.apache.solr.client.solrj.impl.HttpSolrClient;
import org.apache.solr.client.solrj.impl.Krb5HttpClientBuilder;
import org.apache.solr.client.solrj.impl.SolrHttpClientBuilder;
import org.apache.solr.client.solrj.request.CollectionAdminRequest;
import org.apache.solr.client.solrj.request.ConfigSetAdminRequest;
import org.apache.solr.client.solrj.response.CollectionAdminResponse;
import org.apache.solr.common.cloud.SolrZkClient;
import org.apache.solr.common.cloud.ZkConfigManager;
import org.noggit.JSONParser;
import org.noggit.ObjectBuilder;
import org.springframework.stereotype.Component;

@Component
public class SolrStorageMigrationUtility extends BaseLoader {

	private static final String SOLR_ZK_HOSTS = "ranger.audit.solr.zookeepers";
	private static final String ADMIN_USER_KEYTAB = "ranger.admin.kerberos.keytab";
	private static final String ADMIN_USER_PRINCIPAL = "ranger.admin.kerberos.principal";
	private static final String ADMIN_NAME_RULES = "hadoop.security.auth_to_local";
	private static final String SSL_ENABLED_PARAM = "ranger.service.https.attrib.ssl.enabled";
	private static final String RANGER_ADMIN_HOME = "ranger.admin.home";
	private final static String CONFIG_SET_LOCATION = "ranger.audit.solr.configset.location";
	private final static String SOLR_COLLECTION_NAME = "ranger.audit.solr.collection.name";
	private static final String DEFAULT_COLLECTION_NAME = "ranger_audits";
	private final static String SOLR_CONFIG_NAME = "ranger.audit.solr.config.name";
	private static final String DEFAULT_CONFIG_NAME = "ranger_audits";
	private static final String AUTHENTICATION_TYPE = "hadoop.security.authentication";
	private static final String AUTH_TYPE_KERBEROS = "kerberos";
	private static final String SOLR_USER = "ranger.solr.audit.user";
	private static final String SOLR_USER_PASSWORD = "ranger.solr.audit.user.password";
	private static final String SOLR_URL = "ranger.audit.solr.urls";

	private static final String COLLECTION = "--collection";
	private static final String CONFIG = "--config";

	private static final String BACKUP = "backup";
	private static final String DELETE = "delete";
	private static final String RESTORE = "restore";
	private static final String UPLOAD = "upload";
	private static String ranger_home_dir;
	private static String solr_collection_name;
	private static String solr_config_name;
	private static SolrZkClient zkClient;

	private static File configSetFolder;
	private static String customConfigSetLocation;
	private static Path path_for_cloud_mode;

	private static CloudSolrClient solrCloudClient = null;
	private static SolrClient solrClient;
	private static boolean isKERBEROS = false;
	private static boolean solr_cloud_mode;
	private static boolean isSSLEnabled = false;
	private static String solr_url;

	public static void main(String[] args) {
		try {
			validate(args);
			SolrStorageMigrationUtility solrStorageMigraUtil = (SolrStorageMigrationUtility) CLIUtil
					.getBean(SolrStorageMigrationUtility.class);
			solrStorageMigraUtil.init();
			createSolrClient();
			String baseUrl = solr_cloud_mode ? getBaseUrl() : solr_url;
			loginAndProceed(args, baseUrl);
		} catch (Exception e) {
			System.out.println("Failed to get the instanciate of SolrStorageMigrationUtility, Error : " + e.getMessage());
			System.exit(1);
		}
	}

	@Override
	public void init() throws Exception {
		initProps();
	}

	private static void initProps() {
		customConfigSetLocation = PropertiesUtil.getProperty((CONFIG_SET_LOCATION));
		String sslEnabledProp = PropertiesUtil.getProperty(SSL_ENABLED_PARAM);
		isSSLEnabled = ("true".equalsIgnoreCase(sslEnabledProp));

		solr_collection_name = PropertiesUtil.getProperty(SOLR_COLLECTION_NAME, DEFAULT_COLLECTION_NAME);
		solr_config_name = PropertiesUtil.getProperty(SOLR_CONFIG_NAME, DEFAULT_CONFIG_NAME);

		final String authType = PropertiesUtil.getProperty(AUTHENTICATION_TYPE);
		System.out.println("Auth Type : " + authType);
		if (StringUtils.isNotEmpty(authType) && authType.equalsIgnoreCase(AUTH_TYPE_KERBEROS)) {
			isKERBEROS = true;
		}

		solr_cloud_mode = StringUtils.isNotEmpty(PropertiesUtil.getProperty(SOLR_ZK_HOSTS));
		if (!solr_cloud_mode) {
			solr_url = PropertiesUtil.getProperty(SOLR_URL);
			if (StringUtils.isEmpty(solr_url)) {
				System.out.println("Please provide ZooKeeper URL or Solr URL.");
				System.exit(1);
			}
		}

		if (StringUtils.isNotEmpty(customConfigSetLocation)) {
			configSetFolder = new File(customConfigSetLocation);
		} else {
			path_for_cloud_mode = Paths.get(ranger_home_dir, "contrib", "solr_for_audit_setup");
			configSetFolder = path_for_cloud_mode.toFile();
		}
	}

	private static void proceed(String[] args, String baseUrl) {
		final String command = args[1];
		String action = args[2];
		String location = null;
		String backupName = null;
		switch (command) {
		case COLLECTION:
			if (action.equalsIgnoreCase(BACKUP)) {
				location = args[3];
				backupName = args[4];
				System.out.println("\n Action : " + command + " " + action + "\n Location : " + location
						+ "\n Backup Name : " + backupName);
				backup(backupName, location);
			} else if (action.equalsIgnoreCase(DELETE)) {
				System.out.println("Deleting collection '" + solr_collection_name + "'.");
				Scanner reader = new Scanner(System.in);
				System.out.println("Are you sure? Please make sure to take the backup of '" + solr_collection_name
						+ "' collection before deleting it, If already backed up, press 'y' to contnue or press any key and hit enter to abort the operation");

				final String decision = reader.next();
				if (decision != null && "y".equalsIgnoreCase(decision.trim())) {
					deleteCollection();
				}
				System.exit(0);
				reader.close();
			} else if (action.equalsIgnoreCase(RESTORE)) {
				location = args[3];
				backupName = args[4];
				System.out.println("\n\n Command : " + command + "\n Action : " + action + "\n Location : " + location
						+ "\n Backup Name : " + backupName + "\n");
				restore(backupName, location);
			} else {
				System.out.println("Invalid command");
				showUsage();
				System.exit(400);
			}
			break;
		case CONFIG:
			if (action.equalsIgnoreCase(DELETE)) {
				deleteConfigSet();
			} else if (action.equalsIgnoreCase(UPLOAD)) {
				if (uploadConfiguration(baseUrl)) {
					System.out.println("Config-Set with the name '" + solr_config_name + "' created successfully.");
					System.exit(0);
				} else {
					System.out.println("Failed to create Config-Set with the name '" + solr_config_name + "'");
					System.exit(1);
				}
			}
			break;
		default:
			System.out.println("Invalid command");
			showUsage();
			System.exit(1);
			break;
		}
	}

	private static void loginAndProceed(String[] args, String baseUrl) {
		Subject sub = null;
		if (isKERBEROS) {
			String keytab = PropertiesUtil.getProperty(ADMIN_USER_KEYTAB);
			String principal = null;
			final String hostName = PropertiesUtil.getProperty("ranger.service.host");
			try {
				principal = SecureClientLogin.getPrincipal(PropertiesUtil.getProperty(ADMIN_USER_PRINCIPAL), hostName);
			} catch (IOException ignored) {
				System.out.println("Failed to get ranger.admin.kerberos.principal. Reason: " + ignored.toString());
				System.exit(1);
			}
			String nameRules = PropertiesUtil.getProperty(ADMIN_NAME_RULES);
			if (StringUtils.isBlank(nameRules)) {
				System.out.println("Name is empty. Setting Name Rule as 'DEFAULT'");
				nameRules = "DEFAULT";
			}
			try {
				sub = SecureClientLogin.loginUserFromKeytab(principal, keytab, nameRules);
			} catch (Exception e) {
				System.out.println("Failed to get login subject, Error : " + e.getMessage());
				System.exit(1);
			}

			try {
				Subject.doAs(sub, new PrivilegedAction<Void>() {
					@Override
					public Void run() {
						proceed(args, baseUrl);
						return null;
					}
				});
			} catch (Exception e) {
				System.out.println("'" + args[1] + "' '" + args[2] + "' failed , Error : " + e.getMessage());
				System.exit(1);
			}
		} else {
			proceed(args, baseUrl);
		}
	}

	private static void setBasicAuth(CollectionAdminRequest<?> req) {
		final String solr_user = PropertiesUtil.getProperty(SOLR_USER);
		final String solr_user_pssword = PropertiesUtil.getProperty(SOLR_USER_PASSWORD);
		if (StringUtils.isEmpty(solr_user) || StringUtils.isEmpty(solr_user_pssword)) {
			System.out.println("Solr Username or password should not be empty!");
			System.exit(1);
		}
		req.setBasicAuthCredentials(solr_user, solr_user_pssword);
	}

	private static void backup(String backupName, String location) {
		try {
			CollectionAdminRequest.Backup backupReq = CollectionAdminRequest.backupCollection(solr_collection_name,
					backupName);
			backupReq.setLocation(location);
			if (!isKERBEROS) {
				setBasicAuth(backupReq);
			}

			int status = backupReq.process(solrClient).getStatus();
			if (status == 0) {
				System.out.println("\n Backup of collection '" + solr_collection_name
						+ "' was successful, \n Backup location : " + location + ", \n Backup Name : " + backupName);
				System.exit(0);
			} else {
				System.out.println("\n Backup of collection failed, Status : " + status);
				System.exit(1);
			}
		} catch (Exception e) {
			System.out.println("\n Backup of collection '" + solr_collection_name + "' was unsuccessful, Error : "
					+ e.getMessage());
			System.exit(1);
		}
	}

	private static void deleteCollection() {
		try {
			CollectionAdminRequest.Delete deleteCollectionReq = CollectionAdminRequest
					.deleteCollection(solr_collection_name);
			if (!isKERBEROS) {
				setBasicAuth(deleteCollectionReq);
			}

			CollectionAdminResponse deleteCollectionResponse = deleteCollectionReq.process(solrClient);
			if (deleteCollectionResponse != null && deleteCollectionResponse.getStatus() != 0) {
				System.out.println("Unable to delete the collection = " + deleteCollectionResponse);
				System.exit(1);
			} else {
				System.out.println("Collection '" + solr_collection_name + "' deleted.");
				System.exit(0);
			}
		} catch (Exception e) {
			System.out.println(
					"Unable to delete the collection '" + solr_collection_name + "', Error :  " + e.getMessage());
			System.exit(1);
		}
	}

	private static void restore(String backupName, String location) {
		CollectionAdminRequest.Restore restoreReq = CollectionAdminRequest.restoreCollection(solr_collection_name,
				backupName);
		restoreReq.setLocation(location);
		if (!isKERBEROS) {
			final String solr_user = PropertiesUtil.getProperty(SOLR_USER);
			final String solr_user_pssword = PropertiesUtil.getProperty(SOLR_USER_PASSWORD);
			if (StringUtils.isEmpty(solr_user) || StringUtils.isEmpty(solr_user_pssword)) {
				System.out.println("Solr Username or password should not be empty!");
				System.exit(1);
			}
			restoreReq.setBasicAuthCredentials(solr_user, solr_user_pssword);
		}
		try {
			final int status = restoreReq.process(solrClient).getStatus();
			if (status == 0) {
				System.out.println(
						"Restoration of '" + backupName + "' into '" + solr_collection_name + "' was successful.");
				System.exit(0);
			}
			System.out.println("Failed to restore '" + backupName + "' into '" + solr_collection_name
					+ "' collection, Restore statu : " + status);
			System.exit(1);
		} catch (Exception e) {
			System.out.println("Failed to restore '" + backupName + "' into '" + solr_collection_name
					+ "' collection, Error : " + e.getMessage());
			System.exit(1);
		}
	}

	private static void deleteConfigSet() {
		try {
			ConfigSetAdminRequest.Delete configSetAdmin = new ConfigSetAdminRequest.Delete();
			configSetAdmin.setConfigSetName(solr_config_name);
			System.out.println("Deleting config-set '" + solr_config_name + "'");

			if (!isKERBEROS) {
				final String solr_user = PropertiesUtil.getProperty(SOLR_USER);
				final String solr_user_pssword = PropertiesUtil.getProperty(SOLR_USER_PASSWORD);
				if (StringUtils.isEmpty(solr_user) || StringUtils.isEmpty(solr_user_pssword)) {
					System.out.println("Solr Username or password should not be empty!");
					System.exit(1);
				}
				configSetAdmin.setBasicAuthCredentials(solr_user, solr_user_pssword);
			}
			int status = configSetAdmin.process(solrClient).getStatus();
			if (status == 0) {
				System.out.println("Config-set '" + solr_config_name + "' deleted successfully.");
				System.exit(0);
			} else {
				System.out.println("Failed to delete config-set, Status : " + status);
				System.exit(1);
			}
		} catch (Exception e) {
			System.out.println("Failed to delete config-set, Error : " + e.getMessage());
			System.exit(1);
		}
	}

	private static void validateBkpNameAndLocation(String[] args) {
		String location = null;
		try {
			location = args[3];
			if (StringUtils.isEmpty(location) || !location.startsWith("/")) {
				System.out.println("Please provide valid backup location, Provided : " + location);
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("Please provide backup location.");
			showUsage();
			System.exit(1);
		}

		String backupName = null;
		try {
			backupName = args[4];
			if (StringUtils.isEmpty(backupName)) {
				System.out.println("Please specify the valid backup name, Provided : " + backupName);
				showUsage();
				System.exit(1);
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("Please specify backup name");
			showUsage();
			System.exit(1);
		}
	}

	private static void validate(String[] args) {
		if (args == null || args.length < 3) {
			showUsage();
		}

		ranger_home_dir = PropertiesUtil.getProperty((RANGER_ADMIN_HOME));
		ranger_home_dir = StringUtils.isEmpty(ranger_home_dir) ? args[0] : ranger_home_dir;
		if (StringUtil.isEmpty(ranger_home_dir)) {
			System.out.println("Ranger Home Directory is missing. Kindly set it.");
			System.exit(1);
		}

		System.setProperty("ranger.admin.home", args[0]);
		final String command = args[1];
		if (StringUtils.isEmpty(command)) {
			showUsage();
		}

		String action = args[2];
		if (StringUtils.isEmpty(action)) {
			showUsage();
		}

		switch (command) {
		case COLLECTION:
			if (action.equalsIgnoreCase(BACKUP)) {
				validateBkpNameAndLocation(args);
			} else if (action.equalsIgnoreCase(DELETE)) {
				System.out.println("Deleting Collection");
			} else if (action.equalsIgnoreCase(RESTORE)) {
				validateBkpNameAndLocation(args);
			} else {
				showUsage();
			}
			break;
		case CONFIG:
			if (action.equalsIgnoreCase(DELETE)) {
				System.out.println("Deleting Config");
			} else if (action.equalsIgnoreCase(UPLOAD)) {
				System.out.println("Uploading Config");
			} else {
				showUsage();
			}
			break;
		default:
			showUsage();
			System.exit(1);
			break;
		}
	}

	@SuppressWarnings("rawtypes")
	private static long execute(HttpUriRequest req) throws Exception {
		Map responseMap = null;
		String response;
		Map resHeader = null;
		try {
			HttpEntity resEntity = solrCloudClient.getLbClient().getHttpClient().execute(req).getEntity();
			response = EntityUtils.toString(resEntity, StandardCharsets.UTF_8);
			responseMap = (Map) ObjectBuilder.getVal(new JSONParser(new StringReader(response)));
			resHeader = (Map) responseMap.get("responseHeader");
		} catch (Exception e) {
			System.out.println("Request Failed, Error : " + e.getMessage());
			System.exit(1);
		}

		final long status = (long) resHeader.get("status");
		if (status != 0) {
			try {
				System.out.println("Request Failed, Error : " + ((Map) responseMap.get("exception")).get("msg")
						+ ", status - " + status);
			} catch (Exception e) {
				System.out.println("Request Failed, Error : " + ((Map) responseMap.get("error")).get("msg")
						+ ", status - " + status);
				System.exit(1);
			}
		}
		return status;
	}

	private static void showUsage() {
		System.out.println("Invalid request,\nUsage : " + "\n=> --collection backup /backup_location backup_name"
				+ "\n=> --collection delete" + "\n=> --config delete" + "\n=> --config upload"
				+ "\n=> --collection restore /backup_loaction backup_name");
		System.exit(1);
	}

	private static String getBaseUrl() {
		Set<String> nodes = solrCloudClient.getClusterStateProvider().getLiveNodes();
		String[] nodeArr = nodes.toArray(new String[0]);
		// getting nodes URL as 'port_solr', so converting it to 'port/solr'
		return nodeArr[0].replaceAll("_", "/");
	}

	private static List<String> getZkHosts() {
		String zkHosts = "";
		List<String> zookeeperHosts = null;
		if (!StringUtil.isEmpty(PropertiesUtil.getProperty(SOLR_ZK_HOSTS))) {
			zkHosts = PropertiesUtil.getProperty(SOLR_ZK_HOSTS).trim();
			zookeeperHosts = new ArrayList<String>(Arrays.asList(zkHosts.split(",")));
		}
		return zookeeperHosts;
	}

	private static boolean createSolrClient() {
		List<String> zookeeperHosts = null;
		try {
			if (isKERBEROS) {
				setHttpClientBuilderForKrb();
			}
			if (solr_cloud_mode) {
				zookeeperHosts = getZkHosts();
				solrCloudClient = new CloudSolrClient.Builder(zookeeperHosts, Optional.empty()).build();
				solrCloudClient.setDefaultCollection(solr_collection_name);
				solrCloudClient.connect();
				zkClient = solrCloudClient.getZkStateReader().getZkClient();
				solrClient = solrCloudClient;
				solr_cloud_mode = true;
			} else {
				HttpSolrClient.Builder builder = new HttpSolrClient.Builder();
				builder.withBaseSolrUrl(solr_url);
				solrClient = builder.build();
			}
			return true;
		} catch (Exception ex) {
			System.out.println("Can't connect to Solr server. ZooKeepers=" + zookeeperHosts + ", Solr Url " + solr_url
					+ ", collection=" + solr_collection_name + ex);
			return false;
		}
	}

	@SuppressWarnings("resource")
	private static void setHttpClientBuilderForKrb() {
		Krb5HttpClientBuilder krbBuild = new Krb5HttpClientBuilder();
		SolrHttpClientBuilder kb = krbBuild.getBuilder();
		HttpClientUtil.setHttpClientBuilder(kb);
	}

	private static boolean uploadConfiguration(String baseUrl) {
		try {
			if (zkClient != null) {
				ZkConfigManager zkConfigManager = new ZkConfigManager(zkClient);

				boolean configExists = zkConfigManager.configExists(solr_config_name);
				if (!configExists) {
					try {
						System.out
								.println("Config does not exist with name " + solr_config_name + ", Creating new one.");
						String zipOfConfigs = null;
						if (configSetFolder.exists() && configSetFolder.isFile()) {
							zipOfConfigs = configSetFolder.getAbsolutePath();
						} else {
							String[] files = configSetFolder.list();
							if (files != null) {
								for (String aFile : files) {
									if (aFile != null) {
										if (aFile.equals("solr_audit_conf.zip")) {
											zipOfConfigs = configSetFolder + "/" + aFile;
											break;
										}
									}
								}
							}
						}
						if (zipOfConfigs == null) {
							throw new FileNotFoundException("Could Not Find Configs Zip File : " + configSetFolder);
						}
						System.out.println("Config Set Location : " + zipOfConfigs);
						File file = new File(zipOfConfigs);
						byte[] arrByte = Files.readAllBytes(file.toPath());
						ByteBuffer byteBuffer = ByteBuffer.wrap(arrByte);
						String protocol = isSSLEnabled ? "https" : "http";
						String uploadConfigsUrl = String.format(
								"%s://%s/admin/configs?action=UPLOAD&overwrite=true&cleanup=true&name=%s", protocol,
								baseUrl.toString(), solr_config_name);
						postDataAndGetResponse(uploadConfigsUrl, byteBuffer);
						return true;
					} catch (Exception ex) {
						System.out.print("Error while uploading configs, Error : " + ex.getMessage());
						return false;
					}
				} else {
					System.out.println("Config already exists with name " + solr_config_name);
					return true;
				}
			} else {
				System.out.println(
						"Solr is in cloud mode and could not find the zookeeper client for performing upload operations. ");
				return false;
			}
		} catch (Exception ex) {
			System.out.println("Error while uploading configuration, Error :  " + ex);
			return false;
		}

	}

	private static void postDataAndGetResponse(String uri, ByteBuffer bytarr)
			throws IOException {
		HttpPost httpPost = null;
		httpPost = new HttpPost(uri);
		httpPost.setHeader("Content-Type", "application/octet-stream");
		httpPost.setEntity(new ByteArrayEntity(bytarr.array(), bytarr.arrayOffset(), bytarr.limit()));

		try {
			long status = execute(httpPost);
			if (status == 0) {
				System.out.println("Config set created with name : '" + solr_config_name + "'");
				System.exit(0);
			}
		} catch (Exception e) {
			System.out.println("Error while creating  config set with name : '" + solr_config_name + "'");
			System.exit(1);
		} finally {
			httpPost.releaseConnection();
		}
	}

	@Override
	public void printStats() {
		// TODO Auto-generated method stub

	}

	@Override
	public void execLoad() {
		// TODO Auto-generated method stub

	}

}
