/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.ranger.tagsync.source.atlas;

import org.apache.commons.lang.StringUtils;
import org.apache.ranger.plugin.model.RangerPolicy;
import org.apache.ranger.plugin.model.RangerServiceResource;
import org.apache.ranger.tagsync.source.atlasrest.RangerAtlasEntity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.Map;

public class AtlasS3ResourceMapper extends AtlasResourceMapper {
    public static final String ATLAS_TYPE_S3_DIRECTORY = "aws_s3_v2_bucket";
    public static final String ATLAS_TYPE_S3_BUCKET = "aws_s3_v2_directory";

    public static final String RANGER_TYPE_BUCKET = "bucket";
    public static final String RANGER_TYPE_PATH = "path";

    private static final String[] SUPPORTED_ENTITY_TYPES = {ATLAS_TYPE_S3_DIRECTORY, ATLAS_TYPE_S3_BUCKET};
    private static final Logger LOG = LoggerFactory.getLogger(AtlasS3ResourceMapper.class);

    public AtlasS3ResourceMapper() {
        super("s3", SUPPORTED_ENTITY_TYPES);
    }

    @Override
    public RangerServiceResource buildResource(RangerAtlasEntity entity) throws Exception {
        String qualifiedName = (String)entity.getAttributes().get(AtlasResourceMapper.ENTITY_ATTRIBUTE_QUALIFIED_NAME);

        String entityType  = entity.getTypeName();
        if (!entityType.equals(ATLAS_TYPE_S3_BUCKET) && !entityType.equals(ATLAS_TYPE_S3_DIRECTORY)) {
            throw new Exception("Unhandled atlas entity type " + entityType);
        }

        if (StringUtils.isEmpty(qualifiedName)) {
            throw new Exception("attribute '" +  ENTITY_ATTRIBUTE_QUALIFIED_NAME + "' not found in entity");
        }

        if (!qualifiedName.startsWith("s3a://")) {
            throw new Exception("Not an s3 path: " + qualifiedName);
        }

        /**
         *  The qualified name is of format s3a://<bucket>/<path>@<cluster>
         *  "s3a://qe-s3-bucket-daily/cc-yarn-0max5p-dd/tables/test_table2/@cm"
         */
        String entityGuid  = entity.getGuid();
        String clusterName = getClusterNameFromQualifiedName(qualifiedName);
        assertNotEmpty(clusterName, "Cluster name");

        String serviceName = getRangerServiceName(clusterName);
        String bucketPath = getResourceNameFromQualifiedName(qualifiedName).replaceFirst("s3a://", "");
        String[] parts = bucketPath.split("/", 2);
        String bucket = parts[0];
        assertNotEmpty(bucket, "bucket name");
        String path = "";
        if (parts.length > 1) {
            path = parts[1];
            assertNotEmpty(path, "path");
        }

        LOG.trace("Building RangerServiceResource with bucket={}, path={}", bucket, path);
        Map<String, RangerPolicy.RangerPolicyResource> elements = new HashMap<>();
        elements.put(RANGER_TYPE_BUCKET, new RangerPolicy.RangerPolicyResource(bucket));
        elements.put(RANGER_TYPE_PATH, new RangerPolicy.RangerPolicyResource("/" + path, Boolean.FALSE, Boolean.TRUE));

        RangerServiceResource ret = new RangerServiceResource(entityGuid, serviceName, elements);

        return ret;
    }

    private void assertNotEmpty(String data, String varName) throws Exception {
        if (StringUtils.isEmpty(data)) {
            throw new Exception(varName + " is empty");
        }
    }
}
